package MyTreeMap_indels;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#	Version 0.6.2: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap_indels.pm:
#	This package makes a set of maps of the most parsimonious insertions/deletions
#	scenarios onto an input sequence tree.
#	The main algorithm is based on Farris's (1977) Dollo parsimony principle.
#	(CREATION STARTED on May 3 (Thu), 2012, ENDED on May ?? (???), 2012)
#
#	REVISED on Oct 24 (Wed), 2012:
#	To make the algorithm closer to perfection,
#	the subroutin "recede_absence" was modified so that it will utilize
#	flanking blocks in broad sense as well as those in strict sense.
#
#	REVISED on Nov 13 (Wed), 2013:
#	The module can now correctly handle trees containing two-branch nodes.
#
#	REVISED on Nov 15 (Fri), 2013:
#	An optional flag (global variable) was added to enable the subroutine 'enum_psm_indel_scns_in_cls_blk'
#	to ignore null columns when necessary.
#	[ This may be useful when we deal with a 'culled' MSA that resulted from removing some sequences
#	from an alignmentof a larger number of sequences reconstructed via an aligning program. ]
#
#	REVISESD on Mar 8 (Sat) & 9 (Sun), 2014:
#	A new subroutine, 'enum_psm_scns_gen_null_blk', was created.
#	It enables the subroutine 'enum_psm_indel_scns_in_cls_blks' to enumerate the parsimonious indel scenarios
#	that resulted in "NULL-cluster", which consists only of gaps.
#
#	REVISED on Mar 10 (Mon), 2014:
#	The package now introduces a global variable, $UPPER_BOUND_ANAL_GAPLEN.
#	If a gap-block cluster contains a run of gaps longer than this variable,
#	the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses of the cluster.
#	Accompanying the introduction of this global variable,
#	a new subroutine 'scan_gbcls_for_long_gaps' was created.
#
#	REVISED on Mar 11 (Tue), 2014:
#	TEMPORARILY Modified the subroutine 'incorporate_new_event' to TENTATIVELY cope with
#	the rare cases where the new 'deletion' event encounters BOTH left- and right-neighboring events.
#
#	(I WILL SERIOUSLY ADDRESS THIS ISSUE LATER. ... Probably, a solution would be to leave 'deletion' events unmarged when they are intervened by the 'presence' state in any 'downstream' nodes.)
#
#
#	REVISED on Mar 11 (Tue), 2014 (II):
#	The subroutine now introduces an upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS = 200,
#	for the cluster-length itself as well.
# 
# MyTreeMap_D.pm:
#	This package maps states and their changes
#	onto the nodes and the branches, respectively, of a phylogenetic tree,
#	according to Farris's (1977) Dollo parsimony principle.
#	In a nutshell, Farris's (1977) Dollo parsimony seeks for the evolutionary scenario
#	that minimizes the number of changes, UNDER THE CONSTRAINT that
#	the 'creation' of a state occurs at most once in each column.
#	(CREATION STARTED on Apr 19 (Thu), 2012, ENDED on Apr ?? (???), 2012)
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#	(REVISED on Nov 27 (Sun), 2011)
#	The subroutine 'min_penalties_sankoff_bo' now excludes from later consideration
#	the characters on each node that will NEVER be included in any parsiminious states.
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray union_nmarray );

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
	fetch_all_offsprings_nr
		);
	## [ADDED 'fetch_all_offsprings_nr' on 2019/05/14.]
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	create_initial_set_indels
	enum_psm_scns_gen_null_blk
	incorporate_new_event
	create_bestset_mvd_events
	recede_absence
	cleanup_br2indels
	incorporate_null_block
	enum_psm_indel_scns_in_cls_blks
	extract_cls_gblocks
	msa_rep_gpatterns
	scan_gbcls_for_long_gaps
	enum_psm_indel_scns_for_msa
	);


	# This specifies the level of tests on main subroutines,
	# i.e. 'enum_psm_indel_scns_for_msa', ..., and ....
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# This specifies the level of tests on the satellite subroutines of 'enum_psm_indel_scns_for_msa',
	# such as  'enum_indel_scns_in_cls_blks', 'extract_cls_glbocks', and 'msa_rep_gpatterns'. #
##our $TEST_LEVEL1 = 0;
our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 5;

	# This specifies the level of tests on major satellite subroutines of 'enum_indel_scns_in_cls_blks'.
##our $TEST_LEVEL1hf = 0;
our $TEST_LEVEL1hf = 1;
##our $TEST_LEVEL1hf = 2;
##our $TEST_LEVEL1hf = 5;

	# This specifies the level of tests on other important satellite subroutines.
##our $TEST_LEVEL2 = 0;
our $TEST_LEVEL2 = 1;
##our $TEST_LEVEL2 = 2;
##our $TEST_LEVEL2 = 5;

	# This specifies the level of tests on the subroutine 'create_initial_set_indels'.
##our $TEST_LEVEL2i = 0;
our $TEST_LEVEL2i = 1;
##our $TEST_LEVEL2i = 2;
##our $TEST_LEVEL2i = 3;
##our $TEST_LEVEL2i = 5;

	# This specifies the level of tests on the subroutine 'create_bestset_mvd_events'.
##our $TEST_LEVEL2c = 0;
our $TEST_LEVEL2c = 1;
##our $TEST_LEVEL2c = 2;
##our $TEST_LEVEL2c = 3;
##our $TEST_LEVEL2c = 5;

	# This specifies the level of tests on the subroutine 'incorporate_null_block'.
##our $TEST_LEVEL2d = 0;
our $TEST_LEVEL2d = 1;
##our $TEST_LEVEL2d = 2;
##our $TEST_LEVEL2d = 3;
##our $TEST_LEVEL2d = 5;

	# This specifies the level of tests on minor satellite subroutines.
our $TEST_LEVEL3 = 0;
##our $TEST_LEVEL3 = 1;
##our $TEST_LEVEL3 = 2;
##our $TEST_LEVEL3 = 5;

	# This specifies the level of tests on the satellite subroutine 'recede_spec_deletion'.
our $TEST_LEVEL3d = 0;
##our $TEST_LEVEL3d = 1;
##our $TEST_LEVEL3d = 2;
##our $TEST_LEVEL3d = 5;

	# This specifies the level of tests on the satellite subroutine 'enum_psm_scns_gen_null_blk'.
	# (ADDED on Mar 8 (Sat), 2014.)
our $TEST_LEVEL_GEN_NULL_BLK = 0;
##our $TEST_LEVEL_GEN_NULL_BLK = 5;

	# This specifies the level of tests on the satellite subroutine 'scan_gbcls_for_long_gaps'. #
	# (ADDED on Mar 10 (Mon), 2014.)
our $TEST_LEVEL_SCN = 0;
##our $TEST_LEVEL_SCN = 5;

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


	## A flag that determines whether the subroutine 'enum_psm_indel_scns_in_cls_blks'
	## should leave null columns untouched or not.
	## = 0 when handle null columns, regarding them as resulting from indel events within the phylogeny of aligned sequences (useful when handling simulated MSAs),
	## = 1 when ignore (i.e., leave untouched) null columns, regarding them as resulting from the removal of sequences
	##	that hold residues in the columns (useful when handling 'culled' MSAs made from MSAs reconstructed via aligners).
	## (ADDED on Nov 15 (Fri), 2013)
	#
our $FL_IGNORE_NULL_COLUMNS = 0;

	## A flag that determines whether or not you should conjure the subroutine 'enum_psm_scns_gen_null_blk',
	## which enumerates parsimonious scenarios generating null-clusters (namely, gap-block clusters consisting only of null columns).
	## =1 if you want to, 
	## = 0 if you don't want to.
	# (ADDED on Mar 8 (Sat), 2014.)
	#
our $FL_HANDLE_NULL_CLS = 1;
($FL_IGNORE_NULL_COLUMNS) and ($FL_HANDLE_NULL_CLS  = 0); # When ignoring null-columns, null-clusters must always be ignored as well.

	## This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	# (ADDED on Mar 10 (Mon), 2014.)
	#
our $UPPER_BOUND_ANAL_GAPLEN = 150;

	## This variable sets an upper-bound for the length of the gap-block cluster itself.
	## (ADDED on Mar 11 (Tue), 2014.):w
our $UPPER_BOUND_ANAL_LEN_GBCLS = 200;

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
	#
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# create_initial_set_indels:
#	This satellite subroutine of 'enum_psm_indel_scns_in_cls_blks' creates
#	the initial 'tentative' candidate of the parsimonious indel scenario,
#	out of the raw output of the Dollo parsimony algorithm on the cluster of alignment blocks.
#	Each indel event is made by merging the neighboring (or "quasi-neighboring") events of the same type along the same branch.
#	The "quasi_neighboring" events are indel events NOT separated by (a) block(s) with $char_pres at BOTH ends of the branch in question.
#
#	REVISED on May 13 (Sun), 2012:
#	The format for the input set of indel blocks was made more general,
#	to make the subroutine a bit more flexible.
#	(The change accompanied the change in the master subroutine 'enum_psm_indel_scns_in_cls_blk'.)
#
# INPUT: ($topnode_id, \%node2pa, $char_pres, $char_abs, \@blocks_in_cls, $indx_indx_gp, \@set_node2dparsstat, \@set_br2change), where
#	%noce2pa = ($node_id => $parent_id, ...) for all nodes except the top node.
#	[ %node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state; ] -- It turned out to be unnecessary.
#
#	@blocks_in_cls represents a cluster of contiguous blocks,
#	each consisting of sites with a particular gap-pattern, where
#	$indx_gp = $blocks_in_cls[$b]->[$indx_indx_gp] (or $blocks_in_cls[$b] itself if $indx_indx_gp is undefined) is the index specifying the gap pattern in @set_node2parsstat and @set_br2change,
#	\@set_node2parsstat and \@set_br2change are output of 'map_parsim_states_dollo_msa'
#	on the 'secondary MSA', which is a concatenated set of 1-column representations of the gap patterns, where
#	%{$set_node2dparsstat[$k]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $k th gap pattern (i.e. the $k th column in the secondary MSA) onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%{$set_br2change[$k]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches (under the Dollo parsimony) for the $k th gap pattern, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node;
#
# OUTPUT: ($ct_indels0, \%node2blwise_gstat0, \%br2indels0, \@list_blank_blks), where
#	%node2blwise_gstat0 = ($node_id => \@blwise_gstat0, ...) for all nodes,
#	with $blwise_gstat0[$b] = $char_pres/abs being the state in the $b th block in the cluster;
#	%{$br2indels0} = ($branch_id => \@list_indels0, ...) for the initial set of indels, with
#	@{$list_indels0[$id]} = ($stat_parent, $stat_node, \@indices_blk0) describing the $id th initial indel event at the branch, with @indices_blk0 listing the indices in @{$blocks_in_cls};
#	@list_blank_blks lists the indices of the blocks consisting only of $char_abs.
#
sub create_initial_set_indels ($\%$$\@$\@\@) {

	my ($topnode_id, $node2pa, $char_pres, $char_abs, $blocks_in_cls, $indx_indx_gp, $set_node2dparsstat, $set_br2change) = @_;

	my $ct_blks = @{$blocks_in_cls};
    (print STDOUT "  < Subroutine 'create_initial_set_indels' (#{blocks} = $ct_blks) >\n\n") if ($TEST_LEVEL2i>0);

	my %node2blwise_gstat0;
	my %br2indels0;
	my $ct_indels0 = 0;
	my @list_blank_blks = ();

	for (my $b=0; $b<$ct_blks; $b++) { # Main outer for-loop (over blocks) #

	    my $block = $blocks_in_cls->[$b];
	    my $indx_gp = (defined $indx_indx_gp) ? $block->[$indx_indx_gp] : $block;
	    ##my $indx_gp = $blocks_in_cls->[$b][0];

	(print STDOUT "  [ The $b th block (", '$indx_gp = ', $indx_gp, ") ]\n\n") if ($TEST_LEVEL2i>1);

	    my $node2dparsstat = $set_node2dparsstat->[$indx_gp];
	    my $br2change = $set_br2change->[$indx_gp];

	    foreach my $node (keys %{$node2dparsstat}) { # 1st middle foreach-loop (over nodes) #
		my ($stat, $ct_pres_ext) = @{$node2dparsstat->{$node}};

		my $blwise_gstat0 = $node2blwise_gstat0{$node};
		(defined $blwise_gstat0) or ($blwise_gstat0 = $node2blwise_gstat0{$node} = []);

		$blwise_gstat0->[$b] = $stat;
		##push @{$blwise_gstat0}, $stat;
	    }		# End of the 1st middle foreach-loop (over nodes) #

	(print STDOUT '  +++ Finished filling in  %node2blwise_gstat0 with the states in the ', $b, ' th block. +++', "\n\n") if ($TEST_LEVEL2i>1);

	    unless (0< scalar keys %{$br2change}) {
			# Handle a block withOUT CHARACTER CHANGEs. #

		   # A BUG fixed on May 14 (Mon), 2012. #
		my $stat_top = $node2dparsstat->{$topnode_id}[0];
		##my $stat_top = $node2dparsstat->{$topnode_id};
		if ($stat_top eq $char_abs) { # This is normal (ALL the NODES show 'absence'). #
		    ## $ct_indels0++; ## DISABLED on May 14 (Mon), 2012 to coordinate the behaviors with 'incorporate_null_block'. ##
		    push @list_blank_blks, $b;
		    next;
		} else { # This is not permitted (ALL the NODES show 'presence'). #
		    print STDOUT "[ERROR 'create_initial_set_indels'] ILLEGAL INPUT: The $b th block in the cluster (the $indx_gp th column in the 2ndary MSA) has NO STATE CHANGE and STATE at the TOP (='$stat_top') is NOT 'ABSENCE' (= '$char_abs'), which is impossible: $! !\n";
		    return (undef, undef, undef, undef);
		}
	    }

	(print STDOUT '  +++ Now, start incorporating the state changes in the ', $b, ' th block into %bl2indels0. +++', "\n\n") if ($TEST_LEVEL2i>1);

	    foreach my $br (keys %{$br2change}) { # 2nd middle foreach-loop (over branches) #

	(print STDOUT '   ( Branch ID: $br = ', "$br)\n")  if ($TEST_LEVEL2i>1);

		my ($stat_pa, $stat_node) = @{$br2change->{$br}};
		my $list_indels0 = $br2indels0{$br};

		my @new_indel0 = ($stat_pa, $stat_node, [$b]);
		unless (defined $list_indels0) { # Initialize the list of indels. #
		    $br2indels0{$br} = [\@new_indel0];
		    $ct_indels0++;
		    next;
		}

#	%{$br2indels0} = ($branch_id => \@list_indels0, ...) for the initial set of indels, with
#	@{$list_indels0[$id]} = ($stat_parent, $stat_node, \@indices_blk0) describing the $id th initial indel event at the branch, with @indices_blk0 listing the indices in @{$blocks_in_cls};

    if ($TEST_LEVEL2i>2) {
	print STDOUT "\n", ' ' x 3, '[ Contents of @{$list_indels0} ]', "\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx', 'stat_pa', 'stat_node', '[indices,involved,blocks]'), "\n\n";
	for (my $id=0; $id<@{$list_indels0}; $id++) {
	    my ($stat_pa, $stat_node, $indices_blk) = @{$list_indels0->[$id]};
	    my $cnct_indices = join (',', @{$indices_blk});
	    print STDOUT ' ' x 3, join ("\t", $id, $stat_pa, $stat_node, "\[$cnct_indices\]"), "\n";
	}
	print STDOUT "\n";
    }
		my $pa = $node2pa->{$br};
		my $blwise_gstat0 = $node2blwise_gstat0{$br};
		my $blwise_gstat0_pa = $node2blwise_gstat0{$pa};

    if ($TEST_LEVEL2i>2) {
	print STDOUT ' ' x 3, '[ Blockwise States ]', "\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx_Block', 'state_pa', 'state_node'), "\n\n";
	for (my $i=0; $i<$ct_blks; $i++) {
	    my ($gstat0_pa, $gstat0_node) = ($blwise_gstat0_pa->[$i], $blwise_gstat0->[$i]);
	    (defined $gstat0_pa) or ($gstat0_pa = 'undef');
	    (defined $gstat0_node) or ($gstat0_node = 'undef'); 
	    print STDOUT ' ' x 3, join ("\t", $i, $gstat0_pa, $gstat0_node), "\n";
	}
	print STDOUT "\n";
    }

		my $last_indel0 = $list_indels0->[$#{$list_indels0}];
		my ($lstat_pa, $lstat_node, $lindices_blk0) = @{$last_indel0};
		my $last_indx_blk0 = $lindices_blk0->[$#{$lindices_blk0}];

		my $slast_indel0 = (@{$list_indels0}>1) ? $list_indels0->[$#{$list_indels0}-1] : [];
		my ($slstat_pa, $slstat_node, $slindices_blk0) = @{$slast_indel0};
		my $slast_indx_blk0 = (@{$list_indels0}>1) ? $slindices_blk0->[$#{$slindices_blk0}] : undef;
		if ((defined $slast_indx_blk0) and ($slast_indx_blk0 > $last_indx_blk0)) {
			# Swap the last and the second last indel events. #
		    ($last_indel0, $slast_indel0) = ($slast_indel0, $last_indel0);
		    ($lstat_pa, $lstat_node, $lindices_blk0) = @{$last_indel0};
		    $last_indx_blk0 = $lindices_blk0->[$#{$lindices_blk0}];
		    ($slstat_pa, $slstat_node, $slindices_blk0) = @{$slast_indel0};
		    $slast_indx_blk0 = $slindices_blk0->[$#{$slindices_blk0}];
		}

    if ($TEST_LEVEL2i>2) {
	print STDOUT "\n", ' ' x 3, '[ Last and Second-last Indel Events ]', "\n\n";
	print STDOUT ' ' x 3, join ('Last/second', 'state_pa', 'state_node', '[indices,involved,blocks]', 'indx_last_block'), "\n\n";
	my $cnct_lindices = join (',', @{$lindices_blk0});
	print STDOUT ' ' x 3, join ("\t", 'Last', $lstat_pa, $lstat_node, "\[$cnct_lindices\]", $last_indx_blk0), "\n";
	if (defined $slast_indx_blk0) {
	    my $cnct_slindices = join (',', @{$slindices_blk0});
	    print STDOUT ' ' x 3, join ("\t", 'Second', $slstat_pa, $slstat_node, "\[$cnct_slindices\]", $slast_indx_blk0), "\n";
	} else {
	    print STDOUT ' ' x 3, join ("\t", 'Second', 'undef'), "\n";
	}
	print STDOUT "\n";
    }

		my $fl_barrier = 0;
		my $end_barrier = $b-1;

		for (my $j=$end_barrier; $j>$last_indx_blk0; $j--) { # 1st inner for-loop #
		    if (($blwise_gstat0->[$j] eq $char_pres) and ($blwise_gstat0_pa->[$j] eq $char_pres)) {
			$fl_barrier = 1;
			last;
		    }
		}					# End of the 1st inner for-loop #

    if ($TEST_LEVEL2i>2) {
	print STDOUT ' ' x 3, '$end_barrier = ', $end_barrier, "\n";
	print STDOUT ' ' x 3, '$fl_barrier  = ', $fl_barrier, " .\n\n";
    }

		if ($fl_barrier == 1) {
			# There is a barrier between the current and the last indels. #
		    push @{$list_indels0}, \@new_indel0;
		    $ct_indels0++;
		    next;
		} elsif (($stat_pa eq $lstat_pa) and ($stat_node eq $lstat_node)) {
			# No barrier, and the current and last indels are of the same type. #
		    push @{$lindices_blk0}, $b;
		    next;
		} elsif (@{$list_indels0}==1) {
			# So far, only one indel event is listed along the branch. #
		    push @{$list_indels0}, \@new_indel0;
		    $ct_indels0++;
		    next;
		}

			# The current and the last indels are of different types, #
			# but are independent of each other. #

		my $end_sbarrier = $lindices_blk0->[0] -1;

    (print STDOUT ' ' x 3, '... Now searching for a barrier between the current event and the 2nd-last event ...', "\n\n") if ($TEST_LEVEL2i>3);

		my $fl_barrier2=0;
		for (my $j=$end_sbarrier; $j>$slast_indx_blk0; $j--) { # 2nd inner for-loop #

    if ($TEST_LEVEL2i>3) {
	print STDOUT ' ' x 5, "{ The $j th block }\n\n";
	print STDOUT ' ' x 5, '(state_pa, state_node) = (', join (', ', $blwise_gstat0_pa->[$j], $blwise_gstat0->[$j]), ")\n\n";
    }

		    if (($blwise_gstat0->[$j] eq $char_pres) and ($blwise_gstat0_pa->[$j] eq $char_pres)) {
				## A critial BUG fixed on Sep 29 (Sat), 2012. ##
			$fl_barrier2 = 1;
			##$fl_barrier = 1;
			last;
		    }
		}					# End of the 2nd inner for-loop #

    if ($TEST_LEVEL2i>2) {
	print STDOUT ' ' x 3, '$end_sbarrier = ', $end_sbarrier, "\n";
	print STDOUT ' ' x 3, '$fl_barrier2  = ', $fl_barrier2, " .\n\n";
    }

		if ($fl_barrier2 == 1) {
			# A barrier intervenes between the current and the second last indels.
		    push @{$list_indels0}, \@new_indel0;
		    $ct_indels0++;
		} else {
			# No intervening barrier. #

		    unless (($stat_pa eq $slstat_pa) and ($stat_node eq $slstat_node)) {
			print STDOUT "[ERROR 'create_initial_set_indels'] ILLEGAL INPUT: The current indel event ($stat_pa -> $stat_node) at the $b th block (i.e. the $indx_gp th column in the secondary MSA) CLASHES with the second last indel event ($slstat_pa -> $slstat_node): $! !\n";
			return (undef, undef, undef, undef);
		    }

			#### I do not understand why the barrier between the current indel event and the 2nd last event was NOT detected, while it DOES exist... (Sep 23 (Sun), 2012) ###
			### RESTART FROM HERE ###

		    push @{$slindices_blk0}, $b;
		}

	    }				# End of the 2nd middle foreach-loop (over branches) #

	(print STDOUT "\n", '  +++ Finished incorporating the state changes in the ', $b, ' th block into %bl2indels0. +++', "\n\n") if ($TEST_LEVEL2i>1);
	    
	}			# End of the main outer for-loop (over blocks) #

    (print STDOUT "\n  < END of the subroutine 'create_initial_set_indels' />\n\n") if ($TEST_LEVEL2i>0);

	return ($ct_indels0, \%node2blwise_gstat0, \%br2indels0, \@list_blank_blks);

}	# End of the subroutine 'create_initial_set_indels' #


# enum_psm_scns_gen_null_blk:
#	This satellite subroutine of 'enum_psm_indel_scns_in_cls_blks' enumerates the parsimonious scenarios
#	each of which results in a null-cluster, i.e., a gap-block cluster consisting only of a single null-block.
#	(CREATED in revision of this package on Mar 8 (Sat), 2014.)
#
# INPUT: ($top_node_id, \%node2pa, \%node2ch, $char_pres, $char_abs, $indx_null_blk, $fl_full), where
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#
#	$indx_null_blk specifies the location of the null-block in the array @blocks_in_cls;
#
#	$fl_full = 1 if you want the full output,
#		 = 0 if you want a simpler output.
#
# OUTPUT: ($min_totct_indels, \@set_psm_scns_null_blk), where
#	$min_totct_indels is the total number of insertion/deletion events in the most parsimonious scenario(s);
#	@{$set_psm_scns_null_blk[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th parsimonious scenario, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk = ($indx_null_blk) listing the indices (in @blocks_in_cls)
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$b] being $char_pres/abs for the $b th block (in the cluster) at the node $node_id.
#
#	If $fl_full = 0, %node2gapstat is given only at the top node.
#
sub enum_psm_scns_gen_null_blk ($\%\%$$$$) {

	my ($topnode_id, $node2pa, $node2ch, $char_pres, $char_abs, $indx_null_blk, $fl_full) = @_;

    (print STDOUT "   < Subroutine 'enum_psm_scns_gen_null_blk' (top-node ID: $topnode_id) >\n\n") if ($TEST_LEVEL_GEN_NULL_BLK >0); 

	my @set_psm_scns_null_blk = ();

		## First, enumerate the scenarios involving the top-node, if necessary. ##
		# (In this case, the gap-state at the top-node is 'presence', and the states at other nodes are 'absence'.)

	my $sub_topnodes = $node2ch->{$topnode_id};
	unless (@{$sub_topnodes}>0) {
	    print STDOUT "[ERROR 'enum_psm_scns_gen_null_blk'] ILLEGAL INPUT: The top-node (ID: $topnode_id) has NO CHILDREN: $! !\n";
	    return (undef, undef);
	}

	if (@{$sub_topnodes} < 3) { # If the top-node has only 1 or 2 children. #

	    my %br2indels;
	    foreach my $br (@{$sub_topnodes}) { # 1st auxiliary foreach-loop (over branches emanating from the topnode). #
		$br2indels{$br} = [ [$char_pres, $char_abs, [$indx_null_blk]] ];

	    }				# End of the 1st auxiliary foreach-loop (over branches emanating from the topnode). #

		# The gap-state at the top-node is 'presence', those at the other nodes are 'absence'. #

	    my @blwise_gstat_top = ();
	    $blwise_gstat_top[$indx_null_blk] = $char_pres;
	    my %node2blwise_gapstat = ($topnode_id => \@blwise_gstat_top);

	    if ($fl_full) {
		foreach my $node (keys %{$node2pa}) { # 2nd auxiliary foreach-loop (over non-top nodes). #
		    my @blwise_gstat = ();
		    $blwise_gstat[$indx_null_blk] = $char_abs;
		    $node2blwise_gapstat{$node} = \@blwise_gstat;
		}			# End of the 2nd auxiliary foreach-loop (over non-top nodes). #
	    }

	    push @set_psm_scns_null_blk, [\%br2indels, \%node2blwise_gapstat];

	    if (@{$sub_topnodes} < 2) { # The top node has only one child. Then, the scenario with #{indel events} = 1 should be the parsimonious. #
		return (1, \@set_psm_scns_null_blk);
	    }

	} # End of "if (@{$sub_topnodes} < 3) {...}"


		## Then, enumerate the other scenarios. ##

	foreach my $node (keys %{$node2pa}) { # Main outer foreach-loop (over non-top nodes). #

		# Part1: There is always a scenario in which an insertion occurs and a deletion follows along the same branch. #
		# (In this case, the gap states are 'absence' at all the nodes.)

	    my %br2indels1;
	    $br2indels1{$node} = [[$char_abs, $char_pres, [$indx_null_blk]], [$char_pres, $char_abs, [$indx_null_blk]] ];
	    my @blwise_gstat1_top = ();
	    $blwise_gstat1_top[$indx_null_blk] = $char_abs;
	    my %node2blwise_gapstat1 = ($topnode_id => \@blwise_gstat1_top);

	    if ($fl_full) {
		foreach my $node1 (keys %{$node2pa}) { # 1st main inner foreach-loop (over non-top nodes). #
		    my @blwise_gstat1 = ();
		    $blwise_gstat1[$indx_null_blk] = $char_abs;
		    $node2blwise_gapstat1{$node1} = \@blwise_gstat1;
		}			# End of the 1st main inner foreach-loop (over non-top nodes). #
	    }

	    push @set_psm_scns_null_blk, [\%br2indels1, \%node2blwise_gapstat1];

	    my $children = $node2ch->{$node};

	    ((defined $children) and (@{$children} ==1)) or next;

		# Part2: WHEN THE NODE HAS EXACTLY ONE CHILD, there is also a scenario in which
		#	 an insertion occurs above the node and a deletion occurs below the node.
		# (In this case, the gap state is 'presence' at the node, and the states are 'absence' elsewhere.)

	    my %br2indels2;
	    $br2indels2{$node} = [ [$char_abs, $char_pres, [$indx_null_blk]] ];
	    $br2indels2{$children->[0]} = [ [$char_pres, $char_abs, [$indx_null_blk]] ];

		# The top-node always has the 'absence' gap-state. #
	    my @blwise_gstat2_top = ();
	    $blwise_gstat2_top[$indx_null_blk] = $char_abs;
	    my %node2blwise_gapstat2 = ($topnode_id => \@blwise_gstat2_top);

	    if ($fl_full) {
		foreach my $node2 (keys %{$node2pa}) { # 2nd main inner foreach-loop (over non-top nodes). #
		    my @blwise_gstat2 = ();
		    $blwise_gstat2[$indx_null_blk] = ($node == $node2) ? $char_pres : $char_abs;
		    $node2blwise_gapstat2{$node2} = \@blwise_gstat2;
		}			# End of the 2nd main inner foreach-loop (over non-top nodes). #
	    }

	    push @set_psm_scns_null_blk, [\%br2indels2, \%node2blwise_gapstat2];

	}			# End of the main outer foreach-loop (over non-top nodes). #


    (print STDOUT "\n   < Subroutine 'enum_psm_scns_gen_null_blk' Ended />\n\n") if ($TEST_LEVEL_GEN_NULL_BLK>0);

	return (2, \@set_psm_scns_null_blk);

}	# End of the subroutine 'enum_psm_scns_gen_null_blk'. #


# create_list_to_be_examined:
#	This satellite subroutine of 'enum_psm_indel_scns_in_cls_blks' creates
#	a list of indel events to be examined in order to make
#	the truly parsimonious indel scenario(s).
#
#	REVISED on Nov 13 (Wed), 2013:
#	Now, the subroutine correctly handles trees with two-branch nodes.
#
# [INPUT: (\%node2ct_exoffs, $ct_seqs, $char_abs, \%br2indels), where ] -- Obsolete as of Nov 13 (Wed), 2013.
# INPUT: (\%node2ch, \%node2ct_exoffs, $ct_seqs, $char_abs, \%br2indels), where
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	%{$br2indels} = ($branch_id => \@list_indels, ...) for the set of indels, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices in @{$blocks_in_cls}.
#
# OUTPUT: \@list_to_be_examined, where
#	@{$list_to_be_examined[$k]->[$i]} = ($branch_id, index in the list @{$br2indels{$branch_id}})
#	is for the $i th indel event that result in $k exterior nodes with $char_abs (for 0<$k<$ct_seqs).
#	@{$list_to_be_examined[0]} should always be empty.
#
##sub create_list_to_be_examined (\%$$\%) {
#	my ($node2ct_exoffs, $ct_seqs, $char_abs, $br2indels) = @_;
#
sub create_list_to_be_examined (\%\%$$\%) {
	## my $list_indels_to_be_examined0 = create_list_to_be_examined (%{$node2ch}, %{$node2ct_exoffs}, $ct_seqs, $char_abs, %{$br2indels0});

	my ($node2ch, $node2ct_exoffs, $ct_seqs, $char_abs, $br2indels) = @_;

    (print STDOUT "  < Subroutine 'create_list_to_be_examined' (#{sequences} = $ct_seqs) >\n\n") if ($TEST_LEVEL2>0);

	my @list_to_be_examined = ();
	for (my $k=0; $k<$ct_seqs; $k++) { $list_to_be_examined[$k] = []; }

	foreach my $br (sort {$a <=> $b} keys %{$br2indels}) {  # Main outer foreach-loop #

	    my $ct_exoffs = $node2ct_exoffs->{$br};
	    my $list_indels = $br2indels->{$br};
	    my $ct_indels = @{$list_indels};
	    my $ct_children = (defined $node2ch->{$br}) ? scalar (@{$node2ch->{$br}}) : 0; # ADDED in revision on Nov 13 (Wed), 2013.

	(print STDOUT "    [ Branch ID = $br (#{indels} = $ct_indels) ]\n") if ($TEST_LEVEL2>1);

	    for (my $indx=0; $indx<$ct_indels; $indx++) { # Main inner for-loop #
		my ($stat_pa, $stat_node, $indices_blk) = @{$list_indels->[$indx]};

			# ADDED in revision on Nov 13 (Wed), 2013. #
		(($stat_node eq $char_abs) and ($ct_children == 0)) and next; # The 'absence' state canNOT recede any more if it is at the lower-node that has no children.

		my $k = ($stat_node eq $char_abs)  ? $ct_exoffs : $ct_seqs - $ct_exoffs;

		push @{$list_to_be_examined[$k]}, [$br, $indx];

	    }				# End of the main inner for-loop #
	}			# End of the main outer foreach-loop #

    (print STDOUT "\n   < Subroutine 'create_list_to_be_examined' ended />\n\n") if ($TEST_LEVEL2>0);

	return \@list_to_be_examined;

}	# End of the subroutine 'create_list_to_be_examined' #


# list_mvd_events:
#	This satellite subroutine of 'recede_absence' lists
#	the new indel events resulting from receding an 'absence' state
#	from one end of the branch along which the input indel event occurred.
#
# INPUT: ($br_id, $stat_le, $char_abs, $char_pres, \%node2pa, \%node2ch), where
#	$br_id is the ID of the branch where the indel event occurred;
#	$stat_le is the state of the lower-end of the branch;
#	$char_abs/pres is the character representing the 'absence'/'presence' state;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#
# OUTPUT: \@mvd_events, where
#	@{$mvd_events[$e]} = ($new_br_id, $new_stat_ue, $new_stat_le)
#	is for the $e th event resulting from the absence recession, with
#	$new_br_id specifying the branch where the event occurs,
#	$new_stat_ue/le representing the new state at the upper-end/lower-end of the branch.
#	@mvd_events is empty if the 'absence' state cannot recede any further.
#
sub list_mvd_events ($$$$\%\%) {

	my ($br_id, $stat_le, $char_abs, $char_pres, $node2pa, $node2ch) = @_;

    (print STDOUT "     < Subroutine 'list_mvd_events' (Branch ID = $br_id) >\n\n") if ($TEST_LEVEL3>0);

	my @mvd_events = ();

	if ($stat_le eq $char_abs) { # The lower-end shows the 'absence' state.

	    my $new_branches = $node2ch->{$br_id};
	    unless (defined $new_branches) {
		# The lower-end is an exterior node,
		# so the 'absence' state cannot recede any further.
		return [];
	    }
	    foreach my $new_br (@{$new_branches}) { # 1st main foreach-loop #
		push @mvd_events, [$new_br, $char_pres, $char_abs];
	    }			# End of the 1st main foreach-loop #

	} else { # The upper-end shows the 'absence' state. #

	    my $pa_br = $node2pa->{$br_id};
	    if (defined $node2pa->{$pa_br}) { # The 'parent branch exists'. #
		push @mvd_events, [$pa_br, $char_abs, $char_pres];
	    }
	    my $new_branches = $node2ch->{$pa_br};
	    foreach my $new_br (@{$new_branches}) { # 2nd main foreach-loop #
		($new_br == $br_id) and next;
		push @mvd_events, [$new_br, $char_pres, $char_abs];
	    }			# End of the 2nd main foreach-loop #
	    
	}

    (print STDOUT "\n     < Subroutine 'list_mvd_events' ended />\n\n") if ($TEST_LEVEL3>0);

	return \@mvd_events;

}	# End of the subroutine 'list_mvd_events' #


# search_flanking_blocks:
#	This satellite subroutine of 'recede_absence'
#	searches for flanking blocks of the indel event in question.
#	A flanking block is defined as the closest block with the 'presence' state
#	at both ends of the branch along which the event occurred.
#
# INPUT: ($node_ue, $node_le, \@sbjct_indel_event, \%node2blwise_gstat, $char_abs), where
#	[ $br_id is the ID of the lower-end node of the branch along which the event occurred; ] --- Obsolete
#	$node_ue/le is the ID of the node that is at the upper-end/lower-end of the branch along which the event occurred;
#	@sbjct_indel_event = ($stat_ue, $stat_le, \@indices_sb_blk) describes the subject indel event, with 
#	$stat_ue/le being the state at the upper-end/lower-end node of the branch, and
#	@indices_sb_blk storing the indices of the blocks involved in the event;
#
#	%node2blwise_gstat = ($node_id => \@blwise_gstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gstats[$b] being $char_pres/abs for the $b th block (in the cluster) at the node $node_id;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: ($if_success, $lflank_blk, $rflank_blk), where
#	$if_success = 1 if the subroutine successfully ended,
#		    = 0 if it failed;
#	$l/rflank_blk is the index of the left/right flanking block.
#
sub search_flanking_blocks ($$\@\%$) {

	my ($node_ue, $node_le, $sbjct_indel_event, $node2blwise_gstat, $char_abs) = @_;

	my ($sb_stat_ue, $sb_stat_le, $indices_sb_blk) = @{$sbjct_indel_event};
	my $lm_sb_blk = $indices_sb_blk->[0];
	my $rm_sb_blk = $indices_sb_blk->[$#{$indices_sb_blk}];

    (print STDOUT "     < Subroutine 'search_flanking_blocks' (leftmost subject block = $lm_sb_blk; rightmost subject block = $rm_sb_blk) >\n\n") if ($TEST_LEVEL3>0);

	my $blwise_gstats_ue = $node2blwise_gstat->{$node_ue};
	my $blwise_gstats_le = $node2blwise_gstat->{$node_le};
	my $ct_blks = @{$blwise_gstats_ue};

	my $lflank_blk;
	for (my $b=$lm_sb_blk-1; $b>=0; $b--) { # 1st main for-loop #
	    my $stat_ue = $blwise_gstats_ue->[$b];
	    my $stat_le = $blwise_gstats_le->[$b];
	    if (($stat_ue eq $sb_stat_ue) and ($stat_le eq $sb_stat_le)) {
		print STDOUT "[ERROR 'search_flanking_blocks'] IMPOSSIBLE! While looking for the left flank block, I encountered the block ($b) with the states at the lower- and the upper-ends identical to those of the subject event (from the $lm_sb_blk th block to the $rm_sb_blk th block): $! !\n";
		return (0, undef, undef);
	    }
	    (($stat_ue eq $char_abs) or ($stat_le eq $char_abs)) and next;
	    $lflank_blk = $b;
	    last; 
	}		# End of the 1st main for-loop #

    (print STDOUT "       --- Finished searching for the left-flanking block, now start searching for the right-flanking one ...\n\n") if ($TEST_LEVEL3>1);

	my $rflank_blk;
		# A BUG FIXED on May 13 (Sun), 2012. #
	for (my $b=$rm_sb_blk +1; $b<$ct_blks; $b++) { # 2nd main for-loop #
	##for (my $b=$rm_sb_blk; $b<$ct_blks; $b++) { # 2nd main for-loop #
	    my $stat_ue = $blwise_gstats_ue->[$b];
	    my $stat_le = $blwise_gstats_le->[$b];
	    if (($stat_ue eq $sb_stat_ue) and ($stat_le eq $sb_stat_le)) {
		print STDOUT "[ERROR 'search_flanking_blocks'] IMPOSSIBLE! While looking for the right flanking block, I encountered the block ($b) with the states at the lower- and the upper-ends identical to those of the subject event (from the $lm_sb_blk th block to the $rm_sb_blk th block): $! !\n";
		return (0, undef, undef);
	    }
	    (($stat_ue eq $char_abs) or ($stat_le eq $char_abs)) and next;
	    $rflank_blk = $b;
	    last; 
	}		# End of the 2nd main for-loop #

    (print STDOUT "\n     < Subroutine 'search_flanking_blocks' ended />\n\n") if ($TEST_LEVEL3>0);

	return (1, $lflank_blk, $rflank_blk);

}	# End of the subroutine 'search_flanking_blocks' #


# search_flanking_blocks2:
#	This subroutine was created by modifying 'search_flanking_blocks'
#	in order to search for flanking blocks in a broad sense,
#	including the blocks where either of the ends of the branch shows the 'presence' state.
#	(CREATED on Oct 23 (Tue), 2012 => IMPORTED on Oct 24 (Wed), 2012)
#
# search_flanking_blocks:
#	This satellite subroutine of 'recede_absence'
#	searches for flanking blocks of the indel event in question.
#	A flanking block is defined as the closest block with the 'presence' state
#	at both ends of the branch along which the event occurred.
#
# INPUT: ($node_ue, $node_le, \@sbjct_indel_event, \%node2blwise_gstat, $char_abs), where
#	[ $br_id is the ID of the lower-end node of the branch along which the event occurred; ] --- Obsolete
#	$node_ue/le is the ID of the node that is at the upper-end/lower-end of the branch along which the event occurred;
#	@sbjct_indel_event = ($stat_ue, $stat_le, \@indices_sb_blk) describes the subject indel event, with 
#	$stat_ue/le being the state at the upper-end/lower-end node of the branch, and
#	@indices_sb_blk storing the indices of the blocks involved in the event;
#
#	%node2blwise_gstat = ($node_id => \@blwise_gstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gstats[$b] being $char_pres/abs for the $b th block (in the cluster) at the node $node_id;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: ($if_success, $lflank_blk, $rflank_blk), where
#	$if_success = 1 if the subroutine successfully ended,
#		    = 0 if it failed;
#	$l/rflank_blk is the index of the left/right flanking block.
#
sub search_flanking_blocks2 ($$\@\%$) {

	my ($node_ue, $node_le, $sbjct_indel_event, $node2blwise_gstat, $char_abs) = @_;

	my ($sb_stat_ue, $sb_stat_le, $indices_sb_blk) = @{$sbjct_indel_event};
	my $lm_sb_blk = $indices_sb_blk->[0];
	my $rm_sb_blk = $indices_sb_blk->[$#{$indices_sb_blk}];

    (print STDOUT "     < Subroutine 'search_flanking_blocks2' (leftmost subject block = $lm_sb_blk; rightmost subject block = $rm_sb_blk) >\n\n") if ($TEST_LEVEL3>0);

	my $blwise_gstats_ue = $node2blwise_gstat->{$node_ue};
	my $blwise_gstats_le = $node2blwise_gstat->{$node_le};
	my $ct_blks = @{$blwise_gstats_ue};

	my $lflank_blk;
	for (my $b=$lm_sb_blk-1; $b>=0; $b--) { # 1st main for-loop #
	    my $stat_ue = $blwise_gstats_ue->[$b];
	    my $stat_le = $blwise_gstats_le->[$b];
	    if (($stat_ue eq $sb_stat_ue) and ($stat_le eq $sb_stat_le)) {
		print STDOUT "[ERROR 'search_flanking_blocks2'] IMPOSSIBLE! While looking for the left flank block, I encountered the block ($b) with the states at the lower- and the upper-ends identical to those of the subject event (from the $lm_sb_blk th block to the $rm_sb_blk th block): $! !\n";
		return (0, undef, undef);
	    }
	    (($stat_ue eq $char_abs) and ($stat_le eq $char_abs)) and next;
	    $lflank_blk = $b;
	    last; 
	}		# End of the 1st main for-loop #

    (print STDOUT "       --- Finished searching for the left-flanking block, now start searching for the right-flanking one ...\n\n") if ($TEST_LEVEL3>1);

	my $rflank_blk;
		# A BUG FIXED on May 13 (Sun), 2012. #
	for (my $b=$rm_sb_blk +1; $b<$ct_blks; $b++) { # 2nd main for-loop #
	##for (my $b=$rm_sb_blk; $b<$ct_blks; $b++) { # 2nd main for-loop #
	    my $stat_ue = $blwise_gstats_ue->[$b];
	    my $stat_le = $blwise_gstats_le->[$b];
	    if (($stat_ue eq $sb_stat_ue) and ($stat_le eq $sb_stat_le)) {
		print STDOUT "[ERROR 'search_flanking_blocks2'] IMPOSSIBLE! While looking for the right flanking block, I encountered the block ($b) with the states at the lower- and the upper-ends identical to those of the subject event (from the $lm_sb_blk th block to the $rm_sb_blk th block): $! !\n";
		return (0, undef, undef);
	    }
	    (($stat_ue eq $char_abs) and ($stat_le eq $char_abs)) and next;
	    $rflank_blk = $b;
	    last; 
	}		# End of the 2nd main for-loop #

    (print STDOUT "\n     < Subroutine 'search_flanking_blocks2' ended />\n\n") if ($TEST_LEVEL3>0);

	return (1, $lflank_blk, $rflank_blk);

}	# End of the subroutine 'search_flanking_blocks2' #


# incorporate_new_event:
#	This subroutine of 'recede_absence' incorporates a new indel event 
#	into a map of events onto the branches.
#
#	TEMPORARILY REVISED on Mar 11 (Tue), 2014:
#	The subroutine tentatively cope with rare situations where
#	the 'new' event has BOTH left-neighboring and right-neighboring indel events.
#
# INPUT: (\@new_event, \@indices_new_blk, \%br2indels, \%node2blwise_gstat, $char_abs, \%node2pa), where
#	@new_event = ($new_br_id, $new_stat_ue, $new_stat_le) represents the new indel event,
#	with $new_br_id specifying the branch, and $new_stat_ue/le specifying the state at the upper-end/lower-end;
#	@indices_new_blk stores the indices of the blocks involved in the new event;
#
#	%br2indels = ($br_id => \@list_indels, ...) maps indel events onto the branches,
#	with @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls)
#	of the blocks involved in the indel event;
#
#	%node2blwise_gstat = ($node_id => \@blwise_gstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gstats[$b] being $char_pres/abs for the $b th block (in the cluster) at the node $node_id;
#
#	$char_abs is the character representing the 'absence' state;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node.
#
# OUTPUT: $if_absorbed, where
#	$if_absorbed = 1 if the new indel event is absorbed into the existing event,
#		     = 0 if the event remain independent of any other events.
#		     = undef if there is some error.
#
#	REMEMBER that the main job of this subroutine is to incorporate the new event
#	into the map of events onto the branches.
#
sub incorporate_new_event (\@\@\%\%$\%) {

	my ($new_event, $indices_new_blk, $br2indels, $node2blwise_gstat, $char_abs, $node2pa) = @_;

	my ($new_br_id, $new_stat_ue, $new_stat_le) = @{$new_event}; 

    (print STDOUT "     < Subroutine 'incorporate_new_event' (New branch ID = $new_br_id) >\n\n") if ($TEST_LEVEL3>0);

	my @new_indel = ($new_stat_ue, $new_stat_le, $indices_new_blk);

	my $list_indels_on_new_br = $br2indels->{$new_br_id};
	unless ((defined $list_indels_on_new_br) and (@{$list_indels_on_new_br}>0)) { # No events so far along the branch. #
	    $br2indels->{$new_br_id} = [\@new_indel];

    (print STDOUT "\n     < Subroutine 'incorporate_new_event' ended PREMATURELY (because there was NO events along the branch $new_br_id). />\n\n") if ($TEST_LEVEL3>0);

	    return 0;
	}
	my @list_indels_on_new_br = @{$list_indels_on_new_br};
	$br2indels->{$new_br_id} = \@list_indels_on_new_br;
	my $ct_indels= @list_indels_on_new_br;

	my $lm_new_blk = $indices_new_blk->[0]; # Leftmost block involved in the new event.
	my $rm_new_blk = $indices_new_blk->[$#{$indices_new_blk}]; # Rightmost block involved in the new event.

    (print STDOUT "     ++ Now the main process begins (#{indels to be examined} = $ct_indels). ++\n\n") if ($TEST_LEVEL3>1);

	my ($indx_left_nb_event, $indx_right_nb_event); # Left neighboring, right neighboring, and "embracing" events of the same type. #
	my $rm_blk_lnb; # Rightmost block in the left neighboring event.
	my $lm_blk_rnb; # Leftmost block in the right neighboring event. 

	for (my $id=0; $id<$ct_indels; $id++) { # 1st main for-loop (over indel events along $new_br_id) #

	(print STDOUT "      [ The $id th indel ]\n") if ($TEST_LEVEL3>1);

	    my $indel = $list_indels_on_new_br[$id];
	    (@{$indel}>0) or next; # Skip an empty event. #
	    my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
	    (($stat_ue eq $new_stat_ue) and ($stat_le eq $new_stat_le)) or next; # Keep only the events of the same type. #

	    my $lm_blk = $indices_blk->[0];
	    my $rm_blk = $indices_blk->[$#{$indices_blk}];

	    if ($lm_blk<$lm_new_blk) {
		if ($rm_new_blk <$rm_blk) {
			# The current indel "embrace"s the new indel
			# => incorporate the latter into the former.

		    my $mgd_indices_blk = union_nmarray (@{$indices_blk}, @{$indices_new_blk});
		    my @indel = ($stat_ue, $stat_le, $mgd_indices_blk);
		    $list_indels_on_new_br[$id] = \@indel;
		    return 1;
		}

		# The event is on the left of the  new event. #
		# => Compare with the current left-neighboring event.
		unless ((defined $rm_blk_lnb) and ($rm_blk < $rm_blk_lnb)) {
		    $indx_left_nb_event = $id;
		    $rm_blk_lnb = $rm_blk;
		}

	    } elsif ($rm_new_blk < $rm_blk) {
		# The event is on the right of the new event. #
		# => Compare with the current right-neighboring event.
		unless ((defined $lm_blk_rnb) and ($lm_blk_rnb < $lm_blk)) {
		    $indx_right_nb_event = $id;
		    $lm_blk_rnb = $lm_blk;
		}
	    }

	}		# End of the 1st main foreach-loop (over indel events along $new_br_id) #

    (print STDOUT "\n      ++ The main process Ended ++\n\n") if ($TEST_LEVEL3>1);


		# Examine whether the neighboring event, if any, is genuine or not. #

	my ($new_node_ue, $new_node_le) = ($node2pa->{$new_br_id}, $new_br_id);
	my $blwise_gstats_ue = $node2blwise_gstat->{$new_node_ue};
	my $blwise_gstats_le = $node2blwise_gstat->{$new_node_le};

	if (defined $indx_left_nb_event) {
	    # Examine whether the subject and the left neighboring events are separated by (a) 'barrier'(s). #
    (print STDOUT '      ++ Examine whether the $indx_left_nb_event (=', $indx_left_nb_event, ') th event is genuinely left-neighboring or not. ++', "\n\n") if ($TEST_LEVEL3>1);
 
	    for (my $b = $lm_new_blk-1; $b>$rm_blk_lnb; $b--) { # 1st trailing for-loop #
		(($blwise_gstats_ue->[$b] eq $char_abs) or ($blwise_gstats_le->[$b] eq $char_abs)) and next;
		# Separated by a barrier. => The neighboring event is not genuine. #
		$indx_left_nb_event = undef;
		last;
	    }			# End of the 1st trailing for-loop #
	}
	if (defined $indx_right_nb_event) {
	    # Examine whether the two events are separated by (a) 'barrier'(s). #

    (print STDOUT '      ++ Examine whether the $indx_right_nb_event (=', $indx_right_nb_event, ') th event is genuinely right-neighboring or not. ++', "\n\n") if ($TEST_LEVEL3>1);

	    for (my $b = $rm_new_blk+1; $b<$lm_blk_rnb; $b++) { # 2nd trailing for-loop #
		(($blwise_gstats_ue->[$b] eq $char_abs) or ($blwise_gstats_le->[$b] eq $char_abs)) and next;
		# Separated by a barrier. => The neighboring event is not genuine. #
		$indx_right_nb_event = undef;
		last;
	    }			# End of the 2nd trailing for-loop #
	}

		# Finally, incorporate the new event into the map. #


		# Blockwise gap-states at the 'destination' node of the new 'deletion'. #
		# (TEMPORARILY ADDED to tentatively cope with rare situations (on Mar 11 (Tue), 2014).) #
	my $blwise_gstats_dest = ($new_stat_le eq $char_abs) ? $blwise_gstats_le : $blwise_gstats_ue ;
	my $incorporating;
	if (defined $indx_left_nb_event) {
	    if (defined $indx_right_nb_event) {

		    # TEMPORARILY MODIFIED to tentatively cope with rare situations (on Mar 11 (Tue), 2014). #
		    # In this case, the blocks involved in the subject 'new' indel event
		    # must be split by some intervening 'presence' states at the 'destination' node.
		my @intervening = ();
		for (my $g = $lm_new_blk+1; $g<$rm_new_blk; $g++) { ($blwise_gstats_dest->[$g] ne $char_abs) and (push @intervening, $g); }
		if (@intervening>0) {
		    print STDOUT "[WARNING 'incorporate_new_event'] The new event has both left- and right-neighboring indel events, intervened by the 'presence' states in the blocks, (", join (', ', @intervening), ") at the 'destination' node of the new 'deletion' event. ... I will tentatively STOP PROCESSING this new event ...: $! !\n\n";
    (print STDOUT "\n     < Subroutine 'incorporate_new_event' stopped (with the new indel event NOT incorporated into any existing event) />\n\n") if ($TEST_LEVEL3>0); 
		    return 0;

		    ### I WILL SERIOUSLY ADDRESS THIS PROBLEM LATER!! ###

		} else {
		    print STDOUT "[ERROR 'incorporate_new_event'] IMPOSSIBLE! There are BOTH left- and right-neighboring indel events for the event, while there is NO intervening 'presence' state at the 'destination' node of the new 'deletion' : $! !\n";
		    return undef;
		}
		##print STDOUT "[ERROR 'incorporate_new_event'] IMPOSSIBLE! There are BOTH left- and right-neighboring indel events for the event : $! !\n";
		##return undef;
	    }
	    $incorporating = $indx_left_nb_event;
	} elsif (defined $indx_right_nb_event) {
	    $incorporating = $indx_right_nb_event;
	}

	if (defined $incorporating) {
	    my ($stat_ue, $stat_le, $indices_old_blk) = @{$list_indels_on_new_br[$incorporating]};
	    my $mgd_indices_blk = union_nmarray (@{$indices_old_blk}, @{$indices_new_blk});

	    my @indel = ($stat_ue, $stat_le, $mgd_indices_blk);
	    $list_indels_on_new_br[$incorporating] = \@indel;

    (print STDOUT "\n     < Subroutine 'incorporate_new_event' ended (with the new indel event incorporated into the $incorporating th event) />\n\n") if ($TEST_LEVEL3>0); 

	    return 1;

	} else {
	    # There is no genuinely neighboring or "embracing" events. #
	    push @list_indels_on_new_br, \@new_indel;

    (print STDOUT "\n     < Subroutine 'incorporate_new_event' ended (with the new indel event NOT incorporated into any existing event) />\n\n") if ($TEST_LEVEL3>0); 

	    return 0;
	}

}	# End of the subroutine 'incorporate_new_event' #


# enum_events_involve_blk:
#	This satellite subroutine of 'recede_absence' enumerates the events
#	involving a specified gap-block.
#	It is used to see whether the event is worth trying to 'recede' in a complex way,
#	using 'create_bestset_mvd_events' and 'incorporate_new_event'.
#	(CREATED in revision on Mar 20 (Thu), 2014)
#
# INPUT: ($indx_blk, \%br2indels), where
#	$indx_blk is the index of the specified block in the array of blocks constituting a cluster;
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls)
#	of the blocks involved in the indel event;
#
# OUTPUT: \@list_rlv_events, where
#	@{$list_rlv_events[$re]} = ($branch_id, $indx_event) for the $re th event involving the block,
#	with $indx_event being the index specifying the event in @{$br2indels{$branch_id}.
#
sub enum_events_involve_blk ($\%) {

	my ($indx_blk, $br2indels) = @_;

	my @list_rlv_events = ();

	foreach my $br (keys %{$br2indels}) { # Main foreach-loop (over branches). #
	    my $indels = $br2indels->{$br};
	    my $ct_indels = @{$indels};
	    for (my $id = 0; $id<$ct_indels; $id++) { # Middle for-loop (over indels). #
		my ($stat_ue, $stat_le, $indices_blk) = @{$indels->[$id]};
		foreach my $blk (@{$indices_blk}) { # Inner foreach-loop (over involved blocks). #
		    if ($blk == $indx_blk) {
			push @list_rlv_events, [$br, $id];
			last;
		    }
		}			# End of the inner foreach-loop (over involved blocks). #
	    }			# End of the middle for-loop (over indels). #
	}		# End of the main foreach-loop (over branches). #

	return \@list_rlv_events;

}		# End of the subroutine 'enum_events_involve_blk'. #


# create_bestset_mvd_events:
#	This satellite subroutine of 'recede_absence' creates
#	a 'best' set of moved indel events, which has the least events that cannot be absorbed into the neighboring events.
#	If multiple events tie, the set with the least events is chosen.
#	NOTE that the one-element set of the subject event is always excluded.
#
# INPUT: ($sb_br_id, \@sbjct_indel_event, $lflank_blk, $rflank_blk, \%node2blwise_gstat, \%node2pa, \%node2ch, $char_pres, $char_abs), where
#	$sb_br_id is the subject branch ID;
#	@sbjct_event = ($stat_ue, $stat_le, \@indices_sb_blk) describes the subject indel event, with
#	$stat_ue/le representing the state at the upper-end/lowe-end of the branch along which the event occurs,
#	@indices_sb_blk storing the indices of the blocks involved in the event;
#
#	$l/rflank_blk is the index of the left/right-flanking block;
#
#	%node2blwise_gstat = ($node_id => \@blwise_gstats, ...) maps states onto nodes, with
#	$blwise_gstats[$b] representing the gap state of the $b th block on the node $node_id;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@children, ...) for all interior nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state.
#
# OUTPUT: ($exp_diff_totct_indels, \@bestset_mvd_events, %node2flipped_state), where
#	$exp_diff_totct_indels is the expected difference in the total number of indels resulting from the best set of moved events;
#
#	@{$bestset_mvd_events[$e]} = ($new_br_id, $new_stat_ue, $new_stat_le)	
#	represents the $e th moved event in the 'best' set, with
#	$new_br_id being the ID of the branch where the new indel occurs,
#	$new_state_ue/le being the new state at the lower/upper-end of the branch;
#
#	%node2flipped_state = ($node_id => 1, ...) for the nodes (including the pivotal node) whose state flipped from $char_abs to $char_pres as a result of moving the events.
#	
sub create_bestset_mvd_events ($\@$$\%\%\%$$) {

	my ($sb_br_id, $sbjct_indel_event, $lflank_blk, $rflank_blk, $node2blwise_gstat, $node2pa, $node2ch, $char_pres, $char_abs) = @_;

    (print STDOUT "   < Subroutine 'create_bestset_mvd_events' (Subject branch ID = $sb_br_id) >\n\n") if ($TEST_LEVEL2>1);

	# @{$mvd_events->[$e]} = ($new_br_id, $new_stat_ue, $new_stat_le)
	# is for the $e th event resulting from the absence recession, with
	# $new_br_id specifying the branch where the event occurs,
	# $new_stat_ue/le representing the new state at the upper-end/lower-end of the branch.

	my ($stat_pa, $stat_node, $indices_sb_blk) = @{$sbjct_indel_event};
	my $pivotal_node = ($stat_node eq $char_abs) ? $sb_br_id : $node2pa->{$sb_br_id}; # The 'end'-node  with the 'absence' state.
	my %uphill; # = ($br_id/$node_id  => 1, ...) for branches/nodes representing 'uphill' events (from the subject branch/node)
	if ($stat_pa eq $char_abs) {
	    my $uh_node = $sb_br_id;
	    while (defined $uh_node) { # Preliminary while-loop #
		$uphill{$uh_node} = 1;
		$uh_node = $node2pa->{$uh_node};
	    }			# End of the preliminary while-loop #
	}
    if ($TEST_LEVEL2c>2) {
	print STDOUT "\n{Nodes on the 'uphill' track} = ", join (', ', sort {$a <=> $b} keys %uphill), ".\n\n";
    } 

		# The following auxiliary arrays/hashes will be defined
		# for nodes/branches 'under' the pivotal node.

	my %br2rel_depth; # ($branch_id => $relative depth, ...), $relative_depth = 0 for the subject branch;
	my %node2eff_ch; # ($node_id => \@effective_chnodes, ...), where \@effective_chnodes consists only of children for $uphill{$node_id} is undefined, otherwise it consists of the child nodes (except those with defined $hphill_events{$node}) and the parent node of $node_id.
	my %br2eff_ch; # ($branch_id => \@effective_chbras, ...), where \@effective_chbras consists only of child branches for $uphill$br_id} = undef, otherwise it consists of the sibling branches and the parent branch.
	my %br2flanking_deletion; # = ($branch_id => 'L'/'R'/'LR' if deletion occurred in the left/right/both flanking block(s), ...)
	my %node2flank_pres; # = ($node_id => 1, ...) for the nodes where the BOTH flanking blocks show the 'presence' state.

    (print STDOUT '    ++ The 1st process (PRELIMINARY): Constructing %br2rel_depth, %node2eff_ch, %br2eff_ch, and %br2flanking_deletion. ++', "\n\n") if ($TEST_LEVEL2>1);

	my $rel_depth=0;
	my @subjects = ($sb_br_id);
	while (@subjects>0) { # 1st main outer while-loop #
	    my @new_subjects = ();

	    my $ct_sbjcts = @subjects;
	(print STDOUT "\n    [ Relative depth = $rel_depth (#{subject branches} = $ct_sbjcts) ]\n\n") if ($TEST_LEVEL2>1);

	    while (my $br = shift @subjects) { # Middle while-loop #

	(print STDOUT "     ( Branch ID = $br )\n") if ($TEST_LEVEL2>1);

		$br2rel_depth{$br} = $rel_depth;
		my $pa_br = $node2pa->{$br};
		my @eff_chnodes = my @eff_chbras = ();
		my ($node, $upstream) = (defined $uphill{$br}) ? ($pa_br, $br) : ($br, $pa_br);

		my $blwise_gstats_nd = $node2blwise_gstat->{$node};
		my $blwise_gstats_us = $node2blwise_gstat->{$upstream};

		my ($gstat_nd_lf, $gstat_nd_rf) = @{$blwise_gstats_nd}[$lflank_blk, $rflank_blk]; # At the subject node. #
		my ($gstat_us_lf, $gstat_us_rf) = @{$blwise_gstats_us}[$lflank_blk, $rflank_blk]; # At the upstream node. #

			## A BUG fixed on May 17 (Thu), 2012. ##
		if (($gstat_nd_lf ne $char_abs) and ($gstat_nd_rf ne $char_abs)) {
		##if (($gstat_nd_lf ne $char_abs) and ($gstat_nd_lf ne $char_abs)) {
		    $node2flank_pres{$node} = 1;
		} else {
		    my @deletions = ();
		    (($gstat_us_lf ne $char_abs) and ($gstat_nd_lf eq $char_abs)) and (push @deletions, 'L');
		    (($gstat_us_rf ne $char_abs) and ($gstat_nd_rf eq $char_abs)) and (push @deletions, 'R');

			# A BUG fixed on Mar 20 (Thu), 2014. #
		    if (@deletions>0) {
			$br2flanking_deletion{$br} = join ('', @deletions);
			next; # Once encounter (a) deletion(s) in the flanking region(s), not explore 'under' such a branch.
		    }
		    ##(@deletions>0) and ($br2flanking_deletion{$br} = join ('', @deletions));
		    ##next; # Once encounter (a) deletion(s) in the flanking region(s), not explore 'under' such a branch.
		}

		(defined $node2ch->{$node}) or next; # Skip the downhill exterior branch. #

		if (defined $uphill{$br}) {
		    foreach my $ch_br (@{$node2ch->{$node}}) { ($ch_br != $br) and (push @eff_chbras, $ch_br); }
		    @eff_chnodes = @eff_chbras;
		    my $pa_node = $node2pa->{$node};
		    if (defined $pa_node) {
			push @eff_chnodes, $pa_node;
			push @eff_chbras, $pa_br;
		    }
		} else {
		    @eff_chnodes = @eff_chbras = @{$node2ch->{$br}};
		}
		$br2eff_ch{$br} = \@eff_chbras;
		$node2eff_ch{$node} = \@eff_chnodes;
		push @new_subjects, @eff_chbras;
	    }		# End of the middle while-loop #

	    @subjects = @new_subjects;
	    $rel_depth++;

	}	# End of the 1st main outer while-loop #

    if ($TEST_LEVEL2c>2) {
	print STDOUT "\n", '    ++ The 1st process ENDED ++', "\n\n";
	print STDOUT '{Keys of %node2flank_pres} = ', join (', ', sort {$a <=> $b} keys %node2flank_pres), ".\n\n";
	print STDOUT '%br2rel_depth = ( # $branch_id => $relative depth', "\n";
	foreach my $br (sort {$a <=> $b} keys %br2rel_depth) { print STDOUT "\t", $br, ' => ', $br2rel_depth{$br}, ",\n"; }
	print STDOUT "\t", ');', "\n\n";

	print STDOUT '%node2eff_ch = ( # $node_id => [effective,chnode,IDs], ', "\n";
	foreach my $node (sort {$a <=> $b} keys %node2eff_ch) { print STDOUT "\t", $node, ' => [', join (', ', @{$node2eff_ch{$node}}), '],', "\n"; }
	print STDOUT "\t", ');', "\n\n";

	print STDOUT '%br2eff_ch = ( # $br_id => [effective,chbranch,IDs], ', "\n";
	foreach my $br (sort {$a <=> $b} keys %br2eff_ch) { print STDOUT "\t", $br, ' => [', join (', ', @{$br2eff_ch{$br}}), '],', "\n"; }
	print STDOUT "\t", ');', "\n\n";

	print STDOUT '%br2flanking_deletion = ( # $br_id => \'L/R/LR\' if deletion (in direction of increasing relative depth) occurred in the left/right/both flanking block(s), ', "\n";
	foreach my $br (sort {$a <=> $b} keys %br2flanking_deletion) { print STDOUT "\t", $br, ' => ', $br2flanking_deletion{$br}, ",\n"; }
	print STDOUT "\t", ');', "\n\n";
	
    }

		# A bottom-up process to count the deletion events that cannot be absorbed into the flanking events. #
	my %br2ct_nonabsorbed; # ($branch_id => [#{deletions NOT absorbed if the deletion occurs at the branch}, #{deletions NOT absorbed if the deletions occur downstream}, \@downstream_branches_with_absorbed_deletions (only when the 1st or 2nd element is zero) ], ...)

	##my $totct_sb_branches = scalar keys %br2rel_depth;

    (print STDOUT "\n", '    ++ The 2nd process (MAIN 1): Constructing %br2ct_nonabsorbed. ++', "\n\n") if ($TEST_LEVEL2>1);

	foreach my $br (sort {$br2rel_depth{$b} <=> $br2rel_depth{$a}} keys %br2rel_depth) { # 2nd main outer foreach-loop #

	(print STDOUT "    [ Branch ID: $br ]\n") if ($TEST_LEVEL2>1);

	    if (defined $br2flanking_deletion{$br}) {
		$br2ct_nonabsorbed{$br} = [0, undef, [$br]];
		next;
	    }
	    my $eff_chbras = $br2eff_ch{$br};
	    unless (defined $eff_chbras) {
		$br2ct_nonabsorbed{$br} = [1, undef];
		next;
	    } 

	    my $ct_nabs_ds = 0;
	    my @set_ds_abs_brs_chbr = ();
	    foreach my $chbr (@{$eff_chbras}) { # Middle foreach-loop #
		my ($ct_nabs_chbr, $ct_nabs_ds_chbr, $ds_abs_brs_chbr) = @{$br2ct_nonabsorbed{$chbr}};
		if (($ct_nabs_chbr == 0) or ((defined $ct_nabs_ds_chbr) and ($ct_nabs_ds_chbr == 0)) ) {
		    push @set_ds_abs_brs_chbr, @{$ds_abs_brs_chbr};
		} else {
		    # At least one deletion (along the branch in question) cannot be absorbed. #
		    $ct_nabs_ds++;
		}
	    }			# End of the middle foreach-loop #

	    my @ct_nabs = (1, $ct_nabs_ds);
	    if ($ct_nabs_ds == 0) {
		my @std_set_ds_abs_brs_chbr = sort {$a <=> $b} @set_ds_abs_brs_chbr;
		push @ct_nabs, \@std_set_ds_abs_brs_chbr;
	    } 
	    $br2ct_nonabsorbed{$br} = \@ct_nabs;
	    
	}			# End of the 2nd main outer foreach-loop #

    if ($TEST_LEVEL2c>2) {
	print STDOUT "\n", '     ++ The 2nd process ENDED ++', "\n\n";
	print STDOUT '%br2ct_nonabsorbed = ( # $branch_id => #{deletions nonabsorbed if deletion at the branch}/#{deletion nonabsorbed if deletion downstream}/[downstream,branches,with,absorbed,deletions]', "\n\n";
	foreach my $br (sort {$a <=> $b} keys %br2ct_nonabsorbed) {
	    my ($ct_nabs_sbj, $ct_nabs_ds_sbj, $set_ds_abs_brs_sbj) = @{$br2ct_nonabsorbed{$br}};
	    (defined $ct_nabs_ds_sbj) or ($ct_nabs_ds_sbj = 'undef');
	    my $cnct = '[';
	    if (defined $set_ds_abs_brs_sbj) {
 		$cnct .= join (',', @{$set_ds_abs_brs_sbj});
	    }
	    $cnct .= ']';
	    print STDOUT "\t", $br, ' => ', join ('/', $ct_nabs_sbj, $ct_nabs_ds_sbj, $cnct), ",\n";
	}
	print STDOUT "\t", ');', "\n\n";
    }

	my ($ct_nabs_sbjct, $ct_nabs_ds_sbjct, $set_ds_abs_brs_sbjct) = @{$br2ct_nonabsorbed{$sb_br_id}};
	my $exp_diff_totct_indels = $ct_nabs_ds_sbjct - 1;

	if ($exp_diff_totct_indels>0) {
		# No chance of improvement anyway. #

    (print STDOUT "\n   < Subroutine 'create_bestset_mvd_events' Ended (with ", '$exp_diff_totct_indels = ', $exp_diff_totct_indels, " > 0, which means no chance of improving the scenario(s) ) />\n\n") if ($TEST_LEVEL2>0);

	    return ($exp_diff_totct_indels, undef, undef);
	}

		# There could be a chance of improvement. #

    (print STDOUT "\n", '    ++ The 3rd process (Main 2): Constructing @bestset_mvd_events. ++', "\n\n") if ($TEST_LEVEL2>1);

	my @bestset_mvd_events = ();

	my $eff_chbras_sbjct = $br2eff_ch{$sb_br_id};
	foreach my $chbr (@{$eff_chbras_sbjct}) { # 3rd main foreach-loop #

	(print STDOUT "     ( 'Child' branch ID: $chbr )\n") if ($TEST_LEVEL2>1);

	    my ($ct_nabs_chbr, $ct_nabs_ds_chbr, $set_ds_abs_brs_chbr) = @{$br2ct_nonabsorbed{$chbr}}; 

	    my @event_hosting_branches = (defined $set_ds_abs_brs_chbr) ? @{$set_ds_abs_brs_chbr} : ($chbr) ;
	    foreach my $br (@event_hosting_branches) { # Inner foreach-loop # 
		my @mvd_event = (defined $uphill{$br}) ?
			($br, $char_abs, $char_pres) : ($br, $char_pres, $char_abs);
		push @bestset_mvd_events, \@mvd_event;
	    }					# End of the inner foreach-loop #
	}				# End of the 3rd main foreach-loop #

    if ($TEST_LEVEL2c>2) {
	print STDOUT "\n", '    ++ The 3rd process ENDED ++', "\n\n";
	print STDOUT 'Content of @bestset_mvd_events', "\n\n";
	print STDOUT join ("\t", 'Branch_ID', 'state_ue', 'state_le'), "\n\n";
	foreach my $event (@bestset_mvd_events) { print STDOUT join ("\t", @{$event}), "\n"; } 
	print STDOUT "\n";
    }


    (print STDOUT "\n", '    ++ The 4th process (Main 3): Constructing %node2flipped_state. ++', "\n\n") if ($TEST_LEVEL2>1);

	my %node2flipped_state = ($pivotal_node => 1); # Nodes where the state flips from $char_abs to $char_pres. #

	my @subjects2 = @{$eff_chbras_sbjct}; # Set of subject branches. #

	while (@subjects2>0) { # The 4th main outer while-loop #

	    my $ct_sbjcts2 = @subjects2;
    (print STDOUT "    [ #{subject branches} = $ct_sbjcts2 ]\n\n") if ($TEST_LEVEL2>1);

	    my @new_subjects2 = ();
	    while (my $br = shift @subjects2) { # Middle while-loop #

	(print STDOUT "    ( Branch ID = $br )\n") if ($TEST_LEVEL2>1);

		my $eff_chbras = $br2eff_ch{$br};
		(defined $eff_chbras) or next; # If there is no 'effective' child branches, we cannot flip the state at the tip of the branch.

		my ($ct_nabs_br, $ct_nabs_ds_br, $set_ds_abs_brs_br) = @{$br2ct_nonabsorbed{$br}};
		($ct_nabs_br == 0) and next; # If the deletion itself is absorbed into the flanking event(s), the tip should remain $stat_abs.
		(defined $ct_nabs_ds_br) or next; # This variable is undefined when there is a 'dead-end', where the tip should remain $stat_abs. #
		($ct_nabs_ds_br>0) and next; # If this variable is nonzero, the 'recession' should stop here.

			# The 'absence' should 'recede' further. #

		my $ds_node = (defined $uphill{$br}) ? $node2pa->{$br}: $br; # Node at the 'downstream' of the branch. #
		$node2flipped_state{$ds_node} = 1;
		push @new_subjects2, @{$eff_chbras};

	    }				# End of the middle while-loop #

	    @subjects2 = @new_subjects2;
	}		# End of the 4th main outer while-loop #

    if ($TEST_LEVEL2c>2) {
	print STDOUT "\n", '    ++ The 4th process ENDED ++', "\n\n";
	print STDOUT "{Nodes whose state will be flipped from '$char_abs' to '$char_pres'} = ", join (', ', sort {$a <=> $b} keys %node2flipped_state), ".\n\n";
    }

    (print STDOUT "\n    +++ Finished ALL the 4 processes +++\n\n") if ($TEST_LEVEL1>1);


    (print STDOUT "\n   < Subroutine 'create_bestset_mvd_events' Ended />\n\n") if ($TEST_LEVEL2>0);

		## A BUG fixed on May 17 (Thu), 2012. ##
	return ($exp_diff_totct_indels, \@bestset_mvd_events, \%node2flipped_state);
	##return ($exp_diff_totct_indels, \@bestset_mvd_events, %node2flipped_state);

}	# End of the subroutine 'create_bestset_mvd_events' #


# recede_absence:
#	This core satellite subroutine of 'enum_psm_indel_scns_in_cls_blks' tries to
#	'recede' $char_abs from the relevant end of a specified branch.
#	As a 'side-effect', this subroutine removes the last entry in the list of indel events to be examined.
#
#	REVISED on Nov 13 (Wed), 2013:
#	Now, the subroutine can correctly handle trees containing two-branch nodes.
#
# INPUT: (\%node2blwise_gstat, \%br2indels, \@indels_to_be_examined, $char_abs, $char_pres, \%node2pa, \%node2ch, \%node2ct_exoffs, $ct_seqs), where
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$b] being $char_pres/abs for the $b th block (in the cluster) at the node $node_id;
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls)
#	of the blocks involved in the indel event;
#
#	@{$indels_to_be_examined->[$k][$i]}
#	= ($branch_id, index in the list @{$br2indels{$branch_id}})
#	is for the $i th indel event that result in $k exterior nodes with $char_abs (for 0<$k<$ct_seqs).
#	@{$indels_to_be_examined[0]} should always be empty;
#
#	$char_abs/pres is the character representing the 'absence'/'presence' state;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	$ct_seqs is the number of all exterior nodes;
#
# OUTPUT: ($diff_totct_indels, \%new_node2blwise_gstat, \%new_br2indels, \@new_indels_to_be_examined), where
#	$diff_totct_indels is the change in the total number of indel events resulting from this session of 'absence recession';
#	%new_node2blwise_gstat is %node2blwise_gstat after this session (and follows the same format as the latter);
#	%new_br2indels is %br2indels after this session (and follows the same format as the latter).
#	The indel event that were eliminated as a result of the 'absence recession'
#	is represented as an empty array (to avoid the confusion resulting from the physical removal of the entry);
#	@new_indels_to_be_examined is @indels_to_be_examined after this session (and follows the same format as the latter).
#	
sub recede_absence (\%\%\@$$\%\%\%$) {

	my ($node2blwise_gstat, $br2indels, $indels_to_be_examined, $char_abs, $char_pres, $node2pa, $node2ch, $node2ct_exoffs, $ct_seqs) = @_;

	my $k = $#{$indels_to_be_examined};
	my $list_indels = $indels_to_be_examined->[$k];

    (print STDOUT " < The subroutine 'recede_absence' (Subject: The ", $#{$list_indels}, " th indel event in the $k th list in ", '@{$indels_to_be_examined}', ") >\n\n") if ($TEST_LEVEL1hf>0);

		# MOVED below in revision on Nov 13 (Wed), 2013. #
	## my $examined_indel = pop @{$list_indels};
	## unless (defined $examined_indel) {
	##     print STDOUT "[ERROR 'recede_absence'] Found NO elements in the list of indels resulting in $k exterior nodes with 'absence': $! !\n";
	##     return (undef, undef, undef, undef);
	## }

		# MODIFIED in revision on Nov 13 (Wed), 2013. #
	if (($k < 2) and (@{$list_indels}==0)) {
	##if ($k < 2) {
	     # The 'absence' state cannot recede any further. #
    (print STDOUT "\n < PREMATURELY END 'recede_absence' because ", '$k (=', $k, ') < 2 and @{$indels_to_be_examined->[$k]} is empty />', "\n\n") if ($TEST_LEVEL1hf>0); 
    ##(print STDOUT "\n < PREMATURELY END 'recede_absence' because ", '$k (=', $k, ') < 2. />', "\n\n") if ($TEST_LEVEL1hf>0); 
	    return (0, undef, undef, undef);
	}

		# MOVED from above in revision on Nov 13 (Wed), 2013. #
	my $examined_indel = pop @{$list_indels};
	unless (defined $examined_indel) {
	    print STDOUT "[ERROR 'recede_absence'] Found NO elements in the list of indels resulting in $k exterior nodes with 'absence': $! !\n";
	    return (undef, undef, undef, undef);
	}

	my ($br_id, $indx_indel) = @{$examined_indel};

	my $list_indels_at_sbbr = $br2indels->{$br_id};
	my $sbjct_indel_event = $list_indels_at_sbbr->[$indx_indel];

	unless (@{$sbjct_indel_event}>0) {
		# If the event is "empty", do nothing. #
    (print STDOUT "\n < PREMATURELY END 'recede_absence' because the indel event to be examined (the $indx_indel th event on the branch $br_id) is EMPTY. />\n\n") if ($TEST_LEVEL1hf>0);
	     return (0, undef, undef, undef);
	}

	my ($stat_pa, $stat_node, $indices_sb_blk) = @{$sbjct_indel_event};
	my $pivotal_node = ($stat_node eq $char_abs) ? $br_id : $node2pa->{$br_id}; # The 'end'-node  with the 'absence' state.


   if ($TEST_LEVEL1hf>3) {
	print STDOUT "\n  [ Input Data ]\n\n";

	print STDOUT '  ( Content of %{$node2blwise_gstat} ):', "\n\n";
	print STDOUT '  ', join ("\t", 'Node_ID', 'Block-wise,gap,states'), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat}) {
	    my $cnct_gstat = join (',', @{$node2blwise_gstat->{$node}});
	    print STDOUT '  ', join ("\t", $node, $cnct_gstat), "\n";
	}
	print STDOUT "\n\n", '  ( Content of %{$br2indels} ):', "\n\n";
	print STDOUT '  ', join ("\t", 'Branch_ID', 'Stat_ue/stat_le/blocks,involved', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2indels}) {
	    my @displayed = ();
	    foreach my $indel_info (@{$br2indels->{$br}}) {
		((defined $indel_info) and (@{$indel_info}>0)) or next; # Skip a null event. #
		my ($stat_ue, $stat_le, $blks_involved) = @{$indel_info};
		my $cnct_blks_involved = (defined $blks_involved) ? join (',', @{$blks_involved}) : 'none';
		push @displayed, join ('/', $stat_ue, $stat_le, $cnct_blks_involved);
	    }
	    print STDOUT '  ', join ("\t", $br, @displayed), "\n";
	}

	print STDOUT "\n\n", '  ( Subject indel event ):', "\n\n";
	print STDOUT '  Branch_ID = ', $br_id, " ,\n";
	print STDOUT '  Indel index = ', $indx_indel, " .\n\n";
	print STDOUT '  State_ue = ', $stat_pa, " ,\n";
	print STDOUT '  State_le = ', $stat_node, " ,\n";
	print STDOUT '  Blocks involved: ', join (', ', @{$indices_sb_blk}), " .\n\n";
	print STDOUT "  Pivotal node ID = $pivotal_node .\n\n\n";

   } # End of "if ($TEST_LEVEL1hf>3) {...}".


	# @{$mvd_events->[$e]} = ($new_br_id, $new_stat_ue, $new_stat_le)
	# is for the $e th event resulting from the absence recession, with
	# $new_br_id specifying the branch where the event occurs,
	# $new_stat_ue/le representing the new state at the upper-end/lower-end of the branch.
	# @{$mvd_events} is empty if the 'absence' state cannot recede any further.
	#
	my $mvd_events = list_mvd_events ($br_id, $stat_node, $char_abs, $char_pres, %{$node2pa}, %{$node2ch});
	unless (defined $mvd_events) {
	    print STDOUT "[ERROR 'recede_absence'] The subroutine 'list_mvd_events' FAILED on the $indx_indel th event along the branch $br_id: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $ct_mvd_events = @{$mvd_events};
    (print STDOUT "  --- Subroutine 'list_mvd_events' ended (with #{moved events} = $ct_mvd_events) ...\n\n") if ($TEST_LEVEL1hf>1);

	unless (@{$mvd_events}>0) {
	     # The 'absence' state cannot recede any further. (II) #
    (print STDOUT "\n < Subroutine 'recede_absence' Ended (because the 'absence' state cannot recede at all) />\n\n") if ($TEST_LEVEL1hf>0);
	    return (0, undef, undef, undef);
	}

   if ($TEST_LEVEL1hf>3) {

	print STDOUT "  [ $ct_mvd_events 'New' events will result if the subject event is 'recede'ed. ]\n\n";
	print STDOUT '  ', join ("\t", 'No.', 'New_branch_ID', 'new_state_ue', 'new_state_le'), "\n\n";
	for (my $ne = 0; $ne<$ct_mvd_events; $ne++) { print STDOUT '  ', join ("\t", $ne, @{$mvd_events->[$ne]}), "\n"; }
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL1hf>3) {...}"



		# Look for flanking blocks that can absorb new indel events. #

			# Search for flanking blocks in strict sense. #
			# (MODIFIED to make the algorithm closer to perfection (on Oct 24 (Wed), 2012)) #
	my ($if_success, $lflank_blk1, $rflank_blk1) = search_flanking_blocks2 ($node2pa->{$br_id}, $br_id, @{$sbjct_indel_event}, %{$node2blwise_gstat}, $char_abs);
	##my ($if_success, $lflank_blk, $rflank_blk) = search_flanking_blocks ($node2pa->{$br_id}, $br_id, @{$sbjct_indel_event}, %{$node2blwise_gstat}, $char_abs);
	unless ($if_success) {
	    print STDOUT "[ERROR 'recede_absence'] The subroutine 'search_flanking_blocks2' FAILED on the $indx_indel th event along the branch $br_id: $! !\n";
	    return (undef, undef, undef, undef);
	}

		# MODIFIED in revision on Nov 13 (Wed), 2013. #
		# (This modification enables to recede the 'absence' state from a two-branch node, even if there is no flanking block.)

	unless ((defined $lflank_blk1) or (defined $rflank_blk1) or ($ct_mvd_events<2)) {
	##unless ((defined $lflank_blk1) or (defined $rflank_blk1)) {
		# There is no flanking block that can absorbe new indel events
		# resulting from the 'absence recession'. #
    (print STDOUT "\n < Subroutine 'recede_absence' Ended (because there is no flanking block that can absorbe any of the $ct_mvd_events new indel events) />\n\n") if ($TEST_LEVEL1hf>0);
	     return (1, undef, undef, undef);
	}

   if ($TEST_LEVEL1hf>3) {
	print STDOUT "   [ Flanking blocks ]\n\n";
	print STDOUT '   left-flanking block = ', ((defined $lflank_blk1) ? $lflank_blk1 : 'undef'), " ,\n";
	print STDOUT '   right-flanking_block = ', ((defined $rflank_blk1) ? $rflank_blk1 : 'undef'), " .\n\n\n";
   }

    (print STDOUT "  || Initialize a new parsimonious indel scenario. ||\n\n") if ($TEST_LEVEL1hf>1);

		# New block-wise states at the pivotal node. #
	my @new_pv_gstates = @{$node2blwise_gstat->{$pivotal_node}};
	my $ct_blks = @new_pv_gstates;
	foreach my $indx (@{$indices_sb_blk}) { $new_pv_gstates[$indx] = $char_pres; }

		# Make a new (modified) map of indel events to branches (and a new list of events to be examined). #
	my %new_br2indels = %{$br2indels};
	my @new_list_indels_at_sbbr = @{$list_indels_at_sbbr};
	$new_list_indels_at_sbbr[$indx_indel] = [];
	$new_br2indels{$br_id} = \@new_list_indels_at_sbbr;

	my @new_indels_to_be_examined = ();
	foreach my $list_indels (@{$indels_to_be_examined}) { # Preliminary foreach-loop #
	    if (@{$list_indels}>0) {
		my @new_list_indels = @{$list_indels};
		push @new_indels_to_be_examined, \@new_list_indels;
	    } else {
		push @new_indels_to_be_examined, $list_indels;
	    }
	}		# End of the preliminary foreach-loop #

    (print STDOUT "  || NOW, the 1st attempt (to move $ct_mvd_events indel events across the pivotal node $pivotal_node). ||\n\n") if ($TEST_LEVEL1hf>1);

	my @not_absorbed = (); # moved indel events that are not absorbed into neighboring events. #

	for (my $me=0; $me<$ct_mvd_events; $me++) { # 1st main outer for-loop #
	    my $mvd_event = $mvd_events->[$me];
	    my ($new_br_id, $new_stat_ue, $new_stat_le) = @{$mvd_event};

    (print STDOUT "  [ The $me th event (along the branch ID = $new_br_id) ]\n") if ($TEST_LEVEL1hf>1);

	    my $new_ct_exoffs = $node2ct_exoffs->{$new_br_id};
	    my $new_k = ($new_stat_le eq $char_abs) ? $new_ct_exoffs : $ct_seqs - $new_ct_exoffs;
	    my $if_absorbed = incorporate_new_event (@{$mvd_event}, @{$indices_sb_blk}, %new_br2indels, %{$node2blwise_gstat}, $char_abs, %{$node2pa});

	    unless (defined $if_absorbed) {
		print STDOUT "[ERROR 'recede_absence'] The subroutine 'incorporate_new_event' FAILED on the new deletion event along the branch $new_br_id: $! !\n";
		return (undef, undef, undef, undef);
	    }
    (print STDOUT "   --- Subroutine 'incorporate_new_event' ended ...\n\n") if ($TEST_LEVEL1hf>1);

	    unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		my $ct_indels_in_br = scalar @{$new_br2indels{$new_br_id}};
		my @new_list_indels = @{$new_indels_to_be_examined[$new_k]};

		if ( ($new_stat_le eq $char_pres) or ((defined $node2ch->{$new_br_id}) and (@{$node2ch->{$new_br_id}}>0)) ) { # Only when the 'absence' state in the new indel event is 'recedable', it is incorporated into the list of indels to be examined. ## Condition ADDED in revision on Nov 13 (Wed), 2013. ##
		    push @new_list_indels, [$new_br_id, $ct_indels_in_br - 1];
		}
		$new_indels_to_be_examined[$new_k] = \@new_list_indels;
		push @not_absorbed, $mvd_event;
	    }
	}			# End of the 1st main outer for-loop #

	my $diff_totct_indels = scalar (@not_absorbed) - 1;

    (print STDOUT "   ... The 1st attempt ENDED (with ", '$diff_totct_indels = ', $diff_totct_indels, ") ...\n\n") if ($TEST_LEVEL1hf>1);

	##if ($diff_totct_indels>1) {
	##	# The new scenario will be dismissed anyway. #
	##    return ($diff_totct_indels, undef, undef, undef);
	##} els
	if ($diff_totct_indels<1) {
	    my %new_node2blwise_gstat = %{$node2blwise_gstat};
	    $new_node2blwise_gstat{$pivotal_node} = \@new_pv_gstates;
    (print STDOUT "\n < Subroutine 'recede_absence' NORMALLY Ended (I) (with ", '$diff_totct_indels = ', $diff_totct_indels, " < 1). />\n\n") if ($TEST_LEVEL1hf>0);
	    return ($diff_totct_indels, \%new_node2blwise_gstat, \%new_br2indels, \@new_indels_to_be_examined);

			# MODIFIED to make the algorithm closer to perfection. I (on Oct 24 (Wed), 2012) #
	} 
	#} elsif ( !((defined $lflank_blk) and (defined $rflank_blk)) ) {

			# Search for flanking blocks in strict sense. #
			# (ADDED to make the algorithm closer to perfection (on Oct 24 (Wed), 2012)) #

	my ($if_success2, $lflank_blk2, $rflank_blk2) = search_flanking_blocks ($node2pa->{$br_id}, $br_id, @{$sbjct_indel_event}, %{$node2blwise_gstat}, $char_abs);
	unless ($if_success2) {
	    print STDOUT "[ERROR 'recede_absence'] The subroutine 'search_flanking_blocks' FAILED on the $indx_indel th event along the branch $br_id: $! !\n";
	    return (undef, undef, undef, undef);
	}

		# MODIFIED to prevent the subroutine from becoming too slow (on Mar 20 (Thu), 2014). #

	my @flk_blk_pairs0  = ();
	((defined $lflank_blk1) and (defined $rflank_blk1)) and (push @flk_blk_pairs0, [$lflank_blk1, $rflank_blk1]);
	((defined $lflank_blk2) and ($lflank_blk1 != $lflank_blk2)) and (push @flk_blk_pairs0, [$lflank_blk2, $lflank_blk1]);
	((defined $rflank_blk2) and ($rflank_blk1 != $rflank_blk2)) and (push @flk_blk_pairs0, [$rflank_blk1, $rflank_blk2]);

	unless (@flk_blk_pairs0>0) {

		# [ There is  one flanking block that can absorb the deletion events. ] -- Obsolete
		# There is NO potential flanking block pair that can absorb the deletion events.
		# Therefore, there is no chance of improvement.

    (print STDOUT "\n < Subroutine 'recede_absence' Ended (with ", '$diff_totct_indels = ', $diff_totct_indels, " < 1, and LACKING either of the flanking blocks). />\n\n") if ($TEST_LEVEL1hf>0);

	    return ($diff_totct_indels, undef, undef, undef);
	}


	my $set_dwn_brs = fetch_all_offsprings_nr ($br_id, %{$node2ch});
	my %if_uphill; # = ($br_id => 1, ...) for the branches along which the 'deletion' would be uphill. (ADDED on Mar 21 (Fri), 2014.) #
	if ($stat_node eq $char_abs) { # $br_id is the 'lower-end' of the 'deletion'. #
	    shift @{$set_dwn_brs};

	} else { # $br_id is the 'upper-end' of the 'deletion'. #
	    my $pa_br = $node2pa->{$br_id};
	    ##(defined $node2pa->{$pa_br}) and (push @{$set_dwn_brs}, $pa_br); ## DISABLED on Mar 21 (Fri), 2014. ##
	    my @set_all_brs = keys %{$node2pa};
	    (undef, undef, $set_dwn_brs) = diff_nmarray (@{$set_dwn_brs}, @set_all_brs);

		# ADDED on Mar 21 (Fri), 2014. #
	    my $uphill = $br_id;
	    while (defined $uphill) { # Preliminary while-loop (along the uphill path). #
		$if_uphill{$uphill} = 1;
		$uphill = $node2pa->{$uphill};
	    }				# End of the preliminary while-loop (along the uphill path). #
	}

	my $ct_events = 0;
	foreach my $br (keys %{$br2indels}) { $ct_events += @{$br2indels->{$br}}; } # A BUG fixed on 2019/05/14.
	##foreach my $br (keys %{$br2indels}) { $ct_events += @{$br2indels}; }

	my @flk_blk_pairs = ();
	if ($ct_events < 2 * scalar (@{$set_dwn_brs}) ) {
	    my %if_dwn; # = ($br_id => 1, ...) for all branches 'downstream' of $br_id. #
	    foreach my $br (@{$set_dwn_brs}) { $if_dwn{$br} = 1; }

	    foreach my $pair (@flk_blk_pairs0) { # Preliminary foreach-loop. #
		my ($lflank_blk, $rflank_blk) = @{$pair};
		my $lflk_events = enum_events_involve_blk ($lflank_blk, %{$br2indels});
	 	my $rflk_events = enum_events_involve_blk ($rflank_blk, %{$br2indels});

    if ($TEST_LEVEL1hf>3) {
	print STDOUT "  [ For the flanking pair: (left, right) = ($lflank_blk, $rflank_blk) ]\n\n";
    }

    if ($TEST_LEVEL1hf>=5) {
	print STDOUT "   {Left-flanking events (branch_ID/event_ID)}:   ";
	foreach my $event (@{$lflk_events}) { print STDOUT join ('/', @{$event}), ', '; }
	print STDOUT " ;\n\n";
	print STDOUT "   {Right-flanking events (branch_ID/event_ID)}:  ";
	foreach my $event (@{$rflk_events}) { print STDOUT join ('/', @{$event}), ', '; }
	print STDOUT " .\n\n";
    }


			## MODIFIED in revision on Mar 21 (Fri), 2014. ##

		my %if_relv; # = ("$br_id/$event_id" => 1, ...) for the relevant event. (This avoids the double-counting of relevant events.)
		foreach my $event (@{$lflk_events}, @{$rflk_events}) { # Preliminary inner foreach-loop (over flanking events). #
		    my ($br_id, $event_id) = @{$event};
		    (defined $if_dwn{$br_id}) or next;
		    my ($stat_ue, $stat_le, $involved_blks) = @{$br2indels->{$br_id}[$event_id]};
		    my $stat_ds_end = (defined $if_uphill{$br_id}) ? $stat_ue : $stat_le; # State at the 'downstream' end of the branch. 
		    ($stat_ds_end eq $char_abs) or next; # Skip unless the event is 'deletion'.
		    $if_relv{"$br_id/$event_id"} = 1;
		}			# End of the preliminary inner foreach-loop (over flanking events). #
		my @rlv_events = sort keys %if_relv;

		##my @rlv_events = ();
		##foreach my $event (@{$lflk_events}, @{$rflk_events}) { (defined $if_dwn{$event->[0]}) and (push @rlv_events, $event); }
		my $ct_rlv_events = @rlv_events;

    if ($TEST_LEVEL1hf>3) {
	print STDOUT "   {There are $ct_rlv_events seemingly relevant events (branch_ID/event_ID)}:  "; 

			## MODIFIED in revision on Mar 21 (Fri), 2014. ##
	print STDOUT join (', ', @rlv_events), " .\n\n";
	foreach my $event (@rlv_events) { print STDOUT join ('/', @{$event}), ', '; }
	##foreach my $event (@rlv_events) { print STDOUT join ('/', @{$event}), ', '; }
	##print STDOUT " .\n\n";
    }
			# MODIFIED to accommodate more possible parsimonious scenarios. (on Mar 20 (Thu), 2014). #
	 	($ct_rlv_events >1) and (push @flk_blk_pairs, $pair);
	 	##($ct_rlv_events >2) and (push @flk_blk_pairs, $pair);

	    }			# End of the preliminary foreach-loop. #
	} else {
	    @flk_blk_pairs = @flk_blk_pairs0;
	}

	##my @flk_blk_pairs  = ();
	##((defined $lflank_blk1) and (defined $rflank_blk1)) and (push @flk_blk_pairs, [$lflank_blk1, $rflank_blk1]);
	##((defined $lflank_blk2) and ($lflank_blk1 != $lflank_blk2)) and (push @flk_blk_pairs, [$lflank_blk2, $lflank_blk1]);
	##((defined $rflank_blk2) and ($rflank_blk1 != $rflank_blk2)) and (push @flk_blk_pairs, [$rflank_blk1, $rflank_blk2]);

	unless (@flk_blk_pairs>0) {
			# END of "MODIFIED to make the algorithm closer to perfection. I (on Oct 24 (Wed), 2012)" #

		# [ There is  one flanking block that can absorb the deletion events. ] -- Obsolete
		# There is NO potential flanking block pair that can absorb the deletion events.
		# Therefore, there is no chance of improvement.

		# MODIFIED in revision on Mar 20 (Thu), 2014. #
    (print STDOUT "\n < Subroutine 'recede_absence' Ended (with ", '$diff_totct_indels = ', $diff_totct_indels, " < 1, because the flanking blocks are NOT EXPECTED to ABSORB the 'absence recession' in any complex ways). />\n\n") if ($TEST_LEVEL1hf>0);
    ##(print STDOUT "\n < Subroutine 'recede_absence' Ended (with ", '$diff_totct_indels = ', $diff_totct_indels, " < 1, and LACKING either of the flanking blocks). />\n\n") if ($TEST_LEVEL1hf>0);

	    return ($diff_totct_indels, undef, undef, undef);
	}

		# When $diff_totct_indels > 0 and the event has two flanking blocks,
		# try to recede the 'absence' further. #

	my $ct_flk_blk_prs = @flk_blk_pairs;
    (print STDOUT "  || NOW, see if we can make a 2nd attempt (to move the indel event along the branch $br_id), with $ct_flk_blk_prs potential flanking pairs. ||\n\n") if ($TEST_LEVEL1hf>1);


		# MODIFIED to make the algorithm closer to perfection. II (on Oct 24 (Wed), 2012) #
	my ($exp_diff_totct_indels, $bestset_mvd_events, $node2flipped_state);

	foreach my $flk_blk_pair (@flk_blk_pairs) { # 2nd main outer foreach-loop (over potential pairs of flanking blocks) #

	    my ($lflank_blk, $rflank_blk) = @{$flk_blk_pair};

    (print STDOUT "    [ Flanking pair = ($lflank_blk, $rflank_blk) ]\n\n") if ($TEST_LEVEL1hf>1);

	    # $exp_diff_totct_indels is the expected difference in the total number of indels resulting from the best set of moved events;
	    # @{$bestset_mvd_events->[$e]} = ($new_br_id, $new_stat_ue, $new_stat_le)	
	    # represents the $e th moved event in the 'best' set, with
	    # $new_br_id being the ID of the branch where the new indel occurs,
	    # $new_state_ue/le being the new state at the upper/lower-end of the branch;
	    # %{$node2flipped_state} = ($node_id => 1, ...) for the nodes (including the pivotal node) whose state flipped from $char_abs to $char_pres as a result of moving the events.
	    #	
	    ##my ($exp_diff_totct_indels, $bestset_mvd_events, $node2flipped_state)
	    my ($tmp_exp_diff_totct_indels, $tmp_bestset_mvd_events, $tmp_node2flipped_state)
		= create_bestset_mvd_events ($br_id, @{$sbjct_indel_event}, $lflank_blk, $rflank_blk, %{$node2blwise_gstat}, %{$node2pa}, %{$node2ch}, $char_pres, $char_abs);
	    unless (defined $tmp_exp_diff_totct_indels) {
		print STDOUT "[ERROR 'recede_absence'] The subroutine 'create_bestset_mvd_events' FAILED on the $indx_indel th event along the branch $br_id, with flanking blocks ($lflank_blk, $rflank_blk): $! !\n";
		return (undef, undef, undef, undef);
	    }
    (print STDOUT "  --- Subroutine 'create_bestset_mvd_events' ended (with ", '$tmp_exp_diff_totct_indels = ', $tmp_exp_diff_totct_indels, "). ...\n\n") if ($TEST_LEVEL1hf>1);
    ##(print STDOUT "  --- Subroutine 'create_bestset_mvd_events' ended (with ", '$exp_diff_totct_indels = ', $exp_diff_totct_indels, "). ...\n\n") if ($TEST_LEVEL1hf>1);

	   unless ((defined $exp_diff_totct_indels) and ($exp_diff_totct_indels <= $tmp_exp_diff_totct_indels)) {
		($exp_diff_totct_indels, $bestset_mvd_events, $node2flipped_state)
		   = ($tmp_exp_diff_totct_indels, $tmp_bestset_mvd_events, $tmp_node2flipped_state) ;
	   }

	}			# End of the 2nd main outer foreach-loop (over potential pairs of flanking blocks) #

		# END of "MODIFIED to make the algorithm closer to perfection. II (on Oct 24 (Wed), 2012)" #


	if ($exp_diff_totct_indels>0) {
		# There should be no chance of improvement by the further recession. #

    (print STDOUT "\n < Subroutine 'recede_absence' Ended (with ", '$exp_diff_totct_indels = ', $exp_diff_totct_indels, " > 0, which means there should be no chance of improving the scenario(s)). />\n\n") if ($TEST_LEVEL1hf>0);

	    return ($exp_diff_totct_indels, undef, undef, undef);
	}

		# Re-create a new (modified) map of indel events to branches
		# (and a new list of events to be examined). #

    (print STDOUT "  || Initialize a 2nd new parsimonious indel scenario. ||\n\n") if ($TEST_LEVEL1hf>1);

	%new_br2indels = %{$br2indels};
	@new_list_indels_at_sbbr = @{$list_indels_at_sbbr};
	$new_list_indels_at_sbbr[$indx_indel] = [];
	$new_br2indels{$br_id} = \@new_list_indels_at_sbbr;

	@new_indels_to_be_examined = ();
	foreach my $list_indels (@{$indels_to_be_examined}) { # 2nd preliminary foreach-loop #
	    if (@{$list_indels}>0) {
		my @new_list_indels = @{$list_indels};
		push @new_indels_to_be_examined, \@new_list_indels;
	    } else {
		push @new_indels_to_be_examined, $list_indels;
	    }
	}				# End of the 2nd preliminary foreach-loop #

	my $ct_best_mvd_events = @{$bestset_mvd_events};

    (print STDOUT "  || NOW, the 2nd attempt (to move $ct_best_mvd_events indel events across the pivotal node $pivotal_node). ||\n\n") if ($TEST_LEVEL1hf>1);

	my @not_absorbed2 = (); # moved indel events that are not absorbed into neighboring events. #

	for (my $me=0; $me<$ct_best_mvd_events; $me++) { # 2nd main outer for-loop #
	    my $mvd_event = $bestset_mvd_events->[$me];
	    my ($new_br_id, $new_stat_ue, $new_stat_le) = @{$mvd_event};

    (print STDOUT "  [ The $me th event (along the branch ID = $new_br_id) ]\n") if ($TEST_LEVEL1hf>1);

	    my $new_ct_exoffs = $node2ct_exoffs->{$new_br_id};
	    my $new_k = ($new_stat_le eq $char_abs) ? $new_ct_exoffs : $ct_seqs - $new_ct_exoffs;
	    my $if_absorbed = incorporate_new_event (@{$mvd_event}, @{$indices_sb_blk}, %new_br2indels, %{$node2blwise_gstat}, $char_abs, %{$node2pa});

	    unless (defined $if_absorbed) {
		print STDOUT "[ERROR 'recede_absence'] The subroutine 'incorporate_new_event' FAILED on the new deletion event along the branch $new_br_id: $! !\n";
		return (undef, undef, undef, undef);
	    }
    (print STDOUT "   --- Subroutine 'incorporate_new_event' ended ...\n\n") if ($TEST_LEVEL1hf>1);

	    unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		my $ct_indels_in_br = scalar @{$new_br2indels{$new_br_id}};
		my @new_list_indels = @{$new_indels_to_be_examined[$new_k]};
		push @new_list_indels, [$new_br_id, $ct_indels_in_br - 1];
		$new_indels_to_be_examined[$new_k] = \@new_list_indels;
		push @not_absorbed2, $mvd_event;
	    }
	}			# End of the 2nd main outer for-loop #

	$diff_totct_indels = scalar (@not_absorbed2) - 1;

    (print STDOUT "   ... The 2nd attempt ENDED (with", '$diff_totct_indels = ', $diff_totct_indels, ") ...\n\n") if ($TEST_LEVEL1hf>1);

	if ($diff_totct_indels>0) {
	    ## The new scenario will be dismissed anyway. ##

    (print STDOUT "\n < Subroutine 'recede_absence' Ended (with ", '$diff_totct_indels = ', $diff_totct_indels, " > 0, which means that the scenario(s) couldn't improved). />\n\n") if ($TEST_LEVEL1hf>0);

	    return ($diff_totct_indels, undef, undef, undef);
	}

		# Construct a new map of block-wise gap states onto nodes by flipping states (from $char_abs to $char_pres) where necessary. #

    (print STDOUT '   +++ Construct %new_node2blwise_gstat ', "by flipping states from $char_abs to $char_pres. +++ \n\n") if ($TEST_LEVEL1hf>1);

	my %new_node2blwise_gstat = %{$node2blwise_gstat};

	foreach my $node (sort {$a <=> $b} keys %{$node2flipped_state}) { # Auxliary foreach-loop #
	# %{$node2flipped_state} = ($node_id => 1, ...) for the nodes (including the pivotal node) whose state flipped from $char_abs to $char_pres as a result of moving the events.

	    my @new_blwise_gstats = @{$new_node2blwise_gstat{$node}};
	    foreach my $indx (@{$indices_sb_blk}) { $new_blwise_gstats[$indx] = $char_pres; }
	    $new_node2blwise_gstat{$node} = \@new_blwise_gstats;
	}			# End of the auxiliary foreach-loop #


    (print STDOUT "\n < Subroutine 'recede_absence' NORMALLY Ended (II) (with $diff_totct_indels). />\n\n") if ($TEST_LEVEL1hf>0);

	return ($diff_totct_indels, \%new_node2blwise_gstat, \%new_br2indels, \@new_indels_to_be_examined);

}	# End of the subroutine 'recede_absence' #


# cleanup_br2indels:
#	This subroutine "cleans up" the hash %br2indels = ($br_id => \@list_indels, ...)
#	by removing empty arrays lurking in @indels.
#
# INPUT: \%br2indels, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls)
#	of the blocks involved in the indel event.
#	Sometimes @{$list_indels[$id]} can be empty (i.e. ()).
#
# OUTPUT: $fl_succ, where
#	$fl_succ = 1 if the subroutine successfully ended,
#	 	 = 0 if it failed.
#
sub cleanup_br2indels (\%) {

	my ($br2indels) = @_;

    (print STDOUT "   < Subroutine 'cleanup_br2indels' >\n\n") if ($TEST_LEVEL2>0);

	foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # Main outer foreach-loop #

	(print STDOUT "   [ Branch ID = $br ]\n") if ($TEST_LEVEL2>0);

	    my $list_indels = $br2indels->{$br};
	    my $if_emptyarray = 0;
	    foreach my $indel (@{$list_indels}) { # 1st inner foreach-loop #
		(@{$indel}>0) and next;
		$if_emptyarray = 1;
		last;
	    }			# End of the 1st inner foreach-loop #

	    $if_emptyarray  or next; # Do nothing if there are no empty arrays.

	    my $ct_indels = @{$list_indels};
	    for (my $k=0; $k<$ct_indels; $k++) { # 2nd inner foreach-loop #
		my $indel = shift @{$list_indels}; # Remove an entry from the front end.

			# Return the entry to the back end, if it is defined and not empty.
		((defined $indel) and (@{$indel}>0)) and (push @{$list_indels}, $indel);

	    }			# End of the 2nd inner foreach-loop #

	    unless (@{$list_indels}>0) { # If there is no event left, delete the hash entry. #
		delete $br2indels->{$br};
	    }
	    
	}			# End of the main outer foreach-loop #

    (print STDOUT "\n   < Subroutine 'cleanup_br2indels' />\n\n") if ($TEST_LEVEL2>0);

	return 1;

}	# End of the subroutine 'cleanup_br2indels' #


# examine_flanking_events:
#	This satellite subroutine of 'incorporate_null_block'
#	examines the both flanking blocks to see if they experience indel events
#	along a specified branch.
#
# INPUT: ($br, \%br2lflank_event, \%br2rflank_event, \%br2compensating, \$ct_nonadd_indels), where
#	$br is the ID of the branch in question;
#	%br2l/rflank_event = ($br_id => [$stat_ue_l/rfl, $stat_le_l/rfl], ...)
#	is a map of left/right-flanking indel events onto the branches;
#	%br2compensating = ($br_id => [$stat_le_lfl, $stat_le_rfl], ...) stores
#	branches where both flanking blocks experience indel events that are opposite to each other;
#	$ct_nonadd_indels is the number of indels with $curr_minct_add_indels = 0.
#
# OUTPUT: ($if_succ, $stat_ue, $stat_le, $sgn), where
#	$if_succ = 1 if the subroutine ended successfully,
#		= 0 if it failed;
#	$stat_ue/le is the character representing the state at the upper-/lower-end of the branch;
#	$sgn = -1, +1, or 0 if only left-flanking, right-flanking, or both blocks has/have such an indel event.
#
sub examine_flanking_events ($\%\%\%\$) {

	my ($br, $br2lflank_event, $br2rflank_event, $br2compensating, $ct_nonadd_indels) = @_;

	my $lfl_event = $br2lflank_event->{$br};
	my $rfl_event = $br2rflank_event->{$br};

    (print STDOUT "     < Subroutine 'examine_flanking_events' (", '$ct_nonadd_indels = ', ${$ct_nonadd_indels}, ") >\n\n") if ($TEST_LEVEL3>0);

	my ($stat_pa, $stat_node, $sgn);
	if (defined $lfl_event) {
	    if (defined $rfl_event) {
	        my ($stat_pa_lfl, $stat_node_lfl) = @{$lfl_event};
	        my ($stat_pa_rfl, $stat_node_rfl) = @{$rfl_event};

	        if ($stat_node_lfl eq $stat_node_rfl) {
		    ($stat_pa, $stat_node) = ($stat_pa_lfl, $stat_node_lfl);
		    $sgn = 0;
	        } else {
		    # The two events compensate for each other. #
		    $br2compensating->{$br} = [$stat_node_lfl, $stat_node_rfl];
		    ${$ct_nonadd_indels}++;
	        }
	    } else {
		($stat_pa, $stat_node) = @{$lfl_event};
		$sgn = -1;
	    }
	} elsif (defined $rfl_event) {
	    ($stat_pa, $stat_node) = @{$rfl_event};
	    $sgn = 1;
	}

    (print STDOUT "\n     < Subroutine 'examine_flanking_events' ended />\n\n") if ($TEST_LEVEL3>0);

	return (1, $stat_pa, $stat_node, $sgn);

}	# End of the subroutine 'examine_flanking_events' #


# recede_spec_deletion:
#	This satellite subroutine of 'incorporate_null_block' recedes
#	a specified 'deletion' event to the neighboring events,
#	by flipping the $char_abs state at one of the end-nodes.
#
# INPUT: ($br_del, $bl, $node_abs, \%node2ch, \%node2pa, $char_abs, $char_pres, \%node2blwise_gstat0, \%br2indels0), where
#	$br_del is the branch along which the subject deletion occurred;
#	$bl is the index of the block in which the subject deletion occurred;
#	$node_abs is the end-node of $br_del where the state is $char_abs;
#
#	%node2ch = ($node_id => \@child_nodes, ...) for ALL interior nodes;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	$char_abs/pres is the character representing the 'absence'/'presence' state;
#
#	%node2blwise_gstat0 = ($node_id => \@blwise_gstat0, ...) represents the initial gap states for all nodes,
#	with $blwise_gstat0[$b] = $char_pres/abs being the state in the $b th block in the cluster;
#
#	%br2indels0 = ($branch_id => \@list_indels0, ...) for the input set of indels, with
#	@{$list_indels0[$id]} = ($stat_ue, $stat_le, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices in @{$blocks_in_cls};
#	
#	
#
# OUTPUT: (\%node2blwise_gstat, \%br2indels), which define the set of indel events after the recession, where
#	%node2blwise_gstat and %br2indels follow the same format as %node2blwise_gstat0 and %br2indels0,
#	respectively.
#
#	The subroutine returns undef if it failed.
#
sub recede_spec_deletion ($$$\%\%$$\%\%) {

	my ($br_del, $bl, $node_abs, $node2ch, $node2pa, $char_abs, $char_pres, $node2blwise_gstat0, $br2indels0) = @_;

    (print STDOUT ' ' x 5, "< Subroutine 'recede_spec_deletion' (Subject branch: $br_del; End-node with 'absence': $node_abs; Subject block: $bl) >\n\n") if ($TEST_LEVEL3d>0);

    if ($TEST_LEVEL3d>1) {
	print STDOUT ' ' x 5, "[ Input data ]\n\n";
	print STDOUT ' ' x 5, 'Content of %{$node2blwise_gstat0}:', "\n\n";
	print STDOUT ' ' x 5, join ("\t", 'Node_ID', "'block-wise,gap,states,(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat0}) {
	    my $cnct_gstats = join ('', @{$node2blwise_gstat0->{$node}});
	    print STDOUT ' ' x 5, join ("\t", $node, "'$cnct_gstats'"), "\n";
	}
	print STDOUT "\n\n", ' ' x 5, 'Content of %{$br2indels0}:', "\n\n";
	print STDOUT ' ' x 5, join ("\t", 'Branch_ID', 'stat_ue/stat_le/indices,of,involved,blocks', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) {
	    my $list_indels = $br2indels0->{$br};
	    my @out_indels = ();
	    foreach my $indel (@{$list_indels}) {
		my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
		my $cnct_indices_blk = join (',', @{$indices_blk});
		push @out_indels, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
	    }
	    print STDOUT ' ' x 5, join ("\t", $br, @out_indels), "\n";
	}
	print STDOUT "\n\n";
    }

	my @cmpl_brs = ();
	if ($br_del == $node_abs) { # If 'absence' is at the lower-end. #
	    my $children = $node2ch->{$node_abs};
	    unless (defined $children) { 
		print STDOUT "[ERROR 'recede_spec_deletion'] CanNOT FIND the child nodes of the lower-end node '$node_abs' of the branch '$br_del', whose state is '$char_abs': $! !\n\n";
		return (undef, undef);
	    }
	    @cmpl_brs = @{$children};
	} else { # If 'absence' is at the upper-end. #
	    (defined $node2pa->{$node_abs}) and (push @cmpl_brs, $node_abs); # $node_abs is also an lower-end of the parent branch.
	    my $siblings = $node2ch->{$node_abs};
	    foreach my $sbl (@{$siblings}) { ($sbl == $br_del) or (push @cmpl_brs, $sbl); }
	}

	my ($stat0_ue, $stat0_le) = ($br_del == $node_abs) ?
		($char_pres, $char_abs) : ($char_abs, $char_pres);

	my @new_indels_brdel = ();
	my $sbjct_indices_blk;
	foreach my $indel (@{$br2indels0->{$br_del}}) { # 1st foreach-loop #
	    my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
	    unless (($stat_ue eq $stat0_ue) and ($stat_le eq $stat0_le)) {
			# Skip if the patterns of indels don't match. #
		push @new_indels_brdel, $indel;
		next;
	    }
	    my ($indx_lm, $indx_rm) = @{$indices_blk}[0, $#{$indices_blk}];
	    if (($bl < $indx_lm) or ($indx_rm < $bl)) {
			# Skip if the subject block is outside the set. #
		push @new_indels_brdel, $indel;
		next;
	    }
	    my $if_belong=0;
	    foreach my $indx (@{$indices_blk}) { # Inner foreach-loop #
		if ($indx == $bl) {
		    $if_belong=1;
		    last;
		}
	    }			# End of the inner foreach-lop #
	    if ($if_belong) {
		$sbjct_indices_blk = $indices_blk;
	    } else {
		push @new_indels_brdel, $indel;
	    }
	}			# End of the 1st foreach-loop #

	unless (defined $sbjct_indices_blk) {
	    print STDOUT "[ERROR 'recede_spec_deletion'] CanNOT FIND an indel event  with the specified pattern ($stat0_ue -> $stat0_le) along the branch '$br_del': $! !\n\n";
	    return (undef, undef);
	}

		## Create a new map of indels onto branches. ##

	my %br2indels = %{$br2indels0};

		# A BUG fixed on May 21 (Mon), 2012. #
	if (@new_indels_brdel>0) {
	    $br2indels{$br_del} = \@new_indels_brdel;
	} else {
	    delete $br2indels{$br_del};
	}
	##$br2indels{$br_del} = \@new_indels_brdel;

	foreach my $cmpl_br (@cmpl_brs) { # 2nd foreach-loop #

		# A BUG fixed on May 21 (Mon), 2012.
		# (NOTE that, after the recession, $node_abs has the state $char_pres.) #
	    my ($stat_ue, $stat_le) = ($cmpl_br ==  $node_abs) ?
	    	($char_abs, $char_pres) : ($char_pres, $char_abs) ;
	    ##my ($stat_ue, $stat_le) = ($cmpl_br eq $node_abs) ?  ($char_abs, $char_pres) : ($char_pres, $char_abs) ;
	    ## my ($stat_ue, $stat_le) = ($cmpl_br eq $node_abs) ? ($char_pres, $char_abs) : ($char_abs, $char_pres) ;

	    my $set_indels = $br2indels{$cmpl_br};

		# Another BUG fixed on May 21 (Mon), 2012. #
	    my @new_set_indels = (defined $set_indels) ? @{$set_indels} : ();
	    push @new_set_indels, [$stat_ue, $stat_le, $sbjct_indices_blk];
	    $br2indels{$cmpl_br} = \@new_set_indels;
	    ##(defined $set_indels) or ($set_indels = $br2indels{$cmpl_br} = []);
	    ##push @{$set_indels}, [$stat_ue, $stat_le, $sbjct_indices_blk];

	}			# End of the 2nd foreach-loop #


		## Create a new map of blockwise states onto nodes. ##

	my %node2blwise_gstat = %{$node2blwise_gstat0};
	my @new_bw_gstat_nabs = @{$node2blwise_gstat{$node_abs}};

	foreach my $indx_blk (@{$sbjct_indices_blk}) { $new_bw_gstat_nabs[$indx_blk] = $char_pres; }
	$node2blwise_gstat{$node_abs} = \@new_bw_gstat_nabs;

    (print STDOUT "\n     < Subroutine 'recede_spec_deletion' ended />\n\n") if ($TEST_LEVEL3d>0);

	return (\%node2blwise_gstat, \%br2indels);

}	# End of the subroutine 'recede_spec_deletion' #


# incorporate_null_block:
#	This satellite subroutine of 'enum_psm_indel_scns_in_cls_blks'
#	incorporate a "null block", which is a block consisting only of $char_abs,
#	into a parsimonious scenario in a manner that minimizes the number of additional indel events.
#
# INPUT: ($bl, \@in_psm_scn, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ch, $prev_minct_add_indels), where 
#	$bl is the index of the "null block" to be incorporated;
#
#	@in_psm_scn = (\%br2indels, \%node2blwise_gstat) is an input parsimonious scenario, where
#	%{$br2indels} = ($branch_id => \@list_indels, ...) for the set of indels, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices in @{$blocks_in_cls};
#	%node2blwise_gstat = ($node_id => \@blwise_gstat, ...) for all nodes,
#	with $blwise_gstat[$b] = $char_pres/abs being the state in the $b th block in the cluster;
#
#	$char_abs/pres is the character representing the 'absence'/'presence' state;
#	$topnode_id is the ID of the top node;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_nodes, ...) for all interior nodes;
#	$prev_minct_add_indels is the previous minimum number of additional indels introduced,
#	which should be either undef, 0, or 1.
#
# OUTPUT: ($curr_minct_add_indels, \@set_expanded_psm_scns), where
#	$curr_minct_add_indels is the minimum number of additional indels introduced in this very session;
#	@{$set_expanded_psm_scns[$ps]} = (\%br2newset_indels, \%node2new_blwise_gstat)
#	specifies the $ps th parsimonious scenario AFTER incorporating the null block in question,
#	with %br2newset_indels following the same format as %br2indels,
#	and %node2new_blwise_gstat following the same format as %node2blwise_gstat.
#
sub incorporate_null_block ($\@$$$\%\%$) {

	my ($bl, $in_psm_scn, $char_abs, $char_pres, $topnode_id, $node2pa, $node2ch, $prev_minct_add_indels) = @_;

	my ($br2indels, $node2blwise_gstat) = @{$in_psm_scn};

	my $top_blwise_gstats = $node2blwise_gstat->{$topnode_id};
	my $ct_blks = @{$top_blwise_gstats};

    (print STDOUT "  < Subroutine 'incorporate_null_block'  (Subject block ID = $bl; #{blocks} = $ct_blks) >\n\n") if ($TEST_LEVEL2>0);

		# First, examine the indel events involving the flanking blocks. #

	# %{$br2l/rflank_event} = ($br_id => [$stat_pa, $stat_node], ...) for the left/right-flanking event.
	#
	my ($br2lflank_event, $ct_lf_events) = ($bl>0) ?
	    locate_parsim_changes (%{$node2pa}, %{$node2blwise_gstat}, $bl-1) : ({}, 0) ;
	unless (defined $br2lflank_event) {
	    print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'locate_parsim_changes' FAILED on the left-flanking block of the $bl th block, which is 'null': $! !\n";
	    return (undef, undef);
	}
    (print STDOUT "   --- Subroutine 'locate_parsim_changes' ended (with ", '$ct_lf_events = ', $ct_lf_events, ") ...\n\n") if ($TEST_LEVEL2>0);

	my ($br2rflank_event, $ct_rf_events) = ($bl + 1 < $ct_blks) ?
	    locate_parsim_changes (%{$node2pa}, %{$node2blwise_gstat}, $bl+1) : ({}, 0) ;
	unless (defined $br2rflank_event) {
	    print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'locate_parsim_changes' FAILED on the right-flanking block of the $bl th block, which is 'null': $! !\n";
	    return (undef, undef);

	}
    (print STDOUT "   --- Subroutine 'locate_parsim_changes' ended (with ", '$ct_rf_events = ', $ct_rf_events, ") ...\n\n") if ($TEST_LEVEL2>0);

	my @br_lfl_events = keys %{$br2lflank_event};
	my @br_rfl_events = keys %{$br2rflank_event};
	my $br_flanking_events = union_nmarray (@br_lfl_events, @br_rfl_events);

		# Examine whether the flanking indel events at either flanking block are
		# isolated or combine together with those at the other flanking block.

	my %br2compensating;
		#  = ($br_id => [$stat_node_lfl, $stat_node_rfl], ...),
		# for the branches where the left-flanking and right-flanking blocks experience the opposite indel events.
	my %br2single;
		# Any indel event, except those in %br2compensating, could be in this category.
		# = ($br_id => [$stat_pa, $stat_node, $sgn], ...), where $sgn = -1, +1, or 0 if only left flanking, right flanking, or both has/have such an indel event, respectively.

	my %pv2triple; # = ($pivotal_node => \@triple, ...).
		# ALL of the parent and sibling branches have indel events, with pivotal node sharing $char_pres.
		# @{$triple[k]} = ($br_id, $sgn), where $sgn = -1, +1, or 0 if only left flanking, right flanking, or both has/have such an indel event, respectively.
	my %pv2combi; # = ($pivotal_node => \@combi, ...).
		# Only one among the parent and sibling branches has NO indel event, and the pivotal node shares $char_pres.
		# @{$combi[k]} = ($br_id, $sgn), where $sgn = -1, +1, or 0 if only left flanking, right flanking, or both has/have such an indel event, respectively.
		# The last element of @combi should be the event to be added onto the branch withOUT indel event (and in the $b th block).

	my %pv2already; # = ($node_id => 1, ...) for the pivotal nodes already examined.

	my $ct_nonadd_indels = 0; # #{indels with $curr_minct_add_indels = 0}

	my $ct_br_fl_events = @{$br_flanking_events};
    (print STDOUT "   +++ The 1st Process:  Sorting out the flanking indel events (along the $ct_br_fl_events branches). +++\n\n") if ($TEST_LEVEL2>1);

	foreach my $br (@{$br_flanking_events}) { # 1st main outer foreach-loop #

    (print STDOUT "    [ Branch ID = $br ]\n\n") if ($TEST_LEVEL2>1);

		# Skip the branch already examined. #
	    if ((defined $br2compensating{$br}) or (defined $br2single{$br})) { 
    (print STDOUT "     .. Already Examined ..\n\n") if ($TEST_LEVEL2>1);
		next;
	    }

	    my ($if_succ, $stat_pa, $stat_node, $sgn) = examine_flanking_events ($br, %{$br2lflank_event}, %{$br2rflank_event}, %br2compensating, $ct_nonadd_indels);
	    unless ($if_succ) {
		print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'examine_flanking_events' FAILED on the branch '$br': $! !\n";
		return (undef, undef);
	    }
    (print STDOUT "    --- Subroutine 'examine_flanking_events' ended ...\n\n") if ($TEST_LEVEL2>1);

	    unless (defined $sgn) {
		if (defined $br2compensating{$br}) {
    		    print STDOUT "[NOTICE] The branch '$br' is flanked by compensating indel events, so I will skip examining its complementary branches ...\n\n";
		    my $pa_br = $node2pa->{$br};
		    $pv2already{$br} = $pv2already{$pa_br} = 1;
		    next;
		} else {
		    print STDOUT "[ERROR 'incorporate_null_block'] ILLEGAL OUTPUT of 'examine_flanking_events': ", '$sgn', " is UNDEFINED, yet ", '$br2compensating{$br}', " is also UNDEFINED for the branch '$br': $! !\n";
		    return (undef, undef);
		}
	    }

	    $br2single{$br} = [$stat_pa, $stat_node, $sgn]; 

	    my $pv; # Pivotal node.
	    my @complementary_brs = (); # element = [$br_id, $if_pv_eq_ue] for branches that 'complement' $br, with $if_pv_eq_ue = 1 if the pivotal node is the upper-end of the branch, = 0 if it is the lower-end.
	    if ($stat_node ne $char_abs) {
		# The lower-end of $br is the pivotal point.
		$pv = $br;
		if (defined $pv2already{$pv}) {
	(print STDOUT "    ... The pivotal node ($pv) was already examined ...\n\n") if ($TEST_LEVEL2>1);
		    next;
		}

		my $children = $node2ch->{$br};
		unless (defined $children) { # The 'pivotal point' is the exterior node.
	(print STDOUT "    ... The pivotal node ($pv) is exterior, so the branch '$br' should have NO compensatory branches ...\n\n") if ($TEST_LEVEL2>1);
		    next;
		}

		foreach my $ch (sort {$a <=> $b} @{$children}) { push @complementary_brs, [$ch, 1]; }

	    } else {
		# The upper-end of $br is the pivotal point.
		my $pa_br = $pv = $node2pa->{$br};
		if (defined $pv2already{$pv}) {
	(print STDOUT "    ... The pivotal node ($pv) was already examined ...\n\n") if ($TEST_LEVEL2>1);
		    next;
		}

			## A BUG fixed on May 15 (Tue), 2012. ##
		(defined $node2pa->{$pa_br}) and (push @complementary_brs, [$pa_br, 0]); # If $node2pa->{$pa_br} is undefined, there should not be a branch above it.
		## push @complementary_brs, [$pa_br, 0];
		my $siblings = $node2ch->{$pa_br};
		foreach my $sbl (sort {$a <=> $b} @{$siblings}) { ($sbl == $br) or (push @complementary_brs, [$sbl, 1]); }
	    }

	    my $ct_cmpl_brs = @complementary_brs;

	(print STDOUT "\n     { Pivotal node ID = $pv (#{complementary branches} = $ct_cmpl_brs) }.\n\n") if ($TEST_LEVEL1>1);

	    my @combi = ([$br, $sgn]);
	    my @no_event = (); # element = $br_id, for the branches withOUT events with $char_pres at the pivital node.

	    foreach my $br_info (@complementary_brs) { # Main middle foreach-loop #
		my ($br_c, $if_pv_eq_ue) = @{$br_info};
		my $compensating = $br2compensating{$br_c};
		if (defined $compensating) {
		    print STDOUT "[WARNING 'incorporate_null_block'] The compensatory branch '$br_c' has compensating flanking indel events (, which is actually impossible) !!\n\n";
		    push @no_event, $br_c, $br_c;
		    last;
		    ##my $sgn_c = ($compensating->[0] eq $char_abs) ? -1 : 1; # Regard that the pivotal is on the upper-end. #
		    ##$if_pv_eq_ue or ($sgn_c = - $sgn_c); # Pivotal was actually on the lower-end. #
		    ##push @combi, [$br_c, $sgn];
		    ##next;
		}
		my $single = $br2single{$br_c};
		if (defined $single) {
		    my $stat_pv_c = $if_pv_eq_ue ? $single->[0] : $single->[1];
		    my $sgn_c = $single->[2];
		    if ($stat_pv_c eq $char_abs) {
			push @no_event, $br_c;
		    } else {
			push @combi, [$br_c, $sgn_c];
		    }
		    next;
		}
		my ($if_succ2, $stat_pa_c, $stat_node_c, $sgn_c) = examine_flanking_events ($br_c, %{$br2lflank_event}, %{$br2rflank_event}, %br2compensating, $ct_nonadd_indels);
		unless ($if_succ2) {
		    print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'examine_flanking_events' FAILED on the branch '$br_c': $! !\n";
		    return (undef, undef);
		}
	(print STDOUT "    --- Subroutine 'examine_flanking_events' ended on the compensatory branch '$br_c' ...\n\n") if ($TEST_LEVEL2>1);
		unless (defined $stat_pa_c) { # No indel events. #
		    push @no_event, $br_c;
		    next;
		}
		$br2single{$br_c} = [$stat_pa_c, $stat_node_c, $sgn_c]; 
		my $stat_pv_c = $if_pv_eq_ue ? $stat_pa_c : $stat_node_c;
		if ($stat_pv_c eq $char_abs) {
		    push @no_event, $br_c;
		} else {
		    push @combi, [$br_c, $sgn_c];
		}
	    }				# End of the main middle foreach-loop #

	    $pv2already{$pv} = 1;
	    my $ct_no_event = @no_event;

	    if ($ct_no_event>1) { # We cannot use collectively the indel events around the pivotal node. #
    (print STDOUT "      ... The pivotal node '$pv' has $ct_no_event branches that canNOT be used to compensate for the indel along the branch '$br', thus it is DISMISSED ...\n\n") if ($TEST_LEVEL2>1);
		next;
	    }

	    if (@no_event<1) { # The indel events around the pivotal node can completely absorbe the "null" block. #

    (print STDOUT "      ... The indel events around the pivotal node '$pv' can COMPLETELY absorbe the 'null' block ...") if ($TEST_LEVEL2>1);

		$pv2triple{$pv} = \@combi;
		$ct_nonadd_indels++;

	    } else { # There is only one branch which has no flanking indel events. #

    (print STDOUT "      ... The indel events around the pivotal node '$pv' can absorbe the 'null' block by introducing JUST ONE ADDITIONAL event ...\n\n") if ($TEST_LEVEL2>1);

		push @combi, \@no_event;
		$pv2combi{$pv} = \@combi;
	    }

	}			# End of the 1st main outer foreach-loop #

		# Filter the potential expanded scenarios. #

	if ((defined $prev_minct_add_indels) and ($prev_minct_add_indels == 0) and ($ct_nonadd_indels == 0)) {
		# The resulting expanded scenarios will NOT be used anyway. #
    (print STDOUT "\n    < Subroutine 'incorporate_null_block' PREMATURELY Ended, because no expanded scenarios to be produced here will even tie the existing one(s). />\n\n") if ($TEST_LEVEL2>0);
	    return (1, undef, undef);
	} 

		# Now, expand the input parsimonious scenario
		# by incorporating the null block.

	# @{$set_expanded_psm_scns[$ps]} = (\%br2newset_indels, \%node2new_blwise_gstat)
	# specifies the $ps th parsimonious scenario AFTER incorporating the null block in question,
	# with %br2newset_indels following the same format as %br2indels,
	# and %node2new_blwise_gstat following the same format as %node2blwise_gstat.
	#
	my @set_expanded_psm_scns0 = ();
	my @indices_sb_blk = ($bl); # The "subject blocks" to be absorbed. #

	# my ($br2indels, $node2blwise_gstat) = @{$in_psm_scn};


    (print STDOUT "   +++ The 2nd Process:   Incorporate the 'null' block into pairs of compensating flanking events (along the ", scalar (keys %br2compensating), " branches). +++\n\n") if ($TEST_LEVEL2>1);

	foreach my $br (sort {$a <=> $b} keys %br2compensating) { # The 2nd main outer foreach-loop #

    (print STDOUT "    [ Branch ID = $br ]\n\n") if ($TEST_LEVEL2>1);

	    my %new_br2indels = %{$br2indels};
	    ##my %new_node2blwise_gstat = %{$node2blwise_gstat};

	    foreach my $new_stat_le (@{$br2compensating{$br}}) { # Middle foreach-loop #
	     ## my %br2compensating;
		#  = ($br_id => [$stat_node_lfl, $stat_node_rfl], ...),
		# for the branches where the left-flanking and right-flanking blocks experience the opposite indel events.

		my $new_stat_ue = ($new_stat_le eq $char_abs) ? $char_pres : $char_abs ;
		my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

		my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels, %{$node2blwise_gstat}, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le), which is one of the compensating events involving the block $bl, along the branch $br: $! !\n";
		    return (undef, undef);
	        }
	        unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le), which is one of the compensating events involving the block $bl, along the branch $br: $! !\n";
		    return (undef, undef);
		}
    (print STDOUT "     ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le), which is one of the compensating events involving the block $bl, along the branch $br...\n\n") if ($TEST_LEVEL2>1);

	    }			# End of the middle foreach-loop #

	    push @set_expanded_psm_scns0, [\%new_br2indels, $node2blwise_gstat];

	}			# End of the 2nd main outer foreach-loop #


    (print STDOUT "   +++ The 3rd Process:   Incorporate the 'null' block into perfect combinations of flanking events (around the ", scalar (keys %pv2triple), " pivotal nodes). +++\n\n") if ($TEST_LEVEL2>1);

	foreach my $pv (sort {$a <=> $b} keys %pv2triple) { # The 3rd main outer foreach-loop #
		## my %pv2triple; # = ($pivotal_node => \@triple, ...).
		# ALL of the parent and sibling branches have indel events, with pivotal node sharing $char_pres.
		# @{$triple[k]} = ($br_id, $sgn), where $sgn = -1, +1, or 0 if only left flanking, right flanking, or both has/have such an indel event, respectively.

	    my $surrounding_brs = $pv2triple{$pv};
	    my $ct_surr_brs = @{$surrounding_brs};

	(print STDOUT "    [ Pivotal node = $pv (#{surrounding branches} = $ct_surr_brs) ]\n\n") if ($TEST_LEVEL2>1);

	    my %new_br2indels = %{$br2indels};
	    my %new_node2blwise_gstat = %{$node2blwise_gstat};

	    for (my $i=0; $i<$ct_surr_brs; $i++) { # Middle for-loop #

		my $br_info = $surrounding_brs->[$i];
		my ($br, $sgn) = @{$br_info};

	(print STDOUT "       { The $i th branch (branch ID = $br) }\n") if ($TEST_LEVEL2>1);

		my ($new_stat_ue, $new_stat_le) = ($br eq $pv) ? 
			($char_abs, $char_pres) : ($char_pres, $char_abs) ;
		my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

		my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels, %new_node2blwise_gstat, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the pivotal node $pv): $! !\n\n";
		    return (undef, undef);
	        }
	        unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the pivotal node $pv): $! !\n\n";
		    return (undef, undef);
		}

	(print STDOUT "    ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the pivotal node $pv) ...\n\n") if ($TEST_LEVEL2>1);

	    }			# End of the middle for-loop #

	    # Flip the state at the pivotal node (in the block in question) #

	    my @blwise_gstat_pv = @{$new_node2blwise_gstat{$pv}};
	    $blwise_gstat_pv[$bl] = $char_pres;
	    $new_node2blwise_gstat{$pv} = \@blwise_gstat_pv;

	    push @set_expanded_psm_scns0, [\%new_br2indels, \%new_node2blwise_gstat];

	(print STDOUT "\n     ~~~ Finished processing the pivotal node = $pv ~~~\n\n") if ($TEST_LEVEL2>1);

	}			# End of the 3rd main outer foreach-loop #

	if (@set_expanded_psm_scns0>0) { # There is at least one expanded scenario with NO additional event introduced. #
	    my $ct_exp_psm_scns0 = @set_expanded_psm_scns0;
    (print STDOUT "\n   < Subroutine 'incorporate_null_block' normally Ended (with $ct_exp_psm_scns0 PERFECTly expanded parsimonious scenarios) />\n\n") if ($TEST_LEVEL2>0);

	    return (0, \@set_expanded_psm_scns0);

	} elsif ((defined $prev_minct_add_indels) and ($prev_minct_add_indels == 0)) {
		# Going further is useless. #
    (print STDOUT "\n    < Subroutine 'incorporate_null_block' PREMATURELY Ended, because no perfectly expanded scenarios were produced here, while there already exist some such scenarios. />\n\n") if ($TEST_LEVEL2>0);
	    return (1, undef);
	}

# OUTPUT: ($curr_minct_add_indels, \@set_expanded_psm_scns), where
#	$curr_minct_add_indels is the minimum number of additional indels introduced in this very session;

	## my ($b, $in_psm_scn, $char_abs, $char_pres, $topnode_id, $node2pa, $node2ch, $prev_minct_add_indels) = @_;

		# Basically, the following processes will
		# incorporate the null block by introducing an additional indel event, which will not be absorbed.

    (print STDOUT "   +++ The 4th Process:   Incorporate the 'null' block into pairs, each consisting of an existing flanking event and a new compensatory event (along the ", scalar (keys %br2single), " branches). +++\n\n") if ($TEST_LEVEL2>1);

	my @set_expanded_psm_scns = ();

	foreach my $br (sort {$a <=> $b} keys %br2single) { # The 4th main outer foreach-loop #
	## my %br2single;
		# Any indel event, except those in %br2compensating, could be in this category.
		# = ($br_id => [$stat_pa, $stat_node, $sgn], ...), where $sgn = -1, +1, or 0 if only left flanking, right flanking, or both has/have such an indel event, respectively.

    (print STDOUT "    [ Branch ID = $br ]\n\n") if ($TEST_LEVEL2>1); 

	   my ($new_stat_ue, $new_stat_le, $sgn) = @{$br2single{$br}};

	   my %new_br2indels = %{$br2indels};
	   ## my %new_node2blwise_gstat = %{$node2blwise_gstat};

	    my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

	    my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels, %{$node2blwise_gstat}, $char_abs, %{$node2pa});
	    unless (defined $if_absorbed) {
		print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le), which was to be absorved, involving the block $bl along the branch $br: $! !\n\n";
		return (undef, undef);
	    }
	    unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le), which was to be absorved, involving the block $bl along the branch $br: $! !\n\n";
		return (undef, undef);
	    }
	(print STDOUT "      ... Subroutine 'incorporate_new_event' successfully absorved the event ($new_stat_ue -> $new_stat_le), which was to be absorved, involving the block $bl along the branch $br ...\n\n") if ($TEST_LEVEL2>1);

	    my @additional = ($br, $new_stat_le, $new_stat_ue);

	    my $if_absorbed2 = incorporate_new_event (@additional, @indices_sb_blk, %new_br2indels, %{$node2blwise_gstat}, $char_abs, %{$node2pa});
	    unless (defined $if_absorbed2) {
		print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the ADDITINAL event ($new_stat_le -> $new_stat_ue), involving the block $bl along the branch $br: $! !\n\n";
		return (undef, undef);
	    }
		# A Bug FIXED on May 14 (Mon), 2012. #
	    if ($if_absorbed2) { # The new event was somehow absorbed into any existing event. #
	    ##if ($if_absorbed) { # The new event was somehow absorbed into any existing event. #
		print STDOUT "[WARNING 'incorporate_null_block'] Somehow, the subroutine 'incorporate_new_event' ABSORBED the ADDITIONAL event ($new_stat_le -> $new_stat_ue), involving the block $bl along the branch $br !!\n\n";

	        push @set_expanded_psm_scns0, [\%new_br2indels, $node2blwise_gstat];
	    } else {
	(print STDOUT "    ... Subroutine 'incorporate_new_event' successfully incorporated (while not absorbing) the ADDITIONAL event ($new_stat_le -> $new_stat_ue), involving the block $bl along the branch $br ...\n\n") if ($TEST_LEVEL2>1);

	        push @set_expanded_psm_scns, [\%new_br2indels, $node2blwise_gstat];
	    }
	}			# End of the 4th main outer foreach-loop #

    (print STDOUT "\n   +++ The 4th process ENDED (#{expanded scenarios with ZERO additional event} = ", scalar (@set_expanded_psm_scns0), ", #{expanded scenarios with ONE additional event} = ", scalar (@set_expanded_psm_scns), ") +++\n\n") if ($TEST_LEVEL2>1);

    (print STDOUT "\n   +++ The 5th Process:   Incorporate the 'null' block into sub-perfect combinations of flanking events (around the ", scalar (keys %pv2combi), " pivotal nodes). +++\n\n") if ($TEST_LEVEL2>1);

	foreach my $pv (sort {$a <=> $b} keys %pv2combi) { # The 5th main outer foreach-loop #
		## my %pv2combi; # = ($pivotal_node => \@combi, ...).
		# Only one among the parent and sibling branches has NO indel event, and the pivotal node shares $char_pres.
		# @{$combi[k]} = ($br_id, $sgn), where $sgn = -1, +1, or 0 if only left flanking, right flanking, or both has/have such an indel event, respectively.
		# The last element of @combi should be the event to be added onto the branch withOUT indel event (and in the $b th block).

	    my %new_br2indels = %{$br2indels};
	    my %new_node2blwise_gstat = %{$node2blwise_gstat};

	    my @combi = @{$pv2combi{$pv}};
	    my $additional = pop @combi;
	    my ($br_add, $sgn_add) = @{$additional};

	    my $ct_comp_brs = @combi;

	(print STDOUT "    [ Pivotal node = $pv (#{compensatory branches} = $ct_comp_brs, additional branch ID = $br_add) ]\n\n") if ($TEST_LEVEL2>1);

	    for (my $i=0; $i<$ct_comp_brs; $i++) { # Middle foreach-loop #

		my $br_info = $combi[$i];
		my ($br, $sgn) = @{$br_info};

	(print STDOUT "      { The $i th branch (branch ID = $br) }\n") if ($TEST_LEVEL2>1);

		my ($new_stat_ue, $new_stat_le) = ($br eq $pv) ? 
			($char_abs, $char_pres) : ($char_pres, $char_abs) ;
		my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

		my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels, %new_node2blwise_gstat, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the pivotal node $pv and an additional event along the branch $br_add): $! !\n\n";
		    return (undef, undef);
	        }
	        unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the pivotal node $pv and an additional event along the branch $br_add): $! !\n\n";
		    return (undef, undef);
		}
	(print STDOUT "      ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the pivotal node $pv) ...\n\n") if ($TEST_LEVEL2>1);

	    }			# End of the middle foreach-loop #

	    # Flip the state at the pivotal node (in the block in question) #

	    my @blwise_gstat_pv = @{$new_node2blwise_gstat{$pv}};
	    $blwise_gstat_pv[$bl] = $char_pres;
	    $new_node2blwise_gstat{$pv} = \@blwise_gstat_pv;

		# Incorporate the additional event. #

	    my ($add_stat_ue, $add_stat_le) = ($br_add eq $pv) ? 
		    ($char_abs, $char_pres) : ($char_pres, $char_abs) ;
	    my @to_be_added = ($br_add, $add_stat_ue, $add_stat_le);

	    my $if_absorbed = incorporate_new_event (@to_be_added, @indices_sb_blk, %new_br2indels, %new_node2blwise_gstat, $char_abs, %{$node2pa});
	    unless (defined $if_absorbed) {
	         print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the ADDITIONAL event ($add_stat_ue -> $add_stat_le) of the block $bl along the branch $br_add (with the pivotal node $pv): $! !\n";
	        return (undef, undef);
	    }
	    if ($if_absorbed) { # Somehow, the new event WAS absorbed into any existing event. #

	         print STDOUT "[WARNING 'incorporate_null_block'] Somehow, the subroutine 'incorporate_new_event' ABSORBED the ADDITIONAL event ($add_stat_ue -> $add_stat_le) of the block $bl along the branch $br_add (with the pivotal node $pv): $! !\n";

	        push @set_expanded_psm_scns0, [\%new_br2indels, \%new_node2blwise_gstat];
	    } else {
		# A BUG fixed on May 15 (Tue), 2012. #
	(print STDOUT "    ... Subroutine 'incorporate_new_event' successfully incorporated (while not absorbing) the ADDITIONAL event ($add_stat_ue -> $add_stat_le), involving the block $bl along the branch $br_add ...\n\n") if ($TEST_LEVEL2>1);
	##(print STDOUT "    ... Subroutine 'incorporate_new_event' successfully incorporated (while not absorbing) the ADDITIONAL event ($add_stat_le -> $add_stat_ue), involving the block $bl along the branch $br_add ...\n\n") if ($TEST_LEVEL2>1);

	        push @set_expanded_psm_scns, [\%new_br2indels, \%new_node2blwise_gstat];
	    }

	(print STDOUT "\n     ~~~ Finished processing the pivotal node = $pv ~~~\n\n") if ($TEST_LEVEL2>1);
	}			# End of the 5th main outer foreach-loop #

    (print STDOUT "\n   +++ The 5th process ENDED (#{expanded scenarios with ZERO additional event} = ", scalar (@set_expanded_psm_scns0), ", #{expanded scenarios with ONE additional event} = ", scalar (@set_expanded_psm_scns), ") +++\n\n") if ($TEST_LEVEL2>1);


	if (@set_expanded_psm_scns0>0) {
		## If there are any parsimonious scenarios with the null block ($bl) incorporated withOUT additional indel events, return them (,although this should NOT usually occur ...) ##

	    print STDOUT "[WARNING 'incorporate_null_block'] Somehow, UNEXPECTEDLY, the null block '$bl' was incorporated withOUT the need of any ADDITIONAL indel event !\n";
	    my $ct_exp_psm_scns0 = @set_expanded_psm_scns0;

    (print STDOUT "\n  < Subroutine 'incorporate_null_block' Ended (with $ct_exp_psm_scns0 perfectly expanded scenarios) />\n\n") if ($TEST_LEVEL2>0);

	    return (0, \@set_expanded_psm_scns0);

	}

		## ADDED on May 20 (Sun), 2012. ##

	## my @br_lfl_events = keys %{$br2lflank_event};
	## my @br_rfl_events = keys %{$br2rflank_event};
	if ((@br_lfl_events>0) and (@br_rfl_events>0)) { # A Big IF. #

    (print STDOUT "\n   +++ The 6th Process (SPECIAL SESSION):   Try to incorporate the 'null' block into 1 'insertion' + 2 'deletions', which resulted from a transformation of 2 'insertion' events on a pair 'sibling' branches by flipping the state from '$char_abs' to '$char_pres' at the shared node. +++\n\n") if ($TEST_LEVEL2>1);

	  my %br2already2;
	  foreach my $br_lf (sort {$a <=> $b} keys %{$br2lflank_event}) { # The 6th main outer foreach-loop (over branches with left-flanking events) #

	    (defined $br2already2{$br_lf}) and next;
	    (defined $br2rflank_event->{$br_lf}) and next;

	    my ($stat_ue_lf, $stat_le_lf) = @{$br2lflank_event->{$br_lf}};
	    my $op_pv = ($stat_ue_lf eq $char_abs) ? $node2pa->{$br_lf} : $br_lf; # The 'opposite' pivotal node, where the state is 'absence'.

	    my $blwise_gstats_opv0 = $node2blwise_gstat->{$op_pv};
	    ($blwise_gstats_opv0->[$bl+1] eq $char_abs) or next; # If the state at $op_pv in the left-flanking block is NOT $char_abs, it must have been covered in the previous processes.


	    my @complementary = ();
	    if ($br_lf == $op_pv) { # The 'opposite' pivotal node is at the lower-end of the left-flanking branch. #
		my $ch_bras = $node2ch->{$br_lf};
		(defined $ch_bras) or next;
		@complementary = @{$ch_bras};

	    } else { # The 'opposite' pivotal node is at the upper-end of the left-flanking branch. #
		(defined $node2pa->{$op_pv}) and (push @complementary, $op_pv); # $op_pv is NOT the top node, and therefore is a lower-end of the parent branch. #
		my $sbl_bras = $node2ch->{$op_pv};
		foreach my $sbl (@{$sbl_bras}) { ($sbl == $br_lf) or (push @complementary, $sbl); }
	    }

	    foreach my $cmp_br (@complementary) { $br2already2{$cmp_br} = 1; }
	    (@complementary>2) and next;

	    my $br_rf; # Branch for the right-flanking event.
	    my $br_rem; # The remaining branch.
	    foreach my $cmp_br (@complementary) { # 1st preliminary middle foreach-loop #
		(defined $br2lflank_event->{$cmp_br}) and last; # Skip branches accommodating a left flanking event.
		my $rfl_event = $br2rflank_event->{$cmp_br};
		if (defined $rfl_event) {
		    my ($stat_ue_rf, $stat_le_rf) = @{$rfl_event}; 
		    my $stat_opv_rf = ($cmp_br eq $op_pv) ? $stat_le_rf : $stat_ue_rf;
		    unless ($stat_opv_rf eq $char_abs) {
			print STDOUT "[ERROR 'incorporate_null_block'] The upper-end of the branch '$cmp_br', which is a complementary of the branch '$br_lf', has a state '$stat_opv_rf' at the 'opposite' pivotal node '$op_pv', while it should have the state '$char_abs' there: $! !\n\n";
			return (undef, undef);
		    }
		    $br_rf = $cmp_br;

		} else { # The branch accommodates NO flanking events. #
		    $br_rem = $cmp_br;
		}
	    }			# End of the 1st preliminary middle foreach-loop #

	    ((defined $br_rf) and (defined $br_rem)) or next;  # Skip if both $br_rf and $br_rem is NOT defined.

    if ($TEST_LEVEL2d>2) {
	print STDOUT "\n     ++ Found a 'trio' of branches to be examined ++\n\n";
	print STDOUT "       {Branch w/ left-flanking event}  = $br_lf,\n";
	print STDOUT "       {Branch w/ right-flanking event} = $br_rf,\n";
	print STDOUT "       {Branch w/ NO events}            = $br_rem.\n\n";
    }

    (print STDOUT "      + (Trial A)  Recede the left-flanking event (branch: $br_lf, block: ", $bl-1, ") +\n\n") if ($TEST_LEVEL2d>1);

	    my ($node2blwise_gstat_a0, $br2indels_a0) = recede_spec_deletion ($br_lf, $bl-1, $op_pv, %{$node2ch}, %{$node2pa}, $char_abs, $char_pres, %{$node2blwise_gstat}, %{$br2indels});
	    unless (defined $node2blwise_gstat_a0) {
		print STDOUT "[ERROR 'incorporate_null_block'] Subroutine 'recede_spec_deletion' FAILED to recede the left-flanking event at '$br_lf' in the set of blocks containing '", $bl -1, "': $! !\n\n";
		return (undef, undef);
	    }
    (print STDOUT "\n       --- 'recede_spec_deletion' ended ...\n\n") if ($TEST_LEVEL2d>1);

    if ($TEST_LEVEL2d>2) { # TEST OUTPUT for 'recede_spec_deletion' in the Trial A. #
	print STDOUT "      [ Output of 'recede_spec_deletion' ]\n\n";
	print STDOUT '      Content of %{$node2blwise_gstat_a0}:', "\n\n";
	print STDOUT ' ' x 6, join ("\t", '#Node_ID', "'block-wise,gap,states,(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat_a0}) {
	    my $cnct_bw_gstats = join ('', @{$node2blwise_gstat_a0->{$node}});
	    print STDOUT ' ' x 6, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	}
	print STDOUT "\n\n", ' ' x 6, 'Content of %{$br2indels_a0}:', "\n\n";
	print STDOUT ' ' x 6, join ("\t", '#Branch_ID', 'stat_ue/stat_le/indices,involved,blocks', '...'), "\n\n";
	foreach my $br_id (sort {$a <=> $b} keys %{$br2indels_a0}) {
	    my $list_indels = $br2indels_a0->{$br_id};
	    my @out_indels = ();
	    foreach my $indel (@{$list_indels}) {
		my ($stat_ue, $stat_le, $indices_blks) = @{$indel};
		my $cnct_indices = join (',', @{$indices_blks});
		push @out_indels, join ('/', $stat_ue, $stat_le, $cnct_indices);
	    }
	    print STDOUT ' ' x 6, join ("\t", $br_id, @out_indels), "\n";
	}
	print STDOUT "\n\n";
    } # END of "TEST OUTPUT for 'recede_spec_deletion' in the Trial A". #

	    ### $node2blwise_gstat_a0->{$op_pv}[$bl] = $char_pres; ### -- Actually, this was SUPERFLUOUS!!

=NOW, it turned out that the trio of newly creaetd 'deletions' canNOT generate a mapping of parsimonious indel scenarios....
    (print STDOUT "\n       .. (sub-trial. 1) Absorbe the newly created 'deletions' at the trio of branches ($br_lf, $br_rf, $br_rem) surrounding the node '$op_pv' in the block '$bl' ..\n\n") if ($TEST_LEVEL2d>1);

##	    my %new_node2blwise_gstat_a1 = %{$node2blwise_gstat_a0};
##	    my %new_br2indels_a1 = %{$br2indels_a0};

		# Merge the newly created 'insertion' with the old one, both along $br_rf  #
		## (NOW, this part turned out UNnecessary...)
##	    my @new_indels_brrf_a = @{$new_br2indels_a1{$br_rf}};
##	    my $new_ins_a = pop @new_indels_brrf_a;
##	    my ($new_stat_ue_a, $new_stat_le_a, $new_indices_blk_a) = @{$new_ins_a};
##	    $new_br2indels_a1{$br_rf} = \@new_indels_rbrf_a;
##	    my @to_be_absorbed_a = ($br_rf, $new_stat_ue_a, $new_stat_le_a);
##	    my $if_absorbed_a = incorporate_new_event (@to_be_absorbed_a, @{$new_indices_blk_a}, %new_br2indels_a1, %new_node2blwise_gstat_a1, $char_abs, %{$node2pa});
##	    unless ((defined $if_absorbed_a) and $if_absorbed_a) {
##		print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to ABSORB the new 'insertion' ($new_stat_ue_a -> $new_stat_le_a) along the branch ", '$br_rf = ' , "$br_rf and on the indices = (", join (', ', @{$new_indices_blk_a}), "): $! !\n";
##		return (undef, undef);
##	    }
##	# %{$br2indels} = ($branch_id => \@list_indels, ...) for the set of indels, with
##	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices in @{$blocks_in_cls};


	    foreach my $br ($br_lf, $br_rf, $br_rem) { # 1st main middle foreach-loop #
		my ($new_stat_ue, $new_stat_le) = ($br eq $op_pv) ? ($char_pres, $char_abs) : ($char_abs, $char_pres) ;
		my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

		my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels_a1, %new_node2blwise_gstat_a1, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv': $! !\n\n";
		    return (undef, undef);
	        }
	        unless (($br eq $br_lf) or ($if_absorbed)) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br ", '(ne $br_lf)', " (with the 'opposite' pivotal node '$op_pv'): $! !\n\n";
		    return (undef, undef);
		}
	(print STDOUT "        ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node $op_pv) ...\n\n") if ($TEST_LEVEL2d>2);
	    }				# End of the 1st main middle foreach-loop #

	    push @set_expanded_psm_scns, [\%new_br2indels_a1, \%new_node2blwise_gstat_a1];

 	## End of "NOW, it turned out that the trio of newly creaetd 'deletions' canNOT generate a mapping of parsimonious indel scenarios...."
=cut

    (print STDOUT "\n       .. (sub-trial. 2) Absorbe the newly created pair of 'insertion' and 'deletion' along the branch '$br_rf) in the block '$bl' ...\n\n") if ($TEST_LEVEL2d>1);

	    my %new_node2blwise_gstat_a2 = %{$node2blwise_gstat_a0};
	    my %new_br2indels_a2 = %{$br2indels_a0};

	    foreach my $to_be_absorbed ([$br_rf, $char_abs, $char_pres], [$br_rf, $char_pres, $char_abs]) { # 2nd main middle foreach-loop #
		my ($br, $new_stat_ue, $new_stat_le) = @{$to_be_absorbed};

		my $if_absorbed = incorporate_new_event (@{$to_be_absorbed}, @indices_sb_blk, %new_br2indels_a2, %new_node2blwise_gstat_a2, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv': $! !\n\n";
		    return (undef, undef);
	        }
	        unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv'): $! !\n\n";
		    return (undef, undef);
		}
	(print STDOUT "        ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node $op_pv) ...\n\n") if ($TEST_LEVEL2d>2);
	    } 			# End of the 2nd main middle foreach-loop #

	    push @set_expanded_psm_scns, [\%new_br2indels_a2, \%new_node2blwise_gstat_a2];


    (print STDOUT "      + (Trial B)  Recede the right-flanking event (branch: $br_rf, block: ", $bl+1, ") +\n\n") if ($TEST_LEVEL2d>1);

		# A BUG fixed on May 21 (Mon), 2012. #
	    my ($node2blwise_gstat_b0, $br2indels_b0) = recede_spec_deletion ($br_rf, $bl+1, $op_pv, %{$node2ch}, %{$node2pa}, $char_abs, $char_pres, %{$node2blwise_gstat}, %{$br2indels});
	    ##my ($node2blwise_gstat_b0, $br2indels_b0) = recede_spec_deletion ($br_rf, $b+1, $op_pv, %{$node2ch}, %{$node2pa}, $char_abs, $char_pres, %{$node2blwise_gstat}, %{$br2indels});
	    unless (defined $node2blwise_gstat_b0) {
		print STDOUT "[ERROR 'incorporate_null_block'] Subroutine 'recede_spec_abs' FAILED to recede the right-flanking event at '$br_rf' in the set of blocks containing '", $bl +1, "': $! !\n\n";
		return (undef, undef);
	    }
    (print STDOUT "\n       --- 'recede_spec_deletion' ended ...\n\n") if ($TEST_LEVEL2d>1);

    if ($TEST_LEVEL2d>2) { # TEST OUTPUT for 'recede_spec_deletion' in the Trial B. #
	print STDOUT "      [ Output of 'recede_spec_deletion' ]\n\n";
	print STDOUT '      Content of %{$node2blwise_gstat_b0}:', "\n\n";
	print STDOUT ' ' x 6, join ("\t", '#Node_ID', "'block-wise,gap,states,(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat_b0}) {
	    my $cnct_bw_gstats = join ('', @{$node2blwise_gstat_b0->{$node}});
	    print STDOUT ' ' x 6, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	}
	print STDOUT "\n\n", ' ' x 6, 'Content of %{$br2indels_b0}:', "\n\n";
	print STDOUT ' ' x 6, join ("\t", '#Branch_ID', 'stat_ue/stat_le/indices,involved,blocks', '...'), "\n\n";
	foreach my $br_id (sort {$a <=> $b} keys %{$br2indels_b0}) {
	    my $list_indels = $br2indels_b0->{$br_id};
	    my @out_indels = ();
	    foreach my $indel (@{$list_indels}) {
		my ($stat_ue, $stat_le, $indices_blks) = @{$indel};
		my $cnct_indices = join (',', @{$indices_blks});
		push @out_indels, join ('/', $stat_ue, $stat_le, $cnct_indices);
	    }
	    print STDOUT ' ' x 6, join ("\t", $br_id, @out_indels), "\n";
	}
	print STDOUT "\n\n";
    } # END of "TEST OUTPUT for 'recede_spec_deletion' in the Trial B". #

	    ### $node2blwise_gstat_b0->{$op_pv}[$bl] = $char_pres; ### --- Actually, this was SUPERFLUOUS!!

=NOW, it turned out that the trio of newly creaetd 'deletions' canNOT generate a mapping of parsimonious indel scenarios....

    (print STDOUT "\n       .. (sub-trial. 1) Absorbe the newly created 'deletions' at the trio of branches ($br_lf, $br_rf, $br_rem) surrounding the node '$op_pv' in the block '$bl' ..\n\n") if ($TEST_LEVEL2d>1);

	    my %new_node2blwise_gstat_b1 = %{$node2blwise_gstat_b0};
	    my %new_br2indels_b1 = %{$br2indels_b0};
	    foreach my $br ($br_lf, $br_rf, $br_rem) { # 3rd main middle foreach-loop #
		my ($new_stat_ue, $new_stat_le) = ($br eq $op_pv) ? ($char_pres, $char_abs) : ($char_abs, $char_pres) ;
		my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

		my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels_b1, %new_node2blwise_gstat_b1, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv': $! !\n\n";
		    return (undef, undef);
	        }
	        unless (($br == $br_rf) or $if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv'): $! !\n\n";
		    return (undef, undef);
		}
	(print STDOUT "        ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node $op_pv) ...\n\n") if ($TEST_LEVEL2d>2);
	    }				# End of the 3rd main middle foreach-loop #

	    push @set_expanded_psm_scns, [\%new_br2indels_b1, \%new_node2blwise_gstat_b1];

 	## END of "NOW, it turned out that the trio of newly creaetd 'deletions' canNOT generate a mapping of parsimonious indel scenarios...."
=cut


    (print STDOUT "\n       .. (sub-trial. 2) Absorbe the newly created pair of 'insertion' and 'deletion' along the branch '$br_rf) in the block '$bl' ...\n\n") if ($TEST_LEVEL2d>1);

	    my %new_node2blwise_gstat_b2 = %{$node2blwise_gstat_b0};
	    my %new_br2indels_b2 = %{$br2indels_b0};

		# A BUG fixed on May 21 (Mon), 2012. #
	    foreach my $to_be_absorbed ([$br_lf, $char_abs, $char_pres], [$br_lf, $char_pres, $char_abs]) { # 4th main middle foreach-loop #
	    ##foreach my $to_be_absorbed ([$br_rf, $char_abs, $char_pres], [$br_rf, $char_pres, $char_abs]) { # 4th main middle foreach-loop #
		my ($br, $new_stat_ue, $new_stat_le) = @{$to_be_absorbed};

		my $if_absorbed = incorporate_new_event (@{$to_be_absorbed}, @indices_sb_blk, %new_br2indels_b2, %new_node2blwise_gstat_b2, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv': $! !\n\n";
		    return (undef, undef);
	        }
	        unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv'): $! !\n\n";
		    return (undef, undef);
		}
	(print STDOUT "        ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node $op_pv) ...\n\n") if ($TEST_LEVEL2d>2);
	    } 			# End of the 4th main middle foreach-loop #

	    push @set_expanded_psm_scns, [\%new_br2indels_b2, \%new_node2blwise_gstat_b2];


    (print STDOUT "      + (Trial C)  Recede the both left- and right-flanking events (branches: $br_lf and $br_rf; blocks: ", join (' and ', $bl-1, $bl+1), ") +\n\n") if ($TEST_LEVEL2d>1);

		# A BUG fixed on May 21 (Mon), 2012. #
	    my ($node2blwise_gstat_c0, $br2indels_c0) = recede_spec_deletion ($br_rf, $bl+1, $op_pv, %{$node2ch}, %{$node2pa}, $char_abs, $char_pres, %{$node2blwise_gstat_a0}, %{$br2indels_a0});
	    ##my ($node2blwise_gstat_c0, $br2indels_c0) = recede_spec_deletion ($br_rf, $b+1, $op_pv, %{$node2ch}, %{$node2pa}, $char_abs, $char_pres, %{$node2blwise_gstat_a0}, %{$br2indels_a0});
	    unless (defined $node2blwise_gstat_c0) {
		print STDOUT "[ERROR 'incorporate_null_block'] Subroutine 'recede_spec_abs' FAILED to recede the right-flanking event at '$br_rf' in the set of blocks containing '", $bl +1, "', after successfully receding the left-flanking event at '$br_lf' in the set of blocks containing '", $bl-1, "': $! !\n\n";
		return (undef, undef);
	    }
    (print STDOUT "\n       --- 'recede_spec_deletion' ended ...\n\n") if ($TEST_LEVEL2d>1);

    if ($TEST_LEVEL2d>2) { # TEST OUTPUT for 'recede_spec_deletion' in the Trial C. #
	print STDOUT "      [ Output of 'recede_spec_deletion' ]\n\n";
	print STDOUT '      Content of %{$node2blwise_gstat_c0}:', "\n\n";
	print STDOUT ' ' x 6, join ("\t", '#Node_ID', "'block-wise,gap,states,(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat_c0}) {
	    my $cnct_bw_gstats = join ('', @{$node2blwise_gstat_c0->{$node}});
	    print STDOUT ' ' x 6, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	}
	print STDOUT "\n\n", ' ' x 6, 'Content of %{$br2indels_c0}:', "\n\n";
	print STDOUT ' ' x 6, join ("\t", '#Branch_ID', 'stat_ue/stat_le/indices,involved,blocks', '...'), "\n\n";
	foreach my $br_id (sort {$a <=> $b} keys %{$br2indels_c0}) {
	    my $list_indels = $br2indels_c0->{$br_id};
	    my @out_indels = ();
	    foreach my $indel (@{$list_indels}) {
		my ($stat_ue, $stat_le, $indices_blks) = @{$indel};
		my $cnct_indices = join (',', @{$indices_blks});
		push @out_indels, join ('/', $stat_ue, $stat_le, $cnct_indices);
	    }
	    print STDOUT ' ' x 6, join ("\t", $br_id, @out_indels), "\n";
	}
	print STDOUT "\n\n";
    } # END of "TEST OUTPUT for 'recede_spec_deletion' in the Trial C". #


	    $node2blwise_gstat_c0->{$op_pv}[$bl] = $char_pres; ### This IS necessary!!

		# Merge the newly created two indel scenarios along $br_rem. #
	    my $indels_br_rem = $br2indels_c0->{$br_rem};
	    my $indel_c2 = pop @{$indels_br_rem};
	    my $indel_c1 = pop @{$indels_br_rem};
	    unless (($indel_c2->[0] eq $indel_c1->[0]) and ($indel_c1->[1] eq $indel_c2->[1])) {
		print STDOUT "[ERROR 'incorporate_null_block'] Somehow, the two newly created events along the branch '$br_rem' have CLASHING patterns: (", join (' -> ', @{$indel_c1}[0,1]), ") and (", join (' -> ', @{$indel_c2}[0,1]), "), while actually they should match each other: $! !\n\n";
		return (undef, undef);
	    }
	    my $mgd_indices_blk = union_nmarray (@{$indel_c1->[2]}, @{$indel_c2->[2]});
	    my @new_indel_c1 = (@{$indel_c1}[0,1], $mgd_indices_blk);
	    push @{$indels_br_rem}, \@new_indel_c1;
	    ##$indel_c1->[2] = $mgd_indices_blk;
	    ##push @{$indels_br_rem}, $indel_c1;

    (print STDOUT "\n       .. (sub-trial. (only one)) Absorbe the newly created 'deletions' at the trio of branches ($br_lf, $br_rf, $br_rem) surrounding the node '$op_pv' in the block '$bl' ..\n\n") if ($TEST_LEVEL2d>1);

	    my %new_node2blwise_gstat_c1 = %{$node2blwise_gstat_c0};
	    my %new_br2indels_c1 = %{$br2indels_c0};
	    foreach my $br ($br_lf, $br_rf, $br_rem) { # 5th main middle foreach-loop #

			# A BUG fixed on May 21 (Mon), 2012. #
			# (NOTE that the 'opposite' pivotal node should have 'presence' in the block $bl.)
		my ($new_stat_ue, $new_stat_le) = ($br == $op_pv) ? ($char_abs, $char_pres) : ($char_pres, $char_abs) ;
		##my ($new_stat_ue, $new_stat_le) = ($br eq $op_pv) ? ($char_abs, $char_pres) : ($char_pres, $char_abs) ;
		##my ($new_stat_ue, $new_stat_le) = ($br eq $op_pv) ? ($char_pres, $char_abs) : ($char_abs, $char_pres) ;

		my @to_be_absorbed = ($br, $new_stat_ue, $new_stat_le);

		my $if_absorbed = incorporate_new_event (@to_be_absorbed, @indices_sb_blk, %new_br2indels_c1, %new_node2blwise_gstat_c1, $char_abs, %{$node2pa});
		unless (defined $if_absorbed) {
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' FAILED to incorporate the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv': $! !\n\n";
		    return (undef, undef);
	        }
	        unless ($if_absorbed) { # The new event was NOT absorbed into any existing event. #
		     print STDOUT "[ERROR 'incorporate_null_block'] The subroutine 'incorporate_new_event' DID NOT ABSORB the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node '$op_pv'): $! !\n\n";
		    return (undef, undef);
		}
	(print STDOUT "        ... Subroutine 'incorporate_new_event' successfully absorbed the event ($new_stat_ue -> $new_stat_le) of the block $bl along the branch $br (with the 'opposite' pivotal node $op_pv) ...\n\n") if ($TEST_LEVEL2d>2);
	    }				# End of the 5th main middle foreach-loop #

	    push @set_expanded_psm_scns, [\%new_br2indels_c1, \%new_node2blwise_gstat_c1];

	(print STDOUT "\n     ++ Ended examining the 'trio' of branches: ", '($br_lf, $br_rf, $br_rem) = ', "($br_lf, $br_rf, $br_rem). ++\n\n") if ($TEST_LEVEL2d>2);

	  }		# End of the 6th main outer foreach-loop (over branches with left-flanking events) #
    (print STDOUT "\n   +++ The 6th process ENDED (#{expanded scenarios with ZERO additional event} = ", scalar (@set_expanded_psm_scns0), ", #{expanded scenarios with ONE additional event} = ", scalar (@set_expanded_psm_scns), ") +++\n\n") if ($TEST_LEVEL2>1);

	} # End of the Big IF. #


	## my ($br2rflank_event, $ct_rf_events) = ($bl + 1 < $ct_blks) ?
	##    locate_parsim_changes (%{$node2pa}, %{$node2blwise_gstat}, $bl+1) : ({}, 0) ;

	# %{$br2indels} = ($branch_id => \@list_indels, ...) for the set of indels, with
	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices in @{$blocks_in_cls};
	# %node2blwise_gstat = ($node_id => \@blwise_gstat, ...) for all nodes,
	# with $blwise_gstat[$b] = $char_pres/abs being the state in the $b th block in the cluster;
	# $char_abs/pres is the character representing the 'absence'/'presence' state;
	# %br2l/rflank_event = ($br_id => [$stat_ue_l/rfl, $stat_le_l/rfl], ...)
	# is a map of left/right-flanking indel events onto the branches;


		## Finally, return the set of parsimonious scenarios with the null block ($bl) incorporated. ##

	if (@set_expanded_psm_scns>0) { # This is the normal ending. #

	    my $ct_exp_psm_scns = @set_expanded_psm_scns;
    (print STDOUT "\n  < Subroutine 'incorporate_null_block' Normally Ended (with $ct_exp_psm_scns expanded scenarios, each with an additional event) />\n\n") if ($TEST_LEVEL2>0);

	    return (1, \@set_expanded_psm_scns);

	} else { # Somehow, there is no successfully expanded parsimonious scenarios. #

	    print STDOUT "[WARNING 'incorporate_null_block'] Somehow, there was NO EXPANDED PARSIMONIOUS SCENARIO with zero or one additional indel event, when trying to incorporate the null block '$bl': $! !\n";

    (print STDOUT "\n  < Subroutine 'incorporate_null_block' Ended (withOUT any useful expanded scenarios) />\n\n") if ($TEST_LEVEL2>0);

	    return (1, undef);
	}

}	# End of the subroutine 'incorporate_null_block' #


# enum_psm_indel_scns_in_cls_blks:
#	This subroutine enumerates the most parsimonious scenarios of insertions/deletions
#	in a cluster of contiguous blocks, each consisting of sites with an identical presence/absence pattern.
#
#	REVISED on May 13 (Sun), 2012:
#	Made the format for the input set of indel blocks a bit more general,
#	so make the subroutine more flexible.
#	(This revision accompanied the revision of 'create_initial_set_indels'.)
#
#	REVISED on Nov 13 (Wed), 2013:
#	Now, the subroutine can correctly handle trees with two-branch nodes.
#	(This revision accompanied the revisions of 'create_list_to_be_examined' and 'recede_absence'.)
#
#	REVISED on Mar 8 (Sat), 2013:
#	Now, a new satellite subroutine, 'enum_psm_scns_gen_null_blk', was introduced, which can correctly enumerates
#	the parsimonious indel scenarios that generates a cluster consisting only of a single null-block.
#
# INPUT: ($top_node_id, \%node2pa, \%node2ch, \%node2ct_exoffs, $char_pres, $char_abs, \@blocks_in_cls, $indx_indx_gp, \@set_node2dparsstat, \@set_br2change, $fl_full), where
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#
#	@blocks_in_cls represents a cluster of contiguous blocks,
#	each consisting of sites with a particular gap-pattern, where
#	$indx_gp = $blocks_in_cls[$b]->[$indx_indx_gp] (or $blocks_in_cls[$b] itself if $indx_indx_gp is undefined) is the index specifying the gap pattern in @set_node2parsstat and @set_br2change;
#
#	\@set_node2parsstat and \@set_br2change are output of 'map_parsim_states_dollo_msa'
#	on the 'secondary MSA', which is a concatenated set of 1-column representations of the gap patterns, where
#	%{$set_node2dparsstat[$k]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $k th gap pattern onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%{$set_br2change[$k]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches (under the Dollo parsimony) for the $k th gap pattern, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node;
#
#	$fl_full = 1 if you want the full output,
#		 = 0 if you want a simpler output.
#
# OUTPUT: ($min_totct_indels, \@set_psm_scns_in_cls), where
#	$min_totct_indels is the total number of insertion/deletion events in the most parsimonious scenario(s);
#	@{$set_psm_scns_in_cls[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th parsimonious scenario, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls)
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$b] being $char_pres/abs for the $b th block (in the cluster) at the node $node_id.
#
#	If $fl_full = 0, %node2gapstat is given only at the top node.
#
sub enum_psm_indel_scns_in_cls_blks ($\%\%\%$$\@$\@\@$) {

	my ($topnode_id, $node2pa, $node2ch, $node2ct_exoffs, $char_pres, $char_abs, $blocks_in_cls, $indx_indx_gp, $set_node2dparsstat, $set_br2change, $fl_full) = @_;

	my $ct_seqs = $node2ct_exoffs->{$topnode_id};

    (print STDOUT " << Subroutine 'enum_psm_indel_scns_in_cls_blks' (#{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL1>0); 
    if ($TEST_LEVEL1>3) {
	my $size_set_gpatterns = @{$set_node2dparsstat};
	my @indices_gp = ();
	print STDOUT "[ Input data ]\n\n";
	print STDOUT '$topnode_id = ', $topnode_id, ".\n";
	print STDOUT '$char_pres/$char_abs = ', "$char_pres/$char_abs.\n"; 
	print STDOUT 'Size of @{$set_node2dparsstat} (and of @{$set_br2change}) = ', $size_set_gpatterns, "\n\n";
	print STDOUT 'Indices of gap patterns in the current cluster of blocks (in positional order):', "\n\n";
	print STDOUT join ("\t", 'Index_in_@{$blocks_in_cls}', 'Index_in_@{$set_node2dparsstat}'), "\n\n";
	for (my $bl=0; $bl<@{$blocks_in_cls}; $bl++) {
	    my $block = $blocks_in_cls->[$bl];
	    my $indx_gp = (defined $indx_indx_gp) ? $block->[$indx_indx_gp] : $block;
	    print STDOUT join ("\t", $bl, $indx_gp), "\n";
	}
	print STDOUT "\n\n";
    }

		# Create a initial 'tentative candidate' of the parsimonious indel scenario. #

	# %{$node2blwise_gstat0} = ($node_id => \@blwise_gstat0, ...) for all nodes,
	# with $blwise_gstat0[$b] = $char_pres/abs being the state in the $b th block in the cluster;
	# %{$br2indels0} = ($branch_id => \@list_indels0, ...) for the initial set of indels, with
	# @{$list_indels0[$id]} = ($stat_parent, $stat_node, \@indices_blk0) describing the $id th initial indel event at the branch, with @indices_blk0 listing the indices in @{$blocks_in_cls};
	# @{$list_blank_blks} lists the indices of the blocks consisting only of $char_abs.
	#
	my ($ct_indels0, $node2blwise_gstat0, $br2indels0, $list_blank_blks) = create_initial_set_indels ($topnode_id, %{$node2pa}, $char_pres, $char_abs, @{$blocks_in_cls}, $indx_indx_gp, @{$set_node2dparsstat}, @{$set_br2change});
	unless (defined $ct_indels0) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'create_initial_set_indels' FAILED : $! !\n";
	    return (undef, undef);
	}
    (print STDOUT "\n --- Subroutine 'create_initial_set_indels' Ended (with #{initial indels} = $ct_indels0) ...\n\n") if ($TEST_LEVEL1>1);

    if ($TEST_LEVEL1>2) {
	print STDOUT "[ Output of 'create_initial_set_indels' (#{initial indels} = $ct_indels0) ]\n\n";

	print STDOUT 'Content of %{$node2blwise_gstat0}:', "\n\n";
	print STDOUT join ("\t", 'Node_ID', "'initial_block-wise_gap_state(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat0}) {
	    my $cnct_gstat0 = join ('', @{$node2blwise_gstat0->{$node}});
	    print STDOUT join ("\t", $node, "'$cnct_gstat0'"), "\n";
	}
	print STDOUT "\n";

	print STDOUT 'Content of %{$br2indels0}:', "\n\n";
	print STDOUT join ("\t", 'Branch_ID', '#{exteior_offsprings}', 'State_ue/State_le/indices,in,@{$blocks_in_cls}', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) {
	    my $ct_exoffs = $node2ct_exoffs->{$br};
	    my @out_list_indels0 = ();
	    foreach my $indel (@{$br2indels0->{$br}}) {
		my ($stat_ue, $stat_le, $indices_blk0) = @{$indel};
		my $cnct_indices_blk0 = join (',', @{$indices_blk0});
		push @out_list_indels0, join ('/', $stat_ue, $stat_le, $cnct_indices_blk0);
	    }
	    print STDOUT join ("\t", $br, $ct_exoffs, @out_list_indels0), "\n";
	}
	print STDOUT "\n";

	print STDOUT '@{$list_blank_blks} = (', join (', ', @{$list_blank_blks}), ").\n\n";
    }

	my $ct_blank_blks = @{$list_blank_blks}; 


		# Handle an exceptional case,
		# where the gap-block cluster consists only of a null-block (i.e., a blank block). #
		# (ADDED in revision on Mar 8 (Sat), 2014.) #

	if (($ct_blank_blks == 1) and (@{$blocks_in_cls} == 1) and ($FL_HANDLE_NULL_CLS)) {
	    print STDOUT "[WARNING 'enum_psm_indel_scns_in_cls_blks'] This cluster consists ONLY of a single NULL-BLOCK: $! !\n";

	    my $indx_null_blk = 0;
	    my ($min_totct_indels, $set_psm_scns_null_blk) = enum_psm_scns_gen_null_blk ($topnode_id, %{$node2pa}, %{$node2ch}, $char_pres, $char_abs, $indx_null_blk, $fl_full);
	    unless (defined $set_psm_scns_null_blk) {
		print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'create_list_to_be_examined' FAILED: $! !\n";
		return (undef, undef);
	    }

	    return ($min_totct_indels, $set_psm_scns_null_blk);
	}


		# Create a list of indel events to be examined #

	# @{$list_to_be_examined->[$k][$i]}
	# = ($branch_id, index in the list @{$br2indels{$branch_id}})
	# is for the $i th indel event that result in $k exterior nodes with $char_abs (for 0<$k<$ct_seqs).
	# @{$list_indels_to_be_examined[0]} should always be empty.
	#
		# MODIFIED on Nov 13 (Wed), 2013, to enable the subroutine to correctly handle two-branch nodes. #
	my $list_indels_to_be_examined0 = create_list_to_be_examined (%{$node2ch}, %{$node2ct_exoffs}, $ct_seqs, $char_abs, %{$br2indels0});
	##my $list_indels_to_be_examined0 = create_list_to_be_examined (%{$node2ct_exoffs}, $ct_seqs, $char_abs, %{$br2indels0});
	unless (defined $list_indels_to_be_examined0) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'create_list_to_be_examined' FAILED: $! !\n";
	    return (undef, undef);
	}
    (print STDOUT "\n --- Subroutine 'create_list_to_be_examined' Ended ...\n\n") if ($TEST_LEVEL1>1);

    if ($TEST_LEVEL1>2) {
	my $size_list_to_be_exmd = @{$list_indels_to_be_examined0};
	print STDOUT "[ Output of 'create_list_to_be_examined' (Size_list = $size_list_to_be_exmd) ]\n\n";
	print STDOUT 'Content of @{$list_indels_to_be_examined0}:', "\n\n";
	print STDOUT join ("\t", '#{Exterior $char_abs\'s resulting from the event}', 'indx/$branch_id/index_in_@{$br2indels{$branch_id}}', '...'), "\n\n";
	for (my $k=0; $k<$size_list_to_be_exmd; $k++) {
	    my $indels = $list_indels_to_be_examined0->[$k];
	    my $ct_indels=@{$indels};
	    my @out_list_indels = ();
	    for (my $i=0; $i<$ct_indels; $i++) {
		my ($br_id, $indx_indel) = @{$indels->[$i]};
		push @out_list_indels, join ('/', $i, $br_id, $indx_indel);
	    }
	    print STDOUT join ("\t", $k, @out_list_indels), "\n";
	}
	print STDOUT "\n\n";
    }


		# Initialize the variables/arrays to be used for the main process. #

	my $min_totct_indels0 = undef;
	my @set_psm_scns_in_cls0 = ();
	my @stack_scns_to_be_examined = ([$ct_indels0, $node2blwise_gstat0, $br2indels0, $list_indels_to_be_examined0]);

		# Now the main process begins. #
    (print STDOUT "\n{{ Beginning of the Main Process }}\n\n") if ($TEST_LEVEL1>1);
     my $iter_no = 0;

	while (my $scn_to_be_examined = shift @stack_scns_to_be_examined) { # Main outermost while-loop (over scenarios to be examined) #

	    my ($totct_indels, $node2blwise_gstat, $br2indels, $indels_to_be_examined) = @{$scn_to_be_examined};
	    my $lclmin_totct_indels = $totct_indels;

	if ($TEST_LEVEL1>1) {
	    $iter_no++;
	    my $ct_stacked = @stack_scns_to_be_examined;
	    print STDOUT "\n{{ The $iter_no th Iteration:  #{Stacked scenarios} = $ct_stacked\; Initial Min(Total#{indels}) = $lclmin_totct_indels }}\n\n";
	}

	    ##while (@{$indels_to_be_examined}>0) { # Main outer while-loop  #
	    ##while (@{$indels_to_be_examined}>2) { # Main outer while-loop (over the set of indel events resulting in the largest number of exterior nodes with $char_abs) #
	    while (@{$indels_to_be_examined}>1) { # Main outer while-loop (over the set of indel events that could possibly be receded). ## MODIFIED on Nov 13 (Wed), 2013. ##
		# @{$indels_to_be_examined->[0]} should always be empty,
		# [@{$indels_to_be_examined->[1]} contain only deletions along exteior branches (, which therefore cannot recede any further). ] ---  This logic does not hold any longer (as of Nov 13 (Wed), 2013).

		my $indx_to_be_examined = $#{$indels_to_be_examined};
		my $list_indels = $indels_to_be_examined->[$indx_to_be_examined];

		while (@{$list_indels}>0) { # Main middle while-loop (over the indels to be examined) # 

		    my $indx_exmd_indel = $#{$list_indels};
	if ($TEST_LEVEL1>1) {
	    (print STDOUT "\n{ To be examined: The $indx_exmd_indel th indel event of the $indx_to_be_examined th set of events. }\n\n") if ($TEST_LEVEL1>1);
	}

			## Newly ADDED on Sep 22, 2012, to make the subroutine more robust against unexpected upstream behaviors. ##
		    my $examined_event = $list_indels->[$indx_exmd_indel];
		    unless ((defined $examined_event) and (@{$examined_event}>0)) {
			print STDOUT "\n[WARNING 'enum_psm_indel_scns_in_cls_blks'] The $indx_exmd_indel th indel event of the $indx_to_be_examined th set of events, which should have been examined now, is UNDEFINED or EMPTY! So I will remove it...\n\n";
			pop @{$list_indels};
			next;
		    }

		    my ($diff_totct_indels, $new_node2blwise_gstat, $new_br2indels, $new_indels_to_be_examined) = recede_absence (%{$node2blwise_gstat}, %{$br2indels}, @{$indels_to_be_examined}, $char_abs, $char_pres, %{$node2pa}, %{$node2ch}, %{$node2ct_exoffs}, $ct_seqs);
		    unless (defined $diff_totct_indels) {
			print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'recede_absence' FAILED : $! !\n";
			return (undef, undef);
		    }
	(print STDOUT "\n --- Subroutine 'recede_absence' Ended ...\n\n") if ($TEST_LEVEL1>1);

		    if ($diff_totct_indels>0) {
			# Just dismiss the new scenario (, which may often be undefined). #
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " => Dismiss the new scenario.\n\n") if ($TEST_LEVEL1>1);
			next;
		    } elsif ($diff_totct_indels<0) { 
			# Replace the old scenario with the new scenario. #
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " => Replace the old scenario(s) with the new scenario.\n\n") if ($TEST_LEVEL1>1);

			$lclmin_totct_indels += $diff_totct_indels;
			$node2blwise_gstat = $new_node2blwise_gstat;
			$br2indels = $new_br2indels;
			$indels_to_be_examined = $new_indels_to_be_examined;

				## A BUG fixed (on Sep 23 (Sun), 2012). ##
			$indx_to_be_examined = $#{$indels_to_be_examined};
			$list_indels = $indels_to_be_examined->[$indx_to_be_examined];

		    } elsif (defined $new_node2blwise_gstat) {
			# Add the new scenario to the stack of scenarios to be examined. #
			##push @{$new_indels_to_be_examined}, $list_indels;
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " => Append the new scenario to the stack of scenarios to be examined.\n\n") if ($TEST_LEVEL1>1);
			push @stack_scns_to_be_examined, [$lclmin_totct_indels, $new_node2blwise_gstat, $new_br2indels, $new_indels_to_be_examined];
		    } else {
			# Do nothing, because the 'absence' state in question cannot recede any further. #
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " (AND", '%{$new_node2blwise_gstat}', " is UNDEFINED) => STOP receding the 'absence' in the event just examined.\n\n") if ($TEST_LEVEL1>1);
			next;
		    }

		}	# End of the main middle while-loop (over the indels to be examined) #

	    (print STDOUT "\n +++ End of the $indx_to_be_examined th set of events. +++\n\n") if ($TEST_LEVEL1>1);

	    pop @{$indels_to_be_examined}; # Remove the last entry, which has been emptied. #

	    } # End of the main outer while-loop (over the set of indel events resulting in the largest number of exterior nodes with $char_abs) #

	(print STDOUT "\n ~~~ The Set of parsimonious scenarios will be renewed ~~~\n\n") if ($TEST_LEVEL1>1);

	    my @cand_psm_scn = ($br2indels, $node2blwise_gstat); # A candidate parsimonious scenario. #

	    unless (defined $min_totct_indels0) {
		# Initialize the set of global parsimonious scenarios. #
	(print STDOUT 'INITIALIZE the set of parsimonious scenarios (with $min_totct_indels0 = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);

		@set_psm_scns_in_cls0 = (\@cand_psm_scn);
		$min_totct_indels0 = $lclmin_totct_indels;
		next;
	    }

		# Compare the local minimum with the current minimum. #

	    if ($lclmin_totct_indels < $min_totct_indels0) {
		# Replace the set of parsimonious scenarios. #
	(print STDOUT 'REPLACE the OLD parsimonious scenario(s) with the NEW one (with $min_totct_indels0 = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);
		
		@set_psm_scns_in_cls0 = (\@cand_psm_scn);
		$min_totct_indels0 = $lclmin_totct_indels;

	    } elsif ($lclmin_totct_indels > $min_totct_indels0) {
		# Just discard the candidate parsimonious scenario. #

	(print STDOUT 'DISMISS the NEW parsimonious scenario (with $lclmin_totct_indels = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);

		next;

	    } else {
		# Append the candidate parsimonious scenario. (Keep the total#indels as it is.) #
	(print STDOUT 'APPEND the NEW parsimonious scenario to the OLD set of scenario(s) (with $min_totct_indels0 = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);

		push @set_psm_scns_in_cls0, \@cand_psm_scn;
	    }

	}	# End of the main outermost while-loop (over scenarios to be examined) #


    (print STDOUT "\n = The Main Process Ended =\n\n") if ($TEST_LEVEL1>0);

    if ($TEST_LEVEL1>1) {
	print STDOUT "Min(Tot#{indels}) = $min_totct_indels0.\n\n";
    }
    if ($TEST_LEVEL1>2) {
	my $ct_psm_scns_in_cls0 = @set_psm_scns_in_cls0;
	print STDOUT "[ Interim output (Set of $ct_psm_scns_in_cls0 parsimonious scenarios before incorporating 'null' blocks) ]\n\n";
	print STDOUT 'Content of @set_psm_scns_in_cls0:', "\n\n";

	for (my $ps=0; $ps<$ct_psm_scns_in_cls0; $ps++) { # Outermost testing for-loop #
	    print STDOUT "{ The $ps th scenario }\n\n";

	    my ($br2indels, $node2blwise_gstat) = @{$set_psm_scns_in_cls0[$ps]};

	    print STDOUT 'Content of %{$node2blwise_gstat}:', "\n\n";
	    print STDOUT join ("\t", 'Node_ID', "'initial_block-wise_gap_state(concatenated)'"), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat}) {
		my $cnct_gstat = join ('', @{$node2blwise_gstat->{$node}});
		print STDOUT join ("\t", $node, "'$cnct_gstat'"), "\n";
	    }
	    print STDOUT "\n";

	    print STDOUT 'Content of %{$br2indels}:', "\n\n";
	    print STDOUT join ("\t", 'Branch_ID', '#{exteior_offsprings}', 'State_ue/State_le/indices,in,@{$blocks_in_cls}', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) {
		my $ct_exoffs = $node2ct_exoffs->{$br};
		my @out_list_indels = ();
		foreach my $indel (@{$br2indels->{$br}}) {
		    (@{$indel}>0) or next; # Skip an empty entry. (ADDED to fix a BUG on May 13 (Sun), 2012) #
		    my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
		    my $cnct_indices_blk = join (',', @{$indices_blk});
		    push @out_list_indels, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		}
		print STDOUT join ("\t", $br, $ct_exoffs, @out_list_indels), "\n";
	    }
	    print STDOUT "\n\n";

	}			# End of the outermost testing for-loop #

    }

		# Incorporate the "null" blocks, each of which consists only of $char_abs. #

	if ($FL_IGNORE_NULL_COLUMNS) { ## (This "if (...) {...} else {...}" block was ADDED in revision on Nov 15 (Fri), 2013). ##

    (print STDOUT " = Although the MSA has $ct_blank_blks 'null' blocks, we will just leave them untouched, because the flag, ", '$FL_IGNORE_NULL_COLUMNS ', "(=$FL_IGNORE_NULL_COLUMNS), tells me to do so. =\n\n") if ($TEST_LEVEL1>0);
    (print STDOUT "   --- But I will anyway clean up the ", scalar (@set_psm_scns_in_cls0), " raw parsimonious scenarios obtained so far ...\n\n\n") if ($TEST_LEVEL1>1);

	} else {

    (print STDOUT " = NOW, let us incorporate the $ct_blank_blks 'null' blocks =\n\n") if ($TEST_LEVEL1>0);
    if ($TEST_LEVEL1>1) {
	print STDOUT '"Null" blocks = (', join (', ', @{$list_blank_blks}), ").\n\n";
	print STDOUT '#{Parsimonious scenarios to accomodate the null blocks} = ', scalar (@set_psm_scns_in_cls0), ".\n";
	print STDOUT "\n\n";
    }

	} ## END of "if ($FL_IGNORE_NULL_COLUMNS) {...} else {...}".

    my $indx_psm_scn0 = 0;

	my @set_psm_scns_in_cls = ();
	my $min_totct_indels = undef;

	while (my $psm_scn0 = shift @set_psm_scns_in_cls0) { # Second main outer while-loop (over 'initial' parsimonious scenarios) #

	(print STDOUT "\n[[ The $indx_psm_scn0 th 'initial' parsimonious scenario (#{indels} = $min_totct_indels0) ]]\n\n") if ($TEST_LEVEL1>1);

	    unless ( cleanup_br2indels (%{$psm_scn0->[0]}) ) {
		print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'cleanup_br2indels' FAILED on the $indx_psm_scn0 th scenario: $! !\n";
		return (undef, undef);
	    }
	(print STDOUT "\n --- Subroutine 'cleanup_br2indels' ended ...\n\n")  if ($TEST_LEVEL1>1);


	    if ($FL_IGNORE_NULL_COLUMNS) { ## Skip the process to incorporate null blocks (ADDED in revision on Nov 15 (Fri), 2013). ##
	$indx_psm_scn0++;
		push @set_psm_scns_in_cls, $psm_scn0;
		(defined $min_totct_indels) or ($min_totct_indels = $min_totct_indels0);
		next;
	    }

	    if ($ct_blank_blks==0) {
	(print STDOUT " ... APPEND the 'initial' scenario to the final set of parsimonious scenarios, bcause there are NO 'null' blocks to be incorporated ...\n\n") if ($TEST_LEVEL1>1);
	$indx_psm_scn0++;
		push @set_psm_scns_in_cls, $psm_scn0;
		(defined $min_totct_indels) or ($min_totct_indels = $min_totct_indels0);
		next;
	    }

	    my $lcl_min_totct_indels = $min_totct_indels0;
	    my @set_lcl_psm_scns = ($psm_scn0);

	    for (my $indx_bblk=0; $indx_bblk<$ct_blank_blks; $indx_bblk++) { # Middle for-loop (over blank blocks) #
		my $bbl = $list_blank_blks->[$indx_bblk];
		my $ct_lcl_psm_scns = @set_lcl_psm_scns;

	(print STDOUT "\n[ The $indx_bblk th blank block (the $bbl th in the original set of blocks) ; #{accomodating parsimonious scenarios} = $ct_lcl_psm_scns ]\n\n") if ($TEST_LEVEL1>1);

		my @new_lclset_psm_scns = ();
		my $indx_lcl_psm_scn = 0;
		my $mindiff_totct_indels2;

		while (my $lcl_psm_scn = shift @set_lcl_psm_scns) { # Inner while-loop (over local parsimonious scenarios) #

		    $indx_lcl_psm_scn++;
	(print STDOUT "\n{ The $indx_lcl_psm_scn th accomodating parsimonious scenario }\n\n") if ($TEST_LEVEL1>1);
	

			## Incorporate the $b th blank block into $lcl_psm_scn. ##

		    my ($diff_totct_indels2, $set_expanded_psm_scns) = incorporate_null_block ($bbl, @{$lcl_psm_scn}, $char_abs, $char_pres, $topnode_id, %{$node2pa}, %{$node2ch}, $mindiff_totct_indels2); 
		    unless (defined $diff_totct_indels2) {
			print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'incorporate_null_block' FAILED to incorporate the $bbl th block into the $indx_lcl_psm_scn th accomodating parsimonious scenario: $! !\n";
			return (undef, undef);
		    }
	(print STDOUT " --- Subroutine 'incorporate_null_block' ended ...\n\n") if ($TEST_LEVEL1>1);

		    unless (defined $set_expanded_psm_scns) {
			# ALL 'expanded' scenarios had $diff_totct_indels2 > $mindiff_totct_indels2. #
	(print STDOUT '$diff_totct_indels2 = ', $diff_totct_indels2, ' (> $mindiff_totct_indels2 = ', $mindiff_totct_indels2, ") => DISMISS the new parsimonious scenarios.\n\n") if ($TEST_LEVEL1>1);
			next;
		    }

		    if ((defined $mindiff_totct_indels2) and ($mindiff_totct_indels2 == $diff_totct_indels2)) {
	(print STDOUT '$diff_totct_indels2 = ', $diff_totct_indels2, ' (= $mindiff_totct_indels2 )', " => APPEND the new parsimonious scenarios.\n\n") if ($TEST_LEVEL1>1);
			push @new_lclset_psm_scns, @{$set_expanded_psm_scns};
		    } else {
			   # Either $mindiff_totct_indels2 = undef or $mindiff_totct_indels2 > $diff_totct_indels2. #
	if ($TEST_LEVEL1>1) {
	    if (defined $mindiff_totct_indels2) {
		print STDOUT '$diff_totct_indels2 = ', $diff_totct_indels2, ' (< $mindiff_totct_indels2 = ', $mindiff_totct_indels2, ") => REPLACE the OLD parsimonious scenarios with the NEW ones.\n\n";
	    } else {
		print STDOUT '$diff_totct_indels2 = ', $diff_totct_indels2, ' (with $mindiff_totct_indels2 UNDEFINED)', " => INITIALIZE the set of parsimonious scenarios.\n\n";
	    }
	}
			$mindiff_totct_indels2 = $diff_totct_indels2;
			@new_lclset_psm_scns = @{$set_expanded_psm_scns};
		    }

		}		# End of the inner while-loop (over local parsimonious scenarios) #
		@set_lcl_psm_scns = @new_lclset_psm_scns; 
		$lcl_min_totct_indels += $mindiff_totct_indels2;

	if ($TEST_LEVEL1>1) {
	    print STDOUT "\n +++ Ended the attempts to incorporate the $indx_bblk th blank block  (= the $bbl th block) +++\n\n";
	    print STDOUT 'Min[diff(Tot#{indels})] = ', $mindiff_totct_indels2, ".\n";
	    print STDOUT 'Current min(Tot#{indels}) = ', $lcl_min_totct_indels, ".\n\n";
	}

	    }			# End of the middle for-loop (over blank blocks) #

	(print STDOUT "\n ~~~ ENDED the attempts to expand the $indx_psm_scn0 th 'initial' parsimonious scenario ~~~\n\n") if ($TEST_LEVEL1>1);

	$indx_psm_scn0++;

	    unless (defined $min_totct_indels) {
			# Initialize the full set of parsimonious scenarios. #
	(print STDOUT 'INITIALIZE @set_psm_scns_in_cls  (with $min_totct_indels = ', $lcl_min_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);
		$min_totct_indels = $lcl_min_totct_indels;
		@set_psm_scns_in_cls = @set_lcl_psm_scns;
		next; ## ADDED to fix a BUG on May 14 (Mon), 2012. ##
	    }

	    if ($lcl_min_totct_indels < $min_totct_indels) {
			# Replace the full set of parsimonious scenarios. #
	(print STDOUT 'REPLACE the OLD @set_psm_scns_in_cls (with $min_totct_indels = ', $min_totct_indels, ') WITH the NEW ONE (with $min_totct_indels = ', $lcl_min_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);
		$min_totct_indels = $lcl_min_totct_indels;
		@set_psm_scns_in_cls = @set_lcl_psm_scns;

	    } elsif ($lcl_min_totct_indels > $min_totct_indels) {
			# Just discard the local parsimonious scenarios. #
	(print STDOUT 'DISCARD the NEW @set_lcl_psm_scns (with $lcl_min_totct_indels = ', $lcl_min_totct_indels, ' > $min_totct_indels = ', $min_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);
		next;

	    } else {
			# Append the local scenarios to the full set. #
	(print STDOUT 'APPEND the NEW @set_lcl_psm_scns (with $lcl_min_totct_indels = ', $lcl_min_totct_indels, ' = $min_totct_indels).', "\n\n") if ($TEST_LEVEL1>1);
		push @set_psm_scns_in_cls, @set_lcl_psm_scns;
	    }

	}	# End of the second main outer while-loop (over 'initial' parsimonious scenarios) #

    (print STDOUT "\n = Finished icorporating the 'null' blocks =\n\n") if ($TEST_LEVEL1>0);

    (print STDOUT "\n << Subroutine 'enum_psm_indel_scns_in_cls_blks' Ended />>\n\n") if ($TEST_LEVEL1>0);

	return ($min_totct_indels, \@set_psm_scns_in_cls);

}	# End of the subroutine 'enum_psm_indel_scns_in_cls_blks' #


# extract_cls_gblocks:
#	This subroutine extracts, from an input MSA,
#	clusters of neighboring blocks, each of which consists
#	of contiguous columns of a particular gap pattern.
#	The clusters are delimited by the, 'all-presence' columns,
#	i.e. columns containing NO gaps.
#
#	NOTE that this subroutine eats up the input MSA.
#
# INPUT: (\@msa, $char_pres, $char_abs, \%gpattern2indx, \$ct_gpattens), where
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#	%gpattern2indx = ($gpattern => $indx_gp, ...) for the
#	gap patterns encountered so far, with $indx_gp assigned in the order of occurrences;
#	$ct_gpatterns is the number of gap pattens encountered so far.
#
# OUTPUT: \@set_cls_gblocks, where
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention.
#
sub extract_cls_gblocks (\@$$\%\$) {

	my ($msa, $char_pres, $char_abs, $gpattern2indx, $ct_gpatterns) = @_;

	my $ct_seqs = @{$msa};
    (print STDOUT "<< Subroutine 'extract_cls_gblocks' (#{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL1>0);

	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'extract_cls_gblocks'] NO SEQUENCE in the input MSA: $! !\n";
	    return undef;
	}
	my $len_char_abs = length $char_abs;
	unless ($len_char_abs>0) {
	    print STDOUT "[ERROR 'extract_cls_gblocks'] The 'absence' character (= '$char_abs') is NULL: $! !\n";
	    return undef;
	}
	my $len_char_pres = length $char_pres;
	unless ($len_char_pres == $len_char_abs) {
	    print STDOUT "[ERROR 'extract_cls_gblocks'] The length (=$len_char_abs) of the 'absence' character, '$char_abs', CLASHES with that (=$len_char_pres) of the 'presence' character, '$char_pres': $! !\n";
	    return undef;
	}
	my $len_seq  = length ${$msa->[0]};
	unless ($len_seq>0) {
	    print STDOUT "[ERROR 'extract_cls_gblocks'] We do NOT HANDLE NULL MSAs: $! !\n";
	    return undef;
	}
	my $ct_clms = int ($len_seq / $len_char_abs);
	unless ($ct_clms>0) {
	    print STDOUT "[ERROR 'extract_cls_gblocks'] The MSA has NO full column: $! !\n";
	    return undef;
	}

		### Main Process ###

	my $if_in_cluster = 0;
	my @set_cls_gblocks = ();

	for (my $l=0; $l<$ct_clms; $l++) { # The main outer for-loop (over columns in the MSA) #

	    my $gpattern = '';
	    my $ct_abs = 0;
	    for (my $k=0; $k<$ct_seqs; $k++) { # The main inner for-loop (over sequences in the MSA) #
		my $char = substr (${$msa->[$k]}, 0, $len_char_abs, '');
		if ($char eq $char_abs) {
		    $gpattern .= $char_abs;
		    $ct_abs++;
		} else {
		    $gpattern .= $char_pres;
		}
	    }			# End of the main inner for-loop (over sequences in the MSA) #

	    unless ($ct_abs) { # The column consists only of 'presence' states. #
		$if_in_cluster = 0;
		next;
	    }

		# Retrieve or assign an index to the gap pattern. #
	    my $indx_gp = $gpattern2indx->{$gpattern};
	    unless (defined $indx_gp) {
		$indx_gp = $gpattern2indx->{$gpattern} = ${$ct_gpatterns};
		${$ct_gpatterns}++;
	    }

		# @{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)

	    unless ($if_in_cluster) { # A new cluster of blocks begins. #
		$if_in_cluster = 1;
		push @set_cls_gblocks, [ [$indx_gp, 1, $l, $l] ];
		next;
	    }

	     my $last_cls_gblks = $set_cls_gblocks[$#set_cls_gblocks];
	     my $last_gblk = $last_cls_gblks->[$#{$last_cls_gblks}];
	     my $last_indx_gp = $last_gblk->[0];

	    if ($indx_gp == $last_indx_gp) {
			# The pattern is the same as that of the last block of columns. #
			# => Extend the last block. #

		unless ($l == $last_gblk->[3] + 1) {
		    print STDOUT "[ERROR 'extract_cls_gblocks'] Somehow, the last block (with the end coordinate ", $last_gblk->[3], ") is NOT the LEFT NEIGHBOR of the current column, wheras both have the following gap-pattern: $! !\n";
		    print STDOUT "    '$gpattern'.\n\n";
		    return undef;
		}
		$last_gblk->[1]++;
		$last_gblk->[3]++;

	    } else {
			# The gap pattern differs from that of the last block. #
			# => Create a new block. #

		push @{$last_cls_gblks}, [$indx_gp, 1, $l, $l];
	    }

	}			# End of the main outer for-loop (over columns in the MSA) #

    (print STDOUT "\n<< Subroutine 'extract_cls_gblocks' Ended />>\n\n") if ($TEST_LEVEL1>0);

	return \@set_cls_gblocks;

}	# End of the subroutine 'extract_cls_gblocks' #


# msa_rep_gpatterns:
#	This subroutine creates an MSA representation of gap patterns,
#	out of the hash %gpattern2indx.
#
# INPUT: (\%gpattern2indx, $len_stat, $ct_seqs), where
#	%gpattern2indx = ($gpattern => $indx_gp, ...) for the
#	gap patterns encountered so far, with $indx_gp assigned in the order of occurrences.
#	substr ($gpattern, $len_stat*$k, $len_stat) is the 'presence'/'absence' state of the
#	$k th sequence in the MSA specifying the gap pattern;
#
# OUTPUT: (\@msa_rep, \@set_gpatterns), where
#	substr (${$msa_rep[$k]}, $len_stat*$indx_gp, $len_stat) is the
#	state of the $k th sequence in the $indx_gp th gap pattern;
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap pattern.
#
sub msa_rep_gpatterns (\%$$) {

	my ($gpattern2indx, $len_stat, $ct_seqs) = @_;

	my $ct_gpatterns = scalar keys %{$gpattern2indx};

    (print STDOUT "\n<< Subroutine 'msa_rep_gpattenrs' (#{sequences} = $ct_seqs; #{gap patterns} = $ct_gpatterns; unit state length = $len_stat) >>\n\n") if ($TEST_LEVEL1>0);

	my @msa_rep = ();
	for (1..$ct_seqs) { my $init = ''; push @msa_rep, \$init; }
	
	($ct_gpatterns>0) or ( return (\@msa_rep, []) );

	my @set_gpatterns = ();

	foreach my $gpattern (keys %{$gpattern2indx}) { # 1st foreach-loop #
	    my $indx = $gpattern2indx->{$gpattern};
	    $set_gpatterns[$indx] = \$gpattern;
	}			# End of the 1st foreach-loop #

	for (my $indx = 0; $indx<$ct_gpatterns; $indx++) { # 2nd for-loop (over the gap patterns) #
	    my $rf_gp = $set_gpatterns[$indx];
	    unless (defined $rf_gp) {
		print STDOUT "[ERROR 'msa_rep_gpatterns'] The $indx th gap pattern is UNDEFINED: $! !\n\n";
		return (undef, undef);
	    }
		# Copy the string referenced by an element in @set_gpatterns. #
	    my $gpattern = ${$rf_gp};

	    my $len_gp = length $gpattern;
	    unless ($len_gp == $ct_seqs * $len_stat) {
		print STDOUT "[ERROR 'msa_rep_gpattens'] The $indx th gap pattern has the length = $len_gp, which does NOT EQUAL the product of #{sequences} (=$ct_seqs) and the unit state length (=$len_stat): $! !\n\n";
		return (undef, undef);
	    }

	    for (my $k=0; $k<$ct_seqs; $k++) { # Inner for-loop (over the sequences) #
		my $stat = substr ($gpattern, 0, $len_stat, '');
		${$msa_rep[$k]} .= $stat;
	    }		# End of the inner for-loop (over the sequences) #

	}			# End of the 2nd for-loop (over the gap pattens) #

    (print STDOUT "\n<< 'msa_rep_gpatterns' Ended />>\n\n") if ($TEST_LEVEL1>0);

	return (\@msa_rep, \@set_gpatterns);

}	# End of the subroutine 'msa_rep_gpatterns' #


# scan_gbcls_for_long_gaps:
#	This satellite subroutine of 'enum_psm_indel_scns_for_msa' scans a gap-block cluster
#	searching for gaps that are longer than a specified upper-bound.
#
# INPUT: (\@gblocks, \@set_gpatterns, $upper_bound_glen, $char_abs), where
#	@{$gblocks[$gb]} = ($indx_gp, $len_gb, $start_gb, $end_gb) is
#	for the $gb th block in the subject gap-block cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start_gb/$end_gb being the starting/ending coordinate of the block
#	in the 0-based convention;
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap pattern in the input MSA, created by concatenating $char_pres/abs in the sequences;
#	$upper_bound_glen is the upper-bound of the gap length;
#	$char_abs is the character representing an 'absence' state.
#
# OUTPUT: \@list_long_gaps, where
#	@{$list_long_gaps[$lg]} = ($indx_seq, $len, $start, $end, \@cntr_gblks), with
#	$indx_seq specifies the sequence with the gap,
#	$len is the length of the gap (= #{gap-characters in the run}),
#	$start/$end is the start/end coordinate of the gap,
#	@cntr_gblks stores the gap-blocks contributing to the gap.
#
sub scan_gbcls_for_long_gaps (\@\@$$) {

	my ($gblocks, $set_gpatterns, $up_bd_glen, $char_abs) = @_;

	my $ct_gblks = @{$gblocks};
	($ct_gblks>0) or (return []); # Return an empty list if the cluster is empty. #

	my $cl_start = $gblocks->[0][2];
	my $cl_end = $gblocks->[$ct_gblks - 1][3];
	my $cl_len = $cl_end - $cl_start + 1;
	($cl_len > $up_bd_glen) or (return []);  # Return an empty list if the cluster is NOT longer than the upper-bound.

    (print STDOUT "  < Subroutine 'scan_gbcls_for_long_gaps' (region = \[$cl_start\, $cl_end\] \; {cluster_length} = $cl_len \; #{gap-blocks} = $ct_gblks \; {upper-bounnd of gap-length} = $up_bd_glen \; \"absence\" = '$char_abs' >\n\n") if ($TEST_LEVEL_SCN>0);

	my @set_chwise_gpatterns = ();
	for (my $g=0; $g<$ct_gblks; $g++) { # Preliminary for-loop (over gap-blocks in the cluster). #
	    my $indx_gp = $gblocks->[$g][0];
	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    my @chwise_gpattern = split //, $gpattern;
	    $set_chwise_gpatterns[$g] = \@chwise_gpattern;
	}			# End of the preliminary for-loop (over gap-blocks in the cluster). #

	my $ct_seqs = @{$set_chwise_gpatterns[0]};

	my @list_long_gaps = ();

	for (my $s=0; $s<$ct_seqs; $s++) { # Main outer for-loop (over sequences in the MSA). #

	    my @gap_run = ();
	    for (my $g=0; $g<$ct_gblks; $g++) { # Main inner for-loop (over gap-blocks in the cluster). #

		my $char = $set_chwise_gpatterns[$g]->[$s];

		if ($char ne $char_abs) {
		##if ($char eq $char_abs) {...} # A BUG fixed on Mar 10 (Mon), 2014. #

			# If the block-wise state is 'presence'. #
		    if (@gap_run>0) {
			if ($gap_run[1] > $up_bd_glen) {
			    my @copy = @gap_run;
			    push @list_long_gaps, \@copy;
			}
			@gap_run = (); # Empty the gap-run. #
		    }
		    next;
		}

			# If the block-wise state is 'absence'. #

		my ($indx_gp, $len_gb, $start_gb, $end_gb) = @{$gblocks->[$g]};

		if (@gap_run>0) {
			# Update the gap-run. #
		    $gap_run[1] += $len_gb;
		    $gap_run[3] = $end_gb;
		    push @{$gap_run[4]}, $g;

		} else {
			# Initialize the gap-run. #
		    @gap_run = ($s, $len_gb, $start_gb, $end_gb, [$g]); 
		}

	    }				# End of the main inner for-loop (over gap-blocks in the cluster). #

	    if ((@gap_run>0) and ($gap_run[1] > $up_bd_glen)) {
		push @list_long_gaps, \@gap_run;
	    }

	}			# End of the main outer for-loop (over sequences in the MSA). #

    if ($TEST_LEVEL_SCN>2) {
	my $ct_lgps = @list_long_gaps;
	print STDOUT "\n  [ There are $ct_lgps long gaps ]\n\n";
	print STDOUT   "  ", join ("\t", 'No.', 'Indx_seq', 'Length', 'Start', 'End', 'contributing,gap,blocks'), "\n\n";

	for (my $lg=0; $lg<$ct_lgps; $lg++) { # Testing for-loop (over long-gaps). #
	    my ($indx_seq, $len, $start, $end, $gap_blocks) = @{$list_long_gaps[$lg]};
	    my $cnct_gblks = join (',', @{$gap_blocks});
	    print STDOUT "  ", join ("\t", $lg, $indx_seq, $len, $start, $end, $cnct_gblks), "\n";
	}			# End of the testing for-loop (over long-gaps). #
	print STDOUT "\n";
    }

    (print STDOUT "\n  < Subroutine 'scan_gbcls_for_long_gaps' ENDED />\n\n") if ($TEST_LEVEL_SCN>0);

	return \@list_long_gaps;

}	# End of the subroutine 'scan_gbcls_for_long_gaps'. #


# enum_psm_indel_scns_for_msa:
#	This is one of the main subroutines in this Perl module.
#	to enumerate all parsimonious indel scenarios
#	each of which realizes the input MSA.
#	It puts together several subroutines in this Perl module,
#	out of which 'enum_psm_indel_scns_in_cls_blks' and 'map_parsim_psm_dollo_msa'
#	play core roles,
#	and 'extract_cls_gblocks' and 'msa_rep_gpatterns'
#	perform necessary pre-processes of the input data.
#
#	NOTE that this subroutine eats up the input MSA.
#	
# INPUT: (\%node2ch, \%node2pa, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres), where
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA
#	(NOTE that these names must exactly match those used in %name2id);
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
# OUTPUT: (\@set_cls_gblocks, \@min_totcts_indels_for_cls, \@sets_psm_scns_for_cls,
#		\@set_gpatterns, \@set_node2dparsstat, \@set_br2change), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is
#	for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention.
#
#	$min_totcts_indels_for_cls[$cl] is the total number of insertion/deletion events in the most parsimonious scenario(s) realizing the $cl th cluster of gap-pattern-blocks;
#
#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id.
#
#	The above 3 arrays comprises the MAIN OUTPUT,
#	and the following 3 arrays are bi-produces,
#	which could be used for some other purposes (including debugging).
#
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap pattern in the input MSA, created by concatenating $char_pres/abs in the sequences;
#
#	\@set_node2dparsstat and \@set_br2change are output of 'map_parsim_states_dollo_msa'
#	on the 'secondary MSA', which is a concatenated set of 1-column representations of the gap patterns, where
#	%{$set_node2dparsstat[$k]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $k th gap pattern onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#
#	%{$set_br2change[$k]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches (under the Dollo parsimony) for the $k th gap pattern, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node;
#
#	$fl_full = 1 if you want the full output,
#		 = 0 if you want a simpler output.
#
sub enum_psm_indel_scns_for_msa (\%\%\%\@\@$$) {

	my ($node2ch, $node2pa, $name2id, $seqnames, $in_msa, $char_abs, $char_pres) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs == scalar (@{$in_msa})) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] ILLEGAL INPUT: The number of sequence names (=$ct_seqs) CLASHES with the number of sequences themselves (=", scalar (@{$in_msa}), ") in the input MSA: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

    (print STDOUT "<<< Subroutine 'enum_psm_indel_scns_for_msa' (Top node ID: $topnode_id\; #{sequences} = $ct_seqs) >>>\n\n") if ($TEST_LEVEL>0);


    (print STDOUT '/// The 1st Process (preliminary): Prepare %{$node2ct_exoffs} for the input tree ///', "\n\n") if ($TEST_LEVEL>1);

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] Subroutine 'make_node2abs_depth' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n--- Subroutine 'make_node2abs_depth' Ended ...\n\n") if ($TEST_LEVEL>1);

	my $node2ct_exoffs = make_node2ct_exoffs (%{$node2ch}, %{$node2depth});
	unless (defined $node2ct_exoffs) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] Subroutine 'make_node2ct_exoffs' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n--- Subroutine 'make_node2ct_exoffs' Ended ...\n\n") if ($TEST_LEVEL>1);

    if ($TEST_LEVEL>2) {
	print STDOUT '[ TEST 1: Output of the 1st Process (preliminary) ]', "\n\n";
	print STDOUT join ("\t", 'Node_ID', 'depth', '#{ext_offsprings}'), "\n\n";
	foreach my $node (sort { $node2depth->{$a} <=> $node2depth->{$b} } keys %{$node2depth}) {  # Auxiliary foreach-loop #
	    my $depth = $node2depth->{$node};
	    my $ct_exoffs = $node2ct_exoffs->{$node};
	    print STDOUT join ("\t", $node, $depth, $ct_exoffs), "\n";
	}				# End of the auxiliary foreach-loop #
	print STDOUT "\n\n";
    }

    (print STDOUT '/// The 2nd Process (main): Examine the gap pattern structure of the input MSA ///', "\n\n") if ($TEST_LEVEL>1);

	my %gpattern2indx;
	my $ct_gpatterns = 0;
	my $set_cls_gblocks = extract_cls_gblocks (@{$in_msa}, $char_pres, $char_abs, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The subroutine 'extract_cls_gblocks' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
	my $ct_cls = @{$set_cls_gblocks};
    (print STDOUT "\n --- 'extract_cls_gblocks' Ended (#{gap clusters} = $ct_cls) ...\n\n") if ($TEST_LEVEL>1);

	unless ($ct_gpatterns == scalar keys %gpattern2indx) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The number of gap patterns (=$ct_gpatterns) CLASHES with the number of keys in the hash '", '%gpattern2indx', "' (=", scalar (keys %gpattern2indx), "): $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

	# substr (${$sec_msa[$k]}, $indx_gp, 1) is the presence/absence state of the $k th sequence in the $indx_gp th gap pattern.
	# ${$set_gpatterns[$indx_gp]} is the string made by concatenating the states in the $indx_gp th gap pattern.
	#
	my ($sec_msa, $set_gpatterns) = msa_rep_gpatterns (%gpattern2indx, length ($char_abs), $ct_seqs);
	unless (defined $sec_msa) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The subroutine 'msa_rep_gpatterns' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n --- 'msa_rep_gpatterns' Ended ...\n\n") if ($TEST_LEVEL>1);

    if ($TEST_LEVEL>2) {
	print STDOUT "[[TEST2: Output of the 2nd Process (main) ]]\n\n";
	
	print STDOUT "[ The MSA representation of the gap patterns in the input MSA ]\n\n";
	unless ( pr_ODEN_input (@{$seqnames}, @{$sec_msa}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The subroutine 'pr_ODEN_input'  FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
	print STDOUT "\n\n[ List of gap patterns (in order of their first occurrences) ]\n\n";
	print STDOUT join ("\t", 'Indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	for (my $indx_gp = 0; $indx_gp<$ct_gpatterns; $indx_gp++) { # Auxiliary for-loop #
	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print STDOUT join ("\t", $indx_gp, "'$gpattern'"), "\n";
	} 					# End of the auxiliary for-loop #

	print STDOUT "\n\n[ Clusters of contiguous blocks with particular gap patterns (#{clusters} = $ct_cls) ]\n\n";

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer auxiliary for-loop (over clusters) #

	    my $cls_gblks = $set_cls_gblocks->[$cl];
	    my $ct_gblks = @{$cls_gblks};

	    print STDOUT ' ' x 3, "{ The $cl th cluster (#{gap-blocks} = $ct_gblks) }\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	    for (my $bl=0; $bl<$ct_gblks; $bl++) { # Inner for-loop (over the gap-blocks) #
		my ($indx_gp, $len_gb, $start, $end) = @{$cls_gblks->[$bl]};

		my $gpattern = ${$set_gpatterns->[$indx_gp]};
		print STDOUT ' ' x 3, join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
	    }			# End of the inner for-loop (over the gap-blocks) #
	    print STDOUT "\n\n";
	}		# End of the outer auxiliary for-loop (over clusters) #

    }		# End of the test output for the 2nd process. #


		## Handle a case with NO gap-block clusters. (ADDED on Sep 22 (Sat), 2012) ##

	## $ct_cls = @{$set_cls_gblocks};
	if ($ct_cls == 0) {
	    print STDOUT "\n[WARNING 'enum_psm_indel_scns_for_msa'] NO CLUSTERS of gap-blocks were extracted from the input MSA, so the MSA should involve no indels...\n\n";
	    my @min_totcts_indels_for_cls = my @sets_psm_scns_for_cls = ();
	    my @set_node2dparsstat = my @set_br2change = ();
	    return ($set_cls_gblocks, \@min_totcts_indels_for_cls, \@sets_psm_scns_for_cls, $set_gpatterns, \@set_node2dparsstat, \@set_br2change);
	}


    (print STDOUT '/// The 3rd Process (main): Assign the Dollo parsimonious scenario to each of the $ct_gpatterns gap patterns ///', "\n\n") if ($TEST_LEVEL>1);

	my ($set_node2dparsstat, $set_br2change) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	unless (defined $set_node2dparsstat) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The subroutine 'map_parsim_states_dollo_msa' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT " --- 'map_parsim_states_dollo_msa' Ended ...\n\n") if ($TEST_LEVEL>1);

	unless ($ct_gpatterns == @{$set_node2dparsstat}) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The number of maps of Dollo parsimonious states onto nodes (= ", scalar (@{$set_node2dparsstat}), ") CLASHES with that of gap patterns (= $ct_gpatterns): $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}
	unless ($ct_gpatterns == @{$set_br2change}) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The number of maps of Dollo parsimoinous indel events onto branches (= ", scalar (@{$set_br2change}), ") CLASHES with that of gap patterns (= $ct_gpatterns): $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

    if ($TEST_LEVEL>2) {
	print STDOUT "\n[[ Output of the 3rd process (main) ]]\n\n";
	print STDOUT '[ Contents of @{$set_node2dparsstaat} and @{$set_br2change}) ]', "\n\n";

	for (my $indx_gp=0; $indx_gp < $ct_gpatterns; $indx_gp++) { # Outer auxiliary for-loop #
	    my $node2dparsstat = $set_node2dparsstat->[$indx_gp];
	    my $br2change = $set_br2change->[$indx_gp];

	    print STDOUT ' ' x 3, "{ The $indx_gp th gap pattern }\n\n";
	    print STDOUT ' ' x 3, '( Map of Dollo parsimonious states onto nodes )', "\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Node_ID', 'State', '#{children with #{exterior offsprings with \'presence\'}>0}'), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2dparsstat}) { print STDOUT ' ' x 3, join ("\t", $node, @{$node2dparsstat->{$node}}), "\n"; }
	    print STDOUT "\n";
	    print STDOUT ' ' x 3, '( Map of indel events onto branches )', "\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Branch_ID', 'state(upper-end)', 'state(lower-end)'), "\n\n"; 
	    foreach my $br (sort {$a <=> $b} keys %{$br2change}) { print STDOUT ' ' x 3, join ("\t", $br, @{$br2change->{$br}}), "\n"; }
	    print STDOUT "\n\n";
 
	}			# End of the outer auxiliary for-loop #
    }

    (print STDOUT '/// The 5th Process (main): Enumerate parsimonious indel scenarios for each of the ', $ct_cls, ' clusters of gap-pattern blocks  ///', "\n\n") if ($TEST_LEVEL>1);

	my $fl_full_enum = 1; # = 1 if you want the parsimonious scenarios to be fully enumerated, = 0 otherwise.
	my $indx_indx_gp = 0; # Index to specify where the index of the gap-pattern is stored.

	my @min_totcts_indels_for_cls = my @sets_psm_scns_for_cls = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main outermost for-loop (over the clusters) #

	    my $blocks_in_cls = my $cls_gblks = $set_cls_gblocks->[$cl];
	    my $ct_gblks = @{$cls_gblks};

    (print STDOUT ' ' x 3, "{ The $cl th cluster (#{gap-blocks} = $ct_gblks) }\n\n") if ($TEST_LEVEL>1);

    if ($TEST_LEVEL>3) {
	print STDOUT ' ' x 3, "( Gap-pattern blocks in the cluster )\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	for (my $bl=0; $bl<$ct_gblks; $bl++) { # Inner auxiliary for-loop (over the gap-blocks) #
	    my ($indx_gp, $len_gb, $start, $end) = @{$cls_gblks->[$bl]};

	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print STDOUT ' ' x 3, join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
	}			# End of the inner auxiliary for-loop (over the gap-blocks) #
	print STDOUT "\n\n";
    }

		# Before advancing to the core sub-routine, 'enum_psm_indel_scns_in_cls_blks',
		# screen the gap-block cluster with its length itself.
		# (ADDED in revision on Mar 11 (Tue), 2014.) #

	my $cl_start = $cls_gblks->[0][2];
	my $cl_end = $cls_gblks->[$ct_gblks - 1][3];
	my $cl_len = $cl_end - $cl_start + 1;

	unless ($cl_len <= $UPPER_BOUND_ANAL_LEN_GBCLS) {
	    print STDOUT "[WARNING 'enum_psm_indel_scns_for_msa'] The $cl th cluster is of length $cl_len, which is LONGER THAN the upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS... I will skip analyzing this cluster: $! !\n\n";
	    $min_totcts_indels_for_cls[$cl] = 'N/A';
	    $sets_psm_scns_for_cls[$cl] = []; # Assign an empty set.
	    next;
	}

		# Before advancing to the core sub-routine, 'enum_psm_indel_scns_in_cls_blks',
		# scan the gap-block cluster for long gaps.
		# (ADDED in revision on Mar 10 (Mon), 2014.) #

	my $list_long_gaps = scan_gbcls_for_long_gaps (@{$cls_gblks}, @{$set_gpatterns}, $UPPER_BOUND_ANAL_GAPLEN, $char_abs);
	unless (defined $list_long_gaps) {
	     print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The subroutine 'scan_gbcls_for_long_gaps' FAILED on the $cl th cluster: $! !\n\n";
	     return (undef, undef, undef, undef, undef, undef);
	}
	unless (@{$list_long_gaps} == 0) {
	    my $ct_long_gaps = @{$list_long_gaps};
	    print STDOUT "[WARNING 'enum_psm_indel_scns_for_msa'] The $cl th cluster contains $ct_long_gaps gaps longer than $UPPER_BOUND_ANAL_GAPLEN. So, I will skip analyzing this cluster: $! !\n\n"; 
	    $min_totcts_indels_for_cls[$cl] = 'N/A';
	    $sets_psm_scns_for_cls[$cl] = []; # Assign an empty set.
	    next;
	}

	# INPUT: (\%node2ch, \%node2pa, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres),
	# my ($node2ch, $node2pa, $name2id, $seqnames, $in_msa, $char_abs, $char_pres) = @_;

	    my ($min_totct_indels, $set_psm_scns) = enum_psm_indel_scns_in_cls_blks ($topnode_id, %{$node2pa}, %{$node2ch}, %{$node2ct_exoffs}, $char_pres, $char_abs, @{$blocks_in_cls}, $indx_indx_gp, @{$set_node2dparsstat}, @{$set_br2change}, $fl_full_enum); 
	    unless (defined $min_totct_indels) {
	 	print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The CORE subroutine 'enum_psm_indel_scns_in_cls_blks' FAILED on the $cl th cluster: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef);
	    }
	    my $ct_scns = @{$set_psm_scns};

    (print STDOUT "\n --- 'enum_psm_indel_scns_in_cls_blks' Ended (Min#{indels} = $min_totct_indels, #{parsimonious scenarios} = $ct_scns) ...\n\n") if ($TEST_LEVEL>1);

    if ($TEST_LEVEL>3) {
	print STDOUT ' ' x 3, "[ TEST: Output of 'enum_psm_indel_scns_in_cls_blks' ]\n\n";
	for (my $ps=0; $ps<$ct_scns; $ps++) { # Axiliary outer for-loop (over parsimonious scenarios) #

	    print STDOUT ' ' x 3, "| The $ps th scenario |\n\n";

	    my ($br2indels, $node2bw_gstats) =  @{$set_psm_scns->[$ps]};

	    print STDOUT ' ' x 3, "( Map of block-wise gap states onto the nodes )\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Node_ID', "'block_wise_gap_states(concatenated)'"), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats}) { # 1st inner foreach-loop #
		my $cnct_bw_gstats = join ('', @{$node2bw_gstats->{$node}});
		print STDOUT ' ' x 3, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	    }				# End of the 1st inner foreach-loop #
	    print STDOUT "\n\n";

	    print STDOUT ' ' x 3, "( Map of indel events onto the branches )\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner foreach-loop #
		my @out_indel_info = ();
		foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop #
		    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
		    (defined $stat_ue) or ($stat_ue = 'undef');
		    (defined $stat_le) or ($stat_le = 'undef');
		    my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
		    push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		}			# End of the innermost foreach-loop #
		print STDOUT ' ' x 3, join ("\t", $br, @out_indel_info), "\n";
	    }				# End of the 2nd inner foreach-loop #
	    print STDOUT "\n\n";

	}			# End of the auxiliary outer for-loop (over parsimonious scenarios) #

	# @{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
	# %br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
	# of the blocks involved in the indel event,
	# %node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
	# interior nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id.

    }	# End of the Test output for 'enum_psm_indel_scns_in_cls_blks' #

	    $min_totcts_indels_for_cls[$cl] = $min_totct_indels;
	    $sets_psm_scns_for_cls[$cl] = $set_psm_scns;

	}			# End of the main outermost for-loop (over the clusters) #

    (print STDOUT "\n/// The 5th process (main) Ended ///\n\n") if ($TEST_LEVEL>1);

    (print STDOUT "\n<<< Subroutine 'enum_psm_indel_scns_for_msa' ENDED />>>\n\n") if ($TEST_LEVEL>0);

	return ($set_cls_gblocks, \@min_totcts_indels_for_cls, \@sets_psm_scns_for_cls, $set_gpatterns, $set_node2dparsstat, $set_br2change);

	# OUTPUT: (\@set_cls_gblocks, \@min_totcts_indels_for_cls, \@sets_psm_scns_for_cls,
	#	\@set_gpatterns, \@set_node2dparsstat, \@set_br2change), where

}	# End of the subroutine 'enum_psm_indel_scns_for_msa' #

# 
# INPUT: ($top_node_id, \%node2pa, \%node2ch, \%node2ct_exoffs, $char_pres, $char_abs, \@blocks_in_cls, $indx_indx_gp, \@set_node2dparsstat, \@set_br2change, $fl_full), where
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#
#	@blocks_in_cls represents a cluster of contiguous blocks,
#	each consisting of sites with a particular gap-pattern, where
#	$indx_gp = $blocks_in_cls[$b]->[$indx_indx_gp] (or $blocks_in_cls[$b] itself if $indx_indx_gp is undefined) is the index specifying the gap pattern in @set_node2parsstat and @set_br2change;
#
#	\@set_node2parsstat and \@set_br2change are output of 'map_parsim_states_dollo_msa'
#	on the 'secondary MSA', which is a concatenated set of 1-column representations of the gap patterns, where
#	%{$set_node2dparsstat[$k]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $k th gap pattern onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%{$set_br2change[$k]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches (under the Dollo parsimony) for the $k th gap pattern, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node;
#


1;	# a "spell" which makes the module "use"-able and/or "require"-able #



