package MyPRANK_Scores;
#
###################################################################################
#	
#	ComplLiMment_P: A package to calculate the Complete-likelihood for a given Multiple-sequence alignment
#			(almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyPRANK_Scores.pm:
#	This Perl package is a collection of subroutines
#	necessary to calculate the MSA scores for PRANK.
#	(CREATION STARTED on Aug 12 (Tue), 2014, ENDED on Sep 27 (Sat), 2014)
#

use strict;
use warnings;

use MyTaxonomy qw(
	if_ancestor fetch_ancestry calculate_abs_depth
	fetch_ext_offsprings_nr fetch_all_offsprings_nr
	);

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTreeManipu_III qw( make_node2abs_depth );

=set aside for potential use
use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );
=cut

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	calc_transition_mtrx_TN
	calc_br2tr_mtrx_TN
	init_prob_descresstats_cond_ancstat
	calc_prob_descresstats_cond_ancstat
	calc_prob_descresstats
	set_prank_align_hmm
	prank_score_2profile_aln
	);
	## mk_node2set_sw_gapstats


	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = qw( A G C T ); # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = 'log(ZERO)';

	## Default PRANK parameters. ##

	# Gap opening rate (relative to the substitution rate), which corresponds to 'r' in Loytynoja and Goldman (2005). (It should probably correspond to 1/2 of the total (relative) indel rate in Dawg (Cartwright, 2005).
our $GAPRATE_DNA = 0.025; # For DNA.
our $GAPRATE_PRO = 0.005; # For proteins.

	# Gap extension probability, which corresponds to  'epsilon' in Loytynoja and Goldman (2005).
our $GAPEXT_DNA  = 0.75; # For DNA.
our $GAPEXT_PRO  = 0.5; # For proteins.

	#  Match extension probability, which corresponds to 'gamma' in Loytynoja and Goldman (2005).
our $MATCHEXT = 0; # (In Dawg, it should be zero. But I'm not sure about the actual default value for PRANK...)


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'calc_prob_descresstats_cond_ancstat'. #
our $TEST_LEVEL_cpdca = 0;
##our $TEST_LEVEL_cpdca = 5;

	# For 'prank_score_2profile_aln'. #
our $TEST_LEVEL_ps2pa = 0;
##our $TEST_LEVEL_ps2pa = 6;

### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


# calc_transition_mtrx_TN:
#	This subroutine calculates (the integrated version of) the transition matrix
#	for a branch of a given length (measured in the number of expected substitutions per site),
#	in the Tamura-Nei nucleotide substitution model.
#	(CREATED on Aug 12 (Tue), 2014)
#
# INPUT: (\@in_equil_freqs, $a1, $a2, $b, $dist), where
#	$in_equil_freqs[$b] is the equilibrium frequency of the nucleotide $NUCLEOTIDES[$b];
#	$a1 and $a2 are the rate parameters for the transitional substitutions between A and G, and between C and T, respectively;
#	$b is the rate parameter for the transversional substitutions;
#
#	More specifically, we have the Tamura-Nei instantaneous substitution matrix as:
#	[M^{TN}_{$i,$j}] =	| ---           $a1 * pi_G	$b * pi_C	$b * pi_T  |
#				| $a1 * pi_A	---       	$b * pi_C	$b * pi_T  |
#				| $b * pi_A	$b * pi_G	---      	$a2 * pi_T |
#				| $b * pi_A	$b * pi_G	$a2 * pi_C	---        |
#	Here, M^{TN}_{$i,$j} represents the substitution rate from $NUCLEOTIDES[i] to $NUCLEOTIDES[$j].
#	I omitted the diagonal elements, which are always specified so that each row sums up to give zero.
#
#	$dist is the distance between the both ends of a branch,
#	measured in terms of the expected number of substitutions per nucleotide site.
#
# OUTPUT: \@tr_matrix_tn, where
#	$tr_matrix_tn[$i]->[$j] is the element of the Tamura-Nei transition matrix,
#	which gives the probability that $NUCLEOTIDES[$j] is found on one end of the branch,
#	on condition that $NUCLEOTIDES[$i] is found on the other end. 
#	(That is, the conditional probability of the transition from the $i th to the $j th nucleotide.)
#
sub calc_transition_mtrx_TN (\@$$$$) {

	my ($in_equil_freqs, $a1, $a2, $b, $dist) = @_;

	my ($pi_A, $pi_G, $pi_C, $pi_T) = @{$in_equil_freqs}; # Base frequencies.

	my $pi_R = $pi_A + $pi_G; # Purine frequency.
	my $pi_Y = $pi_C + $pi_T;  # Pyrimidine frequency.

	my ($tpi_A, $tpi_G) = ($pi_R>0) ? ($pi_A/$pi_R, $pi_G/$pi_R) : (1/2, 1/2); # Relative frequencies between purines.

	unless ($pi_R>0) {
	    print STDOUT "\n[WARNING 'calc_transition_mtrx_TN'] The purine frequency is ZERO ... I arbitrarily set ", '($tpi_A, $tpi_G) = ', "($tpi_A, $tpi_G) ...\n\n";
	}

	my ($tpi_C, $tpi_T) = ($pi_Y>0) ? ($pi_C/$pi_Y, $pi_T/$pi_Y) : (1/2, 1/2); # Relative frequencies between pyrimidines.

	unless ($pi_Y>0) {
	    print STDOUT "\n[WARNING 'calc_transition_mtrx_TN'] The pyrimidine frequency is ZERO ... I arbitrarily set ", '($tpi_C, $tpi_T) = ', "($tpi_C, $tpi_T) ...\n\n";
	}

		# Column vectors. #
	my @cvec1 = (1, 1, 1, 1);
	my @cvec2 = ($tpi_G, -$tpi_A, 0, 0);
	my @cvec3 = (0, 0, $tpi_T, -$tpi_C);
	my @cvec4 = ($pi_Y, $pi_Y, -$pi_R, -$pi_R);

		# Row vectors. #
	my @rvec1 = ($pi_A, $pi_G, $pi_C, $pi_T);
	my @rvec2 = (1, -1, 0, 0);
	my @rvec3 = (0, 0, 1, -1);
	my @rvec4 = ($tpi_A, $tpi_G, -$tpi_C, -$tpi_T);

		# Exponential factors. #
	my $n_fac = 2*($a1 * $pi_A * $pi_G + $a2 * $pi_C * $pi_T + $b * $pi_R * $pi_Y); # Normalization factor.
	my $exp2 = exp ( - ($pi_R * $a1 + $pi_Y * $b) * $dist /$n_fac );
	my $exp3 = exp ( - ($pi_Y * $a2 + $pi_R * $b) * $dist /$n_fac );
	my $exp4 = exp ( - $b * $dist /$n_fac );

	my @tr_matrix_tn = ();

	for (my $i=0; $i< $CT_NUCLS; $i++) { # Main outer for-loop (over initial nucleotides). #

	    $tr_matrix_tn[$i] = []; # Initialize the row.

	    for (my $j=0; $j<$CT_NUCLS; $j++) { # Main inner for-loop (over final nucleotides). #

		$tr_matrix_tn[$i]->[$j] =
		   $cvec1[$i] * $rvec1[$j] + $exp2 * $cvec2[$i] * $rvec2[$j] 
		   + $exp3 * $cvec3[$i] * $rvec3[$j] + $exp4 * $cvec4[$i] * $rvec4[$j] ;

	    }				# End of the main inner for-loop (over final nucleotides). #

	}			# End of the main outer for-loop (over initial nucleotides).

	return \@tr_matrix_tn ;

}	# End of the subroutine 'calc_transition_mtrx_TN'. #


# calc_br2tr_mtrx_TN:
#	This subroutine calculates (the integrated version of) the transition matrices
#	in the Tamura-Nei nucleotide substitution model
#	for tree branches, each with a given length (measured in the number of expected substitutions per site),
#	(CREATED on Aug 12 (Tue), 2014)
#
#	For more details on each transition matrix, see the heading of 'calc_transition_mtrx_TN'.
#
# INPUT: (\%br_attr, \@in_equil_freqs, $a1, $a2, $b), where
#	%br_attr = ($branch_id => \@attr, ...), with
#	$attr[$INDX_LENGTH] giving the branch length (measured in the expected number of base substitutions per site;
#
#	$in_equil_freqs[$b] is the equilibrium frequency of the nucleotide $NUCLEOTIDES[$b];
#	$a1 and $a2 are the rate parameters for the transitional substitutions between A and G, and between C and T, respectively;
#	$b is the rate parameter for the transversional substitutions.
#
# OUTPUT: \%br2tr_mtrx_tn, where
#	@{$br2tr_mtrx_tn{$br_id}} is the transition matrix in the Tamura-Nei model 
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx_tn{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDE[$j]
#	given that the parent of $br_id shows $NUCLEOTIDE[$i].
#
sub calc_br2tr_mtrx_TN (\%\@$$$) {

	my ($br_attr, $in_equil_freqs, $a1, $a2, $b) = @_;

	my %br2tr_mtrx_tn;

	foreach my $br_id (keys %{$br_attr}) { # Main foreach-loop (over branches). #

	    my $attr = $br_attr->{$br_id};
	    my $brlen = $attr->[$INDX_LENGTH];

	    my $tr_matrix_tn = calc_transition_mtrx_TN (@{$in_equil_freqs}, $a1, $a2, $b, $brlen);
	    unless (defined $tr_matrix_tn) {
		print STDOUT "[ERROR 'calc_br2tr_mtrx_TN'] 'calc_transition_mtrx_TN' FAILED on the branch $br_id : $! !\n\n";
		return undef;
	    }

	    $br2tr_mtrx_tn{$br_id} = $tr_matrix_tn;

	}			# End of the main foreach-loop (over branches). #

	return \%br2tr_mtrx_tn;

}	# End of the subroutine 'calc_br2tr_mtrx_TN'. #


# init_prob_descresstats_cond_ancstat:
#	This satellite subroutine of 'calc_prob_descresstats_cond_ancstat' initializes the main output of
#	the master subroutine, %node2set_sw_cond_probs, so that
#
#	$node2set_sw_cond_probs{$node}->[$s][$i] = 1 if the $s th site of the sequence at the exterior node $node is occupied by $NUCLEOTIDES[$i],
#						= 0 if occupied by aother nucleotide, and
#	$node2set_sw_cond_probs{$node}->[$s] = [] (empty) if the site is occupied by a gap.
#
#	This subroutine leaves the input MSA intact.
#
#	(CREATED on Sep 19 (Fri), 2014)
#
# INPUT: (\%node2set_sw_cond_probs, \@in_msa, \@seqnames, \%name2id, $gap), where
#
#	%node2set_sw_cond_probs is the main output of 'calc_prob_descresstats_cond_ancstat', which this subroutine will initialize;
#
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$seqnames[$k] is the name (or ID) of the $k th sequence in the input MSA;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node (i.e., root) of the tree;
#
#	$gap is the character representing a gap.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub init_prob_descresstats_cond_ancstat (\%\@\@\%$) {

	my ($node2set_sw_cond_probs, $in_msa, $seqnames, $name2id, $gap) = @_;

	my $ct_seqs = @{$in_msa};

	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat'] ILLEGAL INPUT: The input MSA contains NO Sequences: $! !\n\n";
	    return 0;
	}
	unless (@{$seqnames} == $ct_seqs) {
	    my $ct_seqs2 = @{$seqnames};
	    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat'] ILLEGAL INPUT: The number of sequences in the input MSA (=$ct_seqs) CLASHES the number of sequence names (=$ct_seqs2): $! !\n\n";
	    return 0;
	}
	$gap = uc $gap;

	my $ct_sites = length ( ${$in_msa->[0]} );

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences in the MSA). #

	    my $seqname = $seqnames->[$k];
	    my $node_id = $name2id->{$seqname};
	    unless (defined $node_id) {
		print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat'] ILLEGAL INPUT: The $k th sequence in the MSA (name: $seqname) does NOT correspond to ANY EXTERIOR NODE: $! !\n\n";
		return 0;
	    }

	    my $cp_seq = uc ${$in_msa->[$k]};  # Copy the sequence. #

	    my @set_sw_cond_probs = ();

	    for (my $s=0; $s<$ct_sites; $s++) { # Main inner for-loop (over sites). #

		my $char = substr ($cp_seq, 0, 1, '');

		if ($char eq $gap) {
		    $set_sw_cond_probs[$s] = []; # Assign an empty array. #
		    next;
		}

		my $indx = $NUCL2INDX{$char};
		unless (defined $indx) {
		    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat'] ILLEGAL INPUT: The $s th site of the $k th sequence (name: $seqname) is OCCUPIED by an UNEXPECTED CHARACTER, '$char': $! !\n\n";
		    return 0;
		}

		my @probs = ();
		for (my $c=0; $c<$CT_NUCLS; $c++) { $probs[$c] = ($c == $indx) ? 1 : 0; }

		$set_sw_cond_probs[$s] = \@probs;

	    }			# End of the main inner for-loop (over sites). #

	    $node2set_sw_cond_probs->{$node_id} = \@set_sw_cond_probs;

	}			# End of the main outer for-loop (over sequences in the MSA). #

	return 1; # The subroutine successfully ended. #

}	# End of the subroutine 'init_prob_descresstats_cond_ancstat'. #


# calc_prob_descresstats_cond_ancstat:
#	This subroutine calculates probabilities of the residue states of all the descendants
#	of an ancestral node, conditional on the residue state of that ancestral node,
#	of every site in an input MSA.
#	The subroutine leaves the input MSA intact.
#	(CREATED on Sep 19 (Fri), 2014)
#
#	[Formula]
#	The probability, p(n)_i = Pr [{s(d)}_n | s(n) = N_i ],
#	that the descendant sequences of node n, {d}_n, show the states, {s(d}}_n, at the s-th site,
#	given that the s-th site at node n is nucleotide N_i, can be calculated by the following formula:
#
#	p(n)_i = Prod_{m in ch(n)}  [ Sum_{j} { t(n;m)_{i,j} * p(m)_j } ],
#
#	where t(n;m)_{i,j} is the substitution transition probability that
#	node m has nucleotide N_j, on condition that node n has nucleotide N_i.
#
#	[NOTE] This subroutine itself can handle multifurcated trees as well,
#		although PRANK analyses cannot, in general.
#
# INPUT: (\@in_msa, \@seqnames, \%name2id, \%node2ch, \%br2tr_mtrx, $gap), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$seqnames[$k] is the name (or ID) of the $k th sequence in the input MSA;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node (i.e., root) of the tree;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes of the tree;
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	$gap is the character representing a gap.
#
# OUTPUT: \%node2set_sw_cond_probs, where
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#
#	But $set_sw_cond_probs[$s] = [] (empty) if all the descendants of $node_id have their $s th sites occupied by gaps.
#
sub calc_prob_descresstats_cond_ancstat (\@\@\%\%\%$) {

	my ($in_msa, $seqnames, $name2id, $node2ch, $br2tr_mtrx, $gap) = @_;

	$gap = uc $gap;

	my $ct_seqs = @{$in_msa};
	my $ct_sites = length (${$in_msa->[0]});

	my $topnode_id = $name2id->{$TOP_NODE};

   (print STDOUT "\n< Subroutine 'cal_prob_descresstats_cond_ancstat'  (#{sequences} = $ct_seqs; #{sites} = $ct_sites; topnode ID = $topnode_id; Gap = '$gap') >\n\n") if ($TEST_LEVEL_cpdca>0);

		# Preliminary (I): Initialize the output hash. #

	my %node2set_sw_cond_probs;

	unless ( init_prob_descresstats_cond_ancstat (%node2set_sw_cond_probs, @{$in_msa}, @{$seqnames}, %{$name2id}, $gap) ) {
	    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstat'] Subroutine 'init_prob_descresstats_cond_ancstat' FAILED : $! !\n\n";
	    return undef;
	}
   (print STDOUT "\n ... 'init_prob_descresstats_cond_ancstat' Ended ...\n\n") if ($TEST_LEVEL_cpdca >1);


   if ($TEST_LEVEL_cpdca>3) {

	print STDOUT "\n[ Initialized Output, '", '%node2set_sw_cond_probs', "' ]\n\n";

	foreach my $node (sort {$a <=> $b} keys %node2set_sw_cond_probs) { # Outer testing foreach-loop (over exterior nodes). #

	    my $set_sw_cond_probs = $node2set_sw_cond_probs{$node};

	    print STDOUT "( Conditional probabilities for Node ID: $node )\n\n";
	    print STDOUT join ("\t", 'Site_no.', @NUCLEOTIDES), "\n\n";

	    for (my $s=0; $s<$ct_sites; $s++) { print STDOUT join ("\t", $s, @{$set_sw_cond_probs->[$s]}), "\n"; }
	    print STDOUT "\n\n";

	}			# End of the outer testing foreach-loop (over exterior nodes). #

   }	# END of "if ($TEST_LEVEL_cpdca>2) {...}"


		# Preliminary (II): Assign depths to the nodes. #

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstat'] Subroutine 'make_node2abs_depth' FAILED : $! !\n\n";
	    return undef;
	}
   (print STDOUT "\n ... 'make_node2abs_depth' Ended ...\n\n") if ($TEST_LEVEL_cpdca >1);

   if ($TEST_LEVEL_cpdca>2) {
	print STDOUT "\n[ Depths assigned to the nodes ]\n\n";
	print STDOUT join ("\t", 'Node_ID', 'Depth'), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}), "\n"; }
	print STDOUT "\n\n";
   }


		## MAIN PROCESS: Calculate the conditional probabilities in a BOTTOM-UP manner. ##

   (print STDOUT "\n{{ MAIN PROCESS: Calculate the conditional probabilities from bottom (leaves) to top (root)... }}\n\n") if ($TEST_LEVEL_cpdca>0);


	foreach my $node_id (sort { $node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2ch}) { # Main outer foreach-loop (over interior nodes). #

	    my $depth = $node2depth->{$node_id};

    (print STDOUT "{ Node ID: $node_id (depth = $depth) }\n\n") if ($TEST_LEVEL_cpdca>1);

	    my $child_nodes = $node2ch->{$node_id};
	    my $ct_children = @{$child_nodes};

   (print STDOUT "There are $ct_children child nodes: ", join (', ', @{$child_nodes}), "\n\n") if ($TEST_LEVEL_cpdca>2) ;

	    my @setset_sw_ch_cond_probs = ();
	    my @set_tr_mtrices = ();
	    foreach my $ch (@{$child_nodes}) { # Middle preliminary foreach-loop (over child nodes). #

		my $set_sw_ch_cond_probs = $node2set_sw_cond_probs{$ch};
		unless (defined $set_sw_ch_cond_probs) {
		    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstats'] NO SET of conditional PROBABILITIES was ASSIGNED TO the child '$ch' of the node '$node_id': $! !\n\n";
		    return undef;
		}
		push @setset_sw_ch_cond_probs, $set_sw_ch_cond_probs;

		my $tr_mtrx = $br2tr_mtrx->{$ch};
		unless (defined $tr_mtrx) {
		    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstats'] NO TRANSITION MATRIX was ASSIGNED TO the branch from the node '$node_id' to its child '$ch': $! !\n\n";
		    return undef;
		}
		push @set_tr_mtrices, $tr_mtrx;

	    }		# End of the middle preliminary foreach-loop (over child nodes). #

   if ($TEST_LEVEL_cpdca>3) {

	print STDOUT "[ Transition matrices along the branches from the node '$node_id' ]\n\n";
	for (my $c=0; $c<$ct_children; $c++) {

	    my $ch_id = $child_nodes->[$c];
	    my $tr_mtrx = $set_tr_mtrices[$c];

	    print STDOUT "| Down to the $c th child (ID: $ch_id) |\n\n";
	    print STDOUT join ("\t", 'From\To', @NUCLEOTIDES), "\n\n";
	    for (my $i=0; $i<$CT_NUCLS; $i++) { print STDOUT join ("\t", $NUCLEOTIDES[$i], @{$tr_mtrx->[$i]}), "\n"; } 
	    print STDOUT "\n\n";
	}

   }

   (print STDOUT "[ Site-wise calculations ... ]\n\n") if ($TEST_LEVEL_cpdca>2) ;

	    my @set_sw_cond_probs = ();

	    for (my $s=0; $s<$ct_sites; $s++) { # Main middle for-loop (over sites). #


   (print STDOUT ' ' x 2,  "|| The $s th site ||\n\n") if ($TEST_LEVEL_cpdca>3) ;

			# Initialize the conditional probabilities at the site. # 
		my @cond_probs = ();
		for (my $i=0; $i<$CT_NUCLS; $i++) { $cond_probs[$i] = 1; }

		my $ct_empty = 0;

		for (my $c=0; $c<$ct_children; $c++) { # Main inner-middle for-loop (over child nodes). #

		    my $ch_id = $child_nodes->[$c];

   (print STDOUT ' ' x 4, "| The $c th child (ID: $ch_id) |\n\n") if ($TEST_LEVEL_cpdca>3);

		    my $ch_cond_probs = $setset_sw_ch_cond_probs[$c]->[$s];
		    unless (@{$ch_cond_probs}>0) { # If the probabilities conditional on the child states are NOT assigned.

   (print STDOUT ' ' x 4, "... Calculation was SKIPPED because the child-conditional probabilities are UNDEFINED ...\n\n") if ($TEST_LEVEL_cpdca>=5);

			$ct_empty++;
			next;
		    }

		    my $tr_mtrx = $set_tr_mtrices[$c];

   if ($TEST_LEVEL_cpdca>=5) {
	print STDOUT ' ' x 4, "{Descendant state probabilities conditional on Res($ch_id) = (", join (',', @NUCLEOTIDES), ")} = (", join (', ', @{$ch_cond_probs}), ") .\n\n";

	print STDOUT ' ' x 4, "( Transition matrix from $node_id to $ch_id )\n\n";
	print STDOUT ' ' x 4, join ("\t", 'From\To', @NUCLEOTIDES), "\n\n";
	for (my $i=0; $i<$CT_NUCLS; $i++) { print ' ' x 4, join ("\t", $NUCLEOTIDES[$i], @{$tr_mtrx->[$i]}), "\n"; }
	print STDOUT "\n\n";
   }

   (print STDOUT ' ' x 4, "{ Calculation of conditional probabilities of the residue states of descendant sequences of the child node '$ch_id', conditional on the state of the node '$node_id' }\n\n") if ($TEST_LEVEL_cpdca>=5);

   (print STDOUT ' ' x 6, join ("\t", "Res($node_id)", 'Cond_Prob(child_spec_descendants)' ), "\n\n") if ($TEST_LEVEL_cpdca>5);

		    for (my $i=0; $i<$CT_NUCLS; $i++) { # Main inner for-loop (over potential residue states at the subject node). #
				# Here, the formula in the header will be calculated. #

			my $row_tr_probs = $tr_mtrx->[$i];

			my $ch_spec_prob = 0; # Probability factor specific to the $c th child. It corresponds to a factor in the product over children.
			for (my $j=0; $j<$CT_NUCLS; $j++) { $ch_spec_prob += $row_tr_probs->[$j] * $ch_cond_probs->[$j]; }

			$cond_probs[$i] *= $ch_spec_prob;

   (print STDOUT ' ' x 6, join ("\t", $NUCLEOTIDES[$i], $ch_spec_prob), "\n") if ($TEST_LEVEL_cpdca>5); 

		    }				# End of the main inner for-loop (over potential residue states at the subject node).
   (print STDOUT "\n") if ($TEST_LEVEL_cpdca>5);
   (print STDOUT "\n") if ($TEST_LEVEL_cpdca>=5);

		}			# End of the main inner-middle for-loop (over child nodes). #

		if ($ct_empty == $ct_children) { @cond_probs = (); } # If all the descendant sequences are occupied by gaps, conditional probabilities cannot be calculated.

		$set_sw_cond_probs[$s] = \@cond_probs;

   if ($TEST_LEVEL_cpdca>=5) {
	print STDOUT ' ' x 4, "[ The resulting probabilies of residue states of descendants of the node '$node_id' conditional on its residue state ]\n\n";
	if (@cond_probs>0) {
	    print STDOUT ' ' x 4, join ("\t", "Res($node_id)", 'Cond_Prob(all_descendants)'), "\n\n";
	    for (my $i=0; $i<$CT_NUCLS; $i++) { print STDOUT ' ' x 4, join ("\t", $NUCLEOTIDES[$i], $cond_probs[$i]), "\n"; }
	    print STDOUT "\n\n";
	} else {
	    print STDOUT ' ' x 4, "... The probabilities are UNDEFINED because all descendants are occupied by gaps at the site ...\n\n\n";
	}
   }

	    }			# End of the main middle for-loop (over sites). #

	    $node2set_sw_cond_probs{$node_id} = \@set_sw_cond_probs;

	}			# End of the main outer foreach-loop (over interior nodes). #


   (print STDOUT "\n< End of subroutine 'cal_prob_descresstats_cond_ancstat' />\n\n") if ($TEST_LEVEL_cpdca>0);

	return \%node2set_sw_cond_probs;

}	# End of the subroutine 'calc_prob_descresstats_cond_ancstat'. #


# calc_prob_descresstats:
#	This subroutine receives the output of 'calc_prob_descresstats_cond_ancstat',
#	as well as the equilibrium residue frequencies of the substitution models,
#	and calculates the probabilities that the residue states of the descendant sequences
#	of each node at each site are realized.
#
#	[Formula]
#	Let p(n)_i = Pr [{s(d)}_n | s(n) = N_i ] the probability
#	that the descendant sequences of node n, {d}_n, show the states, {s(d}}_n, at the s-th site,
#	given that the s-th site at node n is nucleotide N_i.
#	Then, the UNCONDITIONAL probability p(n) that {d}_n show the states, {s(d)}_n, at the s-th site
#	is calculated as:
#
#	p(n) = Sum_{i} pi_i * p(n)_i,
#
#	where pi_i is the equilibrium frequency of nucleotide N_i.
#
# INPUT: (\%node2set_sw_cond_probs, \@equil_res_freqs), where
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#
#	But $set_sw_cond_probs[$s] = [] (empty) if all the descendants of $node_id have their $s th sites occupied by gaps;

#	$equil_res_freqs[$i] is the equilibrium frequency of $NUCLEOTIDES[$i].
#
# OUTPUT: \%node2set_sw_probs, where
#	%node2set_sw_probs = ($node_id => \@set_sw_probs, ...) for ALL nodes, where
#	$set_sw_probs[$s] is the probability that all the descendant sequences of $node_id show the residue that are actually found in the $s th site.
#	But $set_sw_probs[$s] = undef if all the descendant sequences are occupied by gaps at the $s th site.
#
sub calc_prob_descresstats (\%\@) {

	my ($node2set_sw_cond_probs, $equil_res_freqs) = @_;

	my %node2set_sw_probs ;

	my $ct_sites;

	foreach my $node_id (sort {$a <=> $b} keys %{$node2set_sw_cond_probs}) { # Main outer foreach-loop (over nodes). #

	    my $set_sw_cond_probs = $node2set_sw_cond_probs->{$node_id};

	    (defined $ct_sites) or ($ct_sites = @{$set_sw_cond_probs});

	    my @set_sw_probs = ();

	    for (my $s=0; $s<$ct_sites; $s++) { # Main middle for-loop (over sites). #

		my $sw_cond_probs = $set_sw_cond_probs->[$s];

		unless (@{$sw_cond_probs}>0) {
		    $set_sw_probs[$s] = undef;
		    next;
		}

			# Here, the formula in the heading will be calculated !! #
		my $sw_prob = 0;
		for (my $i = 0; $i<$CT_NUCLS; $i++) { $sw_prob += $equil_res_freqs->[$i] * $sw_cond_probs->[$i]; }

		$set_sw_probs[$s] = $sw_prob;

	    }			# End of the main middle for-loop (over sites). #

	    $node2set_sw_probs{$node_id} = \@set_sw_probs;

	}		# End of the main outer foreach-loop (over nodes). #

	return \%node2set_sw_probs;

}	# End of the subroutine 'calc_prob_descresstats'. #


# set_prank_align_hmm:
#	This subroutine creates a "transition matrix" under the Alignment HMM used in PRANK.
#	The output is in the form of a hash that can be used as an input of the CORE SUBROUTINE, 'prank_score_2profile_aln'.
#	(CREATED on Sep 23 (Tue), 2014.)
#
# INPUT: ($pw_dist_sbst, $gaprate, $gapext, $matchext), where
#	$pw_dist_sbst is the distance between the two child nodes whose profiles (sub-MSAs) are to be aligned,
#	in terms of the expected number of substitutions per site. It corresponds to |vec(x)| + |vec(y)| in Loytynoja and Goldman (2005);
#	$gaprate is the gap-opening rate (relative to the substitution rate).
#	It corresponds to the PRANK option with the same name, and to 'r' in Loytynoja and Goldman (2005). But it corresponds to 1/2 of the (relative) total indel rate (i.e., the insertion rate plus the deletion rate) in Dawg;
#	$gapext is the gap extension probability, which should correspond to 'epsilon' in Loytynoja and Goldman (2005);
#	$matchext is the extension probability of matches, which corresponds to 'gamma' in Loytynoja and Goldman (2005).
#
# OUTPUT: \%trans_states2score, which can be used as an input of 'prank_score_2profile_aln', where
#	%trans_states2score = ("$state_last:$state_pres" => $score, ....), where $state_last/pres is the state at the last/present site (column) in the currently reconstructed alignment, and
#	$state_last/pres = 'M' (if the two sites are aligned),
#			 = 'X' (if residues in the 1st profile is aligned with the gap-only column in the 2nd profile),
#			 = 'Y' (if residues in the 2nd profile is aligned with the gap-only column in the 1st profile),
#			 = 'S' (for the alignment start),
#			 = 'E' (for the alignment end).
#	$score is the score itself, not the logarithm;
#
sub set_prank_align_hmm ($$$$) {

	my ($pw_dist_sbst, $gaprate, $gapext, $matchext) = @_;

	my $eps = $gapext;
	my $del = 1 - exp (- $pw_dist_sbst * $gaprate);
	my $gam = $matchext;

	my %trans_states2score = ( # "$state_last:$state_pres" => $score, ...

		'S:X' => (1-2*$del) * (1-$gam) * $del,
		'S:Y' => (1-2*$del) * (1-$gam) * $del,
		'S:M' => (1-2*$del) * ($gam + (1-$gam)*(1-2*$del)),
		'X:X' => $eps + (1-$eps)*$del,
		'Y:Y' => $eps + (1-$eps)*$del,
		'M:M' => $gam + (1-$gam)*(1-2*$del),
		'X:Y' => (1-$eps)*$del,
		'Y:X' => (1-$eps)*$del,
		'X:M' => (1-$eps)*(1-2*$del),
		'Y:M' => (1-$eps)*(1-2*$del),
		'M:X' => (1-$gam)*$del,
		'M:Y' => (1-$gam)*$del,
		'X:E' => 1-$eps,
		'Y:E' => 1-$eps,
		'M:E' => 1-$gam,
		'S:E' => (1-2*$del) * (1-$gam),
		);


	return \%trans_states2score;

}	# End of the subroutine 'set_prank_align_hmm'. #


=This subroutine seems unnecessary after all. Thus it was frozen.

# mk_node2set_sw_gapstats:
#	This subroutine receives the output of 'calc_prob_descresstats',
#	and returns the gapstates of all sites at all nodes.
#
# INPUT: (\%node2set_sw_probs, \%node2ch), where
#	%node2set_sw_probs = ($node_id => \@set_sw_probs, ...) for ALL nodes, where
#	$set_sw_probs[$s] is the probability that all the descendant sequences of $node_id show the residue that are actually found in the $s th site.
#	But $set_sw_probs[$s] = undef if all the descendant sequences are occupied by gaps at the $s th site.
#	This hash is typically an output of 'calc_prob_descresstats';
#
#	%node2ch = ($node_id => \@children, ...) for all interior nodes.
#
# OUTPUT: \%node2set_sw_gapstats, where
#	%node2set_sw_gapstats = ($node_id => \@set_sw_gapstats, ...) for ALL nodes, where
#
#	$set_sw_gapstats[$s] = 1, 0, -1 if the $s th site of the profile (sub-MSA) at the node $node_id
#
#	consists of "residue vs residue," "residue vs gap" or "gap vs residue," and "gap vs gap," respectively,
#
#	of the sub-sub-MSAs at its child nodes.
#
sub mk_node2set_sw_gapstats (\%\%) {

	my ($node2set_sw_probs, $node2ch) = @_;

	my %node2set_sw_gapstats;

	foreach my $node_id (sort {$a <=> $b} keys) { # Main outer foreach-loop (over nodes). #

	    my $children = $node2ch->{$node_id};

	    unless (defined $children) {
			# An exterior node. => Assign the gap-states according to its own gap states alone.
			# (The states can never be 0 (i.e., gap vs residue).)

		my $set_sw_probs = $node2set_sw_probs->{$node_id};
		my $ct_sites = @{$set_sw_probs};
		my @set_sw_gapstats = ();
		for (my $s=0; $s<$ct_sites; $s++) { $set_sw_gapstats[$s] = (defined $set_sw_probs->[$s]) ? 1 : -1; }

		$node2set_sw_gapstats{$node_id} = \@set_sw_gapstats;
		next;
	    }

		# An interior node. #

	    unless (@{$children}==2) {
		my $ct_ch = @{$children};
		print STDOUT "[ERROR 'mk_node2set_sw_gapstats'] ILLEGAL INPUT: The number of child nodes of the node '$node_id' is $ct_ch, which differs from 2 that is expected: $! !\n\n";
		return undef;
	    }

	    my ($ch1, $ch2) = @{$children};
	    my $set_sw_probs_ch1 = $node2set_sw_probs->{$ch1};
	    my $set_sw_probs_ch2 = $node2set_sw_probs->{$ch2};

	    my $ct_sites = @{$set_sw_probs_ch1};

	    my @set_sw_gapstats = ();

	    for (my $s=0; $s<$ct_sites; $s++) { # Main inner for-loop (over sites). #

		if (defined $set_sw_probs_ch1->[$s]) { # Child 1 is occupied by some residues.

		    $set_sw_gapstats[$s] = (defined $set_sw_probs_ch2->[$s]) ? 1 : 0;

		} else { # Child 1 is occupied by gaps alone.

		    $set_sw_gapstats[$s] = (defined $set_sw_probs_ch2->[$s]) ? 0 : -1;
		}
	    }			# End of the main inner for-loop (over sites). #

	    $node2set_sw_gapstats{$node_id} = \@set_sw_gapstats;
	    
	}			# End of the main outer foreach-loop (over nodes). #


	return \%node2set_sw_gapstats;

}	# End of the subroutine 'mk_node2set_sw_gapstats'. #

=cut # END of "This subroutine seems unnecessary after all. Thus it was frozen."


# prank_score_2profile_aln:
#	This subroutine scores a given profile-profile alignment (i.e., the alignment of two sub-MSAs) at a progressive alignment step
#	corresponding to a specified node (of the guide tree), according to the scheme of PRANK.
#	As a SIDE-EFFECT, the subroutine also assigns ZERO probabilities to the nodes ancestral to a branch,
#	at the site that are "confirmed" as inserted along the branch.
#	(CREATION STARTED on Sep 22 (Mon), 2014, ENDED icompletely on Sep 23 (Tue), 2014)
#	(MODIFICATION STARTED on Sep 24 (Wed), 2014, ENDED on Sep ?? (???), 2014)
#
# INPUT: ($node_z, \%node2set_sw_probs, \%node2ch, \%node2pa, \%trans_states2score, $state_lf, $state_rf, $if_open_forever, $if_ignore_term_gaps), where
#	$node_z is the node corresponding to the progressive alignment step.
#	That is, the profiles (sub-MSAs) reconstructed at the child nodes of $node_z, called $node_x and $node_y,
#	 will be aligned (or, actually, was aligned);
#
#	%node2set_sw_probs = ($node_id => \@set_sw_probs, ...) for ALL nodes, where
#	$set_sw_probs[$s] is the probability that all the descendant sequences of $node_id show the residue that are actually found in the $s th site.
#	But $set_sw_probs[$s] = undef if all the descendant sequences are occupied by gaps at the $s th site.
#	This hash is typically an output of 'calc_prob_descresstats';
#
#	%node2ch = ($node_id => \@child_nodes, ....) for interior nodes;
#	%node2pa = ($node_id => $parent, ...) for nodes except the top node (root);
#
#	%trans_states2score = ("$state_last:$state_pres" => $score, ....), where $state_last/pres is the state at the last/present site (column) in the currently reconstructed alignment, and
#	$state_last/pres = 'M' (if the two sites are aligned),
#			 = 'X' (if residues in the 1st profile is aligned with the gap-only column in the 2nd profile),
#			 = 'Y' (if residues in the 2nd profile is aligned with the gap-only column in the 1st profile),
#			 = 'S' (for the alignment start),
#			 = 'E' (for the alignment end).
#	$score is the score itself, not the logarithm;
#
#	$state_lf/rf is the state at the left/right-flanking column (excluding all-gap columns) of the currently reconstructed alignment;
#
#	$if_open_forever = 1 if you want the "confirmed" insertions to open forever (i.e., never alignable with residues),
#			 = 0 if you want to leave the "confirmed" insertions to be alignable with residues again.
#
#	$if_ignore_term_gaps = 1 if you ignore the gaps at the end of the alignment,
#			     = 0 if you include them them into the scoring.
#
# OUTPUT: ($log_prank_score, $log_sbst_score, $log_hmm_score), where
#	$log_prank_score is the logarithm of the score of the alignment.
#		(= $LOG_ZERO if the score becomes zero (i.e., the alignment is impossible));
#	$log_sbst_score is the logarithm of the portion of the score concerning substitutions;
#	$log_hmm_score is the logarithm of the portion of the score concerning gap-patterns (specified by the alignment HMM of PRANK).
#
#	Usually, $log_prank_score = $log_sbst_score + $log_hmm_score.
#
sub prank_score_2profile_aln ($\%\%\%\%$$$$) {

	my ($node_z, $node2set_sw_probs, $node2ch, $node2pa, $trans_states2score, $state_lf, $state_rf, $if_open_forever, $if_ignore_tgaps) = @_;


	unless (($state_lf eq 'M') or ($state_lf eq 'S')) {
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The left-flanking state, '$state_lf', is UNEXPECTED: $! !\n\n";
	    return (undef, undef, undef);
	}
	unless (($state_rf eq 'M') or ($state_rf eq 'E')) {
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The right-flanking state, '$state_rf', is UNEXPECTED: $! !\n\n";
	    return (undef, undef, undef);
	}


		# Preliminary process (I): Prepare the sets of site-wise probabilities at the $node_z, and its child nodes, namely $node_x and $node_y. #

	my $sw_probs_z = $node2set_sw_probs->{$node_z};
	my $ct_sites = @{$sw_probs_z};

	my $children_z = $node2ch->{$node_z};

	unless ((defined $children_z) and (@{$children_z}==2)) {
	    my $ct_ch = (defined $children_z) ? @{$children_z} : 0;
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The subject node '$node_z' has $ct_ch children, which is different from 2 as we expected: $! !\n\n";
	    return (undef, undef, undef);
	}
	my ($node_x, $node_y) = @{$children_z};
	my $sw_probs_x = $node2set_sw_probs->{$node_x};
	my $sw_probs_y = $node2set_sw_probs->{$node_y};


   (print STDOUT "< Subroutine 'prank_score_2profile_aln' (", '$node_z = ', $node_z, ', $node_x = ', $node_x, ', $node_y = ', $node_y, ', $ct_sites = ', $ct_sites, ', $state_lf = ', $state_lf, ', $state_rf = ', $state_rf, ', $if_open_forever = ', $if_open_forever, ', $if_ignore_tgaps = ', $if_ignore_tgaps, ') >', "\n\n") if ($TEST_LEVEL_ps2pa>0);


   if ($TEST_LEVEL_ps2pa>2) {

	print STDOUT "[ Site-wise descendant residue probabilities at the subject nodes ]\n\n";
	print STDOUT join ("\t", 'Site', 'Node_Z', 'Node_X', 'Node_Y'), "\n\n";

	for (my $s=0; $s<$ct_sites; $s++) { # Testing for-loop (over sites). #
	    my @out = ($s);
	    foreach my $prob ($sw_probs_z->[$s], $sw_probs_x->[$s], $sw_probs_y->[$s]) { push @out, (defined $prob) ? $prob : 'undef'; }
	    print STDOUT join ("\t", @out), "\n";
	}		# End of the testing for-loop (over sites). #
	print STDOUT "\n\n";
   }


		# Preliminary process (II): Set the site regions to be examined (in the 0-based half-open convention). #

	my $lb = 0;
	my $ub = $ct_sites;

	if ($if_ignore_tgaps) {
	   if ($state_lf eq 'S') {
		for (my $s=0; $s<$ct_sites; $s++) { # 1st preliminary for-loop (over sites). #
		    ((defined $sw_probs_x->[$s]) and (defined $sw_probs_y->[$s])) and last;
		    $lb++;
		}			# End of the 1st preliminary for-loop (over sites). #
	   }

	   if ($state_rf eq 'E') {
		for (my $s=$ct_sites-1; $s>=$lb; $s--) { # 2nd preliminary for-loop (over sites). #
		    ((defined $sw_probs_x->[$s]) and (defined $sw_probs_y->[$s])) and last;
		    $ub--;
		}			# End of the 2nd preliminary for-loop (over sites). #
	   }
	}

   if ($TEST_LEVEL_ps2pa>1) {
	print STDOUT "[ Boundaries of the sites to be examined (in the 0-based half-open convention) ]\n\n";
	print STDOUT "{Lower-boundary} = $lb ,\n";
	print STDOUT "{Upper-boundary} = $ub .\n";
	print STDOUT "\n\n";
   }

		### MAIN PROCESS: Score the profile-profile alignment reconstructed at $node_z. ###

   (print STDOUT "[[ MAIN PROCESS: Score the profile-profile alignment reconstructed at $node_z. ]]\n\n\n") if ($TEST_LEVEL_ps2pa>1);

	my $state_last = $state_lf;
	my $last_site = -1;

	my $log_sbst_score = my $log_hmm_score = 0;

	for (my $s=$lb; $s<$ub; $s++) { # Main outer for-loop (over sites to be examined). #

   (print STDOUT "|| Site no. = $s ", '(last_site = ', "$last_site) ||\n\n") if ($TEST_LEVEL_ps2pa>2);

	   my $prob_z_res = $sw_probs_z->[$s];
	   (defined  $prob_z_res) or next; # Skip if ALL descendant sequences are occupied by gaps. #

	   my $prob_x_res = $sw_probs_x->[$s];
	   my $prob_y_res = $sw_probs_y->[$s];

   if ($TEST_LEVEL_ps2pa>3) {
	print STDOUT '$prob_z_res = ', ((defined $prob_z_res) ? $prob_z_res : 'undef'), " ,\n";
	print STDOUT '$prob_x_res = ', ((defined $prob_x_res) ? $prob_x_res : 'undef'), " ,\n"; 
	print STDOUT '$prog_y_res = ', ((defined $prob_y_res) ? $prob_y_res : 'undef'), " .\n";
	print STDOUT "\n\n";
   }

	   my ($sw_probs_gapped_node, $sw_probs_ch_possins, $sw_probs_gapped_ch);
		# The sets of site-wise probabilities at the gapped node (must be either $node_x or $node_y),
		#  at the children of the other node ($node_x or $node_y), the second with possible insertion and the third occupied by gaps alone,
	   my $ch_possins;

	   my $tentative_state_pres;

	   if ((defined $prob_x_res) and (defined $prob_y_res)) { # Residues in $node_x are aligned with those in $node_y.

		$tentative_state_pres = 'M';

			# Increment the logarithmic substitution score. #

		my $increment = $LOG_ZERO;
		my $log_sbst_score_bf = $log_sbst_score;

		unless ($log_sbst_score eq $LOG_ZERO) {
		   if ($prob_z_res>0) {
			$increment = log ($prob_z_res / ($prob_x_res * $prob_y_res) );
			$log_sbst_score += $increment;
		   } else {
			$log_sbst_score = $LOG_ZERO ; 
		   }
		}

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT '[ $tentative_state_pres = ', $tentative_state_pres, " ]\n\n";
	print STDOUT '$log_sbst_score(before) = ', $log_sbst_score_bf, " ,\n";
	print STDOUT 'increment = ', $increment, " ,\n";
	print STDOUT '$log_sbst_score(after) = ', $log_sbst_score , " .\n";
	print STDOUT "\n\n";
   }

	   } elsif (!(defined $prob_x_res) ) { # ALL descendant sequences of $node_x are occupied by gaps.

		$tentative_state_pres = 'Y';

		my $ch_y = $node2ch->{$node_y};
		my ($chy1, $chy2) = (defined $ch_y) ? @{$ch_y} : (undef, undef) ;
		my $sw_probs_chy1 = (defined $chy1) ? $node2set_sw_probs->{$chy1} : undef;
		my $sw_probs_chy2 = (defined $chy2) ? $node2set_sw_probs->{$chy2} : undef;

		my $prob_chy1_res = (defined $chy1) ? $sw_probs_chy1->[$s] : undef;
		my $prob_chy2_res = (defined $chy2) ? $sw_probs_chy2->[$s] : undef;

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT '[ $tentative_state_pres = ', $tentative_state_pres, " ]\n\n";
	print STDOUT '$prob_chy1_res = ', ((defined $prob_chy1_res) ? $prob_chy1_res : 'undef'), " ,\n";
	print STDOUT '$prob_chy2_res = ', ((defined $prob_chy2_res) ? $prob_chy2_res : 'undef'), " ,\n";
	print STDOUT "\n\n";
   }

		if (!(defined $ch_y) or ((defined $prob_chy1_res) and (defined $prob_chy2_res)) ) { # Cannot "confirm" an insertion yet at this node. Thus cannot be skipped over.

			# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

		} else { # An insertion may be "confirm"ed. #

		    $sw_probs_gapped_node = $sw_probs_x;
		    ($ch_possins, $sw_probs_ch_possins, $sw_probs_gapped_ch) = (defined $prob_chy1_res) ? ($chy1, $sw_probs_chy1, $sw_probs_chy2) : ($chy2, $sw_probs_chy2, $sw_probs_chy1);
		}

	   } elsif (!(defined $prob_y_res) ) { # ALL descendant sequences of $node_y are occupied by gaps.

		$tentative_state_pres = 'X';

		my $ch_x = $node2ch->{$node_x}; 
		my ($chx1, $chx2) = (defined $ch_x) ? @{$ch_x} : (undef, undef);
		my $sw_probs_chx1 = (defined $chx1) ? $node2set_sw_probs->{$chx1} : undef;
		my $sw_probs_chx2 = (defined $chx2) ? $node2set_sw_probs->{$chx2} : undef;

		my $prob_chx1_res = (defined $chx1) ? $sw_probs_chx1->[$s] : undef;
		my $prob_chx2_res = (defined $chx2) ? $sw_probs_chx2->[$s] : undef;

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT '[ $tentative_state_pres = ', $tentative_state_pres, " ]\n\n";
	print STDOUT '$prob_chx1_res = ', ((defined $prob_chx1_res) ? $prob_chx1_res : 'undef'), " ,\n";
	print STDOUT '$prob_chx2_res = ', ((defined $prob_chx2_res) ? $prob_chx2_res : 'undef'), " ,\n";
	print STDOUT "\n\n";
   }

		if (!(defined $ch_x) or ((defined $prob_chx1_res) and (defined $prob_chx2_res)) ) { # Cannot "confirm" an insertion yet at this node. Thus cannot be skipped over.

			# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

		} else { # An insertion may be "confirm"ed. #

		    $sw_probs_gapped_node = $sw_probs_y;
		    ($ch_possins, $sw_probs_ch_possins, $sw_probs_gapped_ch) = (defined $prob_chx1_res) ? ($chx1, $sw_probs_chx1, $sw_probs_chx2) : ($chx2, $sw_probs_chx2, $sw_probs_chx1);
		}

	   } # END of "if ((defined $prob_x_res) and (defined $prob_y_res)) {...} elsif {...} elsif {...}." #


	   if (!(defined $ch_possins)) { # This site was already processed because it is already clear that the site cannot be skipped over.


		my $state_pres = $tentative_state_pres;

			# Increment the logarithmic HMM score. #

		my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
		unless (defined $hmm_score_mfac) {
		    print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s th site: $! !\n\n";
		    return (undef, undef, undef);
		}
		my $log_hmm_score_bf = $log_hmm_score;
		my $increment = log ($hmm_score_mfac);
		$log_hmm_score += $increment;

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT '[ The HMM state FIXED: $state_pres = ', $state_pres, ' ($state_last = ', $state_last, ") ]\n\n";
	print STDOUT '$hmm_score_mfac = ', $hmm_score_mfac, " .\n\n";
	print STDOUT '$log_hmm_score(before) = ', $log_hmm_score_bf, " ,\n";
	print STDOUT "{Increment}     = $increment ,\n";
	print STDOUT '$log_hmm_score(after)  = ', $log_hmm_score, " .\n";
	print STDOUT "\n\n";
   }
			# Update the variables. #
		$state_last = $state_pres;
		$last_site = $s;
		next;
	   }

		### Further examine the possible insertion. ###
 
   (print STDOUT "[ The Gap in this site might be due to an Insertion ({child node with possible insertion} = $ch_possins) ... will be examined further ... ]\n\n\n") if ($TEST_LEVEL_ps2pa>3);

   if ($TEST_LEVEL_ps2pa>5) {
	print STDOUT "[ Site-wise descendant residue probabilities at the child nodes ]\n\n";
	print STDOUT join ("\t", 'Site', 'gapped_node(XorY)',  'Child(w/possible_insertion)', 'Child(gapped)'), "\n\n";
	for (my $s1=$lb; $s1<$ub; $s1++) {  # Testing for-loop (over sites). #
	    my @out = ($s1);
	    foreach my $prob ($sw_probs_gapped_node->[$s1], $sw_probs_ch_possins->[$s1], $sw_probs_gapped_ch->[$s1]) { push @out, ((defined $prob) ? $prob : 'undef'); } 
	    print STDOUT join ("\t", @out), "\n";
	}			# End of the testing for-loop (over sites). #
	print STDOUT "\n\n";
   }

		# (1) Extend the region with the same gap-pattern (down to the child nodes). #

	   my $region_end = $s+1;
	   my $last_region_end = $s;

	   while ($region_end < $ub) { # Main inner while-loop (over sites with the possible insertion). #

		unless (defined $sw_probs_z->[$region_end]) { # Skip sites where all sequences are occupied by gaps.
		    $region_end++;
		    next;
		}

		my $prob_gapped_node = $sw_probs_gapped_node->[$region_end];
		my $prob_ch_possins = $sw_probs_ch_possins->[$region_end];
		my $prob_gapped_ch = $sw_probs_gapped_ch->[$region_end];

		(!(defined $prob_gapped_node) and (defined $prob_ch_possins) and !(defined $prob_gapped_ch)) or last; # If the gap-pattern discontinues, end the extension.

			# Update the site-recorders. #
		$last_region_end = $region_end;
		$region_end++;

	   }			# End of the main inner while-loop (over sites with the possible insertion). #

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT "[ (1) Boundaries of the region with a possible insertion (excluding sites with all gaps) ]\n\n";
	print STDOUT '{Left-flanking site} (= $last_site) = ', $last_site, " ,\n";
	print STDOUT '{Region start} (= $s) = ', $s, " ,\n";
	print STDOUT '{Region end} (= $last_region_end) = ', $last_region_end, " ,\n";
	print STDOUT '{Right-flanking site} (= $region_end) = ', $region_end, " .\n";
	print STDOUT "\n\n";
   }


		# (2) Examine the left-flanking site. #

		# (When $last_site = -1, we believe that the gaps in the subject child and the anti-subject grandchild  started exactly at the current site.) #
	   my $prob_gapped_node_lastres = ($last_site>=0) ? $sw_probs_gapped_node->[$last_site] : 1 ;
	   my $prob_ch_possins_lastres = ($last_site>=0) ? $sw_probs_ch_possins->[$last_site] : 1 ;
	   my $prob_gapped_ch_lastres = ($last_site>=0) ? $sw_probs_gapped_ch->[$last_site] : 1 ;

	   my ($prob_ch_possins_vlastres, $prob_gapped_ch_vlastres) = ($prob_ch_possins_lastres, $prob_gapped_ch_lastres);
	   my $virtual_last_site = $last_site;
	   if (!(defined $prob_ch_possins_lastres)  and !(defined $prob_gapped_ch_lastres)) { # The child with possible insertion is occupied by gaps alone.
		my $s1;
		for ($s1 = $last_site-1; $s1>=$lb; $s1--) { # 1st auxiliary inner for-loop (backward over sites to find the virtual left-flanking site).
		    $prob_ch_possins_vlastres = $sw_probs_ch_possins->[$s1];
		    $prob_gapped_ch_vlastres  = $sw_probs_gapped_ch->[$s1];
		    (!(defined $prob_ch_possins_vlastres) and !(defined $prob_gapped_ch_vlastres)) or last;
		}		# End of the 1st auxiliary inner for-loop (backward over sites to find the virtual left-flanking site). #
		if (!(defined $prob_ch_possins_vlastres) and !(defined $prob_gapped_ch_vlastres)) {
			# In this case, the virtual last site must reach the left-flanking site of the MSA itself.
		   $virtual_last_site = -1;
		   $prob_ch_possins_vlastres = $prob_gapped_ch_vlastres = 1;
		} else {
		   $virtual_last_site = $s1;
		}
	   }

	   my $if_non_smpl_insertion_lf = ((defined $prob_gapped_node_lastres) and (defined $prob_ch_possins_vlastres) and !(defined $prob_gapped_ch_vlastres)) ? 1 : 0 ; # If the parenthesized condition holds, the gap should not be due to a simple insertion.

	   my $if_confirm1_lf = ((defined $prob_gapped_node_lastres) and (defined $prob_gapped_ch_vlastres)) ? 1 : 0; # The left-end of the new gap must coincide with the left-end  of the pre-existing gap (within the sequence).

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT "[ (2) Examining the left-flanking site of the region. ]\n\n";
	print STDOUT '$prob_gapped_node_lastres = ', ((defined $prob_gapped_node_lastres) ? $prob_gapped_node_lastres : 'undef'), " ,\n" ;
	print STDOUT '$prob_ch_possins_lastres = ', ((defined $prob_ch_possins_lastres) ? $prob_ch_possins_lastres : 'undef'), " ,\n";
	print STDOUT '$prob_gapped_ch_lastres = ', ((defined $prob_gapped_ch_lastres) ? $prob_gapped_ch_lastres : 'undef'), " .\n\n" ;

	print STDOUT '{Virtual left-flanking site} (= $virtual_last_site) = ', $virtual_last_site, " .\n\n";
	print STDOUT '$prob_ch_possins_vlastres = ', ((defined $prob_ch_possins_vlastres) ? $prob_ch_possins_vlastres : 'undef'), " ,\n";
	print STDOUT '$prob_gapped_ch_vlastres = ', ((defined $prob_gapped_ch_vlastres) ? $prob_gapped_ch_vlastres : 'undef'), " .\n\n" ;

	print STDOUT '$if_non_smpl_insertion_lf = ', $if_non_smpl_insertion_lf, " ,\n";
	print STDOUT '$if_confirm1_lf = ', $if_confirm1_lf, " .\n";
	print STDOUT "\n\n";
   }


		### RESTART FROM HERE ###

		# (3) Examine the right-flanking site. #

		# (When $region_end >= $ub, we believe that the gaps in the subject child and the anti-subject grandchild ended exactly at the region end).
	   my $prob_gapped_node_rgend = ($region_end < $ub) ? $sw_probs_gapped_node->[$region_end] : 1 ;
	   my $prob_ch_possins_rgend = ($region_end < $ub) ? $sw_probs_ch_possins->[$region_end] : 1 ;
	   my $prob_gapped_ch_rgend = ($region_end < $ub) ? $sw_probs_gapped_ch->[$region_end] : 1 ;

	   my ($prob_ch_possins_vrgend, $prob_gapped_ch_vrgend) = ($prob_ch_possins_rgend, $prob_gapped_ch_rgend);

	   my $virtual_region_end = $region_end;
	   if (!(defined $prob_ch_possins_rgend) and !(defined $prob_gapped_ch_rgend)) { # The child with possible insertion is occupied by gaps alone.
		my $s1;
		for ($s1 = $region_end+1; $s1<$ub; $s1++) { # 2nd auxiliary inner for-loop (forward over sites to find the virtual right-flanking site). 
		    $prob_ch_possins_vrgend = $sw_probs_ch_possins->[$s1];
		    $prob_gapped_ch_vrgend  = $sw_probs_gapped_ch->[$s1];
		    (!(defined $prob_ch_possins_vrgend) and !(defined $prob_gapped_ch_vrgend)) or last;
		}		# End of the 2nd auxiliary inner for-loop (forward over sites to find the virtual right-flanking site). #
		if (!(defined $prob_ch_possins_vrgend) and !(defined $prob_gapped_ch_vrgend)) {
			# In this case, the virtual region end must reach the right-flanking site of the MSA itself.
		    $virtual_region_end = @{$sw_probs_z};
		    $prob_ch_possins_vrgend = $prob_gapped_ch_vrgend = 1;
		} else {
		    $virtual_region_end = $s1;
		}
	   }

	   my $if_non_smpl_insertion_rf = ((defined $prob_gapped_node_rgend) and (defined $prob_ch_possins_vrgend) and !(defined $prob_gapped_ch_vrgend)) ? 1 : 0 ; # If the parenthesized condition holds, the gap should not be due to a simple insertion.
	   my $if_confirm1_rf = ((defined $prob_gapped_node_rgend) and (defined $prob_gapped_ch_vrgend)) ? 1 : 0; # The right-end of the new gap must coincide with the right-end of the pre-existing gap (within the sequence). #


   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT "[ (3) Examining the right-flanking site of the region. ]\n\n";
	print STDOUT '$prob_gapped_node_rgend = ', ((defined $prob_gapped_node_rgend) ? $prob_gapped_node_rgend : 'undef'), " ,\n" ;
	print STDOUT '$prob_ch_possins_rgend = ', ((defined $prob_ch_possins_rgend) ? $prob_ch_possins_rgend : 'undef'), " ,\n";
	print STDOUT '$prob_gapped_ch_rgend = ', ((defined $prob_gapped_ch_rgend) ? $prob_gapped_ch_rgend : 'undef'), " .\n\n" ;

	print STDOUT '{Virtual right-flanking site} (= $virtual_region_end) = ', $virtual_region_end, " .\n\n";
	print STDOUT '$prob_ch_possins_vrgend = ', ((defined $prob_ch_possins_vrgend) ? $prob_ch_possins_vrgend : 'undef'), " ,\n";
	print STDOUT '$prob_gapped_ch_vrgend = ', ((defined $prob_gapped_ch_vrgend) ? $prob_gapped_ch_vrgend : 'undef'), " .\n\n" ;

	print STDOUT '$if_non_smpl_insertion_rf = ', $if_non_smpl_insertion_rf, " ,\n";
	print STDOUT '$if_confirm1_rf = ', $if_confirm1_rf, " .\n";
	print STDOUT "\n\n";
   }


	    if (!($if_non_smpl_insertion_lf) and !($if_non_smpl_insertion_rf)) {

		# We can skip over these sites. #
		# In this case, $state_last should remain unchanged. #

   (print STDOUT "[INFORMATION] This region, \[$s, $region_end\), will be SKIPPED OVER ...\n\n") if ($TEST_LEVEL_ps2pa>3);

		if ($if_open_forever and $if_confirm1_lf and $if_confirm1_rf) {

			# Forbid the "confirm"ed insertion to be aligned with residues from outer lineages.

   (print STDOUT "[INFORMATION] The INSERTION in this region was CONFIRMED, and thus will be OPENED Forever ...\n\n") if ($TEST_LEVEL_ps2pa>3);

		    for (my $s2=$s; $s2<=$last_region_end; $s2++) { # Main 2nd inner for-loop (over sites with the same gap-pattern). #
			(defined $sw_probs_z->[$s2]) or next; 

			my $anc_node = $node_z;

			while (defined $anc_node) { # Innermost while-loop (climbing the ancestry of $node_id). #

			    $node2set_sw_probs->{$anc_node}[$s2] = 0; # Set the probability of descendant residues to be zero, because they are NOT descended from any residue at this node.

			    $anc_node = $node2pa->{$anc_node}; # Update the ancestral node.

			}			# End of the innermost while-loop (climbing the ancestry of $node_id). #
			
		    }			# End of the main 2nd inner for-loop (over sites with the same gap-pattern). #
		}

   (print STDOUT "\n") if ($TEST_LEVEL_ps2pa>3);

		    # Update the site-recorders. #

		$last_site = $last_region_end; # It might have been better that this line be in the "if ($if_non_insertion_lf or $if_non_insertion_rf)" block above.
		$s = $region_end - 1;

		next;
	   }


		## (5) Now, calculate the HMM score,
		## while looking into the states of the grand-children in the lineage of possible insertion. ##

   (print STDOUT "[ (5) Calculate the HMM score over the region, \[$s, $last_region_end\], (with initial ", '$state_last',  " = $state_last), while looking into the staes of grand-children ... ]\n\n\n") if ($TEST_LEVEL_ps2pa>3);

	   my ($sw_probs_grch_possins, $sw_probs_gapped_grch);
		# The sets of site-wise probabilities at the children of the child with possible insertion, the 1st with possible insertion and the 2nd occupied by gaps alone. (These are defined only if relevant).

	   my $ch_ch_possins = $node2ch->{$ch_possins};
	   my ($grch1, $grch2) = (defined $ch_ch_possins) ? @{$ch_ch_possins} : (undef, undef);
	   my $sw_probs_grch1 = (defined $ch_ch_possins) ? $node2set_sw_probs->{$grch1} : undef;
	   my $sw_probs_grch2 = (defined $ch_ch_possins) ? $node2set_sw_probs->{$grch2} : undef;

   if ($TEST_LEVEL_ps2pa>5) {
	print STDOUT "[ Site-wise descendant residue probabilities at grand-child nodes with (a) possible insertion(s).  ]\n\n";
	if (defined $ch_ch_possins) {
	    print STDOUT join ("\t", 'Site', 'grand_ch1', 'grand_ch2'), "\n\n";
	    for (my $s2 = $lb; $s2<$ub; $s2++) {
		my @out = ($s2);
		foreach my $prob ($sw_probs_grch1->[$s2], $sw_probs_grch2->[$s2]) { push @out, ((defined $prob) ? $prob: 'undef'); }
		print STDOUT join ("\t", @out), "\n";
	    }

	} else {

	   print STDOUT "... NO CHILD NODES of the child node '$ch_possins' with possible insertion ...\n";
	}
	print STDOUT "\n\n";
   }


	   my @possible_ins_regions = (); # element = [$pi_rgstart, $pi_rgend], where the coordinates are in the full-closed converntion.

	   my $state_pres = $tentative_state_pres;

	   my $lfsite_posins_rg = $virtual_last_site;
	   for (my $s2=$s; $s2<=$last_region_end; $s2++) { # Main 2nd inner for-loop (over sites with the same gap-pattern). #

		(defined $sw_probs_z->[$s2]) or next;

		if ((defined $ch_ch_possins) and !((defined $sw_probs_grch1->[$s2]) and (defined $sw_probs_grch2->[$s2])) ) { # One of the grand-children is occupied by gaps only.

		    ($sw_probs_grch_possins, $sw_probs_gapped_grch) = (defined $sw_probs_grch1->[$s2]) ? ($sw_probs_grch1, $sw_probs_grch2) : ($sw_probs_grch2, $sw_probs_grch1)  ; 
		    my $pi_rgstart = my $pi_rgend = $s2;
		    my $s3=$s2+1;
		    for (; $s3<=$last_region_end; $s3++) { # Inner-inner for-loop (over sites with the same gap-pattern down to the grandchildren). #
			(defined $sw_probs_z->[$s3]) or next;
			((defined $sw_probs_grch_possins->[$s3]) and !(defined $sw_probs_gapped_grch->[$s3])) or last;

			$pi_rgend = $s3; # The site exhibits the same gap-pattern. #

		    } # End of the inner-inner for-loop (over sites with the same gap-pattern down to the grandchildren). #


			## The new condition ADDED to implement the PRANK scoring scheme more faithfully (on Sep 26 (Fri), 2014). ##

		    my $prob_gapped_grch_lf_posins_rg = ($lfsite_posins_rg >= $lb) ? $sw_probs_gapped_grch->[$lfsite_posins_rg] : 1;
		    my $rfsite_posins_rg = $s3;
		    for (; $rfsite_posins_rg<$ub; $rfsite_posins_rg++) { (defined $sw_probs_z->[$s3]) and last; }
		    my $prob_gapped_grch_rf_posins_rg = ($rfsite_posins_rg < $ub) ? $sw_probs_gapped_grch->[$rfsite_posins_rg] : 1;
		    if ((defined $prob_gapped_grch_lf_posins_rg) and (defined $prob_gapped_grch_rf_posins_rg)) {
		    
			# This region will be skipped over. #

			push @possible_ins_regions, [$pi_rgstart, $pi_rgend];

   (print STDOUT ' ' x 3, "[INFORMATION] The region, \[$pi_rgstart, $s3\), was SKIPPED OVER ...\n\n") if ($TEST_LEVEL_ps2pa>=5);

		    } else {

			# This region will also be counted into the HMM scoring. #
			# (This block was also ADDED to implement the PRANK scoring scheme more faithfully (on Sep 26 (Fri), 2014).

			for (my $s5=$pi_rgstart; $s5<=$pi_rgend; $s5++) { # Auxiliary inner-inner for-loop (over sites in the region with a confounding pattern but was finally judged as non-insertion). #

			    (defined $sw_probs_z->[$s5]) or next; 

   (print STDOUT ' ' x 3, "| Site = $s5  (", '$state_last = ', $state_last, ', $state_pres = ', $state_pres, ") |\n\n") if ($TEST_LEVEL_ps2pa>=5);

			    my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
			    unless (defined $hmm_score_mfac) {
				print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s5 th site (in the extended region starring from the $s th site): $! !\n\n";
				return (undef, undef, undef);
			    }
			    my $log_hmm_score_bf = $log_hmm_score;
			    my $increment = log ($hmm_score_mfac);
			    $log_hmm_score += $increment;

				# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

   if ($TEST_LEVEL_ps2pa>5) {
	print STDOUT ' ' x 3, '$hmm_score_mfac = ', $hmm_score_mfac, " .\n\n";
	print STDOUT ' ' x 3, '$log_hmm_score(before) = ', $log_hmm_score_bf, " ,\n";
	print STDOUT ' ' x 3, "{Increment}  = $increment ,\n";
	print STDOUT ' ' x 3, '$log_hmm_score(after)  = ', $log_hmm_score, " .\n";
	print STDOUT "\n\n";
   }

				# Update the last state. #

			    $state_last = $state_pres;

			} # END of the auxiliary inner-inner for-loop (over sites in the region with a confounding pattern but was finally judged as non-insertion). #

		    } # END of "if ((defined $prob_gapped_grch_lf_posins_rg) and (defined $prob_gapped_grch_rf_posins_rg)) {...} else {...}." #

		    $lfsite_posins_rg = $pi_rgend;
		    $s2 = $s3-1; # so that the next round will start at $s2 = $s3 (>= $pi_rgend+1).
		    next;
		}

			# A tangible gap is considered to be inserted in this site. #
			# The site's state should automatically be $state_pres. #

   (print STDOUT ' ' x 3, "| Site = $s2  (", '$state_last = ', $state_last, ', $state_pres = ', $state_pres, ") |\n\n") if ($TEST_LEVEL_ps2pa>=5);

		my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
		unless (defined $hmm_score_mfac) {
			print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s2 th site (in the extended region starring from the $s th site): $! !\n\n";
			return (undef, undef, undef);
		}
		my $log_hmm_score_bf = $log_hmm_score;
		my $increment = log ($hmm_score_mfac);
		$log_hmm_score += $increment;

			# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

   if ($TEST_LEVEL_ps2pa>5) {
	print STDOUT ' ' x 3, '$hmm_score_mfac = ', $hmm_score_mfac, " .\n\n";
	print STDOUT ' ' x 3, '$log_hmm_score(before) = ', $log_hmm_score_bf, " ,\n";
	print STDOUT ' ' x 3, "{Increment}  = $increment ,\n";
	print STDOUT ' ' x 3, '$log_hmm_score(after)  = ', $log_hmm_score, " .\n";
	print STDOUT "\n\n";
   }

			# Update the last state. #
			# (If no site shows this type of pattern,
			#  where both grand-children are occupied by residues,
			#  the $state_last should NEVER be updated.)

		$state_last = $state_pres;

		$lfsite_posins_rg = $s2;

	   }			# End of the main 2nd inner for-loop (over sites with the same gap-pattern). #


	   my $if_confirm2 = 0; # By default, not "confirm" the possible insertion. #

	   if ($if_open_forever and (@possible_ins_regions==1) and ($possible_ins_regions[0]->[1] == $last_region_end)) {

			## (6) Seek the possibility of "confirm"ing the insertion. ##

		my $pi_rgstart = $possible_ins_regions[0]->[0]; 
		my $if_nontrivial=0;
		for (my $s3=$s; $s3<$pi_rgstart; $s3++) { (defined $sw_probs_z->[$s3]) and ($if_nontrivial = 1) and last; }

		unless ($if_nontrivial) { # No nontrivial sites between $s and $pi_rgstart. #

   (print STDOUT "[ (6) Examine whether an insertion of the homogeneous gap-pattern region, \[$s, $last_region_end\], can be 'confirm'ed or not ... ]\n\n") if ($TEST_LEVEL_ps2pa>3);

		    my $prob_gapped_grch_vlastres = ($virtual_last_site>=0) ? $sw_probs_gapped_grch->[$virtual_last_site] : 1 ;
		    my $if_confirm2_lf = ((defined $prob_gapped_node_lastres) and (defined $prob_gapped_grch_vlastres)) ? 1 : 0;

		    my $prob_gapped_grch_vrgend = ($virtual_region_end < $ub) ? $sw_probs_gapped_grch->[$virtual_region_end] : 1 ;
		    my $if_confirm2_rf = ((defined $prob_gapped_node_rgend) and (defined $prob_gapped_grch_vrgend)) ? 1 : 0;
		    ($if_confirm2_lf and $if_confirm2_rf) and ($if_confirm2 = 1);

   if ($TEST_LEVEL_ps2pa>=5) {
	print STDOUT '$prob_gapped_node_lastres = ', ((defined $prob_gapped_node_lastres) ? $prob_gapped_node_lastres : 'undef'), " ,\n";
	print STDOUT '$prob_gapped_grch_vlastres = ', ((defined $prob_gapped_grch_vlastres) ? $prob_gapped_grch_vlastres : 'undef'), " ,\n";
	print STDOUT '$if_confirm2_lf = ', $if_confirm2_lf, " .\n\n";

	print STDOUT '$prob_gapped_node_rgend = ', ((defined $prob_gapped_node_rgend) ? $prob_gapped_node_rgend : 'undef'), " ,\n";
	print STDOUT '$prob_gapped_grch_vrgend = ', ((defined $prob_gapped_grch_vrgend) ? $prob_gapped_grch_vrgend : 'undef'), " ,\n";
	print STDOUT '$if_confirm2_rf = ', $if_confirm2_rf, " .\n\n";
	print STDOUT "\n";
   }

		}
	   }

	   if ($if_confirm2) {

			# The possible insertion was "confirm"ed at the last minute,
			# thus make the sites unalignable with residues from outer sequences.

		for (my $s2=$s; $s2<=$last_region_end; $s2++) { # Main 2nd inner for-loop (over sites with the same gap-pattern). #

		    (defined $sw_probs_z->[$s2]) or next; 

		    my $anc_node = $node_z;

		    while (defined $anc_node) { # Innermost while-loop (climbing the ancestry of $node_id). #

			$node2set_sw_probs->{$anc_node}[$s2] = 0; # Set the probability of descendant residues to be zero, because they are NOT descended from any residue at this node.

			$anc_node = $node2pa->{$anc_node}; # Update the ancestral node.

		    }			# End of the innermost while-loop (climbing the ancestry of $node_id). #
			
		}		# End of the main 2nd inner for-loop (over sites with the same gap-pattern). #

   (print STDOUT "[INFORMATION] The Gap in the region was 'confirm'ed to be due to an insertion, thus the gap will be open forever ...\n\n\n") if ($TEST_LEVEL_ps2pa>3);

	   }

		# Update the site-recorders. #

	   $last_site = $last_region_end; # It might have been better that this line be in the "if ($if_non_insertion_lf or $if_non_insertion_rf)" block above.
	   $s = $region_end - 1;

	}			# End of the main outer for-loop (over sites to be examined). #

		# Finally, add (the logarithm of) the HMM score for ending the regional MSA. #

	my $fnl_hmm_score_mfac = $trans_states2score->{"$state_last:$state_rf"};
	unless (defined $fnl_hmm_score_mfac) {
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_rf' at the end of the regional MSA: $! !\n\n";
	    return (undef, undef, undef);
	}

	my $log_hmm_score_bf = $log_hmm_score;
	my $fnl_increment = log ($fnl_hmm_score_mfac);
	$log_hmm_score += $fnl_increment;

   if ($TEST_LEVEL_ps2pa>5) {
	print STDOUT "[ Final Calculation of the HMM Score (adding transition from '$state_last' to '$state_rf') ]\n\n";
	print STDOUT '$fnl_hmm_score_mfac = ', $fnl_hmm_score_mfac, " .\n\n";
	print STDOUT '$log_hmm_score(before) = ', $log_hmm_score_bf, " ,\n";
	print STDOUT '{Final increment} = ', $fnl_increment, " ,\n";
	print STDOUT '$log_hmm_score(after) = ', $log_hmm_score, " .\n";
	print STDOUT "\n\n";
   }

		### RESTART FROM HERE (II) ###


	my $log_prank_score = ($log_sbst_score eq $LOG_ZERO) ? $LOG_ZERO : $log_sbst_score + $log_hmm_score;

   (print STDOUT "< End of the subroutine 'prank_score_2profile_aln' (", '$log_prank_score = ', $log_prank_score, ', $log_sbst_score = ', $log_sbst_score, ', $log_hmm_score = ', $log_hmm_score, ') >', "\n\n") if ($TEST_LEVEL_ps2pa>0);

	return ($log_prank_score, $log_sbst_score, $log_hmm_score);

}	# End of the subroutine 'prank_score_2profile_aln'. #


=This version of the subroutine does not faithfully implement the scoring scheme in PRANK, which only concerns whether or not each of sequence profiles to be aligned consists of "residue vs gap" or "residue vs residue" ("gap vs gap" is just a newly introducted gap).

# prank_score_2profile_aln:
#	This subroutine scores a given profile-profile alignment (i.e., the alignment of two sub-MSAs) at a progressive alignment step
#	corresponding to a specified node (of the guide tree), according to the scheme of PRANK.
#	As a SIDE-EFFECT, the subroutine also assigns ZERO probabilities to the nodes ancestral to a branch,
#	at the site that are "confirmed" as inserted along the branch.
#	(CREATION STARTED on Sep 22 (Mon), 2014, ENDED on Sep 23? (Tue), 2014?)
#
# INPUT: ($node_id, \%node2set_sw_probs, \%node2ch, \%node2pa, \%trans_states2score, $state_lf, $state_rf, $if_open_forever, $if_ignore_term_gaps), where
#	$node_id is the node corresponding to the progressive alignment step.
#	That is, the profiles (sub-MSAs) reconstructed at the child nodes of $node_id will be aligned (or, actually, was aligned);
#
#	%node2set_sw_probs = ($node_id => \@set_sw_probs, ...) for ALL nodes, where
#	$set_sw_probs[$s] is the probability that all the descendant sequences of $node_id show the residue that are actually found in the $s th site.
#	But $set_sw_probs[$s] = undef if all the descendant sequences are occupied by gaps at the $s th site.
#	This hash is typically an output of 'calc_prob_descresstats';
#
#	%node2ch = ($node_id => \@child_nodes, ....) for interior nodes;
#	%node2pa = ($node_id => $parent, ...) for nodes except the top node (root);
#
#	%trans_states2score = ("$state_last:$state_pres" => $score, ....), where $state_last/pres is the state at the last/present site (column) in the currently reconstructed alignment, and
#	$state_last/pres = 'M' (if the two sites are aligned),
#			 = 'X' (if residues in the 1st profile is aligned with the gap-only column in the 2nd profile),
#			 = 'Y' (if residues in the 2nd profile is aligned with the gap-only column in the 1st profile),
#			 = 'S' (for the alignment start),
#			 = 'E' (for the alignment end).
#	$score is the score itself, not the logarithm;
#
#	$state_lf/rf is the state at the left/right-flanking column (excluding all-gap columns) of the currently reconstructed alignment;
#
#	$if_open_forever = 1 if you want the "confirmed" insertions to open forever (i.e., never alignable with residues),
#			 = 0 if you want to leave the "confirmed" insertions to be alignable with residues again.
#
#	$if_ignore_term_gaps = 1 if you ignore the gaps at the end of the alignment,
#			     = 0 if you include them them into the scoring.
#
# OUTPUT: ($log_prank_score, $log_sbst_score, $log_hmm_score), where
#	$log_prank_score is the logarithm of the score of the alignment.
#		(= $LOG_ZERO if the score becomes zero (i.e., the alignment is impossible));
#	$log_sbst_score is the logarithm of the portion of the score concerning substitutions;
#	$log_hmm_score is the logarithm of the portion of the score concerning gap-patterns (specified by the alignment HMM of PRANK).
#
#	Usually, $log_prank_score = $log_sbst_score + $log_hmm_score.
#
sub prank_score_2profile_aln ($\%\%\%\%$$$$) {

	my ($node_id, $node2set_sw_probs, $node2ch, $node2pa, $trans_states2score, $state_lf, $state_rf, $if_open_forever, $if_ignore_tgaps) = @_;


	unless (($state_lf eq 'M') or ($state_lf eq 'S')) {
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The left-flanking state, '$state_lf', is UNEXPECTED: $! !\n\n";
	    return (undef, undef, undef);
	}
	unless (($state_rf eq 'M') or ($state_rf eq 'E')) {
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The right-flanking state, '$state_rf', is UNEXPECTED: $! !\n\n";
	    return (undef, undef, undef);
	}


		# Preliminary process (I): Prepare the sets of site-wise probabilities at the subject node, child nodes, and grand-child nodes. #

	my $sw_probs_node = $node2set_sw_probs->{$node_id};
	my $ct_sites = @{$sw_probs_node};

	my $children = $node2ch->{$node_id};

	unless ((defined $children) and (@{$children}==2)) {
	    my $ct_ch = (defined $children) ? @{$children} : 0;
	    print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The subject node '$node_id' has $ct_ch children, which is different from 2 as we expected: $! !\n\n";
	    return (undef, undef, undef);
	}
	my @set_sw_probs_ch = ();
	foreach my $ch (@{$children}) { push @set_sw_probs_ch, $node2set_sw_probs->{$ch}; }

	my ($ch1, $ch2) = @{$children};

	my $gr_children1 = $node2ch->{$ch1};
	my $gr_children2 = $node2ch->{$ch2};

	my @sw_probs_grch1 = my @sw_probs_grch2 = ();

	if (defined $gr_children1) {
	    unless (@{$gr_children1}==2) {
		my $ct_grch1 = @{$gr_children1};
		print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The child node '$ch1' of the node '$node_id' has $ct_grch1 children, which is different from 2 as we expected: $! !\n\n";
		return (undef, undef, undef);
	    }
	    foreach my $grch1 (@{$gr_children1}) { push @sw_probs_grch1, $node2set_sw_probs->{$grch1}; }
	}

	if (defined $gr_children2) {
	    unless (@{$gr_children2}==2) {
		my $ct_grch2 = @{$gr_children2};
		print STDOUT "[ERROR 'prank_score_2profile_aln'] ILLEGAL INPUT: The child node '$ch2' of the node '$node_id' has $ct_grchs2 children, which is different from 2 as we expected: $! !\n\n";
		return (undef, undef, undef);
	    }
	    foreach my $grch2 (@{$gr_children2}) { push @sw_probs_grch2, $node2set_sw_probs->{$grch2}; }
	}

	my @set_sw_probs_grch = (\@sw_probs_grch1, \@sw_probs_grch2);

		# Preliminary process (II): Set the site regions to be examined (in the 0-based half-open convention). #

	my $lb = 0;
	my $ub = $ct_sites;

	if ($if_ignore_tgaps) {
	   if ($state_lf = 'S') {
		for (my $s=0; $s<$ct_sites; $s++) { # 1st preliminary for-loop (over sites). #
		    ((defined $set_sw_probs_ch[0]->[$s]) and (defined $set_sw_probs_ch[1]->[$s])) and last;
		    $lb++;
		}			# End of the 1st preliminary for-loop (over sites). #
	   }

	   if ($state_rf = 'E') {
		for (my $s=$ct_sites-1; $s>=$lb; $s--) { # 2nd preliminary for-loop (over sites). #
		    ((defined $set_sw_probs_ch[0]->[$s]) and (defined $set_sw_probs_ch[1]->[$s])) and last;
		    $ub--;
		}			# End of the 2nd preliminary for-loop (over sites). #
	   }
	}

		### MAIN PROCESS: Score the profile-profile alignment reconstructed at $node_id. ###

	my $state_last = $state_lf;
	my $last_site = -1;

	my $log_sbst_score = my $log_hmm_score = 0;

	for (my $s=$lb; $s<$ub; $s++) { # Main outer for-loop (over sites to be examined). #

	   my $prob_node_res = $sw_probs_node->[$s];
	   (defined  $prob_node_res) or next; # Skip if ALL descendant sequences are occupied by gaps. #

	   my $prob_ch0_res = $set_sw_probs_ch[0]->[$s];
	   my $prob_ch1_res = $set_sw_probs_ch[1]->[$s];

	   my $indx_possible_insertion; # Specifies the child lineage where an insertion is suspected.
	   my $indx_possible_insertion2; # Specifies the grand-child onto which an insertion is suspected.
	   my $state_pres;

	   if ((defined $prob_ch0_res) and (defined $prob_ch1_res)) {

		$state_pres = 'M';
		my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
		unless (defined $hmm_score_mfac) {
		    print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s th site: $! !\n\n";
		    return (undef, undef, undef);
		}
		$log_hmm_score += log ($hmm_score_mfac);

		my $sbst_score_term = ($prob_node_res > 0) ? log ($prob_node_res / ($prob_ch0_res*$prob_ch1_res) ) : $LOG_ZERO ; 
		unless ($log_sbst_score eq $LOG_ZERO) {
		   ($sbst_score_term eq $LOG_ZERO) ? ($log_sbst_score = $LOG_ZERO) : ($log_sbst_score += $sbst_score_term) ;
		}

	   } elsif (!(defined $prob_ch0_res) ) { # ALL descendant sequences of $childlen->[0] are occupied by gaps.

		my $sw_probs_sbj_grch_res = (@set_sw_probs_grch>0) ? $set_sw_probs_grch[1] : [[], []];
		my $prob_sbj_grch0_res = $sw_probs_sbj_grch_res->[0][$s];
		my $prob_sbj_grch1_res = $sw_probs_sbj_grch_res->[1][$s];

		if ((@set_sw_probs_grch==0) or ((defined $prob_sbj_grch0_res) and (defined $prob_sbj_grch1_res)) ) { # Cannot "confirm" an insertion yet at this node.
		    $state_pres = 'Y';
		    my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
		    unless (defined $hmm_score_mfac) {
			print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s th site: $! !\n\n";
			return (undef, undef, undef);
		    }
		    $log_hmm_score += log ($hmm_score_mfac);

			# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

		} else { # An insertion may be "confirm"ed. #
		    $indx_possible_insertion = 1;
		    $indx_possible_insertion2 = (defined $prob_sbj_grch0_res) ? 0 : 1;
		}

	   } elsif (!(defined $prob_ch1_res) ) { # ALL descendant sequences of $children->[1] are occupied by gaps.

		my $sw_probs_sbj_grch_res = (@set_sw_probs_grch>0) ? $set_sw_probs_grch[0] : [[], []];
		my $prob_sbj_grch0_res = $sw_probs_sbj_grch_res->[0][$s];
		my $prob_sbj_grch1_res = $sw_probs_sbj_grch_res->[1][$s];

		if ((@set_sw_probs_grch==0) or ((defined $prob_sbj_grch0_res) and (defined $prob_sbj_grch1_res)) ) { # Cannot "confirm" an insertion yet at this node.
		    $state_pres = 'X';
		    my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
		    unless (defined $hmm_score_mfac) {
			print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s th site: $! !\n\n";
			return (undef, undef, undef);
		    }
		    $log_hmm_score += log ($hmm_score_mfac);

			# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

		} else { # An insertion may be "confirm"ed. #
		    $indx_possible_insertion = 0;
		    $indx_possible_insertion2 = (defined $prob_sbj_grch0_res) ? 0 : 1;
		}
	   }

	   unless (defined $indx_possible_insertion) { # Cannot yet confirm that the gap introduced here, if any, is due to an insertion in a grandchildren (or its descendant).
		$state_last = $state_pres;
		$last_site = $s;
		next;
	   }

		## Further examine the possible insertion. ##

	   my $sw_probs_sbj_ch = $set_sw_probs_ch[1 - $indx_possible_insertion]; # The child node with all gaps.
	   my $sw_probs_sbj_grch = $set_sw_probs_grch[$indx_possible_insertion]->[$indx_possible_insertion2]; # The grandchild node with residues.
	   my $sw_probs_anti_grch = $set_sw_probs_grch[$indx_possible_insertion]->[1 - $indx_possible_insertion2]; # The grandchild node with all gaps. 

		# (1) Extend the region with the same gap-pattern. #

	   my $region_end = $s+1;
	   my $last_region_end = $s;

	   while ($region_end < $ub) { # Main inner while-loop (over sites with the possible insertion). #

		unless (defined $sw_probs_node->[$region_end]) { # Skip sites where all sequences are occupied by gaps.
		    $retion_end++;
		    next;
		}

		my $prob_sbj_ch = $sw_probs_sbj_ch->[$region_end];
		my $prob_sbj_grch = $sw_probs_sbj_grch->[$region_end];
		my $prob_anti_grch = $sw_probs_anti_grch->[$region_end];

		(!(defined $prob_sbj_ch) and (defined $prob_sbj_grch) and !(defined $prob_anti_grch)) or last; # If the gap-pattern discontinues, end the extension.

			# Update the site-recorders. #
		$last_region_end = $region_end;
		$region_end++;

	   }			# End of the main inner while-loop (over sites with the possible insertion). #


		# (2) Examine the left-flanking site. #

		# (When $last_site = -1, we believe that the gaps in the subject child and the anti-subject grandchild  started exactly at the current site.) #
	   my $prob_sbj_ch_lastres = ($last_site>=0) ? $sw_probs_sbj_ch->[$last_site] : 1 ;
	   my $prob_sbj_grch_lastres = ($last_site>=0) ? $sw_probs_sbj_grch->[$last_site] : 1 ;
	   my $prob_anti_grch_lastres = ($last_site>=0) ? $sw_probs_anti_grch->[$last_site] : 1 ;

	   my $if_non_insertion_lf = ((defined $prob_sbj_ch_lastres) and (defined $prob_sbj_grch_lastres) and !(defined $prob_anti_grch_lastres)) ? 1 : 0 ; # If the parenthesized condition holds, the gap should not be due to an insertion.
	   my $if_confirm_lf = ((defined $prob_sbj_ch_lastres) and (defined $prob_anti_grch_lastres)) ? 1 : 0;

		# (3) Examine the right-flanking site. #

		# (When $region_end >= $ub, we believe that the gaps in the subject child and the anti-subject grandchild ended exactly at the region end).
	   my $prob_sbj_ch_rgend = ($region_end < $ub) ? $sw_probs_sbj_ch->[$region_end] : 1 ;
	   my $prob_sbj_grch_rgend = ($region_end < $ub) ? $sw_probs_sbj_grch->[$region_end] : 1 ;
	   my $prob_anti_grch_rgend = ($region_end < $ub) ? $sw_probs_anti_grch->[$region_end] : 1 ;

	   my $if_non_insertion_rf = ((defined $prob_sbj_ch_rgend) and (defined $prob_sbj_grch_rgend) and !(defined $prob_anti_grch_rgend)) ? 1 : 0 ; # If the parenthesized condition holds, the gap should not be due to an insertion.
	   my $if_confirm_rf = ((defined $prob_sbj_ch_rgend) and (defined $prob_anti_grch_rgend)) ? 1 : 0;


	   if ($if_non_insertion_lf or $if_non_insertion_rf) {
		# The gap should NOT be due to an insertion. => These sites canNOT be skipped over. #

		$state_pres = ($indx_possible_insertion==0) ? 'X' : 'Y' ;

		for (my $s2=$s; $s2<=$last_region_end; $s2++) { # Main 2nd inner for-loop (over sites with the same gap-pattern). #
		    (defined $sw_probs_node->[$s2]) or next;

		    my $hmm_score_mfac = $trans_states2score->{"$state_last:$state_pres"};
		    unless (defined $hmm_score_mfac) {
			print STDOUT "[ERROR 'prank_score_2profile_aln'] UNDEFINED SCORE for the transition from '$state_last' to '$state_pres' at the $s2 th site (in the extended region starring from the $s th site): $! !\n\n";
			return (undef, undef, undef);
		    }
		    $log_hmm_score += log ($hmm_score_mfac);

			# (Do nothing about $log_sbst_score, because the summed term here is log (1) = 0.) #

		    $state_last = $state_pres;

		}			# End of the main 2nd inner for-loop (over sites with the same gap-pattern). #

	   } else {
		# We can skip over these sites. #
		# In this case, $state_last should remain unchanged. #
		# (But, what about $last_site? .. The choice could change when possible insertions be "confirm"ed.)

		if ($if_open_forever and $if_confirm_lf and $if_confirm_rf) {

			# Forbid the "confirm"ed insertion to be aligned with residues from outer lineages.

		    for (my $s2=$s; $s2<=$last_region_end; $s2++) { # Main 3rd inner for-loop (over sites with the same gap-pattern). #
			(defined $sw_probs_node->[$s2]) or next; 

			my $anc_node = $node_id;

			while (defined $anc_node) { # Innermost while-loop (climbing the ancestry of $node_id). #

			    $node2set_sw_probs->{$anc_node}[$s2] = 0; # Set the probability of descendant residues to be zero, because they are NOT descended from any residue at this node.

			    $anc_node = $node2pa->{$anc_node}; # Update the ancestral node.

			}			# End of the innermost while-loop (climbing the ancestry of $node_id). #
			
		    }			# End of the main 3rd inner for-loop (over sites with the same gap-pattern). #
		
		}
	   }


		# Update the site-recorders. #

	   $last_site = $last_region_end; # It might have been better that this line be in the "if ($if_non_insertion_lf or $if_non_insertion_rf)" block above.
	   $s = $region_end - 1;

	}			# End of the main outer for-loop (over sites to be examined). #


	my $log_prank_score = ($log_sbst_score eq $LOG_ZERO) ? $LOG_ZERO : $log_sbst_score + $log_hmm_score;

	return ($log_prank_score, $log_sbst_score, $log_hmm_score);

}	# End of the subroutine 'prank_score_2profile_aln'. #

=cut # END of "This version of the subroutine does not faithfully implement the scoring scheme in PRANK, which only concerns whether or not each of sequence profiles to be aligned consists of "residue vs gap" or "residue vs residue" ("gap vs gap" is just a newly introducted gap)."


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


