package MyDetect_cmplx_error_cands;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyDetect_cmplx_error_cands.pm:
#	This package is a collection of subroutines
#	necessary for detecting the candidates of regions containing complex errors.
#	[CREATION STARTED on December 10th (Sat), 2016, ENDED on Mmm dd (www), yyyy.]
#
#	[REVISED on 2019/06/12]:
#	Added subroutines, 'mk_ingredients_acls_gpd_segs' & '',
#	to facilitate the aritficial clustering of gapped segments.
#

use strict;
use warnings;

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTaxonomy qw(
	fetch_common_ancestors
	);
	# if_ancestor fetch_ancestry calculate_abs_depth
	# fetch_ext_offsprings_nr fetch_all_offsprings_nr

use MyLikelihoods_respattern_msa_dna qw( );
	# calc_node2ntfreqs_td
	# calc_prob_descresstats_cond_ancstat2
	# calc_joint_prob_ancresstats_and_ancstat2
	# calc_clm_probs

=Set aside for potential use (I)

use MyTreeManipu_III qw( make_node2abs_depth );

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );

use MyProbDistri qw( binomial_cum_via_betai );

=cut # END of "Set aside for potential use (I)" #
 

	# Added in REVISION on 2019/06/12. #

use MyTreeMap_indels_sppl qw(
	mk_eventset
	);
use MyTreeMap_indels_spt_odr2 qw(
	br_list_gblocks
	);
	# End of "Added in REVISION on 2019/06/12." #

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	selfcompare_indel_hosting_branches
	compare_indel_hosting_branches
	mk_self_nn_events
	mk_ingredients_acls_gpd_segs
	artf_cluster_gapped_segs
	artf_cluster_gapped_segs_len
	init_stats_artf_clusters
	stats_artf_clusters
	mk_labels_stats_artf_clusters
	pr_heading_stats_artf_clusters
	pr_stats_artf_clusters
	detect_cmplx_error_cand_acls0
	wrapper_detect_cmplx_error_cand_acls0
	);
	# Added 'mk_ingredients_acls_gpd_segs' on 2019/06/12.

#sub mk_ingredients_acls_gpd_segs (\@\@\@$\%\%\%); # Added & Disabled on 2019/06/12.

	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = $MyLikelihoods_respattern_msa_dna::LOG_ZERO;

	## Representative Sets of Parameters for 'artf_cluster_gapped_segs' and 'artf_cluster_gapped_segs_len'. ##
	## (ADDED on Jan 14 (Sun), 2018.) ##

	# Parameters for 'artf_cluster_gapped_segs' (cum%{complex} = 55.12% & cum%{non-complex} = 89.81% on 15 mammals)
our @PARAMS_ACGS_15MAMM_NCMPLX90
	= ( # $upper_bound_spsize, $thrsh_spsize0, $mag_thrsh, [ list of [$totct_indels, $totct_ins - $totct_del] ]
	20, 4, 2,
	[[1,-1], [1,1], [2,-2], [2,0], [2,2], [3,-1], [3,-3], [3,1], [4,-4], [3,3], [4,-2], [6,-6], [5,-5], [4,0]]
	);

	# Parameters (1) for 'artf_cluster_gapped_segs_len' (cum%{complex} = 35.88% & cum%{non-complex} = 78.71% on 15 mammals)
our @PARAMS_ACGS_LEN_15MAMM_NCMPLX80a
	= ( # $upper_bound_spsize, $thrsh_spsize0, $mag_thrsh, [ list of [$totct_indels, $totct_ins - $totct_del] ]
	20, 6, 2,
	[[1,-1], [1,1], [2,-2], [2,0], [2,2], [3,-1], [3,-3], [3,1], [3,3], [4,-4], [4,-2]] 
	);

	# Parameters (2) for 'artf_cluster_gapped_segs_len' (cum%{complex} = 35.10% & cum%{non-complex} = 77.50% on 15 mammals)
our @PARAMS_ACGS_LEN_15MAMM_NCMPLX80b
	= ( # $upper_bound_spsize, $thrsh_spsize0, $mag_thrsh, [ list of [$totct_indels, $totct_ins - $totct_del] ]
	15, 5, 2,
	[[1,-1], [1,1], [2,-2], [2,0], [2,2], [3,-1], [3,-3]] 
	);

	# Parameters (3) for 'artf_cluster_gapped_segs_len' (cum%{complex} = 52.61% & cum%{non-complex} = 89.44% on 15 mammals)
our @PARAMS_ACGS_LEN_15MAMM_NCMPLX90a
	= ( # $upper_bound_spsize, $thrsh_spsize0, $mag_thrsh, [ list of [$totct_indels, $totct_ins - $totct_del] ]
	10, 2, 2,
	[[1,-1], [1,1], [2,-2], [2,0], [2,2], [3,-3], [3,-1], [3,1], [4,-4], [3,3], [5,-5], [4,-2], [5,-3]] 
	);

	# Parameters (4) for 'artf_cluster_gapped_segs_len' (cum%{complex} = 54.08% & cum%{non-complex} = 90.06% on 15 mammals)
our @PARAMS_ACGS_LEN_15MAMM_NCMPLX90b
	= ( # $upper_bound_spsize, $thrsh_spsize0, $mag_thrsh, [ list of [$totct_indels, $totct_ins - $totct_del] ]
	10, 3, 2,
	[[1,-1], [1,1], [2,-2], [2,0], [2,2], [3,-3], [3,-1], [3,1], [3,3], [4,-4], [5,-5], [4,-2], [4,0], [5,-3], [5,-1]] 
	);

	## END of "(ADDED on Jan 14 (Sun), 2018.)" ##


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'mk_ingredients_acls_gpd_segs'. # (Added on 2019/06/12.)
our $TEST_LEVEL_migacgss = 0;
##our $TEST_LEVEL_migacgss = 5;

	# For 'artf_cluster_gapped_segs'. #
our $TEST_LEVEL_acgs = 0;
##our $TEST_LEVEL_acgs = 6;

	# For 'artf_cluster_gapped_segs_len'. #
our $TEST_LEVEL_acgsl = 0;
##our $TEST_LEVEL_acgsl = 6;

	# For 'detect_cmplx_error_cand_acls0'. #
our $TEST_LEVEL_dceca = 0;
##our $TEST_LEVEL_dceca = 6;

	# For 'wrapper_detect_cmplx_error_cand_acls0'. # (Added on 2019/06/12.)
our $TEST_LEVEL_wrdceca = 0;
##our $TEST_LEVEL_wrdceca = 6;


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.



# selfcompare_indel_hosting_branches:
#	This subroutine compares a set of indel events with itself,
#	especially in terms of the branches along which the events occur.
#
#	[CREATED on July 14 (Thu), 2016, in creation of 'summary_examine_complex.6p_aug.raw.pl'.]
#	[ => IMPORTED from 'summary_examine_complex.6pp_aug.raw.pl' on Dec 16 (Fri), 2016.]
#
# INPUT: (\@eventset, \%equiv_br, \%br2nn), where
#	@{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) for the $ev th event in the set,
#	   with $br being the ID of the branch along which it occurred,
#	   with $stat_ue being the state ("presence"/"absence") at the upper-end of the branch,
#	   with $beg/$end/$len being the beginning coordinate/ending coordinate/length of the event;
#
#	%equiv_br = ($br_id => $equiv_br_id, ...) for branches with their equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...).
#
# OUTPUT: (\@on_coinc_brs, \@on_nn_brs), where
#	
#	@{$on_coinc_brs[$k]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $k th pair of events along the same branch ($br1 or $br2), with
#	  $indx_ev1/2 being the index of the relevant 1st/2nd event (in the order stored in @eventset);
#
#	@{$on_nn_brs[$m]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $m th pair of events along the nearest-neighboring branches ($br1 and $br2), with
#	   $indx_ev1/2 being the index of the relevant 1st/2nd event (in the order stored in @eventset).
#
sub selfcompare_indel_hosting_branches (\@\%\%) {

	my ($eventset, $equiv_br, $br2nn) = @_;

	my $ct_events = @{$eventset};

	my @on_coinc_brs = my @on_nn_brs = ();

	for (my $ev2=1; $ev2 < $ct_events; $ev2++) { # Outer for-loop (over 'later' events). #

	    my ($br2, $state_ue2, $beg2, $end2, $len2) = @{$eventset->[$ev2]};
	    my $eq_br2 = $equiv_br->{$br2};
	    my $nn2 = $br2nn->{$br2};
	    (defined $nn2) or ($nn2 = {});

	    my $if_relv = 0;

	    for (my $ev1=0; $ev1 < $ev2; $ev1++) { # Inner for-loop (over 'earlier' events). #

		my ($br1, $state_ue1, $beg1, $end1, $len1) = @{$eventset->[$ev1]};

		if ( ($br1 == $br2)  or
			((defined $eq_br2) and ($br1 == $eq_br2)) ) {

		    push @on_coinc_brs, [$br1, $br2, $ev1, $ev2];
		    # $if_relv=1;
		} elsif (defined $nn2->{$br1}) {

		    push @on_nn_brs, [$br1, $br2, $ev1, $ev2];
		    # $if_relv=1;
		}
	    }					# End of the inner for-loop (over 'earlier' events). #

	    ## (!($if_relv) and ($state_ue2 eq $char_abs)) and (push @irrlv_insertions2, $ev2);

	}				# End of the outer for-loop (over 'later' events). #

	return (\@on_coinc_brs, \@on_nn_brs);

}	# End of the subroutine 'selfcompare_indel_hosting_branches'. #


# compare_indel_hosting_branches:
#	This subroutine compares a set of indel events with another set,
#	especially in terms of the branches along which the events occur.
#	[CREATED on July 14 (Thu), 2016, in creation of 'summary_examine_complex.6p_aug.raw.pl'.]
#	[ => IMPORTED from 'summary_examine_complex.6pp_aug.raw.pl' on Dec 18 (Sun), 2016. ]
#
# INPUT: (\@eventset1, \@eventset2, \%equiv_br, \%br2nn, $char_abs, \%node2pa), where
#
#	@{$eventset1/2[$ev]} = ($br, $stat_ue, $beg, $end, $len) for the $ev th event in the 1st/2nd set,
#	   with $br being the ID of the branch along which it occurred,
#	   with $stat_ue being the state ("presence"/"absence") at the upper-end of the branch,
#	   with $beg/$end/$len being the beginning coordinate/ending coordinate/length of the event;
#
#	%equiv_br = ($br_id => $equiv_br_id, ...) for branches with their equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...);
#	$char_abs is the string representing the absence state;
#	%node2pa = ($node_id => $parent_id, ...);
#
# OUTPUT: (\@on_coinc_brs, \@on_nn_brs, \@irrlv_insertions2), where
#	
#	@{$on_coinc_brs[$k]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $k th pair of events along the same branch ($br1 or $br2), with
#	  $indx_ev1/2 being the index of the relevant event in @eventset1/2;
#
#	@{$on_nn_brs[$m]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $m th pair of events along the nearest-neighboring branches ($br1 and $br2), with
#	   $indx_ev1/2 being the index of the relevant event in @eventset1/2;
#
#	@irrlv_insertions2 stores the indices of the insertions in @eventset2 along the branches that neither coincide with, nor is the nearest-neighbors of, any branches hosting the events in @eventset1.
#
sub compare_indel_hosting_branches (\@\@\%\%$\%) {

	my ($eventset1, $eventset2, $equiv_br, $br2nn, $char_abs, $node2pa) = @_;

	my $ct_events1 = @{$eventset1};
	my $ct_events2 = @{$eventset2};

	my @on_coinc_brs = my @on_nn_brs = my @irrlv_insertions2 = ();

	for (my $ev2=0; $ev2 < $ct_events2; $ev2++) { # Outer for-loop (over events in @{$eventset2}). #

	    my ($br2, $state_ue2, $beg2, $end2, $len2) = @{$eventset2->[$ev2]};
	    my $eq_br2 = $equiv_br->{$br2};
	    my $nn2 = $br2nn->{$br2};
	    (defined $nn2) or ($nn2 = {});

	    my $if_relv = 0;

	    for (my $ev1=0; $ev1 < $ct_events1; $ev1++) { # Inner for-loop (over events in @{$eventset1}). #

		my ($br1, $state_ue1, $beg1, $end1, $len1) = @{$eventset1->[$ev1]};

		if ( ($br1 == $br2)  or
			((defined $eq_br2) and ($br1 == $eq_br2)) ) {
		    push @on_coinc_brs, [$br1, $br2, $ev1, $ev2];
		    $if_relv=1;
		} elsif (defined $nn2->{$br1}) {
		    push @on_nn_brs, [$br1, $br2, $ev1, $ev2];
		    $if_relv=1;
		} elsif (!($if_relv) and ($state_ue2 eq $char_abs) ) {
				# ADDED on July 14 (Thu), 2016. #
		    my ($common_anc, $after_div1, $after_div2) = fetch_common_ancestors ($br1, $br2, %{$node2pa});
		    if (@{$after_div1} == 0) { # $br1 is ancestor of $br2.
			($state_ue1 eq $char_abs) and ($if_relv = 1); # event along $br1 is insertion. #
		    } elsif (@{$after_div2} == 0) { # $br2 is ancestor of $br1.
			$if_relv = 1;
		    }
		}
	    }					# End of the inner for-loop (over events in @{$eventset1}). #

	    (!($if_relv) and ($state_ue2 eq $char_abs)) and (push @irrlv_insertions2, $ev2);

	}				# End of the outer for-loop (over events in @{$eventset2}). #

	return (\@on_coinc_brs, \@on_nn_brs, \@irrlv_insertions2);

}	# End of the subroutine 'compare_indel_hosting_branches'. #


# mk_self_nn_events:
#	This subroutine creates a hash, %self_nn_events,
#	which associates each particular pair of nearest neighboring branches
#	with a set of pairs of 'nearest-neighboring' events.
#
#	[CREATED on Dec 16 (Fri), 2016.]
#
# INPUT: (\@eventset, \%br2eq_br, \%br2nn, $gs), where
#
#	@{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) for the $ev th event in the set,
#	   with $br being the ID of the branch along which it occurred,
#	   with $stat_ue being the state ("presence"/"absence") at the upper-end of the branch,
#	   with $beg/$end/$len being the beginning coordinate/ending coordinate/length of the event;
#
#	%equiv_br = ($br_id => $equiv_br_id, ...) for branches with their equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...);
#
#	$gs is the identifier (usually index) of the subject gapped segment.
#
# OUTPUT: \%self_nn_events, where
#
#	%self_nn_events = ($br2 => {$br1 => \@set_eventpairs, ...}, ..), where
#	@{$set_eventpairs[$ep]} = ($gs, $indx_ev1, $gs, $indx_ev2) for the $ep th pair of events
#	occurring in the nearest-neighbor branch pair, ($br1, $br2).
#
sub mk_self_nn_events (\@\%\%$) {


	my ($eventset, $br2eq_br, $br2nn, $gs) = @_;


		# @{$on_coinc_brs[$k]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $k th pair of events along the same branch ($br1 or $br2), with
		# @{$on_nn_brs[$m]} = ($br1, $br2, $indx_ev1, $indx_ev2) for the $m th pair of events along the nearest-neighboring branches ($br1 and $br2), with
		#
	my ($on_coinc_brs, $on_nn_brs) = selfcompare_indel_hosting_branches (@{$eventset}, %{$br2eq_br}, %{$br2nn});
	unless (defined $on_nn_brs) {
	    print STDERR "[ERROR 'mk_self_nn_events'] Subroutine 'selfcompare_indel_hosting_branches' FAILED: $! !\n\n";
	    return undef;
	}

	my %self_nn_events;

	foreach my $pairinfo0 (@{$on_nn_brs}) { # Outer foreach-loop (over event pairs). #

	    my ($br01, $br02, $indx_ev01, $indx_ev02) = @{$pairinfo0};

	    ##foreach my $pairinfo ([$br01, $br02, $indx_ev01, $indx_ev02]) {
	    foreach my $pairinfo ([$br01, $br02, $indx_ev01, $indx_ev02],
	    		[$br02, $br01, $indx_ev02, $indx_ev01]) { # Inner foreach-loop (over possible representations of a pair). #

	        my ($br1, $br2, $indx_ev1, $indx_ev2) = @{$pairinfo};
	        my $hash = $self_nn_events{$br2};
	        (defined $hash) or ($hash = $self_nn_events{$br2} = {});
	        my $array = $hash->{$br1};
	        (defined $array) or ($array = $hash->{$br1} = []);
	        push @{$array}, [$gs, $indx_ev1, $gs, $indx_ev2];
	    }						# End of the inner foreach-loop (over possible representations of a pair). #
	    ##}

	}				# End of the outer foreach-loop (over eventpairs). #

	return \%self_nn_events;

}	# END of 'mk_self_nn_events'. #


# mk_ingredients_acls_gpd_segs:
#	This subroutine produces some main ingredients of the subroutine,
#	'artf_cluster_gapped_segs(_len)'.
#	It is mostly based on the for-loop immediately above 'artf_cluster_gapped_segs(_len)'
#	in the script, 'test_prepare_sliding_windows.pl'.
#	[CREATED on 2019/06/12.]
#
# INPUTS: (\@set_cls_gpblocks, \@init_psm_hsts_for_cls, \@set_exnodes, $char_abs, \%node2ch, \%br2eq, \%br2nn), where
#
#	@set_cls_gpblocks stores information on the gapped-segments, i.e., clusters of gap-pattern-blocks.
#	More precisely,
#	($indx_gp, $len, $beg, $end) = @{$set_cls_gpblocks->[$cls][$gpb]} are information on the $gpb th gap-pattern-block in the $cls th gapped segment, where
#	$indx_gp is the index of the gap-pattern,
#	$len, $beg, $end are the length, beginning coordinate, and end coordinate, respectively,
#	 of the gap-pattern-block;
#	[\@set_cls_gpblocks is an important input of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm"];
#
#
#	@{$init_psm_hsts_for_cls[$cls]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious history resulting in the $cls th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#	[\@init_psm_hsts_for_cls is an important output of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm"];
#
#	@set_exnodes stores the IDs of all external nodes;
#
#	$char_abs is the character representing an 'absence' state (i.e., gap);
#
#	%node2ch = ($node_id => \@child_node_ids, ...) for all external nodes;
#
#	%br2eq = ($br_id => $equiv_br_id, ...) for the branches with their own equivalents;
#	%br2nn = ($br_id => {$nearest_neighboring_br_id => 1, ...}, ...) for the branches with their own nearest neighbors;
#	[Typically, \%br2eq and \%br2nn are among the outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm".]
#
# OUTPUTS: (\@bds_gpd_segs_in_msa, \@set_eventsets, \@set_self_nn_events, \@set_lists_gblocks), where
#
#	($beg_gs, $end_gs) = @{$bds_gpd_segs_in_msa[$gs]} are the beginning and end coordinates of the $gs th gapped segment in the subject MSA;
#
#	@eventset = @{$eventsets_in_msa[$gs]} stores the events in the $gs th gapped segment,
#	with @{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) being the information on the $ev th event;
#	[It is basically an output of 'mk_eventset' in "MyTreeMap_indels_sppl.pm"];
#
#	\%self_nn_events = $set_self_nn_events[$gs] is the hash recording the self-nearest-neighboring event pairs in the $gs th gapped segment, with
#	%self_nn_events = ($br2 => {$br1 => \@set_eventpairs, ...}, ...),
#	where @{$set_eventpairs[$ep]} = ($gs, $indx_ev1, $gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events;
#	[This hash is basically an output of 'mk_self_nn_events' above];
#
#	@{$set_lists_gblocks[$gs]->[$gb]} provides information on the $gb th gap-block in the $gs th gapped segment. More precisely, 
#	($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@aff_exnodes, $e) = @{$set_lists_gblocks[$cl]->[$gb], where
#	the gap-block is separated by, and on the $up_or_lw (='U' or 'L') side of, branch $br,
#	$lm_gpb and $rm_gpb are the indices of the left-most and right-most gap-pattern blocks,
#	@invlvd_gpbs lists the gap-pattern-blocks involved;
#	@aff_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}};
#	[@br_list_gblocks = @{$set_lists_gblocks[$gs]} is the output of 'br_list_gblocks' (in "MyTreeMap_indels_spt_odr2.pm").];
#
sub mk_ingredients_acls_gpd_segs (\@\@\@$\%\%\%) {

	my ($set_cls_gpblocks, $init_psm_hsts_for_cls, $set_exnodes, $char_abs, $node2ch, $br2eq, $br2nn) = @_;

	my $ct_cls = @{$set_cls_gpblocks};

    (print STDOUT "< Subroutine 'mk_ingredients_acls_gpd_segs' (#{gapped-segments} = $ct_cls) >\n\n") if ($TEST_LEVEL_migacgss>0);

	my @bds_gpd_segs_in_msa = my @set_eventsets = my @set_self_nn_events = ();
	my @set_lists_gblocks = ();

	for (my $cl = 0; $cl < $ct_cls; $cl++) { # The Main for-loop (over gapped segments). #

	    my $cls_gpblks = $set_cls_gpblocks->[$cl];
		## my ($indx_gp, $len, $beg, $end) = @{$cls_gpblks->[$bl]};

	    my $ct_gpbs = @{$cls_gpblks};
	    my $beg_cls = $cls_gpblks->[0][2];
	    my $end_cls = $cls_gpblks->[$ct_gpbs-1][3];
	    push @bds_gpd_segs_in_msa, [$beg_cls, $end_cls];

	    my ($br2indels0, $node2bw_gstats0) = @{$init_psm_hsts_for_cls->[$cl]};


	    my $eventset = mk_eventset (%{$br2indels0}, @{$cls_gpblks}, 2, 3, 1);
	    unless (defined $eventset) {
		print STDOUT "[ERROR 'mk_ingredients_acls_gpd_segs'] The subroutine 'mk_eventset' FAILED on the $cl th gapped segment: $! !\n\n";
		return (undef, undef, undef, undef);
	    }
	    my $ct_events = @{$eventset};

   if ($TEST_LEVEL_migacgss >2) {
	print STDOUT ' ' x 2, "{ The $cl th gapped segment (#{events} = $ct_events) }\n\n";

	print STDOUT ' ' x 2, '[ The content of @{$eventset}, the output of ', "'mk_eventset' ]\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Indx', 'branch', 'stat_ue', 'start_ev', 'end_ev', 'len_ev'), "\n\n";
	for (my $e=0; $e < $ct_events; $e++) { # Inner for-loop (over events). #
	    print STDOUT ' ' x 2, join ("\t", $e, @{$eventset->[$e]}), "\n";
	}				# End of the inner for-loop (over events). #
	print STDOUT "\n\n";
   } # END of "if ($TEST_LEVEL_migacgss >2) {...}"

	    my $self_nn_events = mk_self_nn_events (@{$eventset}, %{$br2eq}, %{$br2nn}, $cl);
	    unless (defined $self_nn_events) {
		print STDOUT "[ERROR 'mk_ingredients_acls_gpd_segs'] The subroutine 'mk_self_nn_events' FAILED on the $cl th gapped segment: $! !!\n";
		return (undef, undef, undef, undef);
	    }


   if ($TEST_LEVEL_migacgss >2) {
	print STDOUT ' ' x 2, '[ The content of @{$self_nn_events}, the output of ', "'mk_self_nn_events' ]\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Branch2', 'branch1', 'indx_gs', 'indx_ev1', 'indx_gs', 'indx_ev2'), "\n\n";
	foreach my $br2 (sort {$a <=> $b} keys %{$self_nn_events}) { # Inner foreach-loop (over the 2nd branches). #
	    my $br2set_eventpairs = $self_nn_events->{$br2};
	    foreach my $br1 (sort {$a <=> $b} keys %{$br2set_eventpairs}) { # Inner-inner foreach-loop (over the 1st branches). #
		my $set_eventpairs = $br2set_eventpairs->{$br1};
		foreach my $eventpair (@{$set_eventpairs}) { # Innermost foreach-loop (over the event pairs). #
		    print STDOUT ' ' x 2, join ("\t", $br2, $br1, @{$eventpair}), "\n";
		}				# End of the innermost foreach-loop (over the event pairs). #
	    }						# End of the inner-inner foreach-loop (over the 1st branch). #
	} 					# End of the inner foreach-loop (over the 2nd branches). #
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_migacgss >2) {...}"

	    $set_eventsets[$cl] = $eventset;
	    $set_self_nn_events[$cl] = $self_nn_events;

	    my $br_list_gblocks = br_list_gblocks (%{$br2indels0}, @{$set_exnodes}, $char_abs, %{$node2ch});
	    unless (defined $br_list_gblocks) {
		print STDOUT "[ERROR 'mk_ingredients_acls_gpd_segs'] The subroutine 'br_list_gblocks' FAILED on the $cl th gapped segment: $!  !!\n";
		return (undef, undef, undef, undef);
	    }


   if ($TEST_LEVEL_migacgss >2) {
	my $ct_gbs = @{$br_list_gblocks};
	print STDOUT "\n";
	print STDOUT ' ' x 2, '[ The content of @{$br_list_gblocks}, the output of ', "'br_list_gblocks' (#{gap-blocks} = $ct_gbs) ]\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_lw(L)', 'lm_gpb', 'rm_gpb', '[involved,gap,pattern,blocks]', '(affected,external,nodes)', 'within_branch_index'), "\n\n";

	for (my $gb = 0; $gb < $ct_gbs; $gb++) {
	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $aff_exnodes, $e) = @{$br_list_gblocks->[$gb]};
	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs}); 
	    $cnct_invlvd_gpbs = "[$cnct_invlvd_gpbs]";
	    my $cnct_aff_exnodes = join (',', @{$aff_exnodes});
	    $cnct_aff_exnodes = "($cnct_aff_exnodes)";
	    print STDOUT ' ' x 2, join ("\t", $gb, $br, $up_or_lw, $lm_gpb, $rm_gpb, $cnct_invlvd_gpbs, $cnct_aff_exnodes, $e), "\n";
	}
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_migacgss >2) {...}"

	    $set_lists_gblocks[$cl] = $br_list_gblocks;

	}  # End of the Main for-loop (over gapped segments). #


   (print STDOUT "\n< Subroutine 'mk_ingredients_acls_gpd_segs' Ended />\n\n") if ($TEST_LEVEL_migacgss>0);


	return (\@bds_gpd_segs_in_msa, \@set_eventsets, \@set_self_nn_events, \@set_lists_gblocks);

}	# End of the subroutine, 'mk_ingredients_acls_gpd_segs'. #



# artf_cluster_gapped_segs:
#	This subroutine artificially cluster the gapped segments.
#	It was created based on the main process of the script, 'summary_examine_complex.6pp_aug.raw.pl'.
#
#	[ METHOD ]
#	The gapped segments will be clustered together if the following conditions are fulfilled:
#	(i) Total{spacer sizes between the two gapped segments} <= $upper_bound_spsize ;
#	(ii) [ Total{spacer sizes between the two gapped segments} <= $thrsh_spsize0 ] OR
#	     [ [ Max_size {gapped segment 1, gapped segment 2} >= Total{spacer sizes} * $mag_thrsh ] AND
#	     [ Min_size {gapped segment 1, gapped segment 2} >= Total{spacer sizes} / $mag_thrsh ] ] ;
#
#	(iii) EITHER the two gapped segments contain indel events along the same branch OR
#	      { they contain indel events along neighboring branches that form a triangle with another indel event (from either of the segments or from yet another 'neighboring' gapped segment) };
#
#	[ CREATION STARTED on Dec 17 (Sat), 2016, ENDED on Dec 18 (Sun), 2016.]	
#
# INPUT: (\@gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, \@eventsets_in_msa, \@set_self_nn_events, $upper_bound_spsize, \%br2eq_br, \%br2nn, $char_abs, \%node2pa, $thrsh_spsize0, $mag_thrsh), where
#
#	($beg_gs, $end_gs) = @{$gpd_segs_in_msa[$gs]}[$indx_beg_gs,$indx_end_bs] are the beginning and ending coordinates of the $gs th gapped segment in the subject MSA;
#
#	@eventset = @{$eventsets_in_msa[$gs]} stores the events in the $gs th gapped segment,
#	with @{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) being the information on the $ev th event. (It is basically an output of 'mk_eventset' in "MyTreeMap_indels_sppl.pm".);
#
#	\%self_nn_events = $set_self_nn_events[$gs] is the hash recording the self-nearest-neighboring event pairs in the $gs th gapped segment, with
#	%self_nn_events = ($br2 => {$br1 => \@set_eventpairs, ...}, ...),
#	where @{$set_eventpairs[$ep]} = ($gs, $indx_ev1, $gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. (This hash is basically an output of 'mk_self_nn_events'.);
#
#	$upper_bound_spsize is the parameter for the condition (i) above;
#
#	%br2eq_br = ($branch_id => $equivalent_branch_id, ...) for branches with equivalents.
#	(It is basically an output of 'create_br2eq_br' in "MyTreeMap_indels_sppl.pm".);
#
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...) for all branches.
#	(It is basically an output of 'create_br2nn' in "MyTreeMap_indels_sppl.pm".);
#
#	$char_abs is the character representing the absence of the site;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node (i.e., root);
#
#	$thrsh_spsize0 and $mag_thrsh are the parameters for the condition (ii) above;
#
#
# OUTPUT: (\@clusters, \%gs2prev_nn_events), where
#
#	@{$clusters[$cl]} stores the indices of the gapped segments that constitute the $cl th artificial cluster.
#
#	%gs2prev_nn_events = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...),
#	    with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events).
#	    Incidentally, the suffixes, "_n" and "_c" indicate "next" and "current", respectively.
#
#
sub artf_cluster_gapped_segs (\@$$\@\@$\%\%$\%$$) {

	my ($gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, $eventsets_in_msa, $set_self_nn_events, $upper_bound_spsize, $br2eq_br, $br2nn, $char_abs, $node2pa, $thrsh_spsize0, $mag_thrsh) = @_;

	my $ct_gpd_segs = @{$gpd_segs_in_msa};

   ##(print STDOUT "<  Clustering with {Basic Spacer-Size Threshold} = $thrsh_spsize0 >\n\n") if ($TEST_LEVEL>1);

	my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #
			# END of "MOVED from within the 1st middle while-loop, to fix a BUG." #

	my $last_cluster = [0];

=For DEBUGGING (1)
	$last_cluster = [14];
	$ct_gpd_segs = 22;
=cut
=For DEBUGGING (2)
	$last_cluster = [35];
	$ct_gpd_segs = 42;
=cut

	my @clusters = ($last_cluster);

	while ($last_cluster->[$#{$last_cluster}] < $ct_gpd_segs-1) { # Main outer while-loop (untill all gapped segments are examined). #

   if ($TEST_LEVEL_acgs>0) {
	print STDOUT "\n{{{ Initial 'last' cluster = (", join (', ', @{$last_cluster}), ") }}}\n\n";
   }

		my $indx_c_gs = 0;
		while ($indx_c_gs < @{$last_cluster}) { # Middle while-loop (until the "last" cluster is exhausted). #


	            my $c_gs = $last_cluster->[$indx_c_gs]; # Current gapped segment. #
	            my ($c_beg_coord, $c_end_coord) = @{$gpd_segs_in_msa->[$c_gs]}[$indx_beg_gs,$indx_end_gs];
		    my $c_gs_size = $c_end_coord - $c_beg_coord + 1;

   if ($TEST_LEVEL_acgs>0) {
	print STDOUT "\n{{ The 'Current' segment = the $indx_c_gs th segment in the 'last' cluster (= the $c_gs th segment in the entire MSA) ; [beg, end] = [$c_beg_coord, $c_end_coord] ; size = $c_gs_size }}}\n\n";
   }
		    my $c_eventset = $eventsets_in_msa->[$c_gs];

			# push @eventsets_in_msa, \@eventset; element of @eventset = [$br, $stat_ue, $beg, $end, $len] 

		    my $c_self_nn_events = $set_self_nn_events->[$c_gs]; # = {$br2 => {$br1 => \@set_eventpairs, ...}, ...}, where @{$set_eventpairs[$ep]} = ($c_gs, $indx_ev1, $c_gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. #

		    my $tot_spsize = 0;

		    my $last_n_end_coord = $c_end_coord; # A BUG-FIXATION measure (A-1). #

		    for (my $n_gs = $c_gs+1; $n_gs < $ct_gpd_segs; $n_gs++) { # Inner for-loop (over "next" gapped segments). #

			my ($n_beg_coord, $n_end_coord) = @{$gpd_segs_in_msa->[$n_gs]}[$indx_beg_gs,$indx_end_gs];
			my $n_gs_size = $n_end_coord - $n_beg_coord + 1;

     if ($TEST_LEVEL_acgs>0) {
	print STDOUT "\n{ The 'Next' segment = the $n_gs th segment in the entire MSA ;  [beg, end] = [$n_beg_coord, $n_end_coord] ; size = $n_gs_size }}\n\n";
     }

				# A BUG-FIXATION measure (A-2). #
			$tot_spsize += $n_beg_coord - $last_n_end_coord - 1;
			$last_n_end_coord = $n_end_coord;
			##$tot_spsize += $n_beg_coord - $c_end_coord - 1;
			####my $tot_spsize += $n_beg_coord - $c_end_coord - 1; # A BUG fixed on July 15 (Fri), 2016. #

    if ($TEST_LEVEL_acgs>1) {
	print STDOUT "(Total spacer size) = $tot_spsize .\n\n";
    }

				## Replace the conditions with what I really meant. (1) ##
			($tot_spsize > $upper_bound_spsize) and last; # In this case, we give up extending the "last" cluster based on the current gapped segment. #
			## ($tot_spsize > $thrsh_spsize_lng) and last; # In this case, we give up extending the "last" cluster based on the current gapped segment. # ... OBSOLETE...

			my $n_eventset = $eventsets_in_msa->[$n_gs];
			    # push @eventsets_in_msa, \@eventset; element of @eventset = [$br, $stat_ue, $beg, $end, $len] 

			# @{$on_coinc_brs->[$k]} = ($br_c, $br_n, $indx_c_ev, $indx_n_ev) for the $k th pair of events along the same branch.
			# @{$on_nn_brs->[$m]} = ($br_c, $br_n, $indx_c_ev, $indx_n_ev) for the $m th pair of events along the nearest-neighboring branches.
			# @{$irrlv_insertions_n} stores the indices of the insertions in the "next" gapped segment along the branches that neither coincide with, nor is the nearest-neighbors of, any "current" branches.
			# 
			my ($on_coinc_brs, $on_nn_brs, $irrlv_insertions_n)
				= compare_indel_hosting_branches (@{$c_eventset}, @{$n_eventset}, %{$br2eq_br}, %{$br2nn}, $char_abs, %{$node2pa});

				# %{$br2nn}; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

			my $incl_spsize = $n_gs_size; # The increment of the spacer size. #

			foreach my $indx_ev (@{$irrlv_insertions_n}) { $incl_spsize -= $n_eventset->[$indx_ev][4]; }

   if ($TEST_LEVEL_acgs>2) {
	print STDOUT "(Increment of the spacer-size that will be originated from the 'next' segment) = $incl_spsize .\n\n";
   }

			##$tot_spsize += $incl_spsize; # The "next" branch will be incorporated into the "spacer", after excluding the irrelevant insertions. # A BUG ... it was too early to incorporate it into the "spacer"!!

				## Replace the conditions with what I really meant. (2) ##

			my ($max_gs_size, $min_gs_size) = ($n_gs_size > $c_gs_size) ?
					($n_gs_size, $c_gs_size) : ($c_gs_size, $n_gs_size) ;

			unless (($tot_spsize <= $thrsh_spsize0) or 
				(($max_gs_size >= $tot_spsize * $mag_thrsh) and 
				($min_gs_size >= $tot_spsize / $mag_thrsh)) ) {
				##($max_gs_size >= $tot_spsize * $mag_thrsh) or
				##($min_gs_size >= $tot_spsize / $mag_thrsh) ) {

				# In this case, this "next" segment is NEVER considered as interacting with the current segment.

			    $tot_spsize += $incl_spsize; # A measure to fix the BUG which is one block above. #

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (A) ...\n\n") if ($TEST_LEVEL_acgs>2);

			    next;
			}

			$tot_spsize += $incl_spsize; # A measure to fix the BUG which is two blocks above. #

			if (@{$on_nn_brs}>0) {

				# Store the information on the interacting event pairs for later use. #

			    my $br2br2nn_events = $gs2prev_nn_events{$n_gs};
			    (defined $br2br2nn_events) or ($br2br2nn_events = $gs2prev_nn_events{$n_gs} = {});

			    foreach my $info_evpair (@{$on_nn_brs}) { # 2nd inner-inner foreach-loop (over nn event pairs). #
				my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
				my $br2nn_events = $br2br2nn_events->{$br_n};
				(defined $br2nn_events) or ($br2nn_events = $br2br2nn_events->{$br_n} = {});
				my $nn_events = $br2nn_events->{$br_c};
				(defined $nn_events) or ($nn_events = $br2nn_events->{$br_c} = []);
				push @{$nn_events}, [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev];
			    
			    }			# End of the 2nd inner-inner foreach-loop (over nn event pairs). #
				## my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #

			} # END of "if (@{$on_nn_brs}>0) {...}" #

			my $last_gs_in_cls = $last_cluster->[$#{$last_cluster}];

			unless ($last_gs_in_cls < $n_gs) {

   (print STDOUT " ... The 'next' segment is already incorporated in the cluster. Thus, skip the subsequent analyses ...\n\n") if ($TEST_LEVEL_acgs>2);

			    next; # In this case, $n_gs is already incorporated in the cluster, thus we don't have to examine further. #
			}

				### Examine whether or not this "next" segment can interact with the current segment. ###

				# First, examine the triangular interaction involving a past segment. #
				# (MOVED from below and augmented, to fix a BUG) #
			my $br2br2prev_nn_events = $gs2prev_nn_events{$c_gs};
			(defined $br2br2prev_nn_events) or ($br2br2prev_nn_events = {});

			my ($lm_igs, $rm_igs); # The indices for the leftmost and rightmost 'past' gapped segments interacting with $c_gs and $n_gs. 
			foreach my $info_evpair (@{$on_nn_brs}) { # 3rd inner-inner foreach-loop (over nn event pairs). #
			    my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
			    ##my $nn_c = $br2nn{$br_c}; # It's probably unnecessary any more. #
			    my $nn_n = $br2nn->{$br_n};
				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

			    my $br2prev_nn_events = $br2br2prev_nn_events->{$br_c};
			    (defined $br2prev_nn_events) or next; # Skip if the event along $br_c in $c_gs has no past neighboring events. #
			    foreach my $br (keys %{$br2prev_nn_events}) { # Outer-innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c. #
				(defined $nn_n->{$br}) or next; # Skip if $br is not the nearest-neighbor of $br_n as well. #
				my $prev_nn_events = $br2prev_nn_events->{$br};
				foreach my $pev_info (@{$prev_nn_events}) { # Inner-innermost foreach-loop (over the past interacting events). #
				    my ($p_igs, $indx_p_iev, $c_gs2, $indx_c_ev2) = @{$pev_info};
				    ((defined $lm_igs) and ($lm_igs <= $p_igs)) or ($lm_igs = $p_igs);
				    ((defined $rm_igs) and ($p_igs <= $rm_igs)) or ($rm_igs = $p_igs);
				}				# End of the inner-innermost foreach-loop (over the past interacting events). #
			    }				# End of the outer-innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c). #
			}				# End of the 5th inner-inner foreach-loop (over nn event pairs). #

			if (defined $lm_igs) {
				# Interaction between $n_gs and $c_gs (II-b). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-b) ...\n\n") if ($TEST_LEVEL_acgs>2);

					# MODIFIED to FIX a BUG (i). #
			    my $counter = 0;
			    while ($lm_igs < $last_cluster->[0]) { $last_cluster = pop @clusters; $counter++; } # Get back to older clusters until the updated $last_cluster accommodates $lm_igs. #
			    ##while ($lm_igs < $last_cluster->[0]) { $last_cluster = pop @clusters; } # Get back to older clusters until the updated $last_cluster accommodates $lm_igs. #

			    my ($gs_lm_in_lc, $gs_rm_in_lc) = @{$last_cluster}[0, $#{$last_cluster}];

				# Update the $last_cluster and the $indx_c_gs. #
			    push @{$last_cluster}, ($gs_rm_in_lc + 1 .. $n_gs);

					# MODIFIED to FIX a BUG (ii). #
			    ($counter>0) and (push @clusters, $last_cluster);
			    ##push @clusters, $last_cluster;
			    $indx_c_gs = $c_gs - $gs_lm_in_lc;

   if ($TEST_LEVEL_acgs>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n";
	print STDOUT '    $indx_c_gs  = ', $indx_c_gs, " .\n\n\n";
   }

			    next;
			}

				# END of "(MOVED from below and augmented, to fix a BUG)" #

			if (@{$on_coinc_brs}>0) {

				# Interaction between $n_gs and $c_gs (I). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (I) ...\n\n") if ($TEST_LEVEL_acgs>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_acgs>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;
			}

			unless (@{$on_nn_brs}>0) {
				# In this case, this "next" segment is NEVER considered as interacting with the current segment.

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (B) ...\n\n") if ($TEST_LEVEL_acgs>2);
			    next;
			}

				## In the following, $n_gs is regarded as interacting with $c_gs
				## ONLY IF the nearest-neighboring events form a triangle encompassing a node.

		        my $n_self_nn_events = $set_self_nn_events->[$n_gs]; # = {$br2 => {$br1 => \@set_eventpairs, ...}, ...}, where @{$set_eventpairs[$ep]} = ($n_gs, $indx_ev1, $n_gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. #

=MOVED to above interaction type (I), to fix a BUG. (1)
=cut

			my $if_interact = 0;

			foreach my $info_evpair (@{$on_nn_brs}) { # 5th inner-inner foreach-loop (over nn event pairs). #
			    my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
			    my $nn_c = $br2nn->{$br_c};
			    my $nn_n = $br2nn->{$br_n};
				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

				# (1) Examine the self-nearest-neighboring events in $n_gs. #

			    my $br2self_nn_n = $n_self_nn_events->{$br_n};
			    if (defined $br2self_nn_n) {
				foreach my $br (keys %{$br2self_nn_n}) { # 1st innermost foreach-loop (over branches self-neighboring $br_n). #
				    if (defined $nn_c->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 1st innermost foreach-loop (over branches self-neighboring $br_n). #
				$if_interact and last;
			    }

				# (2) Examine the self-nearest-neighboring events in $c_gs. #

			    my $br2self_nn_c = $c_self_nn_events->{$br_c};
			    if (defined $br2self_nn_c) {
				foreach my $br (keys %{$br2self_nn_c}) { # 2nd innermost foreach-loop (over branches self-neighboring $br_c). #
				    if (defined $nn_n->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 2nd innermost foreach-loop (over branches self-neighboring $br_c). #
				$if_interact and last;
			    }

=MOVED TO ABOVE interaction type (I), and augmented, to fix a BUG (2).
=cut # END of "MOVED TO ABOVE interaction type (I), and augmented, to fix a BUG (2)."

			}				# End of the 5th inner-inner foreach-loop (over nn event pairs). #

			if ($if_interact) {
				# Interaction between $n_gs and $c_gs (II-a). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-a) ...\n\n") if ($TEST_LEVEL_acgs>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_acgs>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} else {

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (C) ...\n\n") if ($TEST_LEVEL_acgs>2);
			}

		    }						# End of the inner for-loop (over "next" gapped segments). #

		    $indx_c_gs++;

   if ($TEST_LEVEL_acgs>3) {
	print STDOUT '    | Updated $indx_c_gs  = ', $indx_c_gs, " | \n\n\n";
    }

		}				# End of the middle while-loop (until the "last" cluster is exhausted). #

		my $last_gs_clustered = $last_cluster->[$#{$last_cluster}];
		($last_gs_clustered == $ct_gpd_segs-1) and last;

			# Create a new cluster. # 

		my @new_last_cluster = ($last_gs_clustered + 1);
		$last_cluster = \@new_last_cluster;
		push @clusters, $last_cluster;

   if ($TEST_LEVEL_acgs>3) {
	print STDOUT "    | Information updated by default: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

	}					# End of the main outer while-loop (until all gapped segments are examined). #

	return (\@clusters, \%gs2prev_nn_events);

}	# END of 'artf_cluster_gapped_segs'. #


# artf_cluster_gapped_segs_len:
#	This subroutine artificually clusters the gapped segments with more lenient conditions
#	than its predecessor, 'artf_cluster_gapped_segs'.
#	It was created by incorporating the differences of the script, 'summary_examine_complex.6pp_aug_len.raw.pl', from 'summary_examine_complex.6pp_aug.raw.pl'.
#
#	[ METHOD ]
#	The gapped segments will be clustered together if the following conditions are fulfilled:
#	(i) and (ii) are the same as those for 'artf_cluster_gapped_segs'.
#	(iii) EITHER the two gapped segments contain indel events along the same branch OR
#	      { they contain indel events along neighboring branches } . That is, the indel events do NOT need to form a triangle with yet another event ANY LONGER.
#
#	[ CREATED on Dec 26 (Mon), 2016. ]
#
#
# artf_cluster_gapped_segs:
#	This subroutine artificially clusters the gapped segments.
#	It was created based on the main process of the script, 'summary_examine_complex.6pp_aug.raw.pl'.
#
#	[ METHOD ]
#	The gapped segments will be clustered together if the following conditions are fulfilled:
#	(i) Total{spacer sizes between the two gapped segments} <= $upper_bound_spsize ;
#	(ii) [ Total{spacer sizes between the two gapped segments} <= $thrsh_spsize0 ] OR
#	     [ [ Max_size {gapped segment 1, gapped segment 2} >= Total{spacer sizes} * $mag_thrsh ] AND
#	     [ Min_size {gapped segment 1, gapped segment 2} >= Total{spacer sizes} / $mag_thrsh ] ] ;
#
#	(iii) EITHER the two gapped segments contain indel events along the same branch OR
#	      { they contain indel events along neighboring branches that form a triangle with another indel event (from either of the segments or from yet another 'neighboring' gapped segment) };
#
#	[ CREATION STARTED on Dec 17 (Sat), 2016, ENDED on Dec 18 (Sun), 2016.]	
#
# INPUT: (\@gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, \@eventsets_in_msa, \@set_self_nn_events, $upper_bound_spsize, \%br2eq_br, \%br2nn, $char_abs, \%node2pa, $thrsh_spsize0, $mag_thrsh), where
#
#	($beg_gs, $end_gs) = @{$gpd_segs_in_msa[$gs]}[$indx_beg_gs,$indx_end_bs] are the beginning and ending coordinates of the $gs th gapped segment in the subject MSA;
#
#	@eventset = @{$eventsets_in_msa[$gs]} stores the events in the $gs th gapped segment,
#	with @{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) being the information on the $ev th event. (It is basically an output of 'mk_eventset' in "MyTreeMap_indels_sppl.pm".);
#
#	\%self_nn_events = $set_self_nn_events[$gs] is the hash recording the self-nearest-neighboring event pairs in the $gs th gapped segment, with
#	%self_nn_events = ($br2 => {$br1 => \@set_eventpairs, ...}, ...),
#	where @{$set_eventpairs[$ep]} = ($gs, $indx_ev1, $gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. (This hash is basically an output of 'mk_self_nn_events'.);
#
#	$upper_bound_spsize is the parameter for the condition (i) above;
#
#	%br2eq_br = ($branch_id => $equivalent_branch_id, ...) for branches with equivalents.
#	(It is basically an output of 'create_br2eq_br' in "MyTreeMap_indels_sppl.pm".);
#
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...) for all branches.
#	(It is basically an output of 'create_br2nn' in "MyTreeMap_indels_sppl.pm".);
#
#	$char_abs is the character representing the absence of the site;
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node (i.e., root);
#
#	$thrsh_spsize0 and $mag_thrsh are the parameters for the condition (ii) above;
#
#
# OUTPUT: (\@clusters, \%gs2prev_nn_events), where
#
#	@{$clusters[$cl]} stores the indices of the gapped segments that constitute the $cl th artificial cluster.
#
#	%gs2prev_nn_events = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...),
#	    with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events).
#	    Incidentally, the suffixes, "_n" and "_c" indicate "next" and "current", respectively.
#
#
sub artf_cluster_gapped_segs_len (\@$$\@\@$\%\%$\%$$) {

	my ($gpd_segs_in_msa, $indx_beg_gs, $indx_end_gs, $eventsets_in_msa, $set_self_nn_events, $upper_bound_spsize, $br2eq_br, $br2nn, $char_abs, $node2pa, $thrsh_spsize0, $mag_thrsh) = @_;

	my $ct_gpd_segs = @{$gpd_segs_in_msa};

   ##(print STDOUT "<  Clustering with {Basic Spacer-Size Threshold} = $thrsh_spsize0 >\n\n") if ($TEST_LEVEL>1);

	my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #
			# END of "MOVED from within the 1st middle while-loop, to fix a BUG." #

	my $last_cluster = [0];

=For DEBUGGING (1)
	$last_cluster = [14];
	$ct_gpd_segs = 22;
=cut
=For DEBUGGING (2)
	$last_cluster = [35];
	$ct_gpd_segs = 42;
=cut

	my @clusters = ($last_cluster);

	while ($last_cluster->[$#{$last_cluster}] < $ct_gpd_segs-1) { # Main outer while-loop (untill all gapped segments are examined). #

   if ($TEST_LEVEL_acgsl>0) {
	print STDOUT "\n{{{ Initial 'last' cluster = (", join (', ', @{$last_cluster}), ") }}}\n\n";
   }

		my $indx_c_gs = 0;
		while ($indx_c_gs < @{$last_cluster}) { # Middle while-loop (until the "last" cluster is exhausted). #


	            my $c_gs = $last_cluster->[$indx_c_gs]; # Current gapped segment. #
	            my ($c_beg_coord, $c_end_coord) = @{$gpd_segs_in_msa->[$c_gs]}[$indx_beg_gs,$indx_end_gs];
		    my $c_gs_size = $c_end_coord - $c_beg_coord + 1;

   if ($TEST_LEVEL_acgsl>0) {
	print STDOUT "\n{{ The 'Current' segment = the $indx_c_gs th segment in the 'last' cluster (= the $c_gs th segment in the entire MSA) ; [beg, end] = [$c_beg_coord, $c_end_coord] ; size = $c_gs_size }}}\n\n";
   }
		    my $c_eventset = $eventsets_in_msa->[$c_gs];

			# push @eventsets_in_msa, \@eventset; element of @eventset = [$br, $stat_ue, $beg, $end, $len] 

		    my $c_self_nn_events = $set_self_nn_events->[$c_gs]; # = {$br2 => {$br1 => \@set_eventpairs, ...}, ...}, where @{$set_eventpairs[$ep]} = ($c_gs, $indx_ev1, $c_gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. #

		    my $tot_spsize = 0;

		    my $last_n_end_coord = $c_end_coord; # A BUG-FIXATION measure (A-1). #

		    for (my $n_gs = $c_gs+1; $n_gs < $ct_gpd_segs; $n_gs++) { # Inner for-loop (over "next" gapped segments). #

			my ($n_beg_coord, $n_end_coord) = @{$gpd_segs_in_msa->[$n_gs]}[$indx_beg_gs,$indx_end_gs];
			my $n_gs_size = $n_end_coord - $n_beg_coord + 1;

     if ($TEST_LEVEL_acgsl>0) {
	print STDOUT "\n{ The 'Next' segment = the $n_gs th segment in the entire MSA ;  [beg, end] = [$n_beg_coord, $n_end_coord] ; size = $n_gs_size }}\n\n";
     }

				# A BUG-FIXATION measure (A-2). #
			$tot_spsize += $n_beg_coord - $last_n_end_coord - 1;
			$last_n_end_coord = $n_end_coord;
			##$tot_spsize += $n_beg_coord - $c_end_coord - 1;
			####my $tot_spsize += $n_beg_coord - $c_end_coord - 1; # A BUG fixed on July 15 (Fri), 2016. #

    if ($TEST_LEVEL_acgsl>1) {
	print STDOUT "(Total spacer size) = $tot_spsize .\n\n";
    }

				## Replace the conditions with what I really meant. (1) ##
			($tot_spsize > $upper_bound_spsize) and last; # In this case, we give up extending the "last" cluster based on the current gapped segment. #
			## ($tot_spsize > $thrsh_spsize_lng) and last; # In this case, we give up extending the "last" cluster based on the current gapped segment. # ... OBSOLETE...

			my $n_eventset = $eventsets_in_msa->[$n_gs];
			    # push @eventsets_in_msa, \@eventset; element of @eventset = [$br, $stat_ue, $beg, $end, $len] 

			# @{$on_coinc_brs->[$k]} = ($br_c, $br_n, $indx_c_ev, $indx_n_ev) for the $k th pair of events along the same branch.
			# @{$on_nn_brs->[$m]} = ($br_c, $br_n, $indx_c_ev, $indx_n_ev) for the $m th pair of events along the nearest-neighboring branches.
			# @{$irrlv_insertions_n} stores the indices of the insertions in the "next" gapped segment along the branches that neither coincide with, nor is the nearest-neighbors of, any "current" branches.
			# 
			my ($on_coinc_brs, $on_nn_brs, $irrlv_insertions_n)
				= compare_indel_hosting_branches (@{$c_eventset}, @{$n_eventset}, %{$br2eq_br}, %{$br2nn}, $char_abs, %{$node2pa});

				# %{$br2nn}; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

			my $incl_spsize = $n_gs_size; # The increment of the spacer size. #

			foreach my $indx_ev (@{$irrlv_insertions_n}) { $incl_spsize -= $n_eventset->[$indx_ev][4]; }

   if ($TEST_LEVEL_acgsl>2) {
	print STDOUT "(Increment of the spacer-size that will be originated from the 'next' segment) = $incl_spsize .\n\n";
   }

			##$tot_spsize += $incl_spsize; # The "next" branch will be incorporated into the "spacer", after excluding the irrelevant insertions. # A BUG ... it was too early to incorporate it into the "spacer"!!

				## Replace the conditions with what I really meant. (2) ##

			my ($max_gs_size, $min_gs_size) = ($n_gs_size > $c_gs_size) ?
					($n_gs_size, $c_gs_size) : ($c_gs_size, $n_gs_size) ;

			unless (($tot_spsize <= $thrsh_spsize0) or 
				(($max_gs_size >= $tot_spsize * $mag_thrsh) and 
				($min_gs_size >= $tot_spsize / $mag_thrsh)) ) {
				##($max_gs_size >= $tot_spsize * $mag_thrsh) or
				##($min_gs_size >= $tot_spsize / $mag_thrsh) ) {

				# In this case, this "next" segment is NEVER considered as interacting with the current segment.

			    $tot_spsize += $incl_spsize; # A measure to fix the BUG which is one block above. #

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (A) ...\n\n") if ($TEST_LEVEL_acgsl>2);

			    next;
			}

			$tot_spsize += $incl_spsize; # A measure to fix the BUG which is two blocks above. #

=DISABLED in creation of 'artf_cluster_gapped_segs_len' (I) (as "DISABLED in creation of 'summary_examine_complex.6pp_aug_len.raw.pl'. (II)")

			if (@{$on_nn_brs}>0) {

				# Store the information on the interacting event pairs for later use. #

			    my $br2br2nn_events = $gs2prev_nn_events{$n_gs};
			    (defined $br2br2nn_events) or ($br2br2nn_events = $gs2prev_nn_events{$n_gs} = {});

			    foreach my $info_evpair (@{$on_nn_brs}) { # 2nd inner-inner foreach-loop (over nn event pairs). #
				my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
				my $br2nn_events = $br2br2nn_events->{$br_n};
				(defined $br2nn_events) or ($br2nn_events = $br2br2nn_events->{$br_n} = {});
				my $nn_events = $br2nn_events->{$br_c};
				(defined $nn_events) or ($nn_events = $br2nn_events->{$br_c} = []);
				push @{$nn_events}, [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev];
			    
			    }			# End of the 2nd inner-inner foreach-loop (over nn event pairs). #
				## my %gs2prev_nn_events; # = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...), with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events). #

			} # END of "if (@{$on_nn_brs}>0) {...}" #

=cut # END of "DISABLED in creation of 'artf_cluster_gapped_segs_len' (I) (as "DISABLED in creation of 'summary_examine_complex.6pp_aug_len.raw.pl'. (II)")"


			my $last_gs_in_cls = $last_cluster->[$#{$last_cluster}];

			unless ($last_gs_in_cls < $n_gs) {

   (print STDOUT " ... The 'next' segment is already incorporated in the cluster. Thus, skip the subsequent analyses ...\n\n") if ($TEST_LEVEL_acgsl>2);

			    next; # In this case, $n_gs is already incorporated in the cluster, thus we don't have to examine further. #
			}

				### Examine whether or not this "next" segment can interact with the current segment. ###

				## REPLACED the type (II) conditions with a more lenient one,
				## by IMPORTING a block from 'summary_examine_complex.6p_aug_len.raw.pl',
				## with the mediation of 'summary_examine_complex.6pp_aug_len.raw.pl',
				## into 'artf_cluster_gapped_segs_len'. ##

			if (@{$on_coinc_brs}>0) {

				# Interaction between $n_gs and $c_gs (I). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (I) ...\n\n") if ($TEST_LEVEL_acgsl>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} elsif (@{$on_nn_brs}>0) {

				# Interaction between $n_gs and $c_gs (II-len). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-len) ...\n\n") if ($TEST_LEVEL_acgsl>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} else {

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (C) ...\n\n") if ($TEST_LEVEL_acgsl>2);

			}

=DISABLED in creation of 'artf_cluster_gapped_segs_len' (II) (as "DISABLED in creation of 'summary_examine_complex.6pp_aug_len.raw.pl'. (III)")

				# First, examine the triangular interaction involving a past segment. #
				# (MOVED from below and augmented, to fix a BUG) #
			my $br2br2prev_nn_events = $gs2prev_nn_events{$c_gs};
			(defined $br2br2prev_nn_events) or ($br2br2prev_nn_events = {});

			my ($lm_igs, $rm_igs); # The indices for the leftmost and rightmost 'past' gapped segments interacting with $c_gs and $n_gs. 
			foreach my $info_evpair (@{$on_nn_brs}) { # 3rd inner-inner foreach-loop (over nn event pairs). #
			    my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
			    ##my $nn_c = $br2nn{$br_c}; # It's probably unnecessary any more. #
			    my $nn_n = $br2nn->{$br_n};
				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

			    my $br2prev_nn_events = $br2br2prev_nn_events->{$br_c};
			    (defined $br2prev_nn_events) or next; # Skip if the event along $br_c in $c_gs has no past neighboring events. #
			    foreach my $br (keys %{$br2prev_nn_events}) { # Outer-innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c. #
				(defined $nn_n->{$br}) or next; # Skip if $br is not the nearest-neighbor of $br_n as well. #
				my $prev_nn_events = $br2prev_nn_events->{$br};
				foreach my $pev_info (@{$prev_nn_events}) { # Inner-innermost foreach-loop (over the past interacting events). #
				    my ($p_igs, $indx_p_iev, $c_gs2, $indx_c_ev2) = @{$pev_info};
				    ((defined $lm_igs) and ($lm_igs <= $p_igs)) or ($lm_igs = $p_igs);
				    ((defined $rm_igs) and ($p_igs <= $rm_igs)) or ($rm_igs = $p_igs);
				}				# End of the inner-innermost foreach-loop (over the past interacting events). #
			    }				# End of the outer-innermost foreach-loop (over branches hosting previous events nearest-neighboring $br_c). #
			}				# End of the 5th inner-inner foreach-loop (over nn event pairs). #

			if (defined $lm_igs) {
				# Interaction between $n_gs and $c_gs (II-b). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-b) ...\n\n") if ($TEST_LEVEL_acgsl>2);

					# MODIFIED to FIX a BUG (i). #
			    my $counter = 0;
			    while ($lm_igs < $last_cluster->[0]) { $last_cluster = pop @clusters; $counter++; } # Get back to older clusters until the updated $last_cluster accommodates $lm_igs. #
			    ##while ($lm_igs < $last_cluster->[0]) { $last_cluster = pop @clusters; } # Get back to older clusters until the updated $last_cluster accommodates $lm_igs. #

			    my ($gs_lm_in_lc, $gs_rm_in_lc) = @{$last_cluster}[0, $#{$last_cluster}];

				# Update the $last_cluster and the $indx_c_gs. #
			    push @{$last_cluster}, ($gs_rm_in_lc + 1 .. $n_gs);

					# MODIFIED to FIX a BUG (ii). #
			    ($counter>0) and (push @clusters, $last_cluster);
			    ##push @clusters, $last_cluster;
			    $indx_c_gs = $c_gs - $gs_lm_in_lc;

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n";
	print STDOUT '    $indx_c_gs  = ', $indx_c_gs, " .\n\n\n";
   }

			    next;
			}

				# END of "(MOVED from below and augmented, to fix a BUG)" #

			if (@{$on_coinc_brs}>0) {

				# Interaction between $n_gs and $c_gs (I). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (I) ...\n\n") if ($TEST_LEVEL_acgsl>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;
			}

			unless (@{$on_nn_brs}>0) {
				# In this case, this "next" segment is NEVER considered as interacting with the current segment.

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (B) ...\n\n") if ($TEST_LEVEL_acgsl>2);
			    next;
			}

				## In the following, $n_gs is regarded as interacting with $c_gs
				## ONLY IF the nearest-neighboring events form a triangle encompassing a node.

		        my $n_self_nn_events = $set_self_nn_events->[$n_gs]; # = {$br2 => {$br1 => \@set_eventpairs, ...}, ...}, where @{$set_eventpairs[$ep]} = ($n_gs, $indx_ev1, $n_gs, $indx_ev2) for the $ep-th pair of nearest-neighboring events. #

#=MOVED to above interaction type (I), to fix a BUG. (1)
#=cut

			my $if_interact = 0;

			foreach my $info_evpair (@{$on_nn_brs}) { # 5th inner-inner foreach-loop (over nn event pairs). #
			    my ($br_c, $br_n, $indx_c_ev, $indx_n_ev) = @{$info_evpair};
			    my $nn_c = $br2nn->{$br_c};
			    my $nn_n = $br2nn->{$br_n};
				# %br2nn; # = ($br_id => {$id_nearest_neighbor => 1, ...}, ...)

				# (1) Examine the self-nearest-neighboring events in $n_gs. #

			    my $br2self_nn_n = $n_self_nn_events->{$br_n};
			    if (defined $br2self_nn_n) {
				foreach my $br (keys %{$br2self_nn_n}) { # 1st innermost foreach-loop (over branches self-neighboring $br_n). #
				    if (defined $nn_c->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 1st innermost foreach-loop (over branches self-neighboring $br_n). #
				$if_interact and last;
			    }

				# (2) Examine the self-nearest-neighboring events in $c_gs. #

			    my $br2self_nn_c = $c_self_nn_events->{$br_c};
			    if (defined $br2self_nn_c) {
				foreach my $br (keys %{$br2self_nn_c}) { # 2nd innermost foreach-loop (over branches self-neighboring $br_c). #
				    if (defined $nn_n->{$br}) {
					$if_interact = 1;
					last;
				    }
				}				# End of the 2nd innermost foreach-loop (over branches self-neighboring $br_c). #
				$if_interact and last;
			    }

#=MOVED TO ABOVE interaction type (I), and augmented, to fix a BUG (2).
#=cut # END of "MOVED TO ABOVE interaction type (I), and augmented, to fix a BUG (2)."

			}				# End of the 5th inner-inner foreach-loop (over nn event pairs). #

			if ($if_interact) {
				# Interaction between $n_gs and $c_gs (II-a). #

    (print STDOUT " ... Interaction between the 'current' cluster ($c_gs) and the 'next' cluster ($n_gs), type (II-a) ...\n\n") if ($TEST_LEVEL_acgsl>2);

			    push @{$last_cluster}, ($last_gs_in_cls + 1 .. $n_gs);

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT "    | Updated information: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

			    next;

			} else {

   (print STDOUT " ... The 'next' segment will NOT be incorporated into the cluster (at this point), type (C) ...\n\n") if ($TEST_LEVEL_acgsl>2);
			}

=cut # END of "DISABLED in creation of 'artf_cluster_gapped_segs_len' (II) (as "DISABLED in creation of 'summary_examine_complex.6pp_aug_len.raw.pl'. (III)")"

		    }						# End of the inner for-loop (over "next" gapped segments). #

		    $indx_c_gs++;

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT '    | Updated $indx_c_gs  = ', $indx_c_gs, " | \n\n\n";
    }

		}				# End of the middle while-loop (until the "last" cluster is exhausted). #

		my $last_gs_clustered = $last_cluster->[$#{$last_cluster}];
		($last_gs_clustered == $ct_gpd_segs-1) and last;

			# Create a new cluster. # 

		my @new_last_cluster = ($last_gs_clustered + 1);
		$last_cluster = \@new_last_cluster;
		push @clusters, $last_cluster;

   if ($TEST_LEVEL_acgsl>3) {
	print STDOUT "    | Information updated by default: |\n\n";
	print STDOUT '    @{$last_cluster} = (', join (', ', @{$last_cluster}), ") .\n\n\n";
    }

	}					# End of the main outer while-loop (until all gapped segments are examined). #

	return (\@clusters, \%gs2prev_nn_events);

}	# END of 'artf_cluster_gapped_segs_len'. #



# init_stats_artf_clusters:
#	This subroutine initializes statistics holders on the artificial clusters,
#	which will be used in 'stats_artf_clusters'.
#
# 	[CREATED on Dec 18 (Sun), 2016.]
#
# INPUT: (\@tags_gs_type, \@ctpairs), where
#
#	@tags_gs_type = ('B', 'A', 'Total').
#	# @TAGS_GS_TYPE = ('B', 'A', 'Total');
#
#	$ctpairs[$cp] = sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel) for the $cp th count-pair under consideration.
#
# OUTPUT: (\%ctpair2bins_gstype, \@bins_gstype_total), where
#
#	%ctpair2bins_gstype = ($label_ctpair => \@bins_gstype, ...), with
#	$label_ctpair = EITHER an element in @ctpairs OR 'Others' otherwise, and
#	@bins_gstype = (#{category 'B' gapped segments in the cluster}, #{category 'A' gapped segments in the cluster}) (Initialized by 'init_stats_artf_clusters'.);
#
#	@bins_gstype_total = (Total#{category 'B' gapped segments}, Total#{category 'A' gapped segments}) (Initialized by 'init_stats_artf_clusters'.);
#
sub init_stats_artf_clusters (\@\@) {

	my ($tags_gs_type, $ctpairs)  = @_;

	unless (($tags_gs_type->[0] eq 'B') and ($tags_gs_type->[1] eq 'A') and ($tags_gs_type->[2] eq 'Total') ) {
	    print STDOUT "[ERROR 'init_stats_artf_clusters'] An UNEXPECTED input: ", '@{$tags_gs_type} = (', join (', ', @{$tags_gs_type}), ") is NOT equal to (B, A, Total): $! !\n\n";
	    return (undef, undef); 
	}

	my %ctpair2bins_gstype;
	foreach my $ctpair (@{$ctpairs}, 'Others') {
	    my @bins_gstype = (0, 0);
	    $ctpair2bins_gstype{$ctpair} = \@bins_gstype;
	}
	my @bins_gstype_total = (0, 0);

	## $set_ctpair2bins_gstype[$i_thr] = \%ctpair2bins_gstype;
	## $set_bins_gstype_total[$i_thr]  = \@bins_gstype_total;
		## ... Although they were in the bulk of 'summary_examine_complex.6pp_aug.raw.pl', they have been removed from here.


	return (\%ctpair2bins_gstype, \@bins_gstype_total);

}	# END of the subroutine 'init_stats_artf_clusters'. #


# stats_artf_clusters:
#	This subroutine takes statistics on the artificial clusters.
#
#	[CREATED on Dec 18 (Sun), 2016.]
#
# INPUT: (\@clusters, \@gpd_segs_in_msa, \%ctpair2bins_gstype, \@bins_gstype_total), where
#
#	@{$clusters[$cl]} stores the indices of the gapped segments that constitute the $cl th artificial cluster. (It is basically an output of 'artf_cluster_gapped_segs'.);
#
#	@{$gpd_segs_in_msa[$gs]} = ($category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel) is for the $gs th gapped segment in the MSA. (Its indel-related information could be calculated using the outputs of 'counts_totlens_indels' in "MyTreeMap_indels_sppl.pm".);
#
#	%ctpair2bins_gstype = ($label_ctpair => \@bins_gstype, ...), with
#	$label_ctpair = EITHER sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel) for count-pairs under consideration OR 'Others' otherwise, and
#	@bins_gstype = (#{category 'B' gapped segments in the cluster}, #{category 'A' gapped segments in the cluster}) (Initialized by 'init_stats_artf_clusters'.);
#
#	@bins_gstype_total = (Total#{category 'B' gapped segments}, Total#{category 'A' gapped segments}) (Initialized by 'init_stats_artf_clusters'.);
#
# OUTPUT: 1 if the subroutine successfully ended,
#	  0 otherwise.
#
sub stats_artf_clusters (\@\@\%\@) {

	my ($clusters, $gpd_segs_in_msa, $ctpair2bins_gstype, $bins_gstype_total) = @_;


	foreach my $cluster (@{$clusters}) { # Outer foreach-loop (over clusters in the subject MSA). #

		my $sbct_gpd_segs = @{$cluster};
		my $ct_A = 0;
		my $totct_clms = 0;
		my $tot_totct_indel = my $tot_diffct_indel = my $tot_totlen_indel = my $tot_difflen_indel = 0;

		foreach my $gs (@{$cluster}) { # Inner foreach-loop (over gapped segments in the subject cluster). #

			# => Each element of @gpd_segs_in_msa will become:
			# [$category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel]
		    my ($category, $beg_coord, $end_coord, $ct_clms, $totct_indel, $diffct_indel, $totlen_indel, $difflen_indel) = @{$gpd_segs_in_msa->[$gs]};

		    ($category eq 'A') and ($ct_A++);
		    $totct_clms += $ct_clms;
		    $tot_totct_indel += $totct_indel;
		    $tot_diffct_indel += $diffct_indel;
		    $tot_totlen_indel += $totlen_indel;
		    $tot_difflen_indel += $difflen_indel;

		}		# End of the inner foreach-loop (over gapped segments in the subject cluster). #

			# MODIFIED in creation of 'summary_examine_complex.6pp_aug.raw.pl'. (IIb) #
		my $ct_B = $sbct_gpd_segs - $ct_A;
		#my $pc_A = 100 * $ct_A /$sbct_gpd_segs;


			# MODIFIED in creation of 'summary_examine_complex.6p.raw.pl'. (IIb) #
		my $ctpair = sprintf "(%d,%d)", $tot_totct_indel, $tot_diffct_indel;

			# => Further MODIFIED in creation of 'summary_examine_complex.6pp_aug.raw.pl'. (IIc) #

		my $bins_gstype = $ctpair2bins_gstype->{$ctpair};
		(defined $bins_gstype) or ($bins_gstype = $ctpair2bins_gstype->{'Others'});

		foreach my $bins ($bins_gstype, $bins_gstype_total) { # 2nd inner foreach-loop (over the relevant sets of bins). #
		    $bins->[0] += $ct_B;
		    $bins->[1] += $ct_A;
		    ##$bins->[2] += $sbct_gpd_segs;
		}					# End of the 2nd inner foreach-loop (over the relevant sets of bins). #



	}		    # End of the outer foreach-loop (over clusters in the subject MSA). #


	return 1;

}	# END of 'stats_artf_clusters'. #


# mk_labels_stats_artf_clusters:
#	This subroutine creates a set of labels
#	used when printing out the statistics on the artificial clusters.
#
#	[ CREATED on Dec 19 (Mon), 2016. ]
#
# INPUT: (\@tags_gs_type), where
#
#	@tags_gs_type stores the tags for the types of gapped segments.
#	(As a usual expectation, @tags_gs_type = ('B', 'A', 'Total').)
#
# OUTPUT: \@labels_out, where
#
#	@labels_out stores the labels for the outputs that will be printed out.
#	(Absolute frequencies, ratio of Pure A to Pure B, percentages, and cumulative percentages, etc.)
#
sub mk_labels_stats_artf_clusters (\@) {

	my ($tags_gs_type) = @_;

	my @labels_out = ('(X,Y) \ Type_gs', '', @{$tags_gs_type}, 'PureA / PureB');
	foreach my $intvl (@{$tags_gs_type}) { (uc($intvl) =~ m/^TOT/) or (push @labels_out, "$intvl (%)"); }
	foreach my $intvl (@{$tags_gs_type}) { (uc($intvl) =~ m/^TOT/) or (push @labels_out, "$intvl (cum%)"); }

	return \@labels_out;

}	# END of 'mk_labels_stats_artf_clusters'. #


# pr_heading_stats_artf_clusters:
#	This subroutine prints out the headline and some legends
#	for the statistics on the artificial clusters.
#
# INPUT: (\*${$fh_out}, $section, $descr_conditions), where
#
#	$fh_out is the reference to a filehandle through which the statistics will be printed out;
#
#	$section labels this particular section of the print-out;
#
#	$descr_conditions is a string that briefly describes the set of conditions, like 'threshold value', 
#	under each of which an independent set of statistics were taken.
#
# OUTPUT: 1 if the subroutine ended successfully,
#	  0 otherewise.
#
sub pr_heading_stats_artf_clusters (\*$$) {

	my ($fh_out, $section, $descr_conditions) = @_;

	print $fh_out "\n\n[$section] 2D distributions, for each of the $descr_conditions, of (The type (A or B) of the segment that constituent gapped segments belong to) (abscissa) and (Tot[Tot#{indels}], Tot[#{ins} - #{del}]) (ordinates) for the artificial clusters of gapped segments:\n";

	print $fh_out "\n( NOTE1: In each cell is the number of gapped segments in the specified class of artificial clusters. )\n";

	print $fh_out "(NOTE2: X = Tot[#{ins} + #{del}], Y = Tot[#{ins} - #{del}]. )\n\n"; # MODIFIED on July 2 (Sat), 2016. #
	##print $fh_out "(NOTE2: X = Tot[Tot#{indels}], Y = Tot[#{ins} - #{del}]. )\n\n";

	return 1;

}	# END of 'pr_heading_stats_artf_clusters'. #


# pr_stats_artf_clusters:
#	This subroutine prints out the statistics on the artificial clusters,
#	mainly taken via 'stats_artf_clusters'.
#
#	[CREATED on Dec 19 (Mon), 2016, based on the trailing part of 'summary_examine_complex.6pp_aug.raw.pl'.]
#
# INPUT: (\*${$fh_out}, \@ctpairs, \%ctpair2bins_gstype, \@bins_gstype_total, \@labels_out), where
#
#	$fh_out is the reference to a filehandle through which the statistics will be printed out;
#
#	$ctpairs[$cp] = sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel) for the $cp th count-pair under consideration.
#	(It was also used as an input of 'init_stats_artf_clusters'.)
#
#	%ctpair2bins_gstype = ($label_ctpair => \@bins_gstype, ...), with
#	$label_ctpair = EITHER sprintf ("(%d,%d)", $tot_totct_indel, $tot_diffct_indel) for count-pairs under consideration OR 'Others' otherwise, and
#	@bins_gstype = (#{category 'B' gapped segments in the cluster}, #{category 'A' gapped segments in the cluster})
#	(Initialized by 'init_stats_artf_clusters' and nurtured by 'stats_artf_clusters'.);
#
#	@bins_gstype_total = (Total#{category 'B' gapped segments}, Total#{category 'A' gapped segments}) 
#	(Initialized by 'init_stats_artf_clusters' and nurtured by 'stats_artf_clusters'.);
#
#	@labels_out stores the labels for the outputs that will be printed out.
#	(Absolute frequencies, ratio of Pure A to Pure B, percentages, and cumulative percentages, etc.)
#	(It should be created by 'mk_labels_stats_artf_clusters'.)
#
# OUTPUT: 1 if the subroutine ended successfully,
#	  0 otherwise.
#
sub pr_stats_artf_clusters (\*\@\%\@\@) {

	my ($fh_out, $ctpairs, $ctpair2bins_gstype, $bins_gstype_total, $labels_out) = @_;


	my $ct_bins = @{$bins_gstype_total};

	my %ctpair2ratio_AB; # For those count pairs with #{pure B} > 0.
	my %ctpair2ct_pureA; # For those count pairs with #{pure B} = 0.

	foreach my $ctpair (@{$ctpairs}) {
	    my $bins_gstype = $ctpair2bins_gstype->{$ctpair};
	    my ($ct_pureB, $ct_pureA) = @{$bins_gstype}[0, 1];
	    if ($ct_pureB>0) {
		$ctpair2ratio_AB{$ctpair} = $ct_pureA/$ct_pureB;
	    } else {
		$ctpair2ct_pureA{$ctpair} = $ct_pureA;
	    }
	}

	my @std_ctpairs = sort { $ctpair2ratio_AB{$a} <=> $ctpair2ratio_AB{$b} } keys %ctpair2ratio_AB;
	push @std_ctpairs, sort { $ctpair2ct_pureA{$a} <=> $ctpair2ct_pureA{$b} } keys %ctpair2ct_pureA;

	## print STDOUT "\n( For clustering segments with {spacer-size} <= $thr }\n\n"; # Although it was originally in the trailing part of 'summary_examine_complex.6pp_aug.raw.pl', it has been removed. A similar heading should be prepended to this entire output. #

	my @cum_freqs = ();
	for (1 .. $ct_bins) { push @cum_freqs, 0; }

	print $fh_out join ("\t", @{$labels_out}), "\n\n";

	foreach my $ctpair (@std_ctpairs, 'Others') {
	    my $bins_gstype = $ctpair2bins_gstype->{$ctpair};
	    my ($ct_pureB, $ct_pureA) = @{$bins_gstype}[0, 1];
	    my $ratio_AB = ($ct_pureB>0) ? $ct_pureA/$ct_pureB : 'undef' ; 

	    my @freqs = ();
	    my $totct = 0;
	    for (my $i=0; $i<$ct_bins; $i++) {
		my $ct = $bins_gstype->[$i];
		$totct += $ct;

		my $totct_clm = $bins_gstype_total->[$i];
		my $freq = ($totct_clm > 0) ? 100 * $ct / $totct_clm : 'undef';

		$freqs[$i] = $freq;

		if (($freq eq 'undef') or ($cum_freqs[$i] eq 'undef')) {
		   $cum_freqs[$i] = 'undef';
		} else {
		   $cum_freqs[$i] += $freq;
		}
	    }
	    my @out = (sprintf ("%-10s", $ctpair), '');
	    foreach my $ct ( @{$bins_gstype}, $totct) { push @out, sprintf ("%7d", $ct); }

	    push @out, (($ratio_AB eq 'undef') ? sprintf ("%13s", $ratio_AB) : sprintf ("%13.8f", $ratio_AB));
	    foreach my $freq (@freqs) { push @out, (($freq eq 'undef') ? sprintf ("%6s", $freq) : sprintf ("%6.2f%%", $freq)); }
	    foreach my $cfreq (@cum_freqs) { push @out, (($cfreq eq 'undef') ? sprintf ("%6s", $cfreq) : sprintf ("%6.2f%%", $cfreq)); }

	    print $fh_out join ("\t", @out), "\n";
	}

	print $fh_out "\n";

	my $totalct_sbjcts = 0;
	foreach my $total (@{$bins_gstype_total}) { $totalct_sbjcts += $total; }
	my ($totct_pureB, $totct_pureA) = @{$bins_gstype_total}[0, 1];

	my $ratio_AB_total = ($totct_pureB > 0) ? $totct_pureA/$totct_pureB : 'undef';

	my @out_total = (sprintf ("%-10s", 'Total'), '');
	foreach my $ct (@{$bins_gstype_total}, $totalct_sbjcts) { push @out_total, sprintf ("%7d", $ct); }

	push @out_total, (($ratio_AB_total eq 'undef') ? sprintf ("%13s", $ratio_AB_total) : sprintf ("%13.8f", $ratio_AB_total));
	for (my $i=0; $i<$ct_bins; $i++) { push @out_total, (($bins_gstype_total->[$i]>0) ? sprintf ("%6.2f%%", 100) : 'undef'); }

	print $fh_out join ("\t", @out_total), "\n\n";

	return 1;

}	# END of 'pr_stats_artf_clusters'. #


# detect_cmplx_error_cand_acls0:
#	This subroutine detects artificial clusters (of gapped segments) that are likely to contain complex errors,
#	based simply on the numbers of insertions and deletions in the Dollo-parsimonious indel history
#	that can result in the cluster.
#	[CREATION STARTED on Jan 14 (Sun), 2018, ENDED on Jan dd (www), 2018.]
#
# INPUT: (\@artf_clusters, \@eventsets_in_msa, $char_abs, \@list_indel_cts_for_ncmplx), where
#
#	@{$artf_clusters[$ac]} stores the indices of the gapped segments that constitute the $ac th artificial cluster;
#
#	@eventset = @{$eventsets_in_msa[$gs]} stores the events in the $gs th gapped segment,
#	with @{$eventset[$ev]} = ($br, $stat_ue, $beg, $end, $len) being the information on the $ev th event. (It is basically an output of 'mk_eventset' in "MyTreeMap_indels_sppl.pm".);
#
#	@{$list_indel_cts_for_ncmplx[$k]} = ($totct_ins + $totct_del, $totct_ins - $totct_del) 
#	gives the counts of insertions and deletions for the $k th most likely category for non-complex artificial clusters.
#	(It is typically the 3rd element of @PARAMS_ACGS_XXX.)
#
# OUTPUT: (\@cmplx_cands, \@ncmplx_cands), where
#
#	@{$cmplx_cands[$cc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $cc th artificial cluster likely to be involved in complex errors, where
#	$indx_ac is the index in @artf_cluster, $totct_indels = $totct_ins + $totct_dels, and  $diffct_indels = $totct_ins - $totct_dels in the artificial cluster;
#	@{$ncmplx_cands[$nc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $nc th artificial cluster UNlikely to be involved in complex errors.
#
sub detect_cmplx_error_cand_acls0 (\@\@$\@) {

	my ($artf_clusters, $eventsets_in_msa, $char_abs, $list_indel_cts_for_ncmplx) = @_;

	my $ct_acls = @{$artf_clusters};

    (print STDOUT "< Subroutine 'detect_cmplx_error_cand_acls0' (#{input artificial clusters} = $ct_acls) >\n\n")  if ($TEST_LEVEL_dceca > 0);


	my %ctpair_ncmplx; # = ("$totct_indels:$diffct_indels" => 1, ...) for the counts of indels indicating that the artificial cluster is NOT involved in complex errors. ($totct_indels = $ct_ins + $ct_del; $diffct_indels = $ct_ins - $ct_del)

   (print STDOUT "[ List of 'Tot#{indels}:Diff#{indels}' indicating NON-involvement of complex errors ]\n\n") if ($TEST_LEVEL_dceca > 2);

	foreach my $ctpair (@{$list_indel_cts_for_ncmplx}) { # Preliminary foreach-loop (over "non-complex" count pairs). #
	    my $cnct_ctpair = join (':', @{$ctpair});
	    $ctpair_ncmplx{$cnct_ctpair} = 1;
   (print STDOUT "  $cnct_ctpair,") if ($TEST_LEVEL_dceca > 2);
	}					  # End of the preliminary foreach-loop (over "non-complex" count pairs). #

   (print STDOUT " .\n\n\n") if ($TEST_LEVEL_dceca > 2);

   (print STDOUT "\n[ MAIN Process:  Sorting artificial clusters into 'Complex' and 'Non-Complex' ]\n\n") if ($TEST_LEVEL_dceca > 1);

	my @cmplx_cands = my @ncmplx_cands = ();

	for (my $ac = 0; $ac < $ct_acls; $ac++) { # Main outer for-loop (over artificial clusters). #

	    my $acls = $artf_clusters->[$ac];

   (print STDOUT ' ' x 2, "{ The $ac th artificial cluster: (", join (', ', @{$acls}), ") }\n\n") if ($TEST_LEVEL_dceca > 2);

	    my $ct_ins = my $ct_dels = 0;

   (print STDOUT ' ' x 2, join ("\t", 'Indx_gs', 'Indx_ev', 'br_ID', 'stat_ue', 'beg', 'end', 'len'), "\n\n") if ($TEST_LEVEL_dceca > 5);

	    foreach my $indx_gs (@{$acls}) { # Middle foreach-loop (over gapped segments). #

		my $eventset = $eventsets_in_msa->[$indx_gs];
		my $ct_events = @{$eventset};
		for (my $e = 0; $e < $ct_events; $e++) {  # Inner for-loop (over indel events). #
		    my $event = $eventset->[$e];
		    ($event->[1] eq $char_abs) ? ($ct_ins++) : ($ct_dels++);

   (print STDOUT ' ' x 2, join ("\t", $indx_gs, $e, @{$event}), "\n") if ($TEST_LEVEL_dceca > 5);

		}				# End of the inner for-loop (over indel events). #
	    }			   # End of the middle foreach-loop (over gapped segments). #

   (print STDOUT "\n\n") if ($TEST_LEVEL_dceca > 5);

	    my ($totct_indels, $diffct_indels) = ($ct_ins + $ct_dels, $ct_ins - $ct_dels);
	    my $cnct_ctpair = join (':', $totct_indels, $diffct_indels);

   (print STDOUT ' ' x 2, '(#{ins}, #{del}, Tot#{indels}, Diff#{indels}) = ', "($ct_ins, $ct_dels, $totct_indels, $diffct_indels).\n\n") if ($TEST_LEVEL_dceca > 3);

	    if (defined $ctpair_ncmplx{$cnct_ctpair}) {
		push @ncmplx_cands, [$ac, $totct_indels, $diffct_indels];

   (print STDOUT  ' ' x 2, "Classification = 'Non-Complex'.\n\n") if ($TEST_LEVEL_dceca > 3);

	    } else {

		push @cmplx_cands, [$ac, $totct_indels, $diffct_indels];

   (print STDOUT  ' ' x 2, "Classification = 'Complex'.\n\n") if ($TEST_LEVEL_dceca > 3);
	    }

	}				# End of the main outer for-loop (over artificial clusters). #


    (print STDOUT "\n< Subroutine 'detect_cmplx_error_cand_acls0' Ended />\n\n")  if ($TEST_LEVEL_dceca > 0);

	return (\@cmplx_cands, \@ncmplx_cands);

}	# END of 'detect_cmplx_error_cand_acls0'. #


# wrapper_detect_cmplx_error_cand_acls0:
#	This subroutine is a wrapper of the MAIN subroutine, 'detect_cmplx_error_cand_acls0'.
#	It enables a ONE-STOP DETECTION of complex-error candidates
#	from the sets of gapped segments & initical candidates of parsimonious indel histories.
#	Specifically, the subroutine lumps together the following subroutines:
#	'mk_ingredients_acls_gpd_segs',
#	'artf_cluster_gapped_segs(_len)',
#	'detect_cmplx_error_cand_acls0',
#	into a single subroutine.
#	[CREATED on 2019/06/12]
#
# INPUTS: (\@set_cls_gpblocks, \@init_psm_hsts_for_cls, \@set_exnodes, $char_abs, \%node2ch, \%br2eq, \%br2nn, $if_len, $upper_bound_spsize, $thrsh_spsize0, $mag_thrsh, \@list_indel_cts_for_ncmplx, \*${$fh_log}), where
#
#	@set_cls_gpblocks stores information on the gapped-segments, i.e., clusters of gap-pattern-blocks.
#	More precisely,
#	($indx_gp, $len, $beg, $end) = @{$set_cls_gpblocks->[$cls][$gpb]} are information on the $gpb th gap-pattern-block in the $cls th gapped segment, where
#	$indx_gp is the index of the gap-pattern,
#	$len, $beg, $end are the length, beginning coordinate, and end coordinate, respectively,
#	 of the gap-pattern-block;
#	[\@set_cls_gpblocks is an important input of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm"];
#
#
#	@{$init_psm_hsts_for_cls[$cls]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious history resulting in the $cls th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#	[\@init_psm_hsts_for_cls is an important output of 'cal_init_psm_cands_for_sgl_msa' in "MyTreeMap_indels_spt_odr2.pm"];
#
#	@set_exnodes stores the IDs of all external nodes;
#
#	$char_abs is the character representing an 'absence' state (i.e., gap);
#
#	%node2ch = ($node_id => \@child_node_ids, ...) for all external nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) for all nodes except the root;
#
#	%br2eq = ($br_id => $equiv_br_id, ...) for the branches with their own equivalents;
#	%br2nn = ($br_id => {$nearest_neighboring_br_id => 1, ...}, ...) for the branches with their own nearest neighbors;
#	[Typically, \%br2eq and \%br2nn are among the outputs of 'enum_psm_indel_hsts_for_msa_finer_PRE' in "MyTreeMap_indels_spt_odr_finer.pm".];
#
#	$if_len is a switch to choose a subroutine for artificially clustering the gapped segments.
#	Set $if_len = 0 to use 'artf_cluster_gapped_segs',
#		    = 1 to use 'artf_cluster_gapped_segs_len';
#
#	$upper_bound_spsize is the parameter for the condition (i) described in the header of 'artf_cluster_gapped_segs(_len)';
#
#	$thrsh_spsize0 and $mag_thrsh are the parameters for the condition (ii) described in the heder of 'artf_cluster_gapped_segs(_len)';
#
#	@list_indel_cts_for_ncmplx lists the insertion-deletion count pairs for non-complex artificial clusters. More precisely,
#	@{$list_indel_cts_for_ncmplx[$k]} = ($totct_ins + $totct_del, $totct_ins - $totct_del) 
#	gives the counts of insertions and deletions for the $k th most likely category for non-complex artificial clusters;
#	[It is typically the 3rd element of @PARAMS_ACGS_XXX.];
#
#	$fh_log is the reference to the file-handle for some test outputs;
#
# OUTPUTS: (\@cmplx_cands, \@ncmplx_cands, \@artf_clusters, \%gs2prev_nn_events, \@set_lists_gblocks), where
#
#	@{$cmplx_cands[$cc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $cc th artificial cluster likely to be involved in complex errors, where
#	$indx_ac is the index in @artf_cluster, $totct_indels = $totct_ins + $totct_dels, and  $diffct_indels = $totct_ins - $totct_dels in the artificial cluster;
#
#	@{$ncmplx_cands[$nc]} = ($indx_ac, $totct_indels, $diffct_indels) gives information on the $nc th artificial cluster UNlikely to be involved in complex errors;
#
#	[\@cmplx_cands and \@ncmplx_cands are the outputs of 'detect_cmplx_error_cand_acls0' above];
#
#
#	@{$atrf_clusters[$acl]} stores the indices of the gapped segments that constitute the $acl th artificial cluster;
#
#	%gs2prev_nn_events = ($n_gs => {$br_n => {$br_c => \@set_eventpairs, ...}, ...}, ...),
#	    with @{$set_eventpairs[$ep]} = [$c_gs, $indx_c_ev, $n_gs, $indx_n_ev] for the $ep th "interacting" event pair along the nearest-neighboring branch pair, ($br_c, $br_n), (excluding the self-nn-events).
#	    Incidentally, the suffixes, "_n" and "_c" indicate "next" and "current", respectively;
#
#	[\@artf_clusters and \%gs2prev_nn_events are the outputs of 'artf_cluster_gapped_segs(_len)' above];
#
#	@{$set_lists_gblocks[$gs]->[$gb]} provides information on the $gb th gap-block in the $gs th gapped segment. More precisely, 
#	($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@aff_exnodes, $e) = @{$set_lists_gblocks[$cl]->[$gb], where
#	the gap-block is separated by, and on the $up_or_lw (='U' or 'L') side of, branch $br,
#	$lm_gpb and $rm_gpb are the indices of the left-most and right-most gap-pattern blocks,
#	@invlvd_gpbs lists the gap-pattern-blocks involved;
#	@aff_exnodes lists the IDs of external nodes affected by the gap-block,
#	$e is the index of the bounding indel event in @{$br2indels{$br}};
#
#	[@br_list_gblocks = @{$set_lists_gblocks[$gs]} is the output of 'br_list_gblocks' (in "MyTreeMap_indels_spt_odr2.pm").];
#
sub wrapper_detect_cmplx_error_cand_acls0 (\@\@\@$\%\%\%\%$$$$\@\*) {

	my ($set_cls_gpblocks, $init_psm_hsts_for_cls, $set_exnodes, $char_abs, $node2ch, $node2pa, $br2eq, $br2nn, $if_len, $ub_spsize, $thrsh_spsize0, $mag_thrsh, $list_indel_cts_for_ncmplx, $fh_log) = @_;

	my $ct_cls = @{$set_cls_gpblocks};
	my $ct_exnodes = @{$set_exnodes};

    (print STDOUT "<< Subroutine 'wrapper_detect_cmplx_error_cand_acls0' (#{gapped segments}=$ct_cls, #{external_nodes}=$ct_exnodes, \$if_len = $if_len)  />/>\n\n") if ($TEST_LEVEL_wrdceca > 0);
    (print $fh_log "<< Subroutine 'wrapper_detect_cmplx_error_cand_acls0' (#{gapped segments}=$ct_cls, #{external_nodes}=$ct_exnodes, \$if_len = $if_len)  />/>\n\n") if ($TEST_LEVEL_wrdceca > 0);

   if ($TEST_LEVEL_wrdceca >1) { # (1) #

	print STDOUT "\n";
	print STDOUT "[NOTE1]  SWITCH ON '\$MyDetect_cmplx_error_cands::TEST_LEVEL_migacgss', if you want to know what's going on with 'mk_ingredients_acls_gpd_segs' in detail.\n";
	print STDOUT "[NOTE2a] SWITCH ON '\$MyDetect_cmplx_error_cands::TEST_LEVEL_acgs', if you want to know what's going on with 'artf_cluster_gapped_segs' in detail.\n";
	print STDOUT "[NOTE2b] SWITCH ON '\$MyDetect_cmplx_error_cands::TEST_LEVEL_acgsl', if you want to know what's goin on with 'artf_cluster_gapped_segs_len' in detail.\n";
	print STDOUT "[NOTE3]  SWITCH ON '\$MyDetect_cmplx_error_cands::TEST_LEVEL_dceca', if you want to know what's goin on with 'detect_cmplx_error_cand_acls0' in detail.\n";
	print STDOUT "\n\n";

	print $fh_log '[ Input Parameters ]', "\n\n";

	print $fh_log '$if_len = ', $if_len, " .\n\n";
	print $fh_log '$ub_spsize = ', $ub_spsize, " ,\n";
	print $fh_log '$thrsh_spsize0 = ', $thrsh_spsize0, " ,\n";
	print $fh_log '$mag_thrsh = ', $mag_thrsh, " .\n\n";

	print $fh_log '@{$list_indel_cts_for_ncmplx} = ( # element = [#{insertions} + #{deletions}, #{insertions} - #{deletions}],', "\n";
	foreach my $ct_pair (@{$list_indel_cts_for_ncmplx}) {
		print $fh_log "\t[", join (', ', @{$ct_pair}), "],\n";
	}
	print $fh_log "\t\t) .\n";
	print STDOUT "\n\n";

   } # End of "if ($TEST_LEVEL_wrdceca >1) { # (1) # ...}"


	my ($bds_gpd_segs_in_msa, $set_eventsets, $set_self_nn_events, $set_lists_gblocks)
		= mk_ingredients_acls_gpd_segs (@{$set_cls_gpblocks}, @{$init_psm_hsts_for_cls}, @{$set_exnodes}, $char_abs, %{$node2ch}, %{$br2eq}, %{$br2nn});
	unless (defined $set_lists_gblocks) {
		print STDOUT "[ERROR 'wrapper_detect_cmplx_error_cand_acls0'] The subroutine 'mk_ingredients_acls_gpd_segs' FAILED on the input tree & initial indel history candidate: $! !\n\n";
		return (undef, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... Subroutine 'mk_ingredients_acls_gpd_segs' Ended ...\n\n") if ($TEST_LEVEL_wrdceca > 1);
   (print $fh_log "\n ... Subroutine 'mk_ingredients_acls_gpd_segs' Ended ...\n\n") if ($TEST_LEVEL_wrdceca > 1);


	my ($artf_cls, $gs2prev_nn_events) = $if_len ? 
	artf_cluster_gapped_segs_len (@{$bds_gpd_segs_in_msa}, 0, 1, @{$set_eventsets}, @{$set_self_nn_events}, $ub_spsize, %{$br2eq}, %{$br2nn}, $char_abs, %{$node2pa}, $thrsh_spsize0, $mag_thrsh) :
	artf_cluster_gapped_segs (@{$bds_gpd_segs_in_msa}, 0, 1, @{$set_eventsets}, @{$set_self_nn_events}, $ub_spsize, %{$br2eq}, %{$br2nn}, $char_abs, %{$node2pa}, $thrsh_spsize0, $mag_thrsh);
	unless (defined $artf_cls) {
		print STDOUT "[ERROR 'wrapper_detect_cmplx_error_cand_acls0'] The MAIN subroutine '", 
			($if_len ? 'artf_cluster_gapped_segs_len' : 'artf_cluster_gapped_segs'), 
			"' FAILED: $! !!\n\n";
		return (undef, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... Subroutine '", ($if_len ? 'artf_cluster_gapped_segs_len' : 'artf_cluster_gapped_segs'), "' Ended ...\n\n") if ($TEST_LEVEL_wrdceca > 1);
   (print $fh_log "\n ... Subroutine '", ($if_len ? 'artf_cluster_gapped_segs_len' : 'artf_cluster_gapped_segs'), "' Ended ...\n\n") if ($TEST_LEVEL_wrdceca > 1);

	my $ct_acls = @{$artf_cls};

   if ($TEST_LEVEL_wrdceca >1) { # (2) #

	print $fh_log "\n", '[ The content of @{$artf_cls}, the main output of ', 
		($if_len ? "'artf_cluster_gapped_segs_len'" : "'artf_cluster_gapped_segs'"),
 		" (#{artificial clusters} = $ct_acls) ]\n\n";
	print $fh_log join ("\t", 'Indx_acls', '[indices,of,constituent,gapped,segments]'), "\n\n";
	for (my $ac = 0; $ac < $ct_acls; $ac++) { # The 1st main for-loop (over artificial clusters). #
	    my $cnct_indices_gsegs = join (',', @{$artf_cls->[$ac]});
	    print $fh_log  join ("\t", $ac, "[$cnct_indices_gsegs]"), "\n";
	}				# End of the 1st main for-loop (over artificial clusters). #
	print $fh_log "\n\n";

   } # End of "if ($TEST_LEVEL_wrdceca >1) { # (2) # ...}"


	my ($cmplx_cands, $ncmplx_cands) = detect_cmplx_error_cand_acls0 (@{$artf_cls}, @{$set_eventsets}, $char_abs, @{$list_indel_cts_for_ncmplx});
	unless (defined $cmplx_cands) {
		print STDOUT "[ERROR 'wrapper_detect_cmplx_error_cand_acls0'] The MAIN subroutine 'detect_cmplx_error_cand_acls0' FAILED: $! !!\n";
		return (undef, undef, undef, undef, undef);
	}
   (print STDOUT "\n ... Subroutine 'detect_cmplx_error_cand_acls0' Ended ...\n\n") if ($TEST_LEVEL_wrdceca > 1);
   (print $fh_log "\n ... Subroutine 'detect_cmplx_error_cand_acls0' Ended ...\n\n") if ($TEST_LEVEL_wrdceca > 1);

	my $ct_cands = @{$cmplx_cands};
	my $ct_ncands = @{$ncmplx_cands};

    (print STDOUT "<< Subroutine 'wrapper_detect_cmplx_error_cand_acls0' ENDED (w/ #{candidates}=$ct_cands, #{non-candidates}=$ct_ncands) />/>\n\n") if ($TEST_LEVEL_wrdceca > 0);
    (print $fh_log "<< Subroutine 'wrapper_detect_cmplx_error_cand_acls0' ENDED (w/ #{candidates}=$ct_cands, #{non-candidates}=$ct_ncands) />/>\n\n") if ($TEST_LEVEL_wrdceca > 0);


	return ($cmplx_cands, $ncmplx_cands, $artf_cls, $gs2prev_nn_events, $set_lists_gblocks)

}	# End of the subroutine, 'wrapper_detect_cmplx_error_cand_acls0'. #



1;		# a "spell" which makes the module "use"-able and/or "require"-able #


