package MyHash_2;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyHash_2: a package for various hash manipulations 
#	[CREATION ENDED on 2006/12/09.]
#
#	[REVISED on 2019/02/22.] 'pr_comp_array2' was created.
#

use strict;
use warnings;
use Exporter;
our @ISA = qw(Exporter);
##our @EXPORT = qw( merge_hash diff_hash);
our @EXPORT_OK = qw(
	merge_hash diff_hash 
	pr_comp_hash pr_comp_array
	pr_comp_array2
	cp_comp_array cp_comp_hash
	construct_relayed_hash
	);

sub pr_comp_array (\@\*$);
sub pr_comp_array2 (\@\*$); # ADDED on 2019/02/22.
sub pr_comp_hash (\%\*$);
sub cp_comp_array (\@);
sub cp_comp_hash (\%);

our $MYTAB = ' ' x 3;


# pr_comp_hash:
#	This subroutine can print out a composite hash,
#	whose keys may be the references to scalars, arrays, and/or hashes.
#
# INPUT: (\%hash, \*$fh, $level),
#	where $level increases as we go deeper into the composite structure.
#
# OUTPUT: 1 if succeeded in printing out the whole hash,
#	  0 if failed.
#
sub pr_comp_hash (\%\*$) {

	my ($hash, $fh, $level) = @_;

	my $mytab = $MYTAB x $level;
	print $fh $mytab . "[HASH] (\n\n";

	foreach my $key (sort keys %{$hash}) { # the outermost foreach #

	    my $value = $hash->{$key};

	    print $fh $mytab . "Key= $key\,\n";
	    print $fh $mytab . $MYTAB . "Value= ";
	    unless (defined $value) {
		print $fh "'undef' \;\n";
		next;
	    }

	    my $ref = ref $value;
	    unless (defined $ref) {
		print $fh $value . " \;\n";
		next;
	    }

	    if ($ref eq 'SCALAR') {
		print $fh ${$value} . " \;\n";

	    } elsif ($ref eq 'ARRAY') {
		my $flag = pr_comp_array (@{$value}, *${$fh}, $level+1);
		unless ($flag) {
		    printf STDERR "'pr_comp_array' FAILED at the level %d.\n", $level +1;
		    return $flag;
		}
		print $fh " \;\n\n";
	    } elsif ($ref eq 'HASH') {
		my $flag = pr_comp_hash (%{$value}, *${$fh}, $level+1);
		unless ($flag) {
		    printf STDERR "'pr_comp_hash' FAILED at the level %d.\n", $level + 1;
		    return $flag;
		}
		print $fh " \;\n\n";
	    } else {
		print $fh  $value . " \;\n";
	    }

	}			# end of the outermost foreach #

	print $fh "\n";
	print $fh $mytab . ")";

	($level == 0) and (print $fh "\n\n");

	return 1;

}		# end of the subroutine 'pr_comp_hash' #


# pr_comp_array:
#	This subroutine can print out a composite array,
#	whose elements may be the references to scalars, arrays, and/or hashes.
#
# INPUT: (\@array, \*$fh, $level),
#	where $level increases as we go deeper into the composite structure.
#
# OUTPUT: 1 if succeeded in printing out the whole hash,
#	  0 if failed.
#
sub pr_comp_array (\@\*$) {

	my ($array, $fh, $level) = @_;

	my $mytab = $MYTAB x $level;
	print $fh $mytab . "[ARRAY] (\n";

	foreach my $value (@{$array}) { # the outermost foreach #

	    print $fh $mytab . $MYTAB . "ELEMENT= ";
	    unless (defined $value) {
		print $fh "'undef' \;\n";
		next;
	    }

	    my $ref = ref $value;
	    unless (defined $ref) {
		print $fh $value . " \;\n";
		next;
	    }

	    if ($ref eq 'SCALAR') {
		print $fh ${$value} . " \;\n";

	    } elsif ($ref eq 'ARRAY') {
		my $flag = pr_comp_array (@{$value}, *${$fh}, $level+1);
		unless ($flag) {
		    printf STDERR "'pr_comp_array' FAILED at the level %d.\n", $level +1;
		    return $flag;
		}
		print $fh " \;\n\n";

	    } elsif ($ref eq 'HASH') {
		my $flag = pr_comp_hash (%{$value}, *${$fh}, $level+1);
		unless ($flag) {
		    printf STDERR "'pr_comp_hash' FAILED at the level %d.\n", $level + 1;
		    return $flag;
		}
		print $fh " \;\n\n";

	    } else {
		print $fh $value . " \;\n";
	    }

	}			# end of the outermost foreach #

	print $fh $mytab . ")";

	($level == 0) and (print $fh "\n\n") ;

	return 1;

}		# end of the subroutine 'pr_comp_array' #



# pr_comp_array2:
#	This modified version of 'pr_comp_array' outputs
#	the index of each element as well,
#	in addition to the contents of the  element (as the predecessor does).
#	[CREATED on 2019/02/22.]
#
# pr_comp_array:
#	This subroutine can print out a composite array,
#	whose elements may be the references to scalars, arrays, and/or hashes.
#
# INPUT: (\@array, \*$fh, $level),
#	where $level increases as we go deeper into the composite structure.
#
# OUTPUT: 1 if succeeded in printing out the whole hash,
#	  0 if failed.
#
sub pr_comp_array2 (\@\*$) {

	my ($array, $fh, $level) = @_;

	my $mytab = $MYTAB x $level;
	print $fh $mytab . "[ARRAY] (\n";

	my $ct_elements = @{$array};

	for (my $k = 0; $k < $ct_elements; $k++) { # The outermost for-loop (over elements). # MODIFIED in creation of 'pr_comp_array2'.
	##foreach my $value (@{$array}) { # the outermost foreach #

	    my $value = $array->[$k]; # Added on 2019/02/22.

	    print $fh $mytab . $MYTAB . "The $k th ELEMENT= "; # MODIFIED on 2019/02/22.
	    #print $fh $mytab . $MYTAB . "ELEMENT= ";
	    unless (defined $value) {
		print $fh "'undef' \;\n";
		next;
	    }

	    my $ref = ref $value;
	    unless (defined $ref) {
		print $fh $value . " \;\n";
		next;
	    }

	    if ($ref eq 'SCALAR') {
		print $fh ${$value} . " \;\n";

	    } elsif ($ref eq 'ARRAY') {
		my $flag = pr_comp_array2 (@{$value}, *${$fh}, $level+1); # MODIFIED on 2019/02/22.
		##my $flag = pr_comp_array (@{$value}, *${$fh}, $level+1);
		unless ($flag) {
		    printf STDERR "'pr_comp_array2' FAILED at the level %d.\n", $level +1;
		    return $flag;
		}
		print $fh " \;\n\n";

	    } elsif ($ref eq 'HASH') {
		my $flag = pr_comp_hash (%{$value}, *${$fh}, $level+1);
		unless ($flag) {
		    printf STDERR "'pr_comp_hash' FAILED at the level %d.\n", $level + 1;
		    return $flag;
		}
		print $fh " \;\n\n";

	    } else {
		print $fh $value . " \;\n";
	    }

	##}			# end of the outermost foreach #
	} # End of the outermost for-loop (over elements). # MODIFIED in creation of 'pr_comp_array2'. #

	print $fh $mytab . ")";

	($level == 0) and (print $fh "\n\n") ;

	return 1;

}		# end of the subroutine 'pr_comp_array2' #


# cp_comp_array:
#	This subroutine makes a copy of a composite array
#	in a recurrent manner.
#
# INPUT: \@comp_array,
#
# OUTPUT: \@cp_comp_array.
#
sub cp_comp_array (\@) {

	my ($comp_array) = @_;

	my @cp_comp_array = ();

	foreach my $element (@{$comp_array}) { # The outer foreach-loop #

	    unless (defined $element) { # Undefined #
		push @cp_comp_array, undef;
		next;
	    }

	    my $ref = ref $element;

	    unless ($ref) {  # NOT a reference #
		push @cp_comp_array, $element;
		next;
	    }

	    if ($ref eq 'SCALAR') {
		my $cp_element = ${$element};
		push @cp_comp_array, \$cp_element;
	    } elsif ($ref eq 'ARRAY') {
		my $cp_element = cp_comp_array (@{$element});
		push @cp_comp_array, $cp_element;
	    } elsif ($ref eq 'HASH') {
		my $cp_hash = cp_comp_hash (%{$element});
		push @cp_comp_array, $cp_hash;
	    } else {
		print STDERR "[WARNING 'cp_comp_array'] Found an element ('$element') that are NEITHER an ordinary scalar NOT a reference to a scalar, array, or hash.\n";
		push @cp_comp_array, $element;
	    }

	}			# End of the outer foreach-loop #

	return \@cp_comp_array ;

}		# End of the subroutine 'cp_comp_array' #


# cp_comp_hash:
#	This subroutine makes a copy of a composite hash.
#
# INPUT: \%comp_hash.
#
# OUTPUT: \%cp_comp_hash.
#
sub cp_comp_hash (\%) {

	my ($comp_hash) = @_;

	my %cp_comp_hash ;

	foreach my $key (keys %{$comp_hash}) { # The outer foreach-loop #
	    my $value = $comp_hash->{$key};

	    unless (defined $value) { # Undefined 
		$cp_comp_hash{$key} = $value;
		next;
	    }

	    my $ref = ref $value;
	    unless ($ref) { # NOT a reference #
		$cp_comp_hash{$key} = $value;
		next;
	    }

	    if ($ref eq 'SCALAR') {
		my $cp_value = ${$value};
		$cp_comp_hash{$key} = \$cp_value;
	    } elsif ($ref eq 'ARRAY') {
		my $cp_value = cp_comp_array (@{$value});
		$cp_comp_hash{$key} = $cp_value;
	    } elsif ($ref eq 'HASH') {
		my $cp_value = cp_comp_hash (%{$value});
		$cp_comp_hash{$key} = $cp_value;
	    } else {
		print STDERR "[WARNING 'cp_comp_hash'] Found a value ('$value') that are NEITHER an ordinary scalar NOR a reference to a scalar, array, or hash.\n";
		$cp_comp_hash{$key} = $value;
	    } 
	}			# End of the outer foreach-loop #

	return \%cp_comp_hash;

}		# End of the subroutine 'cp_comp_hash' #



# merge_hash: a subroutine to merge two hashes, #
# both with similar lists of keys and values  #
sub merge_hash (\%\%) {

    local $| = 1;
    my ($equiv, $ne_value, $extra1, $extra2) = (0,0,0,0); # counters #
    my ($list1, $list2) = @_;
    my %merged;

    my @sortkey1 = sort {$a <=> $b} keys %$list1;
    my @sortkey2 = sort {$a <=> $b} keys %$list2;

   print "Now I'm going to merge the two hashes.\n\n";
   my ($i,$j) = (0,0);
   while ($i < @sortkey1 and $j < @sortkey2) {

	if($sortkey1[$i] == $sortkey2[$j]) {

	    if($$list1{$sortkey1[$i]} eq  $$list2{$sortkey2[$j]}) {

		print "an equivalent element was found:";
		print "$sortkey1[$i]:\t $$list1{$sortkey1[$i]}\n";
		$merged{$sortkey1[$i]} = $$list1{$sortkey1[$i]};
		$equiv++;

	    } else {

		$ne_value++;
		print "For key = $sortkey1[$i]:\t{ $$list1{$sortkey1[$i]} }\t";
		print "does not equal\t{ $$list2{$sortkey2[$j]} }\n";
		$merged{$sortkey1[$i]} = "$$list1{$sortkey1[$i]}\t:\t$$list2{$sortkey2[$j]}";

	    }
	    $i++;
	    $j++;

	} elsif ( $sortkey1[$i] < $sortkey2[$j] ) {

	    #print "an extra key was found in hash1: $sortkey1[$i]\n";
	    $merged{$sortkey1[$i]} = $$list1{$sortkey1[$i]};
	    $extra1++;
	    $i++;

	} else {

	    #print "an extra key was found in hash2: $sortkey2[$j]\n";
	    $merged{$sortkey2[$j]} = $$list2{$sortkey2[$j]};
	    $extra2++;
	    $j++;

	}		# end of  the if-elsif-else construction #

    } 			# end of the while-loop # 

    for (my $k = $i; $k < @sortkey1; $k++) {
	$merged{$sortkey1[$k]} = $$list1{$sortkey1[$k]};
	$extra1++;
    }
    #$extra1 += @sortkey1 - $i;

    for ( my $m = $j; $m < @sortkey2; $m++) {
	$merged{$sortkey2[$m]} = $$list2{$sortkey2[$m]};
	$extra2++;
    }
    #$extra2 += @sortkey2 - $j;

    print "\n\n\#(equivalent elements) = $equiv\n" .
	  "\#(pairs ofinequivalent values for the same keys) = $ne_value\n" .
	  "\#(keys in hash1 not in hash2) = $extra1\n" .
	  "\#(keys in hash2 not in hash1) = $extra2\n\n";

    return %merged;

}				# end of the subroutine merge_hash #

# diff_hash: a subroutine to recieve as arguments two hashes, #
# both of which have similar lists of keys and values,  #
# and returns two refecences to hashes,
# which are unoverlapping parts of the input hashes.
# It also prints out  a hash whose keys consists the intersection of #
# the two hash-key sets, and each of whose values is the differecne #
# between the corespoonding argument hash values.   #
# PREREQUIITE: values of the two argument hashes must be numbers.
sub diff_hash (\%\%) {

    local $| = 1;
    my ($common, $extra1, $extra2) = (0,0,0); # counters #
    my ($list1, $list2) = @_;
    my %diff;
    my %extra_keys1;
    my %extra_keys2;

   print "Now I'm going to compare the two hashes.\n\n";
   for ( keys %{$list1}) {				# the 1st outer for #

	my $base_number = ${$list1}{$_};
	if (defined ${$list2}{$_} ) {		# outer if-else #

	    $common++;
	    if ( ${$list2}{$_} =~ /\D/ ) {      # inner if-else #

		my @temp = split( /\D/, ${$list2}{$_} );
		my $subtractee = shift(@temp);
		my $diff_list = $subtractee - $base_number;
		for ( @temp) {				# inner for #
			(my $tmp = $_) =~  s/\s//g;
			($tmp =~ /^\d+$/) or next;
			$diff_list .= ':' . ($tmp - $base_number);
			#$diff_list .= ":$diff";
		}					# end of the inner for #
		$diff{$_} = $diff_list;

	    } else {					# middle of the inner if-else #
		$diff{$_} = ${$list2}{$_} - $base_number;
	    }					# end of the inner if-else #

	} else {					# middle of the outer if-else #

	    $extra_keys1{$_} = $base_number;
	    $extra1++;

	} 					# end of outer if-else #
    }							# end of the 1st outer for #

    for (keys %{$list2} ) { 			# the 2nd outer for #

	(defined ${$list1}{$_} ) and next;
	$extra_keys2{$_} = ${$list2}{$_};
	$extra2++;

    }						# end of the 2nd outer for #
	
    print "\n\n\#(common keys) = $common\n" .
	  "\#(keys in hash1 but not in hash2) = $extra1\n" .
	  "\#(keys in hash2 but not in hash1) = $extra2\n\n";

    print "\nList of overlapping hash keys with inequivalent values:\n";
    for (sort keys %diff) {
	$diff{$_} and  print "$_:\t$diff{$_}\n";
    }

    #print "Keys in hash1 but not in hash2:\n";
    #for (sort keys %extra_keys1) { print "$_:\t$extra_keys1{$_}\n"; }
    #print "\n\n";
    #print "Keys in hash2 but not in hash1:\n";
    #for (sort keys %extra_keys2) { print "$_:\t$extra_keys2{$_}\n"; }
    #print "\n\n";

    return (\%extra_keys1, \%extra_keys2, \%diff);

}		# end of the subroutine 'diff_hash' #


# construct_relayed_hash:
#	This subroutine receives more than 2 hashes,
#	then constructs a hash whose key and value are 'relay'ed by
#	the intermediate hashes.
#
# INPUT: (\@hashes, \@flags_alarm), where
#	@hashes is a set of hashes to be relayed,
#	by equating the value of %{$hashes[$k]} with the keys of %{$hashes[$k+1]}; 
#	$flags_alarm[$k] = 1 if you issue a warning when a value of %{$hashes[$k]}
#			is not found in the set of keys for %{$hashes[$k+1]},
#			= 0 otherwise.
#	We always require @hashes == @flags_alarm + 1.
#
# OUTPUT: \%relayed, where
#	%relayed is formed in the following manner:
#	a key of %{$hashes[0]} -> a value of %{$hashes[0]} = a key of %{$hashes[1]}
#	  -> ... -> a value of %{$hashes[$#hashes -1]} = a key of %{$hashes[$#hashes]}
#			-> a value of %{$hashes[$#hashes]}.
#
sub construct_relayed_hash (\@\@) {

	my ($hashes, $flags_alarm) = @_;

	my $ct_relays = @{$flags_alarm};
	my $ct_hashes = @{$hashes};
	unless ($ct_hashes == $ct_relays + 1) {
	    print STDERR "ILLEGAL INPUT for the subroutine 'construct_relayed_hash': (\#{hashes}, \#{alarms}) = ($ct_hashes, $ct_relays).\n";
	    return undef;
	}

	my $relayed = $hashes->[0];
		# (a key of %{$hashes->[0]} => a value of  %{$hashes->[$k]}, ...)
		# after the $k th round of relaying (or before the $k +1 th round).

	for (my $k=1; $k<$ct_hashes; $k++) { # the outer for-loop #
	    my %new_relayed; # (a key of %{$hashes->[0]} => a value of %{$hashes->[$k]}, ...)

	    foreach my $key (keys %{$relayed}) { # the inner for-loop #
		my $medium = $relayed->{$key};
		(defined $medium) or next;
		my $value = $hashes->[$k]{$medium};
		if ($flags_alarm->[$k-1] and !(defined $value) ) {
		    printf STDERR "[WARNING 'construct_relayed_hash'] The %s th input hash does NOT have a value corresponding to the key '$medium'.\n", $k+1;
		    next;
		}
		$new_relayed{$key} = $value;
	    }				# end of the inner for-loop #
	    $relayed = \%new_relayed;
	}			# end of the outer for-loop #

	return $relayed;

}		# end of the subroutine 'construct_relayed_hash' #


1;	# a "spell" which enables this package to be "use"-able and "require"-able #
