package MyTreeMap_indels_spt_odr2;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyTreeMap_indels_spt_odr2.pm
#	This package contains some subroutines that advance
#	the functions of some subroutines in 'MyTreeMap_indels_spt_odr.pm'.
#
#	[CREATION STARTED on Feb 4 (Sat), 2017, ENDED on Feb xx (www), 2017.]
#
#	[REVISED on 2019/07/14]:
#	Added a new subroutine, 'cmp_gblocks2',
#	which can handle arrays of information on the gap-blocks
#	in more general formats.
#
#	[REVISED on 2019/07/28]:
#	Add a yet new subroutine, 'cmp_gblocks3' by revising 'cmp_gblocks2'.
#	The new version should better adapt to the 'move's of the gap-blocks.
#
# MyTreeMap_indels_spt_odr.pm:
#	This modified version of the package 'MyTreeMap_indels.pm' searches for
#	the most parsimonious indel scenarios while respecting the spatial order
#	between the gap-blocks, even when they appear to be swappable with each other,
#	as long as each of them has 'presence' states in the (temporal) 'downstream' of the other.
#	(CREATION STARTED on Mar 18 (Tue), 2014, ENDED on Mar 20 (Thu), 2014)
#
#	REVISED on Mar 22 (Sat) & 23 (Sun), 2014:
#	Now, the subroutine, 'incorporate_null_block', was overhauled by introducing a new satellite subroutine,
#	'search_flanking_blocks_spt_odr2', and became much closer to being perfect.
#
#	REVISED on Mar 24 (Mon), 2014:
#	Now, the subroutine, 'recede_absence_spt_odr', was revised by incorporating the subroutine,
#	'search_flanking_blocks_spt_odr2', and became closer to being perfect.
#
# MyTreeMap_indels.pm:
#	This package makes a set of maps of the most parsimonious insertions/deletions
#	scenarios onto an input sequence tree.
#	The main algorithm is based on Farris's (1977) Dollo parsimony principle.
#	(CREATION STARTED on May 3 (Thu), 2012, ENDED on May ?? (???), 2012)
#
#	REVISED on Oct 24 (Wed), 2012:
#	To make the algorithm closer to perfection,
#	the subroutin "recede_absence" was modified so that it will utilize
#	flanking blocks in broad sense as well as those in strict sense.
#
#	REVISED on Nov 13 (Wed), 2013:
#	The module can now correctly handle trees containing two-branch nodes.
#
#	REVISED on Nov 15 (Fri), 2013:
#	An optional flag (global variable) was added to enable the subroutine 'enum_psm_indel_scns_in_cls_blk'
#	to ignore null columns when necessary.
#	[ This may be useful when we deal with a 'culled' MSA that resulted from removing some sequences
#	from an alignmentof a larger number of sequences reconstructed via an aligning program. ]
#
#	REVISESD on Mar 8 (Sat) & 9 (Sun), 2014:
#	A new subroutine, 'enum_psm_scns_gen_null_blk', was created.
#	It enables the subroutine 'enum_psm_indel_scns_in_cls_blks' to enumerate the parsimonious indel scenarios
#	that resulted in "NULL-cluster", which consists only of gaps.
#
#	REVISED on Mar 10 (Mon), 2014:
#	The package now introduces a global variable, $UPPER_BOUND_ANAL_GAPLEN.
#	If a gap-block cluster contains a run of gaps longer than this variable,
#	the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses of the cluster.
#	Accompanying the introduction of this global variable,
#	a new subroutine 'scan_gbcls_for_long_gaps' was created.
#
#	REVISED on Mar 11 (Tue), 2014:
#	TEMPORARILY Modified the subroutine 'incorporate_new_event' to TENTATIVELY cope with
#	the rare cases where the new 'deletion' event encounters BOTH left- and right-neighboring events.
#
#	(I WILL SERIOUSLY ADDRESS THIS ISSUE LATER. ... Probably, a solution would be to leave 'deletion' events unmarged when they are intervened by the 'presence' state in any 'downstream' nodes.)
#
#
#	REVISED on Mar 11 (Tue), 2014 (II):
#	The subroutine now introduces an upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS = 200,
#	for the cluster-length itself as well.
# 
# MyTreeMap_D.pm:
#	This package maps states and their changes
#	onto the nodes and the branches, respectively, of a phylogenetic tree,
#	according to Farris's (1977) Dollo parsimony principle.
#	In a nutshell, Farris's (1977) Dollo parsimony seeks for the evolutionary scenario
#	that minimizes the number of changes, UNDER THE CONSTRAINT that
#	the 'creation' of a state occurs at most once in each column.
#	(CREATION STARTED on Apr 19 (Thu), 2012, ENDED on Apr ?? (???), 2012)
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#	(REVISED on Nov 27 (Sun), 2011)
#	The subroutine 'min_penalties_sankoff_bo' now excludes from later consideration
#	the characters on each node that will NEVER be included in any parsiminious states.
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray union_nmarray ); # ADDED on 2019/02/05.

use MyTaxonomy qw(
	fetch_ext_offsprings_nr
		); # RESURRECTED on 2019/01/29.
	## fetch_ancestry 
	## fetch_last_common_ancestor_ov2
	## fetch_common_ancestors
	## fetch_all_offsprings_nr
	## fetch_ext_offsprings fetch_all_offsprings

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	## make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	); # RESURRECTED on after 'FAIL3' of 'test_partit_into_eff_indep_indel_blocks.pl', on March 21st (Tue), 2017. #
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

use MyTreeMap_indels_spt_odr qw(
	extract_cls_gblocks
	msa_rep_gpatterns
	scan_gbcls_for_long_gaps
	create_initial_set_indels_spt_odr
	);
	## enum_psm_indel_scns_for_msa


=set aside for later use (IMPORTED from 'MyTreeMap_indels_spt_odr.pm')

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );


use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);

=cut # END of "set aside for later use (IMPORTED from 'MyTreeMap_indels_spt_odr.pm')"


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	cal_init_psm_cands_for_sgl_msa
	cmp_gblocks
	cmp_gblocks2
	cmp_gblocks3
	br_list_gblocks
	mk_aux_datasets_on_gblocks
	br_parsimonize_dollo_indel_history
	bch_br_parsimonize_dollo_indel_histories
	); # (Added 'cmp_gblocks' on 2019/02/13.)
	   # (Added 'cmp_gblocks2' on 2019/07/14.)


sub cmp_gblocks (\@\@);
sub cmp_gblocks2 (\@\@$$); # ADDED on 2019/07/14. #
sub cmp_gblocks3 (\@\@$$$); # ADDED on 2019/07/28. #

sub br_list_gblocks (\%\@$\%);
sub mk_aux_datasets_on_gblocks ($\@);
sub br_parsimonize_dollo_indel_history ($\@\%\%$\%\%$$\%);

	# This specifies the level of tests on the subroutine, 
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

		# For the subroutine, 'cal_init_psm_cands_for_sgl_msa':
		# (IMPORTED from 'classify_msa_errors_via_mblks.2.pl' in creation of 'MyTreeMap_indels_spt_odr2.pm'.)
our $TEST_LEVEL_cipc2 = 0;
##our $TEST_LEVEL_cipc2 = 5;

		# For the subroutine, 'br_parsimonize_dollo_indel_history'. (ADDED on 2019/02/06.)
our $TEST_LEVEL_br_psm_did_hs = 0;
#our $TEST_LEVEL_br_psm_did_hs = 3;

		# For the subroutine, 'bch_br_parsimonize_dollo_indel_histories'. (ADDED on 2019/02/06.)
our $TEST_LEVEL_bch_br_psm_did_hs = 0;
#our $TEST_LEVEL_bch_br_psm_did_hs = 2;



	## IMPORTED with modifications from 'classify_msa_errors_via_mblks.2.pl', in creation of 'MyTreeMap_indels_spt_odr2.pm'. #

		# Upper-bound on the accepted length of a gap-block cluster (i.e., gapped segment) in the reference/reconstructed MSA.
##our $UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS ;
our $UPPER_BOUND_ANAL_LEN_GBCLS = 2000; # The 1st Measure to avoid unexpected immature termination of the program due to MSAs with long gapped segments.

		# Upper-bound on the accepted length of a run of gaps in the reference/reconstructed MSA.
##our $UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN ;
our $UPPER_BOUND_ANAL_GAPLEN = 2000; # The 2nd Measure to avoid unexpected immature termination of the program due to MSAs with long gapped segments.

	## END of "IMPORTED with modifications from 'classify_msa_errors_via_mblks.2.pl', in creation of 'MyTreeMap_indels_spt_odr2.pm'." #

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###


# cal_init_psm_cands_for_sgl_msa:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of a single input MSA.
#	It also outputs the (non-redundant) list of gap-patterns and the information on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSA.
#	It was created by modifying the subroutine 'cal_init_psm_cands_for_twin_msas'.
#	[CREATED on Oct 9 (Fri), 2015]
#	[ => IMPORTED from 'classify_msa_errors_via_mblks.2.pl' on Feb 4 (Sat), 2017]
#
# cal_init_psm_cands_for_twin_msas:
#	This subroutine calculates the initial candidate of the parsimonious scenario(s) for each gap-block cluster (i.e., gapped segment)
#	of the reference and reconstructed MSAs,
#	it also outputs the (non-redundant) list of gap-patterns and the informaion on their Dollo parsimonious scenarios.
#	NOTE that the subroutine eats up the input MSAs.
#	(CREATED on Oct 19 (Sun), 2014)
#
# INPUT: (\%node2ch, \%node2pa, \%node2depth, \%node2ct_exoffs, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres, \@info_gpatterns0), where
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#
#	%node2depth = ($node_id => $absolute_depth, ...) is for all the nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) is for all the nodes;
#
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA (reference/reconstructed)
#	(NOTE that these names must exactly match those used in %name2id);
#
#	${$in_msa$k]} is the $k th sequence embedded in the input MSA;
#
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
#	@info_gpatterns0 = (\@set_gpatterns0, \@set_node2dparsstat0, \@set_br2change0) gives information on the gap-patterns found in the reference & reconstructed MSAs, and the Dollo parsimonious scenarios yielding these gap-patterns, where
#	${$set_gpatterns0[$indx_gp]} is the string representing the $ndx_gp th gap-pattern in the input MSA, created by concatenating $CHAR_PRES/ABS in the sequences;
#	%{$set_node2dparsstat0[$indx_gp]} = ($node_id => [$state, $ct_pres_ext], ...) is for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern;
#	%{$set_br2change0[$indx_gp]} = ($branch_id => [$state_ue, $state_le], ...) for the Dollo parsimonious scenario yielding the $indx_gp th gap-pattern.
#
# OUTPUT: (\@info_init_psm_scn, \@info_gpatterns), where
#
#	@info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the reference/reconstructed MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$init_totcts_indels_for_cls[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	@info_gpatterns =  (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change) gives information on the gap-patterns found in the input MSA, and the Dollo parsimonious scenarios yielding these gap-patterns, (on top of those initially given,) where
#
#	@set_gpatterns follows the same format as @info_gpatterns0 in the input,
#	@set_node2dparsstat follows the same format as @set_node2dparsstat0 in the input,
#	@set_br2change follows the same format as @set_br2change0 in the input.
#
##sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$) { ....}
##	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa, $char_abs, $char_pres) = @_;
#
#		... MODIFIED on Oct 14 (Wed), 2015...
#
sub cal_init_psm_cands_for_sgl_msa (\%\%\%\%\%\@\@$$\@) {

	my ($node2ch, $node2pa, $node2depth, $node2ct_exoffs, $name2id, $seqnames, $in_msa, $char_abs, $char_pres, $info_gpatterns0) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};

    (print STDOUT "<< Subroutine 'cal_init_psm_cands_for_sgl_msa' (Top node ID: $topnode_id\; #{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL_cipc2>0);

		# Main Process 1: Extract gapped segments and their constituent gap-pattern blocks from the input MSA. #
	

			# MODIFIED on Oct 14 (Mon), 2015. #
	my ($set_gpatterns0, $set_node2dparsstat0, $set_br2change0) = @{$info_gpatterns0}; # Initial information of the gap-patterns and their Dollo parsimonious scenarios. #

	my %gpattern2indx;
	my $ct_gpatterns = my $ct_gpatterns0 = @{$set_gpatterns0};
	for (my $i=0; $i<$ct_gpatterns0; $i++) { $gpattern2indx{${$set_gpatterns0->[$i]}} = $i; }
	##my $ct_gpatterns = 0;

	my $set_cls_gblocks = extract_cls_gblocks (@{$in_msa}, $char_pres, $char_abs, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'extract_cls_gblocks' FAILED: $! !\n\n";
	    return (undef, undef);
	}
	my $ct_cls = @{$set_cls_gblocks};
    (print STDOUT "\n --- 'extract_cls_gblocks' Ended (#{gap clusters} = $ct_cls) ...\n\n") if ($TEST_LEVEL_cipc2>1);

	# substr (${$sec_msa[$k]}, $indx_gp, 1) is the presence/absence state of the $k th sequence in the $indx_gp th gap pattern.
	# ${$set_gpatterns[$indx_gp]} is the string made by concatenating the states in the $indx_gp th gap pattern.
	#
	my ($sec_msa, $set_gpatterns) = msa_rep_gpatterns (%gpattern2indx, length ($char_abs), $ct_seqs);
	unless (defined $sec_msa) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'msa_rep_gpatterns' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT "\n --- 'msa_rep_gpatterns' Ended ...\n\n") if ($TEST_LEVEL_cipc2>1);


		# Scan the information of gapped segments for screening those with gapped segment(s) and/or gap(s) that are TOO LONG. #

	foreach my $cls_gblk (@{$set_cls_gblocks}) { # Screening foreach-loop (over gapped segments in the input MSA). #

	    my $ct_gblks = @{$cls_gblk};
	    my $start_cls = $cls_gblk->[0][2];
	    my $end_cls = $cls_gblk->[$ct_gblks-1][3];
	    my $cl_len = $end_cls - $start_cls+1;

	    unless ($cl_len <= $UPPER_BOUND_ANAL_LEN_GBCLS) {
		print STDOUT "[WARNING 'cal_init_psm_cands_for_sgl_msa'] Found a cluster of length $cl_len, which is LONGER THAN the upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS... I will skip analyzing this input MSA: $! !\n\n";
		return ([], [], []);
	    }

	    my $list_long_gaps = scan_gbcls_for_long_gaps (@{$cls_gblk}, @{$set_gpatterns}, $UPPER_BOUND_ANAL_GAPLEN, $char_abs);
	    unless (defined $list_long_gaps) {
		print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'scan_gbcls_for_long_gaps' FAILED: $! !\n\n";
		return (undef, undef, undef);
	    }
	    unless (@{$list_long_gaps} == 0) {
		my $ct_long_gaps = @{$list_long_gaps};
		print STDOUT "[WARNING 'cal_init_psm_cands_for_sgl_msa'] Found $ct_long_gaps gaps longer than $UPPER_BOUND_ANAL_GAPLEN. So, I will skip analyzing this input MSA: $! !\n\n"; 
		return ([], [], []);
	    }

	}		# End of the screening foreach-loop (over gapped segments in reference & reconstructed MSAs). #


		## Main Process 2: Calculate the Dollo parsimonious indel scenario for each gap-pattern. ##

 	my ($set_node2dparsstat, $set_br2change) = ([], []); # ADDED on Oct 14 (Wed), 2015. #
    if ($ct_gpatterns>$ct_gpatterns0) { # ADDED on Oct 14 (Wed), 2015. #

		# MODIFIED on Oct 14 (Wed), 2015. #

	foreach my $sec_seq (@{$sec_msa}) { substr (${$sec_seq}, 0, $ct_gpatterns0 * length ($char_abs), ''); } # Remove the first $ct_gpatterns0 gap-patterns, whose Dollo parsimonious scenarios were already calculated.

	my ($set_node2dparsstat1, $set_br2change1) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	##my ($set_node2dparsstat, $set_br2change) = map_parsim_states_dollo_msa ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$sec_msa}, $char_abs, $char_pres, 1) ;
	unless (defined $set_node2dparsstat1) {
	    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'map_parsim_states_dollo_msa' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT " --- 'map_parsim_states_dollo_msa' Ended ...\n\n") if ($TEST_LEVEL_cipc2>1);

		# ADDED on Oct 14 (Wed), 2015. #

	@{$set_node2dparsstat} = (@{$set_node2dparsstat0}, @{$set_node2dparsstat1});
	@{$set_br2change} = (@{$set_br2change0}, @{$set_br2change1});

   } else {
		# In this case, the initial (input) set exhausts the gap-patterns in the MSA. #
		# (ADDED on Oct 14 (Wed), 2015.) #

	$set_node2dparsstat = $set_node2dparsstat0;
	$set_br2change = $set_br2change0;

   } # END of "if ($ct_gpatterns>$ct_gpatterns0) {...} else {...}" # ADDED on Oct 14 (Wed), 2015. #


	my @info_gpatterns = ($set_gpatterns, $set_node2dparsstat, $set_br2change);


		## Main Process 3: Calculate the initial candidate of the parsimonious indel scenario resulting in the gap-configurations of each gapped segment in the input MSA. ##

			# Create a set of all nodes.
	my @set_all_nodes = sort keys %{$node2pa};
	unshift @set_all_nodes, $topnode_id;

			# Initialize the output. #
	my @info_init_psm_scn = ($set_cls_gblocks, [], []);

		# @info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the input MSA gap configuration, where

	my $indx_indx_gp = 0; # Index to specify where the index of the gap-pattern is stored.

	foreach my $info_in_out (
		['input MSA', \@info_init_psm_scn]) { # Main outer foreach-loop (for the input MSA). #

	    my ($title_msa, $info_init_psm_scn) = @{$info_in_out};

	    my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};

	    my $ct_cls = @{$set_cls_gblocks};

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

		my $blocks_in_cls = $set_cls_gblocks->[$cl];

		my ($ct_indels0, $node2blwise_gstat0, $br2indels0, $list_blank_blks) = create_initial_set_indels_spt_odr ($topnode_id, %{$node2pa}, %{$node2ch}, @set_all_nodes, $char_pres, $char_abs, @{$blocks_in_cls}, $indx_indx_gp, @{$set_node2dparsstat}, @{$set_br2change});
		unless (defined $ct_indels0) {
		    print STDOUT "[ERROR 'cal_init_psm_cands_for_sgl_msa'] The subroutine 'create_initial_set_indels' FAILED on the $cl th cluster of '$title_msa': $! !\n\n";
		    return (undef, undef, undef);
		}
    (print STDOUT "\n --- Subroutine 'create_initial_set_indels_spt_odr' Ended (with #{initial indels} = $ct_indels0) ...\n\n") if ($TEST_LEVEL_cipc2>1);

		$init_totcts_indels_for_cls->[$cl] = $ct_indels0;
		$init_psm_scn_for_cls->[$cl] = [$br2indels0, $node2blwise_gstat0];

	# @{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 

	    }				# End of the main inner for-loop (over clusters (i.e., gapped segments) in each MSA. #

	}					# End of the main outer foreach-loop (for the input MSA). #

    (print STDOUT "<< END of the subroutine 'cal_init_psm_cand_for_sgl_msa' />>\n\n") if ($TEST_LEVEL_cipc2>0);

	return (\@info_init_psm_scn, \@info_gpatterns);

}	# End of the subroutine 'cal_init_psm_cands_for_sgl_msa'. #



# cmp_gblocks:
#	This satellite subroutine of 'br_list_gblocks'
#	just compares two gap-blocks.
#	[CREATED on 2019/01/29; REVISED on 2019/02/13.]
#
#	NOTE:
#	The output will be used for sorting a set of gap-blocks.
#	Whether the 1st block should come 'earlier than', 'simultaneously with', or 'later than' the 2nd one
#	will be determined according to the rule given at the NOTE of 'br_list_gblocks'.
#
# INPUT: (\@gblock1, \@gblock2), where
#
#	[ @gblock1 = ($br1, $up_or_lw1, $lm_gpb1, $rm_gpb1, \@invlvd_gpbs1, \@affected_exnodes1, $e1) gives information on the 1st gap-block. ] ... OBSOLETE as of 2019/02/13.
#	@gblock1 = ($br1, $up_or_lw1, $lm_pos1, $rm_pos1, \@invlvd_gpbs1, \@affected_exnodes1, $e1) gives information on the 1st gap-block.
#	$br1 is the ID of the branch delimiting the 1st gap-block,
#	$up_or_lw1 = 'U'/'L' if the 1st gap-block extends on the upper/lower side of the branch,
#	[ $lm/rm_gpb1 is the index of the left-most/right-most gap-pattern block, ] ... OBSOLETE as of 2019/02/13!!
#	$lm/rm_pos1 is the index of the left-most/right-most position (typically, gap-pattern-block or column),
#	@invlvd_gpbs1 lists the gap-pattern blocks involved,
#	@affected_exnodes1 lists the IDs of external nodes affected by the gap-block;
#	$e1 is the index of the bounding indel event in @{$br2indels{$br}}.
#	
#	The format of @gblock2 is identical to that of @gblock1, except that the suffix '1' is replaced with '2'.
#
# OUTPUT: $sgn, where
#	$sgn = -1 if @gblock1 should come earlier than @gblock2,
#	     = +1 if @gblock1 should come later than @gblock2,
#	     = 0  if they should come simultaneously.
#
sub cmp_gblocks (\@\@) {

	my ($gblock1, $gblock2) = @_;

		# REVISED on 2019/02/13. #
	my ($br1, $up_or_lw1, $lm_pos1, $rm_pos1, $invlvd_gpbs1, $aff_exnodes1, $e1) = @{$gblock1};
	my ($br2, $up_or_lw2, $lm_pos2, $rm_pos2, $invlvd_gpbs2, $aff_exnodes2, $e2) = @{$gblock2};
	# my ($br1, $up_or_lw1, $lm_gpb1, $rm_gpb1, $invlvd_gpbs1, $aff_exnodes1, $e1) = @{$gblock1};
	# my ($br2, $up_or_lw2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2, $aff_exnodes2, $e2) = @{$gblock2};

		# (1) Compare the number of external nodes affected. #

	my $ct_aff1 = @{$aff_exnodes1};
	my $ct_aff2 = @{$aff_exnodes2};
	if ($ct_aff1 > $ct_aff2) {
	     return -1;
	} elsif ($ct_aff1 < $ct_aff2) {
	     return +1;
	}

		# (2) Compare the IDs of the external nodes affected. #

	for (my $k =0; $k < $ct_aff1; $k++) {
	     my $aff1 = $aff_exnodes1->[$k];
	     my $aff2 = $aff_exnodes2->[$k];
	     if ($aff1 < $aff2) {
		return -1;
	     } elsif ($aff1 > $aff2) {
		return +1;
	     }
	}

		# (3) Compare the left-ends of the gap-blocks. #

	if ($lm_pos1 < $lm_pos2) { # REVISED on 2019/02/13.
	#if ($lm_gpb1 < $lm_gpb2) {
	     return -1;
	} elsif ($lm_pos1 > $lm_pos2) { # REVISED on 2019/02/13.
	#} elsif ($lm_gpb1 > $lm_gpb2) {
	     return +1;
	} else {
	      return 0;
	}

}	# END of the subroutine, 'cmp_gblocks'. #


# cmp_gblocks2:
#	This modified version of 'cmp_gblocks' can handle
#	arrays of information on the gap-blocks
#	in more general formats.
#	[CREATED on 2019/07/14.]
#
# cmp_gblocks:
#	This satellite subroutine of 'br_list_gblocks'
#	just compares two gap-blocks.
#	[CREATED on 2019/01/29; REVISED on 2019/02/13.]
#
#	NOTE:
#	The output will be used for sorting a set of gap-blocks.
#	Whether the 1st block should come 'earlier than', 'simultaneously with', or 'later than' the 2nd one
#	will be determined according to the rule given at the NOTE of 'br_list_gblocks'.
#
#
# INPUT: (\@gblock1, \@gblock2, $indx_lm_pos, $indx_affd_exnodes), where
#
#	@gblock1 gives information on the 1st gap-block;
#	More precisely,
#	@gblock1[$indx_lm_pos, $indx_affd_exnodes] = ($lm_pos1, \@affected_exnodes1), where
#	$lm_pos1 is the index of the left-most position (typically, gap-pattern-block or column),
#	@affected_exnodes1 lists the IDs of external nodes affected by the gap-block;
#	
#	The format of @gblock2 is identical to that of @gblock1, except that the suffix '1' is replaced with '2';
#
# OUTPUT: $sgn, where
#	$sgn = -1 if @gblock1 should come earlier than @gblock2,
#	     = +1 if @gblock1 should come later than @gblock2,
#	     = 0  if they should come simultaneously.
#
sub cmp_gblocks2 (\@\@$$) {

	my ($gblock1, $gblock2, $indx_lm_pos, $indx_affd_exnodes) = @_;

		# REVISED in creation of 'cmp_gblocks2' (on 2019/07/14). #
	my ($lm_pos1, $affd_exnodes1) = @{$gblock1}[$indx_lm_pos, $indx_affd_exnodes];
	my ($lm_pos2, $affd_exnodes2) = @{$gblock2}[$indx_lm_pos, $indx_affd_exnodes];
	##my ($br1, $up_or_lw1, $lm_pos1, $rm_pos1, $invlvd_gpbs1, $aff_exnodes1, $e1) = @{$gblock1};
	##my ($br2, $up_or_lw2, $lm_pos2, $rm_pos2, $invlvd_gpbs2, $aff_exnodes2, $e2) = @{$gblock2};

		# (1) Compare the number of external nodes affected. #

	my $ct_affd1 = @{$affd_exnodes1};
	my $ct_affd2 = @{$affd_exnodes2};
	if ($ct_affd1 > $ct_affd2) {
	     return -1;
	} elsif ($ct_affd1 < $ct_affd2) {
	     return +1;
	}

		# (2) Compare the IDs of the external nodes affected. #

	for (my $k =0; $k < $ct_affd1; $k++) {
	     my $affd1 = $affd_exnodes1->[$k];
	     my $affd2 = $affd_exnodes2->[$k];
	     if ($affd1 < $affd2) {
		return -1;
	     } elsif ($affd1 > $affd2) {
		return +1;
	     }
	}

		# (3) Compare the left-ends of the gap-blocks. #

	if ($lm_pos1 < $lm_pos2) {
	     return -1;
	} elsif ($lm_pos1 > $lm_pos2) {
	     return +1;
	} else {
	      return 0;
	}

}	# END of the subroutine, 'cmp_gblocks2'. #


# cmp_gblocks3:
#	This modified version of 'cmp_gblocks2' can output
#	the ordering result that should BETTER FIT
#	the current rules for the RANKing of the Gap-Blocks.
#	[CREATED on 2019/07/28.]
#
# cmp_gblocks2:
#	This modified version of 'cmp_gblocks' can handle
#	arrays of information on the gap-blocks
#	in more general formats.
#	[CREATED on 2019/07/14.]
#
# cmp_gblocks:
#	This satellite subroutine of 'br_list_gblocks'
#	just compares two gap-blocks.
#	[CREATED on 2019/01/29; REVISED on 2019/02/13.]
#
#	NOTE:
#	The output will be used for sorting a set of gap-blocks.
#	Whether the 1st block should come 'earlier than', 'simultaneously with', or 'later than' the 2nd one
#	will be determined according to the rule given at the NOTE of 'br_list_gblocks'.
#
#
# INPUT: (\@gblock1, \@gblock2, $indx_lm_pos, $indx_rm_pos, $indx_affd_exnodes), where
#
#	@gblock1 gives information on the 1st gap-block;
#	More precisely,
#	@gblock1[$indx_lm_pos, $indx_rm_pos, $indx_affd_exnodes] = ($lm_pos1, $rm_pos1, \@affected_exnodes1), where
#	$lm_pos1 is the index of the left-most position (typically, gap-pattern-block or column),
#	$rm_pos1 is the index of the right-most position (typically, gap-pattern-block or column),
#	@affected_exnodes1 lists the IDs of external nodes affected by the gap-block;
#	
#	The format of @gblock2 is identical to that of @gblock1, except that the suffix '1' is replaced with '2';
#
# OUTPUT: $sgn, where
#	$sgn = -1 if @gblock1 should come earlier than @gblock2,
#	     = +1 if @gblock1 should come later than @gblock2,
#	     = 0  if they should come simultaneously.
#
sub cmp_gblocks3 (\@\@$$$) {

	my ($gblock1, $gblock2, $indx_lm_pos, $indx_rm_pos, $indx_affd_exnodes) = @_;

		# REVISED in creation of 'cmp_gblocks2' (on 2019/07/14). #
		# => Further REVISED in creation of 'cmp_gblocks3' (on 2019/07/28). #
	my ($lm_pos1, $rm_pos1, $affd_exnodes1) = @{$gblock1}[$indx_lm_pos, $indx_rm_pos, $indx_affd_exnodes];
	my ($lm_pos2, $rm_pos2, $affd_exnodes2) = @{$gblock2}[$indx_lm_pos, $indx_rm_pos, $indx_affd_exnodes];
	##my ($lm_pos1, $affd_exnodes1) = @{$gblock1}[$indx_lm_pos, $indx_affd_exnodes];
	##my ($lm_pos2, $affd_exnodes2) = @{$gblock2}[$indx_lm_pos, $indx_affd_exnodes];
	####my ($br1, $up_or_lw1, $lm_pos1, $rm_pos1, $invlvd_gpbs1, $aff_exnodes1, $e1) = @{$gblock1};
	####my ($br2, $up_or_lw2, $lm_pos2, $rm_pos2, $invlvd_gpbs2, $aff_exnodes2, $e2) = @{$gblock2};

		# (1) Compare the number of external nodes affected. #

	my $ct_affd1 = @{$affd_exnodes1};
	my $ct_affd2 = @{$affd_exnodes2};
	if ($ct_affd1 > $ct_affd2) {
	     return -1;
	} elsif ($ct_affd1 < $ct_affd2) {
	     return +1;
	}

		# (2) Compare the IDs of the external nodes affected. #

	for (my $k =0; $k < $ct_affd1; $k++) {
	     my $affd1 = $affd_exnodes1->[$k];
	     my $affd2 = $affd_exnodes2->[$k];
	     if ($affd1 < $affd2) {
		return -1;
	     } elsif ($affd1 > $affd2) {
		return +1;
	     }
	}

		# (3) Consider some rare cases, where the gap-blocks are nested. #
		# (ADDED in creation of 'cmp_gblocks3' (on 2019/07/28).) #
	if (($lm_pos1 == $lm_pos2) and ($rm_pos1 == $rm_pos2)) {

	    return 0; # In this case, the two gap-blocks are regarded as identical. #

	} elsif (($lm_pos1 <= $lm_pos2) and ($rm_pos2 <= $rm_pos1)) {

		# $gblock1 horizontally includes $gblock2. #
	    return +1;

	} elsif (($lm_pos2 <= $lm_pos1) and ($rm_pos1 <= $rm_pos2)) {

		# $gblock2 horizontally includes $gblock1. #
	    return -1;
	}

		# END of "(ADDED in creation of 'cmp_gblocks3' (on 2019/07/28).)" #

		# (5) Compare the left-ends of the gap-blocks. #

	if ($lm_pos1 < $lm_pos2) {
	     return -1;
	} elsif ($lm_pos1 > $lm_pos2) {
	     return +1;
	} else {

		# I hope this case will NEVER occur;
		# In case it does, however, set a rule for ordering.

		# (REVISED in creation of 'cmp_gblocks3' (on 2019/07/28).) #
	    if ($rm_pos1 < $rm_pos2) {
		return -1;
	    } elsif ($rm_pos1 > $rm_pos2) {
		return +1;
	    } else {
		return 0;
	    }
	    ##return 0;
		# END of "(REVISED in creation of 'cmp_gblocks3' (on 2019/07/28).)" #
	}

}	# END of the subroutine, 'cmp_gblocks3'. #


# br_list_gblocks:
#	This satellite subroutine of 'br_parsimonize_history' 
#	receives an input indel history in the form of %br2indels,
#	and returns a brief list of gap-blocks defined by the indels in the input.
#	[CREATED on 2019/01/29]
#
#	NOTE:
#	The output gap-blocks are sorted in the following order:
#	(1) In descending order of the number of affected external nodes; if some blocks tie,
#	(2) In ascending order of the IDs of affected external nodes; if some blocks still tie,
#	(3) In ascending order of the block's horizontal position (especially the left-end).
#
# INPUT: (\%br2indels, \@set_exnodes, $char_abs, \%node2ch), where
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	@set_exnodes lists the IDs of ALL external nodes;
#
#	$char_abs is the character specifying the absence of the residue at the site (Usually, it's the same as the 'gap' character);
#
#	%node2ch = ($node_ID => \@children, ...) for all internal nodes;
#
# OUTPUT: \@br_list_gblocks, where
#
#	@{$br_list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$lm/rm_gpb is the index of the left-most/right-most gap-pattern block,
#	[ $start/$end is the start/end coordinate of the gap-block, ]
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block;
#	$e is the index of the bounding indel event in @{$br2indels{$br}}.
#
sub br_list_gblocks (\%\@$\%) {

	my ($br2indels, $set_exnodes, $char_abs, $node2ch) = @_;

	my @prelist_gblocks = (); # element = [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e];

	foreach my $br (keys %{$br2indels}) { # Main foreach-loop (over branches where indels occur). #

	    my $indels = $br2indels->{$br};
	    unless (defined $indels) { next; }
	    my $ct_indels  = @{$indels};
	    if ($ct_indels == 0) { next; }

	    my $ex_offsprings = fetch_ext_offsprings_nr ($br, %{$node2ch});

	    for (my $e = 0; $e < $ct_indels; $e++) { # Inner foreach-loop (over indels along $br). #

		my $indel = $indels->[$e];
		my ($state_ue, $state_le, $invlvd_gpbs) = @{$indel};
		(defined $state_ue) or next; # Skip a null event. #
		my $up_or_lw = ($state_ue eq $char_abs) ? 'U' : 'L' ;
		my ($lm_gpb, $rm_gpb) = @{$invlvd_gpbs}[0, $#{$invlvd_gpbs}];

		my $affected_exnodes;
		if  ($up_or_lw eq 'L') {
		    my @std_ex_offsprings = sort {$a <=> $b} @{$ex_offsprings};
		    $affected_exnodes = \@std_ex_offsprings;
		} else { # if ($up_or_lw eq 'U')
		    (my $common, $affected_exnodes, my $dummy) = diff_nmarray (@{$set_exnodes}, @{$ex_offsprings});
		}

		push @prelist_gblocks, [$br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affected_exnodes, $e];

	    }	# End of the inner foreach-loop (over indels along $br). #

	}  # End of the main foreach-loop (over branches where indels occur). #


		# Sort the gap-blocks. #

	my @br_list_gblocks = sort { cmp_gblocks (@{$a}, @{$b}) } @prelist_gblocks;

	return \@br_list_gblocks;

} 	# END of the subroutine 'br_list_gblocks'. #


# mk_aux_datasets_on_gblocks:
#	This satellite subroutine of 'br_parsimonize_dollo_indel_history' creates
#	some auxiliary data-sets that will be useful for the brief-parsimonization of the indel history.
#	[CREATED on 2019/01/31.]
#
# INPUT: ($ct_gpbs_in_cls, \@br_list_gblocks), where
#
#	$ct_gpbs_in_cls is the number of gap-pattern blocks in this cluster (gapped segment);
#
#	@{$br_list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$lm/rm_gpb is the index of the left-most/right-most gap-pattern block,
#	[ $start/$end is the start/end coordinate of the gap-block, ]
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block;
#	$e is the index of the bounding indel event in @{$br2indels{$br}}.
#
# OUTPUT: (\%br2gblocks, \@gblocks_w_spec_lm, \@gblocks_w_spec_rm), where
#
#	$br2gblocks{$br}->[$e] gives the index of the gap-block (in @br_list_gblocks) that is defined by the $e th indel event along branch $br;
#
#	@{$gblocks_w_spec_lm[$lm_gpb]} lists the indices of the gap-blocks (in @br_list_gblocks) whose left-most gap-pattern block is the $lm_gpb th one;
#
#	@{$gblocks_w_spec_rm[$rm_gpb]} lists the indices of the gap-blocks (in @br_list_gblocks) whose right-most gap-pattern block is the $rm_gpb th one;
#	
sub mk_aux_datasets_on_gblocks ($\@) {

	my ($ct_gpbs_in_cls, $br_list_gblocks) = @_;

	my %br2gblocks;
	my @gblocks_w_spec_lm = my @gblocks_w_spec_rm = ();
	for (1 .. $ct_gpbs_in_cls) { # Preliminary for-loop. #
	     push @gblocks_w_spec_lm, [];
	     push @gblocks_w_spec_rm, [];
	}		# End of the preliminary for-loop. #

	my $ct_gblks = @{$br_list_gblocks};

	for (my $gb =0; $gb < $ct_gblks; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affected_exnodes, $e) = @{$br_list_gblocks->[$gb]} ;

	     my $gblocks_along_br = $br2gblocks{$br};
	     unless (defined $gblocks_along_br) { $gblocks_along_br = $br2gblocks{$br} = []; }

	    $gblocks_along_br->[$e] = $gb;

	    push @{$gblocks_w_spec_lm[$lm_gpb]}, $gb;
	    push @{$gblocks_w_spec_rm[$rm_gpb]}, $gb;

	}	# End of the main for-loop (over gap-blocks). #

	return (\%br2gblocks, \@gblocks_w_spec_lm, \@gblocks_w_spec_rm);

}	# End of the subroutine 'mk_aux_datasets_on_gblocks'. #


# br_parsimonize_dollo_indel_history:
#	This subroutine receives an indel history based on the Dollo parsimony,
#	and briefly attempts to modify it into a parsimonious history,
#	by 'vertically splitting' a 'parent' gap-block sandwiched by a pair of 'sibling' gap-blocks.
#	[CREATION STARTED on 2019/01/29, ENDED on 2019/02/02; MODIFIED on 2019/02/05, 2019/02/06 (added testing outputs & debugged).]
#
# INPUT: ($ct_gpblocks_in_cls, \@set_exnodes, \%br2indels, \%node2blwise_gstat, $topnode_id, \%node2pa, \%node2ch, $char_abs, $char_pres, \%br2eq), where
#
#	[ @{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block in the subject gapped segment, with
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block (in the 0-based full-closed convention); ] ... OBSOLETE as of 2019/01/31.
#
#	$ct_gpbs_in_cls is the number of gap-pattern blocks in this cluster (gapped segment);
#
#	@set_exnodes stores the IDs of all external nodes;
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes;
#	$char_abs is the character specifying the absence of the residue at the site (Usually, it's the same as the 'gap' character);
#	$char_pres is the character specifying the presence of the residue at the site (Usually, it's the same as the 'gap' character);
#	%br2eq = ($br_id => $equivalent_br_id, ...) for branches with equivalents;
#	[ %br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...). ] ... OBSOLETE as of 2019/01/31.
#
# OUTPUT: (\%new_br2indels, \%new_node2blwise_gstat, \@new_list_gblocks, \@places_gpbs, \@gpbs_at_spec_places, $ct_parsimonized), where
#
#	%new_br2indels gives the updated list of indels, in the same format as %br2indels;
#	%new_node2blwise_gstat gives the updated list of ancestral gap-states, in the same format as %node2blwise_gstat;
#	@new_list_gblocks is the updated list of gap-blocks, where
#	@{$new_list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$lm/rm_gpb is the index of the left-most/right-most gap-pattern block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@affected_exnodes lists the IDs of external nodes affected by the gap-block;
#	$e is the index of the bounding indel event in @{$br2indels{$br}}.
#
#	$places_gpbs[$gpb] tells in what place the $gpb th gap-pattern block resides when this subroutine ends;
#	$gpbs_at_spec_places[$pl] is the index of the gap-pattern block that is in the $pl th place when this subroutine ends;
#
#	$ct_parsimonized is the number of "parsimonized" indels, i.e., the number of gap-blocks that have been vertically split and merged with their "child" gap-blocks.
#
#	NOTE1: @places_gpbs and @gpbs_at_spec_places will be used for updating some data,
#	for example, the alignment itself of the gapped segment in question,
#	the set of columns in the segment, and the set of column-wise probabilities.
#
#	NOTE2: Before this subroutine ends, the gap-pattern blocks will be re-ordered in %new_br2indels, %new_node2blwise_gstat, and @new_list_gblocks.
#
sub br_parsimonize_dollo_indel_history ($\@\%\%$\%\%$$\%) {

	my ($ct_gpbs_in_cls, $set_exnodes, $br2indels, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs, $char_pres, $br2eq) = @_;

	my $ct_exnodes = @{$set_exnodes}; # MOVED from the start of the Main process (on 2019/02/06). #

    (print STDOUT "\n", ' ' x 3, "< Subroutine 'br_parsimonize_dollo_indel_history' (#{gap-pattern-blocks} = $ct_gpbs_in_cls, #{external nodes} = $ct_exnodes, topnode_ID = $topnode_id) >\n\n") if ($TEST_LEVEL_br_psm_did_hs>0);

	my $br_list_gblocks = br_list_gblocks (%{$br2indels}, @{$set_exnodes}, $char_abs, %{$node2ch}) ;
	unless (defined $br_list_gblocks) {
	      print STDERR "[ERROR 'br_parsimonize_dollo_indel_history'] The subroutine 'br_list_gblocks' FAILED: $! !\n";
	      return (undef, undef, undef, undef, undef, undef);
	}
		# @{$br_list_gblocks[$gb]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs, \@affected_exnodes, $e) gives information on the $gb th gap-block.

	my $ct_gblocks = @{$br_list_gblocks}; # MOVED from the start of the Main process (on 2019/02/06). #

    if ($TEST_LEVEL_br_psm_did_hs >2) {

	print STDOUT "\n", ' ' x 3, "[ Initial list of gap-blocks (#{gap-blocks} = $ct_gblocks) ]\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Indx_gb', 'br_ID', 'up(U)_or_low(L)', 'lm_gpb', 'rm_gpb', '(involved,gap,pattern,blocks)', '[affected,external,nodes]', 'within_branch_indx'), "\n\n";

	for (my $gb = 0; $gb < $ct_gblocks; $gb++) {  # 1st test for-loop (over gap-blocks). #

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affected_exnodes, $e) = @{$br_list_gblocks->[$gb]};

	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
	    $cnct_invlvd_gpbs = "($cnct_invlvd_gpbs)";
	    my $cnct_affected_exnodes = join (',', @{$affected_exnodes});
	    $cnct_affected_exnodes = "[$cnct_affected_exnodes]";

	    print STDOUT ' ' x 3, join ("\t", $gb, $br, $up_or_lw, $lm_gpb, $rm_gpb, $cnct_invlvd_gpbs, $cnct_affected_exnodes), "\n";

	} # End of the 1st test for-loop (over gap-blocks). #

	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"


		# Create some auxiliary data-sets that will be useful for the brief-parsimonization. #

	my ($br2gblocks, $gblocks_w_spec_lm, $gblocks_w_spec_rm) = mk_aux_datasets_on_gblocks ($ct_gpbs_in_cls, @{$br_list_gblocks});
	unless (defined $gblocks_w_spec_rm) {
	      print STDERR "[ERROR 'br_parsimonize_dollo_indel_history'] The subroutine 'mk_aux_datasets_on_gblocks' FAILED: $! !\n";
	      return (undef, undef, undef, undef, undef, undef);
	}

		# $br2gblocks{$br}->[$e] gives the index of the gap-block (in @br_list_gblocks) that is defined by the $e th indel event along branch $br;
		#
		# @{$gblocks_w_spec_lm[$lm_gpb]} lists the indices of the gap-blocks (in @br_list_gblocks) whose left-most gap-pattern block is the $lm_gpb th one;
		#
		# @{$gblocks_w_spec_rm[$rm_gpb]} lists the indices of the gap-blocks (in @br_list_gblocks) whose right-most gap-pattern block is the $rm_gpb th one;

    if ($TEST_LEVEL_br_psm_did_hs >2) {

	print STDOUT "\n", ' ' x 3, "[ Outputs of 'mk_aux_datasets_on_gblocks' ]\n\n\n";

	print STDOUT ' ' x 3, '%{$br2gblocks} = ( # $br_ID => [gap, blocks]', "\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2gblocks}) { # 2nd test foreach-loop (over branches).
		print STDOUT ' ' x 3, "\t\t$br", ' => [', join (', ',  @{$br2gblocks->{$br}}), "],\n";
	}	# End of the 2nd test foreach-loop (over branches).
	print STDOUT "\n\n";

	print STDOUT ' ' x 3, '(Contents of @{$gblocks_w_spec_lm})', "\n\n";
	print STDOUT ' ' x 3, join ("\t", 'lm_gpb', '[gap, blocks]'), "\n\n";
	for (my $lm_gpb = 0; $lm_gpb < $ct_gpbs_in_cls; $lm_gpb++) { # 3rd test for-loop (over left-most gap-blocks).
	    my $cnct_gblocks = join (', ', @{$gblocks_w_spec_lm->[$lm_gpb]});
	    print STDOUT ' ' x 3, join ("\t", $lm_gpb, "[$cnct_gblocks]"), "\n";

	}	# End of the 3rd test for-loop (over left-most gap-blocks). 
	print STDOUT "\n\n";

	print STDOUT ' ' x 3, '(Contents of @{$gblocks_w_spec_rm})', "\n\n";
	print STDOUT ' ' x 3, join ("\t", 'rm_gpb', '[gap, blocks]'), "\n\n";
	for (my $rm_gpb = 0; $rm_gpb < $ct_gpbs_in_cls; $rm_gpb++) { # 5th test for-loop (over right-most gap-blocks).
	    my $cnct_gblocks = join (', ', @{$gblocks_w_spec_rm->[$rm_gpb]});
	    print STDOUT ' ' x 3, join ("\t", $rm_gpb, "[$cnct_gblocks]"), "\n";

	}	# End of the 5th test for-loop (over right-most gap-blocks). 
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"


		# Initialize the output data-sets. #

	my @places_gpbs = my @gpbs_at_spec_places = (0 .. $ct_gpbs_in_cls -1);

	my %new_br2indels;
	foreach my $br (keys %{$br2indels}) { # 1st preliminary foreach-loop (over branches with indels). #
	    my $list_indels = $br2indels->{$br};
	    unless (defined $list_indels) { next; }
	    my @cp_list_indels = ();
	    foreach my $indel (@{$list_indels}) { # Inner preliminary foreach-loop (over indels along $br).
		my ($state_ue, $state_le, $invlvd_gpbs) = @{$indel};
		my @cp_invlvd_gpbs = @{$invlvd_gpbs};
		push @cp_list_indels, [$state_ue, $state_le, \@cp_invlvd_gpbs];
	    } # End of the inner preliminary foreach-loop (over indels along $br).
	    $new_br2indels{$br} = \@cp_list_indels;
	}	# End of the 1st preliminary foreach-loop (over branches with indels). #


	my %new_node2blwise_gstat;
	foreach my $node (keys %{$node2blwise_gstat}) { # 2nd preliminary foreach-loop (over all nodes). #
	    my @cp_blwise_gstat = @{$node2blwise_gstat->{$node}};
	    $new_node2blwise_gstat{$node} = \@cp_blwise_gstat;
	}	# End of the 2nd preliminary foreach-loop (over all nodes). #


		# Main Process. #

    (print STDOUT "\n", ' ' x 3, "[[ Main Process ]]\n\n") if ($TEST_LEVEL_br_psm_did_hs>1);

	# my $ct_exnodes = @{$set_exnodes}; # MOVED to just below the start (on 2019/02/06). #
	# my $ct_gblocks = @{$br_list_gblocks}; # MOVED to just below 'br_list_gblocks' (on 2019/02/06). #
	my $ct_parsimonized = 0;

	for (my $gb = 0; $gb < $ct_gblocks; $gb++) { # Main for-loop (over gap-blocks). #
	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affected_exnodes, $e) = @{$br_list_gblocks->[$gb]} ;
		# $up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,

    if ($TEST_LEVEL_br_psm_did_hs>2) {
	print STDOUT "\n", ' ' x 3, "[ The $gb th gap-block (br_ID=$br, up_or_low=$up_or_lw, lm_gpb=$lm_gpb, rm_gpb=$rm_gpb, {involved gpbs} = (", join (',', @{$invlvd_gpbs}), "), {affected exnodes}=[", join (',', @{$affected_exnodes}), "], within_branch_indx=$e) ]\n\n";
    } elsif ($TEST_LEVEL_br_psm_did_hs>1) {
	print STDOUT "\n", ' ' x 3, "[ The $gb th gap-block (br_ID=$br, up_or_low=$up_or_lw, lm_gpb=$lm_gpb, rm_gpb=$rm_gpb) ]\n\n";
    }

	    my ($ch1, $up_or_lw1, $ch2, $up_or_lw2);

	    if ($up_or_lw eq 'L') {
		my $children = $node2ch->{$br};
		if ((defined $children) and (@{$children} ==2)) {
		    ($ch1, $ch2) = @{$children};
		    $up_or_lw1 = $up_or_lw2 = 'L';
		}
	    } else { # if ($up_or_lw eq 'U')
		my $pa = $node2pa->{$br};
		my $sibs = $node2ch->{$pa};
		if ($pa == $topnode_id) {
		    my $eq_br = $br2eq->{$br};
		    if (defined $eq_br) {
			my $children_eqbr = $node2ch->{$eq_br};
			if ((defined $children_eqbr) and (@{$children_eqbr} == 2)) {
			    ($ch1, $ch2) = @{$children_eqbr};
			    $up_or_lw1 = $up_or_lw2 = 'L';
			}
		    } elsif (@{$sibs} ==3) {
			foreach my $sib (@{$sibs}) {
			    if ($sib == $br) { next; }
			    if (defined $ch1) {
				$ch2 = $sib;
				last;
			    } else {
				$ch1 = $sib;
			    }
			}
			$up_or_lw1 = $up_or_lw2 = 'L';
		    }
		} elsif (@{$sibs} ==2) {

			# A BUG fixed on 2019/02/05. => UnDone on the same day!! #
		    #my $eq_pa = $br2eq->{$pa};
		    #if (defined $eq_pa) {
		    #	$ch1 = $eq_pa;
		    #	$up_or_lw1 = 'L';
		    #} else {
		    #	$ch1 = $pa;
		    #	$up_or_lw1 = 'U';
		    #}
		    $ch1 = $pa;
		    $up_or_lw1 = 'U';
			# END of "A BUG fixed on 2019/02/05. => UnDone on the same day!!" #

		    foreach my $sib (@{$sibs}) {
			if ($sib == $br) { next; }
			$ch2 = $sib;
			last;
		    }
		    $up_or_lw2 = 'L';
		}
	    } # End of "if ($up_or_lw eq 'L') {...} else {...}"

	    unless ((defined $ch1) and (defined $ch2)) { next; }

    if ($TEST_LEVEL_br_psm_did_hs >2) {
	print STDOUT ' ' x 3, "/ Passed the 1st screen /\n\n";
	print STDOUT ' ' x 3, '($ch1, $up_or_lw1) = ', "($ch1, $up_or_lw1),\n";
	print STDOUT ' ' x 3, '($ch2, $up_or_lw2) = ', "($ch2, $up_or_lw2).\n\n\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"

	    my $gblocks_ch1 = $br2gblocks->{$ch1};
	    my $gblocks_ch2 = $br2gblocks->{$ch2};

	    unless ((defined $gblocks_ch1) and (defined $gblocks_ch2)) { next; }

	    my $plc_lm_gpb = $places_gpbs[$lm_gpb];
	    my $plc_rm_gpb = $places_gpbs[$rm_gpb];

    if ($TEST_LEVEL_br_psm_did_hs >2) {
	print STDOUT ' ' x 3, "/ Passed the 2nd screen /\n\n";
	print STDOUT ' ' x 3, '{Gap-blocks on $ch1}: ', join (', ', @{$gblocks_ch1}), "\;\n";
	print STDOUT ' ' x 3, '{Gap-blocks on $ch2}: ', join (', ', @{$gblocks_ch2}), ".\n\n";

	print STDOUT ' ' x 3,  '($plc_lm_gpb, $plc_rm_gpb) = ', "($plc_lm_gpb, $plc_rm_gpb).\n\n\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"


		# Examine whether a pair of flanking gap-blocks along the "children" exists or not. #

	    my ($left_flanking_ch1, $right_flanking_ch1);
	    my ($plc_rm_lf_ch1, $plc_lm_rf_ch1);
	    my $if_intersecting1=0;

	    foreach my $gb1 (@{$gblocks_ch1}) {
		my ($br1, $up_or_lw_gb1, $lm_gpb1, $rm_gpb1, $invlvd_gpbs1, $affected_exnodes1, $e1) = @{$br_list_gblocks->[$gb1]} ;
		unless ($up_or_lw_gb1 eq $up_or_lw1) { next; }
		my $plc_lm_gpb1 = $places_gpbs[$lm_gpb1];
		my $plc_rm_gpb1 = $places_gpbs[$rm_gpb1];
		if ($plc_rm_gpb < $plc_lm_gpb1) {
		    if ((defined $right_flanking_ch1) and ($plc_lm_rf_ch1 < $plc_lm_gpb1)) { next; }
		    $right_flanking_ch1 = $gb1;
		    $plc_lm_rf_ch1 = $plc_lm_gpb1;
		} elsif ($plc_rm_gpb1 < $plc_lm_gpb) {
		   if ((defined $left_flanking_ch1) and ($plc_rm_gpb1 < $plc_rm_lf_ch1)) { next; }
		    $left_flanking_ch1 = $gb1;
		    $plc_rm_lf_ch1 = $plc_rm_gpb1;

		} elsif (($plc_lm_gpb1 < $plc_lm_gpb) and ($plc_rm_gpb < $plc_rm_gpb1)) {
		    $if_intersecting1 =1;
		    last;
		}
	    }

	    unless ((defined $left_flanking_ch1) or (defined $right_flanking_ch1)) { next; }  # If no candidates are on $ch1, the subject block cannot be parsimonized. (Added on 2019/02/06.) #

	    if ($if_intersecting1) { next; } # If a child gap-block intersects the current block, the latter cannot be parsimonized.

	    my ($left_flanking_ch2, $right_flanking_ch2);
	    my ($plc_rm_lf_ch2, $plc_lm_rf_ch2);
	    my $if_intersecting2 =0; 

	    foreach my $gb2 (@{$gblocks_ch2}) {
		my ($br2, $up_or_lw_gb2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2, $affected_exnodes2, $e2) = @{$br_list_gblocks->[$gb2]} ;
		unless ($up_or_lw_gb2 eq $up_or_lw2) { next; }
		my $plc_lm_gpb2 = $places_gpbs[$lm_gpb2];
		my $plc_rm_gpb2 = $places_gpbs[$rm_gpb2];
		if ($plc_rm_gpb < $plc_lm_gpb2) {
		    if ((defined $right_flanking_ch2) and ($plc_lm_rf_ch2 < $plc_lm_gpb2)) { next; }
		    $right_flanking_ch2 = $gb2;
		    $plc_lm_rf_ch2 = $plc_lm_gpb2;
		} elsif ($plc_rm_gpb2 < $plc_lm_gpb) {
		   if ((defined $left_flanking_ch2) and ($plc_rm_gpb2 < $plc_rm_lf_ch2)) { next; }
		    $left_flanking_ch2 = $gb2;
		    $plc_rm_lf_ch2 = $plc_rm_gpb2;

		} elsif (($plc_lm_gpb2 < $plc_lm_gpb) and ($plc_rm_gpb < $plc_rm_gpb2)) {
		    $if_intersecting2 =1;
		    last;
		}
	    }

	    unless ((defined $left_flanking_ch2) or (defined $right_flanking_ch2)) { next; }  # If no candidates are on $ch2, the subject block cannot be parsimonized. (Added on 2019/02/06.) #

	    if ($if_intersecting2) { next; } # If a child gap-block intersects the current block, the latter cannot be parsimonized.

    if ($TEST_LEVEL_br_psm_did_hs >2) {
	print STDOUT ' ' x 3, "/ Passed the 3rd screen /\n\n";

	my $out_left_flanking_ch1 = (defined $left_flanking_ch1) ? $left_flanking_ch1 : 'undef';
	my $out_right_flanking_ch1 = (defined $right_flanking_ch1) ? $right_flanking_ch1 : 'undef';
	my $out_plc_rm_lf_ch1 = (defined $plc_rm_lf_ch1) ? $plc_rm_lf_ch1 : 'undef';
	my $out_plc_lm_rf_ch1 = (defined $plc_lm_rf_ch1) ? $plc_lm_rf_ch1 : 'undef';

	my $out_left_flanking_ch2 = (defined $left_flanking_ch2) ? $left_flanking_ch2 : 'undef';
	my $out_right_flanking_ch2 = (defined $right_flanking_ch2) ? $right_flanking_ch2 : 'undef';
	my $out_plc_rm_lf_ch2 = (defined $plc_rm_lf_ch2) ? $plc_rm_lf_ch2 : 'undef';
	my $out_plc_lm_rf_ch2 = (defined $plc_lm_rf_ch2) ? $plc_lm_rf_ch2 : 'undef';


	print STDOUT ' ' x 3, '($left_flanking_ch1, $right_flanking_ch1) = ', "($out_left_flanking_ch1, $out_right_flanking_ch1),\n";
	print STDOUT ' ' x 3, '($plc_rm_lf_ch1, $plc_lm_rf_ch1) = ', "($out_plc_rm_lf_ch1, $out_plc_lm_rf_ch1).\n\n";

	print STDOUT ' ' x 3, '($left_flanking_ch2, $right_flanking_ch2) = ', "($out_left_flanking_ch2, $out_right_flanking_ch2),\n";
	print STDOUT ' ' x 3, '($plc_rm_lf_ch2, $plc_lm_rf_ch2) = ', "($out_plc_rm_lf_ch2, $out_plc_lm_rf_ch2).\n\n\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"

		# (Ended up to here on 2019/01/31.) #

	    my ($left_flanking_gb, $plc_rm_lf_gb, $right_flanking_gb, $plc_lm_rf_gb);
	    ##if ((defined $left_flanking_ch1) and ($right_flanking_ch2)) { # A BUG fixed on 2019/02/04. (1)
	    if ((defined $left_flanking_ch1) and (defined $right_flanking_ch2)) {
		($left_flanking_gb, $right_flanking_gb) = ($left_flanking_ch1, $right_flanking_ch2);
		($plc_rm_lf_gb, $plc_lm_rf_gb) = ($plc_rm_lf_ch1, $plc_lm_rf_ch2);

	    ##} elsif ((defined $left_flanking_ch2) and ($right_flanking_ch1)) { # A BUG fixed on 2019/02/04. (2)
	    ##} elsif ((defined $left_flanking_ch2) and (defined $right_flanking_ch1)) { # A BUG fixed on 2019/02/04. (3)
	    } 
	    if ((defined $left_flanking_ch2) and (defined $right_flanking_ch1)) {
		unless ((defined $left_flanking_gb) and ($plc_rm_lf_ch2 < $plc_rm_lf_gb)) {

			# A BUG Fixed on 2019/02/06!! #
		     ($left_flanking_gb, $right_flanking_gb) = ($left_flanking_ch2, $right_flanking_ch1);
		     ($plc_rm_lf_gb, $plc_lm_rf_gb) = ($plc_rm_lf_ch2, $plc_lm_rf_ch1);
		     ##($left_flanking_gb, $right_flanking_gb) = ($left_flanking_ch1, $right_flanking_ch2);
		     ##($plc_rm_lf_gb, $plc_lm_rf_gb) = ($plc_rm_lf_ch1, $plc_lm_rf_ch2);
			# END of "A BUG Fixed on 2019/02/06!!" #
		} 
	    }

	    unless (defined $left_flanking_gb) { next; } # Skip if no pair of flanking "child" gap-blocks exist. #

    if ($TEST_LEVEL_br_psm_did_hs >2) {
	print STDOUT ' ' x 3, "/ Passed the 5th screen /\n\n";
	print STDOUT ' ' x 3, '($left_flanking_gb, $right_flanking_gb) =  ', "($left_flanking_gb, $right_flanking_gb),\n";
	print STDOUT ' ' x 3, '($plc_rm_lf_gb, $plc_lm_rf_gb) = ', "($plc_rm_lf_gb, $plc_lm_rf_gb).\n\n\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"

		# If a pair exists,
		# attempt to get the flanking block candidates to truly flank the current gap-block.

		# For this purpose, first examine whether the candidates can actually flank the current gap-block, solely by changing the alignment into its equivalent form, if necessary.

	    my @left_padding = (); # element = [$indx_pad, $relation, $plc_lm_pad, $plc_rm_pad]

	    my $affected_exnodes_lf = $br_list_gblocks->[$left_flanking_gb]->[5];

	    my $ex_plc_rm_lf_gb = $plc_rm_lf_gb;
	    while ($ex_plc_rm_lf_gb +1 < $plc_lm_gpb) { # 1st middle while-loop.

		my $desired_gpb = $gpbs_at_spec_places[$ex_plc_rm_lf_gb +1];
		my $padding_cands = $gblocks_w_spec_lm->[$desired_gpb];
		unless (@{$padding_cands}>0) { last; }

		my $if_rlv=0;
		foreach my $cand (@{$padding_cands}) { # Inner foreach-loop (over candidate gap-blocks). #
		    my ($lm_cand, $rm_cand, $affected_exnodes_cand) = @{$br_list_gblocks->[$cand]}[2,3,5];
			# ($br2, $up_or_lw_gb2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2, $affected_exnodes2, $e2) = @{$br_list_gblocks->[$gb2]} ;

		    my ($common, $ex_lf, $ex_cand) = diff_nmarray (@{$affected_exnodes_lf}, @{$affected_exnodes_cand});
		   my $rel ; # How $cand is related to $left_flanking_gb.
		   if (@{$common}>0) {
			if (@{$ex_cand}>0) {
			    $rel = (@{$ex_lf}>0) ? 'ON' : '<';
			}
		   } elsif (scalar (@{$ex_lf}) + scalar (@{$ex_cand}) == $ct_exnodes) {
			$rel = 'Cp';
		   }
		   unless (defined $rel) { next; } # The candidate canNOT pad the gap.
		   my $plc_rm_cand = $places_gpbs[$rm_cand];
		   push @left_padding, [$cand, $rel, $ex_plc_rm_lf_gb+1, $plc_rm_cand];
		   $ex_plc_rm_lf_gb = $plc_rm_cand;
		   $if_rlv=1;
		   last;

		} # End of the inner foreach-loop (over candidate gap-blocks). #

		unless ($if_rlv) { last; }

	    }	# End of the 1st middle while-loop.

	    if ($ex_plc_rm_lf_gb +1 < $plc_lm_gpb) { next; } # FAILED to get the left-flanking candidate to truly flank the current block.

    if ($TEST_LEVEL_br_psm_did_hs >2) {
	my $ct_lpads = @left_padding;
	print STDOUT ' ' x 3, "/ Passed the 6th screen (L) (#{left-padding} = $ct_lpads) /\n\n";

	if ($TEST_LEVEL_br_psm_did_hs>3) {
	    print STDOUT ' ' x 3, '( Contents of @left_padding) ', "\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Indx', 'pad-gb', 'rel', 'plc_lm_pad', 'plc_rm_pad'), "\n\n";
	    for (my $k = 0; $k <$ct_lpads; $k++) {
		print STDOUT ' ' x 3, join ("\t", $k, @{$left_padding[$k]}), "\n";
	    }
	    print STDOUT "\n";
	}
	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"

    
	    my @right_padding = (); # element = [$indx_pad, $relation, $plc_lm_pad, $plc_rm_pad]

	    my $affected_exnodes_rf = $br_list_gblocks->[$right_flanking_gb]->[5];

	    my $ex_plc_lm_rf_gb = $plc_lm_rf_gb;
	    while ($plc_rm_gpb < $ex_plc_lm_rf_gb -1) { # 2nd middle while-loop.

		my $desired_gpb = $gpbs_at_spec_places[$ex_plc_lm_rf_gb -1];
		my $padding_cands = $gblocks_w_spec_rm->[$desired_gpb];
		unless (@{$padding_cands}>0) { last; }

		my $if_rlv=0;
		foreach my $cand (@{$padding_cands}) { # Inner foreach-loop (over candidate gap-blocks). #
		    my ($lm_cand, $rm_cand, $affected_exnodes_cand) = @{$br_list_gblocks->[$cand]}[2,3,5];
			# ($br2, $up_or_lw_gb2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2, $affected_exnodes2, $e2) = @{$br_list_gblocks->[$gb2]} ;

		    my ($common, $ex_rf, $ex_cand) = diff_nmarray (@{$affected_exnodes_rf}, @{$affected_exnodes_cand});
		   my $rel ; # How $cand is related to $left_flanking_gb.
		   if (@{$common}>0) {
			if (@{$ex_cand}>0) {
			    $rel = (@{$ex_rf}>0) ? 'ON' : '<';
			}
		   } elsif (scalar (@{$ex_rf}) + scalar (@{$ex_cand}) == $ct_exnodes) {
			$rel = 'Cp';
		   }
		   unless (defined $rel) { next; } # The candidate canNOT pad the gap.
		   my $plc_lm_cand = $places_gpbs[$lm_cand];
		   push @right_padding, [$cand, $rel, $plc_lm_cand, $ex_plc_lm_rf_gb -1];
		   $ex_plc_lm_rf_gb = $plc_lm_cand;
		   $if_rlv=1;
		   last;

		} # End of the inner foreach-loop (over candidate gap-blocks). #

		unless ($if_rlv) { last; }

	    }	# End of the 2nd middle while-loop.

	    if ($plc_rm_gpb < $ex_plc_lm_rf_gb -1) { next; } # FAILED to get the right-flanking candidate to truly flank the current block.

    if ($TEST_LEVEL_br_psm_did_hs >2) {
	my $ct_rpads = @right_padding;
	print STDOUT ' ' x 3, "/ Passed the 6th screen (R) (#{right-padding} = $ct_rpads) /\n\n";

	if ($TEST_LEVEL_br_psm_did_hs>3) {
	    print STDOUT ' ' x 3, '( Contents of @right_padding) ', "\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Indx', 'pad-gb', 'rel', 'plc_lm_pad', 'plc_rm_pad'), "\n\n";
	    for (my $k = 0; $k <$ct_rpads; $k++) {
		print STDOUT ' ' x 3, join ("\t", $k, @{$right_padding[$k]}), "\n";
	    }
	    print STDOUT "\n";
	}
	print STDOUT "\n";

    } # End of "if ($TEST_LEVEL_br_psm_did_hs >2) {...}"

    
		# Now, it turns out that the subject block can be parsimonized!! #

	    $ct_parsimonized++; # Increment the counter by one.


		# If not yet, actually get the flanking candidates to truly flank the subject gap-block.

	    if (@left_padding>0) {

    print STDOUT (' ' x 3, "| Actually get the left-flanking candidate to truly flank the subject |\n\n") if ($TEST_LEVEL_br_psm_did_hs >2);

		my $lm_lf_gb = $br_list_gblocks->[$left_flanking_gb][2];
		my $plc_lm_lf_gb = $places_gpbs[$lm_lf_gb];
		my $ex_plc_rm_lf_gb = $plc_rm_lf_gb;

		foreach my $info_pad (@left_padding) { # 1st middle foreach-loop (over left-padding blocks). #
		    my ($gb_pad, $rel, $plc_lm_pad, $plc_rm_pad) = @{$info_pad};
		    if ($rel eq '<') { # Just extend the right-end of the left-flanking.
			$ex_plc_rm_lf_gb = $plc_rm_pad;
			next;

		    } else {  # if (($rel eq 'Cp') or ($rel eq 'ON'))
			# Swap the (extended) left-flanking block with the padding block. #
			# (NOTE that, before swapping, the pad is on the right of the left-flanking.)
			my $span_pad = $plc_rm_pad - $plc_lm_pad +1;
			my $span_lf = $ex_plc_rm_lf_gb - $plc_lm_lf_gb +1;
			my @gpbs_in_pad = @gpbs_at_spec_places[$plc_lm_pad .. $plc_rm_pad];
			for (my $k = $plc_lm_lf_gb; $k <= $ex_plc_rm_lf_gb; $k++) {
			    my $gpb_in_lf = $gpbs_at_spec_places[$k];
			    $places_gpbs[$gpb_in_lf] += $span_pad;
			    $gpbs_at_spec_places[$k + $span_pad] = $gpb_in_lf;
			}
			for (my $k=$plc_lm_pad; $k <= $plc_rm_pad; $k++) {
			    my $gpb_in_pad = $gpbs_in_pad[$k-$plc_lm_pad];
			    $places_gpbs[$gpb_in_pad] -= $span_lf;
			    $gpbs_at_spec_places[$k - $span_lf] = $gpb_in_pad;
			}
				# Update the boundaries of the (extended) left-flanking block. #
			$ex_plc_rm_lf_gb += $span_pad;
			$plc_lm_lf_gb += $span_pad;
		    }
		}	# End of the 1st middle foreach-loop (over left-padding blocks). #
	    } # End of "if (@left_padding>0) {...}"

	    if (@right_padding) {

    print STDOUT (' ' x 3, "| Actually get the right-flanking candidate to truly flank the subject |\n\n") if ($TEST_LEVEL_br_psm_did_hs >2);

		my $rm_rf_gb = $br_list_gblocks->[$right_flanking_gb][3];
		my $plc_rm_rf_gb = $places_gpbs[$rm_rf_gb];
		my $ex_plc_lm_rf_gb = $plc_lm_rf_gb;

		foreach my $info_pad (@right_padding) { # 2nd middle foreach-loop (over right-padding blocks). #
		    my ($gb_pad, $rel, $plc_lm_pad, $plc_rm_pad) = @{$info_pad};
		    if ($rel eq '<') { # Just extend the left-end of the right-flanking.
			$ex_plc_lm_rf_gb = $plc_lm_pad;
			next;

		    } else {  # if (($rel eq 'Cp') or ($rel eq 'ON'))
			# Swap the (extended) right-flanking block with the padding block. #
			# (NOTE that, before swapping, the pad is on the left of the right-flanking.)
			my $span_pad = $plc_rm_pad - $plc_lm_pad +1;
			my $span_rf = $plc_rm_rf_gb - $ex_plc_lm_rf_gb +1;
			my @gpbs_in_pad = @gpbs_at_spec_places[$plc_lm_pad .. $plc_rm_pad];
			for (my $k = $ex_plc_lm_rf_gb; $k <= $plc_rm_rf_gb; $k++) {
			    my $gpb_in_rf = $gpbs_at_spec_places[$k];
			    $places_gpbs[$gpb_in_rf] -= $span_pad;
			    $gpbs_at_spec_places[$k - $span_pad] = $gpb_in_rf;
			}
			for (my $k=$plc_lm_pad; $k <= $plc_rm_pad; $k++) {
			    my $gpb_in_pad = $gpbs_in_pad[$k-$plc_lm_pad];
			    $places_gpbs[$gpb_in_pad] += $span_rf;
			    $gpbs_at_spec_places[$k + $span_rf] = $gpb_in_pad;
			}
				# Update the boundaries of the (extended) right-flanking block. #
			$ex_plc_lm_rf_gb -= $span_pad;
			$plc_rm_rf_gb -= $span_pad;
		    }
		}	# End of the 2nd middle foreach-loop (over right-padding blocks). #
	    } # End of "if (@right_padding) {...}"

			# Now, vertically split the subject gap-block,
			# and merge the resulting blocks with the corresponding flanking blocks. #

    (print STDOUT "\n", ' ' x 3, "| Now, actually 'absorb' the subject gap-block into the flanking blocks |\n\n\n") if ($TEST_LEVEL_br_psm_did_hs >2);

	    # my ($left_flanking_gb, $plc_rm_lf_gb, $right_flanking_gb, $plc_lm_rf_gb);
	    # my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs, $affected_exnodes, $e) = @{$br_list_gblocks->[$gb]} ;

	    my ($br_lf, $lm_lf, $rm_lf, $invlvd_gpbs_lf, $e_lf) = @{$br_list_gblocks->[$left_flanking_gb]}[0,2,3,4,6];
	    my ($br_rf, $lm_rf, $rm_rf, $invlvd_gpbs_rf, $e_rf) = @{$br_list_gblocks->[$right_flanking_gb]}[0,2,3,4,6];

		# Expand the left-flanking block by absorbing the corresponding portion of the subject block. #

	    $br_list_gblocks->[$left_flanking_gb][3] = $rm_gpb;
	    my @new_invlvd_gpbs_lf = (@{$invlvd_gpbs_lf}, @{$invlvd_gpbs});
	    $br_list_gblocks->[$left_flanking_gb][4] = \@new_invlvd_gpbs_lf;

	    my @cp_new_invlvd_gpbs_lf = @new_invlvd_gpbs_lf;
	    $new_br2indels{$br_lf}->[$e_lf][2] = \@cp_new_invlvd_gpbs_lf;

	    my @new_gbs_w_rm_lf = ();
	    foreach my $gb5 (@{$gblocks_w_spec_rm->[$rm_lf]}) {
		if ($gb5 != $left_flanking_gb) { push @new_gbs_w_rm_lf, $gb5; }
	    }
	    $gblocks_w_spec_rm->[$rm_lf] = \@new_gbs_w_rm_lf;

	    my $gbs_w_rm_gpb = $gblocks_w_spec_rm->[$rm_gpb];
	    my $ct_gbs_w_rm_gpb = @{$gbs_w_rm_gpb};
	    for (my $k=0; $k < $ct_gbs_w_rm_gpb; $k++) {
		if ($gbs_w_rm_gpb->[$k] == $gb) {
		    $gbs_w_rm_gpb->[$k] = $left_flanking_gb;
		    last;
		}
	    }

		# Expand the right-flanking block by absorbing the corresponding portion of the subject block. #

	    $br_list_gblocks->[$right_flanking_gb][2] = $lm_gpb;
	    my @new_invlvd_gpbs_rf = (@{$invlvd_gpbs}, @{$invlvd_gpbs_rf});
	    $br_list_gblocks->[$right_flanking_gb][4] = \@new_invlvd_gpbs_rf;

	    my @cp_new_invlvd_gpbs_rf = @new_invlvd_gpbs_rf;
	    $new_br2indels{$br_rf}->[$e_rf][2] = \@cp_new_invlvd_gpbs_rf;

	    my @new_gbs_w_lm_rf = ();
	    foreach my $gb5 (@{$gblocks_w_spec_lm->[$lm_rf]}) {
		if ($gb5 != $right_flanking_gb) { push @new_gbs_w_lm_rf, $gb5; }
	    }
	    $gblocks_w_spec_lm->[$lm_rf] = \@new_gbs_w_lm_rf;

	    my $gbs_w_lm_gpb = $gblocks_w_spec_lm->[$lm_gpb];
	    my $ct_gbs_w_lm_gpb = @{$gbs_w_lm_gpb};
	    for (my $k=0; $k < $ct_gbs_w_lm_gpb; $k++) {
		if ($gbs_w_lm_gpb->[$k] == $gb) {
		    $gbs_w_lm_gpb->[$k] = $right_flanking_gb;
		    last;
		}
	    }

		# Mark the subject gap-block as "eliminated".
		# (=> Later, execute the actual elimination.)

	    $br_list_gblocks->[$gb][2] = undef;
	    $br_list_gblocks->[$gb][3] = undef;
	    $br_list_gblocks->[$gb][4] = undef;

		# Actually remove $gb from some data-sets (associating branches with indels or gap-blocks). #

	    my $gbs_on_br = $br2gblocks->{$br};
	    my $indels_on_br = $new_br2indels{$br};
	    my $ct_gbs_on_br = @{$gbs_on_br};
	    for (my $e2 = $e+1; $e2 < $ct_gbs_on_br; $e2++) {
		my $gb5 = $gbs_on_br->[$e2];

		$br_list_gblocks->[$gb5][6]--;

		$gbs_on_br->[$e2-1] = $gb5;
		$indels_on_br->[$e2-1] = $indels_on_br->[$e2];
	    }
	    pop @{$gbs_on_br};
	    pop @{$indels_on_br};

		# ADDED on 2019/02/05. #

		# If each list is emptied, remove it from the hash. #
	    if (@{$gbs_on_br} ==0) { delete $br2gblocks->{$br}; }
	    if (@{$indels_on_br} == 0) { delete $new_br2indels{$br}; }

		# END of "ADDED on 2019/02/05." #

	# push @cp_list_indels, [$state_ue, $state_le, \@cp_invlvd_gpbs];
	# $new_br2indels{$br} = \@cp_list_indels;

		# Update %new_node2blwise_gstat. #

	    if ($up_or_lw eq 'L') {
		my $blwise_gstat = $new_node2blwise_gstat{$br};
		foreach my $gpb (@{$invlvd_gpbs}) { $blwise_gstat->[$gpb] = $char_pres; }
	    } else { # if ($up_or_lw eq 'U') 
		my $pa = $node2pa->{$br};
		my $blwise_gstat_pa = $new_node2blwise_gstat{$pa};
		foreach my $gpb (@{$invlvd_gpbs}) { $blwise_gstat_pa->[$gpb] = $char_pres; }
		my $eq_br = $br2eq->{$br};
		if (defined $eq_br) {
		    my $blwise_gstat_eq = $new_node2blwise_gstat{$eq_br};
		    foreach my $gpb (@{$invlvd_gpbs}) { $blwise_gstat_eq->[$gpb] = $char_pres; }
		}

	    }

		# %node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
		# with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;

		# ($br2, $up_or_lw_gb2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2, $affected_exnodes2, $e2) = @{$br_list_gblocks->[$gb2]} ;

		# $br2gblocks{$br}->[$e] gives the index of the gap-block (in @br_list_gblocks) that is defined by the $e th indel event along branch $br;
		#
		# @{$gblocks_w_spec_lm[$lm_gpb]} lists the indices of the gap-blocks (in @br_list_gblocks) whose left-most gap-pattern block is the $lm_gpb th one;
		#
		# @{$gblocks_w_spec_rm[$rm_gpb]} lists the indices of the gap-blocks (in @br_list_gblocks) whose right-most gap-pattern block is the $rm_gpb th one;
		# my @places_gpbs = my @gpbs_at_spec_places = (0 .. $ct_gpbs_in_cls -1);

	}	# End of the main for-loop (over gap-blocks). #

		# (Ended up to here on 2019/02/01.) #

    (print STDOUT ' 'x 3, "[[ The Main Process Ended. Now, finish some data-sets ... ]]\n\n") if ($TEST_LEVEL_br_psm_did_hs >1);

		# Put the finishing touch (1): Remove the "absorbed" gap-blocks. #

	my @new_list_gblocks = ();
	foreach my $info_gb (@{$br_list_gblocks}) { # 1st trailing foreach-loop (over gap-blocks). #
	    if (defined $info_gb->[2]) { push @new_list_gblocks, $info_gb; }

	}	# End of the 1st trailing foreach-loop (over gap-blocks). #


		# Put the finishing touch (2): If necessary, rearrange the gap-pattern blocks so that their indices match their places. #

	my $if_rearrange = 0;
	for (my $gpb = 0; $gpb < $ct_gpbs_in_cls; $gpb++) { # 2nd trailing foreach-loop (over gap-pattern blocks). #
	    if ($gpb =! $places_gpbs[$gpb]) { 
		$if_rearrange = 1;
		last;
	    }
	}	# End of the 2nd trailing foreach-loop (over gap-pattern blocks). #

	if ($if_rearrange) {
	    foreach my $info_gb (@new_list_gblocks) { # 3rd trailing foreach-loop (over gap-blocks). #
		my ($lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$info_gb}[2,3,4];
			# ($br2, $up_or_lw_gb2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2, $affected_exnodes2, $e2) = @{$br_list_gblocks->[$gb2]} ;

		$info_gb->[2] = $places_gpbs[$lm_gpb];
		$info_gb->[3] = $places_gpbs[$rm_gpb];
		my @new_invlvd_gpbs = ();
		foreach my $gpb5 (@{$invlvd_gpbs}) {
		    push @new_invlvd_gpbs, $places_gpbs[$gpb5];
		}
		$info_gb->[4] = \@new_invlvd_gpbs;
	    }	# End of the 3rd trailing foreach-loop (over gap-blocks). #

	    foreach my $br (keys %new_br2indels) { # 5th trailing foreach-loop (over branches w/ indels). #
		my $indels_on_br = $new_br2indels{$br};
		foreach my $info_indel (@{$indels_on_br}) {
		    my @new_invlvd_gpbs = ();
		    foreach my $gpb5 (@{$info_indel->[2]}) {
			push @new_invlvd_gpbs, $places_gpbs[$gpb5];
		    }
			# push @cp_list_indels, [$state_ue, $state_le, \@cp_invlvd_gpbs];
			# $new_br2indels{$br} = \@cp_list_indels;

		    $info_indel->[2] = \@new_invlvd_gpbs;
		}
		
	    }	# End of the 5th trailing foreach-loop (over branches w/ indels). #

	    foreach my $node (keys %new_node2blwise_gstat) { # 6th trailing foreach-loop (over all nodes).
		my $old_blwise_gstat = $new_node2blwise_gstat{$node};
		my @new_blwise_gstat = ();

		for (my $gpb=0; $gpb < $ct_gpbs_in_cls; $gpb++) {
		    my $plc = $places_gpbs[$gpb];
		    $new_blwise_gstat[$plc] = $old_blwise_gstat->[$gpb];
		}

		$new_node2blwise_gstat{$node} = \@new_blwise_gstat;

	    }	# End of the 6th trailing foreach-loop (over all nodes).

		# %node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
		# with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;

	} # End of "if ($if_rearrange) {...}"

    (print STDOUT "\n", ' ' x 3, "< Subroutine 'br_parsimonize_dollo_indel_history' Ended />\n\n") if ($TEST_LEVEL_br_psm_did_hs>0);

	return (\%new_br2indels, \%new_node2blwise_gstat, \@new_list_gblocks, \@places_gpbs, \@gpbs_at_spec_places, $ct_parsimonized);

}	# END of the subroutine, 'br_parsimonize_dollo_indel_history'. #


# bch_br_parsimonize_dollo_indel_histories:
#	This subroutine receives a set of indel histories based on the Dollo parsimony,
#	and performs the subroutine, 'br_parsimonize_dollo_indel_history',
#	in a batch.
#	If necessary, it also re-arranges (a copy of) the input MSA into its equivalent form
#	and the list of gap-pattern blocks in each gapped segment (cluster),
#	so that the MSA and the list will be completely coordinated with the parsimonized initial indel history.
#	[Creation STARTED on 2019/02/02, ENDED on 2019/02/03; MODIFIED on 2019/02/06 (added testing outputs).]
#
#	NOTE: This subroutine eats up the input MSA. So, if necessary, carbon copy the input MSA beforehand.
#
# INPUT: (\@in_msa, \@info_init_psm_scn, \@set_exnodes, $topnode_id, \%node2pa, \%node2ch, $char_abs, $char_pres, \%br2eq), where
#
#	${$in_msa$k]} is the $k th sequence embedded in the input MSA;
#
#	@info_init_psm_scn = (\@set_cls_gblocks, \@init_totcts_indels_for_cls, \@init_psm_scn_for_cls) gives information on the initial candidate of the parsimonious scenario giving rise to the reference/reconstructed MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$init_totcts_indels_for_cls[$cl] is the total number of insertion/deletion events in the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	@set_exnodes stores the IDs of all external nodes;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes;
#	$char_abs is the character specifying the absence of the residue at the site (Usually, it's the same as the 'gap' character);
#	$char_pres is the character specifying the presence of the residue at the site (Usually, it's the same as the 'gap' character);
#	%br2eq = ($br_id => $equivalent_br_id, ...) for branches with equivalents;
#
# OUTPUT: (\@new_info_init_psm_scn, \@new_in_msa, \@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places), where
#
#	@new_info_init_psm_scn = (\@new_set_cls_gblocks, \@new_init_totcts_indels_for_cls, \@new_init_psm_scn_for_cls) gives information on the PARSIMONIZED initial candidate of the parsimonious indel history (or scenario) giving rise to the reference/reconstructed MSA gap configuration, where
#	@new_set_cls_gblocks, @new_init_totcts_indels_for_cls, @new_init_psm_scn_for_cls
#	conform to the same formats as those for @set_cls_gblocks, @init_totcts_indels_for_cls, and @init_psm_scn_for_cls, respectively, in the INPUT;
#
#	@new_in_msa is the input MSA modified into coordinate with the parsimonized initial indel history;
#	it conforms to the same format as that for @in_msa in the INPUT;
#
#	@{$set_new_lists_gblocks[$cl]} is @new_list_gblocks, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' for the $cl th gapped segment;
#	@{$sets_places_gpbs[$cl]} is @places_gpbs, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' for the $cl th gapped segment;
#	@{$sets_gpbs_at_spec_places[$cl]} is @gpbs_at_spec_places, which is one of the outputs of 'br_parsimonize_dollo_indel_histories' for the $cl th gapped segment.
#
sub bch_br_parsimonize_dollo_indel_histories (\@\@\@$\%\%$$\%) {

	my ($in_msa, $info_init_psm_scn, $set_exnodes, $topnode_id, $node2pa, $node2ch, $char_abs, $char_pres, $br2eq) = @_;

# INPUT: (\@in_msa, \@info_init_psm_scn, \@set_exnodes, $topnode_id, \%node2pa, \%node2ch, $char_abs, $char_pres, \%br2eq), where

	my ($set_cls_gblocks, $init_totcts_indels_for_cls, $init_psm_scn_for_cls) = @{$info_init_psm_scn};

	my $ct_cls = @{$set_cls_gblocks};
	my $ct_seqs = @{$in_msa};

   (print STDOUT "\n<< Subroutine, 'bch_br_parsimonize_dollo_indel_histories' (#{gapped segments} = $ct_cls\; #{sequences} = $ct_seqs\; topnode_ID = $topnode_id) >>\n\n") if ($TEST_LEVEL_bch_br_psm_did_hs>0);


	my @new_set_cls_gblocks = my @new_init_totcts_indels_for_cls = my @new_init_psm_scn_for_cls = ();

	my @new_info_init_psm_scn = (\@new_set_cls_gblocks, \@new_init_totcts_indels_for_cls, \@new_init_psm_scn_for_cls);

	my @set_new_lists_gblocks = my @sets_places_gpbs = my @sets_gpbs_at_spec_places = ();

	my @new_in_msa = ();

	unless ($ct_seqs>0) {
	    return (\@new_info_init_psm_scn, \@new_in_msa,
	\@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places);
	}

	unless ($ct_cls>0) {
	    foreach my $seq (@{$in_msa}) {
		my $cp = ${$seq};
		${$seq} = '';
		push @new_in_msa, \$cp;
	    }
	    return (\@new_info_init_psm_scn, \@new_in_msa, \@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places);
	}

	my $len_msa = length (${$in_msa->[0]});

		# Initialize @new_in_msa. #

	my $le_lm_cls = $set_cls_gblocks->[0][0][2];
	for (my $k =0; $k <$ct_seqs; $k++) {
	    my $left_end = ($le_lm_cls == 0) ? '' : substr (${$in_msa->[$k]}, 0, $le_lm_cls, '');
	    $new_in_msa[$k] = \$left_end;
	} 

		# (Ended up to here on 2019/02/02.) #

	for (my $cl=0; $cl < $ct_cls; $cl++) { # Main outer for-loop (over clusters, i.e., gapped segments).

	    my $gpbs_in_cl = $set_cls_gblocks->[$cl];
	    my $ct_gpbs_in_cl = @{$gpbs_in_cl};
	    my $le_cl = $gpbs_in_cl->[0][2];
	    my $re_cl = $gpbs_in_cl->[$ct_gpbs_in_cl -1][3];
	    my $len_cl = $re_cl - $le_cl +1;
		# @{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)

	    my $totct_indels0_for_cl = $init_totcts_indels_for_cls->[$cl];
	    my ($br2indels0, $node2blwise_gstat0) = @{$init_psm_scn_for_cls->[$cl]};
		# @{$init_psm_scn_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) 

    (print STDOUT "\n\[ The $cl th gapped segment (#{gap-pattern-blocks} = $ct_gpbs_in_cl, left-end = $le_cl, right-end = $re_cl, length = $len_cl, #{indels}_init0 = $totct_indels0_for_cl) ]\n\n") if ( $TEST_LEVEL_bch_br_psm_did_hs >1);

	    my ($new_br2indels0, $new_node2blwise_gstat0, $new_list_gblocks, $places_gpbs, $gpbs_at_spec_places, $ct_parsimonized) = br_parsimonize_dollo_indel_history ($ct_gpbs_in_cl, @{$set_exnodes}, %{$br2indels0}, %{$node2blwise_gstat0}, $topnode_id, %{$node2pa}, %{$node2ch}, $char_abs, $char_pres, %{$br2eq});
	    unless (defined $new_list_gblocks) {
		print STDERR "[ERROR 'bch_br_parsimonize_dollo_indel_histories'] The subtoutine 'br_parsimonize_dollo_indel_history' FAILED on the $cl th gapped segment: $!\n";
		return (undef, undef, undef, undef, undef);
	    }

#sub br_parsimonize_dollo_indel_history ($\@\%\%$\%\%$$\%) {
#
#	my ($ct_gpbs_in_cls, $set_exnodes, $br2indels, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs, $char_pres, $br2eq) = @_;
#
#	...
#
#	return (\%new_br2indels, \%new_node2blwise_gstat, \@new_list_gblocks, \@places_gpbs, \@gpbs_at_spec_places, $ct_parsimonized);
#}

	    $new_init_totcts_indels_for_cls[$cl] = $totct_indels0_for_cl - $ct_parsimonized;

		# Store those data-sets that have already taken account of
		# the changes accompanying the brief parsimonization.

	    $set_new_lists_gblocks[$cl] = $new_list_gblocks;
	    $sets_places_gpbs[$cl] = $places_gpbs;
	    $sets_gpbs_at_spec_places[$cl] = $gpbs_at_spec_places;
	    $new_init_psm_scn_for_cls[$cl] = [$new_br2indels0, $new_node2blwise_gstat0];

		# Examine whether the gap-pattern blocks are rearranged or not. #

	    my $if_rearranged = 0;
	    if ($ct_parsimonized>0) {
		for (my $gpb = 0; $gpb < $ct_gpbs_in_cl; $gpb++) { # 1st middle for-loop (over gap-pattern blocks).
		    if ($gpb != $places_gpbs->[$gpb]) {
			$if_rearranged = 1;
			last;
		    }
		} # End of the 1st middle for-loop (over gap-pattern blocks).
	    }

			# (1) If necessary, rearrange the gap-pattern blocks for the output data-sets.
			# (2) Append the resulting gapped segment of the input MSA to the output MSA. #

	    my @new_in_msa_for_cl = ();
	    for (my $k=0; $k < $ct_seqs; $k++) { # 2nd middle for-loop (over sequences).
		my $seq_in_cl = substr (${$in_msa->[$k]}, 0, $len_cl, '');
		$new_in_msa_for_cl[$k] = \$seq_in_cl;
	    }	# End of the 2nd middle for-loop (over sequences).

	    my @new_gpbs_in_cl = ();

	    if ($if_rearranged) {

			# First, chop the gapped segment of the input MSA into sub-segments that correspond to gap-pattern blocks.
		my @blwise_in_msa = ();
		for (my $gpb = 0; $gpb < $ct_gpbs_in_cl; $gpb++) { # 3rd middle for-loop (over gap-pattern blocks).

		    my ($start_gpb, $end_gpb) = @{$gpbs_in_cl->[$gpb]}[2,3];
			# @{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)

		    my $len_gpb = $end_gpb - $start_gpb +1;
		    my @in_msa_for_gpb = ();
		    for (my $k = 0; $k < $ct_seqs; $k++) { # Inner for-loop (over sequences).
			my $sbseq = substr (${$new_in_msa_for_cl[$k]}, 0, $len_gpb, '');
			$in_msa_for_gpb[$k] = \$sbseq;
		    }	# End of the inner for-loop (over sequences).

		    $blwise_in_msa[$gpb] = \@in_msa_for_gpb;
		}	# End of the 3rd middle for-loop (over gap-pattern blocks).

			# At ths point, ${$new_in_msa_for_cl[$k]} SHOULD be a null string for ALL $k.

			# Rearrange the gap-pattern blocks, both their information and the corresponding MSAs. #

		my $new_start_gpb = $le_cl;
		for (my $plc = 0; $plc < $ct_gpbs_in_cl; $plc++) { # 5th middle for-loop (over new places of gap-pattern blocks).
		    my $gpb = $gpbs_at_spec_places->[$plc];
		    my $info_gpb = $gpbs_in_cl->[$gpb];
		    my ($indx_gp, $len_gpb0, $start_gpb, $end_gpb) = @{$info_gpb};
		    ##my $len_gpb = $end_gpb - $start_gpb +1;
		    ##my $new_end_gpb = $new_start_gpb + $len_gpb -1;
		    my $new_end_gpb = $end_gpb + $new_start_gpb - $start_gpb;
		    $new_gpbs_in_cl[$plc] = [$indx_gp, $len_gpb0, $new_start_gpb, $new_end_gpb];

		    my $in_msa_for_gpb = $blwise_in_msa[$gpb];

		    for (my $k =0; $k <$ct_seqs; $k++) { # Inner for-loop (over sequences).
			${$new_in_msa_for_cl[$k]} .= ${$in_msa_for_gpb->[$k]};
		    }	# End of the inner for-loop (over sequences).

		    #$new_start_gpb += $len_gpb; # Update the start of the gap-pattern block.
		    $new_start_gpb = $new_end_gpb +1; # Update the start of the gap-pattern block.

		}	# End of the 5th middle for-loop (over gap-pattern blocks).
		
	    } else {
		@new_gpbs_in_cl = @{$gpbs_in_cl};

	    } # End of "if ($if_rearranged) {...} else {...}"

		# Store the rearranged set of gap-pattern blocks for the $cl th gapped segment. #
	    $new_set_cls_gblocks[$cl] = \@new_gpbs_in_cl;


		# Merge the gapped segment of MSA into the output MSA.

	    for (my $k=0; $k <$ct_seqs; $k++) { # 6th middle for-loop (over sequences). 
		${$new_in_msa[$k]} .= ${$new_in_msa_for_cl[$k]};
	    }	# End of the 6th middle for-loop (over sequences).


		# Finally, append the right-flanking gapless segment of MSA to the output MSA.

	    if ($cl == $ct_cls -1) { # The right-flanking gapless segment is exactly the remainder of the input MSA.
		for (my $k=0; $k <$ct_seqs; $k++) { # 7th middle for-loop (over sequences).
		    ${$new_in_msa[$k]} .= ${$in_msa->[$k]};
		}	# End of the 7th middle for-loop (over sequences).

	    } else {

		my $le_next_cl = $set_cls_gblocks->[$cl +1][0][2];
		my $len_gapless = $le_next_cl -1 - $re_cl;

		for (my $k=0; $k <$ct_seqs; $k++) { # 8th middle for-loop (over sequences).
		    ${$new_in_msa[$k]} .= substr (${$in_msa->[$k]}, 0, $len_gapless, '');
		}	# End of the 8th middle for-loop (over sequences).

	    } # End of "if ($cl == $ct_cls -1) {...} else {...}"


	}	# End of the main outer for-loop (over clusters, i.e., gapped segments).

   (print STDOUT "\n<< Subroutine, 'bch_br_parsimonize_dollo_indel_histories' ENDED />/>\n\n") if ($TEST_LEVEL_bch_br_psm_did_hs>0);


	return (\@new_info_init_psm_scn, \@new_in_msa, \@set_new_lists_gblocks, \@sets_places_gpbs, \@sets_gpbs_at_spec_places);

}	# END of the subroutine, 'bch_br_parsimonize_dollo_indel_histories'. #

		# (Ended up to here on 2019/02/03.)


		### RESTART FROM HERE ###

1;	# a "spell" which makes the module "use"-able and/or "require"-able #



