package MyBayesian;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyBayesian.pm:  This module is a collection of subroutines which generate or manipulate
#		Bayesian-type probabilities.
#
use strict;
use warnings;
use Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( post_bayes_binomial post_bayes_binomial_old );
			# pv_binom_dbl_fluc ); # moved to 'MyProbdistri.pm'

#use MyPartition_Dichr qw( log_sum );
use MySpecFunc qw( gammln betai );


# post_bayes_binomial: This subroutine generates approximate
#		posterior probabilities of the Bernoulli parameter p
#		using Bayes' rule:
#		P(X|A) =  P(A|X) * pi(X) /Sum_x P(A|X) * pi (X),
#	where X is the parameter whose value we want to estimate,
#	A is the observation, and the prior pi (X) is assumed to be uniform.
#	The tails are cut off if the value of P(X|A) falls short of $error
#	times the maximum value.
#	It determines the interval width based on the stanrard deviation of the beta distribution,
#	thus the assignment should be more imppartial compared to the old version,
#
# INPUT: ($total=#{trials}, $pos=#{events in question}, $p_thrsh, $error)
# OUTPUT:   \@distribution = \([av_p1, prob1], [av_2, prob2], ...),
#	where av_p_m is the average value of the parameter $p in the m-th interval,
#	and prob_m is the integrated posterior probability in the m-th interval.
#	All subintervals have the same width specified by the following formula: 1/ M, with
#	M = max [$MIN_CT_INTVL, int (2 /$sigma) (+1 if 2 /$sigma is fractional) ],
#	where $sigma = sqrt ( ($pos +1) * ($total - $pos +1) / ($total + 3) ) / ($total + 2)
#	is the standard deviation of the beta distribution:
#	p ** $pos * (1 - p) ** ($total - $pos) /B($pos + 1, $total - $pos + 1).
#
#       The conditional probability (likelihood) is given by:  
#	P(A|X) = Norm * $p ** $pos * (1-$p) ** ($total - $pos) ,
#	where Norm is the normalization factor,
#	the observation X = {#trials = $total, #events = $pos},
#	and the inference A = {Bernoulli probability = $p} .
#
# ASSUMPTION :  The (unnormalized) prior probability should have the form:
#	Pri ($p) = 1 for 0 <= $p <= $p_thrsh,
#		 = (1 - $p)** $total / (1 - $p_thrsh) ** $total  for $p_thrsh < $p .
#
sub post_bayes_binomial ($$$$) {

	my ($total, $pos, $p_thrsh, $error) = @_;

	my $MIN_CT_INTVLS = 20;

	($p_thrsh > 1) and ($p_thrsh =1);
	if ( ($total < $pos) or ($total <= 0) or ($pos < 0) ) {
		print STDERR "Illegal input for \'post_bayes_binomial\': (total, pos)  = ($total, $pos).\n";
		return undef;
	}

		## coefficients ##
	my $c_ave1 = ($pos + 1) /($total + 2);  # for the average of $p (< $p_thrsh)
	my $c_ave2 = ($pos + 1) / (2 *$total + 2); # for the average of $p (>= $p_thrsh)
		# for the intergrated probability of $p (>= $p_thrsh) #
	my $c_intp = ($p_thrsh >= 1) ? 1 :
		exp ( gammln ($total +2) + gammln (2* $total - $pos +1)
			- gammln ($total - $pos +1) - gammln (2 *$total +2) - $total * log (1 - $p_thrsh) ) ;

		## Initialization ##

	my $sigma = 1/($total + 2) * sqrt ( ($pos + 1) * ($total - $pos + 1) / ($total +3) ); # standard deviation #
	my $ct_intvls =  2 / $sigma;
	($ct_intvls == int ($ct_intvls) ) or ($ct_intvls = int ($ct_intvls) + 1 );
	($ct_intvls < $MIN_CT_INTVLS) and ($ct_intvls = $MIN_CT_INTVLS);

	my $init = int ($ct_intvls * $pos /$total + 1/2) ; 
	
	($init > $ct_intvls * $p_thrsh )
		and ( $init = ($p_thrsh == 1) ? $ct_intvls : int ($ct_intvls * $p_thrsh ) -1 );
	($init < 1) and ($init = 1);

	my $iprob = betai ($pos +1, $total - $pos + 1, $init /$ct_intvls)
			- betai ($pos +1, $total - $pos + 1, ($init -1) /$ct_intvls);
	my $inumer = betai ($pos +2, $total - $pos +1, $init /$ct_intvls)
			- betai ($pos +2, $total - $pos +1, ($init -1) /$ct_intvls);

	my  @distri = ([$c_ave1* $inumer /$iprob, $iprob] );
	my @intervals = ([($init -1)/$ct_intvls, $init/$ct_intvls] );
	my $denom = $iprob;

	for (my $k = $init -1 ; $k > 0; $k--) {

	    my $prob = betai ($pos +1, $total -$pos +1, $k /$ct_intvls)
			- betai ($pos +1, $total - $pos +1, ($k -1) /$ct_intvls);

	    ($prob < $iprob * $error) and last;

	    $denom += $prob;
	    my $numer = betai ($pos +2, $total -$pos +1, $k /$ct_intvls)
			- betai ($pos +2, $total - $pos +1, ($k -1) /$ct_intvls);

	    unshift @distri , [$c_ave1 * $numer /$prob, $prob];
	    unshift @intervals , [($k-1)/$ct_intvls, $k/$ct_intvls];

	}

	for (my $k = $init +1; $k <= $ct_intvls; $k++) {

			# Determine when we do NOT apply the filter.
	   my $fl_nonfilter = ( ($k -1 < $ct_intvls * $p_thrsh) or ($p_thrsh == 1) ) ? 1 : 0;

	   my $prob0 =  $fl_nonfilter ?
 			betai ($pos +1, $total -$pos +1, $k /$ct_intvls)
		 	    - betai ($pos +1, $total - $pos +1, ($k -1) /$ct_intvls) :
			 betai ($pos +1, 2 *$total -$pos +1, $k /$ct_intvls)
			   - betai ($pos +1, 2 *$total - $pos +1, ($k -1) /$ct_intvls) ;

	    my $prob = $fl_nonfilter ? $prob0 : $prob0 * $c_intp ;

	    ($prob < $iprob * $error) and last;

	    $denom += $prob;
	    my $numer = $fl_nonfilter ? 
 			betai ($pos +2, $total -$pos +1, $k /$ct_intvls)
		 	    - betai ($pos +2, $total - $pos +1, ($k -1) /$ct_intvls) :
			 betai ($pos +2, 2 *$total -$pos +1, $k /$ct_intvls)
			   - betai ($pos +2, 2 *$total - $pos +1, ($k -1) /$ct_intvls) ;

	    my $c_ave = $fl_nonfilter ? $c_ave1 : $c_ave2;
	    push @distri, [$c_ave *$numer/$prob0, $prob];
	    push @intervals, [($k-1)/$ct_intvls, $k/$ct_intvls]; 
	}


		## divide all numerators with the common denominator
	foreach my $element (@distri) { $element->[1] /= $denom; }

	return (\@distri, \@intervals);

}		# post_bayes_binomial #


# post_bayes_binomial_old: This subroutine generates approximate
#		posterior probabilities of the Bernoulli parameter p
#		using Bayes' rule:
#		P(X|A) =  P(A|X) * pi(X) /Sum_x P(A|X) * pi (X),
#	where X is the parameter whose value we want to estimate,
#	A is the observation, and the prior pi (X) is assumed to be uniform.
#	The tails are cut off if the value of P(X|A) falls short of $error
#	times the maximum value.
#	It determines the size of intervals primarily based on the number of trials,
#	thus the assignment is unbiased for different number of positives.
#
# INPUT: ($total=#{trials}, $pos=#{events in question}, $p_thrsh, $error)
# OUTPUT:   \@distribution = \([av_p1, prob1], [av_2, prob2], ...),
#	where av_p_m is the average value of the parameter $p in the m-th interval,
#	and prob_m is the integrated posterior probability in the m-th interval.
#	All subintervals have the same width specified by the following formula: 1/ M, with
#	M = max [$MIN_CT_INTVL, int ($total) (+1 if $total is fractional)
#			 {, int ($total /$pos) if  0< $pos < 1 (+1 if its fractional) } ] .
#
#       The conditional probability (likelihood) is given by:  
#	P(A|X) = Norm * $p ** $pos * (1-$p) ** ($total - $pos) ,
#	where Norm is the normalization factor,
#	the observation X = {#trials = $total, #events = $pos},
#	and the inference A = {Bernoulli probability = $p} .
#
# ASSUMPTION :  The (unnormalized) prior probability should have the form:
#	Pri ($p) = 1 for 0 <= $p <= $p_thrsh,
#		 = (1 - $p)** $total / (1 - $p_thrsh) ** $total  for $p_thrsh < $p .
#
sub post_bayes_binomial_old ($$$$) {

	my ($total, $pos, $p_thrsh, $error) = @_;

	my $MIN_CT_INTVLS = 20;

	($p_thrsh > 1) and ($p_thrsh =1);
	if ( ($total < $pos) or ($total <= 0) or ($pos < 0) ) {
		print STDERR "Illegal input for \'post_bayes_binomial_old\': (total, pos)  = ($total, $pos).\n";
		return undef;
	}

		## coefficients ##
	my $c_ave1 = ($pos + 1) /($total + 2);  # for the average of $p (< $p_thrsh)
	my $c_ave2 = ($pos + 1) / (2 *$total + 2); # for the average of $p (>= $p_thrsh)
		# for the intergrated probability of $p (>= $p_thrsh) #
	my $c_intp = ($p_thrsh >= 1) ? 1 :
		exp ( gammln ($total +2) + gammln (2* $total - $pos +1)
			- gammln ($total - $pos +1) - gammln (2 *$total +2) - $total * log (1 - $p_thrsh) ) ;

		## Initialization ##
	my ($ct_intvls, $init) = ( ($pos > 0) and ($pos < 1) ) ? ($total /$pos, 1) : ($total, int ($pos + 1/2) );
	($ct_intvls == int ($ct_intvls) ) or ($ct_intvls = int ($ct_intvls) + 1 );

	if ($ct_intvls < $MIN_CT_INTVLS) {
	    $ct_intvls  = $MIN_CT_INTVLS;
	    $init =  int ($MIN_CT_INTVLS * $pos/$total +1/2) ;
	}
	($init > $ct_intvls * $p_thrsh )
		and ( $init = ($p_thrsh == 1) ? $ct_intvls : int ($ct_intvls * $p_thrsh ) -1 );
	($init < 1) and ($init = 1);

	my $iprob = betai ($pos +1, $total - $pos + 1, $init /$ct_intvls)
			- betai ($pos +1, $total - $pos + 1, ($init -1) /$ct_intvls);
	my $inumer = betai ($pos +2, $total - $pos +1, $init /$ct_intvls)
			- betai ($pos +2, $total - $pos +1, ($init -1) /$ct_intvls);

	my  @distri = ([$c_ave1* $inumer /$iprob, $iprob] );
	my @intervals = ([($init -1)/$ct_intvls, $init/$ct_intvls] );
	my $denom = $iprob;

	for (my $k = $init -1 ; $k > 0; $k--) {

	    my $prob = betai ($pos +1, $total -$pos +1, $k /$ct_intvls)
			- betai ($pos +1, $total - $pos +1, ($k -1) /$ct_intvls);

	    ($prob < $iprob * $error) and last;

	    $denom += $prob;
	    my $numer = betai ($pos +2, $total -$pos +1, $k /$ct_intvls)
			- betai ($pos +2, $total - $pos +1, ($k -1) /$ct_intvls);

	    unshift @distri , [$c_ave1 * $numer /$prob, $prob];
	    unshift @intervals , [($k-1)/$ct_intvls, $k/$ct_intvls];

	}

	for (my $k = $init +1; $k <= $ct_intvls; $k++) {

			# Determine when we do NOT apply the filter.
	   my $fl_nonfilter = ( ($k -1 < $ct_intvls * $p_thrsh) or ($p_thrsh == 1) ) ? 1 : 0;

	   my $prob0 =  $fl_nonfilter ?
 			betai ($pos +1, $total -$pos +1, $k /$ct_intvls)
		 	    - betai ($pos +1, $total - $pos +1, ($k -1) /$ct_intvls) :
			 betai ($pos +1, 2 *$total -$pos +1, $k /$ct_intvls)
			   - betai ($pos +1, 2 *$total - $pos +1, ($k -1) /$ct_intvls) ;

	    my $prob = $fl_nonfilter ? $prob0 : $prob0 * $c_intp ;

	    ($prob < $iprob * $error) and last;

	    $denom += $prob;
	    my $numer = $fl_nonfilter ? 
 			betai ($pos +2, $total -$pos +1, $k /$ct_intvls)
		 	    - betai ($pos +2, $total - $pos +1, ($k -1) /$ct_intvls) :
			 betai ($pos +2, 2 *$total -$pos +1, $k /$ct_intvls)
			   - betai ($pos +2, 2 *$total - $pos +1, ($k -1) /$ct_intvls) ;

	    my $c_ave = $fl_nonfilter ? $c_ave1 : $c_ave2;
	    push @distri, [$c_ave *$numer/$prob0, $prob];
	    push @intervals, [($k-1)/$ct_intvls, $k/$ct_intvls]; 
	}


		## divide all numerators with the common denominator
	foreach my $element (@distri) { $element->[1] /= $denom; }

	return (\@distri, \@intervals);

}		# post_bayes_binomial_old #


=moved to 'MyProbdistri' in order to eliminate inconsistency

# pv_binom_dbl_fluc: This subroutine estimates the P-value for the unlikeliness
#		of observing $r2 positives out of $n2 samples,
#		given that you have already observed $r1 positives out of $n1 samples
#		taken from the same populatoin.
#		The subroutine takes account of the fluctuation in the Bernoulli
#		probability $p by using a Bayesian method ('pv_binom_dbl_fluc'),
#		and then considers bionomial distributions each consisting of
#		$n2 independent Bernoulli trials of probability $p (variable).
#
sub pv_binom_dbl_fluc ($$$$$$) {

	my ($n1,$r1,$n2,$r2, $p_thrsh, $error) = @_;

	if ( ($n2 < $r2) or ($n2 <= 0) or ($r2 < 0) ) {
	    print STDERR "Illegal input for 'pv_binom_dblfluc': (n2, r2) = ($n2, $r2) .\n";
	    return undef;
	}

	my $distri_p = post_bayes_binomial ($n2, $r2, $p_thrsh, $error);

}		# end of the subroutine pv_inom

=cut

1;	# a spell to enable the subroutines to be used as modules #

