package MyPopgenAlleleFreq;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPopgenAlleleFreq.pm:
#	This package contains modules to perform population-genetics analyses
#	using allele- (or site-)frequency spectra.
#	(CREATION STARTED on Jan 13 (Fri), 2012)
#	

use strict;
use warnings;


	# For TESTING the package #
use MyMalnIO qw( pr_ODEN_input );

=set aside for later use
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	sitefreqspec_2state fold_stfrqspec_2state
	av_pw_seq_diff_2state
	tajima_s_D_2state
	classwise_av_pw_seq_diffs_2state
	haplotypefreqstats
	);


	# This specifies the level of tests on main subroutines.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 5;


# sitefreqspec_2state:
#	This subroutine calculates an 'unfolded' site frequency spectrum
#	for a 2-state model, from an input set of haplotypes of sampled chromosomes.
#	NOTE: Although the algorithm employed here looks rather esoteric,
#	it saves as much memory as possible while keeping the input haplotypes.
#	(CREATED on Jan 13 (Fri), 2012)
#
# INPUT: (\@haplotypes, $char_anc, $char_drv), where
#	${$haplotypes[$k]} ($k = 0, ..., $ssize-1) is the haplotype of the $k th chromosome in the sample;
#	which is made by concatenating its states at the (putatively) neutral segregating sites;
#	$char_anc and $char_drv are characters representing an ancestral state
#	and a derived state, respectively, for each site (or locus).
#
# OUTPUT: (\@sitefreqspec, $ct_segsites), where
#	$sitefreqspec[$k] (with $k = 0, ..., $ssize) is the number of 
#	sites (or loci) where $k out of $ssize chromosomes show the derived state;
#	$ct_segsites (= Sum_{$k=1}^{$ssize-1} $allelefrqspec[$k])
#	is the total number of segregating sites.
#
sub sitefreqspec_2state (\@$$) {

	my ($haplotypes, $char_anc, $char_drv) = @_;

	my $ssize = @{$haplotypes};
	unless ($ssize>0) { # The sample has no haplotypes. #
	    return ([], 0); # Return an empty spectrum. #
	}

    (print STDOUT "    < Subroutine 'sitefreqspec_2state' >\n\n") if ($TEST_LEVEL>1);

	my $ct_sites = length ${$haplotypes->[0]};

	my @cts_drv_sitewise = ();
	for (1 .. $ct_sites) { push @cts_drv_sitewise, 0; }

	for (my $k=0; $k<$ssize; $k++) { # Main outer for-loop (over haplotypes) #

	    my $cp_haplo = ${$haplotypes->[$k]};
	    for (my $l=0; $l<$ct_sites; $l++) { # Main inner for-loop (over sites (or 2-state loci)) #
		my $char = substr ($cp_haplo, 0, 1, '');
		if ($char eq $char_drv) {
		    $cts_drv_sitewise[$l]++;
		} elsif ($char ne $char_anc) {
		    print STDOUT "[ERROR 'sitefreqspec_2state'] ILLEGAL INPUT: The $k th haplotype has a $l th character, $char, which is NEITHER ancestral ($char_anc) NOR derived ($char_drv): $! !\n";
		    return (undef, undef);
		}
	    }		# End of the main inner for-loop (over sites (or 2-state loci)) #
	}		# End of the main outer for-loop (over haplotypes) #

		# Initialize the spectrum. #
	my @sitefreqspec = ();
	for (0 .. $ssize) { push @sitefreqspec, 0; }
	for (my $l=0; $l<$ct_sites; $l++) { # Finishing for-loop (over sites (or 2-state loci)) #
	    my $ct = $cts_drv_sitewise[$l];
	    $sitefreqspec[$ct]++;
	}		# End of the finishing for-loop (over sites (or 2-state loci)) #

		# #{segregating sites} = total#{sites} - #{non-segregating sites}. #
	my $ct_segsites = $ct_sites - $sitefreqspec[0] - $sitefreqspec[$ssize];

    (print STDOUT "    --- Subroutine 'sitefreqspec_2stat' Ended ---\n\n") if ($TEST_LEVEL>1);

	return (\@sitefreqspec, $ct_segsites);

}	# End of the subroutine 'sitefreqspec_2state' #


# fold_stfrqspec_2state:
#	This subroutine receives an 'unfolded' site frequency spectrum,
#	and returns its 'folded' version.
#
# INPUT: \@sitefreqspec, where
#	$sitefreqspec[$k] (with $k = 0, ..., $ssize) gives the
#	number of 2-state sites (or loci) where $k out of $ssize sampled chromosomes
#	show the derived state.
#	(Or the sites where ($k+1) (with $k = 0, ..., $ssize-2) 
#	out of $ssize sampled chromosomes show the derived state.)
#
# OUTPUT: \@folded_sfs, where
#	$folded_sfs[$k] = $sitefreqspec[$k] + $sitefreqspec[$#sitefreqspec-$k] for ($k=0, ..., int (scalar(@sitefreqspec)/2)-1)
#			= $sitefreqspec[$k] for $k = (scalar(@sitefrqspec)-1)/2 if scalar(@sitefreqspec) is odd.
#
sub fold_stfrqspec_2state (\@) {

	my ($sitefreqspec) = @_;

	my $size_spec = @{$sitefreqspec};

	my @folded = ();

	my $ub_mirror = int ($size_spec/2);
	for (my $k=0; $k<$ub_mirror; $k++) { # Main for-loop (over the upper-half of the spectrum) #
	    push @folded, $sitefreqspec->[$k] + $sitefreqspec->[$size_spec-1-$k];
	}	# End of the main for-loop (over the upper-half of the spectrum) #

	if ($size_spec % 2) { # The spectrum has an odd size.
	    my $mp = ($size_spec-1)/2;
	    push @folded, $sitefreqspec->[$mp];
	}

	return \@folded;

}	# End of the subroutine 'fold_stfrqspec_2state' #


# av_pw_seq_diff_2state:
#	This subroutine recieves an output of the subroutine 'allelefreqspec_2state',
#	then calculates the so called pi, or the average pairwise sequence difference:
#	pi = 2/n/(n-1) * Sum_{0 <= i < j <n} diff(i,j)
#	   = 2/n/(n-1) * Sum_{1 <= k <= n} #{sites where k sequences have the derived character} * k*(n-k).
#
# INPUT: ($n, \@sitefreqspec), where
#	$n is the number of sequences in the sample;
#	$sitefreqspec[$k] ($k = 0, 1, ..., $n -1 (or $n)) is the number of
#	sites at which $k out of $n sequences have the derived character.
#
# OUTPUT: $pi.
#
sub av_pw_seq_diff_2state ($\@) {

	my ($n, $sitefreqspec) = @_;

	unless ($n >= 2) {
	    print STDOUT "[ERROR 'av_pw_seq_diff_2state'] ILLEGAL INPUT: The number of sequences, $n, is TOO SMALL: $! !\n";
	    return undef;
	}
	my $size_spec = @{$sitefreqspec};
	unless (($n + 1 == $size_spec) or ($n == $size_spec)) {
	    print STDOUT "[ERROR 'av_pw_seq_diff_2state'] ILLEGAL INPUT: The SIZE of the input site-frequency spectrum (=$size_spec) does NOT MATCH the number of sequences (=$n): $! !\n";
	    return undef; 
	}

    (print STDOUT "    < Subroutine 'av_pw_seq_diff' >\n\n") if ($TEST_LEVEL>1);

	my $pi = 0;
	for (my $k=1; $k<$n; $k++) { # Main for-loop (over #{sequences with the derived character}) #
	    $pi += $sitefreqspec->[$k] * $k * ($n - $k);
	}	# End of the main for-loop #

	$pi /= $n*($n-1)/2;

    (print STDOUT "    --- Subroutine 'av_pw_seq_diff_2state' Ended ---\n\n") if ($TEST_LEVEL>1);

	return $pi;

}	# End of the subroutine 'av_pw_seq_diff_2state' #


# tajima_s_D_2state:
#	This subroutine calculates Tajima's D statistic (Tajima F. 1989. Genetics 123:585-595) 
#	from a set of n sampled chromosomes.
#	It uses the following formula:
#	D = (pi - S/a1) /sqrt(^Var(pi - S/a1)), with the variance estimator
#	^Var(pi - S/a1) = e1*S + e2*S*(S-1),
#	where pi is the average pairwise sequence difference,
#	S is the number of segregating sites.
#	The coefficients are:
#	a1 = Sum_{k=1}^{n-1} 1/k,
#	a2 = Sum_{k=1}^{n-1} 1/k**2,
#	e1 = 1/a1*{(n+1)/3/(n-1) - 1/a1}, and
#	e2 = 1/(a1**2 + a2) * {2(n**2 + n + 3)/9/n/(n-1) - (n+2)/n/a1 + a2/a1**2}.
#
# INPUT: (\@haplotypes, $char_anc, $char_drv), where
#	${$haplotypes[$k]} ($k = 0, ..., $ssize-1) is the haplotype of the $k th chromosome in the sample;
#	which is made by concatenating its states at the (putatively) neutral segregating sites;
#	$char_anc and $char_drv are characters representing an ancestral state
#	and a derived state, respectively, for each site (or locus).
#
# OUTPUT: ($d_taj, $pi, $ct_segsites, \@sitefreqspec), where
#	$d_taj is Tajima's D;
#	$pi is the average pairwise difference between sequences;
#	$ct_segsites is the number of segregating sites;
#	$sitefreqspec[$k] ($k = 0, ..., $ssize) is the number of
#	sites where $k out of $ssize sequences have the derived character.
# 
sub tajima_s_D_2state (\@$$) {

	my ($haplotypes, $char_anc, $char_drv) = @_;

   (print STDOUT "\n< Subroutine 'tajima_s_D_2state' >\n\n") if ($TEST_LEVEL>0);

	my $n = my $ssize = @{$haplotypes};
	unless ($n >= 2) {
	    print STDOUT "[ERROR 'tajima_s_D_2state'] ILLEGAL INPUT: The number of sequences, $n, is TOO SMALL: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $totct_sites = length ${$haplotypes->[0]};

    if ($TEST_LEVEL>2) {
	print STDOUT "{Sample size} = $n\n";
	print STDOUT "Total#{sites} = $totct_sites\n\n";
    }

	my ($sitefreqspec, $ct_segsites) = sitefreqspec_2state (@{$haplotypes}, $char_anc, $char_drv);
	unless (defined $sitefreqspec) {
	    print STDOUT "[ERROR 'tajima_s_D_2state'] The subroutine 'sitefreqspec_2state' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

    (print STDOUT "#{segregating-sites} = $ct_segsites\n\n") if ($TEST_LEVEL>2);

	unless ($ct_segsites>0) {
	    print STDOUT "[ERROR 'tajima_s_D_2state'] There are NO SEGREGATING-SITES, therefor Tajima's D canNOT BE CALCULATED: $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $pi = av_pw_seq_diff_2state ($n, @{$sitefreqspec});
	unless (defined $pi) {
	    print STDOUT "[ERROR 'tajima_s_D_2state'] The subroutine 'av_pw_seq_diff_2state' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

    (print STDOUT "pi = $pi\n\n") if ($TEST_LEVEL>2);

	my $a1 = my $a2 = 0;
	for (my $k=1; $k<$n; $k++) { # Preliminary for-loop (over #{sequences with the derived character} #
	##for (my $k=0; $k<$n; $k++) { # Preliminary for-loop (over #{sequences with the derived character} # -- A BUG Corrected in creation of 'MyPopgenAlleleFreq.pm' (on Jan 15 (Sun), 2012) #
	    $a1 += 1/$k;
	    $a2 += 1/$k/$k;
	}   # End of the preliminary for-loop (over #{sequences with the derived character}) #

	my $e1 = (($n+1)/(3*($n-1)) - 1/$a1)/$a1;
	my $e2 = (2*($n*$n +$n +3)/(9*$n*($n-1)) - ($n+2)/($n*$a1) +$a2/($a1*$a1))/($a1*$a1 + $a2);

    if ($TEST_LEVEL>2) {
	print STDOUT "a1 = $a1\n";
	print STDOUT "a2 = $a2\n";
	print STDOUT "e1 = $e1\n";
	print STDOUT "e2 = $e2\n\n";
    }

	my $est_var = $e1 * $ct_segsites + $e2 * $ct_segsites * ($ct_segsites - 1);

    (print STDOUT "est{Var(pi - S/a1)} = $est_var\n\n") if ($TEST_LEVEL>2);

	my $d_taj = ($pi - $ct_segsites/$a1)/sqrt ($est_var);

    (print STDOUT "\n=== The subroutine 'tajima_s_D_2state' ENDED ===\n\n") if ($TEST_LEVEL>0);

	return ($d_taj, $pi, $ct_segsites, $sitefreqspec);

}	# End of the subroutine 'tajima_s_D_2state' #


# classwise_av_pw_seq_diffs_2state:
#	This subroutine calculates average pairwise sequence differences
#	within and between sub-classes of sampled sequences under the 2-state model.
#	FORMULA:
#	Within a sub-class (of size na), the average difference should be
#	pi(a) = 2/na/(na-1) * Sum_{l=1,...,#{sites}} ka(l)*(na-ka(l)), 
#	where ka(l) is the number of sequences with the derived state at the l th site;
#	Between sub-classes a (with size na) and b (with size nb), the average difference should be:
#	pi(a,b) = 1/na/nb * Sum_{l=1,...,#{sites}} [ka(l)*(nb-kb(l))+(na-ka(l))*kb(l)],
#	where kb(l) is the number of  sequences with the derived state at the l th site.
#
# INPUT: (\@sbcl_haplotypes, $char_anc, $char_drv), where
#	${$sbcl_haplotypes[$c]->[$k]} ($k = 0, ..., $n[$c]-1) is the haplotype of
#	the $k th sequence in the $c th class (with size $n[$c]);
#	$char_anc  and $char_drv are the characters representing
#	the ancestral and derived states, respectively, at each site. 
#
# OUTPUT: (\@clw_av_pw_seq_diffs, \@sbcl_sizes), where
#	$clw_av_pw_seq_diffs[$a]->[$a] ($a = 0, ..., #{classes}-1)
#	is the average pairwise sequence difference within the $a th class;
#	$clw_av_pw_seq_diffs[$a]->[$b] (0 <= $b, $a < #{classes})
#	is the average pairwise sequence difference betweein the $a th and the $b th classes;
#	$sbcl_sizes[$c] ($c = 0, ..., #{classes}-1) is the number of sampled sequences
#	belonging to the $c th sub-class.
#
sub classwise_av_pw_seq_diffs_2state (\@$$) {

	my ($sbcl_haplotypes, $char_anc, $char_drv) = @_;

    (print STDOUT "\n    < Subroutine 'classwise_av_pw_seq_diffs_2state' >\n\n") if ($TEST_LEVEL>1);

	my $ct_sbcls = @{$sbcl_haplotypes};
	($ct_sbcls>0) or (return ([], []));

	my @sbcl_sizes = ();
	for (my $c=0; $c<$ct_sbcls; $c++) { # 1st preliminary for-loop #
	    my $size = scalar (@{$sbcl_haplotypes->[$c]});
	    unless ($size>0) {
		print STDOUT "[ERROR 'classwise_av_pw_seq_diffs_2state'] ILLEGAL INPUT: The $c th class contains NO SEQUENCE: $! !\n";
		return (undef, undef);
	    }
	    push @sbcl_sizes, $size;
	}		# End of the 1st preliminary for-loop #

	my $ct_sites = length ${$sbcl_haplotypes->[0][0]};

	# $cts_derived_at_site_sbcl[$c]->[$l] is the number of sequences in the $c th subclass that have the derived state at the $l th site.
	#
	my @cts_derived_at_site_sbcl = ();

	for (my $c=0; $c<$ct_sbcls; $c++) { # 1st main outer for-loop (over sub-classes) #
	    my $sbcl_size = $sbcl_sizes[$c];
	    my $haplotypes_in_sbcl = $sbcl_haplotypes->[$c];
	    my @sbct_derived_at_site = ();
	    for (1 .. $ct_sites) { push @sbct_derived_at_site, 0; }

	    for (my $k=0; $k<$sbcl_size; $k++) { # 1st main middle for-loop (over sequences) #
		my $cp_haplotype = ${$haplotypes_in_sbcl->[$k]};
		for (my $l=0; $l<$ct_sites; $l++) { # 1st main inner for-loop (over sites) #
		    my $char = substr ($cp_haplotype, 0, 1, '');
		    if ($char eq $char_drv) { $sbct_derived_at_site[$l]++;
		    } elsif ($char ne $char_anc) {
			print STDOUT "[ERROR 'classwise_av_pw_seq_diffs_2state'] The $l th site (='$char') of the $k th sequence in the $c th class is NEITHER ancestral (=$char_anc) NOR derived (=$char_drv) specified as input: $! !\n";
			return (undef, undef);
		    }
		}	# End of the 1st main inner for-loop (over sites) #
	    } 	# End of the 1st main middle for-loop (over sequences) #

	    $cts_derived_at_site_sbcl[$c] = \@sbct_derived_at_site;

	}	# End of the 1st main outer for-loop (over sub-classes) #

	# $clw_av_pw_seq_diffs[$a]->[$a] ($a = 0, ..., #{classes}-1)
	#
	my @clw_av_pw_seq_diffs = ();

	for (my $a=0; $a<$ct_sbcls; $a++) { # 2nd main outer for-loop (over sub-classes) #
	    my $size_a = $sbcl_sizes[$a];
	    my $cts_a_derived_at_site = $cts_derived_at_site_sbcl[$a];

	    my @row_av_pw_sqdiffs = ();
	    for (my $b=0; $b<$a; $b++) { # 2nd-1st main middle for-loop (over 2nd sub-classes) #
		my $size_b = $sbcl_sizes[$b];
		my $cts_b_derived_at_site = $cts_derived_at_site_sbcl[$b];
		my $pi_ab = 0;

		for (my $l=0; $l<$ct_sites; $l++) { # Main inner for-loop (over sites) #
		    my $ct_drv_a = $cts_a_derived_at_site->[$l];
		    my $ct_drv_b = $cts_b_derived_at_site->[$l];
		    $pi_ab += $ct_drv_a * ($size_b - $ct_drv_b) + ($size_a - $ct_drv_a)*$ct_drv_b;
		}			# Main inner for-loop (over sites) #
		$pi_ab /= $size_a * $size_b;
		$row_av_pw_sqdiffs[$b] = $pi_ab;
	    }		# End of the 2nd-1st main middle for-loop (over 2nd sub-classes) #

	    unless ($size_a>1) { # Leave within-subcluster average undefined. #
		$clw_av_pw_seq_diffs[$a] = \@row_av_pw_sqdiffs;
		next;
	    }

	    my $pi_a = 0;
	    for (my $l=0; $l<$ct_sites; $l++) { # 2nd-2nd main middle for-loop (over sites) #
		my $ct_drv_a = $cts_a_derived_at_site->[$l];
		$pi_a += $ct_drv_a* ($size_a - $ct_drv_a);
	    }			# End of the 2nd-2nd main middoe for-loop (over sites) #
	    $pi_a /= $size_a*($size_a-1)/2;
	    $row_av_pw_sqdiffs[$a] = $pi_a;

	    $clw_av_pw_seq_diffs[$a] = \@row_av_pw_sqdiffs;

	}		# End of the 2nd main outer for-loop (over sub-classes) #

    (print STDOUT "\n    --- The subroutine 'classwise_av_pw_seq_diffs_2state' Ended\n\n") if ($TEST_LEVEL>1);

	return (\@clw_av_pw_seq_diffs, \@sbcl_sizes);

}	# End of the subroutine 'classwise_av_pw_seq_diffs_2state' #


# haplotypefreqstats:
#	This subroutine calculates several statistics regarding allele (or haplotype) frequencies.
#	These statistics can be used for haplotype-frequency methods of neutral tests,
#	e.g. the Ewens-Watterson test (Watterson GA. 1978. Genetics 88:405-407),
#	Ewens' test (Ewens WJ. 1973. Theor Popul Biol 4:251-258),
#	Slatkin's exact test (Slatkin M. 1994. Genet Res 64:71-74; Slatkin M. 1996. Genet Res 68:259-260),
#	etc.
#	
# INPUT: (\@haplotypes), where
#	${$haplotypes[$k]} ($k = 0, ..., $ssize-1) is the haplotype of the $k th chromosome in the sample;
#
# OUTPUT: ($ct_alleles, \@std_cfg_allelefreqs, $homo, \%allele2frq), where
#	$ct_alleles is the number of alleles in the sample;
#	$std_cfg_allelefreqs[$k] ($k=0, ..., $ct_alleles-1) is the ($k+1) th largest allele frequency, i.e. the number of sampled chromosomes showing a particular allele;
#	$homo (= 1/$ssize**2 * Sum_{$k=0,...,$ct_alleles-1} $std_cfg_allelefreqs[$k]**2) is the sample homozygosity.
#
sub haplotypefreqstats (\@) {

	my ($haplotypes) = @_;

    (print STDOUT "\n    < Subroutine 'haplotypefreqstats' >\n\n") if ($TEST_LEVEL>1);

	my $ssize = @{$haplotypes};
	unless ($ssize>0) {
	    return (0, [], undef, {});
	}

    (print STDOUT "{Sample size} = $ssize\n\n") if ($TEST_LEVEL>2);

	my %allele2frq;
	for (my $k=0; $k<$ssize; $k++) { # Preliminary for-loop #
	    my $allele = ${$haplotypes->[$k]};
	    (defined $allele2frq{$allele}) or ($allele2frq{$allele} = 0);
	    $allele2frq{$allele}++;
	}		# End of the preliminary for-loop #

		# 1st main foreach-loop (over alleles) #
	my @std_cfg_allelefreqs = ();
	foreach my $allele (sort { $allele2frq{$b} <=> $allele2frq{$a} } keys %allele2frq) { push @std_cfg_allelefreqs, $allele2frq{$allele}; }
	my $ct_alleles = @std_cfg_allelefreqs; 

		# 2nd main foreach-loop (over allele frequencies) #
	my $homo = 0;
	foreach my $frq (@std_cfg_allelefreqs) { $homo += $frq*$frq; }

	$homo /= $ssize*$ssize;

    (print STDOUT "\n    --- The subroutine 'haplotypefreqstats' Ended ---\n\n") if ($TEST_LEVEL>1);

	return ($ct_alleles, \@std_cfg_allelefreqs, $homo, \%allele2frq);

}	# End of the subroutine 'haplotypefreqstats' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #

