package MyTreeMap;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.
#
#	Some subroutines in this module are based on the algorithms or methods described in:
#	"Inferring Phylogenetics",
#	written by Felsenstein J (Sinauer Associates, Inc.; Sunderland, MA, USA; 2004).
#
# MyTreeManipu_IV.pm:
#	This package manipulates trees more actively than the preceding
#	members of the family "MyTreeManipu_xxx.pm".
#	For example, it can even change the topology of a tree,
#	compare to similar trees, and so on.
#
# MyTreeManipu_II.pm:
#	This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );
##use MyFile_manipu qw( open_input_file open_output_file );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
	input_tree_file read_tree_file br_output_tree
		);
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite


use Exporter;

our @ISA = qw(Exporter);

##sub rec_fillin_boot (\@\$);
#sub rec_prod_boot (\@\@);


our @EXPORT_OK = qw(
	fractionate_by_ndepth fractionate_by_ndepth2
	assign_av_node_levels assign_node2closest_exoffs
	initialize_node2inout assign_node2bival_states
	assign_gnode2sp_lca fetch_splca_comps assign_gnode2sp_lca_comp
	assign_node2stat_weights assign_node2minimum_changes
	map_pure_parsimony_states
	infer_stat_by_pars_topdown infer_stat_by_pars_bottomup infer_gapstat_by_pars
	br_map_indels_on_tree map_indels_on_tree bch_map_indels_on_tree
	bch_infer_gapstat_on_nodes parsmap_binstatchg_on_tree
		) ;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


# fractionate_by_ndepth:
#	This subroutine classifies nodes in a tree
#	according to their depths.
#	The 'depth' is defined as the minimum number of nodes
#	between the node in question and the top node,
#	including itself (but excluding the top node).
#
# INPUT: (\%node2children, \%name2id),  where
#	%node2children = (node_ID => [a list of child ID's], ...).
#
# OUTPUT: \@frac_nodes, where
#	@{$frac_nodes[$k]} is a list of IDs of nodes with depth $k.
#
sub fractionate_by_ndepth (\%\%) {

	my ($node2children, $name2id) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	unless (defined $topnode_id) {
	    print STDERR "[ERROR 'fractionate_by_ndepth'] NO ID for the '$TOP_NODE'!\n";
	    return undef;
	}

	my @frac_nodes = ();

		# Initialize the set #
	my @sub_nodes = ($topnode_id); # depth = 0

	while (@sub_nodes>0) { # The outermost while-loop #

	    my @copy = @sub_nodes;
	    push @frac_nodes, \@copy;

	    my $union_children = [];

	    foreach my $node (@sub_nodes) { # The inner foreach-loop #
		my $children = $node2children->{$node};
		(defined $children) or next;
		$union_children = union_array (@{$union_children}, @{$children});
	    }				# End of the inner foreach-loop #

	    @sub_nodes = @{$union_children};

	}		# End of the outermost while-loop #

	return \@frac_nodes;

}		# end of the subroutine 'fractionate_by_ndepth' #


# fractionate_by_ndepth2:
#	This revised version of 'fractionate_by_ndepth'
#	receives top node ID instead of %name2id,
#	so that it can fractionate sub-trees instead of the whole tree,
#	if desired.
#
# fractionate_by_ndepth:
#	This subroutine classifies nodes in a tree
#	according to their depths.
#	The 'depth' is defined as the minimum number of nodes
#	between the node in question and the top node,
#	including itself (but excluding the top node).
#
# INPUT: (\%node2children, $topnode_id),  where
#	%node2children = (node_ID => [a list of child ID's], ...);
#	$topnode_id should be a node in the tree,
#	but it does not necessarily have to be the top node.
#
# OUTPUT: \@frac_nodes, where
#	@{$frac_nodes[$k]} is a list of IDs of nodes with depth $k.
#
sub fractionate_by_ndepth2 (\%$) {

	my ($node2children, $topnode_id) = @_;

	my @frac_nodes = ();

		# Initialize the set #
	my @sub_nodes = ($topnode_id); # depth = 0

	while (@sub_nodes>0) { # The outermost while-loop #

	    my @copy = @sub_nodes;
	    push @frac_nodes, \@copy;

	    my $union_children = [];

	    foreach my $node (@sub_nodes) { # The inner foreach-loop #
		my $children = $node2children->{$node};
		(defined $children) or next;
		$union_children = union_array (@{$union_children}, @{$children});
	    }				# End of the inner foreach-loop #

	    @sub_nodes = @{$union_children};

	}		# End of the outermost while-loop #

	return \@frac_nodes;

}		# end of the subroutine 'fractionate_by_ndepth2' #


# assign_av_node_levels:
#	This subroutine assigns average levels to the nodes in a (sub-)tree.
#	Here the average level of a node is defined by the WPGMA-like arithmetic mean
#	of the paths from the node to the exterior nodes under it.
#	It is calculated iterativly by the formula:
#	av_level($node) = 0 for exterior nodes,
#	av_level($node) = Sum_{$chi in all children of $node} [ br_len($node, $chi) + av_level ($chi)]
#	for interior nodes.
#
# INPUT: ($topnode_id, \%node2chi, \%node2pa, \%br_attr), where
#	$topnode_id is the top node of the (sub-)tree you want to examine;
#	%node2chi= ($node_ID => \@child_nodes, ...);
#	%node2pa = ($node_ID => $parent_node, ...);
#	%br_attr = ($ID of the node immediately below the branch => \@attr, ..), with
#	$attr[$INDX_LENGTH] gives the length of the branch in question.
#
# OUTPUT: \%node2av_level, where
#	%node2av_level = ($node_ID => av_level($node_ID), ...).
#	The subroutine returns undef if the process failed.
#
sub assign_av_node_levels ($\%\%\%) {

	my ($topnode_id, $node2chi, $node2pa, $br_attr) = @_;

	my $frac_nodes = fractionate_by_ndepth2 (%{$node2chi}, $topnode_id);

	my %node2av_level;

		# Assign the average node levels from bottom to top #
	my $depth = @{$frac_nodes};
	while (my $sb_nodes = pop @{$frac_nodes}) { # The outer while-loop #
	   $depth--;
	   while (my $node = shift @{$sb_nodes}) { # The middle while-loop #
		my $children = $node2chi->{$node};

		unless ((defined $children) and (@{$children}>0)) { # Assign level = 0 to the exterior nodes
		   $node2av_level{$node} = 0;
		   next; 
		}
		my $ct_children= @{$children};

		my $sum_level = 0;
		foreach my $chi (@{$children}) { # The inner foreach-loop #
		    my $chi_level = $node2av_level{$chi};
		    unless (defined $chi_level) {
			print STDERR "[ERROR 'assign_av_node_levels'] NO LEVEL is assigned to the child '$chi' of the node '$node' (depth: $depth): $! !\n";
			return undef;
		    }
		    my $attr = $br_attr->{$chi};
		    my $br_len = (defined $attr) ? $attr->[$INDX_LENGTH] : undef;
		    unless (defined $br_len) {
			print STDERR "[ERROR 'assign_av_node_levels'] NO upstream BRANCH LENGTH is assigned to the child '$chi' of the node '$node' (depth: $depth): $! !\n";
			return undef;
		    }
		    $sum_level += $chi_level + $br_len;
		}			# End of the inner foreach-loop #

		$node2av_level{$node} = $sum_level /$ct_children;

	   }			# End of the middle while-loop #
	}			# End of the outer while-loop #

	return \%node2av_level;

}		# End of the subroutin 'assign_av_node_levels' #


# assign_node2closest_exoffs:
#	This subroutine assigns each node a list of the closest external offsprings
#	as well as the distance between the node and the closest offsprings.
#
# INPUT: (\@frac_nodes, \%node2children, \%br_attr, $fl_noneg), where
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d, where the top node has $d = 0);
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%br_attr = ($id  of the node immediately below the branch => \@attributes, ...) with
#	$length = $attributes[$INDX_LENGTH];
#	$fl_noneg = 1 if we do not allow negative branch lengths (replace them with zeros),
#		 = 0 if we do not care whether the branch length is negative/positive.
#
# OUTPUT: \%node2closest_exoffs, where
#	%node2closest_exoffs = ($node_id => [$distance, [list of closest offsprings]], ...).
#
sub assign_node2closest_exoffs (\@\%\%$) {

	my ($frac_nodes, $node2children, $br_attr, $fl_noneg) = @_;

	my %node2closest_exoffs;
	my $max_depth = @{$frac_nodes} - 1;

	for (my $d=$max_depth; $d>= 0; $d--) { # The outermost for-loop #
	    my $sb_nodes = $frac_nodes->[$d];
	    foreach my $node (@{$sb_nodes}) { # The middle foreach-loop #
		my $children = $node2children->{$node};
		unless (defined $children) {
			# external node #
		   $node2closest_exoffs{$node} = [0, [$node]];
		   next;
		}

		my $mindist ;
		my @closest = ();
		foreach my $child (@{$children}) { # The inner foreach-loop #
		    my $attr = $br_attr->{$child};
		    unless ((defined $attr) and (defined $attr->[$INDX_LENGTH])) {
			print STDERR "[CAUTION 'assign_node2closest_exoffs'] NO BRANCH LENGTH assigned to the child '$child' of the node '$node'!\n";
			next;
		    }
		    my $brlen = $attr->[$INDX_LENGTH];
		    ($brlen <0) and ($fl_noneg>0) and ($brlen = 0); ## Turn negative length into zero

		    my $chcloffs = $node2closest_exoffs{$child};
		    unless ((defined $chcloffs) and (defined $chcloffs->[0])) {
		 	print STDERR "[CAUTION 'assign_node2closest_exoffs'] NO CLOSEST OFFSPRING for the chilld '$child' of the node '$node'!\n";
			next;
		    }
		    my $ch_mindist = $brlen + $chcloffs->[0];

		    unless (defined $mindist) { # Initialize
			$mindist = $ch_mindist;
			@closest = @{$chcloffs->[1]};
			next;
		    }

		    if ($ch_mindist < $mindist) {  # Replace
			$mindist = $ch_mindist;
			@closest = @{$chcloffs->[1]};
		    } elsif ($ch_mindist == $mindist) { # Add
			push @closest, @{$chcloffs->[1]};
		    }
		    
		}			# end of the inner foreach-loop #

		$node2closest_exoffs{$node} = [$mindist, \@closest];

	    }			# end of the middle foreach-loop # 

	}			# end of the outermost for-loop #

	return \%node2closest_exoffs;

}		# end of the subroutine 'assign_node2closest_exoffs' #


# assign_gnode2sp_lca:
#	This subroutine assigns each node of a gene tree the last common ancestor of
#	the species the genes under the node belong to.
#
# INPUT: (\@frac_gnodes, \%gnode2children, \%gnode2spnode, \%spnode2parent), where
#	@{$frac_gnodes[$d]} = (IDs of the gene nodes with depth $d, where the top node has $d = 0);
#	%gnode2children = ($gene_node_id => [IDs of child nodes], ...);
#	%gnode2spnode = ($gene_node_id => $species_node_id, ...) for the external nodes;
#	%spnode2parent = ($species_node_id => $parent_id, ...).
#
# OUTPUT: \%gnode2sp_lca, where
#	%gnode2sp_lca = ($gene_node_id => \@info_sp_lca, ...) for all nodes, with
#	@info_sp_lca = ($node_id, $depth) of the lca in the species node assigned to the gene node.
#
sub assign_gnode2sp_lca (\@\%\%\%) {

	my ($frac_gnodes, $gnode2children, $gnode2spnode, $spnode2parent) = @_;

	my $max_depth = @{$frac_gnodes} - 1;
	my %gnode2sp_lca;

	for (my $d=$max_depth; $d>= 0; $d--) { # The outermost for-loop #
	    my $sb_gnodes = $frac_gnodes->[$d];
	    foreach my $gnode (@{$sb_gnodes}) { # The middle foreach-loop #
		my $gchildren = $gnode2children->{$gnode};

		if (defined $gchildren) {
		    my @sp_ch = ();
		    foreach my $gch (@{$gchildren}) { push @sp_ch, $gnode2sp_lca{$gch}->[0]; }
		    my ($sp_lca, $depth) = fetch_last_common_ancestor_ov2 (@sp_ch, %{$spnode2parent});
		    $gnode2sp_lca{$gnode} = [$sp_lca, $depth];
		} else {
		    my $spnode = $gnode2spnode->{$gnode};
		    unless (defined $spnode) {
			print STDERR "[ERROR 'assign_gnode2sp_lca'] NO SPECIES NODE asigned to the gene external node '$gnode'!\n";
			return undef;
		    }
		    my $sp_anc = fetch_ancestry ($spnode, %{$spnode2parent});
		    my $depth = @{$sp_anc} -1;
		    $gnode2sp_lca{$gnode} = [$spnode, $depth];
		}
	    }				# End of the middle foreach-loop #

	}			 # End of the outermost for-loop #

	return \%gnode2sp_lca;

}		# end of the subroutine 'assign_gnode2sp_lca' #


# fetch_splca_comps:
#	This satellite subroutine of 'assign_gnode2sp_lca_comp' 
#	fetches a set of child nodes of a potentially multifurcated node in a species tree.
#
# INPUT: ($gnode, $spnode, $gch, $sp_ch, \%gnode2sp_lca_comp, \%spnode2parent), where
#	$spnode is the species node corresponding to the gene node $gnode;
#	$gch is a child of $gnode;
#	$sp_ch is the species node corresponding to $gch;
#	%gnode2sp_lca_comp = ($gnode_id =>
#	   [IDs of the children of the corresponding species node], ...)
#	for the gene nodes corresponding to the species nodes with more than 2 children;
#	%spnode2children = ($species_node_id => [a set of child IDs], ...).
#
# OUTPUT: \@sp_lca_comps, where
#	@sp_lca_comps is a list of the IDs of the children of $spnode
#	corresponding to $gch.
#	NOTE that it does not necessarily mean the species counterpart of $gch.
#
sub fetch_splca_comps ($$$$\%\%) {

	my ($gnode, $spnode, $gch, $sp_ch, $gnode2sp_lca_comp, $spnode2parent) = @_;

	if ($sp_ch == $spnode) {
	    my $sp_ch_comp = $gnode2sp_lca_comp->{$gch};
	    unless (defined $sp_ch_comp) {
	        print STDERR "[ERROR 'assign_gnode2sp_lca_comp'] NO SPECIES LCA COMPONENTS assigned for the child '$gch' of the gene node '$gnode'.\n";
	        return undef;
	    }
	    return $sp_ch_comp;

	} elsif ($spnode == $spnode2parent->{$sp_ch}) {

	   return [ $sp_ch ];
	} else {

	    (undef, undef, my $sp_ch_anc) = fetch_common_ancestors ($spnode, $sp_ch, %{$spnode2parent}) ;
	    unless ((defined $sp_ch_anc) and (@{$sp_ch_anc}>0)) {
	        print STDERR "[ERROR 'fetch_splca_comps'] NO ANCESTRY below '$spnode' was found for the species node '$sp_ch' corresponding to the child '$gch' of the gene node '$gnode'!\n";
	         return undef;
	    }

	    return [ $sp_ch_anc->[0] ];
	}

}			# End of the subroutine 'fetch_splca_comps' #


# assign_gnode2sp_lca_comp:
#	This subroutine assigns each node of a gene tree that is annotated with
#	a multifurcating species node the list of 'components',
#	i.e. children of the multifurcating species node
#	corresponding to the children of the gene node.
#
# INPUT: (\@frac_gnodes, \%gnode2children, \%gnode2sp_lca, \%spnode2children, \%spnode2parent), where
#	@{$frac_gnodes[$d]} = (IDs of the gene nodes with depth $d, where the top node has $d = 0);
#	%gnode2children = ($gene_node_id => [IDs of child nodes], ...);
#	%gnode2sp_lca = ($gene_node_id => \@info_sp_lca, ...) for all nodes, with
#	@info_sp_lca = ($node_id, $depth) of the lca in the species node assigned to the gene node;
#	%spnode2children = ($species_node_id => [a list of children], ...);
#	%spnode2parent = ($species_node_id => $parent_id, ...).
#
# OUTPUT: \%gnode2sp_lca_comp = ($gene_node_id => \@sp_lca_comp, ...)
#	for gene nodes assigned with the species nodes each with more than 2 children,
#	with @sp_lca_comp = (a list of species child nodes corresponding to
#	the gene children nodes) for the relevant gene node,
#   ### = (the corresponding child species node) for a child gene node. -- disabled ###
#
sub assign_gnode2sp_lca_comp (\@\%\%\%\%) {

	my ($frac_gnodes, $gnode2children, $gnode2sp_lca,
		$spnode2children, $spnode2parent) = @_;

	my $max_depth = @{$frac_gnodes} - 1;
	my %gnode2sp_lca_comp;

	for (my $d=$max_depth; $d>= 0; $d--) { # The outermost for-loop #
	    my $sb_gnodes = $frac_gnodes->[$d];
	    foreach my $gnode (@{$sb_gnodes}) { # The middle foreach-loop #
		my $info_sp_lca = $gnode2sp_lca->{$gnode};
		unless (defined $info_sp_lca) {
		    print STDERR "[ERROR 'assign_gnode2sp_lca_comp'] NO SPECIES LCA assigned to the gene node '$gnode'!\n";
		    return undef;
		}
		my $sp_lca = $info_sp_lca->[0];
		my $sp_children = $spnode2children->{$sp_lca};
		((defined $sp_children) and (@{$sp_children}>2)) or next;

		my $gchildren = $gnode2children->{$gnode};
		unless (defined $gchildren) {
		    printf STDERR "[ERROR 'assign_gnode2sp_lca_comp'] The gene node '$gnode' has NO CHILDREN while it corresponds to the species node '$sp_lca' with %d children !!\n", scalar (@{$sp_children});
		    return undef;
		}

		my @sp_lca_comps = ();
		foreach my $gch (@{$gchildren}) { # The inner foreach-loop #
		    my $info_sp_ch = $gnode2sp_lca->{$gch};
		    unless (defined $info_sp_ch) {
			print STDERR "[ERROR 'assign_gnode2sp_lca_comp'] NO SPECIES LCA assigned to the child '$gch' of the gene node '$gnode'!\n";
			return undef;
		    }
		    my $sp_ch = $info_sp_ch->[0];

		    my $sp_lca_sbcomps = fetch_splca_comps ($gnode, $sp_lca, $gch, $sp_ch, %gnode2sp_lca_comp, %{$spnode2parent});
		    unless (defined $sp_lca_sbcomps) {
			print STDERR "[ERROR 'assign_gnode2sp_lca_comp'] 'fetch_splca_comps' FAILED for the specied node '$sp_lca' and its offspring '$sp_ch' (corresponding to theagene node '$gch')!\n";
			return undef;
		    }
		# Assign sub-components to the child gene node if undefined yet #
		##    (defined $gnode2sp_lca_comp{$gch}) or
		##  	( $gnode2sp_lca_comp{$gch} = $sp_lca_sbcomps); -- DISABLED on Sat, Sep 2, 2006

		    push @sp_lca_comps, @{$sp_lca_sbcomps};
		}			# End of the inner foreach-loop #
		my $nr_sp_lca_comps = nr_array (@sp_lca_comps);
		$gnode2sp_lca_comp{$gnode} = $nr_sp_lca_comps;

	    }				# End of the middle foreach-loop #
	}			 # End of the outermost for-loop #

	return \%gnode2sp_lca_comp;

}		# end of the subroutine 'assign_gnode2sp_lca_comp' #


# initialize_node2inout:
#	This subroutine initializes an input hash %node2inout,
#	by assigning the 'in'/'out' states to external nodes.
#	The hash processed by this subroutine can be used for
#	an input of the subroutine 'assign_node2bival_states' below.
#
# INPUT: (\%node2inout, \%gnode2spnode, \@sp_ingroups), where
#	%node2inout = ($node_id => $state, ...) with
#	  $state = 'in' if the subtree under the node consists only of in-group sequences,
#		 = 'out' if it consists only of out-group sequences,
#	  	 = 'mixed' if it consists both of in- and out-group sequences.
#	This hash SHOULD BE EMPTY when input to this subroutine;
#	%gnode2spnode = ($gene_node_id => $species_node_id, ...)
#	   for the external nodes of the input gene tree;
#	@sp_ingroups = (a list of 'ingroup' species).
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub initialize_node2inout (\%\%\@) {

	my ($node2inout, $gnode2spnode, $sp_ingroups) = @_;

	my %in_spnodes; # ($spnode_id => 1, ...) for 'in-group' species
	foreach my $spnode (@{$sp_ingroups}) { $in_spnodes{$spnode} = 1; }

	foreach my $gnode (keys %{$gnode2spnode}) {
	    my $spnode = $gnode2spnode->{$gnode};
	    (defined $spnode) or next;
	    $node2inout->{$gnode} = (defined $in_spnodes{$spnode}) ? 'in' : 'out' ;
	}

	return 1;

}		# end of the subroutine 'initialize_node2inout' #


# assign_node2bival_states:
#	This subroutine assigns all nodes divalent states.
#	The rules for the assignment are the following:
#	(i) If its children all has the same state, the node is also assigned the same state;
#	(ii) otherwise, it is assigned 'mixed'.
#
# REVISED on Saturday, Sep 9, 2006:
#	It changed the policy on how to deal with external nodes with NO state information.
#	It will suspend the assignment of the state to such external nodes,
#	and the result will NOT affect the assignment of states to their ancesters.
#
# INPUT: (\%node2state, \@frac_nodes, \%node2children), where
#	%node2state = ($node_id => $state, ...), with
#	$state = $state1 or $state2 or 'mixed' depending on the situation;
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d, where the top node has $d = 0);
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#
# REQUISITE: %node2state is properly initialized,
#	namely, multiple-choices of the status for the external nodes are already given.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#	The main job of this subroutine is to fill in %node2state
#	with an appropriate state.
#
sub assign_node2bival_states (\%\@\%) {

	my ($node2state, $frac_nodes, $node2children) = @_;

	my $max_depth = @{$frac_nodes} - 1;

	for (my $d=$max_depth; $d>= 0; $d--) { # The outermost for-loop #
	    my $sb_nodes = $frac_nodes->[$d];
	    foreach my $node (@{$sb_nodes}) { # The middle foreach-loop #
		(defined $node2state->{$node}) and next;
		my $children = $node2children->{$node};
		unless (defined $children) {
		   print STDERR "[CAUTION 'assign_node2bival_states'] The node ID '$node' has NO CHILDREN while having no state information.\n";
		   ###return 0;
		   next;
		}
		my $ct_child = @{$children};
		my %stat2ct;
		foreach my $child (@{$children}) { # The inner foreach-loop #
		    my $ch_stat = $node2state->{$child};
		    unless (defined $ch_stat) {
			print STDERR "[CAUTION 'assign_node2bival_states'] The child '$child' of the node '$node' has NO STATE assigned .\n";
			###return 0;
			next;
		    }
		    if (defined $stat2ct{$ch_stat}) {
			$stat2ct{$ch_stat}++;
		    } else {
			$stat2ct{$ch_stat} = 1;
		    }			# end of the innermost foreach-loop # 
		}			# end of the inner foreach-loop #

		my @states = keys %stat2ct;
		$node2state->{$node} = (@states>1) ? 'mixed' : $states[0];
	    }				# end of the middle foreach-loop # 

	}				# end of the outermost for-loop #

	return 1;

}		# end of the subroutine 'assign_node2bival_states' #


# assign_node2stat_weights:
#	This subroutine assigns all nodes the weight factors for multiple choices of the gap status.
#
# INPUT: (\%node2stat_wt, \@frac_nodes, \%node2children, $indx), where
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]}[0, $indx] = ($state, $weight)
#	    is for the $p th state (anything will be fine, in principle), with
#	    and $weight the fraction of the total weights of supporting child nodes (this should sum up to 1).
#	    The status signs are arranged with lowering priorities;
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d, where the top node has $d = 0);
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#
# REQUISITE: %node2stat_wt is properly initialized,
#	namely, multiple-choices of the status for the external nodes and their weights are already given.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#	The main job of this subroutine is to fill in %node2stat_wt with the value
#	[\@stat_info, ...], with ($state, $weight) = @stat_info[0,$indx], for each $node_id as a key.
#
sub assign_node2stat_weights (\%\@\%$) {

	my ($node2stat_wt, $frac_nodes, $node2children, $indx) = @_;

	my $max_depth = @{$frac_nodes} - 1;

	for (my $d=$max_depth; $d>= 0; $d--) { # The outermost for-loop #
	    my $sb_nodes = $frac_nodes->[$d];
	    foreach my $node (@{$sb_nodes}) { # The middle foreach-loop #
		(defined $node2stat_wt->{$node}) and next;
		my $children = $node2children->{$node};
		unless (defined $children) {
		   print STDERR "[ERROR 'assign_node2stat_weights'] The node ID '$node' has NO CHILDREN while having no state information.\n";
		   return 0;
		}
		my $ct_child = @{$children};
		my %stat2weight;
		foreach my $child (@{$children}) { # The inner foreach-loop #
		    my $ch_stats = $node2stat_wt->{$child};
		    unless (defined $ch_stats) {
			print STDERR "[ERROR 'assign_node2stat_weights'] The child '$child' of the node '$node' has NO STATUS INFORMATION .\n";
			return 0;
		    }
		    foreach my $stat_info (@{$ch_stats}) { # the innermost foreach-loop # 
			my ($status, $weight) = @{$stat_info}[0,1];
			if (defined $stat2weight{$status}) {
			    $stat2weight{$status} += $weight/$ct_child;
			} else {
			   $stat2weight{$status} = $weight/$ct_child;
			}
		    }			# end of the innermost foreach-loop # 
		}			# end of the inner foreach-loop #

		my @list_stats = ();
		foreach my $stat (sort {$stat2weight{$b} <=> $stat2weight{$a} }
			keys %stat2weight) { # the 2nd inner  foreach-loop #
		    my @tmp = ();
		    @tmp[0, $indx] = ($stat, $stat2weight{$stat});
		    push @list_stats, \@tmp;
		}			# end of the 2nd inner foreach-loop #

		$node2stat_wt->{$node} = \@list_stats;
	    }				# end of the middle foreach-loop # 

	}				# end of the outermost for-loop #

	return 1;

}		# end of the subroutine 'assign_node2stat_weights' #


# assign_node2minimum_changes:
#	This subroutine assigns all nodes the minimum number of changes
#	the subtree below the node must undergo when the node take each state.
#
# INPUT: (\%node2stat_wt, \@frac_nodes, \%node2children, \@states, $indx_wt, $indx_ch), where
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]}[0, $indx_wt, $indx_ch] = ($state, $weight, $min_changes)
#	    is for the $p th state (anything will be fine, in principle), with
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    and $min_changes the minimum number of changes the subtree below the node
#	    must undergo when the node take $state;
#	    The status signs are arranged with lowering priorities;
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d, where the top node has $d = 0);
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	@states is the list of all available states, or empty if we employ the
#	'local parsimony' principle.
#
# REQUISITE: %node2stat_wt is properly initialized,
#	namely, multiple-choices of the status for the external nodes,
#	as well as the weight for each state, are already given.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#	The main job of this subroutine is to fill in %node2stat_wt with the value
#	[\@stat_info, ...], with ($state, $min_changes) = @stat_info[0,$indx], for each $node_id as a key.
#
sub assign_node2minimum_changes (\%\@\%\@$$) {

	my ($node2stat_wt, $frac_nodes, $node2children, $states, $indx_wt, $indx_ch) = @_;

	my $max_depth = @{$frac_nodes} - 1;

	for (my $d=$max_depth; $d>= 0; $d--) { # The outermost for-loop #
	    my $sb_nodes = $frac_nodes->[$d];
	    foreach my $node (@{$sb_nodes}) { # The middle foreach-loop #
		my $children = $node2children->{$node};

		unless (defined $children) { # An external node #
		   if (defined $node2stat_wt->{$node}) {
			if (@{$node2stat_wt->{$node}}==1) {
			    $node2stat_wt->{$node}[0][$indx_ch] = 0;
			} else {
			    foreach my $stat_info (@{$node2stat_wt->{$node}}) {
				my $weight = $stat_info->[$indx_wt];
				unless (defined $weight) {
				   print STDERR "[ERROR 'assign_node2minimum_changes'] The EXTERNAL NODE '$node' has a state $stat_info->[0] WITHOUT THE WEIGHT assigned.\n";
				    return 0;
				}
		 		$stat_info->[$indx_ch] =  1 - $weight;
			    }

			}
			next;
		   }
		   print STDERR "[ERROR 'assign_node2minimum_changes'] The node ID '$node' has NO CHILDREN while having no state information.\n";
		   return 0;
		}

		my @states = @{$states};
		unless (@states>0) { # When employing the 'local parsimony'
		    my %state2exist;
		    foreach my $child (@{$children}) { # The 1st inner foreach-loop #
			my $ch_stats = $node2stat_wt->{$child};
			unless (defined $ch_stats) {
			   print STDERR "[ERROR 'assign_node2minimum_chages'] The child '$child' of the node '$node' has NO STATUS INFORMATION .\n";
			   return 0;
			}
			foreach my $stat_info (@{$ch_stats}) { $state2exist{$stat_info->[0]} = 1; }
		    }				# End of the 1st inner foreach-loop #
		    @states = sort keys %state2exist;
		}
		my %stat2min_chgs;
		foreach my $stat (@states) { # The innermiddle foreach-loop #
		    my $min_chgs = 0;		    
		    foreach my $child (@{$children}) { # The inner foreach-loop #
			my $ch_stats = $node2stat_wt->{$child};
			unless (defined $ch_stats) {
			   print STDERR "[ERROR 'assign_node2minimum_chages'] The child '$child' of the node '$node' has NO STATUS INFORMATION .\n";
			   return 0;
			}
			my $sbmin_chgs;
			foreach my $stat_info (@{$ch_stats}) { # the innermost foreach-loop #
			    my ($stat_ch, $sb_changes) = @{$stat_info}[0, $indx_ch];
			    ($stat_ch eq $stat) or ($sb_changes++);
			    ((defined $sbmin_chgs) and ($sbmin_chgs < $sb_changes)) or ($sbmin_chgs = $sb_changes);
			}		# end of the innermost foreach-loop #
			$min_chgs += $sbmin_chgs;
		    }				# End of the inner foreach-loop #
		    $stat2min_chgs{$stat} = $min_chgs;
		}			# End of the innermiddle foreach-loop #

		my %stat2info;
		if (defined $node2stat_wt->{$node}) {
		    foreach my $stat_info (@{$node2stat_wt->{$node}}) { # the 2nd foreach-loop #
			$stat2info{$stat_info->[0]} = $stat_info;
		    }				# end of the 2nd foreach-loop #
		}

		my @list_stats = ();
		foreach my $stat (sort {$stat2min_chgs{$a} <=> $stat2min_chgs{$b}}
			keys %stat2min_chgs) { # the 3rd inner  foreach-loop #
		    my @tmp = (defined $stat2info{$stat}) ? @{$stat2info{$stat}} : ($stat);
		    $tmp[$indx_ch] =  $stat2min_chgs{$stat};

			# A Minor Bug-Fixing on Aug 9 (Sat), 2008
			# (If there is no child nodes harboring $stat,
			# its weight factor should be zero.)
		    (defined $stat2info{$stat}) or ($tmp[$indx_wt] = 0);
		    push @list_stats, \@tmp;
		}			# end of the 3rd inner foreach-loop #

		$node2stat_wt->{$node} = \@list_stats;
	    }				# end of the middle foreach-loop # 

	}				# end of the outermost for-loop #

	return 1;

}		# end of the subroutine 'assign_node2mininum_changes' #


# map_pure_parsimony_states:
#	This subroutine determines the states of internal nodes in a TOP-DOWN manner
#	by applying a pure parsimony, and by using the output of the subroutine
#	'assign_node2minimum_changes'. 
#
# INPUT: (\%node2stat_wt, \%node2children, $$indx_ch, $topnode_id), where
#	%node2stat_wt = ($node_id => \@status, ...), with
#	@{$status[$p]}[0, $indx_ch] = ($state, $changes)
#	are for the $p th state assigned on the node, with
#	    $number the number of supporting child (or grandchild) nodes (resolved),
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    $changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	    The status signs are arranged in the order of lowering priorities;
#	%node2children = ($node_id => [IDs of child nodes], ...);
#
# OUTPUT: \@set_node2parsstat, where
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious stated mapped onto the nodes, with
#	@state_info following the same format as @{$status[$p]} of the input.
#
sub map_pure_parsimony_states (\%\%$$) {

	my ($node2stat_wt, $node2chi, $indx_ch, $topnode_id) = @_;

	my $top_status = $node2stat_wt->{$topnode_id};
	unless ((defined $top_status) and (@{$top_status}>0)) {
	    print STDERR "[ERROR 'map_pure_parsimony_states'] NO STATES assgined to the TOP NODE !!\n";
	    return undef;
	}
		## Determine the parsimonious states on the top node ##
		## ( --> Initialize the output ) ##
	##my @topnode_states = ($top_status->[0]);
	my @set_node2parsstat = ( {$topnode_id => $top_status->[0]} );
	my $top_minch = $top_status->[0][$indx_ch];
	for (my $p=1; $p<@{$top_status}; $p++) { # The 1st for-loop #
	    my $stwise_minch = $top_status->[$p][$indx_ch];
	    if ($stwise_minch < $top_minch) {
			# Reset the state set #
		##@topnode_states = ($top_status->[$p]);
		@set_node2parsstat = ( {$topnode_id => $top_status->[$p]} );
		$top_minch = $stwise_minch;
	    } elsif ($stwise_minch == $top_minch) {
		##push @topnode_states, $top_status->[$p];
		push @set_node2parsstat, {$topnode_id => $top_status->[$p]};
	    }
	}		# End of the 1st for-loop #

		## Initialize a set of nodes whose children are to be examined ##
	my @parents = ($topnode_id);
	while (@parents > 0) { # The main outer while-loop #

	    my @summ_children = ();

	    while (my $parent = shift @parents) { # The outer-middle while-loop #
		my $children = $node2chi->{$parent};
		((defined $children) and (@{$children}>0)) or next;
		push @summ_children, @{$children};

			## Determine the states of children ##
		my $ct_chi = @{$children};
		foreach my $chi (@{$children})  { # The middle foreach-loop #

			# Alternative states for the child #
		    my $chi_status = $node2stat_wt->{$chi};
		    unless ((defined $chi_status) and (@{$chi_status}>0)) {
			print STDERR "[ERROR 'map_pure_parsimony_states'] NO STATES assigned to the NODE '$chi' !!\n";
			return undef;
		    }

		     # When the state is uniquely assigned from the beginning #
		     # (It should at least apply to some external nodes) #
		    if (@{$chi_status} == 1) {
		    	my $ct_opts = @set_node2parsstat;
		    	for (my $opt=0; $opt<$ct_opts; $opt++) { # The inner for-loop #
		    	    $set_node2parsstat[$opt]->{$chi} = $chi_status->[0];
		    	}			# End of the inner for-loop #
		    	next;
		    }

		    my @newset_node2parsstat = ();

			# Options for the parsimonious mapping #
		    my $ct_opt = 0;
		    while (my $node2parsstat = shift @set_node2parsstat) { # The inner-middle while-loop #
			$ct_opt++;
			my $pa_stat = $node2parsstat->{$parent}[0];
			my $minch;
			my @chi_parsstats = ();
			foreach my $chi_statinfo (@{$chi_status}) { # The inner foreach-loop #
			    my ($chi_stat, $changes) = @{$chi_statinfo}[0, $indx_ch];
			    ($chi_stat eq $pa_stat) or ($changes++);
				# at least one change is required
				# on the branch between the child and the parent

			    if ((@chi_parsstats == 0) or ($changes < $minch)) {
				# Reset the set of parsimonious states
			        my @chi_statinfo = @{$chi_statinfo};
				@chi_parsstats = (\@chi_statinfo);
				$minch = $changes;
			    } elsif ($changes == $minch) {
				# Append the current state to the set
			        my @chi_statinfo = @{$chi_statinfo};
				push @chi_parsstats, \@chi_statinfo;
			    }
			}			# End of the inner for-loop #

			unless (@chi_parsstats>0) {
			    print STDERR "[ERROR 'map_pure_parsimony_states'] FAILED to assign parsimonious states to the node '$chi' on the $ct_opt th option of the parsimonious state maps !\n";
			    return undef;
			}
			while (my $chi_parsstat = shift @chi_parsstats) { # The inner while-loop #
			    my %new_node2parsstat = %{$node2parsstat};
			    $new_node2parsstat{$chi} = $chi_parsstat;
			    push @newset_node2parsstat, \%new_node2parsstat;
			} 			# End of the inner while-loop #

		    }			# End of the inner-middle while-loop #

			# (By now, @set_node2parsstat should be empty) #

			# Reset the set of states #
		    @set_node2parsstat = @newset_node2parsstat;

		}				# End of the middle for-loop #
	    }				# End of the outer-middle while-loop #

		# 'Generation change' #
	    @parents = @summ_children;

	}		# End of the main outer while-loop #

	return \@set_node2parsstat ;

}		# End of the subroutine 'map_pure_parsimony_states' #


# infer_stat_by_pars_bottomup:
#	This subroutine determines the states of internal nodes in a BOTTOM-UP manner
#	by applying a 'LOCAL parsimony' principle.
#	Namely, if the minimum number of required changes on the subtree is less than the second minimum
#	BY ONE OR MORE, the minimum state is employed for the node.
#	The subroutine also counts the number of children locally supporting each state.
#	The rule for the counting is:
#	(i) If the child is recorded in %reslv_node2stat, count = 1 for the supported state, = 0 for the remaining;
#	(ii) If it is NOT recorded in %reslv_node2stat, count = the count recorded in $node2stat_wt{$node}.
#	
# INPUT: (\%reslv_node2stat, \%node2stat_wt, \@frac_nodes, \%node2children), where
#	%reslv_node2stat = ($node_id => $state, ...) with
#	$state being the state assigned to the node with the local parsimony principle.
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]} = ($state, $weight, $changes, $number)
#	    is for the $p th state ($char_gap or $char_ng), with
#	    $number the number of supporting child (or grandchild) nodes (resolved),
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    $changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	    The status signs are arranged in the order of lowering priorities;
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d) , where the top node has $d = 0;
#	%node2children = ($node_id => [IDs of child nodes], ...);
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub infer_stat_by_pars_bottomup (\%\%\@\%) {

	my ($reslv_node2stat, $node2stat_wt, $frac_nodes, $node2children) = @_;

	(@{$frac_nodes}>0) or (return 1);
	my $max_depth = @{$frac_nodes} -1;

	for (my $k=$max_depth; $k>= 0; $k--) { # The outermost for-loop #
	    my $sb_nodes = $frac_nodes->[$k];

	    foreach my $node (@{$sb_nodes}) { # The middle foreach-loop #
		my $info_states = $node2stat_wt->{$node};
		unless (defined $info_states) {
		    print STDERR "[ERROR 'infer_stat_by_pars_bottomup'] NO INFORMATION on the STATES found for the node '$node'!\n";
		    return 0;
		}

			## First count the number of supporting child nodes ##
		my $children = $node2children->{$node};

		unless (defined $children) {
			# An external node #

		    if (@{$info_states}==1) {
			$info_states->[0][3] = 1;
			$reslv_node2stat->{$node} = $info_states->[0][0]; # Determine the state #
		    } else {
			foreach my $info_stat (@{$info_states}) { # The 1st inner foreach-loop #
			    $info_stat->[3] = $info_stat->[1];
			}			# End of the 1st inner foreach-loop #
		    }
		    next;
		}

		     # An internal node #

		my %stat2ct_support;
		foreach my $info_stat (@{$info_states}) { $stat2ct_support{$info_stat->[0]} = 0; }
		foreach my $child (@{$children}) { # The 3rd inner foreach-loop #
		    my $reslv_stat = $reslv_node2stat->{$child};
		    if (defined $reslv_stat) {
			$stat2ct_support{$reslv_stat}++; 
		    } else {
			my $info_ch_states = $node2stat_wt->{$child};
			unless (defined $info_ch_states) {
			    print STDERR "[ERROR 'infer_stat_by_pars_bottomup'] NO INFORMATION on the STATES found for the child '$child' of the node '$node'!\n";
			    return 0;
			}
			foreach my $info_stat (@{$info_ch_states}) { # The innermost foreach-loop #
			    $stat2ct_support{$info_stat->[0]} += $info_stat->[3];
			}				# End of the innermost foreach-loop #
		    }
		}			# End of the 3rd inner foreach-loop #
		foreach my $info_stat (@{$info_states}) { $info_stat->[3] = $stat2ct_support{$info_stat->[0]}; }

			## Then resolve the state for the node ##

		if (@{$info_states}==1) { # Trivial case #
		    $reslv_node2stat->{$node} = $info_states->[0][0];
		    next;
		}

			# Sort in terms of the number of changes #
		my @std_info_states = sort {$a->[2] <=> $b->[2] } @{$info_states};
		my ($min, $second) = @std_info_states[0,1];

			# Local parsimony rule #
		#($min->[2] >= $second->[2] +1) and ($reslv_node2stat->{$node} = $min->[0]);
		#($min->[2] <= $second->[2] -1) and ($reslv_node2stat->{$node} = $min->[0]);
		($min->[2] < $second->[2] -1) and ($reslv_node2stat->{$node} = $min->[0]);

	    }				# End of the middle foreach-loop #
	}			# End of the outermost for-loop #

	return 1;

}		# end of the subroutine 'infer_stat_by_pars_bottomup' #


=old version
# infer_stat_by_pars_topdown:
#	This subroutine determines the states of internal nodes in a TOP-DOWN manner
#	by applying a usual parsimony principle.
#	Namely, beginning with a state for the top-node, the subroutine
#	assigns to each descendant node the state that strictly minimizes the number of changes.
#	The subroutine stops descending the tree if it encounters a node with ambiguity.
#	
# INPUT: (\%out_reslv_node2stat, \%in_reslv_node2stat, \%node2stat_wt,
#		\%node2children, $top_node, $top_state), where
#	%out/in_reslv_node2stat = ($node_id => $state, ...) with
#	$state being the state assigned to the node with the local parsimony principle.
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]} = ($state, $weight, $changes, $number)
#	    is for the $p th state ($char_gap or $char_ng), with
#	    $number the number of supporting child (or grandchild) nodes (resolved),
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    $changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	    The status signs are arranged in the order of lowering priorities;
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	$top_node and $top_state are the top node for the subtree in consideration and 
#	a state assigned to it.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub infer_stat_by_pars_topdown (\%\%\%\%$$) {

	my ($out_reslv_node2stat, $in_reslv_node2stat, $node2stat_wt,
		$node2children, $top_node, $top_state) = @_;

	$out_reslv_node2stat->{$top_node} = $top_state;

	my @sb_nodes = ($top_node);

	while (@sb_nodes>0) { # The outermost while-loop #
	    my @next_gen = ();

	    foreach my $node (@sb_nodes) { # The middle foreach-loop #
		my $children = $node2children->{$node};
		(defined $children) or next;
		push @next_gen, @{$children};

		my $fl_par_reslv = 0; # =1 if the parent is already resolved,
				      # =0 otherwise.
		my $par_stat = $in_reslv_node2stat->{$node};
		(defined $par_stat) ? ($fl_par_reslv=1) : ($par_stat = $out_reslv_node2stat->{$node});
		(defined $par_stat) or next;
		my $reslv_node2stat = $fl_par_reslv ? $in_reslv_node2stat : $out_reslv_node2stat ;

		foreach my $child (@{$children}) { # The inner foreach-loop #
		    ((defined $in_reslv_node2stat->{$child}) or
			(defined $out_reslv_node2stat->{$child})) and next;

		    my $info_ch_states = $node2stat_wt->{$child};
		    unless (defined $info_ch_states) {
			print STDERR "[ERROR 'infer_stat_by_pars_topdown'] The child '$child' of the node '$node' has NO INFORMATION on the status!\n";
			return 0;
		    }
		    if (@{$info_ch_states}<2) {
			$reslv_node2stat->{$child} = $info_ch_states->[0][0];
			next;
		    }
		    my @list_ct_changes; # element = [state, #{changes}]
		    foreach my $info_stat (@{$info_ch_states}) { # the innermost foreach-loop #
			my ($stat, $changes) = @{$info_stat}[0, 2];
			($stat eq $par_stat) or ($changes++);
			push @list_ct_changes, [$stat, $changes];
		    }					# end of the innermost foreach-loop #

			# Parsimony principle #
		    my @std_list_changes = sort {$a->[1] <=> $b->[1]} @list_ct_changes;
		    my ($min, $second) = @std_list_changes[0,1];
		    ($min->[1] < $second->[1]) and ($reslv_node2stat->{$child} = $min->[0]);
		    
		}			# end of the inner foreach-loop #

	    }				# end of the middle foreach-loop #

	    @sb_nodes = @next_gen; # Renew the subset of nodes #
	}		# end of the outermost while-loop #

	return 1;

}		# end of the subroutine 'infer_stat_by_pars_topdown' #
=cut


# infer_stat_by_pars_topdown:
#	This subroutine determines the states of internal nodes in a TOP-DOWN manner
#	by applying a usual parsimony principle.
#	Namely, beginning with a state for the top-node, the subroutine
#	assigns to each descendant node the state that strictly minimizes the number of changes.
#	The subroutine stops descending the tree if it encounters a node with ambiguity.
#	
# INPUT: (\%in_alt_node2stat, \%node2stat_wt,
#		\%node2children, $top_node, $top_state), where
#	%{$in_alt_node2stat} = ($node_id => $state, ...),
#	with $state being the state assigned to the node with the local parsimony principle.
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]} = ($state, $weight, $changes, $number)
#	    is for the $p th state ($char_gap or $char_ng), with
#	    $number the number of supporting child (or grandchild) nodes (resolved),
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    $changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	    The status signs are arranged in the order of lowering priorities;
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	$top_node and $top_state are the top node for the subtree in consideration and 
#	a state assigned to it.
#
# OUTPUT: \@outset_alt_node2stat,  where
#	%{$outset_alt_node2stat[$opt]} = ($node_id => $state, ...) is for the $opt th mapping of the gap states,
#	with $state being the state assigned to the node with the local parsimony principle.
#
sub infer_stat_by_pars_topdown (\%\%\%$$) {

	my ($in_alt_node2stat, $node2stat_wt,
		$node2children, $top_node, $top_state) = @_;


	my %cp_node2stat = %{$in_alt_node2stat};
	$cp_node2stat{$top_node} = $top_state;

	##$out_reslv_node2stat->{$top_node} = $top_state;
	my @sb_nodes = ($top_node);

	my @outset_alt_node2stat = (\%cp_node2stat);

    while (@sb_nodes>0) { # The outermost while-loop #
	my @next_gen = ();

	foreach my $node (@sb_nodes) { # The middle foreach-loop #
	    my $children = $node2children->{$node};
	    (defined $children) or next;
	    push @next_gen, @{$children};

	    my @newset_alt_node2stat = ();
	    foreach my $alt_node2stat (@outset_alt_node2stat) { # The inner-middle foreach-loop #
		my $par_stat = $alt_node2stat->{$node};
		(defined $par_stat) or next;
		##my $reslv_node2stat = $fl_par_reslv ? $in_reslv_node2stat : $out_reslv_node2stat ;

		my @sbset_alt_node2stat = ($alt_node2stat);
		foreach my $child (@{$children}) { # The inner foreach-loop #
		    (defined $alt_node2stat->{$child}) and next;

		    my $info_ch_states = $node2stat_wt->{$child};
		    unless (defined $info_ch_states) {
			print STDERR "[ERROR 'infer_stat_by_pars_topdown'] The child '$child' of the node '$node' has NO INFORMATION on the status!\n";
			return 0;
		    }
		    if (@{$info_ch_states}<2) {
			foreach my $alt_node2stat2 (@sbset_alt_node2stat) { # the 1st innermost foeach-loop #
			    $alt_node2stat2->{$child} = $info_ch_states->[0][0];
			}		# end of the 1st innermost foreach-loop #
			next;
		    }
		    my @list_ct_changes; # element = [state, #{changes}]
		    foreach my $info_stat (@{$info_ch_states}) { # the 2nd innermost foreach-loop #
			my ($stat, $changes) = @{$info_stat}[0, 2];
			($stat eq $par_stat) or ($changes++);
			push @list_ct_changes, [$stat, $changes];
		    }					# end of the 2nd innermost foreach-loop #

			# Parsimony principle #
		    my @std_list_changes = sort {$a->[1] <=> $b->[1]} @list_ct_changes;
		    my ($min, $second) = @std_list_changes[0,1];
		    if ($min->[1] < $second->[1]) {
				# When a state is uniquely determined #
			foreach my $alt_node2stat2 (@sbset_alt_node2stat) { # Tne 3rd innermost foreach-loop #
			    $alt_node2stat2->{$child} = $min->[0];
			}	 # end of the 3rd innermost foreach-loop #
			next;
		    }

			# Pick all tie situations #
		    my @new_sbset_alt_node2stat = ();

		    foreach my $stat_info (@std_list_changes) { # The 4th-innermost foreach-loop #
			($stat_info->[1] > $min->[1]) and last; 
			foreach my $alt_node2stat2 (@sbset_alt_node2stat) { # The exta innermost foreach-loop #
			    my %cp_node2stat2 = %{$alt_node2stat2};
			    $cp_node2stat2{$child} = $stat_info->[0];
			    push @new_sbset_alt_node2stat, \%cp_node2stat2;
			}			# end of the extra innermost foreach-loop #
		    }		# end of the 4th innermost foreach-loop #

		    @sbset_alt_node2stat = @new_sbset_alt_node2stat; 
		}			# end of the inner foreach-loop #

		push @newset_alt_node2stat, @sbset_alt_node2stat;

	    }				# end of the inner-middle foreach-loop #

	     @outset_alt_node2stat = @newset_alt_node2stat;
			# Renew the set of gap state mappings #

	}			# end of the middle foreach-loop #

	@sb_nodes = @next_gen; # Renew the subset of nodes #
    }		# end of the outermost while-loop #

	return \@outset_alt_node2stat;

}		# end of the subroutine 'infer_stat_by_pars_topdown' #


# infer_gapstat_by_pars:
#	This subroutine infers gap status of the internal nodes
#	by means of a parsimony principle.
#	If a tie situation is conceived, we prioritize the local parsimony.
#	This means we employ the node status that would minimize
#	the number of changes IN A LOCAL REGION (on the tree).
#	(e.g. majority rule among the child nodes --> children + sibling, etc...).
#
#	RULES FOR RESOLVING THE STATE OF AN INTERNAL NODE:
#	We employ a 'local parsimony principle'.
#	Namely, if there is a tie situation in terms of the global parsimony criterion,
#	we employ the state that would minimize the changes under the node in question.
#	We will implement this criterion via the following procedures:
#	(i) On each internal node, calculate the minimum number of required changes on the subtree below the node when it takes each state seen in the subtree, using a dynamic programming;
#	(ii) First determine the states locally in a bottom-up manner. Namely, if the minimum number of required changes on the subtree exceeds the second minimum BY ONE OR MORE, employ the minimum state for the node;
#	(iii) Then determine the remaining node states in a top-down manner.
#	 As for the top-node, if the minimum number of necessary changes on the whole tree is strictly less than the second minimum, employ the state;
#	(iii) Determine the states on the remaining nodes that would give the minimum changes under a give state of the top node.
#	If there is a tie situation, we will choose the state that gives the STRICTLY minimum number of changes on the subtree.
#	This rule effectively implements the criterion "If all children support the same state, employ the state";
#	(iv) If the above three procedures cannot determine the state, mark the node as 'indecisive'.
#	Then, we will see the number of children supporting the state, as well as the fraction of descendants supporting the state, as an aid to force the decision;
#
# INPUT: (\%node2stat_wt, \@frac_nodes, \%node2children, \%node2parent, \%name2id), where
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]} = ($state, $weight, $changes, $number)
#	    is for the $p th state ($char_gap or $char_ng), with
#	    $number the number of supporting child (or grandchild) nodes (resolved),
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    $changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	    The status signs are arranged in the order of lowering priorities;
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d);
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	%name2id = ($node_name => $node_id, ...) for the external and top nodes.
#
# REQUISITE: %node2stat_wt is properly initialized,
#	namely, multiple-choices of the status, their weight factors
#	and 'local' minimum numbers of required changes are already given for ALL nodes.
#
# OUTPUT: \@set_alt_node2stat, where
#	%{$set_alt_node2stat[$opt]} = ($node_id => $status, ...)  is for the $opt th alternative mapping of the gap states,
#	where $status is the status assigned to the node with the parsimony principle.
#
sub infer_gapstat_by_pars (\%\@\%\%\%) {

	my ($node2stat_wt, $frac_nodes, $node2children, $node2parent, $name2id) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	unless (defined $topnode_id) {
	    print STDERR "[ERROR 'infer_gapstat_by_pars'] NO ID associated to '$TOP_NODE'!\n";
	    ##return (undef, undef);
	    return undef;
	}
	##(@{$frac_nodes}>0) or (return ({}, {}) );
	(@{$frac_nodes}>0) or (return []);
	my $max_depth = @{$frac_nodes} -1;
	my %reslv_node2stat;
	my @set_alt_node2stat;

		# First determine the state locally (in a bottom-up manner),
		# and fill in the number of supporting children.

	infer_stat_by_pars_bottomup (%reslv_node2stat, %{$node2stat_wt}, @{$frac_nodes}, %{$node2children});

		# Then determine the state in a top_down manner.

	my $info_top_states = $node2stat_wt->{$topnode_id};
	##my @inset = (\%reslv_node2stat);	
	if (@{$info_top_states}<2) {
	    my $top_state = $info_top_states->[0][0];

	    my $outset_alt_node2stat =
		infer_stat_by_pars_topdown (%reslv_node2stat,
		%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
		##infer_stat_by_pars_topdown (%reslv_node2stat, %reslv_node2stat,
		##%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
	    @set_alt_node2stat = @{$outset_alt_node2stat};

	} else {
	    my @std_info_top_states = sort { $a->[2] <=> $b->[2] } @{$info_top_states};
	    my ($min, $second) = @std_info_top_states[0,1];

	    if ($min->[2] < $second->[2]) {
			# When the top node has a uniquely determined state
			# (in terms of the parsimony principle) #

		my $top_state = $min->[0];
		my $outset_alt_node2stat =
		infer_stat_by_pars_topdown (%reslv_node2stat,
			%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
		##infer_stat_by_pars_topdown (%reslv_node2stat, %reslv_node2stat,
		##	%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
	        @set_alt_node2stat = @{$outset_alt_node2stat};

	    } else {
			# When the top node does NOT have a uniquely determined state #

		foreach my $info_stat (@std_info_top_states) { # The foreach-loop #
		    my $top_state = $info_stat->[0];
		    my $outset_alt_node2stat =	    
		    infer_stat_by_pars_topdown (%reslv_node2stat,
		    	%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
		    push @set_alt_node2stat, @{$outset_alt_node2stat};
		    ##my %tmp;
		    ##infer_stat_by_pars_topdown (%tmp, %reslv_node2stat,
		    ##	%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
		    ##$alt_node2stat{$top_state} = \%tmp;
		}				# End of the foreach-loop #
	    }
	}

	##return (\%reslv_node2stat, \%alt_node2stat);
	return \@set_alt_node2stat;
}		# end of the subroutine 'infer_gapstat_by_pars' #


=old version
# infer_gapstat_by_pars_OLD:
#	This subroutine infers gap status of the internal nodes
#	by means of a parsimony principle.
#	If a tie situation is conceived, we prioritize the local parsimony.
#	This means we employ the node status that would minimize
#	the number of changes IN A LOCAL REGION (on the tree).
#	(e.g. majority rule among the child nodes --> children + sibling, etc...).
#
#	RULES FOR RESOLVING THE STATE OF AN INTERNAL NODE:
#	We employ a 'local parsimony principle'.
#	Namely, if there is a tie situation in terms of the global parsimony criterion,
#	we employ the state that would minimize the changes under the node in question.
#	We will implement this criterion via the following procedures:
#	(i) On each internal node, calculate the minimum number of required changes on the subtree below the node when it takes each state seen in the subtree, using a dynamic programming;
#	(ii) First determine the states locally in a bottom-up manner. Namely, if the minimum number of required changes on the subtree exceeds the second minimum BY ONE OR MORE, employ the minimum state for the node;
#	(iii) Then determine the remaining node states in a top-down manner.
#	 As for the top-node, if the minimum number of necessary changes on the whole tree is strictly less than the second minimum, employ the state;
#	(iii) Determine the states on the remaining nodes that would give the minimum changes under a give state of the top node.
#	If there is a tie situation, we will choose the state that gives the STRICTLY minimum number of changes on the subtree.
#	This rule effectively implements the criterion "If all children support the same state, employ the state";
#	(iv) If the above three procedures cannot determine the state, mark the node as 'indecisive'.
#	Then, we will see the number of children supporting the state, as well as the fraction of descendants supporting the state, as an aid to force the decision;
#
# INPUT: (\%node2stat_wt, \@frac_nodes, \%node2children, \%node2parent, \%name2id), where
#	%node2stat_wt = ($node_id => \@status, ...), with 
#	    @{$status[$p]} = ($state, $weight, $changes, $number)
#	    is for the $p th state ($char_gap or $char_ng), with
#	    $number the number of supporting child (or grandchild) nodes (resolved),
#	    $weight the fraction of the total weights of supporting child nodes (this should sum up to 1),
#	    $changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	    The status signs are arranged in the order of lowering priorities;
#	@{$frac_nodes[$d]} = (IDs of the nodes with depth $d);
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	%name2id = ($node_name => $node_id, ...) for the external and top nodes.
#
# REQUISITE: %node2stat_wt is properly initialized,
#	namely, multiple-choices of the status, their weight factors
#	and 'local' minimum numbers of required changes are already given for ALL nodes.
#
# OUTPUT: (\%reslv_node2stat, \%alt_node2stat), where
#	%reslv_node2stat = ($node_id => $status, ...) where
#	$status is the status assigned to the node with the parsimony principle.
#	%alt_node2stat = ($topnode_state => {$node_id => $status, ...}, ...) if the top node does NOT have a fixed state,
#			= empty if it does.
#	
#
sub infer_gapstat_by_pars_OLD (\%\@\%\%\%) {

	my ($node2stat_wt, $frac_nodes, $node2children, $node2parent, $name2id) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	unless (defined $topnode_id) {
	    print STDERR "[ERROR 'infer_gapstat_by_pars'] NO ID associated to '$TOP_NODE'!\n";
	    return (undef, undef);
	}
	(@{$frac_nodes}>0) or (return ({}, {}) );
	my $max_depth = @{$frac_nodes} -1;
	my %reslv_node2stat;
	my %alt_node2stat;

		# First determine the state locally (in a bottom-up manner),
		# and fill in the number of supporting children.

	infer_stat_by_pars_bottomup (%reslv_node2stat, %{$node2stat_wt}, @{$frac_nodes}, %{$node2children});

		# Then determine the state in a top_down manner.

	my $info_top_states = $node2stat_wt->{$topnode_id};
	if (@{$info_top_states}<2) {
	    my $top_state = $info_top_states->[0][0];
	
	    infer_stat_by_pars_topdown (%reslv_node2stat, %reslv_node2stat,
		%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
	} else {
	    my @std_info_top_states = sort { $a->[2] <=> $b->[2] } @{$info_top_states};
	    my ($min, $second) = @std_info_top_states[0,1];

	    if ($min->[2] < $second->[2]) {
			# When the top node has a uniquely determined state
			# (in terms of the parsimony principle) #

		my $top_state = $min->[0];
		infer_stat_by_pars_topdown (%reslv_node2stat, %reslv_node2stat,
			%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
	    } else {
			# When the top node does NOT have a uniquely determined state #

		foreach my $info_stat (@std_info_top_states) { # The foreach-loop #
		    my $top_state = $info_stat->[0];
		    my %tmp;
		    infer_stat_by_pars_topdown (%tmp, %reslv_node2stat,
			%{$node2stat_wt}, %{$node2children}, $topnode_id, $top_state);
		    $alt_node2stat{$top_state} = \%tmp;
		}				# End of the foreach-loop #
	    }
	}

	return (\%reslv_node2stat, \%alt_node2stat);

}		# end of the subroutine 'infer_gapstat_by_pars_OLD' #
=cut


# map_indels_on_tree:
#	This subroutine maps insertions/deletions onto the branches
#	based on the mapping of gap states on the nodes.
#
# INPUT: (\%reslv_node2stat \%alt_node2stat, \%node2stat_wt, \%node2children,
#		\%name2id,  $char_gap, $char_ng), where
#	%reslv_node2stat = ($node_id => $status, ...) for the nodes whose
#	status is UNIQUELY resolved with a parsimony principle;
#	%alt_node2stat = ($top_state => {$node_id => $status, ...}, ...) for the nodes
#	whose status is resolved with a parsimony principle GIVEN THE TOP NODE STATE;
#	%node2stat_wt = ($node_id => \@status, ...) is for the $k th indel block, with
#	@{$status[$p]} = ($state, $weight, $change, $number), where
#	$state = $char_gap or $char_ng, depending on whether the node was occupied by gaps or not,
#	$changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	$number is the number of child nodes (or grandchildren) that support the status (recorded in %gapstat_on_nodes),
#	$weight is the total fraction of the weights supporting the status;
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%name2id = ($node_name => $node_id, ...) for the external and top nodes.
#
# OUTPUT: \@options_br2indels, where
#	%{$options_br2indels[$option]} = ($node_id => [[$inferred_change, $weight_up, $weight_down], ...], ...)
#	is for the $option th option, with
#	$node_id is the node immediately below the branch to which the change is attributed,
#	$inferred_change = I for insertion / D for deletion,
#	$weight_up/down = the fraction of weights that supports the state on the up/downstream node.
#
sub map_indels_on_tree (\%\%\%\%\%$$) {

	my ($reslv_node2stat, $alt_node2stat, $node2stat_wt, $node2children,
		 $name2id, $char_gap, $char_ng) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	unless (defined $topnode_id) {
	    print STDERR "[ERROR 'map_indels_on_tree'] NO ID assigned to '$TOP_NODE'!\n";
	     return undef;
	}

	my @std_top_states = ();

	my $fl_reslv = (defined $reslv_node2stat->{$topnode_id}) ? 1 : 0;
	if ($fl_reslv) {
	    push @std_top_states, $reslv_node2stat->{$topnode_id};
	} else {
	    my $info_top_states = $node2stat_wt->{$topnode_id};
	    foreach my $stat_info (sort {$b->[1] <=> $a->[1]} @{$info_top_states}) {
		(defined $alt_node2stat->{$stat_info->[0]}) or next;
		push @std_top_states, $stat_info->[0];
	    }
	}

	my @options_br2indels = ();
	foreach my $top_stat (@std_top_states) { # The outermost foreach-loop #
	    my $aux_node2stat;
	    $fl_reslv  or ($aux_node2stat = $alt_node2stat->{$top_stat});

	    my %br2indels; # (downstream_node => [[change, weight_up, weight_down]], ...)
	    my @sb_nodes = ($topnode_id);
	    while (@sb_nodes>0) { # The outer while-loop #
		my @nextgen = ();

		foreach my $up_node (@sb_nodes) { # The middle foreach-loop #
		    my $children = $node2children->{$up_node};
		    (defined $children) or next; # NO branches are below an external node
		    push @nextgen, @{$children};

		    my %upstat2wt;
		    foreach my $stat_info (@{$node2stat_wt->{$up_node}}) { $upstat2wt{$stat_info->[0]} = $stat_info->[1]; }

		    my $up_stat = $reslv_node2stat->{$up_node};
		    ((defined $up_stat) or $fl_reslv) or ($up_stat = $aux_node2stat->{$up_node});
		    my @up_states = (defined $up_stat) ?  ($up_stat) :
				sort {$upstat2wt{$b} <=> $upstat2wt{$a}} keys %upstat2wt ;

		    foreach my $dw_node (@{$children}) { # The innermiddle foreach-loop #
			my %dwstat2wt;
			foreach my $stat_info (@{$node2stat_wt->{$dw_node}}) { $dwstat2wt{$stat_info->[0]} = $stat_info->[1]; }
			my $dw_stat = $reslv_node2stat->{$dw_node};
			((defined $dw_stat) or $fl_reslv) or ($dw_stat = $aux_node2stat->{$dw_node});
			my @dw_states = (defined $dw_stat) ?  ($dw_stat) :
				sort {$dwstat2wt{$b} <=> $dwstat2wt{$a}} keys %dwstat2wt ;

			my @list_indels = ();

			foreach my $upstat (@up_states) { # the inner foreach-loop #
			    my $up_weight = $upstat2wt{$upstat};
			    foreach my $dwstat (@dw_states) { # the innermost foreach-loop #
				($upstat eq $dwstat) and next; # All we are interested in are changes
				my $ch_label = ($upstat eq $char_gap) ? 'I' : 'D';
				my $dw_weight = $dwstat2wt{$dwstat};
				push @list_indels, [$ch_label, $up_weight, $dw_weight];
			    }				# end of the innermost foreach-loop #
			}			# end of the inner foreach-loop #

			(@list_indels>0) and ($br2indels{$dw_node} = \@list_indels);
		    }				# End of the innermiddle foreach-loop #

		}			# End of the middle foreach-loop #

		@sb_nodes = @nextgen; # Renew the subset #
	    }			# End of the outer while-loop #

	    push @options_br2indels, \%br2indels;

	}				# End of the outermost foreach-loop #

	return \@options_br2indels;

}		# end of the subroutine 'map_indels_on_tree' #


# br_map_indels_on_tree:
#	This subroutine maps insertions/deletions onto the branches
#	based on the mapping of gap states on the nodes.
#	We REQUIRE that the input mapping of gap states on the nodes
#	is completely resolved.
#
# INPUT: (\%reslv_node2stat, \%node2stat_wt, \%node2children,
#		\%name2id,  $char_gap, $char_ng), where
#	%reslv_node2stat = ($node_id => $status, ...) for the nodes whose
#	status is UNIQUELY resolved with a parsimony principle;
#	%node2stat_wt = ($node_id => \@status, ...) is for the $k th indel block, with
#	@{$status[$p]} = ($state, $weight, $change, $number), where
#	$state = $char_gap or $char_ng, depending on whether the node was occupied by gaps or not,
#	$changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	$number is the number of child nodes (or grandchildren) that support the status (recorded in %gapstat_on_nodes),
#	$weight is the total fraction of the weights supporting the status;
#	%node2children = ($node_id => [IDs of child nodes], ...);
#	%name2id = ($node_name => $node_id, ...) for the external and top nodes.
#
# OUTPUT: \%br2indels, where
#	%br2indels = ($node_id => [$inferred_change, $weight_up, $weight_down], ...)
#	is the map of indels onto the branches, with
#	$node_id the node immediately below the branch to which the change is attributed,
#	$inferred_change = I for insertion / D for deletion,
#	$weight_up/down = the fraction of weights that supports the state on the up/downstream node.
#
sub br_map_indels_on_tree (\%\%\%\%$$) {

	my ($reslv_node2stat, $node2stat_wt, $node2children,
		 $name2id, $char_gap, $char_ng) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	unless (defined $topnode_id) {
	    print STDERR "[ERROR 'br_map_indels_on_tree'] NO ID assigned to '$TOP_NODE'!\n";
	     return undef;
	}

	my @std_top_states = ();

	unless (defined $reslv_node2stat->{$topnode_id}) {
	    print STDERR "[ERROR 'br_map_indels_on_tree'] The gap state on '$TOP_NODE' is UNRESOLVED !\n";
	    return undef;
	}

	my %br2indels; # (downstream_node => [[change, weight_up, weight_down]], ...)
	my @sb_nodes = ($topnode_id);
	while (@sb_nodes>0) { # The outer while-loop #
		my @nextgen = ();

		foreach my $up_node (@sb_nodes) { # The middle foreach-loop #
		    my $children = $node2children->{$up_node};
		    (defined $children) or next; # NO branches are below an external node
		    push @nextgen, @{$children};

		    my %upstat2wt;
		    foreach my $stat_info (@{$node2stat_wt->{$up_node}}) { $upstat2wt{$stat_info->[0]} = $stat_info->[1]; }

		    my $up_stat = $reslv_node2stat->{$up_node};
		    unless (defined $up_stat) {
			print STDERR "[ERROR 'br_map_indels_on_tree'] The upstream node '$up_node' has NO RESOLVED gap state!\n";
			return undef;
		    }

		    foreach my $dw_node (@{$children}) { # The innermiddle foreach-loop #
			my %dwstat2wt;
			foreach my $stat_info (@{$node2stat_wt->{$dw_node}}) { $dwstat2wt{$stat_info->[0]} = $stat_info->[1]; }
			my $dw_stat = $reslv_node2stat->{$dw_node};
			unless (defined $dw_stat) {
			    print STDERR "[ERROR 'br_map_indels_on_tree'] The downstream node '$dw_node' has NO RESOLVED gap state!\n";
			    return undef;
			}

			   # We are only interested in changes #
			unless ($up_stat eq $dw_stat) {
			    my $ch_label = ($up_stat eq $char_gap) ? 'I' : 'D';
			    my $up_weight = $upstat2wt{$up_stat};
			    my $dw_weight = $dwstat2wt{$dw_stat};
			    $br2indels{$dw_node} = [$ch_label, $up_weight, $dw_weight];
			}
		    }		# End of the innermiddle foreach-loop #
		}		# End of the middle foreach-loop #

		@sb_nodes = @nextgen; # Renew the subset #

	}			# End of the outer while-loop #

	return \%br2indels;

}		# end of the subroutine 'br_map_indels_on_tree' #



# bch_map_indels_on_tree:
#	This subroutine maps a bunch of insertions/deletions onto branches of
#	a phylogenetic tree based on a parsimony principle.
#	It also outputs the associated gap status on the internal nodes
#	(as well as on the external ones).
#
# INPUT: (\@list_indels, \@seq_names, \%node2children, \%node2parent, \%name2id, $char_gap, $char_ng), where
#	@{$list_indels[$k]} = ($blstart_frame, $blend_frame, \@blstart_seq, \@blend_seq,
#	\@pattern, @sbalign) for the $k th indel unit, with
#	$blstart/end_frame is the starting/ending coordinate of the unit in terms of the alignment frame,
#	$blstart/blend_seq[$m] = the coordinate of the residue of the $m th sequence that starts/ends the block,
#	@pattern = (the sequence numbers to which the gap is assigned),
#	$sbalign[$m] = the subsequence of the $m th sequence embedded in the subalignment;
#	$seq_names[$m] is the name of the $m th sequence;
#	%node2children = ($node_id => [ a list of child IDs], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	%name2id = ($node_name => $node_id, ...)
#	for the external nodes (i.e. sequences) plus the top node;
#	$char_gap indicates that the sequence posesses the gap, and $char_ng indicates the opposite.
#
# OUTPUT: [ OLD: (\@mapped_indels, \@reslv_gapstat_on_nodes, \@alt_gapstat_on_nodes, \@gapstat_wt_on_nodes ), where]
# OUTPUT: (\@mapped_indels, \@alt_gapstat_on_nodes, \@gapstat_wt_on_nodes ), where]
# { OLD:
#	%{$mapped_indels[$k]->[$option]} = ($node_id => [[$inferred_change, $weight_up, $weight_down], ...], ...)
# }
#	%{$mapped_indels[$k]->[$option]} = ($node_id => [$inferred_change, $weight_up, $weight_down], ...)
#	is the $option th option for the $k th indel block, with
#	$node_id the node immediately below the branch to which the change is attributed,
#	$inferred_change = 'I' for insertion / 'D' for deletion,
#	$weight_up/down = the fraction of weights that supports the state on the up/downstream node;
# {OLD DESCRIPTIONS:
#	%{$reslv_gapstat_on_nodes[$k]} = ($node_id => $status, ...) for the nodes whose
#	status is UNIQUELY resolved with a parsimony principle;
#	%{$alt_gapstat_on_nodes[$k]} = ($top_state => {$node_id => $status, ...}, ...) for the nodes
#	whose status is resolved with a parsimony principle GIVEN THE TOP NODE STATE ;
#  }
#	%{$alt_gapstat_on_nodes[$k]->[$opt]} = ($node_id => $status, ...) for the $opt th alternative map for the $k th unit;
#	%{$gapstat_wt_on_nodes[$k]} = ($node_id => \@status, ...) is for the $k th indel block, with
#	@{$status[$p]} = ($state, $weight, $change, $number), where
#	$state = $char_gap or $char_ng, depending on whether the node was occupied by gaps or not,
#	$changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	$number is the number of child nodes (or grandchildren) that support the status (recorded in %gapstat_on_nodes),
#	$weight is the total fraction of the weights supporting the status.
#
sub bch_map_indels_on_tree (\@\@\%\%\%$$) {

	my ($list_indels, $seq_names, $node2children, $node2parent, $name2id,
		$char_gap, $char_ng) = @_;

	my @states = ($char_gap, $char_ng);
	my $ct_indels = @{$list_indels};
	my $ct_seqs = @{$seq_names};
	(($ct_indels>0) and ($ct_seqs>0)) or (return ([], [], []));

	my @seqids_aln = (0..$ct_seqs-1);
	my @mapped_indels = my @reslv_gapstat_on_nodes = ();
	my @alt_gapstat_on_nodes = my @gapstat_wt_on_nodes = ();

	my $frac_nodes = fractionate_by_ndepth (%{$node2children}, %{$name2id});

	for (my $k=0; $k<$ct_indels; $k++) { # The outermost for-loop #

		# Initialize the phylogenetic map #

	    my %node2stat_wt; # (node_ID => \@status, ...) where
			   # @{$status[$p]} = ($state, $weight, $change, $number)

	    my $gapped = $list_indels->[$k][4];
	    (undef, undef, my $nongapped) = diff_nmarray (@{$gapped}, @seqids_aln);

	    foreach my $m (@{$gapped}) { # the 1st inner foreach-loop #
		my $seqname = $seq_names->[$m];
		my $nodeid = $name2id->{$seqname};
		$node2stat_wt{$nodeid} = [[$char_gap, 1, 0, 1]];
	    }				# end of the 1st inner foreach-loop #
	    foreach my $m (@{$nongapped}) { # the 2nd inner foreach-loop # 
		my $seqname = $seq_names->[$m];
		my $nodeid = $name2id->{$seqname};
		$node2stat_wt{$nodeid} = [[$char_ng, 1, 0, 1]];
	    }				# end of the 2nd inner foreach-loop #

	    assign_node2stat_weights (%node2stat_wt, @{$frac_nodes}, %{$node2children}, 1);
	    my @dummy_states = ();
	    assign_node2minimum_changes (%node2stat_wt, @{$frac_nodes}, %{$node2children}, @dummy_states, 1, 2);

	    ##my ($reslv_node2stat, $alt_node2stat)
	    my $set_alt_node2stat
		= infer_gapstat_by_pars (%node2stat_wt, @{$frac_nodes}, %{$node2children}, %{$node2parent}, %{$name2id});

	    my @options_br2indels = ();
	    foreach my $alt_node2stat (@{$set_alt_node2stat}) {
		my $br2indels = br_map_indels_on_tree (%{$alt_node2stat}, %node2stat_wt, %{$node2children}, %{$name2id}, $char_gap, $char_ng);
		push @options_br2indels, $br2indels;
	    }
	    ##my $options_br2indels
		##= map_indels_on_tree (%{$reslv_node2stat}, %{$alt_node2stat}, %node2stat_wt, %{$node2children}, %{$name2id}, $char_gap, $char_ng);

	    push @gapstat_wt_on_nodes, \%node2stat_wt;
	    ##push @reslv_gapstat_on_nodes, $reslv_node2stat;
	    ##push @alt_gapstat_on_nodes, $alt_node2stat;
	    push @alt_gapstat_on_nodes, $set_alt_node2stat;
	    ##push @mapped_indels, $options_br2indels;
	    push @mapped_indels, \@options_br2indels;

	}			# End of the outermost for-loop #

	##return (\@mapped_indels, \@reslv_gapstat_on_nodes,
	##	\@alt_gapstat_on_nodes, \@gapstat_wt_on_nodes);
	return (\@mapped_indels, \@alt_gapstat_on_nodes, \@gapstat_wt_on_nodes);

}		# end of the subroutine 'bch_map_indels_on_tree' #


# bch_infer_gapstat_on_nodes:
#	This subroutine infers the gap states on the internal nodes
#	(as well as on the external ones) of
#	a phylogenetic tree based on a parsimony principle.
#
# INPUT: (\@list_indels, \@seq_names, \%node2children, \%node2parent, \%name2id, $char_gap, $char_ng), where
#	@{$list_indels[$k]} = ($blstart_frame, $blend_frame, \@blstart_seq, \@blend_seq,
#	\@pattern, @sbalign) for the $k th indel unit, with
#	$blstart/end_frame is the starting/ending coordinate of the unit in terms of the alignment frame,
#	$blstart/blend_seq[$m] = the coordinate of the residue of the $m th sequence that starts/ends the block,
#	@pattern = (the sequence numbers to which the gap is assigned),
#	$sbalign[$m] = the subsequence of the $m th sequence embedded in the subalignment;
#	$seq_names[$m] is the name of the $m th sequence;
#	%node2children = ($node_id => [ a list of child IDs], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	%name2id = ($node_name => $node_id, ...)
#	for the external nodes (i.e. sequences) plus the top node;
#	$char_gap indicates that the sequence posesses the gap, and $char_ng indicates the opposite.
#
# OUTPUT: \@gapstat_wt_on_nodes, where
#	%{$gapstat_wt_on_nodes[$k]} = ($node_id => \@status, ...)
#	    is for the $k th indel block, with
#	@{$status[$p]} = ($state, $weight, $change, $number), where
#	$state = $char_gap or $char_ng, depending on whether the node was occupied by gaps or not,
#	$changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	$number is the number of child nodes (or grandchildren) that support the status (recorded in %gapstat_on_nodes),
#	$weight is the total fraction of the weights supporting the status.
#
sub bch_infer_gapstat_on_nodes (\@\@\%\%\%$$) {

	my ($list_indels, $seq_names, $node2children, $node2parent, $name2id,
		$char_gap, $char_ng) = @_;

	my @states = ($char_gap, $char_ng);
	my $ct_indels = @{$list_indels};
	my $ct_seqs = @{$seq_names};
	(($ct_indels>0) and ($ct_seqs>0)) or (return []);

	my @seqids_aln = (0..$ct_seqs-1);
	my @gapstat_wt_on_nodes = ();

	my $frac_nodes = fractionate_by_ndepth (%{$node2children}, %{$name2id});

	for (my $k=0; $k<$ct_indels; $k++) { # The outermost for-loop #

		# Initialize the phylogenetic map #

	    my %node2stat_wt; # (node_ID => \@status, ...) where
			   # @{$status[$p]} = ($state, $weight, $change, $number)

	    my $gapped = $list_indels->[$k][4];
	    (undef, undef, my $nongapped) = diff_nmarray (@{$gapped}, @seqids_aln);

	    foreach my $m (@{$gapped}) { # the 1st inner foreach-loop #
		my $seqname = $seq_names->[$m];
		my $nodeid = $name2id->{$seqname};
		$node2stat_wt{$nodeid} = [[$char_gap, 1, 0, 1]];
	    }				# end of the 1st inner foreach-loop #
	    foreach my $m (@{$nongapped}) { # the 2nd inner foreach-loop # 
		my $seqname = $seq_names->[$m];
		my $nodeid = $name2id->{$seqname};
		$node2stat_wt{$nodeid} = [[$char_ng, 1, 0, 1]];
	    }				# end of the 2nd inner foreach-loop #

	    assign_node2stat_weights (%node2stat_wt, @{$frac_nodes}, %{$node2children}, 1);
	    my @dummy_states = ();
	    assign_node2minimum_changes (%node2stat_wt, @{$frac_nodes}, %{$node2children}, @dummy_states, 1, 2);

	    ## my ($reslv_node2stat, $alt_node2stat)
	    ## my $set_alt_node2stat
	    ##	= infer_gapstat_by_pars (%node2stat_wt, @{$frac_nodes}, %{$node2children}, %{$node2parent}, %{$name2id});

	    push @gapstat_wt_on_nodes, \%node2stat_wt;

	}			# End of the outermost for-loop #

	return \@gapstat_wt_on_nodes;

}		# End of the subroutine 'bch_infer_gapstat_on_nodes' #


# parsmap_binstatchg_on_tree:
#	This subroutine maps changes of binary states onto branches of
#	a phylogenetic tree based on a parsimony principle.
#	It also outputs the associated states on the internal nodes
#	(as well as on the external ones).
#
# INPUT: (\%name2stat, \%node2children, \%node2parent, \%name2id, $char_abs, $char_pres), where
#	%name2stat = ($node_name => $state, ...) for the external nodes;
#	%node2children = ($node_id => [ a list of child IDs], ...);
#	%node2parent = ($node_id => $parent_id, ...);
#	%name2id = ($node_name => $node_id, ...) for the external nodes plus the top node;
#	$char_abs/$char_pres represents the 'absence' and 'presence' of the character (trait)
#	(actually it is translatable to any binary characters).
#
# OUTPUT: (\@options_br2changes, \@set_alt_node2stat, \%node2stat_wt), where
#	%{$options_br2changes->[$option]} = ($node_id => [$inferred_change, $weight_up, $weight_down], ...)
#	is the $option th option for the $k th indel block, with
#	$node_id the node immediately below the branch to which the change is attributed,
#	$inferred_change = 'G'/'L' for gain/loss of the character in question,
#	$weight_up/down = the fraction of weights that supports the state on the up/downstream node;
#	%{$set_alt_node2stat[$opt]} = ($node_id => $status, ...) for the $opt th alternative map;
#	%node2stat_wt = ($node_id => \@status, ...) with
#	@{$status[$p]} = ($state, $weight, $change, $number) for the $p th state (on the node),
#	where
#	$state = $char_gap or $char_ng, depending on whether the node was occupied by gaps or not,
#	$changes is the minimum number of changes the subtree below the node must undergo if the node take that state.
#	$number is the number of child nodes (or grandchildren) that support the status (recorded in %gapstat_on_nodes),
#	$weight is the total fraction of the weights supporting the status.
#
sub parsmap_binstatchg_on_tree (\%\%\%\%$$) {

	my ($name2stat, $node2children, $node2parent, $name2id, $char_abs, $char_pres) = @_;

	my @states = ($char_abs, $char_pres);
	my $ct_extnodes = scalar keys %{$name2stat};
	unless ($ct_extnodes + 1 == scalar keys %{$name2id}) {
	    printf STDERR "[ERROR 'parsmap_binstatchg_on_tree'] ILLEGAL INPUT: The number of state-assigned external nodes ($ct_extnodes) CLASHES with that of ID-assigned one (%d) !\n\n", scalar( keys %{$name2id} ) - 1;
	    return (undef, undef, undef);
	}

	($ct_extnodes>0) or (return ([], [], []));

	my @mapped_indels = my @reslv_gapstat_on_nodes = ();
	my @alt_gapstat_on_nodes = my @gapstat_wt_on_nodes = ();

	my $frac_nodes = fractionate_by_ndepth (%{$node2children}, %{$name2id});

	##for (my $k=0; $k<$ct_indels; $k++) { # The outermost for-loop #

		# Initialize the phylogenetic map #

	my %node2stat_wt; # (node_ID => \@status, ...) where
			   # @{$status[$p]} = ($state, $weight, $change, $number)

	foreach my $node_name (keys %{$name2stat}) { # The 1st foreach-loop #
	    my $nodeid = $name2id->{$node_name};
	    unless (defined $nodeid) {
		print STDERR "[ERROR 'parsmap_binstatchg_on_tree'] ILLEGAL INPUT: NO ID is assigned for the node '$node_name' !\n";
		return (undef, undef, undef);
	    }
	    $node2stat_wt{$nodeid} = [ [$name2stat->{$node_name}, 1, 0, 1]];
	}			# End of the 1st foreach-loop #


		# Assign each internal node various stats,
		# including the number of minimum changes the subtree under it must undergo.

	assign_node2stat_weights (%node2stat_wt, @{$frac_nodes}, %{$node2children}, 1);
	my @dummy_states = ();
	assign_node2minimum_changes (%node2stat_wt, @{$frac_nodes}, %{$node2children}, @dummy_states, 1, 2);

		# Infer sets of states for internal nodes
		# that minimize the total number of changes.

	##my ($reslv_node2stat, $alt_node2stat)
	my $set_alt_node2stat
	   = infer_gapstat_by_pars (%node2stat_wt, @{$frac_nodes}, %{$node2children}, %{$node2parent}, %{$name2id});

		# map changes onto the branches for each set of node states #

	my %converter = ('I' => 'G', 'D' => 'L');
		# Convert the old representation scheme 'I(nsertion)'/'D(eletion)'
		# into the current one 'G(ain)'/'L(oss)'.

	my @options_br2changes = ();
	foreach my $alt_node2stat (@{$set_alt_node2stat}) { # The 2nd foreach-loop # 
	    ##my $br2indels = br_map_indels_on_tree (%{$alt_node2stat}, %node2stat_wt, %{$node2children}, %{$name2id}, $char_gap, $char_ng);
	    my $br2indels = br_map_indels_on_tree (%{$alt_node2stat}, %node2stat_wt, %{$node2children}, %{$name2id}, $char_abs, $char_pres);
	    my %br2changes; # ($node2id => [$inferred_change, $weight_up, $weight_down], ...).

		## Just convert 'I(nsertion)'/'D(eletion)' into 'G(ain)'/'L(oss)' ##

	    foreach my $nodeid (keys %{$br2indels}) {  # inner foreach-loop #
		my $change = $br2indels->{$nodeid};
		my @converted = @{$change};
		my $cvtd_change = $converter{$change->[0]};
		unless (defined $cvtd_change) {
		    print STDERR "[WARNING 'parsmap_binstatchg_on_tree'] The change on the node '$nodeid' is assigned an UNFAMILIAR NAME '$change->[0]' !\n";
		}
		$converted[0] = $cvtd_change;
		$br2changes{$nodeid} = \@converted;
	    }				# End of the inner foreach-loop #
	    push @options_br2changes, \%br2changes;
	}					# End of the 2nd foreach-loop #

	return (\@options_br2changes, $set_alt_node2stat, \%node2stat_wt);

}		# end of the subroutine 'parsmap_binstatchg_on_tree' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


