package MyApBayes_Score_MSA_ExtJC;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyApBayes_Score_MSA_Bch_ExtJC.pm:
#	This is a Perl module that does almost the same jobs as
#	the Perl script 'apprxbayes_scoring_msa.bch.extJC.f.pl'.
#	(CREATION STARTED on Nov 28 (Sun), 2010 and ENDED on Nov 29 (Mon), 2010)
#
# apprxbayes_scoring_msa.bch.extJC.f.pl:
#	This revised version of 'apprxbayes_scoring_msa.bch.extJC.pl'
#	calculates the scores more quickly than its predecesssor.
#	The main change is the replacement of the module 'MyTree_ML.pm'
#	with its fast version 'MyTree_MLf.pm'.
#	(CREATED on September 18 (Sat), 2010)
#
# apprxbayes_scoring_msa.bch.extJC.pl:
#	This Perl script calculates the approximate Bayesian-like score
#	of an input MSA under the extended Jukes-Cantor model.
#	The script was made by modifying the script 'ml_scoring_msa.bch.extJC.pl'.
#
#	It is assumed that the list of candidate trees and their ML values
#	is already available and is given via an input file.
#	Bayesian-like score of an MSA is:
#	exp (Score (MSA| {trees}) ) = Sum_{T in trees} Awgt (T) * Prob (MSA | T),
#	where Awgt (T) is the Akaike weight given as:
#	Awgt (T) = exp (- delta_AIC(T)/2) /[ Sum_{T' in trees} exp (- delta_AIC(T')/2) ],
#	with delta_AIC(T) = AIC(T) - AIC_min. AIC (T) is the Akaike information criterion:
#	AIC (T) = -2 * log (Prob (Anchor_MSA | T)_ML ) + 2*K*N/(N - K -1),
#	with N = #{columns in the anchor MSA},
#	K = #{adjustable parameters in the model (including T)}.
#	(CREATION STARTED on July 12 (Mon), 2010 and ENDED on July 13 (Tue), 2010)
#
# ml_scoring_msa.bch.extJC.pl:
#	This Perl script calculates the ML (actually just the likelihood) score
#	of an input MSA under the extended Jukes-Cantor model.
#	This does NOT use PAML programs, but instead uses the subroutine 'msa_likelihood_extJC'
#	in the Perl module 'MyTreeMap_ML.pm'.
#	The script was made by using the framework of 'ml_scoring_msa.bch.pl'.
#	(CREATION STARTED on June 29 (Tue), 2010 and ENDED on June 30 (Wed), 2010)
#
# ml_scoring_msa.bch.pl:
#	This is a minor revision of 'ml_scoring_msa.pl'
#	that is to be used as a component of the script 'compare_ml_msa.pl',
#	which in turn identifies the best ML alignment within the length constraints.
#
# ml_scoring_msa.pl:
#	This script calculates the ML (actually just the likelihood) score
#	of an input MSA under an input tree (with a given topology and branch lengths).
#	Briefly, the script conducts the following tasks:
#	(1) Divide the input MSA into blocks {Bk}, each consisting of columns with the same gap pattern;
#	(2) For each block Bk, prepare a truncated tree Tk consisting of ungapped sequences
#	(and their ancestral nodes);
#	(3) Feed the block Bk and the truncated tree Tk into the program codeml or baseml in
#	the PAML package, and calculate the likelihood Lk of the block;
#	[ (4) Examine the gap patterns more carefully and enumerate the possible indel scenarios of the lowest order. Then calculate the likelihood of each indel scenario and sum them up to obtain the whole probability Pindel ; ]
#	(5) Multiply the likelihoods {Lk} and Pindel to get the whole likelihood:
#	L (MSA | T) = Pindel * Pi_k Lk (Bk | Tk).
#
#	DEPENDENCY: This script depends on the PAML package.
#
#	TO BE DONE:
#	When the reduced alignment contains only 1 sequence, likelihood will be given by the product of the equilibrium frequencies of the amino acid residues.
#	When it contains only 2 sequences, PAML should be conducted with the molecular clock ON.
#	 --- Done on March 22 (Mon), 2010
#
#	< A BUG FIXED on March 23 (Tue), 2010 >
#	The program 'codeml' appears to handle 2-taxon tree without the 'molecular clock' assumption.
#	It, however, clashes when 'clock = 1' and 'fix_blength = 1' coexists.
#	So, I just removed the exceptional handling when #{sequences} = 2.
#

use strict;
use warnings;

use MyFile_manipu qw( open_input_file2 open_output_file );

use MyArrayManipu_num qw( log_sum_exp );

use MyBin_Oper qw( format_bin_pattern select_by_bin_value );
use MyMalnIO qw( extract_fasta pr_clustal_like_lite_aa pr_ODEN_input pr_phylip_aln);
use MyMalnManipu_UH qw( div_msa_by_gappatterns cls_blcs_by_patterns );

use MyTreeManipu_II qw( extract_sgltree_from_file input_tree_file output_tree );
use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyTreeMap_MLf qw( msa_likelihood_extJC_f );

=MOVED to the main program (e.g. of 'compare_apbayes_msa.bch.extJC.3.pl') on Nov 29 (Mon), 2010
use MyPAML_parser qw( );
	## retrieve_aa_freqs prepare_paml_ctrl
=cut

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( apbayes_scoring_msa_extJC_f readin_set_trees3 log_akaike_weights );

##sub readin_set_trees3 (\$);
##sub log_akaike_weights (\@$);
sub ln_ml_sgl_seq (\$\%);
sub even_brlen_2seq_tree (\%$);
=For the predecessor 'ml_scoring_msa.bch.pl'
sub parse_paml_main_output (\$);
=cut

	### Parameters to be Adjusted ###

##our $DIS_LEN = 60; # Maximum length of the sequences to be displayed in a line

our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);

our $COMMENT = '#';  # Indicator of the comment line in the tree file
##our $GAP = '-'; # Character representing a gap ## MOVED to the main program (e.g. of 'compare_apbayes_msa.bch.extJC.3.pl') on Nov 28 (Mon), 2010


=for the Perl script 'apprxbayes_scoring_msa.bch.extJC.f.pl'
	## Read in the command-line arguments ##
	## -- ADDED in revision to 'ml_scoring_msa.bch.pl' ##
	## (REVISED in creation of 'ml_scoring_msa.bch.extJC.pl' on June 29 (Tue), 2010)  
unless (@ARGV == 4) {
	print STDERR "\nUsage:  [perl] apprxbayes_scoring_msa.bch.extJC.pl {input_alignment} {input_tree_list} {output_file} {minimum_ct_characters}\n\n";
	die "$! .\n\n";
}
my ($inaln, $intree, $outfile, $minct_chars) = @ARGV;
(my $outdir = $outfile) =~ s!/*[^/]+$!!;
($outdir =~  m/\S/) or ($outdir = '.');

	### Check necessary files/directories ###

foreach my $file ($inaln, $intree) {
    unless (( -f $file ) and ( -r $file )){
	print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	die "$! .\n\n";
    }
}

foreach my $dir ($outdir) {
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] 'system (mkdir -p ...)' FAILED to CREATE the output directory '$dir': $? !\n";
	    die "$! .\n\n";
	}
    }
    unless (( -r $dir ) and ( -w $dir )) {
	unless (0 == system ("chmod u+rw $dir") ) {
	    print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] 'system (chmod u+rw ...)' FAILED to CHANGE permission MODES of the output directory '$dir': $? !\n";
	    die "$! .\n\n";
 	}
    }
}

	### Declare the Beginning of the Script ###
print STDERR "<< Program name: $0 >>\n\n";

print STDERR "Minimum number of possible characters in a column: $minct_chars .\n\n";

print STDERR "Input files:\n";
print STDERR "     (Alignment)     $inaln ,\n";
print STDERR "     (List of trees) $intree .\n\n";
print STDERR "Output file: $outfile .\n";

=cut

=MOVED to the main Program (e.g.of the Perl script 'compare_apbayes_msa.bch.extJC.3.pl' on Nov 29 (Mon), 2010)
	### Retrieve amino acid frequencies from the data file $src_data
	### for the substitution model to be used.

our %AA2FREQ; # = ($amino_acid => $equil_freq, ...)

	## REVISED in creation of 'ml_scoring_msa.bch.extJC.pl'
	## (on June 30 (Wed), 2010) ##
print STDERR "\n\n<< Amino-acid frequencies >>\n\n";
print STDERR join ("\t", 'Amino_acid', 'Frequency'), "\n\n";
our @AMINO_ACIDS = @MyPAML_parser::AMINO_ACIDS;

foreach my $aa (@AMINO_ACIDS) { # Preliminary foreach-loop #
	$AA2FREQ{$aa} = 1/$minct_chars; ## Assumption of the extended Jukes-Cantor model ##
	print STDERR join ("\t", $aa, $AA2FREQ{$aa}), "\n";
} 			# End of the preliminary foreach-loop #
print STDERR "\n\n";
=cut

=MOVED to the Main Program (e.g. of the Perl script 'compare_apbayes_msa.bch.extJC.3.pl')

	## Read in the list of Input Trees (and their ML values) ##

print STDERR "<< Information on the Input Trees >>\n\n";

# @{$set_cand_trees->[$t]} = ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr)
#
my ($ct_effsites, $ct_mono_sites, $set_cand_trees) = readin_set_trees3 ($intree);
unless (defined $set_cand_trees) {
	print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] The subroutine 'readin_set_trees3' FAILED on the input tree list '$intree' !\n";
	die "$! .\n\n";
}
my $ct_cand_trees = @{$set_cand_trees};

print STDERR "#{effective_sites} = $ct_effsites\n";
print STDERR "#{invariant_sites} = $ct_mono_sites\n";
print STDERR "#{candidate_trees} = $ct_cand_trees\n\n\n"; 

my @std_set_cand_trees = ($ct_cand_trees < 2) ? @{$set_cand_trees} :
	 sort {$a->[3] <=> $b->[3] } @{$set_cand_trees};

	## Calculate the Akaike weights of the candidate trees
	## (ADDED in creation of 'apprxbayes_scoring_msa.bch.extJC.pl' (on July 12 (Mon), 2010)
my $ln_akaike_wgts = log_akaike_weights (@std_set_cand_trees, 3);
unless (defined $ln_akaike_wgts) {
	print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] The subroutine 'log_akaike_weights' FAILED on the tree list extracted from '$intree' !\n";
	die "$! .\n\n";
}

####=for TESTING the script
for (my $t = 0; $t<$ct_cand_trees; $t++) { # Preliminary for-loop #

	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};
	my $topnode_id = $name2id->{$TOP_NODE};
	my $ln_awgt = $ln_akaike_wgts->[$t];

	print STDERR "\< The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) >\n";
	print STDERR "Sum (-log(Pv)) = $sum_neg_log_pv\n";
	print STDERR "log (Prob(MSA|Tree)_ML) = $ln_ml_msa\n";
	print STDERR "#{free parameters} = $ct_free_params\n";
	print STDERR "AIC = $aic\n";
	print STDERR "BIC = $bic\n\n";

	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *STDERR, $fl_bootlabel, $fl_nodeid) ) {
	    print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] 'output_tree' FAILED to output the $t th candidate tree !\n";
	    die "$! .\n\n";
	}

	print STDERR "\n\nNode_ID --> Parent_ID\n\n";
foreach my $nodeid (sort keys %{$node2pa}) { print STDERR "$nodeid  --> ", $node2pa->{$nodeid}, "\n"; }
	print STDERR "\n\n";
##=for TESTING the subroutine 'read_tree_file'
	print STDERR join ("\t", 'Branch_ID', 'length', 'P-value'), "\n\n";
	foreach my $brid (sort keys %{$br_attr}) { # inner foreach-loop #
	    my $attr = $br_attr->{$brid};
	    my ($brlen, $pvalue) = @{$attr}[$INDX_LENGTH, $INDX_BOOT];
	    (defined $brlen) or ($brlen = 'NA');
	    (defined $pvalue) or ($pvalue = 'NA');
	    print STDERR join ("\t", $brid, $brlen, $pvalue), "\n";
	}				# End of the inner foreach-loop #
	print STDERR "\n\n";
##=cut

}			# end of the preliminary for-loop #
#### exit 0; ### For TESTING the Upper part of the script
####=cut

=cut

=for the Perl script 'apprxbayes_scoring_msa.bch.extJC.f.pl'
	### Read in the Input Alignment (in the multi-FASTA format) ###
# ${$seqs->[$i]} is the $i th sequence embedded in the alignment
# $seqids->[$i] is the ID of the $i th sequence
# $seqlens->[$i] is the length of the $i th sequence
# ${$seqid2desc->{$seqid}} is the description of the sequence with ID $seqid
#
my ($seqs, $seqnames, $seqlens, $seqid2desc) = extract_fasta ($inaln);
unless ((defined $seqs) and (@{$seqs}>0)) {
	print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] The subroutine 'extract_fasta' FAILED on th e input alignment '$inaln' !\n";
	die "$! .\n\n";
}
print STDERR "\n --- Finished reading in the input alignment '$inaln' ...\n\n";
=cut

	### The Main Subroutine that replaces most of the functions ###
	### of the Perl script 'apprxbayes_scoring_msa.bch.extJC.f.pl'. ###
	### (CREATED on Nov 29 (Mon), 2010)

# apbayes_scoring_msa_extJC_f:
#	This subroutine is a modulized version of the main functions of
#	the script 'apprxbayes_scoring_msa.bch.extJC.f.pl'.
#	It receives an MSA and a set of candidate trees (and their weights),
#	as well as some other necessary parameters,
#	and returns the approximate Bayesian score of the MSA under the set of trees.
#
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@seqs, \@seqnames, \@set_cand_trees, \@ln_akaike_wgts, $micct_chars, $gap, \%aa2freq, \*${$fh_log}), where
#	${$seqs[$k]} is the $k th sequence embedded in the alignment;
#	$seqnames[$k] is name of the $k th sequence;
#	@{$set_cand_trees[$t]} = ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, \%node2pa, \%node2chi, \%namee2id, \%id2name, \%br_attr)
#	is for the $t th candidate tree ;
#	$ln_akaike_wgts[$t] is the logarithmic Akaike weight for the $t th tree;
#	$minct_chars is the specified count of characters that are mutually transferable,
#	(but the actual number of characters might be larger than this);
#	$gap is a character specifying a gap;
#	%aa2freq = ($amino_acid => $frequency, ...);
#	$fh_log is (the reference to) the file-handle of the log file.
#
# OUTPUT: (\%tree2score, \@sum_lnl_tw, \@class_lnl_tw), where
#	%tree2score = ($tree_id => $score_treewise, ...), where $tree_id = 0, ..., #{trees} -1,
#	and $score_treewise = $sum_lnl_tw[$tree_id] + $ln_akaike_wgts[$tree_id];
#	$sum_lnl_tw[$t] is the total log likelihood of the input MSA under the $t th tree;
#	$class_lnl_tw[$t]->[$c] is the log likelihood of the $c th block-class
#	in the input MSA under the $t th tree.
#
sub apbayes_scoring_msa_extJC_f (\@\@\@\@$$\%\*) {

  my ($seqs, $seqnames, $set_cand_trees, $ln_akaike_wgts, $minct_chars, $gap, $aa2freq, $fh_log) = @_;
	## my $tree2score = apbayes_scoring_msa_extJC_f (@{$seqaln}, @{$sequence_ids}, @std_set_cand_trees, @{$ln_akaike_wgts}, $minct_chars, $GAP, %AA2FREQ, *${$fh_log});

  my $ct_seqs = @{$seqs};

	# ADDED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) # 

  my $ct_cand_trees = @{$set_cand_trees};

  unless (@{$seqnames} == $ct_seqs) {
	print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] ILLEGAL INPUT: The number of sequence names (", scalar (@{$seqnames}), ") CLASHES with that of the sequences in the alignment ($ct_seqs): $! !\n";
	return undef;
  }
  unless (@{$ln_akaike_wgts} == $ct_cand_trees) {
	print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] ILLEGAL INPUT: The number of weights (", scalar (@{$ln_akaike_wgts}), ") CLASHES with that of the candidate trees ($ct_cand_trees): $! !\n";
	return undef;
  }

=for TESTING the script
  print $fh_log "<< Input alignment (#{sequences} = $ct_seqs) >>\n\n";

  unless ( pr_ODEN_input (@{$seqnames}, @{$seqs}, 60, *${$fh_log}) ) {
	print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'pr_ODEN_input_lite' FAILED on the input alignment: $! !\n";
	return undef;
  }
  print STDERR "\n\n";
=cut

	### Divide the input alignment into blocks of contiguous sites with the  same gap pattern ###

  # ${$blocks->[$b][$k]} = $seq is the $k th subsequence (or string of gaps) in the $b th block;
  # @{$patterns->[$b]} is an array of 32-bit binary numbers
  #	representing a gap pattern of the $b th block
  #  ('1' for an ungapped subsequence, and '0' for a gap);
  # 
	# REVISED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) #
  my ($blocks, $patterns) = div_msa_by_gappatterns (@{$seqs}, $gap);
  ##my ($blocks, $patterns) = div_msa_by_gappatterns (@{$seqs}, $GAP);

  unless (defined $blocks) {
	print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'div_msa_by_gappatterns' FAILED: $! !\n";
	return undef;
  }

  my $ct_blocks = @{$blocks};

  ##my $mod = ($ct_seqs -1) % 32;
  ##my $qtnt = int (($ct_seqs - 1) /32);
  ##my $size_last_bits = $mod + 1;

  print $fh_log "<< Blocks of contiguous sites with the same gap pattern (#{blocks} = $ct_blocks) >>\n\n";

  for (my $b = 0; $b < $ct_blocks; $b++) { # The for-loop to print out blocks #

	my $block = $blocks->[$b];
	my $blen = length ${$block->[0]};

	my $pattern = $patterns->[$b];
	my $bin_pattern = format_bin_pattern (@{$pattern}, $ct_seqs, 32);

	print $fh_log "{ The $b th block (length = $blen) }\n\n";
	print $fh_log "Gap pattern:  '", ${$bin_pattern}, "' .\n\n";
=for TESTING the script
	print $fh_log "Alignment:\n\n";
	unless ( pr_ODEN_input (@{$seqnames}, @{$block}, 60, *${$fh_log}) ) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'pr_ODEN_input_lite' FAILED on the $b th block of the input alignment !\n";
	    return undef;
	}
	print $fh_log "\n\n";
=cut
  }			# End of the for-loop to print out blocks #


	### Cluster the blocks with the same gap pattern,
	### then concatenate the alignments in the same cluster.

  # ${$cls_blocks->[$c][$k]} is the $k th subsequence in the $c th concatenated block
  # of the same pattern.
  # @{$cls_patterns->[$c]} is the gap pattern of the $c th concatenated block;
  # @{$cls_indices->[$c]} lists the indices of the blocks that contribute to the $c th concatenated block.
  #
  my ($cls_blocks, $cls_patterns, $cls_indices) = cls_blcs_by_patterns (@{$blocks}, @{$patterns});

  unless (defined $cls_blocks) {
	print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'cls_blcs_by_patterns'  FAILED on the input MSA: $! !\n";
	return undef;
  }

  my $ct_cls = @{$cls_blocks};

  print $fh_log  "\n<< Clusters of the blocks with the same gap pattern (#{clusters} = $ct_cls) >>\n\n";

	### Conduct the ML analyses, cluster by cluster
	### and for ALL of the candidate TREEs

  my @sum_lnl_tw = (); # $sum_lnl_tw[$t] is the total log likelihood of the whole MSA under the $t th tree.
  my @class_lnl_tw = (); # $class_lnl[$t]->[$c] is the log likelihood of the $c th class under the $t th tree.
  foreach (1 .. $ct_cand_trees) { # Auxiliary foreach-loop #
	push @sum_lnl_tw, 0;
	push @class_lnl_tw, [];
  }			# End of the auxiliary foreach-loop #

  for (my $c = 0; $c < $ct_cls; $c++) { # The for-loop over clusters of MSA columns of the same gap patterns #

	my $cnct_blcs = $cls_blocks->[$c];
	my $clen = length ${$cnct_blcs->[0]};
	my $pattern = $cls_patterns->[$c];
	my $indices = $cls_indices->[$c];

	my $bin_pattern = format_bin_pattern (@{$pattern}, $ct_seqs, 32);

	print $fh_log "\n< The $c th cluster (Total length = $clen) >\n\n";
	print $fh_log "Component blocks:  ", join (', ', @{$indices}), " .\n\n";
	print $fh_log "Gap pattern: '", ${$bin_pattern}, "' .\n\n";

##=for TESTING the script
	print $fh_log "Alignment:\n\n";
	unless ( pr_ODEN_input (@{$seqnames}, @{$cnct_blcs}, 60, *${$fh_log}) ) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'pr_ODEN_input' FAILED on the $c th cluster of blocks in the input alignment: $! !\n";
	    return undef;
	}
	print $fh_log "\n\n";
##=cut

	   # Prepare a 'reduced' alignment by removing sequences containing only gaps #

	my $red_seqnames = select_by_bin_value (1, @{$seqnames}, @{$pattern}, 32);
	my $red_aln = select_by_bin_value (1, @{$cnct_blcs}, @{$pattern}, 32);

	my $sbct_seqs = @{$red_seqnames};

		# Detect Erroneous Interim Output as soon as possible #
		# (MOVED from the 2 blocks below on July 13 (Tue), 2010) #

	unless ($sbct_seqs == @{$red_aln}) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The number of sequences in the reduced alignment (", scalar (@{$red_aln}), ") CLASHES with that of selected sequence names ($sbct_seqs) for the $c th block-cluster in the input MSA: $! !\n";
	    return undef;
	}

		# Handle Exceptional Cases  -- Part I --- #

	unless ($sbct_seqs >0) {
	    print $fh_log "[CAUTION 'apprxbayes_scoring_msa.bch.extJC.pl'] NO selected sequences for the block-cluster $c : $! !\n";
	    print $fh_log "     --- I will simply skip this ...\n\n";

	    for (my $t=0; $t<$ct_cand_trees; $t++) { # 1st special inner for-loop #
		push @{$class_lnl_tw[$t]}, 0;
	    }			# End of the 1st special inner for-loop #

	    next;
	}

=for TESTING the script
	print $fh_log "Reduced Alignment (#{involved sequences} = $sbct_seqs):\n\n";
	unless ( pr_ODEN_input (@{$red_seqnames}, @{$red_aln}, 60, *${$fh_log}) ) {
	    print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] The subroutine 'pr_ODEN_input' FAILED on the $c th cluster of blocks in the input alignment: $! !\n";
	    return undef;
	}
	print $fh_log "\n\n";
=cut

	if ($sbct_seqs == 1) {
	    print $fh_log "[CAUTION 'apprxbayes_scoring_msa.bch.extJC.pl'] The block-cluster $c contains ONLY 1 SEQUENCE, so I will perform a simple estimation of the ML score ...\n\n";

		# REVISED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) #
	    my $lnl = ln_ml_sgl_seq (${$red_aln->[0]}, %{$aa2freq});
	    ##my $lnl = ln_ml_sgl_seq (${$red_aln->[0]}, %AA2FREQ);
	    unless (defined $lnl) {
		print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'ln_ml_sgl_seq' FAILED on the block-cluster $c: $! !\n";
		return undef;
	    }

	    for (my $t = 0; $t<$ct_cand_trees; $t++) { # 2nd special inner for-loop #
		push @{$class_lnl_tw[$t]}, $lnl;
		$sum_lnl_tw[$t] += $lnl;
	    }			# End of the 2nd special inner for-loop #
	    next;
	}

    for (my $t = 0; $t<$ct_cand_trees; $t++) { # middle for-loop over candidate trees #

		# REVISED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) #
	my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$set_cand_trees->[$t]};
	##my ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, $node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$std_set_cand_trees[$t]};

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ln_awgt = $ln_akaike_wgts->[$t];

	print $fh_log "\{ The $t th candidate tree (topnode ID: $topnode_id\; Log_Akaike_weight = $ln_awgt) }\n\n";

	   ## Prepare a 'reduced' phylogenetic tree by excluding sequences containing only gaps ##
	my %sb_name2id;

		## REVISED to speed up the process (on Nov 28 (Sun), 2010) ##

	my ($sb_topnode_id, $sb_node2ch, $sb_node2pa, $sb_br_attr);
	if ($sbct_seqs == $ct_seqs) {
		$sb_topnode_id = $topnode_id;
		$sb_node2ch = $node2ch;
		$sb_node2pa = $node2pa;
		$sb_br_attr = $br_attr;
		%sb_name2id = %{$name2id};
	} else {

	  my @sub_nodes = ();
	  foreach my $seqname (@{$red_seqnames}) { # The inner foreach-loop #
	    my $nodeid = $name2id->{$seqname};
	    unless (defined $nodeid) {
		print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] For the block-cluster $c, the sequence name '$seqname' has NO node ID in the $t th candidate tree: $! !\n\n";
		return undef;
	    }
	    push @sub_nodes, $nodeid;

		## A BUG fixed on June 30 (Wed), 2010 ##
	    $sb_name2id{$seqname} = $nodeid;
	    ##$sb_name2id{$seqname} = $name2id{$nodeid};
	  }				# End of the inner foreach-loop #

	  (my $sb_topnode_ids, $sb_node2ch, $sb_node2pa, $sb_br_attr)
		= construct_neat_subtree (%{$node2pa}, @sub_nodes, %{$br_attr});

	  unless (defined $sb_topnode_ids) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] FAILED to reduce the $t th candidate tree for the block-cluster $c !\n";
	    die "$! .\n\n";
	  }
	  unless (1 == @{$sb_topnode_ids}) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] Found MULTIPLE TOP NODES for the reduced tree of the $t th tree for the block-cluster $c: ", join (', ', @{$sb_topnode_ids}), " !\n";
	    die "$! .\n\n";
	  }
	  $sb_topnode_id = $sb_topnode_ids->[0];
	  $sb_name2id{$TOP_NODE} = $sb_topnode_id;
	}
		## END of "REVISED to speed up the process (on Nov 28 (Sun), 2010)" ##


		# Handle Exceptional Cases  -- Part II --- #

=BUG FIXED on March 23 (Tue), 2010
	if (@{$red_seqnames} == 2) {
	    print STDERR "[CAUTION 'ml_scoring_msa.pl'] The cluster $c has ONLY 2 SEQUENCES --- I will use the Molecular clock assumption ...\n\n";
	    unless ( even_brlen_2seq_tree (%{$sb_br_attr}, $INDX_LENGTH) ) {
		print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'even_brlen_2seq_tree' FAILED on the block-cluster $c !\n";
		die "$! .\n\n";

	    }
	} elsif (@{$sb_node2ch->{$sb_topnode_id}} == 2) {
=cut
	if ((@{$sb_node2ch->{$sb_topnode_id}} == 2) and (@{$red_seqnames}>2)) { # This condition replaces the block removed above (on March 23 (Tue), 2010) #
	    print $fh_log "[CAUTION 'apbayes_scoring_msa_extJC_f'] The reduced tree for the cluster $c is BIFURCATED --- I will make it multifurcated ...\n\n";
	    unless ( bi2multifurcated_top (%{$sb_node2pa}, %{$sb_node2ch}, %sb_name2id, %{$sb_br_attr}, @INDICES_ATTR) ) {
		print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] The subroutine 'bi2multifurcated_top' FAILED on the block-cluster $c: $! !\n";
		return undef;
	    }
	}

	print $fh_log "\n[ Reduced Tree (top_node_ID = $sb_topnode_id) ]\n\n";
	
	my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).

	unless ( output_tree ($sb_topnode_id, %{$sb_node2ch}, %{$id2name}, %{$sb_br_attr}, *${$fh_log}, $fl_bootlabel, $fl_nodeid) ) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] 'output_tree' FAILED to output the reduced tree of the $t th tree for the block-cluster $c: $! !\n";
	    return undef;
	}

		## Copy the reduced alignment, and use it as an input of 'msa_likelihood_extJC'.
	my @cp_red_aln = ();
	foreach my $seq (@{$red_aln}) { my $cp_seq = ${$seq}; push @cp_red_aln, \$cp_seq; }

	my ($cl_ln_lk, $clm_ln_lks, $set_chars) = msa_likelihood_extJC_f ($sb_topnode_id, %{$sb_node2ch}, %{$sb_node2pa}, %sb_name2id, @{$red_seqnames}, @cp_red_aln, $minct_chars, %{$sb_br_attr}, $INDX_LENGTH);
	unless (defined $clm_ln_lks) {
	    print STDERR "[ERROR 'apbayes_scoring_msa_extJC_f'] 'msa_likelihood_extJC' FAILED for the block-cluster $c under the $t th candidate tree: $! !\n";
	    return undef;
	} 

	push @{$class_lnl_tw[$t]}, $cl_ln_lk;
	$sum_lnl_tw[$t] += $cl_ln_lk;

###=for TESTING 'msa_likelihood_extJC'
	my $ct_clms = @{$clm_ln_lks};
	print $fh_log "\n[ Column-wise log likelihoods of the $c th cluster under the $t th candidate tree (Total: $cl_ln_lk, #{columns} = $ct_clms) ]\n\n";
	print $fh_log join ("\t", '#Column_no.', 'ln_lk', '(possible, characters, in, the, column)'), "\n\n";
	for (my $l = 0; $l<$ct_clms; $l++) { # Auxiliary for-loop #
	    my $ln_lk = $clm_ln_lks->[$l];
	    my $chars = $set_chars->[$l];
	    print $fh_log $l, "\t", $ln_lk, "\t(", join (', ', @{$chars}), ")\n";
	}			# End of the auxiliary for-loop #
	print $fh_log "\n\n";
###=cut

    }			# End of the middle for-loop over candidate trees #

  }	# End of the for-loop over the clusters of MSA columns of the same gap patterns #


	## Calculate the approximate Bayes-like score ##

  print $fh_log "\n\n<< Approximate Bayesian-like score >>\n\n";

	# REVISED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) #
  print $fh_log join ("\t", '#Tree_ID', 'score_tree-wise', 'ln(Akaike_weight)', 'Total_ln(ML)_tree') , "\n\n";
  ##print $fh_log join ("\t", '#Tree_ID', 'component'), "\n\n";
  my %tree2score;
  my @score_tree_wise = ();
  for (my $t=0; $t<$ct_cand_trees; $t++) { # 1st trailing for-loop over the trees
	my $ln_awgt = $ln_akaike_wgts->[$t];
	my $lnl_tw = $sum_lnl_tw[$t];
	my $term = $ln_awgt + $lnl_tw;

	push @score_tree_wise, $term;

	  # REVISED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) #
	print $fh_log join ("\t", $t, $term, $ln_awgt, $lnl_tw), "\n";
	##print $fh_log join ("\t", $t, $term), "\n";

	  # ADDED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' (on Nov 29 (Mon), 2010) #
	$tree2score{$t} = $term;
  }		# End of the 1st trailing for-loop over the trees

  my $total_apprx_bayes_score = log_sum_exp (@score_tree_wise);
  unless (defined $total_apprx_bayes_score) {
	print STDERR "\n[ERROR 'apbayes_scoring_msa_extJC_f'] 'log_sum_exp' FAILED to calculate the total Bayesian-like score: $! !\n";
	return undef;
  }

  print $fh_log "\nTotal approximate Bayesian-like score = $total_apprx_bayes_score\n\n";

	## Finish the output (ADDED in creation of 'MyApBayes_Score_MSA_ExtJC.pm' on Nov 29 (Mon), 2010) ##
  $tree2score{'Total'} = $total_apprx_bayes_score;

  return (\%tree2score, \@sum_lnl_tw, \@class_lnl_tw);

}	# End of the main subroutine 'apbayes_scoring_msa_extJC_f' #


=for the Perl script 'apprxbayes_scoring_msa.bch.extJC.f.pl'
	### Output the main results ###
my $fh_out = open_output_file ($outfile, 0);
unless (defined $fh_out) {
	print STDERR "[ERROR 'apprxbayes_scoring_msa.bch.extJC.pl'] The subroutine 'open_output_file' FAILED on the output file '$outfile' !\n";
	die "$! .\n\n";
}

##my @sum_lnl_tw = (); # $sum_lnl_tw[$t] is the total log likelihood of the whole MSA under the $t th tree.
##my @class_lnl_tw = (); # $class_lnl[$t]->[$c] is the log likelihood of the $c th class under the $t th tree.
##my $ln_akaike_wgts = log_akaike_weights (@std_set_cand_trees, 3);

print STDERR "\n\n --- Finally, I will print out the total score and contributions from respective clusters under respective trees ...\n\n";

print $fh_out join ("\t", '#Tree_ID', 'score_tree-wise', 'ln(Akaike_weight)', 'Total_ln(ML)', 'ln(ML_class0)', 'ln(ML_class1)', '...'), "\n\n";

for (my $t = 0; $t < $ct_cand_trees; $t++) { # 2nd trailing for-loop over the trees #
	my $score_tw = $score_tree_wise[$t];
	my $ln_awgt = $ln_akaike_wgts->[$t];
	my $lnl_tw = $sum_lnl_tw[$t];
	my $lnl_cw = $class_lnl_tw[$t];
	print  $fh_out join ("\t", $t, $score_tw, $ln_awgt, $lnl_tw, @{$lnl_cw}), "\n";
}				# End of the 2nd trailing for-loop over the trees #
print $fh_out "\n";
print $fh_out join ("\t", 'Total', $total_apprx_bayes_score, 0, '---'), "\n"; 

close ($fh_out);

print STDERR "\n\n === All the Processes FINISHED !! ===\n\n";
##exit 0;
=cut


	### DEFINITIONS of the Auxiliary SUBROUTINES ###

# readin_set_treees3:
#	This revised version of the subroutine 'readin_set_treees2'
#	also reads in the log-likelihood of the MSA under each candidate tree,
#	and calculates the AIC (Akaike information criterion) and BIC (Bayesian informaiton criterion).
#	(CREATED in creation of 'summ_tests_tree_reconst.mlt_ml.pl' on July 2 (Fri), 2010)
#
#	AIC and BIC are given by:
#	AIC = - 2 * ln (Prob (MSA | tree, model)_ML) + 2 * #{free parameters},
#	BIC = -2 * ln (Prob (MSA| tree, model)_ML) + ln (#{observations}) * #{parameters to be fixed}. 
#	If we assume that the sites evolve independently of each other,
#	#{observations} equals to #{columns in the MSA}.
#	#{parameters} depends on the evolutionary model and the tree,
#	especially, the number of branches.
#	(In the extended JC model, #{parameters} = #{branches}.)
#
#	REVISED on July 9 (Fri), 2010:
#	Now the AIC it calculates is not the common first-order version (above),
#	but the second-order version that can also be used for small samples:
#	AIC_r = - 2 * ln (Prob (MSA | tree, model)_ML) + 2 * K * N/(N-K-1),
#	where K = #{free parameters} and N = #{observations}.
# 
# readin_set_treees2:
#	This revised version of the subroutine 'readin_set_treees'
#	also extracts the number of columns used for the tree reconstruction,
#	as well as the number of sites that showed NO change.
#	(Created in creation of 'summ_tests_tree_reconst.mlt.pl' on May 20 (Thu), 2010)
#
# readin_set_treees:
#	This subroutine reads in an output file of 'tree_fm_sgn_partitions.pl'
#	and extracts information on ALL possible trees.
#
# INPUT: \$infile.
#
# OUTPUT: ($ct_columns, $ct_mono_sites, \@set_trees), where
# [OUTPUT: \@set_trees, where] -- for 'readin_set_tree'
#	[@{$set_trees[$t]} = ($sum_neg_log_pv, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr) ] --- for 'readin_set_tree2'
#	@{$set_trees[$t]} = ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr) 
#	for the $t th possible tree recorded in $infile.
#	$sum_neg_log_pv is the summation over compatible partitions of the minus logarithms of the P-values that dictate the significance of the branches responsible for the partitions.
#	$ln_ml_msa is the (approximate) log-likelihood of the MSA under the tree in question.
#	$ct_free_params is the number of free parameters in the model.
#	$aic is the Akaike information criterion.
#	$bic is the Bayesian information criterion.
#	%node2pa = ($node_id => $parent_id, ...).
#	%node2ch = ($node_id => \@child_ids, ...).
#	%name2id = ($node_name => $node_id, ...) for the top node and the exterior nodes.
#	%id2name = ($node_id => $node_name, ...) for the exterior nodes.
#	%br_attr = ($node_id => [attributes of the branch immediately above the node], ...).
#
sub readin_set_trees3 (\$) {

	my ($infile) = @_;
	local $/ = ''; # Use blank lines as input-record-separators

	my $fh_tmp = open_input_file2 (${$infile});
	unless (defined $fh_tmp) {
	    print STDERR "[ERROR 'readin_set_trees3'] The subroutine 'open_input_file2' FAILED on the input file '${$infile}': $! !\n";
	    return (undef, undef, undef);
	}

	my $in;
		## ADDED in creation of 'readin_set_trees2' (on May 20 (Thu), 2010) ##
	my ($ct_clms, $ct_mono_sites);
	while ($in = <$fh_tmp>) { # 1st preliminary while-loop #
	    ($ct_clms) = ($in =~ m/^\#\{Columns\s+without\s+gaps\}\s*\=\s*(\d+)/ );
	    (defined $ct_clms) and last;
	}		# End of the 1st preliminary while-loop # 
	while ($in = <$fh_tmp>) { # 2nd preliminary while-loop #
	    ($ct_mono_sites) = ($in =~ m/^\<\s*There\s+are\s+(\d+)\s+Mono\-color\s+sites\s*\>/);
	    (defined $ct_mono_sites) and last;
	}		# End of the 2nd preliminary while-loop #

		# Skip unnecessary information #
	while ($in = <$fh_tmp>) { # Leading while-loop #

	    ($in =~ m/^\<\<\s+Sets\s+of\s+significant\s+partitions\s+that\s+are\s+compatible/) and last;
	}			# End of the leading while-loop #

	unless (defined $in) {
	   print STDERR "[ERROR 'readin_set_trees3'] We've RUN OUT OF LINES in the input file '${$infile}' while looking for the start of the tree lists: $! !\n";
	   return (undef, undef, undef);
	}

	my @set_trees = ();

	while (my $in = <$fh_tmp>) { # Main outer while-loop #

=for TESTING the subroutine 'readin_set_trees'
	print STDERR "Input chunk:\n";
	print STDERR $in;
=cut
	    my ($tree_no, $sum_neg_log_pv) = ($in =~ m/^\<\s+The\s+(\d+)\s+th\s+set\s+\(\-Sum_log_Pv\s+\=\s+([^\)]+)\)\s+\>/);
	    (defined $sum_neg_log_pv) or next;

=for TESTING the subroutine 'readin_set_trees'
	print STDERR " --- found the start of the $tree_no th tree ...\n\n";
=cut

	    while (my $in2 = <$fh_tmp>) { ($in2 =~ m/^\{\s+Phylogenetic\s+Tree/ ) and last; }

	    my $intree = <$fh_tmp>;
	    unless (defined $intree) {
		print STDERR "[ERROR 'readin_set_trees3'] There is NO $tree_no th tree of the Newick format in the input file '${$infile}': $! !\n";
		return (undef, undef, undef);
	    }

	    $intree =~ s/^[^\(]+//;
	    $intree =~ s/[^\;]+$//;

	    my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	    unless (input_tree_file ($intree, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
		print STDERR "[ERROR 'readin_set_trees3'] The subroutine 'input_tree_file' FAILED on the $tree_no th tree in the input file '${$infile}': $! !\n";
		return (undef, undef, undef);
	    }

		## ADDED in creation of 'readin_set_trees3' (on July 2 (Fri), 2010) ##
	   my $ln_ml_msa;
	    while  (my $in2 = <$fh_tmp>) { # 2nd inner while-loop #
		($in2 =~ m/^\{\s*Log-likelihood\s+of\s+the\s+MSA/) or next;
		($ln_ml_msa) = ($in2 =~ m/\:\s+(\S+)\s*\}\s*$/);
		last;
	    }			# End of the 2nd inner while-loop #
	    unless (defined $ln_ml_msa) {
		print STDERR "[ERROR 'readin_set_trees3'] CanNOT FIND the log-ML value under the $tree_no th tree: $! !\n";
		return (undef, undef, undef);
	    }

	    my $ct_free_params = scalar keys %br_attr;
	    my $bic = -2 * $ln_ml_msa + $ct_free_params * log ($ct_clms) ;

		# REVISED on July 9 (Fri), 2010 #

	    my $aic = ($ct_clms > $ct_free_params + 1) ?
		-2 * $ln_ml_msa + 2 * $ct_free_params * $ct_clms /($ct_clms - $ct_free_params - 1) : undef;
	    ##my $aic = -2 * $ln_ml_msa + 2 * $ct_free_params;

	    unless (defined $aic) {
		print STDERR "[CAUTION 'readin_set_tree3'] The $tree_no th tree is OVERFITTING (#{columns} = $ct_clms, #{parameters} = $ct_free_params), so I caNOT calculate the AIC !\n";
	    }
	    # AIC_r = - 2 * ln (Prob (MSA | tree, model)_ML) + 2 * K * N/(N-K-1),
	    # where K = #{free parameters} and N = #{observations}.

		# End of "REVISED on July 9 (Fri), 2010" #

		## End of "ADDED in creation of 'readin_set_trees3'" ##

	    ## REVISED in creation of 'readin_set_trees3' (on July 2 (Fri), 2010) ##
	    push @set_trees, [$sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr];
	    ##push @set_trees, [$sum_neg_log_pv, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr];
	# @{$set_trees[$t]} = ($sum_neg_log_pv, $ln_ml_msa, $ct_free_params, $aic, $bic, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr) --- For the CURRENT script 'summ_tests_tree_reconst.mlt_ml.pl' (sorted in order of decreasing $aic)

	}			# End of the main outer while-loop #
	close ($fh_tmp);

	   ## REVISED in creation of 'readin_set_trees2' (on May 20 (Thu), 2010) ##
	return ($ct_clms, $ct_mono_sites, \@set_trees);
	##return \@set_trees;

}	# End of the subroutine 'readin_set_trees3' #


# log_akaike_weights:
#	This subroutine calculates logarithms of Akaike weights
#	using a list of AICs.
#	The Akaike weight of a model is given as:
#	Akaike_weight = exp (- (AIC (model) - AIC_min)/2) / [ Sum_{models} exp (- AIC (model) - AIC_min)/2) ].
#
# INPUT: (\@list_models, $indx_aic), where
#	$list_models[$c]->[$indx_aic], or $list_models[$c] if $indx_aic is undefined,
#	gives the AIC of the $c th models in the list.
#
# OUTPUT: \@ln_akaike_wgts, wehre
#	$ln_akaike_wgts[$c] is the Akaike weight of the $c th model.
#
sub log_akaike_weights (\@$) {

	my ($list_models, $indx_aic) = @_;

	my $ct_models = @{$list_models};
	unless ($ct_models>0) {
	    print STDERR "[ERROR 'log_akaike_weights'] NO MODELS in the list: $! !\n";
	    return undef;
	}
	if ($ct_models == 1) {
	    return [0]; # In this case, the Akaike weight is always 1 for the only model.
	}

	my @aics = ();
	my $min_aic;

	foreach my $model_info (@{$list_models}) { # 1st foreach-loop #
	    my $aic = (defined $indx_aic) ? $model_info->[$indx_aic] : $model_info;
	    push @aics, $aic;
	    ((defined $min_aic) and ($min_aic <= $aic)) or ($min_aic = $aic);
	}				# End of the 1st foreach-loop #

	my @ln_akaike_wgts = ();
	my $norm_factor = 0;
	foreach my $aic (@aics) { # 2nd foreach-loop 
	    my $delta = -($aic - $min_aic)/2;
	    push @ln_akaike_wgts, $delta;
	    $norm_factor += exp ($delta);
	}			# End of the 2nd foreach-loop #

	$norm_factor = log ($norm_factor);

	for (my $c=0; $c<$ct_models; $c++) { # 3rd for-loop #
	    $ln_akaike_wgts[$c] -= $norm_factor;
	}			# End of the 3rd for-loop #

	return \@ln_akaike_wgts;

}	# End of the subroutine 'log_akaike_weights' #


# ln_ml_sgl_seq:
#	This subroutine estimates the log-likelihood of obtaining a single sequence.
#	The log-likelihood is simply calculated as:
#	$lnl = Sum_$r log ($freq[$r]) ,
#	where $freq[$r] is the equilibrium frequency of the $r th amino acid residue.
#
# INPUT: (\$seq, \%aa2freq), where
#	$seq is the sequence;
#	%aa2freq = ($amino_acid (a single capital letter) => $equilibrium_frequency, ...).
#
# OUTPUT: $lnl, which is the log-likelihood.
#
#	The subroutine returns undef if it failed.
#	
sub ln_ml_sgl_seq (\$\%) {

	my ($seq, $aa2freq) = @_;

	my $cp_seq = uc ${$seq};
	my $len = length $cp_seq;

	my $lnl = 0;
	for (my $l=0; $l<$len; $l++) { # For-loop #
	    my $res = substr ($cp_seq, 0, 1, '');
	    my $frq = $aa2freq->{$res};
	    unless (defined $frq) {
		print STDERR "[ERROR 'ln_ml_sgl_seq'] Found an UNEXPECTED RESIDUE '$res' at the $l th site: $! !\n\n";
		return undef;
	    }
	    $lnl += log $frq;
	}		# End of the for-loop #

	return $lnl;

}	# End of the subroutine 'ln_ml_sgl_seq' #


# even_brlen_2seq_tree:
#	This subroutine receives a 2-node tree,
#	and evenly distributes the branch length
#	into the 2 branches, to conform to the molecular clock.
#
# INPUT: (\%br_attr, $indx_brlen), where
#	%br_attr = (ID of the node immediately below the branch => \@attr, ...),
#	with $attr[$indx_brlen] being the branch length.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub even_brlen_2seq_tree (\%$) {

	my ($br_attr, $indx_brlen) = @_;

	my @nodes = keys %{$br_attr};
	unless (@nodes == 2) {
	    print STDERR "[ERROR 'even_brlen_2seq_tree'] The input tree has ", scalar (@nodes), " NODES, instead of 2 nodes : $! !\n";
	    return 0;
	}

	my $tot_brlen = 0;
	foreach my $node (@nodes) { # The 1st foreach-loop #
	    my $attr = $br_attr->{$node};
	    unless (defined $attr) {
		print STDERR "[ERROR 'even_brlen_2seq_tree'] The node '$node' does NOT HAVE AN ATTRIBUTE for the upstream branch: $! !\n";
		return 0;
	    }
	    my $brlen = $attr->[$indx_brlen];
	    unless (defined $brlen) {
		print STDERR "[ERROR 'even_brlen_2seq_tree'] The node '$node' does NOT HAVE A BRANCH LENGTH for the upstream branch: $! !\n";
		return 0;
	    }
	    $tot_brlen += $brlen;
	}		# End of the 1st foreach-loop #

	foreach my $node (@nodes) { # The 2nd foreach-loop #
	    $br_attr->{$node}[$indx_brlen] = $tot_brlen / 2;
	}		# End of the 2nd foreach-loop #

	return 1;

}	# End of the subroutine 'even_brlen_2seq_tree' #


=for the predecessor 'ml_scoring_msa.bch.pl'
# parse_paml_main_output:
#	This subroutine parses the main output file (the 'mlc' file by default).
#
# INPUT: (\$infile), where
#	$infile is the path to the main PAML output file.
#
# OUTPUT: (\%output), where
#	%output = ('lnL' => [$lnl_1, $lnl_2, ...],
#		   'se_lnL' => [$se_lnl_1, $se_lnl_2, ...], ...),
#	with $lnl_k being the logarithmic ML value for the tree # k.
#
#	The subroutine returns undef if it failed.
#
sub parse_paml_main_output (\$) {

	my ($infile) = @_;
	local $/ = ''; # Use blank lines as input-record-separators

	my $fh_tmp = open_input_file2 (${$infile});
	unless (defined $fh_tmp) {
	     print STDERR "[ERROR 'parse_paml_main_output'] The subroutine 'open_output_file2' FAILED on the input file '${$infile}': $! !\n\n";
	     return undef;
	}

	my %output;

	while (my $in = <$fh_tmp>) { # Outer while-loop #

	     my ($k) = ($in =~ m/^TREE\s+\#\s+(\d+)\:/ );
	     (defined $k)  or next;

	    my ($lnl, $se) ;
	    my @lines = split /\n/, $in;
	    while (my $line = shift @lines) { # Inner while-loop #
		my ($tmp_lnl, $tmp_se) = ($line =~ m/^lnL\([^\)]+\)\:\s+(\S+)\s+(\S+)/ );
		(defined $tmp_lnl) or next;
		$lnl = $tmp_lnl;
		$se = $tmp_se;
		last;
	    }			# End of the inner while-loop #
	    unless (defined $lnl) {
		print STDERR "[ERROR 'parse_paml_main_output'] FAILED to extract the log likelihood for the $k th tree: $! !\n\n";
		return undef;
	    }

	    (defined $output{'lnL'}) or ($output{'lnL'} = []);
	    (defined $output{'se_lnL'}) or ($output{'se_lnL'} = []);
	    $output{'lnL'}->[$k-1] = $lnl; 
	    $output{'se_lnL'}->[$k-1] = $se;

	}		# End of the outer while-loop #
	close ($fh_tmp);

	return \%output;

}	# End of the subroutine 'parse_paml_main_output' #
=cut


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


