package MyContingency_table;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyContingency_table.pm: a package to print and manipulate contingency tables.
#
# This packege is in part based on 'Mycluster_distribution.pm'.
#
#

use strict;
use warnings;
use Exporter;
use MyProbDistri qw( chi2_lower );

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
	init_contingency_table init_2dimlist_arrays
	ctbl_rowwise_proportion tbl_rowwise_ratios
	ratios_contingency_tables add_ctbl_margins
	pr_contingency_table  pr_contingency_table_real
	pr_contingency_table_xl pr_2dim_table_xl
	pr_juxtapose_ctbls_xl pr_juxtapose_ctbls_xl2
	p_value_contingency convert_ctbl_to_2d_distri
	categorize_quantity
	);
 

# init_contingency_table: This subroutine initializes a contingency table
#				of size $n rows X $m columns (excluding the total).
#
# NOTE that the order of indices is $rf_table->[indx_ordinate (rows) ][indx_abscissa (columns)].
#
sub init_contingency_table (\@$$) {

	my ($rf_ctable, $n, $m) = @_;

	my @row_template = ();	# a row has $m entries #
	for (1 .. $m) { push @row_template, 0; }

	@{$rf_ctable} = ();	# First empty the input table. #
	for (1 .. $n) {
		my @tmp = @row_template;
		push @{$rf_ctable}, \@tmp;
	}

}			# end of the subroutine init_contingency_table #


# init_2dimlist_arrays:
#	This subroutine initializes a 2-dimensional list of arrays
#	which resembles a contingency table
#			of size $n rows X $m columns (excluding the total).
#	The difference is that each cell represents an array
#	instead of a number as a contingency table holds.
#
# NOTE that the order of indices is $rf_table->[indx_ordinate (rows) ][indx_abscissa (columns)].
#
sub init_2dimlist_arrays (\@$$) {

	my ($rf_ctable, $n, $m) = @_;

	my @row_template = ();	# a row has $m entries #
	for (1 .. $m) { push @row_template, 0; }

	@{$rf_ctable} = ();	# First empty the input table. #
	for (1 .. $n) {
	    my @tmp = ();
	    for (1 .. $m) { push @tmp, []; }
	    push @{$rf_ctable}, \@tmp;
	}

}			# end of the subroutine init_2dimlist_arrays #


# ctbl_rowwise_proportion:
#	This subroutine receives a contingency table:
#	$ct_in_cell = $ctbl[$indx_row]->[$indx_column],
#	and calculates the proportion each cell accounts for in the row.
#
# INPUT: (\@ctbl, $fl_total, $fl_prpt_total), where
#	$ct_in_cell = $ctbl[$indx_row]->[$indx_column],
#	$fl_total = 1 if  the last column is the total of the rest,
#		  = 0 otherwise.
#	$fl_prpt_total = 1 if you want to calculate the proportions for the total sum of the rows,
#			= 0 otherwise.
#
# OUTPUT:  \@prpt_ctbl, where
#	$prpt_ctbl[$indx_row]->[$indx_column]
#	    = {the proportion that the cell accounts for in the relevant row}.
#
sub ctbl_rowwise_proportion (\@$$) {

	my ($ctbl, $fl_total, $fl_prpt_total) = @_;

	my $ct_rows = @{$ctbl};
	($ct_rows > 0) or (return []);

	$fl_total = $fl_total ? 1 : 0;
	my $ct_columns = @{$ctbl->[0]} - $fl_total; # Exclude the 'Total' column #

	my @sum_rows =  ();
	for (0 .. $ct_columns ) { push @sum_rows, 0; }

	my @prpt_ctbl = ();
	for (my $i=0; $i<$ct_rows; $i++) { # the outer for-loop #

	    my $row = $ctbl->[$i];
	    my $totct_in_row = 0;
	    for (my $j=0; $j<$ct_columns; $j++) { $totct_in_row += $row->[$j]; } # the 1st inner for-loop #

	    my @prpt_in_row = ();
	    for (my $j=0; $j<$ct_columns; $j++) { # the 2nd inner for-loop #
		push @prpt_in_row, ( ($totct_in_row > 0) ? $row->[$j]/$totct_in_row : -1 ) ;
		$sum_rows[$j] += $row->[$j];
	    }				# end of the 2nd inner for-loop #

	    push @prpt_ctbl, \@prpt_in_row;
	    $sum_rows[$ct_columns] += $totct_in_row;

	}				# end of the outer for-loop #

	if ($fl_prpt_total) {
	    my @prpt_total = ();
	    my $sum_total = $sum_rows[$ct_columns];
	    for (my $j=0; $j<$ct_columns; $j++) {
		push @prpt_total, ( ($sum_total > 0) ? $sum_rows[$j]/$sum_total : -1 ) ;
	    }
	    push @prpt_ctbl, \@prpt_total;
	}
	return \@prpt_ctbl;

}			# end of the subroutine 'ctbl_rowwise_proportion' #


# tbl_rowwise_ratios:
#	This subroutine receives a table,
#	which is not necesarily a contingency table, with the elements:
#	$value_in_cell = $tbl[$indx_row]->[$indx_column],
#	and calculates the ratio of the value each cell to its reference,
#	which is placed in the last column of the same row.
#
#	DIFFERENCE from 'ctbl_rowwise_proportion':
#	Basically, 'ctbl_rowwise_proportion' can be applied ONLY to
#	contingency tables,
#	while this subroutine can also be used for non-contingency tables,
#	as long as the last column stores reference values.
#
# INPUT: (\@in_tbl, $fl_total, $fl_rt_total, $fl_append_ref), where
#	$vl_in_cell = $in_tbl[$indx_row]->[$indx_column],
#	with $vl_ref = $in_tbl[$indx_row]->[#{columns}-1]
#	giving the reference count;
#	$fl_prpt_total = 1 if you want to calculate the ratios for the total sum of the rows,
#			= 0 otherwise;
#	$fl_append_ref = 1 if you want to append the ratio of the reference to itself;
#			= 0 otherwise;
#
# OUTPUT:  \@tbl_ratios, where
#	$tbl_ratios[$indx_row]->[$indx_column]
#	    = {the input value in the cell} / {the reference value for the row} .
#
sub tbl_rowwise_ratios (\@$$) {

	my ($in_tbl, $fl_rt_total, $fl_append_ref) = @_;

	my $ct_rows = @{$in_tbl};
	($ct_rows > 0) or (return []);

	my $ct_columns = @{$in_tbl->[0]} - 1; # Exclude the 'reference' column #

	my @sum_rows =  ();
	for (0 .. $ct_columns) { push @sum_rows, 0; }

	my @tbl_ratios = ();
	for (my $i=0; $i<$ct_rows; $i++) { # the outer for-loop #

	    my $row = $in_tbl->[$i];
	    my $ref_value = $row->[$ct_columns];

	    my @ratios_in_row = ();
	    for (my $j=0; $j<$ct_columns; $j++) { # the 2nd inner for-loop #
		push @ratios_in_row, ( ($ref_value > 0) ? $row->[$j]/$ref_value : -1 ) ;
		$sum_rows[$j] += $row->[$j];
	    }				# end of the 2nd inner for-loop #

	    $fl_append_ref and (push @ratios_in_row, (($ref_value>0) ? 1 : -1) );
	    push @tbl_ratios, \@ratios_in_row;
	    $sum_rows[$ct_columns] += $ref_value;

	}				# end of the outer for-loop #

	if ($fl_rt_total) {
	    my @ratios_total = ();
	    my $sum_ref = $sum_rows[$ct_columns];
	    for (my $j=0; $j<$ct_columns; $j++) {
		push @ratios_total, (($sum_ref > 0) ? $sum_rows[$j]/$sum_ref : -1);
	    }
	    $fl_append_ref and (push @ratios_total, (($sum_ref>0) ? 1 : -1) );
	    push @tbl_ratios, \@ratios_total;
	}

	return \@tbl_ratios ;

}			# end of the subroutine 'tbl_rowwise_ratios' #


# ratios_contingency_tables:
#	This subroutine receives 2 contingency tables with the identical dimensions,
#	then constructs a table of ratios taken for each cells.
#
# INPUT: (\@ctbl_numer, \@ctbl_denom, $fl_rowsum, $fl_clmsum, $format), where
#	@ctbl_numer is a contingency table whose cells are to be used as numerators,
#	@ctbl_denom is a contingency table whose cells are to be used as denominators.
#	$fl_rowsum = 1 if you want to calculate the ratio of the sums for each row as well,
#		   = 0 otherwise.
#	$fl_clmsum = 1 if you want to calculate the ratio of the sums for each column as  well,
#		   = 0 otherwise.
#
# OUTPUT: \@table_ratios, where
#	$table_ratios[$indx_row]->[$indx_clm]
#	    = $ctbl_numer[$indx_row]->[$indx_clm] /$ctbl_denom[$indx_row]->[$indx_clm].
#
sub ratios_contingency_tables (\@\@$$$) {

	my ($ctbl_numer, $ctbl_denom, $fl_rowsum, $fl_clmsum, $format) = @_;

	my $ct_row = @{$ctbl_numer};
	($ct_row > 0) or (return []);
	unless ($ct_row == @{$ctbl_denom}) {
	    printf STDERR "The numbers of rows for the numerator ($ct_row) CLASHES that for the denominator (%d).\n", scalar (@{$ctbl_denom}) ;
	    return undef;
	}
	my $ct_clm = @{$ctbl_numer->[0]};
	unless ($ct_clm == @{$ctbl_denom->[0]}) {
	   printf STDERR "The numbers of columns for the numerator ($ct_clm) CLASHES that for the denominator (%d).\n", scalar (@{$ctbl_denom->[0]}) ;
	}

	$fl_rowsum = $fl_rowsum ? 1 : 0;
	$fl_clmsum = $fl_clmsum ? 1 : 0;

 	my @numer_clmsums = my @denom_clmsums = ();

	for (1 .. $ct_clm + $fl_rowsum) { # the 1st for-loop #
	    push @numer_clmsums, 0;
	    push @denom_clmsums, 0;
	}			# end of the 1st for-loop #

	my @table_ratios = ();
	for (my $i=0; $i<$ct_row; $i++) { # the loop for rows #

	    my @ratios_for_row = ();
	    my $numer_rowsum  = my $denom_rowsum = 0;
	    for (my $j=0; $j<$ct_clm; $j++) { # the loop for columns #
		my $numer = $ctbl_numer->[$i][$j];
		my $denom = $ctbl_denom->[$i][$j];
		$numer_rowsum += $numer;
		$denom_rowsum += $denom;
		push @ratios_for_row, 
		   ( ($denom == 0) ? 'NA' : sprintf ($format, $numer/$denom) );
		if ($fl_clmsum) {
		    $numer_clmsums[$j] += $numer;
		    $denom_clmsums[$j] += $denom;
		}
	    }				# end of the loop for columns #
	    if ($fl_rowsum) {
		push @ratios_for_row,  
		   ( ($denom_rowsum == 0) ? 'NA' :
			 sprintf ($format, $numer_rowsum /$denom_rowsum) ) ;
		if ($fl_clmsum) {
		    $numer_clmsums[$ct_clm] += $numer_rowsum;
		    $denom_clmsums[$ct_clm] += $denom_rowsum;
		}
	    }
	    push @table_ratios, \@ratios_for_row;

	}			# end of the loop for rows #

	if ($fl_clmsum) {
	    my @ratios_for_clmsum = ();
	    $ct_clm += $fl_rowsum;
	    for (my $j=0; $j<$ct_clm; $j++) { # the loop for column-sums  
		my $numer = $numer_clmsums[$j];
		my $denom = $denom_clmsums[$j];
		push @ratios_for_clmsum,
		    ( ($denom == 0) ? 'NA' : sprintf ($format, $numer/$denom) ) ;
	    }				# end of the loop for column-sums
	    push @table_ratios, \@ratios_for_clmsum;
	}

	return \@table_ratios;

}		# end of the subroutine 'ratios_contingency_tables' #


# add_ctbl_margins:
#	This subroutine receives a contingency table,
#	then adds sums of columns and/or rows as marginal information .
#
# INPUT: (\@ctbl, $fl_rowsum, $fl_clmsum), where
#	$ctbl[$i-1]->[$j-1] gives a cell  in the intersection of
#	the $i-th row and the $j-th column.
#	$fl_rowsum = 1 if you want to add the sums for each row as an additional column,
#		   = 0 otherwise.
#	$fl_clmsum = 1 if you want to add the sums for each column as an additonal row,
#		   = 0 otherwise.
#
# OUTPUT: \@ctbl_extended, where
#	$ctbl_extended[$#ctbl_extended]->[$j-1] gives the sum of the $j-th column,
#	$ctbl_extended[$i-1]->[$last_indx] givs the sum for the $i-th row.
#
sub add_ctbl_margins (\@$$) {

	my ($ctbl, $fl_rowsum, $fl_clmsum) = @_;

	my $ct_row = @{$ctbl};
	($ct_row > 0) or (return []);
	my $ct_clm = @{$ctbl->[0]};
	$fl_rowsum = $fl_rowsum ? 1 : 0;
	$fl_clmsum = $fl_clmsum ? 1 : 0;

 	my @clmsums = ();

	for (1 .. $ct_clm + $fl_rowsum) { push @clmsums, 0; }

	my @ctbl_extended = ();
	for (my $i=0; $i<$ct_row; $i++) { # the loop for rows #

	    my @new_row = ();
	    my $rowsum  = 0;
	    for (my $j=0; $j<$ct_clm; $j++) { # the loop for columns #
		my $cell = $ctbl->[$i][$j];
		$rowsum += $cell;
		push @new_row, $cell;
		$fl_clmsum and ($clmsums[$j] += $cell);
	    }				# end of the loop for columns #
	    if ($fl_rowsum) {
		push @new_row, $rowsum; 
		$fl_clmsum and ($clmsums[$ct_clm] += $rowsum);
	    }
	    push @ctbl_extended, \@new_row;

	}			# end of the loop for rows #

	$fl_clmsum and (push @ctbl_extended, \@clmsums) ;

	return \@ctbl_extended;

}		# end of the subroutine 'add_ctbl_margins' #


# pr_contingency_table: This subroutine prints out a contingency table.
#
# PREREQUISITE: Ordinate does not include 'Total'.
#		(Abscissa may or may not include it).
#	'abscissa' ---  tag for a column.
#	'ordinate'  --- tag for a row.
#
sub pr_contingency_table (\@\@\@\*) {

	my ($rf_ctable, $rf_tags_abscissa, $rf_tags_ordinate, $fh) = @_;
	my $mytab = ' ' x 3;
	my $field_width = 7;
	my $tag_width = length('Categories');
	my $ct_ordinates = @{$rf_tags_ordinate};
	my $ct_abscissa = @{$rf_tags_abscissa};

	unless (@{$rf_ctable} == $ct_ordinates) {
		my $ct_rows = @{$rf_ctable};
		print STDERR "\#{ordinates} (=$ct_ordinates) does not agree with \#{rows} (=$ct_rows) in the input of \'pr_contingency_table\'.";
		return 0;
	}
	unless (@{$rf_ctable->[0]} == $ct_abscissa) {
		my $ct_clms = @{$rf_ctable->[0]};
		print STDERR "\#{abscissa} (=$ct_abscissa) does not agree with \#{columns} (=$ct_clms) in the input of \'pr_contingency_table\'.";
		return 0;
	}

	for my $tag (@{$rf_tags_abscissa}) { (length($tag) > $field_width) and ($field_width = length($tag)); }
	for my $tag (@{$rf_tags_ordinate}) { (length($tag) > $tag_width) and ($tag_width = length($tag) ); }
	my $fl_total = ( uc ($rf_tags_abscissa->[$ct_abscissa -1]) eq 'TOTAL') ? 1 : 0;

	my @tags_abscissa = @{$rf_tags_abscissa};
	unless ($fl_total) {
		push @tags_abscissa, 'Total';
		$ct_abscissa++;
	}
	my $format_headline = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa	. "\n";
	##my $format_row = "%-${tag_width}s\:" . "$mytab%${field_width}d" x $ct_abscissa . "\n";
	my $format_row = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa . "\n"; # a minor revision on Nov 8 (Tue), 2005

	my $headline = sprintf $format_headline, 'Categories', @tags_abscissa;
	print $fh $headline;
	print $fh '-' x length($headline) . "\n";  # upper-delimiting line #

	my @total = ();
	for (1 .. $ct_abscissa) { push @total, 0; }

	for (my $i=0; $i<$ct_ordinates; $i++) {
		my @row = @{$rf_ctable->[$i]};
		unless ($fl_total) {
			my $total = 0;
			for my $ct (@row) { $total += $ct; }
			push @row, $total;
		}
		for (my $j=0; $j<$ct_abscissa; $j++) { $total[$j] += $row[$j]; }

		printf $fh $format_row , $rf_tags_ordinate->[$i], @row;
	}

	print $fh '-' x length($headline) . "\n";  # lower-delimiting line #
	printf $fh $format_row, 'Total', @total;
	printf $fh "\n";

	return 1;

}	# end of the subroutine pr_contingency_table #


# pr_contingency_table_real:
#	This subroutine prints out a contingency table-like table
#	whose cells can hold general real values, instead of integers only.
#	The subroutine takes an option to determine whether or not
#	it will display the total (marginal) values as well.
#	$option = 1 specifies that it should print the total, = 0 otherwise. 
#
# PREREQUISITE: Ordinate (tags for the rows) does not include 'Total'.
#		(Abscissa (tags for the columns) may or may not include it).
#
#	REVISED on July 31 (Thu), 2008:
#		The subroutine now can cope with the situation where
#		the ordinate include 'Total'.
#
sub pr_contingency_table_real (\@\@\@\*$$) {

	my ($rf_ctable, $rf_tags_abscissa, $rf_tags_ordinate, $fh, $format, $option) = @_;
	#my $mytab = ' ' x 3;
	my $mytab = "\t";
	my ($pre_spec, $field_width) = ($format =~ m/^\s*\%([\D0]*)(\d+)/ );
	my ($post_spec) = ($format =~ m/(\.[^\.]+)\s*$/ );
	(defined $pre_spec) or ($pre_spec = '');
	(defined $field_width) or ($field_width = 7);
	my $old_fw = $field_width;
	my $tag_width = length('Categories');
	my $ct_ordinates = @{$rf_tags_ordinate};
	my $ct_abscissa = @{$rf_tags_abscissa};

	unless (@{$rf_ctable} == $ct_ordinates) {
		print STDERR "\#{ordinates} does not agree with \#{rows} in the input of \'pr_contingency_table\'.";
		return 0;
	}
	unless (@{$rf_ctable->[0]} == $ct_abscissa) {
		print STDERR "\#{abscissa} does not agree with \#{columns} in the input of \'pr_contingency_table\'.";
		return 0;
	}

	for my $tag (@{$rf_tags_abscissa}) { (length($tag) > $field_width) and ($field_width = length($tag)); }
	for my $tag (@{$rf_tags_ordinate}) { (length($tag) > $tag_width) and ($tag_width = length($tag) ); }
	my $fl_total = ( uc ($rf_tags_abscissa->[$ct_abscissa -1]) eq 'TOTAL') ? 1 : 0;

		## ADDED on July 31 (Thu), 2008 ##
	my $fl_total2 = ( uc ($rf_tags_ordinate->[$ct_ordinates-1]) eq 'TOTAL') ? 1 : 0;
	

	my @tags_abscissa = @{$rf_tags_abscissa};
	if ($option and !$fl_total) {
		push @tags_abscissa, 'Total';
		$ct_abscissa++;
	}
	my $format_headline = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa	. "\n";
	my $new_format = "%${pre_spec}${field_width}";
	$new_format .= (defined $post_spec) ? $post_spec : "d";
	#my $format_row = "%-${tag_width}s\:" . "$mytab%${field_width}d" x $ct_abscissa . "\n";
	my $format_row = "%-${tag_width}s\:" . "$mytab$new_format" x $ct_abscissa . "\n";

	my $headline = sprintf $format_headline, 'Categories', @tags_abscissa;
	my $len_horiz_line = length ($headline);
	if ($mytab =~ m/\t/ ) {
	    $len_horiz_line += 6 * scalar (@tags_abscissa) ;
	    $fl_total and ($len_horiz_line += 6);
	}
	print $fh $headline;
	print $fh '-' x $len_horiz_line . "\n";  # upper-delimiting line #

	my @total = ();
	for (1 .. $ct_abscissa) { push @total, 0; }

	for (my $i=0; $i<$ct_ordinates; $i++) {
		my @row = @{$rf_ctable->[$i]};
		if ($option and !$fl_total) {
			my $total = 0;
			for my $ct (@row) { $total += $ct; }
			push @row, $total;
		}
		if ($option) {
		   for (my $j=0; $j<$ct_abscissa; $j++) { $total[$j] += $row[$j]; }
		}

		printf $fh $format_row , $rf_tags_ordinate->[$i], @row;
	}

	print $fh '-' x $len_horiz_line . "\n";  # lower-delimiting line #

		## REVISED on July 31 (Thu), 2008 ##
	($option and !$fl_total2) and (printf $fh $format_row, 'Total', @total);
	##$option and (printf $fh $format_row, 'Total', @total);
	printf $fh "\n";

	return 1;

}	# end of the subroutine pr_contingency_table_real #


# pr_contingency_table_xl: This subroutine prints out a contingency table
#			in a format which can feed into the Microsoft Excel.
#
# PREREQUISITE: Ordinate (tags for rows) does not include 'Total'.
#		(Abscissa (tags for columns) may or may not include it).
#
sub pr_contingency_table_xl (\@\@\@\*) {

	my ($rf_ctable, $rf_tags_abscissa, $rf_tags_ordinate, $fh) = @_;
	#my $mytab = ' ' x 3;
	my $mytab = "\t";
	##my $field_width = 7;
	my $field_width = 10;
	my $tag_width = length('Categories');
	my $ct_ordinates = @{$rf_tags_ordinate};
	my $ct_abscissa = @{$rf_tags_abscissa};

	unless (@{$rf_ctable} == $ct_ordinates) {
		print STDERR "\#{ordinates} does not agree with \#{rows} in the input of \'pr_contingency_table\'.";
		return 0;
	}
	unless (@{$rf_ctable->[0]} == $ct_abscissa) {
		print STDERR "\#{abscissa} does not agree with \#{columns} in the input of \'pr_contingency_table\'.";
		return 0;
	}

	for my $tag (@{$rf_tags_abscissa}) { (length($tag) > $field_width) and ($field_width = length($tag)); }
	for my $tag (@{$rf_tags_ordinate}) { (length($tag) > $tag_width) and ($tag_width = length($tag) ); }
	my $fl_total = ( uc ($rf_tags_abscissa->[$ct_abscissa -1]) eq 'TOTAL') ? 1 : 0;

	my @tags_abscissa = @{$rf_tags_abscissa};
	unless ($fl_total) {
		push @tags_abscissa, 'Total';
		$ct_abscissa++;
	}
	my $format_headline = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa	. "\n";
	#my $format_row = "%-${tag_width}s\:" . "$mytab%${field_width}d" x $ct_abscissa . "\n";
	##my $format_row = "%-${tag_width}s" . "$mytab%${field_width}d" x $ct_abscissa . "\n";
	my $format_row = "%-${tag_width}s" . "$mytab%${field_width}s" x $ct_abscissa . "\n";

	my $headline = sprintf $format_headline, 'Categories', @tags_abscissa;
	print $fh $headline;
	#print $fh '-' x length($headline) . "\n";  # upper-delimiting line #
	print $fh "\n";  # upper-delimiting line #

	my @total = ();
	for (1 .. $ct_abscissa) { push @total, 0; }

	for (my $i=0; $i<$ct_ordinates; $i++) {
		my @row = @{$rf_ctable->[$i]};
		unless ($fl_total) {
			my $total = 0;
			for my $ct (@row) { $total += $ct; }
			push @row, $total;
		}
		for (my $j=0; $j<$ct_abscissa; $j++) { $total[$j] += $row[$j]; }

		printf $fh $format_row , $rf_tags_ordinate->[$i], @row;
	}

	#print $fh '-' x length($headline) . "\n";  # lower-delimiting line #
	print $fh "\n";  # lower-delimiting line #
	printf $fh $format_row, 'Total', @total;
	printf $fh "\n";

	return 1;

}	# end of the subroutine pr_contingency_table_xl #



# pr_2dim_table_xl: This subroutine prints out a 2-dimensional table
#			in a format which can feed into the Microsoft Excel.
#
# INPUT: (\@table, \@tags_abscissa, \@tags_ordiname, \*$fh), where
#	$table[$indx_row]->[$indx_column] = $cell,
#	\@tags_abscssa = ( tags for absicca, or columns),
#	\@tags_ordinate = ( tags for ordinate, or rows),
#
# NOTICE: This subroutine does not automatically compute Totals of rows or columns.
#	  It just prints out what it received as arguments.
#	If you want to print out the summations, use 'pr_contingency_table(_fmt/xl).
#
sub pr_2dim_table_xl (\@\@\@\*) {

	my ($rf_table, $rf_tags_abscissa, $rf_tags_ordinate, $fh) = @_;
	#my $mytab = ' ' x 3;
	my $mytab = "\t";
	##my $field_width = 7;
	my $field_width = 10;
	my $tag_width = length('Categories');
	my $ct_ordinates = @{$rf_tags_ordinate};
	my $ct_abscissa = @{$rf_tags_abscissa};

	unless (@{$rf_table} == $ct_ordinates) {
		print STDERR "\#{ordinates} does not agree with \#{rows} in the input of \'pr_2dim_table_xl\'.";
		return 0;
	}
	unless (@{$rf_table->[0]} == $ct_abscissa) {
		print STDERR "\#{abscissa} does not agree with \#{columns} in the input of \'pr_contingency_table\'.";
		return 0;
	}

	for my $tag (@{$rf_tags_abscissa}) { (length($tag) > $field_width) and ($field_width = length($tag)); }
	for my $tag (@{$rf_tags_ordinate}) { (length($tag) > $tag_width) and ($tag_width = length($tag) ); }
	##my $fl_total = ( uc ($rf_tags_abscissa->[$ct_abscissa -1]) eq 'TOTAL') ? 1 : 0;

	my @tags_abscissa = @{$rf_tags_abscissa};
	##unless ($fl_total) {
	##	push @tags_abscissa, 'Total';
	##	$ct_abscissa++;
	##}
	my $format_headline = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa	. "\n";
	#my $format_row = "%-${tag_width}s\:" . "$mytab%${field_width}d" x $ct_abscissa . "\n";
	##my $format_row = "%-${tag_width}s" . "$mytab%${field_width}d" x $ct_abscissa . "\n";
	my $format_row = "%-${tag_width}s" . "$mytab%${field_width}s" x $ct_abscissa . "\n";

	my $headline = sprintf $format_headline, 'Categories', @tags_abscissa;
	print $fh $headline;
	print $fh "\n";  # upper-delimiting line #

	##my @total = ();
	##for (1 .. $ct_abscissa) { push @total, 0; }

	for (my $i=0; $i<$ct_ordinates; $i++) {
		my @row = @{$rf_table->[$i]};
		##unless ($fl_total) {
		##	my $total = 0;
		##	for my $ct (@row) { $total += $ct; }
		##	push @row, $total;
		##}
		##for (my $j=0; $j<$ct_abscissa; $j++) { $total[$j] += $row[$j]; }
		printf $fh $format_row , $rf_tags_ordinate->[$i], @row;
	}

	print $fh "\n";  # lower-delimiting line #
	##printf $fh $format_row, 'Total', @total;
	##printf $fh "\n";

	return 1;

}	# end of the subroutine pr_2dim_table_xl #


# pr_juxtapose_ctbls_xl:
#	This subroutine prints out MULTIPLE 2-dimensional tables with the same dimensions
#	in a juxtaposed manner, so that we can compare elements of different tables
#	corresponding to the same cell vertically.
#	The print out is conducted in a format which can feed into the Microsoft Excel.
#
# INPUT: (\@tables, \@tags_tables, \@tags_abscissa, \@tags_ordiname, $categ_ordi, $categ_absc, \*$fh), where
#	$table[$k-1]->[$indx_row][$indx_column] = $cell in the k-th table,
#	@tags_abscssa = ( tags for absicca, or columns),
#	@tags_ordinate = ( tags for ordinate, or rows),
#
# NOTICE: This subroutine does not automatically compute Totals of rows or columns.
#	  It just prints out what it received as arguments.
#	If you want to print out the summations, use 'pr_contingency_table(_fmt/xl).
#
sub pr_juxtapose_ctbls_xl (\@\@\@\@$$\*) {

	my ($tables, $tags_tables, $tags_abscissa, $tags_ordinate, $categ_ordi, $categ_absc, $fh) = @_;
	my $mytab = "\t";
	my $field_width = 7;
	##my $tag_width = length('Categories');
	my $maxlen_tag_ord = length ($categ_ordi);
	my $maxlen_tag_tbl = 0;

	my $ct_ordinates = @{$tags_ordinate};
	my $ct_abscissa = @{$tags_abscissa};
	my $ct_tbls = @{$tags_tables};

	unless ($ct_tbls > 0) {
	   print STDERR "NO TABLES in the table list for the input of 'pr_juxtapose_ctbls_xl'.\n";
	   return 0;
	}
	unless ($ct_tbls == @{$tables}) {
	    printf STDERR "\#{tables} (= %d) CLASHES with that of tags (= $ct_tbls) in the input of 'pr_juxtapose_ctbls_xl'.\n", scalar (@{$tables}) ;
	    return 0;
	}

	foreach my $tag (@{$tags_abscissa}) {
	    ($field_width < length ($tag) ) and ($field_width = length ($tag) );
	}

	foreach my $tag (@{$tags_ordinate}) {
	    ($maxlen_tag_ord < length ($tag) ) and ($maxlen_tag_ord = length ($tag) );
	}
	foreach my $tag (@{$tags_tables}) {
	    ($maxlen_tag_tbl < length ($tag) ) and ($maxlen_tag_tbl = length ($tag) );
	}


	unless (@{$tables->[0]} == $ct_ordinates) {
		printf STDERR "\#{ordinates} (= $ct_ordinates) does not agree with \#{rows} (= %d) in the input of \'pr_2dim_table_xl\'.", scalar (@{$tables->[0]}) ;
		return 0;
	}
	unless (@{$tables->[0][0]} == $ct_abscissa) {
		printf STDERR "\#{abscissa} (= $ct_abscissa) does not agree with \#{columns} (= %d) in the input of \'pr_contingency_table\'.", scalar (@{$tables->[0][0]}) ;
		return 0;
	}

	my @tags_abscissa = @{$tags_abscissa};
	##my $format_headline = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa	. "\n";
	## my $format_row = "%-${tag_width}s" . "$mytab%${field_width}s" x $ct_abscissa . "\n";
	my $format_headline = "%-${maxlen_tag_ord}s" . "$mytab%-${maxlen_tag_tbl}s" .
			"$mytab%${field_width}s" x $ct_abscissa . "\n";
	my $format_row = "%-${maxlen_tag_ord}s" . "$mytab%-${maxlen_tag_tbl}s" .
			"$mytab%${field_width}s" x $ct_abscissa . "\n";

	##my $headline = sprintf $format_headline, 'Categories', @tags_abscissa;
	my $headline = sprintf $format_headline, $categ_ordi, $categ_absc, @tags_abscissa;
	print $fh $headline;
	print $fh "\n";  # upper-delimiting line #

	for (my $i=0; $i<$ct_ordinates; $i++) {# the loop for the ordinates (rows) #
		##my @row = @{$table->[$i]};
		##printf $fh $format_row , $tags_ordinate->[$i], @row;
	    for (my $k=0; $k<$ct_tbls; $k++) { # the loop for the juxtaposition #
		my $tag_ordi = ($k == 0) ? $tags_ordinate->[$i] : ''; 
		printf $fh $format_row, $tag_ordi, $tags_tables->[$k], @{$tables->[$k][$i]};
	    }				# end of the loop for the juxtaposition #
	    print $fh "\n";
	}				# end of the loop for the ordinates (rows) #

	print $fh "\n";  # lower-delimiting line #

	return 1;

}	# end of the subroutine pr_juxtapose_ctbls_xl #


# pr_juxtapose_ctbls_xl2:
#	This revised version of 'pr_juxtapose_ctbls_xl' can specify the
#	'minimum width' for each field in the table.
#	(CREATED on June 3 (Fri), 2011)
#
# pr_juxtapose_ctbls_xl:
#	This subroutine prints out MULTIPLE 2-dimensional tables with the same dimensions
#	in a juxtaposed manner, so that we can compare elements of different tables
#	corresponding to the same cell vertically.
#	The print out is conducted in a format which can feed into the Microsoft Excel.
#
# INPUT: (\@tables, \@tags_tables, \@tags_abscissa, \@tags_ordiname, $categ_ordi, $categ_absc, \*$fh, $min_fld_width), where
#	$table[$k-1]->[$indx_row][$indx_column] = $cell in the k-th table;
#	@tags_abscssa = ( tags for absicca, or columns);
#	@tags_ordinate = ( tags for ordinate, or rows);
#	$categ_ordi and $categ_absc specify the category names for the ordinates and the absicca, respectively;
#	$min_fld_width is the minimum field width.
#
# NOTICE: This subroutine does not automatically compute Totals of rows or columns.
#	  It just prints out what it received as arguments.
#	If you want to print out the summations, use 'pr_contingency_table(_fmt/xl).
#
sub pr_juxtapose_ctbls_xl2 (\@\@\@\@$$\*$) {

	my ($tables, $tags_tables, $tags_abscissa, $tags_ordinate, $categ_ordi, $categ_absc, $fh, $min_fld_width) = @_;
	my $mytab = "\t";

	   # REVISED in creation of 'pr_juxtapose_ctbls_xl2' on June 3 (Wed), 2011 #
	my $field_width = ($min_fld_width > 7) ? $min_fld_width : 7;
	##my $field_width = 7;
	##my $tag_width = length('Categories');
	my $maxlen_tag_ord = length ($categ_ordi);
	   # ADDED in creation of 'pr_juxtapose_ctbl_xl2' on June 3 (Fri), 2011 #
	($maxlen_tag_ord < $min_fld_width) and ($maxlen_tag_ord = $min_fld_width);
	foreach my $tag (@{$tags_ordinate}) {
	    my $len_tag = length $tag;
	    ($maxlen_tag_ord < $len_tag) and ($maxlen_tag_ord = $len_tag);
	}

	my $maxlen_tag_tbl = 0;

	my $ct_ordinates = @{$tags_ordinate};
	my $ct_abscissa = @{$tags_abscissa};
	my $ct_tbls = @{$tags_tables};

	unless ($ct_tbls > 0) {
	   print STDERR "[ERROR 'pr_juxtapose_ctbls_xl2'] NO TABLES in the table list for the input of 'pr_juxtapose_ctbls_xl'.\n";
	   return 0;
	}
	unless ($ct_tbls == @{$tables}) {
	    printf STDERR "[ERROR 'pr_juxtapose_ctbls_xl2'] \#{tables} (= %d) CLASHES with that of tags (= $ct_tbls) in the input of 'pr_juxtapose_ctbls_xl'.\n", scalar (@{$tables}) ;
	    return 0;
	}

	foreach my $tag (@{$tags_abscissa}) {
	    ($field_width < length ($tag) ) and ($field_width = length ($tag) );
	}

	foreach my $tag (@{$tags_ordinate}) {
	    ($maxlen_tag_ord < length ($tag) ) and ($maxlen_tag_ord = length ($tag) );
	}
	foreach my $tag (@{$tags_tables}) {
	    ($maxlen_tag_tbl < length ($tag) ) and ($maxlen_tag_tbl = length ($tag) );
	}


	unless (@{$tables->[0]} == $ct_ordinates) {
		printf STDERR "[ERROR 'pr_juxtapose_ctbls_xl2'] \#{ordinates} (= $ct_ordinates) does not agree with \#{rows} (= %d) in the input of \'pr_2dim_table_xl\'.", scalar (@{$tables->[0]}) ;
		return 0;
	}
	unless (@{$tables->[0][0]} == $ct_abscissa) {
		printf STDERR "[ERROR 'pr_juxtapose_ctbls_xl2'] \#{abscissa} (= $ct_abscissa) does not agree with \#{columns} (= %d) in the input of \'pr_contingency_table\'.", scalar (@{$tables->[0][0]}) ;
		return 0;
	}

	my @tags_abscissa = @{$tags_abscissa};
	##my $format_headline = "%-${tag_width}s\:" . "$mytab%${field_width}s" x $ct_abscissa	. "\n";
	## my $format_row = "%-${tag_width}s" . "$mytab%${field_width}s" x $ct_abscissa . "\n";
	my $format_headline = "%-${maxlen_tag_ord}s" . "$mytab%-${maxlen_tag_tbl}s" .
			"$mytab%${field_width}s" x $ct_abscissa . "\n";
	my $format_row = "%-${maxlen_tag_ord}s" . "$mytab%-${maxlen_tag_tbl}s" .
			"$mytab%${field_width}s" x $ct_abscissa . "\n";

	##my $headline = sprintf $format_headline, 'Categories', @tags_abscissa;
	my $headline = sprintf $format_headline, $categ_ordi, $categ_absc, @tags_abscissa;
	print $fh $headline;
	print $fh "\n";  # upper-delimiting line #

	for (my $i=0; $i<$ct_ordinates; $i++) {# the loop for the ordinates (rows) #
		##my @row = @{$table->[$i]};
		##printf $fh $format_row , $tags_ordinate->[$i], @row;
	    for (my $k=0; $k<$ct_tbls; $k++) { # the loop for the juxtaposition #
		my $tag_ordi = ($k == 0) ? $tags_ordinate->[$i] : ''; 
		printf $fh $format_row, $tag_ordi, $tags_tables->[$k], @{$tables->[$k][$i]};
	    }				# end of the loop for the juxtaposition #
	    print $fh "\n";
	}				# end of the loop for the ordinates (rows) #

	print $fh "\n";  # lower-delimiting line #

	return 1;

}	# end of the subroutine pr_juxtapose_ctbls_xl2 #


# p_value_contingency: This subroutine receives a contingency table,
#			and returns the DOF, test-statistic, and the P-value.
#
# PREREQUISITE: The table does not include a row for subtotals of columns,
#		(while each row may include its subtotal).
# 		--- $fl_subtotal = 1 if each row includes subtotal, =0 otherwise. 
# 
sub p_value_contingency (\@$) {

	my ($rf_ctable, $fl_subtotal) = @_;

	my $ct_rows = @{$rf_ctable};
	($ct_rows > 1) or (return (0, 0, 1) );
	my $ct_columns = @{$rf_ctable->[0]};
	$fl_subtotal and ($ct_columns--);
	($ct_columns > 1) or (return (0,0, 1) );

	my $tr_ct_rows = $ct_rows;
	my $tr_ct_columns = $ct_columns;

		# prepare subtotals #
	my @tot_rows = ();
	for (1 .. $ct_rows) { push @tot_rows, 0; }
	my @tot_columns = ();
	for (1 .. $ct_columns) { push @tot_columns, 0; }

	for (my $i=0; $i<$ct_rows; $i++) {	# outer for #
		for (my $j=0; $j<$ct_columns; $j++) {	# inner for #

			$tot_rows[$i] += $rf_ctable->[$i][$j];
			$tot_columns[$j] += $rf_ctable->[$i][$j];

		}				# end of the inner for #
	}				# end of the outer for #

		# calculation of the total -- up for streamlining #
	my $total = my $total_2 = 0;

		# A row or a column with zero subtotal is regarded as null #
	foreach my $subtotal (@tot_rows) {
		$subtotal or $tr_ct_rows--;
		 $total += $subtotal;
	}
	foreach my $subtotal (@tot_columns) {
		$subtotal or $tr_ct_columns--;
		$total_2 += $subtotal;
	}

	my $dof = ($tr_ct_rows -1 ) * ($tr_ct_columns -1 );

		## MODIFIED to cope with a table with non-integer entries. (on Jan 18 (Sat), 2014). ##

	my $av_total = ($total+$total_2)/2;
	my $diff_total = ($total > $total_2) ? $total - $total_2 : $total_2 - $total;
	my $eps = 1E-09;

	unless ($diff_total < $eps * $av_total) {
	##unless ($total == $total_2) {
		print STDERR "The total ($total) calculated from rows  clashes with that ($total_2) from columns.\n";
		return ($dof, undef, undef);
	} 

	$total or (return ($dof, 0, 1) );
	$dof or (return (0,0, -1) );	  # This indicates something unusual
					  #  has happened. 

	
		# calculate the test statistic #
	my $test_stat = 0;

	for (my $i=0; $i<$ct_rows; $i++) {	# outer for #
		for (my $j=0; $j<$ct_columns; $j++) {	# inner for #

			my $obs = $rf_ctable->[$i][$j];
			$obs  or next;

			my $expect = $tot_rows[$i]*$tot_columns[$j]/$total;
			$test_stat += $obs * log ($obs/$expect);

		}				# end of the inner for #
	}				# end of the outer for #

	$test_stat *= 2;

		## ADDED to cope with a negative infinitesimal $test_stat. (on Jan 18 (Sat), 2014). ##

	my $eps2 = 1E-07;
	if (($test_stat<0) and ($test_stat > - $eps2)) {
	    print STDERR "The test statistics (=$test_stat) is infinitesimally negative: $! ! ... I will reset it to ZERO ...\n\n";
	    $test_stat = 0;
	}

	my $p_value = chi2_lower ($dof, $test_stat);

	return ($dof, $test_stat, $p_value);

}		# end of the subroutine p_value_contingency #


# convert_ctbl_to_2d_distri:
#	This subroutine convert a contingency table to a
#	2D distribution so that their rows and columns are kept as they are
#	when displayed with 'pr_contingency_table' for the former
#	and with 'batch_pr_file_distri' (or 'pr_file_2d_distri') for the latter.
#
# INPUT: (\@ctbl_row_vs_clm, $fl_tot_row, $fl_tot_clm), where
#	$ctbl_row_vs_clm[$indx_row]->[$indx_clm] is the count of elements
#	belonging to the intersection of the $indx_row th row 
#	and the $indx_clm th column;
#	If $fl_tot_row = 1, the last row stores the subtotal count of
#	the elements belonging to each fixed column category;
#	If $fl_tot_clm = 1, the last column stores the subtotal count of
#	the elements belonging to each fixed row category;
#
# OUPUT: \@bins_clm_vs_row, where
#	$bins_clm_vs_row[$indx_clm]->[$indx_row] is the count of elements
#	belonging to the intersection of the $indx_clm th column
#	and  the $indx_row th row,
#	and $bins_clm_vs_row[$#bins_clm_vs_row]->[$indx_row] is
#	the subtotal count of the elements beloging to the $indx_row th row.
#	
sub convert_ctbl_to_2d_distri (\@$$) {

	my ($ctbl_row_vs_clm, $fl_tot_row, $fl_tot_clm) = @_;

	my $ct_rows = @{$ctbl_row_vs_clm};
	$fl_tot_row and ($ct_rows--);
	($ct_rows>0) or (return undef);
	my $ct_clms = @{$ctbl_row_vs_clm->[0]};
	$fl_tot_clm and ($ct_clms--);
	($ct_clms>0) or (return undef);

	my @bins_clm_vs_row = ();
	for (0 .. $ct_clms) { push @bins_clm_vs_row, []; }

	for (my $row=0; $row<$ct_rows; $row++) {
	    my $tbl_clms_row_fixed = $ctbl_row_vs_clm->[$row];
	    my $tot_row_fixed = 0;
	    for (my $clm=0; $clm<$ct_clms; $clm++) {
		my $ct = $tbl_clms_row_fixed->[$clm];
		push @{$bins_clm_vs_row[$clm]}, $ct;
		$tot_row_fixed += $ct;
	    }
	    push @{$bins_clm_vs_row[$ct_clms]}, $tot_row_fixed;
	}

	return \@bins_clm_vs_row;

}		# End of the subroutine 'convert_ctbl_to_2d_distri' #


# categorize_quantity:
#	This subroutine categorizes a quantity into one of the classes
#	defined by a set of boundaries.
#
# INPUT: ($quantity, \@bounds), where
#	@bounds is a set of boundaries, stored in numerically ascending order.
#
# OUTPUT: $class, where
#	$class  = 1 if ($quantity < $bounds[0]),
#		= $k+1 if ($bounds[$k-1] <= $quantity < $bounds[$k]) for $k<@bounds,
#		= scalar (@bounds) + 1 if ($bounds[$#bounds] <= $quantity).
#
sub categorize_quantity ($\@) {

	my ($quantity, $bounds) = @_;

	my $ct_bd = @{$bounds};
	unless ($ct_bd>0) {
	    print STDERR "[ERROR 'categorize_quantity'] The boundary set is EMPTY: $! !\n";
	    return undef;
	}

	for (my $k=0; $k<$ct_bd; $k++) { ($quantity < $bounds->[$k]) and (return $k+1); }

	return $ct_bd + 1;

}	# End of the subroutine 'categorize_quantity' #


=the following two subroutines are commented out

# batch_pr_file_distri (\@\@$\*$\@) : a module which prints out MULTIPLE  distributions (of cluster sizes) 
#				to a filehandle specified by the 3rd argument.
# a minor revision: The subroutine now receives a 4th argument 'format',
#			which specifies the format of boundary parameters
# a second minor revision: The subroutine now receives a 5th argument 'tags',
#			   which specifies the labels of columns in the table.
#
sub batch_pr_file_distri (\@\@$\*$\@) {

	my ($rf_array_bins, $rf_boundaries, $measured_entity, $fh, $format, $tags) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $mytab = ' ' x 2;
	my $fwidth  =  7;
	my ($format_len) = ($format =~ m/^\%-?(\d+)/ );
	(defined $format_len) or ($format_len = 4);
	my $len_title = 2*$format_len + 10 + length($measured_entity);
	my $border = '-' x ($len_title + length($mytab) + 9 * scalar(@{$rf_array_bins}) ) . "\n";
	my @total = ();
	my @outputs = ();
	for (my $i = 0; $i < @{$rf_array_bins}; $i++) { push @total, 0; }
	for (my $i = 0; $i < @{$rf_array_bins->[0]}; $i++) { push @outputs, ''; }

	my $label = '';

	if (@{$tags}) {

		$label = $tags->[0]; 
		$label .= ' ' x ($len_title -1 - length($label) ) . ":$mytab";
		my $mfwidth = $fwidth + length($mytab);
		for (my $i=1; $i<@{$tags}; $i++) {

			if (length($tags->[$i]) < $mfwidth) {

				$label .= ' ' x ($mfwidth -length($tags->[$i]) ) . $tags->[$i];
			} else {
				$label .= ' ' . $tags->[$i];
			}
		}

		$label .= "\n";
	}

	(defined $measured_entity) or ( $measured_entity = "cluster size");

	for (my $i = 0; $i < @{$rf_array_bins}; $i++) {
		for (my $j = 0; $j < @{$rf_array_bins->[$i]}; $j++) {
			$total[$i] += $rf_array_bins->[$i][$j];
			#$outputs[$j] .= "\t" . $rf_array_bins->[$i][$j];
			$outputs[$j] .= sprintf "$mytab%${fwidth}d" , $rf_array_bins->[$i][$j];
		}
	}

	print $fh $label;
	print $fh $border; 
	#print $fh "\($measured_entity\) < $rf_boundaries->[0]:\t\t$outputs[0]\n";
	print $fh ' ' x ($format_len + 4);
	printf $fh "\($measured_entity\) < $format:$mytab$outputs[0]\n", $rf_boundaries->[0];
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

		#print $fh "$rf_boundaries->[$i -1] =< \($measured_entity\) < $rf_boundaries->[$i]:\t$outputs[$i]\n";
		printf $fh "$format =< \($measured_entity\) < $format:$mytab$outputs[$i]\n", $rf_boundaries->[$i-1], $rf_boundaries->[$i];
	}

	#print $fh "$rf_boundaries->[$num_boundaries -1] =< \($measured_entity\):\t$outputs[$num_boundaries]\n";
	printf $fh "$format =< \($measured_entity\)", $rf_boundaries->[$num_boundaries -1];
	print $fh ' ' x ($format_len + 3) . ":$mytab$outputs[$num_boundaries]\n";

	#my $total = join "\t", @total;
	my $total = '';
	foreach my $tmp (@total) { $total .= sprintf "$mytab%${fwidth}d" , $tmp; }
	print $fh $border;
	print $fh "Total" . ' ' x ($len_title - 1 - length('Total') );
	print $fh ":$mytab$total\n";

}						# end of the module batch_pr_file_distri #


# window_distributor: This subroutine construct a set of histogram-like distributions
#			out of a set of (probably potitional) distributions obtained
#			e.g. by a sliding-window analysis.
#
# Expected input format: $_[0] = [ $window_distribution_1, $window_distribution_2, ... ],
#			where $window_distribution_i = [quantity_wd1, quantity_wd2, ... ]_i-th_set,
#			their sizes do not have to be identical.
#
sub window_distributor (\@\@) {

	my ($rf_window_distri, $rf_bounds) = @_;

	my $ct_wd_sets = @{$rf_window_distri};
	$ct_wd_sets or (return []);

	my @batch_windowsets = ();	# statistics collected for each of the windowsets #
	my @bin_tot_windows = ();	# total statistics #

	foreach (1 .. $ct_wd_sets) { push @batch_windowsets, []; }
	push @batch_windowsets, \@bin_tot_windows;
	batch_initializer (@batch_windowsets, @{$rf_bounds});

	for (my $i=0; $i<$ct_wd_sets; $i++) {

		my $wd_ct = @{$rf_window_distri->[$i]};
		for (my $l=0; $l<$wd_ct; $l++) {

			my $wd_quantity = $rf_window_distri->[$i][$l];

			cluster_distributor ($wd_quantity, @{$batch_windowsets[$i]}, @{$rf_bounds});
			cluster_distributor ($wd_quantity, @bin_tot_windows, @{$rf_bounds});

		}		# end of the inner for #
	}			# end of the outer for #

	return (\@batch_windowsets);

}				# end of the subroutine window_distributor #
=cut


1;			# a "spell" necessary for the package to be "use"-able or "require"-able #


