package MyInverseFunc;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyInverseFunc.pm:
#	This package calculates inverse functions of the functions
#	that are defined in "MySpecFunc.pm" or in "MyProbDistri",
#	mainly via the Newton-Raphson algorithm of solution search.
#

use strict;
use warnings;
use Exporter;

use MySpecFunc qw( gammln gammp gammq betai );

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
	inv_gammp_nwt inv_gammq_nwt
	inv_erff_nwt inv_erffc_nwt nwt_crit_norm_2tail nwt_crit_norm_1tail
	inv_chi2_upper_nwt inv_chi2_lower_nwt nwt_crit_chi2
	inv_betai_nwt inv_betai_lvr inv_betai_lvr2
	);

our $ITMAX = $MySpecFunc::ITMAX; # Maximum allowed number of iterations #
our $EPS = $MySpecFunc::EPS; # Relative accuracy #
##our $FPMIN = $MySpecFunc::FPMIN; # Number near the smallest representable floating-point number
our $INFTY = 2.5e+38; # A large number to be practically regarded as infinity


our $EPS2 = 1e-07; # A "small" value to stabilize the behavior of "inv_gammp_nwt'
our $EPS3 = 1e-04; # A "small" value to stabilize the behavior of "inv_gammp_nwt'
our $EPS4 = 1e-03; # A "small" value to stabilize the behavior of "inv_gammp_nwt'

# inv_gammp_nwt:
#	This subroutine searches for $x such that
#	P ($a, $x) = $y
#	for a given $y and $a, using the Newton-Raphson algorithm.
#	Here P ($a, $x) = 1/[Gamma ($a)] int_0^$x d$t exp (-$t) *$t ** ($a - 1) 
#	is the lower-bounded incomplete gamma function of degree $a.
#
#	An AD-HOC MEASURE (introduced on April 30 (Thu), 2009):
#	Because we are interested in $x >= 0, if $x becomes NEGATIVE OR ZERO
#	after an iteration step, the subroutine will FORCE
#	$x = $EPS and ADVANCE to the next step.
#	[FURTHER REVISION  on April 30 (Thu), 2009]
#	 $EPS (= 1e-15) was replaced to a larger parameter, $EPS2 = 1e-7,
#	in order to stabilize the behavior at the small values of $y,
#	especially with $a = 3/2.
#	[YET FURTHER REVISION on April 30 (Thu), 2009]
#	$EPS2 (= 1e-7) was replaced with $EPS3 = 1e-4.
#	[YET YET FURTHER REVISION on April 30 (Thu), 2009]
#	$EPS3 (= 1e-4) was replaced with $EPS4 = 1e-3.
#
# INPUT: ($a, $y, $max_ite, $tol_x, $tol_fnc), where
#	$a is the degree of the incomplete gamma function;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_drv is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#
sub inv_gammp_nwt ($$$$$) {

	my ($a, $y, $max_ite, $tol_x, $tol_fnc) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_gammp_nwt' >\n\n"; 
	print STDERR "\$a = $a,  \$y = $y .\n\n";
=cut

	unless ($a>0) {
	    print STDERR "[ERROR 'inv_gammp_nwt'] ILLEGAL INPUT: NON-POSITIVE \$a (= $a): $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_gammp_nwt'] ILLEGAL INPUT: The TARGET VALUE of Gammap is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($max_ite > $ITMAX) and ($max_ite = $ITMAX);
	($tol_x < $EPS) and ($tol_x = $EPS);
	($tol_fnc < $EPS) and ($tol_fnc = $EPS); 

	($y == 0) and (return (0, 0, 0, 0));
	($y == 1) and (return ($INFTY, 0, undef, 0));

	my $x = $a; # Initial value
	my $itect = 0;

=for DEBUGGING the subroutine
	print STDERR "\n\$x_init = $x .\n\n";
=cut

	   ## REVISED in debugging the subroutine (on April 30 (Thu), 2009) ##
	my $gamma = exp ( gammln ($a) );
	my $rec_gamma = exp ( - gammln ($a) );
	my $abs_del_x;

	while ($itect < $max_ite) { # The main while-loop # 

	    my $fnc = gammp ($a, $x);
	    unless (defined $fnc) {
		print STDERR "[ERROR 'inv_gammp_nwt'] 'gammp' FAILED on the input \(\$a, \$x) = ($a, $x) at the $itect th iteration: $! !\n";
		return (undef, undef, undef, undef);
	    }

		# Check convergence in terms of functional value #

	    my $del_fnc = $y - $fnc;
	    my $abs_del_fnc = ($del_fnc>0) ? $del_fnc : - $del_fnc ;

=for DEBUGGING the subroutine
	    print STDERR "delta{func} = $del_fnc .\n\n";
=cut


	    if ($abs_del_fnc <= $tol_fnc) {

=for DEBUGGING the subroutine
		print STDERR "\$x_final = $x .\n";
		print STDERR "\n --- The subroutine 'inv_gammp_nwt' ENDED ---\n\n";
=cut
		return ($x, $itect, $abs_del_x, $abs_del_fnc);
	    }

	    $itect++;

		## Update $x ##

	    my $drv = exp ( - $x + ($a - 1) * log ($x) ) * $rec_gamma;

		## REVISED in debugging the subroutine (on April 30 (Thu), 2009) ##
	    my $rec_drv = exp ( $x - ($a - 1) * log ($x) ) * $gamma;
	    my $del_x = $del_fnc * $rec_drv;
	    ##my $del_x = $del_fnc /$drv ;
	    $x += $del_x ;

		## An AD-HOC MEASURE (introduced on April 30 (Thu), 2009) ##
	    if ($x <= 0) {
###=for MONITORING the subroutine
		print STDERR "[CAUTION 'inv_gammp_nwt'] Because the variate \$x = $x is NON-POSITIVE at the $itect th step, I will FORCEFULLY set it to $EPS3 ...\n\n"; 
###=cut
		##$x = $EPS4; # Yet Yet Further revision on April 30 (Thu), 2009
		$x = $EPS3; # Yet Further revision on April 30 (Thu), 2009
		##$x = $EPS2; # Further revision on April 30 (Thu), 2009
		##$x = $EPS;
		next;
	    }

		## Check convergence in terms of the solution ($x) ##
	    $abs_del_x = ($del_x > 0) ? $del_x : - $del_x ;

=for DEBUGGING the subroutine 
	    print STDERR "{ Result of the $itect th iteration step }\n\n";
	    print STDERR "\$x = $x ,\n";
	    print STDERR "derivative = $drv ,\n";
	    print STDERR "delta{\$x} = $del_x .\n\n";
=cut
	    ($abs_del_x <= $tol_x) and last;

	}		# End of the main while-loop #

	my $fnc = gammp ($a, $x);
	my $del_fnc = $y - $fnc;
	my $abs_del_fnc = ($del_fnc>0) ? $del_fnc : - $del_fnc ;

=for DEBUGGING the subroutine
	print STDERR "delta{func} = $del_fnc .\n\n";
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_gammp_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc);

}		# End of the subroutine 'inv_gammp_nwt' #


# inv_gammq_nwt:
#	This subroutine searches for $x such that
#	Q ($a, $x) = $y
#	for a given $y and $a, using the Newton-Raphson algorithm.
#	Here Q ($a, $x) = 1/[Gamma ($a)] int_$x^infty d$t exp (-$t) *$t ** ($a - 1) 
#	is the lower-bounded incomplete gamma function of degree $a.
#	This subroutine makes full use of the relationship
#	Q ($a, $x) = 1 - P ($a, $x).
#
# INPUT: ($a, $y, $max_ite, $tol_x, $tol_fnc), where
#	$a is the degree of the incomplete gamma function;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_drv is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#
sub inv_gammq_nwt ($$$$$) {

	my ($a, $y, $max_ite, $tol_x, $tol_fnc) = @_;

=for DEBUGGING the subroutine
	print STDERR "< The subroutine 'inv_gammq_nwt' >\n\n";
	print STDERR "\$a = $a, \$y = $y .\n\n";
=cut

	my ($x, $itect, $abs_del_x, $abs_del_fnc)
		= inv_gammp_nwt ($a, 1 - $y, $max_ite, $tol_x, $tol_fnc);

	unless (defined $x) {
	    print STDERR "[ERROR 'inv_gammq_nwt'] The main inner subroutine 'inv_gammp_nwt' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_gammq_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc) ;

}		# End of the subroutine 'inv_gammq_nwt' #

# inv_erff_nwt:
#	This subroutine looks for $x (>= 0) such that
#	$y = erf ($x) = 2/sqrt(pi) * Int_0^$x d$t exp (- $t * $t ),
#	using the Newton-Raphson algorithm.
#	It makes full use of the relationship
#	erf ($x) = P (1/2, $x * $x );
#
# INPUT: ($y, $max_ite, $tol_x, $tol_fnc), where
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_drv is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#	
sub inv_erff_nwt ($$$$) {

	my ($y, $max_ite, $tol_x, $tol_fnc) = @_;

=for DEBUGGNG the subroutine
	print STDERR "< Subroutine 'inv_erff_nwt' >\n\n";
	print STDERR "\$y = $y .\n\n";
=cut

	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_erff_nwt'] ILLEGAL INPUT: The TARGET VALUE of Gammap is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($y == 0) and (return (0, 0, 0, 0));
	($y == 1) and (return ($INFTY, 0, undef, 0));

	##($max_ite > $ITMAX) and ($max_ite = $ITMAX);
	##($tol_x < $EPS) and ($tol_x = $EPS);
	##($tol_fnc < $EPS) and ($tol_fnc = $EPS); 

	   # Preliminary round #
	my ($sqx0, $itect0, $abs_del_sqx0, $abs_del_fnc0)
		= inv_gammp_nwt (1/2, $y, 5, 1e-04, 1e-06);
	unless (defined $sqx0) {
	    print STDERR "[ERROR 'inv_erff_nwt'] The PRELIMINARY ROUND of the subroutine 'inv_gammp_nwt' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $tol_sqx = ($sqx0 > 1e-06) ? $tol_x * 2 * (sqrt $sqx0) : $tol_x * 1e-03;

	   # The serious round #
	my ($sqx, $itect, $abs_del_sqx, $abs_del_fnc)
		= inv_gammp_nwt (1/2, $y, $max_ite, $tol_sqx, $tol_fnc);
	unless (defined $sqx) {
	     print STDERR "[ERROR 'inv_erff_nwt'] The FULL-SCALE ROUND of the subroutine 'inv_gammp_nwt' FAILED: $! !\n";
	     return (undef, undef, undef, undef);
	}

	my $x = sqrt $sqx;
	my $abs_del_x = ((defined $abs_del_sqx) and ($x>0)) ?
		$abs_del_sqx /(2 *$x) : $abs_del_sqx ;

=for DEBUGGNG the subroutine
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_erff_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc);

}	# End of the subroutine 'inv_erff_nwt' #


# inv_erffc_nwt:
#	This subroutine looks for $x (>= 0) such that
#	$y = erfc ($x) = 2/sqrt(pi) * Int_$x^infty d$t exp (- $t * $t ),
#	using the Newton-Raphson algorithm.
#	It makes full use of the relationship
#	erfc ($x) = 1 - erf ($x) = 1 - P (1/2, $x * $x );
#
# INPUT: ($y, $max_ite, $tol_x, $tol_fnc), where
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_drv is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#	
sub inv_erffc_nwt ($$$$) {

	my ($y, $max_ite, $tol_x, $tol_fnc) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_erffc_nwt' >\n\n";
	print STDERR "\$y = $y .\n\n";
=cut

	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_erffc_nwt'] ILLEGAL INPUT: The TARGET VALUE of Gammap is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($y == 0) and (return ($INFTY, 0, undef, 0));
	($y == 1) and (return (0, 0, 0, 0));

	##($max_ite > $ITMAX) and ($max_ite = $ITMAX);
	##($tol_x < $EPS) and ($tol_x = $EPS);
	##($tol_fnc < $EPS) and ($tol_fnc = $EPS); 

	   # Preliminary round #
	my ($sqx0, $itect0, $abs_del_sqx0, $abs_del_fnc0)
		= inv_gammp_nwt (1/2, 1 - $y, 5, 1e-04, 1e-06);
	unless (defined $sqx0) {
	    print STDERR "[ERROR 'inv_erffc_nwt'] The PRELIMINARY ROUND of the subroutine 'inv_gammp_nwt' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $tol_sqx = ($sqx0 > 1e-06) ? $tol_x * 2 * (sqrt $sqx0) : $tol_x * 1e-03;

	   # The serious round #
	my ($sqx, $itect, $abs_del_sqx, $abs_del_fnc)
		= inv_gammp_nwt (1/2, 1 - $y, $max_ite, $tol_sqx, $tol_fnc);
	unless (defined $sqx) {
	     print STDERR "[ERROR 'inv_erffc_nwt'] The FULL-SCALE ROUND of the subroutine 'inv_gammp_nwt' FAILED: $! !\n";
	     return (undef, undef, undef, undef);
	}

	my $x = sqrt $sqx;
	my $abs_del_x = ((defined $abs_del_sqx) and ($x>0)) ?
		$abs_del_sqx /(2 *$x) : $abs_del_sqx ;

=for DEBUGGING the subroutine
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_erffc_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc);

}	# End of the subroutine 'inv_erffc_nwt' #


# nwt_crit_norm_2tail:
#	This subroutine estimates, via the Newton-Raphson algorithm,
#	the critical $crit value at the significance level $sig_lvl
#	for the 2-tailed test under the null hypothesis of
#	the standard normal distribution N(0, 1):
#	$sig_lvl = Prob (X < - $crit or X > $crit | N(0, 1) )
#	      = 2/sqrt (2* pi) * Int_$crit^infty d$t exp ( - $t * $t /2 ).
#	This subroutine makes full use of the fact that the above probability
#	equals erffc ($crit/sqrt (2) ).
#
# INPUT: ($sig_lvl, $max_ite, $tol_crt, $tol_prb), where
#	$sig_lvl is the specified significance level;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_crt is the tolerance level of the differences in the approximate critical values;
#	$tol_prb is the tolerance level of the differences in the probability;
#
# OUTPUT: ($crit, $itect, $del_crt, $del_prb), where
#	$crit is the final approximation of the critical value;
#	$itect is the number of iterative steps;
#	$del_crt is the absolute value of the difference between
#	the final two approximations of the critical value;
#	$del_prb is the difference between $sig_lvl and the final 
#	approximation of the probability.
#
sub nwt_crit_norm_2tail ($$$$) {

	my ($sig_lvl, $max_ite, $tol_crt, $tol_prb) = @_;

	unless (($sig_lvl >= 0) and ($sig_lvl <= 1)) {
	    print STDERR "[ERROR 'nwt_crit_norm_2tail'] ILLEGAL INPUT: The specified significance level ($sig_lvl) is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($sig_lvl == 0) and (return ($INFTY, 0, undef, 0));
	($sig_lvl == 1) and (return (0, 0, 0, 0));

		# Rescale the tolerance #
	$tol_crt /= sqrt 2;

	my ($crit, $itect, $del_crt, $del_prb)
		= inv_erffc_nwt ($sig_lvl, $max_ite, $tol_crt, $tol_prb);

	unless (defined $crit) {
	    print STDERR "[ERROR 'nwt_crit_norm_2tail'] The main inernal subroutine 'inv_erffc_nwt' FAILED at the target significance level $sig_lvl: $! !\n";
	    return (undef, undef, undef, undef);
	}

		# Rescale the solution (and its error) #
	$crit *= sqrt 2;
	(defined $del_crt) and ($del_crt *= sqrt 2 );

	return ($crit, $itect, $del_crt, $del_prb);

}		# End of the subroutine 'nwt_crit_norm_2tail' #


# nwt_crit_norm_1tail:
#	This subroutine estimates, via the Newton-Raphson algorithm,
#	the critical $crit value at the significance level $sig_lvl
#	for the 1-tailed test under the null hypothesis of
#	the standard normal distribution N(0, 1):
#
#	$sig_lvl = Prob ( X > $crit | N(0, 1) )
#	      = 1/sqrt (2* pi) * Int_$crit^infty d$t exp ( - $t * $t /2 ).
#
#	This subroutine makes full use of the fact that the above probability
#	equals 1/2 * erfc ($crit/sqrt (2) ) (< 1/2) if $crit > 0,
#	     and 1/2 * (1 + erf (- $crit /sqrt (2) ) ) (> 1/2) if $crit < 0.
#
# INPUT: ($sig_lvl, $max_ite, $tol_crt, $tol_prb), where
#	$sig_lvl is the specified significance level;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_crt is the tolerance level of the differences in the approximate critical values;
#	$tol_prb is the tolerance level of the differences in the probability;
#
# OUTPUT: ($crit, $itect, $del_crt, $del_prb), where
#	$crit is the final approximation of the critical value;
#	$itect is the number of iterative steps;
#	$del_crt is the absolute value of the difference between
#	the final two approximations of the critical value;
#	$del_prb is the difference between the $sig_lvl and the final 
#	approximation of the probability.
#
sub nwt_crit_norm_1tail ($$$$) {

	my ($sig_lvl, $max_ite, $tol_crt, $tol_prb) = @_;

	unless (($sig_lvl >= 0) and ($sig_lvl <= 1)) {
	    print STDERR "[ERROR 'nwt_crit_norm_1tail'] ILLEGAL INPUT: The specified significance level ($sig_lvl) is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($sig_lvl == 0) and (return ($INFTY, 0, undef, 0));
	($sig_lvl == 1) and (return (- $INFTY, 0, undef, 0));
	($sig_lvl == 1/2) and (return (0, 0, 0, 0));

		# Rescale the tolerance #
	$tol_crt /= sqrt 2;

	my ($crit, $itect, $del_crt, $del_prb)
		= ($sig_lvl < 1/2) ? 
		inv_erffc_nwt (2 * $sig_lvl, $max_ite, $tol_crt, $tol_prb) :
		inv_erff_nwt (2 * $sig_lvl - 1, $max_ite, $tol_crt, $tol_prb) ;

	unless (defined $crit) {
	    if ($sig_lvl < 1/2) {
	        print STDERR "[ERROR 'nwt_crit_norm_1tail'] The main inernal subroutine 'inv_erffc_nwt' FAILED at the target significance level $sig_lvl: $! !\n";
	    } else {
	        print STDERR "[ERROR 'nwt_crit_norm_1tail'] The main inernal subroutine 'inv_erff_nwt' FAILED at the target significance level $sig_lvl: $! !\n";
	    }
	    return (undef, undef, undef, undef);
	}

		# Rescale the solution (and its error) #
	$crit *= sqrt 2;
	(defined $del_crt) and ($del_crt *= sqrt 2 );

	($sig_lvl > 1/2) and ($crit = - $crit); # negative critical value #

	return ($crit, $itect, $del_crt, $del_prb);

}		# End of the subroutine 'nwt_crit_norm_1tail' #


# inv_chi2_upper_nwt:
#	This subroutine looks for $x such that
#	Prob (0 <= X < $x | $nu) = $y,
#	for a variate X conforming to the chi-square distribuiton of DOF = $nu,
#	via the Newton-Raphson algorithm.
#	The subroutine makes full use of the relationship
#	Prob (0 <= X < $x | $nu) = 1/[2^{$nu/2} * Gamma($nu/2)] *
#		Int_0^$x d$t $t**($nu/2 - 1) * exp ( - $t)
#	   = P ($nu/2, $x/2),
#	whre P($a, $z) is the lower-bound incomplete gamma function of degree $a.
#
# INPUT: ($nu, $y, $max_ite, $tol_x, $tol_prb), where
#	$nu is the DOF of the chi-square variate;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_prb is the tolerance level of the differences in the probability;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_prb), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_prb is the absolute value of the difference between the final approximate value of the probability and the target value;
#	
sub inv_chi2_upper_nwt ($$$$$) {

	my ($nu, $y, $max_ite, $tol_x, $tol_prb) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_chi2_upper_nwt' >\n\n";
	print STDERR "\$nu = $nu, \$y = $y .\n\n";
=cut

	unless ($nu > 0) {
	    print STDERR "[ERROR 'inv_chi2_upper_nwt'] ILLEGAL INPUT: NON-POSITIVE DOF ($nu) for the chi-square distribution: $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_chi2_upper_nwt'] ILLEGAL INPUT: The TARGET VALUE of Gammap is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($y == 0) and (return (0, 0, 0, 0));
	($y == 1) and (return ($INFTY, 0, undef, 0));

	my ($hfx, $itect, $abs_del_hfx, $abs_del_prb)
	    = inv_gammp_nwt ($nu/2, $y, $max_ite, $tol_x/2, $tol_prb);
	unless (defined $hfx) {
	    print STDERR "[ERROR 'inv_chi2_upper_nwt'] The internal subroutine 'inv_gammp_nwt' FAILED on the target value $y and the degree ", $nu/2, ": $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $x = 2 * $hfx;
	my $abs_del_x = (defined $abs_del_hfx) ? 2 * $abs_del_hfx : undef ;

=for DEBUGGING the subroutine
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_chi2_upper_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_prb);

}	# End of the subroutine 'inv_chi2_upper_nwt' #


# inv_chi2_lower_nwt:
#	This subroutine looks for $x such that
#	Prob ( X > $x | $nu) = $y,
#	for a variate X conforming to the chi-square distribuiton of DOF = $nu,
#	via the Newton-Raphson algorithm.
#	The subroutine makes full use of the relationship
#	Prob ( X > $x | $nu) = 1/[2^{$nu/2} * Gamma($nu/2)] *
#		Int_$x^infty d$t $t**($nu/2 - 1) * exp ( - $t)
#	    = 1 - P ($nu/2, $x/2),
#	whre P($a, $z) is the lower-bound incomplete gamma function of degree $a.
#
# INPUT: ($nu, $y, $max_ite, $tol_x, $tol_prb), where
#	$nu is the DOF of the chi-square variate;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_prb is the tolerance level of the differences in the probability;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_prb), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_prb is the absolute value of the difference between the final approximate value of the probability and the target value;
#	
sub inv_chi2_lower_nwt ($$$$$) {

	my ($nu, $y, $max_ite, $tol_x, $tol_prb) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_chi2_lower_nwt' >\n\n";
	print STDERR "\$nu = $nu,  \$y = $y .\n\n";
=cut

	unless ($nu > 0) {
	    print STDERR "[ERROR 'inv_chi2_lower_nwt'] ILLEGAL INPUT: NON-POSITIVE DOF ($nu) for the chi-square distribution: $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_chi2_lower_nwt'] ILLEGAL INPUT: The TARGET VALUE of Gammap is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($y == 0) and (return ($INFTY, 0, undef, 0));
	($y == 1) and (return (0, 0, 0, 0));

	my ($hfx, $itect, $abs_del_hfx, $abs_del_prb)
	    = inv_gammp_nwt ($nu/2, 1 - $y, $max_ite, $tol_x/2, $tol_prb);
	unless (defined $hfx) {
	    print STDERR "[ERROR 'inv_chi2_lower_nwt'] The internal subroutine 'inv_gammp_nwt' FAILED on the target value ", 1 - $y, " , which is the complement of $y, and the degree ", $nu/2, ": $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $x = 2 * $hfx;
	my $abs_del_x = (defined $abs_del_hfx) ? 2 * $abs_del_hfx : undef ;

=for DEBUGGING the subroutine
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_chi2_lower_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_prb);

}	# End of the subroutine 'inv_chi2_lower_nwt' #


# nwt_crit_chi2:
#	This subroutine estimates, via the Newton-Raphson algorithm,
#	the critical value at the significance level $sig_lvl
#	for the chi-square distribution of DOF = $nu.
#
#	The subroutine makes full use of the relationship:
#	$sig_lvl = Prob (X > $crit | $nu)
#	  = 1/[2^{$nu/2} * Gamma($nu/2)] * Int_$crit^infty d$t $t**($nu/2 - 1) * exp (- $t)
#		= 1 - P ($nu/2, $crit/2).
#
#	Actually, this subroutine is just a wrapper of the subroutine 'inv_chi2_lower_nwt'.
#
# INPUT: ($nu, $sig_lvl, $max_ite, $tol_c, $tol_prb), where
#	$nu is the DOF of the chi-square distribution;
#	$sig_lvl is the significance level specified;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_c is the tolerance level of the differences in $crit;
#	$tol_prb is the tolerance level of the differences in the probability;
#
# OUTPUT: ($crit, $itect, $abs_del_c, $abs_del_prb), where
#	$crit is the final approximate solution for the critical value;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_c is the absolute value of the difference in the approximate critical value between the final two iteration steps;
#	$abs_del_prb is the absolute value of the difference between the final approximate value of the probability and the target value;
#
sub nwt_crit_chi2 ($$$$$) {	

	my ($nu, $sig_lvl, $max_ite, $tol_c, $tol_prb) = @_;

	my ($crit, $itect, $abs_del_c, $abs_del_prb)
		= inv_chi2_lower_nwt ($nu, $sig_lvl, $max_ite, $tol_c, $tol_prb);

	unless (defined $crit) {
	    print STDERR "[ERROR 'nwt_crit_chi2'] The main internal subroutine 'inv_chi2_lower_nwt' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

	return ($crit, $itect, $abs_del_c, $abs_del_prb);

}		# End of the subroutine 'nwt_crit_chi2' #


# inv_betai_nwt:
#	This subroutine searches for $x such that
#	I_$x ($a, $b) = $y
#	for a given $y and ($a, $b), using the Newton-Raphson algorithm.
#	Here I_$x ($a, $b) = 1/[Beta ($a, $b)] int_0^$x d$t $t ** ($a - 1)  * (1 - $t)**($b-1)
#	is the incomplete beta function.
#
#	REVISED on April 15 (Thr), 2010:
#	Now the loop will end prematurely
#	ONLY IF BOTH the $x resolution and the $y resolution are
#	under the specified tolerance level.
#
#	An AD-HOC MEASURE:
#	Because we are interested in 0 <= $x <= 1, if $x becomes NEGATIVE OR ZERO
#	after an iteration step, the subroutine will FORCE
#	$x = $EPS3 (= 1E-04) and ADVANCE to the next step.
#	And if $x becomes > 1 after an iteration step, the subroutine will FORCE
#	$x = 1 - $EPS3 and ADVANCE to the next step.
#	##[YET YET FURTHER REVISION] -- may be necessary in the near future
#	##$EPS3 (= 1e-4) was replaced with $EPS4 = 1e-3.
#
# INPUT: ($a, $b, $y, $max_ite, $tol_x, $tol_fnc), where
#	$a and $b are the parameters of the incomplete gamma function;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_fnc is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#
sub inv_betai_nwt ($$$$$$) {

	my ($a, $b, $y, $max_ite, $tol_x, $tol_fnc) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_betai_nwt' >\n\n"; 
	print STDERR "\$a = $a, \$b = $b, \$y = $y .\n\n";
=cut

	unless (($a>0) and ($b>0)) {
	    print STDERR "[ERROR 'inv_betai_nwt'] ILLEGAL INPUT PARAMETERS: (\$a, \$b) = ($a, $b): $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_beta_nwt'] ILLEGAL INPUT: The TARGET VALUE of I_x ($a, $b) = $y is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($max_ite > $ITMAX) and ($max_ite = $ITMAX);
	($tol_x < $EPS) and ($tol_x = $EPS);
	($tol_fnc < $EPS) and ($tol_fnc = $EPS); 

	($y == 0) and (return (0, 0, 0, 0));
	($y == 1) and (return (1, 0, 0, 0));
###=DISABLE for testing the subroutine
	($a == 1) and (return (($b ==1) ? 1 - $y : 1 - exp (log (1 - $y) /$b), 0, 0, 0));
	($b == 1) and (return (exp (log($y) /$a), 0, 0, 0));
###=cut

		### A BUG FIXED at the Initial Stage (on April 14 (Wed), 2010) ###
	my $betaln = gammln ($a) + gammln ($b) - gammln ($a + $b) ;
	### my $betaln = gammln ($a+$b) - gammln ($a) - gammln ($b);
	my $beta = exp ( $betaln);
	my $rec_gamma = exp ( - $betaln );

		# Initialize $x #
	my $x = ($a + $b == 2) ? 1/2 : ($a - 1) /($a + $b - 2);
	(($x > 0) and ($x < 1)) or ($x = 1/2);

		# Handle Two Extreme Cases #
	my $x_01 = exp ( (log ($y *$a) + $betaln)/$a );
	my $x_02  = 1 - exp ( (log ((1 - $y)*$b) + $betaln)/$b);
	if ($x_01 < $EPS3 ) {
	    ##my $scale = exp ( ($b - 1) * log (1 - $x_01) ) ;
	    ##my $d_scale = ($scale > 1) ? $scale - 1 : 1 - $scale;
	    ##if ($d_scale < $tol_x) {
	    #my $fnc = betai ($a, $b, $x_01);
	    #my $abs_del_fnc = ($y > $fnc) ? $y - $fnc : $fnc - $y;
	    #($abs_del_fnc < $tol_fnc) and (return ($x_01, 0, undef, $abs_del_fnc));
	    ##}
	    $x = ($x_01 > $EPS2) ? $x_01 : $EPS2; # A measure to stabilize the subroutine #
=for MONITORING the subroutine
	    print STDERR "[ATTENTION 'inv_betai_nwt'] Initial value set to $x for (\$a, \$b, \$y) = ($a, $b, $y) !\n\n";
=cut
	} elsif (1 - $x_02 < $EPS3) {
	    ##my $scale = exp ( ($a - 1) * log ($x_02) );
	    ##my $d_scale = ($scale > 1) ? $scale -1 : 1 - $scale;
	    ##if ($d_scale < $tol_x) {
	    #my $fnc = betai ($a, $b, $x_02);
	    #my $abs_del_fnc = ($y > $fnc) ? $y - $fnc : $fnc - $y;
	    #($abs_del_fnc < $tol_fnc) and (return ($x_02, 0, undef, $abs_del_fnc));
	    ##}
	    $x = (1 - $x_02 > $EPS2) ? $x_02 : 1 - $EPS2; # A measure to stabilize the subroutine #
=for MONITORING the subroutine
	    print STDERR "[ATTENTION 'inv_betai_nwt'] Initial value set to $x for (\$a, \$b, \$y) = ($a, $b, $y) !\n\n";
=cut
	}


=for DEBUGGING the subroutine
	print STDERR "\n\$x_init = $x .\n\n";
=cut

	my $ct_force = 0;

	my $itect = 0;
	my $abs_del_x;

	while ($itect < $max_ite) { # The main while-loop # 

	    my $fnc = betai ($a, $b, $x);
	    unless (defined $fnc) {
		print STDERR "[ERROR 'inv_betai_nwt'] 'betai' FAILED on the input \(\$a, \$b, \$x) = ($a, $b, $x) at the $itect th iteration: $! !\n";
		return (undef, undef, undef, undef);
	    }

		# Check convergence in terms of functional value #

	    my $del_fnc = $y - $fnc;
	    my $abs_del_fnc = ($del_fnc>0) ? $del_fnc : - $del_fnc ;

=for DEBUGGING the subroutine
	    print STDERR "delta{func} = $del_fnc .\n\n";
=cut


		## REVISED on April 15 (Thr), 2010 ##
	    if (($abs_del_fnc <= $tol_fnc) and (defined $abs_del_x) and ($abs_del_x <= $tol_x)) {
	    ##if ($abs_del_fnc <= $tol_fnc) {
=for DEBUGGING the subroutine
		print STDERR "\$x_final = $x .\n";
		print STDERR "\n --- The subroutine 'inv_betai_nwt' ENDED ---\n\n";
=cut
		return ($x, $itect, $abs_del_x, $abs_del_fnc);
	    }

	    $itect++;

		## Update $x ##

	    my $drv = exp ( ($a - 1) * log ($x) + ($b - 1) * log (1 - $x) - $betaln );
	    ##my $drv = exp ( ($a - 1) * log ($x) + ($b - 1) * log (1 - $x) ) * $rec_beta;
	    ##my $del_x = $del_fnc /$drv ;

		## REVISED in debugging the subroutine (on April 30 (Thu), 2009) ##
	    my $rec_drv = exp ( - ($a - 1) * log ($x) - ($b - 1) * log (1- $x) + $betaln );
	    ##my $rec_drv = exp ( - ($a - 1) * log ($x) - ($b - 1) * log (1- $x) ) * $beta;
	    my $del_x = $del_fnc * $rec_drv;
	    $x += $del_x ;

		## An AD-HOC MEASURE ##
	    if ($x <= 0) {
=for MONITORING the subroutine
		print STDERR "[CAUTION 'inv_betai_nwt'] Because the variate \$x = $x is NON-POSITIVE at the $itect th step, I will FORCEFULLY set it to $EPS3 ...\n\n"; 
=cut
		$ct_force++;
		##$x = $EPS4; #  Further revision -- May be necesasary in the future 
		$x = $EPS3;
		if ($ct_force>2) {
		    # Tolerate up to 3 consecutive times of such forceful corrections
		    print STDERR "[CAUTION 'inv_betai_nwt'] We have already forcefully corrected the value of \$x 3 consecutive times, so I will leave the loop now !!\n\n";
		    last ;
		}
		next;
	    } elsif ($x >=1) {
=for MONITORING the subroutine
		print STDERR "[CAUTION 'inv_betai_nwt'] Because the variate \$x = $x is NOT LESS THAN 1 at the $itect th step, I will FORCEFULLY set it to 1 - $EPS3 ...\n\n";
=cut
		$ct_force++;
		$x = 1 - $EPS3;
		if ($ct_force>2) {
		    # Tolerate up to 3 consecutive times of such forceful corrections
		    print STDERR "[CAUTION 'inv_betai_nwt'] We have already forcefully corrected the value of \$x 3 consecutive times, so I will leave the loop now !!\n\n";
		    last ;
		}
		next;
	    }

	    $ct_force = 0; # Reset the counter of forceful corrections #

		## Check convergence in terms of the solution ($x) ##
	    $abs_del_x = ($del_x > 0) ? $del_x : - $del_x ;

=for DEBUGGING the subroutine 
	    print STDERR "{ Result of the $itect th iteration step }\n\n";
	    print STDERR "\$x = $x ,\n";
	    print STDERR "derivative = $drv ,\n";
	    print STDERR "delta{\$x} = $del_x .\n\n";
=cut

		# DISABLED in REVISION on April 15 (Thr), 2010 #
	    ##($abs_del_x <= $tol_x) and last;

	}		# End of the main while-loop #

	my $fnc = betai ($a, $b, $x);
	unless (defined $fnc) {
	    print STDERR "[ERROR 'inv_betai_nwt'] 'betai' FAILED after finishing the loop at the $itect th step: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $del_fnc = $y - $fnc;
	my $abs_del_fnc = ($del_fnc>0) ? $del_fnc : - $del_fnc ;

=for DEBUGGING the subroutine
	print STDERR "delta{func} = $del_fnc .\n\n";
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_betai_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc);

}		# End of the subroutine 'inv_betai_nwt' #


# inv_betai_lvr:
#	This subroutine searches for $x such that
#	I_$x ($a, $b) = $y
#	for a given $y and ($a, $b), using the Lever Principle.
#	Here I_$x ($a, $b) = 1/[Beta ($a, $b)] int_0^$x d$t $t ** ($a - 1)  * (1 - $t)**($b-1)
#	is the incomplete beta function.
#
#	THEORY:
#	The incomplete beta function I_$x ($a, $b) is monotonically increasing.
#	So, if you know the two values $y1 = I_$x1 ($a, $b) and $y2 = I_$x2 ($a, $b) sandwiches the desired value $y (i.e. $y1 < $y < $y2),
#	we know that the solution $x is between $x1 and $x2.
#	We try to approximate $x with the middle value $x_m obtained by the Lever Principle:
#	$x_m = { $x2 * ($y - $y1) + $x1 * ($y2 - $y) }/($y2 - $y1).
#
#	Like a binary search, we narrow down the region by
#	choosing ($x_m, $x2) if I_$x_m ($a, $b) < $y,
#	       and ($x1, $x_m) if I_$x_m ($a, $b) > $y.
#
#	REVISED on April 15 (Thr), 2010:
#	Now the loop will end prematurely
#	ONLY IF BOTH the $x resolution and the $y resolution are
#	under the specified tolerance level.
#
# INPUT: ($a, $b, $y, $max_ite, $tol_x, $tol_fnc), where
#	$a and $b are the parameters of the incomplete gamma function;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_fnc is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#
sub inv_betai_lvr ($$$$$$) {

	my ($a, $b, $y, $max_ite, $tol_x, $tol_fnc) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_betai_nwt' >\n\n"; 
	print STDERR "\$a = $a, \$b = $b, \$y = $y .\n\n";
=cut

	unless (($a>0) and ($b>0)) {
	    print STDERR "[ERROR 'inv_betai_lvr'] ILLEGAL INPUT PARAMETERS: (\$a, \$b) = ($a, $b): $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_betai_lvr'] ILLEGAL INPUT: The TARGET VALUE of I_x ($a, $b) = $y is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($max_ite > $ITMAX) and ($max_ite = $ITMAX);
	($tol_x < $EPS) and ($tol_x = $EPS);
	($tol_fnc < $EPS) and ($tol_fnc = $EPS); 

	($y == 0) and (return (0, 0, 0, 0));
	($y == 1) and (return (1, 0, 0, 0));
###=DISABLE for testing the subroutine
	($a == 1) and (return (($b ==1) ? 1 - $y : 1 - exp (log (1 - $y) /$b), 0, 0, 0));
	($b == 1) and (return (exp (log($y) /$a), 0, 0, 0));
###=cut

	my ($x1, $y1) = (0, 0); # Left boundary #
	my ($x2, $y2) = (1, 1); # Right boundary #

	my ($x, $abs_del_x) = (0, 1);
	my $itect = 0;
	while ($itect < $max_ite) { # Main while-loop #

	    $itect++;

		# Middle value #
	    my $x_m = ($x2 * ($y - $y1) + $x1 * ($y2 - $y)) / ($y2 - $y1);
	    my $y_m = betai ($a, $b, $x_m);
	    unless (defined $y_m) {
		print "[ERROR 'inv_betai_lvr'] 'betai' FAILED on the middle value ($x_m) at the $itect th iteration step: $! !\n\n";
		return (undef, undef, undef, undef);
	    }

	    $x = $x_m;
	    my $abs_del_fnc = ($y_m > $y) ? $y_m - $y : $y - $y_m;

		# REVISED on April 15 (Thr), 2010 #
	    $abs_del_x = ($y < $y_m) ? $x_m - $x1 : $x2 - $x_m;
	    ##$abs_del_x = $x2 - $x1;

		# REVISED on April 15 (Thr), 2010 #
	    if (($abs_del_fnc < $tol_fnc) and ($abs_del_x < $tol_x)) {
	    ##if (($abs_del_fnc < $tol_fnc) or ($abs_del_x < $tol_x)) {
		return ($x, $itect, $abs_del_x, $abs_del_fnc);
	    }

	    if ($y < $y_m) {
		# Narrow the right bound #
		($x2, $y2) = ($x_m, $y_m);
	    } else {
		# Narrow the left bound #
		($x1, $y1) = ($x_m, $y_m);
	    }

	}		# End of the main while-loop #

	my $fnc = betai ($a, $b, $x);
	my $del_fnc = $y - $fnc;
	my $abs_del_fnc = ($del_fnc>0) ? $del_fnc : - $del_fnc ;

=for DEBUGGING the subroutine
	print STDERR "delta{func} = $del_fnc .\n\n";
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_betai_nwt' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc);

}		# End of the subroutine 'inv_betai_lvr' #


# inv_betai_lvr2:
#	This subroutine searches for $x such that
#	I_$x ($a, $b) = $y
#	for a given $y and ($a, $b), using the Lever Principle.
#	Here I_$x ($a, $b) = 1/[Beta ($a, $b)] int_0^$x d$t $t ** ($a - 1)  * (1 - $t)**($b-1)
#	is the incomplete beta function.
#	This subroutine conducts simple binary search in the first $ct_preite steps,
#	then moves on to the iteration by lever principle.
#
#	THEORY: (For the description on lever principle, see 'inv_betai_lvr')
#
#	Like a binary search, we narrow down the region by
#	choosing ($x_m, $x2) if I_$x_m ($a, $b) < $y,
#	       and ($x1, $x_m) if I_$x_m ($a, $b) > $y.
#
#	REVISED on April 15 (Thr), 2010:
#	Now the loop will end prematurely
#	ONLY IF BOTH the $x resolution and the $y resolution are
#	under the specified tolerance level.
#
# INPUT: ($a, $b, $y, $max_ite, $ct_preite, $tol_x, $tol_fnc), where
#	$a and $b are the parameters of the incomplete gamma function;
#	$y is the target value;
#	$max_ite gives the upper bound of the number of iteration steps INCLUDING THE PRELIMINARY STEPS;
#	$ct_pre_ite is the number of preliminary iterations by binary searches;
#	$tol_x is the tolerance level of the differences in $x;
#	$tol_fnc is the tolerance level of the differences in the function;
#
# OUTPUT: ($x, $itect, $abs_del_x, $abs_del_fnc), where
#	$x is the final approximate solution;
#	$itect is the count of iterative steps actually conducted;
#	$abs_del_x is the absolute value of the difference in $x between the final two iteration steps;
#	$abs_del_fnc is the absolute value of the difference between the final approximate value of the function and the target value;
#
sub inv_betai_lvr2 ($$$$$$$) {

	my ($a, $b, $y, $max_ite, $ct_preite, $tol_x, $tol_fnc) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'inv_betai_nwt' >\n\n"; 
	print STDERR "\$a = $a, \$b = $b, \$y = $y .\n\n";
=cut

	unless (($a>0) and ($b>0)) {
	    print STDERR "[ERROR 'inv_betai_lvr2'] ILLEGAL INPUT PARAMETERS: (\$a, \$b) = ($a, $b): $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (($y >= 0) and ($y <= 1)) {
	    print STDERR "[ERROR 'inv_betai_lvr2'] ILLEGAL INPUT: The TARGET VALUE of I_x ($a, $b) = $y is NOT within the interval \[0, 1\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	($ct_preite > $ITMAX) and ($ct_preite = $ITMAX);
	($max_ite > $ITMAX) and ($max_ite = $ITMAX);
	($tol_x < $EPS) and ($tol_x = $EPS);
	($tol_fnc < $EPS) and ($tol_fnc = $EPS); 

	($y == 0) and (return (0, 0, 0, 0));
	($y == 1) and (return (1, 0, 0, 0));
###=DISABLE for testing the subroutine
	($a == 1) and (return (($b ==1) ? 1 - $y : 1 - exp (log (1 - $y) /$b), 0, 0, 0));
	($b == 1) and (return (exp (log($y) /$a), 0, 0, 0));
###=cut

	my ($x1, $y1) = (0, 0); # Left boundary #
	my ($x2, $y2) = (1, 1); # Right boundary #

	my ($x, $abs_del_x) = (0, 1);

		# Preliminary steps #
	my $itect = 0;

	while ($itect < $ct_preite) { # Preliminary while-loop #

	    $itect++;

		# Just take the center of the boundaries for $x #
	    my $x_m = ($x1+$x2)/2;
	    my $y_m = betai ($a, $b, $x_m);
	    unless (defined $y_m) {
		print "[ERROR 'inv_betai_lvr2'] 'betai' FAILED on the middle value ($x_m) at the $itect th iteration step (preliminary): $! !\n\n";
		return (undef, undef, undef, undef);
	    }

	    $x = $x_m;
	    my $abs_del_fnc = ($y_m > $y) ? $y_m - $y : $y - $y_m;
	    $abs_del_x /= 2;

		# REVISED on April 15 (Thr), 2010 #
	    if (($abs_del_fnc < $tol_fnc) and ($abs_del_x < $tol_x)) {
	    ##if (($abs_del_fnc < $tol_fnc) or ($abs_del_x < $tol_x)) {
		return ($x, $itect, $abs_del_x, $abs_del_fnc);
	    }

	    if ($y < $y_m) {
		# Narrow the right bound #
		($x2, $y2) = ($x_m, $y_m);
	    } else {
		# Narrow the left bound #
		($x1, $y1) = ($x_m, $y_m);
	    }
	}			# End of the preliminary while-loop #


		# Main Iteration Steps #

	while ($itect < $max_ite) { # Main while-loop #

	    $itect++;

		# Middle value by the Lever Principle #
	    my $x_m = ($x2 * ($y - $y1) + $x1 * ($y2 - $y)) / ($y2 - $y1);
	    my $y_m = betai ($a, $b, $x_m);
	    unless (defined $y_m) {
		print "[ERROR 'inv_betai_lvr2'] 'betai' FAILED on the middle value ($x_m) at the $itect th iteration step (main): $! !\n\n";
		return (undef, undef, undef, undef);
	    }

	    $x = $x_m;
	    my $abs_del_fnc = ($y_m > $y) ? $y_m - $y : $y - $y_m;

		# REVISED on April 15 (Thr), 2010 #
	    $abs_del_x = ($y < $y_m) ? $x_m - $x1 : $x2 - $x_m;
	    ##$abs_del_x = $x2 - $x1;

		# REVISED on April 15 (Thr), 2010 #
	    if (($abs_del_fnc < $tol_fnc) and ($abs_del_x < $tol_x)) {
	    ##if (($abs_del_fnc < $tol_fnc) or ($abs_del_x < $tol_x)) {
		return ($x, $itect, $abs_del_x, $abs_del_fnc);
	    }


	    if ($y < $y_m) {
		# Narrow the right bound #
		($x2, $y2) = ($x_m, $y_m);
	    } else {
		# Narrow the left bound #
		($x1, $y1) = ($x_m, $y_m);
	    }

	}		# End of the main while-loop #

	my $fnc = betai ($a, $b, $x);
	my $del_fnc = $y - $fnc;
	my $abs_del_fnc = ($del_fnc>0) ? $del_fnc : - $del_fnc ;

=for DEBUGGING the subroutine
	print STDERR "delta{func} = $del_fnc .\n\n";
	print STDERR "\$x_final = $x .\n";
	print STDERR "\n --- The subroutine 'inv_betai_lvr2' ENDED ---\n\n";
=cut

	return ($x, $itect, $abs_del_x, $abs_del_fnc);

}		# End of the subroutine 'inv_betai_lvr2' #


1; 	# A 'spell' that makes the molue to be "use"-able and "require"-able #

