package MyPertProb_locPWA_itr_dawg;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPertProb_locPWA_itr_dawg.pm:
#	This Perl module is a collection of subroutines,
#	which are components of the algorithm to perform an iterative calculation
#	of the probability of a given local pairwise sequence alignment (PWA),
#	under Dawg's indel model (Cartwright 2005) with the space- and time-homogeneous indel rates.
#
#	NOTE: In this module,
#	     the "probability" calculated ignores an overall exponential factor, exp (-(Delta^Dawg + 2)*$lambda* t),
#	      because it will be cancelled whenever the multiplication factors are calculated.
#
#	[Creation STARTED on April 28 (Tue), 2015, ENDED on May 8 (Fri), 2015]
#

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray log_sum_exp );
	## union_nmarray

use MyPertProb_locPWA_dawg qw(
	mfac_null_2events
	mfac_asites_vs_dgaps_1event
	mfac_asites_vs_dgaps_2events
	mfac_agaps_vs_dsites_1event
	mfac_agaps_vs_dsites_2events
	);

=set aside for later use
use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	nearly_Simpson

	initialize_probs_case_ii
	composite_probs_since_1st_mut_case_ii
	finish_iter_calc_probs_case_ii

	initialize_probs_case_iii
	composite_probs_until_last_mut_case_iii
	finish_iter_calc_probs_case_iii
	);


	# This specifies the level of tests on the main subroutines,
	# 'xxx', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
our $TEST_LEVEL = 5;

	# This specifies the level of tests on the primary satellite subroutines,
	# 'cal_lnlk1_indel_scn_cls', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 3;
our $TEST_LEVEL1 = 5;


	### DEFINITIONS of the SUBROUTINES ###

	## Basic subroutines. ##

# nearly_Simpson:
#	This satellite subroutine of 'finish_iter_calc_probs_case_ii/iii'
#	calculates an approximation of a 1D integral (divided by the minute time-interval),
#	by applying Simpson's rule to the bulk,
#	and by approximating a minor margin, if at all, via a trapezoid.
#
#	NOTE: It assumes that the integrands were evaluated at EQUI-SPACED time-points.
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: \@integrands,
#	which stores integrands evaluated at equi-spaced time-points.
#
# OUTPUT: $near_simpson,
#	which is the weighted summation of the integrands
#	that approximates the integral divided by the minute time-interval.
#
sub nearly_Simpson (\@) {

	my ($integrands) = @_;

	my $count = @{$integrands};

	if ($count < 2) {
	    return 0; # In this case, the integration should be zero.
	} elsif ($count == 2) {
			# The simplest trapezoid. #
	    return ($integrands->[0] + $integrands->[1])/2 ;
	}

	my $near_simpson = 0;
	my $init = 0;
	my $fin = $count - 1;

	unless ($count % 2 ) { # If there are an even number of time-points.

	    my $left_margin  = $integrands->[0] + $integrands->[1];
	    my $right_margin = $integrands->[$count-2] + $integrands->[$count-1];

	    my $abs_left_margin = ($left_margin < 0) ? - $left_margin : $left_margin ;
	    my $abs_right_margin = ($right_margin < 0) ? - $right_margin : $right_margin ;

	    if ($abs_left_margin < $abs_right_margin) {
		$near_simpson = $left_margin /2;
		$init++;
	    } else {
		$near_simpson = $right_margin /2;
		$fin--;
	    }
	    $count--;
	}

	my $hf_count = int ($count/2);

	for (my $i=0; $i < $hf_count; $i++) { # Main for-loop (over pairs of integrands). #

	   my $even = $init + 2*$i;
	   my $odd  = $even + 1;
	   $near_simpson += (2* $integrands->[$even] + 4 * $integrands->[$odd])/3 ;

	}			# End of the main for-loop (over pairs of integrands). #

	$near_simpson += ($integrands->[$fin] - $integrands->[$init])/3 ;


	return $near_simpson ;

}	# End of the subroutine 'nearly_Simpson'. #


	## Practical subroutines. ##

# initialize_probs_case_ii:
#	This subroutine initializes the probabilities of local PWAs of cases (i) and (ii).
#	In case (i), there are no ancestral or descendnat sites in between two neighboring PASs (preserved ancestral sites).
#	In case (ii), there are some ancestral sites but no descendant sites in between two neighboring PASs.
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: (\@probs, \@time_points, $ub_ct_ancsites, $init_ct_events, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#
#	@probs will store the probabilities initialized here. More precisely,
#	$probs[$i_tp]->[$n_A] ($i_tp = 0, ..., $#time_points; $n_A = 0, ..., $ub_ct_ancsites) is the probability
#	that the sequence had $n_A sites in between the PASs at time $time_points[$i_tp] and that it has no sites in between the PASs at time $time_points[$#time_points];
#
#	@time_points stores the time-points that give the initial point of the time interval during which the probabilities will be calculated.
#	($time_points[$#time_points] gives the final point of the time interval.);
#
#	$ub_ct_ancsites (>= 0) is the upper-bound of #{ancestral sites between the PASs};
#	$init_ct_events (=0, 1, or 2) is the number of indels up to which each initial probability allows for;
#	
#	$lambda_I is the total rate (per unit time per site) of insertions;
#	$lambda_D is the total rate (per unit time per site) of deletions;
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: 1 if the subroutine successfully ended,
#	  0 if it failed.
#
sub initialize_probs_case_ii (\@\@$$$$\@\@$$) {

	my ($probs, $time_points, $ub_ct_ancsites, $init_ct_events, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless ($ub_ct_ancsites >= 0) {
	    print STDOUT "[ERROR 'initialize_probs_case_ii'] ILLEGAL INPUT: ", '$ub_ct_ancsites = ', $ub_ct_ancsites, " : $! !\n\n";
	    return 0;
	}
	unless (($init_ct_events == 0) or ($init_ct_events == 1) or ($init_ct_events == 2)) {
	    print STDOUT "[ERROR 'initialize_probs_case_ii'] ILLEGAL INPUT: ", '$init_ct_events = ', $init_ct_events, " : $! !\n\n";
	    return 0;
	}

	my $ct_tpts = @{$time_points};
	unless ($ct_tpts>1) {
	    print STDOUT "[ERROR 'initialize_probs_case_ii'] ILLEGAL INPUT: There are only $ct_tpts time-points: $! !\n\n";
	    return 0;
	}

    (print STDOUT "<< Subroutine 'initialize_probs_case_ii' >>\n\n") if ($TEST_LEVEL>0);

	my $time_init = $time_points->[0];
	my $time_fin = $time_points->[$ct_tpts-1];

   if ($TEST_LEVEL1>1) {
	print STDOUT "{Upper-bound of #{ancsites}} = $ub_ct_ancsites ,\n";
	print STDOUT "Initial#{events per history} = $init_ct_events ,\n";
	print STDOUT "#{time-points} = $ct_tpts ,\n";
	print STDOUT "{Final time} = $time_fin .\n\n";
   }

	my $lambda = $lambda_I + $lambda_D ;

	for (my $i_tp=0; $i_tp<$ct_tpts; $i_tp++) { # Main outer for-loop (over the time-points). #

	    my $time_init = $time_points->[$i_tp];
	    my $tlapse = $time_fin - $time_init ;

    (print STDOUT "   < The $i_tp th time-point (initial time = $time_init; time-lapse = $tlapse >\n\n") if ($TEST_LEVEL1 > 1);

	    my @array_probs = ();
	    $probs->[$i_tp] = \@array_probs;

		# First, initialize the probability with $n_A = 0. #

	    my $prob0 = 1;

	    if (($init_ct_events == 2) and ($tlapse > 0)) {
		my $mfac0_2ev = mfac_null_2events ($tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D);
		unless (defined $mfac0_2ev) {
		    print STDOUT "[ERROR 'initialize_probs_case_ii'] 'mfac_null_2events' FAILED: $! !\n\n";
		    return 0;
		}

		$prob0 += $mfac0_2ev;
	    }

	    $array_probs[0] = $prob0;

		# Then, initialize the probabilities with $n_A > 0. #

	    for (my $n_A = 1; $n_A <= $ub_ct_ancsites; $n_A++) { # Main inner for-loop (over #{ancestral sites}). #

   (print STDOUT '      ($n_A = ', "$n_A)\n\n") if ($TEST_LEVEL1 > 2);

		if (($init_ct_events == 0) or ($tlapse == 0)) {
		    $array_probs[$n_A] = 0;
		    next;
		}

		my $mfac =  mfac_asites_vs_dgaps_1event ($n_A, $tlapse, $lambda_I, $lambda_D, @{$freqs_D}, $cutoff_D);

		unless (defined $mfac) {
		    print STDOUT "[ERROR 'initialize_probs_case_ii'] 'mfac_asites_vs_dgaps_1event' FAILED for ", '($n_A, $cutoff_D) = ', "($n_A, $cutoff_D) : $! !\n\n";
		    return 0;
		}

    (print STDOUT '         $mfac = ', "$mfac .\n\n") if ($TEST_LEVEL1 > 3);

		if ($init_ct_events == 2) {

		    my ($mfac_2ev, $components) = mfac_asites_vs_dgaps_2events ($n_A, $tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D);
		    ##my $mfac_2ev = mfac_asites_vs_dgaps_2events ($n_A, $tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D); # A BUG fixed on May 7 (Thu), 2015. #

		    unless (defined $mfac_2ev) {
			print STDOUT "[ERROR 'initialize_probs_case_ii'] 'mfac_asites_vs_dgaps_2events' FAILED for ", '($n_A, $cutoff_I, $cutoff_D) = ', "($n_A, $cutoff_I, $cutoff_D) : $! !\n\n";
			return 0;
		    }

		    $mfac += $mfac_2ev;

    (print STDOUT '         $mfac = ', "$mfac .\n\n") if ($TEST_LEVEL1 > 3);
		}


		$mfac *= exp (-$lambda * $n_A * $tlapse) ; 

    (print STDOUT '         $mfac = ', "$mfac .\n\n") if ($TEST_LEVEL1 > 3);

		$array_probs[$n_A] = $mfac ;
		##$array_probs[$n_A] = $mfac * exp (-$lambda * $n_A * $tlapse) ; 

	    }				# End of the main inner for-loop (over #{ancestral sites}). #

	}			# End of the main outer for-loop (over the time-points). #

    (print STDOUT "\n<< Subroutine 'initialize_probs_case_ii' Ended. >>\n\n") if ($TEST_LEVEL1>0);

	return 1;

}	# End of the subroutine 'initialize_probs_case_ii'. #


# composite_probs_since_1st_mut_case_ii:
#	This subroutine calculates the "composite probability densities" that
#	the sequence region (with each specified #{sites in between PASs})
#	experienced an indel event at each specified time-point and
#	that the sequence has no sites between the PASs at the final time-point
#	(i.e., for the histories in cases (i) and (ii)).
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: (\@probs, $last_ub_ct_ancsites, $ub_ct_ancsites, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#
#	@probs stores the probabilities initialized here. More precisely,
#	$probs[$i_tp]->[$n_A] ($i_tp = 0, ..., $#time_points; $n_A = 0, ..., $last_ub_ct_ancsites) is the probability
#	that the sequence had $n_A sites in between the PASs at the $i_tp th time-point
#	and that it has no sites in between the PASs at the final time-point;
#
#	$last_ub_ct_ancsites (>= 0) is the last (i.e., previous) upper-bound of #{ancestral sites between the PASs}.
#	(It must equal the upper-bound of @{$probs[$i_tp]} for all $i_tp.);
#	$ub_ct_ancsites (>= 0) is the present upper-bound of #{ancestral sites between the PASs}.
#	(It will become the upper-bound of @{$cmp_probs[$i_tp]} for all $i_tp.);
#	
#	$lambda_I is the total rate (per unit time per site) of insertions;
#	$lambda_D is the total rate (per unit time per site) of deletions;
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: \@cmp_probs, where
#	$cmp_probs[$i_tp]->[$n_A] is the "composite probability density" starting at the $i_tp th time-point
#	and with $n_A sites in between the PASs there.
#
#	The subroutine returns undef if it failed.
#
sub composite_probs_since_1st_mut_case_ii (\@$$$$\@\@$$) {

	my ($probs, $last_ub_ct_ancsites, $ub_ct_ancsites, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless (($ub_ct_ancsites >= 0) and ($last_ub_ct_ancsites >=0)) {
	    print STDOUT "[ERROR 'composite_probs_since_1st_mut_case_ii'] ILLEGAL INPUT: ", '($last_ub_ct_ancsites, $ub_ct_ancsites) = ', "($last_ub_ct_ancsites, $ub_ct_ancsites) : $! !\n\n";
	    return 0;
	}

	my $ct_tpts = @{$probs};
	unless ($ct_tpts>1) {
	    print STDOUT "[ERROR 'composite_probs_since_1st_mut_case_ii'] ILLEGAL INPUT: There are only $ct_tpts time-points: $! !\n\n";
	    return 0;
	}

    (print STDOUT "<< Subroutine 'composite_probs_since_1st_mut_case_ii' >>\n\n") if ($TEST_LEVEL>0);

   if ($TEST_LEVEL1>1) {
	print STDOUT "{OLD upper-bound of #{ancsites}} = $last_ub_ct_ancsites ,\n";
	print STDOUT "{New upper-bound of #{ancsites}} = $ub_ct_ancsites ,\n";
	print STDOUT "#{time-points} = $ct_tpts .\n\n";
   }

	my @cmp_probs = ();

	for (my $i_tp=0; $i_tp<$ct_tpts; $i_tp++) { # Main outer for-loop (over the time-points). #

    (print STDOUT "   < The $i_tp th time-point >\n\n") if ($TEST_LEVEL1 > 1);

	    my $array_probs = $probs->[$i_tp];

	    my @array_cmp_probs = ();
	    $cmp_probs[$i_tp] = \@array_cmp_probs;

	    for (my $n_A = 0; $n_A <= $ub_ct_ancsites; $n_A++) { # Main middle for-loop (over #{ancestral sites}). #


    (print STDOUT "     ( #{anc_sites} = $n_A )\n\n") if ($TEST_LEVEL1 > 2);

			# Sub-process 1: Summ over events starting with an insertion. #

		my $ub_I = ($cutoff_I < $last_ub_ct_ancsites - $n_A) ? $cutoff_I : $last_ub_ct_ancsites - $n_A; 

		my $cmp_prob_I = 0;

		for (my $l_I = 1; $l_I <= $ub_I; $l_I++) { # 1st inner for-loop (over insertion lengths). #

		    $cmp_prob_I += $freqs_I->[$l_I] * $array_probs->[$n_A + $l_I] ;

		}				# End of the 1st inner for-loop (over insertion lengths). #

		$cmp_prob_I *= ($n_A + 1) *$lambda_I ;

			# Sub-process 2: Summ over events starting with a deletion. #

		my $ub_D = ($cutoff_D < $n_A) ? $cutoff_D : $n_A ;

		my $cmp_prob_D = 0;

		for (my $l_D = 1; $l_D <= $ub_D; $l_D++) { # 2nd inner for-loop (over deletion lengths). #

			$cmp_prob_D += ($n_A - $l_D + 1) * $freqs_D->[$l_D] * $array_probs->[$n_A - $l_D] ;

		}				# End of the 2nd inner for-loop (over deletion lengths). #

		$cmp_prob_D *= $lambda_D ;

		$array_cmp_probs[$n_A] = $cmp_prob_I + $cmp_prob_D ;

	    }					# End of the main middle for-loop (over #{ancestral sites}). #

	}			# End of the main outer for-loop (over the time-points). #

    (print STDOUT "\n<< Subroutine 'composite_probs_since_1st_mut_case_ii' Ended. >>\n\n") if ($TEST_LEVEL1>0);

	return \@cmp_probs ;

}	# End of the subroutine 'composite_probs_since_1st_mut_case_ii'. #


# finish_iter_calc_probs_case_ii:
#	This subroutine calculates the right-hand side of the iteration equation, Eq.(VI-2.7),
#	by using the "composite probability densities" output by 'composite_probs_since_1st_mut_case_ii'.
#	When calculating each probability, it performs a numerical integration based on Sympson's rule.
#
#	NOTE: It is assumed that @time_points stores EQUI-SPACED time-points.
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: (\@cmp_probs, $ub_ct_ancsites, \@time_points, $lambda), where
#	$cmp_probs[$i_tp]->[$n_A] ($i_tp = 0, 1, ..., $#time_points; $n_A = 0, ..., $ub_ct_ancsites)
#	is the "composite probability density" starting at the $i_tp th time-point
#	and with $n_A sites in between the PASs there;
#
#	$ub_ct_ancsites is the upper-bound of #{ancestral sites between the PASs};
#
#	@time_points stores the time points.
#	Especially, $time_points[0] is the initial time, and $time_points[$#time_points] is the final time;
#
#	$lambda (= $lambda_I + $lambda_D) is the rate of the indel events (per unit time per site).
#
# OUTPUT: \@new_probs, where
#	@new_probs stores the probabilities calculated here. More precisely,
#	$new_probs[$i_tp]->[$n_A] ($i_tp = 0, ..., $#time_points; $n_A = 0, ..., $ub_ct_ancsites) is the probability
#	that the sequence had $n_A sites in between the PASs at the $i_tp th time-point (= $time_points[$i_tp])
#	and that it has no sites in between the PASs at the final time (= $time_points[$#time_points]).
#
#	The subroutine returns undef if it failed.
#
sub finish_iter_calc_probs_case_ii (\@$\@$) {

	my ($cmp_probs, $ub_ct_ancsites, $time_points, $lambda) = @_;

	my $ct_tpoints = @{$time_points};
	unless ($ct_tpoints>1) {
	    print STDOUT "[ERROR 'finish_iter_calc_probs_case_ii'] ILLEGAL INPUT: #{time_points} = $ct_tpoints : $! !\n\n";
	    return undef;
	}

	unless (@{$cmp_probs} == $ct_tpoints) {
	    my $size_cmp_probs = @{$cmp_probs};
	    print STDOUT "[ERROR 'finish_iter_calc_probs_case_ii'] ILLEGAL INPUT: The size of ", '@{$cmp_probs}', " (=$size_cmp_probs) CLASHES the number of time-points (=$ct_tpoints) : $! !\n\n";
	    return undef;
	}

	my $time_init = $time_points->[0];
	my $time_fin = $time_points->[$ct_tpoints-1];

	my $time_intvl = $time_fin - $time_init ;
	my $min_time_intvl = $time_intvl /($ct_tpoints -1);

   (print STDOUT "<< Subroutine 'finish_iter_calc_probs_case_ii' >>\n\n") if ($TEST_LEVEL1>0) ;

   if ($TEST_LEVEL1 > 1) {
	print STDOUT "#{time-points} = $ct_tpoints ,\n";
	print STDOUT "{Upper-bound of #{anc_sites}}  = $ub_ct_ancsites .\n\n";
	print STDOUT 'Time interval: [$time_init, $time_fin] = ', "[$time_init, $time_fin] .\n\n";
	print STDOUT '{Minute-time-interval} = ', $min_time_intvl , " .\n\n";
   }

	my @new_probs = ();

	for (my $i_tp = 0; $i_tp < $ct_tpoints; $i_tp++) { # Main outer for-loop (over time-points). #

	    ##my $array_cmp_probs = $cmp_probs->[$i_tp];

	    my @array_new_probs = ();
	    $new_probs[$i_tp] = \@array_new_probs;

	    my $tpoint = $time_points->[$i_tp];

   (print STDOUT "   < The $i_tp th time-point (time = $tpoint) >\n\n") if ($TEST_LEVEL1 > 1);

	    for (my $n_A = 0; $n_A <= $ub_ct_ancsites; $n_A++) { # Main middle for-loop (over #{ancestral sites}). #

   (print STDOUT "     ( #{ancestral sites} = $n_A )\n\n") if ($TEST_LEVEL1 > 2);

		my @integrands = ();
		for (my $j= $i_tp; $j < $ct_tpoints; $j++) { # Inner for-loop (over dummy time-points). #

		    my $cmp_prob = $cmp_probs->[$j][$n_A];
		    my $dtpoint = $time_points->[$j];

		    push @integrands, $cmp_prob * exp (-$lambda * $n_A * ($dtpoint - $tpoint) ) ;

		}				# End of the inner for-loop (over dummy time-points). # 

		my $near_simpson = nearly_Simpson (@integrands);
		unless (defined $near_simpson) {
		    print STDOUT "[ERROR 'finish_iter_calc_probs_case_ii'] 'nearly_Simpson' FAILED for ", '($i_tp, $n_A) = ', "($i_tp, $n_A) : $! !\n\n";
		    return undef;
		}

		$array_new_probs[$n_A] = $near_simpson * $min_time_intvl ;

		($n_A == 0) and ($array_new_probs[$n_A]++); # The "delta ($n_A, 0)" part.

	    }						# End of the main middle for-loop (over #{ancestral sites}). #

	}				# End of the main outer for-loop (over time-points). #

   (print STDOUT "<< Subroutine 'finish_iter_calc_probs_case_ii' Ended. >>\n\n") if ($TEST_LEVEL1>0) ;

	return \@new_probs ;

}	# End of the subroutine 'finish_iter_calc_probs_case_ii'. #



# initialize_probs_case_iii:
#	This subroutine initializes the probabilities of local PWAs of cases (i) and (iii).
#	In case (i), there are no ancestral or descendnat sites in between two neighboring PASs (preserved ancestral sites).
#	In case (iii), there are no ancestral sites but some descendant sites in between two neighboring PASs.
#
#	The subroutine was created by modifying 'initialize_probs_case_ii'.
#	[CREATED on May 7 (Thu), 2015.]
#
#
# initialize_probs_case_ii:
#	This subroutine initializes the probabilities of local PWAs of cases (i) and (ii).
#	In case (i), there are no ancestral or descendnat sites in between two neighboring PASs (preserved ancestral sites).
#	In case (ii), there are some ancestral sites but no descendant sites in between two neighboring PASs.
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: (\@probs, \@time_points, $ub_ct_decsites, $init_ct_events, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#
#	@probs will store the probabilities initialized here. More precisely,
#	$probs[$i_tp]->[$n_D] ($i_tp = 0, ..., $#time_points; $n_A = 0, ..., $ub_ct_decsites) is the probability
#	that the sequence had no sites in between the PASs at time $time_points[0]
#	and that it had $n_D sites in between the PASs at time $time_points[$i_tp].
#
#	@time_points stores the time-points that give the final point of the time interval during which the probabilities will be calculated.
#	($time_points[0] gives the initial point of the time interval.);
#
#	$ub_ct_decsites (>= 0) is the upper-bound of #{descendant sites between the PASs};
#	$init_ct_events (=0, 1, or 2) is the number of indels up to which each initial probability allows for;
#	
#	$lambda_I is the total rate (per unit time per site) of insertions;
#	$lambda_D is the total rate (per unit time per site) of deletions;
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: 1 if the subroutine successfully ended,
#	  0 if it failed.
#
sub initialize_probs_case_iii (\@\@$$$$\@\@$$) {

	my ($probs, $time_points, $ub_ct_decsites, $init_ct_events, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless ($ub_ct_decsites >= 0) {
	    print STDOUT "[ERROR 'initialize_probs_case_iii'] ILLEGAL INPUT: ", '$ub_ct_decsites = ', $ub_ct_decsites, " : $! !\n\n";
	    return 0;
	}
	unless (($init_ct_events == 0) or ($init_ct_events == 1) or ($init_ct_events == 2)) {
	    print STDOUT "[ERROR 'initialize_probs_case_iii'] ILLEGAL INPUT: ", '$init_ct_events = ', $init_ct_events, " : $! !\n\n";
	    return 0;
	}

	my $ct_tpts = @{$time_points};
	unless ($ct_tpts>1) {
	    print STDOUT "[ERROR 'initialize_probs_case_iii'] ILLEGAL INPUT: There are only $ct_tpts time-points: $! !\n\n";
	    return 0;
	}

    (print STDOUT "<< Subroutine 'initialize_probs_case_iii' >>\n\n") if ($TEST_LEVEL>0);

	my $time_init = $time_points->[0];
	my $time_fin = $time_points->[$ct_tpts-1];

   if ($TEST_LEVEL1>1) {
	print STDOUT "{Upper-bound of #{decsites}} = $ub_ct_decsites ,\n";
	print STDOUT "Initial#{events per history} = $init_ct_events ,\n";
	print STDOUT "#{time-points} = $ct_tpts ,\n";
	print STDOUT "{Initial time} = $time_init .\n\n";
	##print STDOUT "{Final time} = $time_fin .\n\n";
   }

	my $lambda = $lambda_I + $lambda_D ;

	for (my $i_tp=0; $i_tp<$ct_tpts; $i_tp++) { # Main outer for-loop (over the time-points). #

	    my $time_fin = $time_points->[$i_tp];
	    my $tlapse = $time_fin - $time_init ;

    (print STDOUT "   < The $i_tp th time-point (final time = $time_fin; time-lapse = $tlapse >\n\n") if ($TEST_LEVEL1 > 1);

	    my @array_probs = ();
	    $probs->[$i_tp] = \@array_probs;

		# First, initialize the probability with $n_D = 0. #

	    my $prob0 = 1;

	    if (($init_ct_events == 2) and ($tlapse > 0)) {
		my $mfac0_2ev = mfac_null_2events ($tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D);
		unless (defined $mfac0_2ev) {
		    print STDOUT "[ERROR 'initialize_probs_case_iii'] 'mfac_null_2events' FAILED: $! !\n\n";
		    return 0;
		}

		$prob0 += $mfac0_2ev;
	    }

	    $array_probs[0] = $prob0;

		# Then, initialize the probabilities with $n_D > 0. #

	    for (my $n_D = 1; $n_D <= $ub_ct_decsites; $n_D++) { # Main inner for-loop (over #{descendant sites}). #

   (print STDOUT '      ($n_D = ', "$n_D)\n\n") if ($TEST_LEVEL1 > 2);

		if (($init_ct_events == 0) or ($tlapse == 0)) {
		    $array_probs[$n_D] = 0;
		    next;
		}

		my $mfac =  mfac_agaps_vs_dsites_1event ($n_D, $tlapse, $lambda_I, $lambda_D, @{$freqs_I}, $cutoff_I);

		unless (defined $mfac) {
		    print STDOUT "[ERROR 'initialize_probs_case_iii'] 'mfac_agaps_vs_dsites_1event' FAILED for ", '($n_D, $cutoff_I) = ', "($n_D, $cutoff_I) : $! !\n\n";
		    return 0;
		}

    (print STDOUT '         $mfac = ', "$mfac .\n\n") if ($TEST_LEVEL1 > 3);

		if ($init_ct_events == 2) {

			# MODIFIED in creation of 'initialize_probs_case_iii'. #
		    my ($mfac_2ev, $components) = mfac_agaps_vs_dsites_2events ($n_D, $tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D);

		    ##my ($mfac_2ev, $components) = mfac_asites_vs_dgaps_2events ($n_A, $tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D);
		    ####my $mfac_2ev = mfac_asites_vs_dgaps_2events ($n_A, $tlapse, $lambda_I, $lambda_D, @{$freqs_I}, @{$freqs_D}, $cutoff_I, $cutoff_D); # A BUG fixed on May 7 (Thu), 2015. #

		    unless (defined $mfac_2ev) {
			print STDOUT "[ERROR 'initialize_probs_case_iii'] 'mfac_agaps_vs_dsites_2events' FAILED for ", '($n_D, $cutoff_I, $cutoff_D) = ', "($n_D, $cutoff_I, $cutoff_D) : $! !\n\n";
			return 0;
		    }

		    $mfac += $mfac_2ev;

    (print STDOUT '         $mfac = ', "$mfac .\n\n") if ($TEST_LEVEL1 > 3);
		}


		##$mfac *= exp (-$lambda * $n_A * $tlapse) ;  # DISABLED, because initially there were NO sites in between the PASs.

    (print STDOUT '         $mfac = ', "$mfac .\n\n") if ($TEST_LEVEL1 > 3);

		$array_probs[$n_D] = $mfac ;
		##$array_probs[$n_A] = $mfac * exp (-$lambda * $n_A * $tlapse) ;  # A relic of the predecessor, 'initialize_probs_case_ii'. #

	    }				# End of the main inner for-loop (over #{descendant sites}). #

	}			# End of the main outer for-loop (over the time-points). #

    (print STDOUT "\n<< Subroutine 'initialize_probs_case_iii' Ended. >>\n\n") if ($TEST_LEVEL1>0);

	return 1;

}	# End of the subroutine 'initialize_probs_case_iii'. #


# composite_probs_until_last_mut_case_iii:
#	This subroutine calculates the "composite probability densities"
#	that the sequence had no sites between the PASs at the initial time-point
#	and that the sequence region experienced an indel event at each specified time-point,
#	ending up with a specified #{sites in between PASs}
#	(i.e., the "densities" for the histories in cases (i) and (iii)).
#
#	The subroutine was created by modifying the predecessor, 'composite_probs_since_1st_mut_case_ii'.
#
#	[CREATED on May 7 (Thu), 2015.]
#
# composite_probs_since_1st_mut_case_ii:
#	This subroutine calculates the "composite probability densities" that
#	the sequence region (with each specified #{sites in between PASs})
#	experienced an indel event at each specified time-point and
#	that the sequence has no sites between the PASs at the final time-point
#	(i.e., for the histories in cases (i) and (ii)).
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: (\@probs, $last_ub_ct_decsites, $ub_ct_decsites, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#
#	@probs stores the probabilities initialized here. More precisely,
#	$probs[$i_tp]->[$n_D] ($i_tp = 0, ..., $#time_points; $n_D = 0, ..., $last_ub_ct_decsites) is the probability
#	that the sequence had no sites in between the PASs at the initial time-point
#	and that it had $n_D sites in between the PASs at the $i_tp th time-point;
#
#	$last_ub_ct_decsites (>= 0) is the last (i.e., previous) upper-bound of #{descendant sites between the PASs}.
#	(It must equal the upper-bound of @{$probs[$i_tp]} for all $i_tp.);
#	$ub_ct_decsites (>= 0) is the present upper-bound of #{descendant sites between the PASs}.
#	(It will become the upper-bound of @{$cmp_probs[$i_tp]} for all $i_tp.);
#	
#	$lambda_I is the total rate (per unit time per site) of insertions;
#	$lambda_D is the total rate (per unit time per site) of deletions;
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: \@cmp_probs, where
#	$cmp_probs[$i_tp]->[$n_A] is the "composite probability density" ending at the $i_tp th time-point
#	and with $n_D sites in between the PASs there.
#
#	The subroutine returns undef if it failed.
#
sub composite_probs_until_last_mut_case_iii (\@$$$$\@\@$$) {

	my ($probs, $last_ub_ct_decsites, $ub_ct_decsites, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless (($ub_ct_decsites >= 0) and ($last_ub_ct_decsites >=0)) {
	    print STDOUT "[ERROR 'composite_probs_until_last_mut_case_iii'] ILLEGAL INPUT: ", '($last_ub_ct_decsites, $ub_ct_decsites) = ', "($last_ub_ct_decsites, $ub_ct_decsites) : $! !\n\n";
	    return 0;
	}

	my $ct_tpts = @{$probs};
	unless ($ct_tpts>1) {
	    print STDOUT "[ERROR 'composite_probs_until_last_mut_case_iii'] ILLEGAL INPUT: There are only $ct_tpts time-points: $! !\n\n";
	    return 0;
	}

    (print STDOUT "<< Subroutine 'composite_probs_until_last_mut_case_iii' >>\n\n") if ($TEST_LEVEL>0);

   if ($TEST_LEVEL1>1) {
	print STDOUT "{OLD upper-bound of #{decsites}} = $last_ub_ct_decsites ,\n";
	print STDOUT "{New upper-bound of #{decsites}} = $ub_ct_decsites ,\n";
	print STDOUT "#{time-points} = $ct_tpts .\n\n";
   }

	my @cmp_probs = ();

	for (my $i_tp=0; $i_tp<$ct_tpts; $i_tp++) { # Main outer for-loop (over the time-points). #

    (print STDOUT "   < The $i_tp th time-point >\n\n") if ($TEST_LEVEL1 > 1);

	    my $array_probs = $probs->[$i_tp];

	    my @array_cmp_probs = ();
	    $cmp_probs[$i_tp] = \@array_cmp_probs;

	    for (my $n_D = 0; $n_D <= $ub_ct_decsites; $n_D++) { # Main middle for-loop (over #{descendant sites}). #


    (print STDOUT "     ( #{dec_sites} = $n_D )\n\n") if ($TEST_LEVEL1 > 2);

			# Sub-process 1: Summ over events ending with a deletion. #

		my $ub_D = ($cutoff_D < $last_ub_ct_decsites - $n_D) ? $cutoff_D : $last_ub_ct_decsites - $n_D; 

		my $cmp_prob_D = 0;

		for (my $l_D = 1; $l_D <= $ub_D; $l_D++) { # 1st inner for-loop (over deletion lengths). #

		    $cmp_prob_D += $array_probs->[$n_D + $l_D] * $freqs_D->[$l_D] ;

		}				# End of the 1st inner for-loop (over deletion lengths). #

		$cmp_prob_D *= ($n_D + 1) * $lambda_D ;

			# Sub-process 2: Summ over events ending with an insertion. #

		my $ub_I = ($cutoff_I < $n_D) ? $cutoff_I : $n_D ;

		my $cmp_prob_I = 0;

		for (my $l_I = 1; $l_I <= $ub_I; $l_I++) { # 2nd inner for-loop (over insertion lengths). #

			$cmp_prob_I += ($n_D - $l_I + 1) * $array_probs->[$n_D - $l_I] * $freqs_I->[$l_I] ;

		}				# End of the 2nd inner for-loop (over insertion lengths). #

		$cmp_prob_I *= $lambda_I ;

		$array_cmp_probs[$n_D] = $cmp_prob_I + $cmp_prob_D ;

	    }					# End of the main middle for-loop (over #{descendant sites}). #

	}			# End of the main outer for-loop (over the time-points). #

    (print STDOUT "\n<< Subroutine 'composite_probs_until_last_mut_case_iii' Ended. >>\n\n") if ($TEST_LEVEL1>0);

	return \@cmp_probs ;

}	# End of the subroutine 'composite_probs_until_last_mut_case_iii'. #


# finish_iter_calc_probs_case_iii:
#	This subroutine calculates the right-hand side of the iteration equation, Eq.(SR-8C.2),
#	by using the "composite probability densities" output by 'composite_probs_since_1st_mut_case_iii'.
#	When calculating each probability, it performs a numerical integration based on Sympson's rule.
#
#	The subroutine was created by modifying the predecessor, 'finish_iter_calc_probs_case_ii'.
#
#	[CREATED on May 7 (Thu), 2015.]
#
# finish_iter_calc_probs_case_ii:
#	This subroutine calculates the right-hand side of the iteration equation, Eq.(VI-2.7),
#	by using the "composite probability densities" output by 'composite_probs_since_1st_mut_case_ii'.
#	When calculating each probability, it performs a numerical integration based on Sympson's rule.
#
#	NOTE: It is assumed that @time_points stores EQUI-SPACED time-points.
#
#	[CREATED on May 6 (Wed), 2015.]
#
# INPUT: (\@cmp_probs, $ub_ct_decsites, \@time_points, $lambda), where
#	$cmp_probs[$i_tp]->[$n_D] ($i_tp = 0, 1, ..., $#time_points; $n_D = 0, ..., $ub_ct_decsites)
#	is the "composite probability density" ending at the $i_tp th time-point
#	and with $n_D sites in between the PASs there;
#
#	$ub_ct_decsites is the upper-bound of #{descendant sites between the PASs};
#
#	@time_points stores the time points.
#	Especially, $time_points[0] is the initial time, and $time_points[$#time_points] is the final time;
#
#	$lambda (= $lambda_I + $lambda_D) is the rate of the indel events (per unit time per site).
#
# OUTPUT: \@new_probs, where
#	@new_probs stores the probabilities calculated here. More precisely,
#	$new_probs[$i_tp]->[$n_D] ($i_tp = 0, ..., $#time_points; $n_D = 0, ..., $ub_ct_ancsites) is the probability
#	that the sequence had no sites in between the PASs at the initial time (= $time_points[0]),
#	and that it had $n_D sites in between the PASs at the $i_tp th time-point (= $time_points[$i_tp]).
#
#	The subroutine returns undef if it failed.
#
sub finish_iter_calc_probs_case_iii (\@$\@$) {

	my ($cmp_probs, $ub_ct_decsites, $time_points, $lambda) = @_;

	my $ct_tpoints = @{$time_points};
	unless ($ct_tpoints>1) {
	    print STDOUT "[ERROR 'finish_iter_calc_probs_case_iii'] ILLEGAL INPUT: #{time_points} = $ct_tpoints : $! !\n\n";
	    return undef;
	}

	unless (@{$cmp_probs} == $ct_tpoints) {
	    my $size_cmp_probs = @{$cmp_probs};
	    print STDOUT "[ERROR 'finish_iter_calc_probs_case_iii'] ILLEGAL INPUT: The size of ", '@{$cmp_probs}', " (=$size_cmp_probs) CLASHES the number of time-points (=$ct_tpoints) : $! !\n\n";
	    return undef;
	}

	my $time_init = $time_points->[0];
	my $time_fin = $time_points->[$ct_tpoints-1];

	my $time_intvl = $time_fin - $time_init ;
	my $min_time_intvl = $time_intvl /($ct_tpoints -1);

   (print STDOUT "<< Subroutine 'finish_iter_calc_probs_case_iii' >>\n\n") if ($TEST_LEVEL1>0) ;

   if ($TEST_LEVEL1 > 1) {
	print STDOUT "#{time-points} = $ct_tpoints ,\n";
	print STDOUT "{Upper-bound of #{dec_sites}}  = $ub_ct_decsites .\n\n";
	print STDOUT 'Time interval: [$time_init, $time_fin] = ', "[$time_init, $time_fin] .\n\n";
	print STDOUT '{Minute-time-interval} = ', $min_time_intvl , " .\n\n";
   }

	my @new_probs = ();

	for (my $i_tp = 0; $i_tp < $ct_tpoints; $i_tp++) { # Main outer for-loop (over time-points). #

	    ##my $array_cmp_probs = $cmp_probs->[$i_tp];

	    my @array_new_probs = ();
	    $new_probs[$i_tp] = \@array_new_probs;

	    my $tpoint = $time_points->[$i_tp];

   (print STDOUT "   < The $i_tp th time-point (time = $tpoint) >\n\n") if ($TEST_LEVEL1 > 1);

	    for (my $n_D = 0; $n_D <= $ub_ct_decsites; $n_D++) { # Main middle for-loop (over #{descendant sites}). #

   (print STDOUT "     ( #{descendant sites} = $n_D )\n\n") if ($TEST_LEVEL1 > 2);

		my @integrands = ();
		for (my $j= 0; $j <= $i_tp; $j++) { # Inner for-loop (over dummy time-points). #

		    my $cmp_prob = $cmp_probs->[$j][$n_D];
		    my $dtpoint = $time_points->[$j];

		    push @integrands, $cmp_prob * exp (-$lambda * $n_D * ($tpoint - $dtpoint) ) ;

		}				# End of the inner for-loop (over dummy time-points). # 

		my $near_simpson = nearly_Simpson (@integrands);
		unless (defined $near_simpson) {
		    print STDOUT "[ERROR 'finish_iter_calc_probs_case_iii'] 'nearly_Simpson' FAILED for ", '($i_tp, $n_D) = ', "($i_tp, $n_D) : $! !\n\n";
		    return undef;
		}

		$array_new_probs[$n_D] = $near_simpson * $min_time_intvl ;

		($n_D == 0) and ($array_new_probs[$n_D]++); # The "delta ($n_D, 0)" part.

	    }						# End of the main middle for-loop (over #{descendant sites}). #

	}				# End of the main outer for-loop (over time-points). #

   (print STDOUT "<< Subroutine 'finish_iter_calc_probs_case_iii' Ended. >>\n\n") if ($TEST_LEVEL1>0) ;

	return \@new_probs ;

}	# End of the subroutine 'finish_iter_calc_probs_case_iii'. #



1;	# a "spell" which makes the module "use"-able and/or "require"-able #


