package MyTreeMap_D;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# 
# MyTreeMap_D.pm:
#	This package maps states and their changes
#	onto the nodes and the branches, respectively, of a phylogenetic tree,
#	according to Farris's (1977) Dollo parsimony principle.
#	In a nutshell, Farris's (1977) Dollo parsimony seeks for the evolutionary scenario
#	that minimizes the number of changes, UNDER THE CONSTRAINT that
#	the 'creation' of a state occurs at most once in each column.
#	(CREATION STARTED on Apr 19 (Thu), 2012, ENDED on Apr ?? (???), 2012)
#
#	[MAJOR REVISION on 2019/05/16]:
#	A NEW subroutine, 'map_parsim_states_dollo_gptns_mrounds', was NEWLY created
#	by Substantially MODIFYING the subroutine, 'map_parsim_states_dollo_msa'.
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#	(REVISED on Nov 27 (Sun), 2011)
#	The subroutine 'min_penalties_sankoff_bo' now excludes from later consideration
#	the characters on each node that will NEVER be included in any parsiminious states.
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_III qw(
	make_node2abs_depth
	);


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	ct_presext_dollo_bu map_node2parsstat_dollo_td
	map_parsim_states_dollo_clm
	map_parsim_states_dollo_msa
	map_parsim_states_dollo_gptns_mrounds
	map_parsim_states_dollo_1clm
	);
	# ADDED 'map_parsim_states_dollo_gptns_mrounds' to the list (on 2019/05/16).
	## map_parsim_gapstates_sankoff locate_parsim_changes
	## map_parsim_states_sankoff_clm

	# This specifies the level of tests on main subroutines.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 0;
##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# This specifies the level of tests on the main satellite subroutine 'part_resolve_mlt_node_bu'.
our $TEST_LEVEL1 = 0;
##our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 5;

	# This specifies the level of tests on important satellite subroutines.
our $TEST_LEVEL2 = 0;
##our $TEST_LEVEL2 = 1;
##our $TEST_LEVEL2 = 2;
##our $TEST_LEVEL2 = 5;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###


# ct_presext_dollo_bu:
#	This satellite subroutine of 'map_parsim_states_dollo_clm'
#	assigns #{exterior descendants with state ne 'absence'} to each node,
#	in a bottom-up manner.
#
# INPUT: (\%node2children, \%node2depth, \%exnode2char, $char_abs), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2char = ($node_id => $character at the node, ...) for the exterior nodes;
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: \%node2ct_pres_extdescs, where
#	%node2ct_pres_extdescs = ($node_id => #{exterior descendants with state ne $char_abs}, ...) for ALL nodes.
#
sub ct_presext_dollo_bu (\%\%\%$) {

	my ($node2ch, $node2depth, $exnode2char, $char_abs) = @_;

	(print STDOUT "   < Subroutine 'ct_presext_dollo_bu' >\n\n") if ($TEST_LEVEL2>0);

	my %node2ct_pres_extdescs;

	my @nodes = sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth};
	foreach my $node (@nodes) { # Main outer foreach-loop #

	    my $char = $exnode2char->{$node};
	    if (defined $char) { # An exterior node #
			# Regard any character except $char_abs as the 'presence' state. #
		$node2ct_pres_extdescs{$node} = ($char eq $char_abs) ? 0 : 1;
		next;
	    }

	    my $children = $node2ch->{$node};
	    unless (defined $children) {
		print STDOUT "[ERROR 'ct_presext_dollo_bu'] The 'interior' node '$node'  has NO CHILDREN: $! !\n";
		return undef;
	    }

	    my $ct_pres = 0;

	    foreach my $ch (@{$children}) { # Middle for-loop (over the children) #
		my $ct_pres_ch = $node2ct_pres_extdescs{$ch};
		unless (defined $ct_pres_ch) {
		    print STDOUT "[ERROR 'ct_presext_dollo_bu'] The child node '$ch' of the node '$node' does NOT have #{exterior descendants with state ne '$char_abs'}: $! !\n";
		    return undef;
		}
		$ct_pres += $ct_pres_ch;

	    }		# End of the middle for-loop (over the children) #

	    $node2ct_pres_extdescs{$node} = $ct_pres;

	}			# End of the main outer foreach-loop #

	(print STDOUT "\n   --  'ct_presext_dollo_bu' Ended --\n\n") if ($TEST_LEVEL2>0);

	return \%node2ct_pres_extdescs;

}		# End of the subroutine 'ct_presext_dollo_bu' #


# map_node2parsstat_dollo_td:
#	This satellite subroutine of 'map_parsim_states_dollo_clm'
#	constructs a (unique) map from the nodes to the Dollo parsimonious states,
#	as well as a map from the branches to the state changes,
#	in a top-down manner.
#
#	[ALGORITHM]
#	For each node, 
#	(i) Count the number of child nodes whose #{exterior descendants with state ne 'absence'} > 0, then
#	(ii) Add 1 to the count if the parent node is assigned the 'presence' state,
#	(iii) Assign 'presence' to the node if the count is > 1,
#	      Otherwise assign 'absence' to it.
#
# INPUT: (\%node2pa, \%node2ch, \%node2depth, \%node2ct_pres_extdescs, $char_abs, $char_pres, $fl_full_mode), where
#	%node2pa = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%node2ct_pres_extdescs = ($node_id => #{exterir descendants with state ne $char_abs}, ...)
#	for ALL nodes;
#	$char_abs/pres is the character representing the 'absence'/'presence' state (class);
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: (\%node2parsstat, \%br2change), where
#	%node2parsstat = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the clumn mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state ne $char_abs};
#	%br2change = ($branch_id => [$stat_parent, $stat_node], ...), where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_node2parsstat_dollo_td (\%\%\%\%$$$) {

	my ($node2pa, $node2ch, $node2depth, $node2ct_pres_extdescs, $char_abs, $char_pres, $fl_full_mode) = @_;

	(print STDOUT "   < Subroutine 'map_node2parsstat_dollo_td' >\n\n") if ($TEST_LEVEL2>0);

	my @nodes = sort {$node2depth->{$a} <=> $node2depth->{$b}} keys %{$node2depth};

	my %node2parsstat;
	my %br2change;

	foreach my $node (@nodes) { # The main outer foreach-loop (over all nodes) #

	    my $children = $node2ch->{$node};
	    my $parent = $node2pa->{$node};

	    my $ct_lin_pres = 0;
	    my $state;

	    if (defined $children) {
		# An interior node #

		foreach my $ch (@{$children}) { # Inner foreach-loop (over child nodes) #
		    my $ct_pres_ch = $node2ct_pres_extdescs->{$ch};
		    ($ct_pres_ch>0) and ($ct_lin_pres++);
		}			# End of the inner foreach-loop (over child nodes) #

	    } else {
		# An exterior node #

		my $ct_pres = $node2ct_pres_extdescs->{$node};
		$state = ($ct_pres>0) ? $char_pres : $char_abs;
	    }

	    if (defined $parent) {

		my $parsstat_pa = $node2parsstat{$parent};

		unless (defined $parsstat_pa) {
		    print STDOUT "[ERROR 'map_node2parsstat_dollo_td'] UNDEFINED PARSIMONIOUS STATE for the parent '$parent' of the node '$node': $! !\n";
		    return (undef, undef);
		}

		my $state_pa = $parsstat_pa->[0];
		($state_pa ne $char_abs) and ($ct_lin_pres++);

		unless (defined $state) { # For an interior node. #
		    $state = ($ct_lin_pres>1) ? $char_pres : $char_abs;
		}

		if ($state ne $state_pa) {
		    # A change occurred along the branch immediately above the node. #
		    $br2change{$node} = [$state_pa, $state];
		}
	    } else { # For the top node. #

		$state = ($ct_lin_pres>1) ? $char_pres : $char_abs;
	    }

	    $node2parsstat{$node} = [$state, $ct_lin_pres];

	}		# End of the main outer foreach-loop (over all nodes) #


	(print STDOUT "\n   -- 'map_node2parsstat_dollo_td' Ended --\n\n") if ($TEST_LEVEL2>0);

	return (\%node2parsstat, \%br2change);

}	# End of the subroutine 'map_node2parsstat_dollo_td' #


# map_parsim_states_dollo_clm:
#	This satellite subroutine of 'map_parsim_states_dollo_msa/1clm'
#	performs Farris's (1977) Dollo parsimony algorithm on a column of MSA,
#	and find a map of states onto nodes, and of state changes onto branches,
#	that are the most parsimonious under the constraint that all 'presence' states share the same origin.
#
#	[ALGORITHM]
#	It classifies the states into two classes, 'presence' and 'absence'.
#	Any pair of 'presence' states in the same column are assumed to share the same origin.
#	Under such a constraint, the algorithm seeks for the evolutionary scenario
#	that minimizes the number of 'presence' -> 'absence' and 'absence' -> 'presence' changes.
#
#	A moment of consideration reveals that such a scenario is UNIQUELY determined by connecting all the exterior nodes with the 'presence' state via the shortest paths to one another.
#
#	A further contemplation clarifies that
#	such a scenario, in turn, can be realized by the following 2-step algorithm:
#	(i) In a bottom-up step, iteratively assign the number of exterior descendants with the 'presence' state to each interior node;
#	(ii) In a top-down step, assign either a 'presence' or an 'absence' state to each node as follows:
#	   (a) first count the children with one exterior descendant or more with the 'presence' state, and add 1 to the number if the state of its parent is 'presence';
#	   (b) assign the 'presence' state to the node if the number calculated above is more than one. Otherwise, assign the 'absence' state to the node.
#	
#
# INPUT: (\%node2children, \%node2parent, \%node2depth, \%exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2abs_pres = ($node_id => $char_abs/pres at the node, ...)
#	for the exterior nodes;
#	$char_abs/pres is the character representing the 'absence'/'presence' state;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: (\%node2parsstat, \%br2change), where
#	%node2parsstat = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the clumn mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%br2change = ($branch_id => [$stat_parent, $stat_node], ...), where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_states_dollo_clm (\%\%\%\%$$$) {

	my ($node2ch, $node2pa, $node2depth, $exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode) = @_;

	(print STDOUT "\n < Subroutine 'map_parsim_states_dollo_clm' >\n\n") if ($TEST_LEVEL1>0);
    if ($TEST_LEVEL1>1) {
	print STDOUT '$char_abs = ', $char_abs, ',   $char_pres = ', $char_pres, " .\n\n";
	print STDOUT '%{$exnode2abs_pres} = (', "\n";
	foreach my $node (sort {$a <=> $b} keys %{$exnode2abs_pres}) {
	    print STDOUT $node, '  =>  ', $exnode2abs_pres->{$node}, ",\n";
	}
	print STDOUT ")\n\n";
    }

		## Assign #{exterior descendants with state ne $char_abs} to each node
		## in a bottom-up manner

	# %{$node2ct_pres_extdescs} = ($node_id => #{exterior descendants with state ne $char_abs}, ...) for all nodes.
	#
	my $node2ct_pres_extdescs = ct_presext_dollo_bu (%{$node2ch}, %{$node2depth}, %{$exnode2abs_pres}, $char_abs);
	unless (defined $node2ct_pres_extdescs) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_clm'] The subroutine 'ct_presext_dollo_bu' FAILED: $! !\n";
	    return (undef, undef);
	}

    if ($TEST_LEVEL1>1) {
	print STDOUT "\n{ Output of 'min_penalties_dollo_bo' }\n\n"; 
	print STDOUT '%{$node2ct_pres_extdescs} = ( # Node_ID => #{exterior descendants with state ne $char_abs}', "\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2ct_pres_extdescs}) {
	    print STDOUT "\t", $node, '  =>   ', $node2ct_pres_extdescs->{$node}, ' ,', "\n";
	}
	print STDOUT "\t)\n\n";
    }

		## Identify the map from the nodes to the parsimonious states,
		## as well as from the branches to the state changes,
		## in a top-down manner

	my ($node2parsstat, $br2change) = map_node2parsstat_dollo_td (%{$node2pa}, %{$node2ch}, %{$node2depth}, %{$node2ct_pres_extdescs}, $char_abs, $char_pres, $fl_full_mode);
	##my $set_node2parsstat = enum_node2parsstat_sankoff_td (%{$node2pa}, %{$node2depth}, %{$node2min_penalties}, @{$pen_mtrx}, $fl_full_mode);
	unless (defined $node2parsstat) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_clm'] The subroutine 'map_node2parsstat_dollo_td' FAILED: $! !\n";
	    return (undef, undef);
	}

    if ($TEST_LEVEL1>1) {
	print STDOUT "\n{ Output of 'map_node2parsstat_dollo_td'  }\n\n";
	print STDOUT '[ Content of %{$node2parsstat} ]', "\n\n";
	print STDOUT join ("\t", '#Node_ID', 'State', '#{presence_lins}'), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2parsstat}) {
	    print STDOUT join ("\t", $node, @{$node2parsstat->{$node}}), "\n";
	}
	print STDOUT "\n\n";
	print STDOUT '[ Content of %{br2change} ]', "\n\n";
	print STDOUT join ("\t", '#Br_ID', 'State_parent', 'State_node'), "\n\n";
	foreach my $brid (sort {$a <=> $b} keys %{$br2change}) {
	    print STDOUT join ("\t", $brid, @{$br2change->{$brid}}), "\n";
	}
	print STDOUT "\n\n";
    }

	(print STDERR " --- 'map_parsim_states_dollo_clm' ENDED ---\n\n") if ($TEST_LEVEL1>0);

	return ($node2parsstat, $br2change);

}	# End of the subroutine 'map_parsim_states_dollo_clm' # 



# map_parsim_states_dollo_msa:
#	This subroutine performs Farris's (1977) Dollo parsimony algorithm on a MSA,
#	and serially find a unique parsimonious map of states onto nodes,
#	as well as that of state changes onto branches,
#	for each of the columns (which may not necessarily be of 1-character length)
#	in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# map_parsim_states_dollo_1clm:
#	This subroutine handles just a single column in a MSA,
#	namely compares just a single trait across taxa,
#	and looks for a (unique) parsimonious scenario
#	according to Farris's (1977) Dollo parsimony algorithm.
#
#	For the description of the algorithm,
#	refer to the header of 'map_parsim_states_dollo_clm'.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, $char_abs, $char_pres, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$char_abs is the character representing the 'absence' state;
#	$char_pres is the character representing the 'presence' state,
#	which will be used only for the output;
#	$fl_full_mode = 1 if you want the full mode,
#		      = 0 if the brief mode is enough.
#
# OUTPUT: (\@set_node2parsstat, \@set_br2change), where
#	%{$set_node2parsstat[$l]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $l th site mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%{$set_br2change[$l]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches for the $l th site, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_states_dollo_msa ($\%\%\%\@\@$$$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa,
	$char_abs, $char_pres, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	(print STDOUT "<< Subroutine 'map_parsim_states_dollo_msa' (#{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL>0);

	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef);
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The number of sequences (", scalar (@{$msa}), ") in the input MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef);
	}
	my $len_char_abs = length $char_abs;
	unless ($len_char_abs>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The 'absence' character (= '$char_abs') is NULL: $! !\n";
	    return (undef, undef);
	}
	my $len_char_pres = length $char_pres;
	unless ($len_char_pres == $len_char_abs) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The length (=$len_char_abs) of the 'absence' character, '$char_abs', CLASHES with that (=$len_char_pres) of the 'presence' character, '$char_pres': $! !\n";
	    return (undef, undef);
	}
	my $len_seq  = length ${$msa->[0]};
	unless ($len_seq>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] We do NOT HANDLE NULL MSAs: $! !\n";
	    return (undef, undef);
	}
	my $ct_clms = int ($len_seq / $len_char_abs);
	unless ($ct_clms>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The MSA has NO full column: $! !\n";
	    return (undef, undef);
	}

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_states_dollo_msa'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return (undef, undef);
	}

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_states_dollo_msa'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return (undef, undef);
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #


		## Main Process ##

	my @set_node2parsstat = my @set_br2change = ();

	for (my $l=0; $l<$ct_clms; $l++) { # The main outer for-loop #

	    (print STDOUT "[ The $l th column ]\n\n") if ($TEST_LEVEL>0);

	    my %exnode2abs_pres; # ($node_id => $char_abs/pres, ...) for the exterior nodes

	    my @clm = ();
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st main inner for-loop #

		my $node_id = $node_ids[$k];

		my $char = substr (${$msa->[$k]}, 0, $len_char_abs, '');
		push @clm, $char;
		##my $char = $clm->[$k];
		($char eq $char_abs) or ($char = $char_pres); # Regard the state as 'presence' if it is not 'absence'.
		$exnode2abs_pres{$node_id} = $char;

	    }				# End of the 1st main inner for-loop #

	    (print STDOUT "Characters in the column: ", join (';', @clm), "\n\n") if ($TEST_LEVEL>1);


	    my ($node2parsstat, $br2change) = map_parsim_states_dollo_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode) ;
	    unless (defined $node2parsstat) {
		print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] 'map_parsim_states_dollo_clm' FAILED: $! !\n";
		return (undef, undef);
	    }

	    $set_node2parsstat[$l] = $node2parsstat;
	    $set_br2change[$l] = $br2change;

	}		# End of the main outer for-loop #

	(print STDOUT "=== End of the Subroutine 'map_parsim_states_dollo_msa' ===\n\n") if ($TEST_LEVEL>0);

	return (\@set_node2parsstat, \@set_br2change);
	##return ($node2parsstat, $br2change);

}	# End of the subroutine 'map_parsim_states_dollo_msa' #


# map_parsim_states_dollo_gptns_mrounds:
#	This satellite subroutine of 'enum_psm_indel_scns_for_msa_hs_mrounds'
#	was created by substantially MODIFYING its predecessor, 'map_parsim_states_dollo_msa'.
#	It basically does the same thing as the predecessor.
#	There are, though, 6 MAJOR DIFFERENCES between the two subroutines:
#	(1) The formats of the input set of gap-patterns differ: in the predecessor, it took the form of an MSA (i.e., a set of possibly gapped sequences), in this version, it takes the form of a set of strings, each representing a gap-pattern;
#	(2) In this new version, @exnodes_in_msa replaces the combination of %name2id and @seqnames;
#	(3) The major outputs of the predecessor, @set_node2(d)parsstat & @set_br2changes,
#	    are now fed into this new subroutine as INPUTS;
#	(5) As a result of (2), the subroutine performs the Dollo parsimony analyses
#	   ONLY on newly incorporated gap-patterns;
#	(6) Instead of constructing %node2depth on the fly, this new version receives it as an input;
#	(7) For speeding up the process, we ASSUME that
#		length ($char_abs) == length ($char_pres) == 1. 
#	[CREATED on 2019/05/16.]
#
# map_parsim_states_dollo_msa:
#	This subroutine performs Farris's (1977) Dollo parsimony algorithm on a MSA,
#	and serially find a unique parsimonious map of states onto nodes,
#	as well as that of state changes onto branches,
#	for each of the columns (which may not necessarily be of 1-character length)
#	in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# map_parsim_states_dollo_1clm:
#	This subroutine handles just a single column in a MSA,
#	namely compares just a single trait across taxa,
#	and looks for a (unique) parsimonious scenario
#	according to Farris's (1977) Dollo parsimony algorithm.
#
#	For the description of the algorithm,
#	refer to the header of 'map_parsim_states_dollo_clm'.
#
# [ INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, $char_abs, $char_pres, $fl_full_mode), where ] ... for the predecessor, 'map_parsim_states_dollo_msa'.
# INPUT: ($topnode_id, \%node2children, \%node2parent,
#		\@exnodeids_in_msa, \@set_gpatterns, \@set_node2dparsstat, \@set_br2change, \%node2depth,
#		$char_abs, $char_pres, $fl_full_mode), where
#
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#
#	[ %name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA; ] ... for the predecessor, 'map_parsim_states_dollo_msa'.
#
#		# ADDED in creation of 'map_parsim_states_dollo_gptns_mrounds'. (0a) #
#
#	$exnodeids_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#	[Typically, @exnodeids_in_msa is an output of, e.g., 'associate_nodeids_w_msa_rows' in "MyANEX_Supple.pm".]
#
#	${$set_gpatterns[$indx_gp]} is the string representing the $indx_gp th gap pattern in the input MSA, created by concatenating $char_pres/abs in the sequences;
#
#	%{$set_node2dparsstat[$l]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $l th site mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#
#	%{$set_br2change[$l]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches for the $l th site, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes;
#
#		# END of "ADDED in creation of 'map_parsim_states_dollo_gptns_mrounds'. (0a)" #
#
#	$char_abs is the character representing the 'absence' state;
#	$char_pres is the character representing the 'presence' state,
#	which will be used only for the output;
#
#	$fl_full_mode = 1 if you want the full mode,
#		      = 0 if the brief mode is enough.
#
# [ OUTPUT: (\@set_node2parsstat, \@set_br2change), where
#	%{$set_node2parsstat[$l]} = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the $l th site mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%{$set_br2change[$l]} = ($branch_id => [$stat_parent, $stat_node], ...)
#	for the map of state changes onto the branches for the $l th site, where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given. ] ... for the predecessor, 'map_parsim_states_dollo_msa'.
#
# OUTPUT: ($ct_new_dollo_hsts), where
#
#	$ct_new_dollo_hsts is the number of new Dollo parsimonious histories computed in this round of the subroutine.
#
## sub map_parsim_states_dollo_msa ($\%\%\%\@\@$$$) {
#	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa,
#	$char_abs, $char_pres, $fl_full_mode) = @_;
#
sub map_parsim_states_dollo_gptns_mrounds ($\%\%\@\@\@\@\%$$$) {

	my ($topnode_id, $node2ch, $node2pa,
		$exnodeids_in_msa, $set_gpatterns, $set_node2dparsstat, $set_br2change, $node2depth,
		$char_abs, $char_pres, $fl_full_mode) = @_; # Major Revision in creation of 'map_parsim_states_dollo_gptns_mrounds'. (0)

		# Major Revision in creation of 'map_parsim_states_dollo_gptns_mrounds'. (1) #

	my $ct_rows = @{$exnodeids_in_msa};
	my $ct_gpatterns = @{$set_gpatterns};
	my $old_ct_gpatterns = @{$set_node2dparsstat};

    (print STDOUT "<< Subroutine 'map_parsim_states_dollo_gptns_mrounds' (#{rows} = $ct_rows, #{gap-patterns} = $ct_gpatterns, #{gap-patterns}_OLD = $old_ct_gpatterns) >>\n\n") if ($TEST_LEVEL>0);

	unless ($ct_rows>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_gptns_mrounds'] NO ROWs in the input MSA: $! !\n";
	    return undef;
	}

	unless ($ct_gpatterns>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_gptns_mrounds'] There are NO GAP-PATTERNS to examine: $! !\n";
	    return 0;
	}

	unless ($old_ct_gpatterns == scalar (@{$set_br2change}) ) {
	    my $old_ct_gpatterns2 = @{$set_br2change};
	    print STDOUT "[ERROR 'map_parsim_states_dollo_gptns_mrounds'] The size of ", '@{$set_node2dparsstat}', "(=$old_ct_gpatterns) CLASHES with that of ", '@{$set_br2change}', "(=$old_ct_gpatterns2): $! !\n";
	    return undef;
	}

=For the predecessor, 'map_parsim_states_dollo_msa'. (1)
	my $ct_seqs = @{$seqnames};
	(print STDOUT "<< Subroutine 'map_parsim_states_dollo_msa' (#{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL>0);

	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef);
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The number of sequences (", scalar (@{$msa}), ") in the input MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef);
	}
	my $len_char_abs = length $char_abs;
	unless ($len_char_abs>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The 'absence' character (= '$char_abs') is NULL: $! !\n";
	    return (undef, undef);
	}
	my $len_char_pres = length $char_pres;
	unless ($len_char_pres == $len_char_abs) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The length (=$len_char_abs) of the 'absence' character, '$char_abs', CLASHES with that (=$len_char_pres) of the 'presence' character, '$char_pres': $! !\n";
	    return (undef, undef);
	}
	my $len_seq  = length ${$msa->[0]};
	unless ($len_seq>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] We do NOT HANDLE NULL MSAs: $! !\n";
	    return (undef, undef);
	}
	my $ct_clms = int ($len_seq / $len_char_abs);
	unless ($ct_clms>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] The MSA has NO full column: $! !\n";
	    return (undef, undef);
	}
=cut # End of "For the predecessor, 'map_parsim_states_dollo_msa'. (1)"

		# End of "Major Revision in creation of 'map_parsim_states_dollo_gptns_mrounds'. (1)" #


		## Prepare auxiliary arrays/hashes ##

		# Major Revision in creation of 'map_parsim_states_dollo_gptns_mrounds'. (2) #

=For the predecessor, 'map_parsim_states_dollo_msa'. (2)

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_states_dollo_msa'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return (undef, undef);
	}

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_states_dollo_msa'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return (undef, undef);
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

=cut # End of "For the predecessor, 'map_parsim_states_dollo_msa'. (2)"

		# End of "Major Revision in creation of 'map_parsim_states_dollo_gptns_mrounds'. (2)" #

		### Main Process ###

		# MAJOR REVISION in creation of 'map_parsim_states_dollo_gptns_mrounds'. (MAIN) #

	##my @node_ids = @{$exnodeids_in_msa}; # $node_ids[$k] is the external node ID corresponding to the $k th row of the input MSA.

	my $ct_new_dollo_hsts = 0;

	for (my $i=$old_ct_gpatterns; $i<$ct_gpatterns; $i++) { # The MAIN outer for-loop (over newly incorporated gap-patterns). #

    (print STDOUT "[ The $i th gap-pattern ]\n\n") if ($TEST_LEVEL>0);

	    my %exnode2abs_pres; # ($node_id => $char_abs/pres, ...) for the external nodes.

	    my @clm = split //, ${$set_gpatterns->[$i]};
	    for (my $k=0; $k<$ct_rows; $k++) { # The 1st main inner for-loop (over MSA rows).

		my $node_id = $exnodeids_in_msa->[$k];
		my $char = $clm[$k];
		unless ($char eq $char_abs) { $char = $char_pres; } # Regard the state as 'presence' if it is NOT 'absence'.
		$exnode2abs_pres{$node_id} = $char;

	    } # End of the 1st main inner for-loop (over MSA rows).

    (print STDOUT "Characters in the column: ", join (';', @clm), "\n\n") if ($TEST_LEVEL>1);

	    my ($node2dparsstat, $br2change) = map_parsim_states_dollo_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode) ;
	    unless (defined $node2dparsstat) {
		print STDOUT "[ERROR 'map_parsim_states_dollo_gptns_mrounds'] 'map_parsim_states_dollo_clm' FAILED: $! !\n";
		return undef;
	    }

	    $set_node2dparsstat->[$i] = $node2dparsstat;
	    $set_br2change->[$i] = $br2change;

	    $ct_new_dollo_hsts++;

	} # End of the MAIN outer for-loop (over newly incorporated gap-patterns). #

=For the predecessor, 'map_parsim_states_dollo_msa'. (MAIN)

	my @set_node2parsstat = my @set_br2change = ();

	for (my $l=0; $l<$ct_clms; $l++) { # The main outer for-loop #

	    (print STDOUT "[ The $l th column ]\n\n") if ($TEST_LEVEL>0);

	    my %exnode2abs_pres; # ($node_id => $char_abs/pres, ...) for the exterior nodes

	    my @clm = ();
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st main inner for-loop #

		my $node_id = $node_ids[$k];

		my $char = substr (${$msa->[$k]}, 0, $len_char_abs, '');
		push @clm, $char;
		##my $char = $clm->[$k];
		($char eq $char_abs) or ($char = $char_pres); # Regard the state as 'presence' if it is not 'absence'.
		$exnode2abs_pres{$node_id} = $char;

	    }				# End of the 1st main inner for-loop #

	    (print STDOUT "Characters in the column: ", join (';', @clm), "\n\n") if ($TEST_LEVEL>1);


	    my ($node2parsstat, $br2change) = map_parsim_states_dollo_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode) ;
	    unless (defined $node2parsstat) {
		print STDOUT "[ERROR 'map_parsim_states_dollo_msa'] 'map_parsim_states_dollo_clm' FAILED: $! !\n";
		return (undef, undef);
	    }

	    $set_node2parsstat[$l] = $node2parsstat;
	    $set_br2change[$l] = $br2change;

	}		# End of the main outer for-loop #

=cut # END of "For the predecessor, 'map_parsim_states_dollo_msa'. (MAIN)"

		# END of "MAJOR REVISION in creation of 'map_parsim_states_dollo_gptns_mrounds'. (MAIN)" #

    (print STDOUT "<< Subroutine 'map_parsim_states_dollo_gptns_mrounds' ENDED (w/ ", '$ct_new_dollow_hsts=', "$ct_new_dollo_hsts) />/>\n\n") if ($TEST_LEVEL>0); # MODIFIED in creation of 'map_parsim_states_dollo_gptns_mrounds'. #
    ##(print STDOUT "=== End of the Subroutine 'map_parsim_states_dollo_msa' ===\n\n") if ($TEST_LEVEL>0);

	return $ct_new_dollo_hsts; # MODIFIED in creation of 'map_parsim_states_dollo_gptns_mrounds'. #
	##return (\@set_node2parsstat, \@set_br2change);
	####return ($node2parsstat, $br2change);

}	# End of the subroutine 'map_parsim_states_dollo_gptns_mrounds' #


# map_parsim_states_dollo_1clm:
#	This subroutine handles just a single column in a MSA,
#	namely compares just a single trait across taxa,
#	and looks for a (unique) parsimonious scenario
#	according to Farris's (1977) Dollo parsimony algorithm.
#
#	For the description of the algorithm,
#	refer to the header of 'map_parsim_states_dollo_clm'.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@clm, $char_abs, $char_pres, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	$clm[$k] is the character of the $k th sequence (or taxon);
#	$char_abs is the character representing the 'absence' state;
#	$char_pres is the character representing the 'presence' state,
#	which will be used only for the output;
#	$fl_full_mode = 1 if you want the full mode,
#		      = 0 if the brief mode is enough.
#
# OUTPUT: (\%node2parsstat, \%br2change), where
#	%node2parsstat = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the clumn mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the state on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#	%br2change = ($branch_id => [$stat_parent, $stat_node], ...), where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch,
#	$stat_node (=$char_abs/pres) is the state of of the node, and
#	$stat_parent (=$char_abs/pres) is the state of the parent node.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_states_dollo_1clm ($\%\%\%\@\@$$$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $clm,
	$char_abs, $char_pres, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	(print STDOUT "<< Subroutine 'map_parsim_states_dollo_1clm' (#{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL>0);

	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_1clm'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef);
	}
	unless ($ct_seqs == @{$clm}) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_1clm'] The number of sequences (", scalar (@{$clm}), ") in the column CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef);
	}
	my $len_char  = length $clm->[0];
	unless ($len_char>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_1clm'] We do NOT HANDLE NULL CHARACTERS: $! !\n";
	    return (undef, undef);
	}
	my $len_char_abs = length $char_abs;
	unless ($len_char_abs>0) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_1clm'] The 'absence' character (= '$char_abs') is NULL: $! !\n";
	    return (undef, undef);
	}
	if ($len_char != $len_char_abs) {
	    print STDOUT "[WARNING 'map_parsim_states_dollo_1clm'] The character seems be of length $len_char, instead of $len_char_abs (= length '$char_abs'), but I will go on anyway ...\n\n";
	}

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_states_dollo_1clm'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return (undef, undef);
	}

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_states_dollo_1clm'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return (undef, undef);
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #


		## Main Process ##

	my %exnode2abs_pres; # ($node_id => $char_abs/pres, ...) for the exterior nodes

	for (my $k=0; $k<$ct_seqs; $k++) { # The 1st main for-loop #

	    my $node_id = $node_ids[$k];

	    my $char = $clm->[$k];
	    ($char eq $char_abs) or ($char = $char_pres); # Regard the state as 'presence' if it is not 'absence'.
	    $exnode2abs_pres{$node_id} = $char;

	}				# End of the 1st main for-loop #


	my ($node2parsstat, $br2change) = map_parsim_states_dollo_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2abs_pres, $char_abs, $char_pres, $fl_full_mode) ;
	unless (defined $node2parsstat) {
	    print STDOUT "[ERROR 'map_parsim_states_dollo_1clm'] 'map_parsim_states_dollo_clm' FAILED: $! !\n";
	    return (undef, undef);
	}

	(print STDOUT "=== End of the Subroutine 'map_parsim_states_dollo_1clm' ===\n\n") if ($TEST_LEVEL>0);

	return ($node2parsstat, $br2change);

}	# End of the subroutine 'map_parsim_states_dollo_1clm' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


