#!/usr/bin/perl
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# srd_map.ver0.6.pl:
#	This Perl script receives a reconstructed multiple sequence alignment (MSA) and a phylogenetic tree,
#	and outputs two versions of "Substitutional Residue-Difference (SRD) Map",
#	each of which displays the substitutional residue-differences along each branch at each site,
#	computed in some way.
#	One is the "Expected" SRD Map, whose SRDs are EXPECTED from the residue states of the extant sequences
#	"upstream" and "downstream" of the branch;
#	The other is the "Observed" SRD Map, whose SRDs are a kind of "OBSERVATION",
#	computed from the residue states of all the extant sequences in the MSA.
#	Both of the SRD Maps are computed based on the likelihood theory with a stochastic model of residue substitutions
#	(see, e.g., J. Felsenstein 1980, 2004; Z.-H. Yang 2006?).
#	(The specific recipe of the computation is described in K. Ezawa 2020c.)
#	The script was created by somewhat modifying the predecessor script, 'detect_purge_cands.ver0.5.pl'.
#	[CREATION STARTED on 2020/03/04, ENDED on 2020/03/07; REVISED on 2020/03/10]
#
#	[NOTE (on 2020/03/04)]:
#	This script puts a gap ('-') at a site along a branch of the SRD Map,
#	if the Dollo-parsimonious indel history shows the "absence" state on either end (or both ends) of the branch at the site.
#
# detect_purge_cands.ver0.5.pl:
#	This Perl script detects regions candidates of "purge"-errors, via 'detect_purge_cands2',
#	and outputs the results in ascending order of the boundary coordinates.
#	The script was created by modifying the test script, 'test_detect_purge_cands2.pl' (at $RIGHT_NOW_WORKING_CHICHIBU8).
#	[CREATION STARTED on 2019/06/16, ENDED on 2019/06/16.]
#
#	[MAJOR REVISION on 2019/07/07]:
#	MOVED (nearly) ALL the subroutines within this script to the package, "MyANEX_IO.pm".
#
#	[MAJOR REVISION on 2020/03/04]:
#	DISABLED the attempt to parsimonize the indel history,
#	and, instead, FED the 'detect_purge_cands2' with the DOLLO parsimonious indel history,
#	because of the following reaaon:
#	if, at a site on a node, the parsimonized history has a "presence" state whereas the
#	Dollo parsimonious history has an "absence" state,
#	the residue-state information is virtually MISSING there, after all.
#	Then, it is likely that the "observed" substitutional difference involving the site-node
#	is indistinguishable from that of random matching,
#	and thus they could often be ERRONEOUSLY identified as a part of a "purge-candidate".
#
# test_detect_purge_cands2.pl:
#	This Perl script examines the behaviors of the main subroutine, 'detect_purge_cands2',
#	of the Perl package "MyDetect_purge_cands.pm".
#	The script was created by slightly modifying the predecessor, 'test_detect_purge_cands.pl'.
#
# test_detect_purge_cands.pl:
#	This Perl script examines the behaviors of the main subroutine, 'detect_purge_cands',
#	of the Perl package "MyDetect_purge_cands.pm".
#	The script was created by taking advantage of the backbone of 'test_calc_pv_obscts_sbsts_wd_br.pl'.
#	[CREATION STARTED on Jan 11 (Thu), 2018, ENDED on Jan dd (www), 2018.]
#
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	); # ADDED 'open_output_file' in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #
	# list_sessions_ext
	# tfm_labels2indices

	# ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). (HD 1) #
use MyArrayManipu qw(
	multiple_sort
	);
	# End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). (HD 1)" #
 
use MyMalnIO qw(
	extract_alignment
	pr_ODEN_input
	);

use MyTreeManipu_II qw(
	output_tree 
	);
	# make_node2parent
 
use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg
	);

use MyTreeMap_indels_spt_odr_finer qw(
	enum_psm_indel_hsts_for_msa_finer_PRE
	enum_psm_indel_hsts_for_msa_finer_1ST_HF
	);

use MyLikelihoods_respattern_msa_dna qw(
	calc_br2tr_mtrx_TN
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);

use MyDetect_purge_cands qw(
	detect_purge_cands2
	); # Added 'detect_purge_cands2' in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). #
	# detect_purge_cands
	## calc_br2cw_expcts_sbsts_ave
	## sum_arrays_over_branches
	## calc_br2cw_expcts_obscts_sbsts
	## calc_pv_obscts_sbsts_wd_br
	## calc_modpv_obscts_sbsts_wd_br

	# Added in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). (HD 2) #
use MyTreeMap_indels_spt_odr2 qw(
	bch_br_parsimonize_dollo_indel_histories
	);
	## cal_init_psm_cands_for_sgl_msa
	# End of "Added in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). (HD 2)" #

	# Added in revision of 'detect_purge_cands.ver0.5.pl' (on 2019/06/19). (HD 3) #
use MyTreeMap_sppl qw(
	wrapper_mk_node2exoffs_msa_rows
	);
	# End of "Added in revision of 'detect_purge_cands.ver0.5.pl' (on 2019/06/19). (HD 3)" #

	### ADDED in MAJOR REVISION on 2019/07/07. ###
use MyANEX_IO qw(
	readin_dawg_tree2
	print_out_bch_br_parsimonize_dollo_indel_histories
	);
	### END of "ADDED in MAJOR REVISION on 2019/07/07." ###


	### ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/05. ###
use MySRD_Map qw(
	mk_srd_map
	print_srd_map
	);
	### END of "ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/05." ###



=OBSOLETE as of 2019/07/07.
sub readin_dawg_tree2 (\$);
sub print_out_bch_br_parsimonize_dollo_indel_histories (\$$$\@\@\@\@\@\@); # Imported from 'anex.ver0.5.pl', in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #
=cut # END of "OBSOLETE as of 2019/07/07."


$| = 1; # Flush the input buffer as soon as it receives data.

my ($plscript_name) = ($0 =~ m!([^/]+)$!);
my $program_name = $plscript_name;



	##### Glogal parameter setting #####

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
##our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

		# Added in revision (on 2019/06/18). #
#our $DIS_LEN = 50; # The maximum length of the (sub)sequence displayed in a line.
our $DIS_LEN = 60; # The maximum length of the (sub)sequence displayed in a line.
		# End of "Added in revision (on 2019/06/18)." #

$MyANEX_IO::DIS_LEN = $DIS_LEN; # Added in REVISION (on 2019/07/07).

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (GL1)

our $MIN_WDSIZE = 3;
our $MAX_WDSIZE = 6;


our $THRSH_PV = 0.05; # P-value threshold (based on the substitution model). 
our $IF_MODPV = 0; # = 1 if you use modified P-value, = 0 if you use normal P-value.

#our $THRSH_PV0 = 1; # P-value threshold (based on the random match model).
#our $THRSH_PV0 = 0; # P-value threshold (based on the random match model).
our $THRSH_PV0 = 0.20; # P-value threshold (based on the random match model).

=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (GL1)"

our $GAP = '-';
our $CHAR_ABS = $GAP;
our $CHAR_PRES = 'N';

our @RESIDUES = our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
##our @RESIDUES = our @NUCLEOTIDES = @MyPRANK_Scores::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our $SV_RATIO_R = 1; # Ratio of purine-type transition rate to transversion rate.
our $SV_RATIO_Y = 1; # Ratio of pyrimidine-type transition rate to transversion rate.

##my @cond_probs_gapped = (); # ADDED in creation of 'test_calc_jnt_prob_ancresstats2.pl'. #
##for (1 .. $CT_NUCLS) { push @cond_probs_gapped, 1; }

	# Equilibrium residue frequencies. #
my @root_resfreqs = (0.25, 0.25, 0.25, 0.25);
##my @root_resfreqs = (0.1, 0.2, 0.3, 0.4);

my @labels_resfreqs = ();
foreach my $res (@NUCLEOTIDES) { push @labels_resfreqs, "f($res)"; }

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (GL2)
our @SBJ_REGIONS = ([0, 200], [500, 700], [1150, 1350]);
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (GL2)"


		# (ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (GL1) #
our $DIS_LEN_SRD = 30; # Maximum number of sites (i.e., SRDs) to be displayed in a line of the SRD Map.
our $CT_EFF_DGTS = 2;  # Number of effective digits with which each SRD will be displayed.
		# End of "(ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (GL1)" #


	### Parameters to adjust the verboseness of the test outputs. ###
	### (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose.)

	# For this script. #
#our $TEST_LEVEL = 5;
our $TEST_LEVEL = 1;

        # For 'calc_prob_descresstats_cond_ancstat2'. #
$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 0;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 5;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = 6;

	# For 'calc_joint_prob_ancresstats_and_ancstat2'. #
	# (ADDED in creation of 'test_calc_jnt_prob_ancresstats2.pl'.)
$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = 0;
##$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa = 6;


	# For 'enum_psm_indel_hsts_for_msa_finer_PRE'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE = 0;

	# For 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'.
$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_1ST = 0;

		# Added in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/18). (TL) #
	# For 'br_parsimonize_dollo_indel_history'. 
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs = 0;

	# For 'bch_br_parsimonize_dollo_indel_histories'.
$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs = 0;
		# End of "Added in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/18). (TL)" #


	# For 'detect_purge_cands'. #
##$MyDetect_purge_cands::TEST_LEVEL = 0;
$MyDetect_purge_cands::TEST_LEVEL = 5;

	##### END of "Glogal parameter setting" #####


	##### ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (ARGV) #####

	### Take in the Command-Line Arguments. ###

unless (@ARGV > 2) {
	print STDERR "\nUsage:  [perl] $program_name {input_msa_file} {input_tree_file} {output_dir} [options]\n";
	print STDERR "   where the [options] can consist of any (or none) of the following:\n\n";

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (ARGV)
	print STDERR "            --MIN_WD_SIZE=integer      for the minimum window size (default=$MIN_WDSIZE);\n";
	print STDERR "            --MAX_WD_SIZE=integer      for the maximum window size (default=$MAX_WDSIZE);\n";
	print STDERR "            --THRSH_PV=real(<1)     for the threshold value for the P-value measuring the deviation from the specified substitution model (default=$THRSH_PV);\n";
	print STDERR "            --IF_MODPV=(0 or 1)     to determine whether to use a modified P-value(=1) or not(=0) (default=$IF_MODPV);\n";
	print STDERR "            --THRSH_PV0=real(<1)    for the threshold value for the P-value measuring the deviation from the random-match model (default=$THRSH_PV0);\n\n";
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (ARGV)"

	print STDERR "            --BASE_FREQS=f(A),f(G),f(C),f(T)   (default=", join (',', @root_resfreqs), "),\n"; 
	print STDERR "                     where f(N) is the frequency of nucleotide 'N' at the root;\n\n";
	print STDERR "            --TN_PARAMS=alpha1,alpha2,beta     (default=$SV_RATIO_R,$SV_RATIO_Y,1),\n";
	print STDERR "                     where alpha1 is the transitional substitution rate between A and G,\n";
	print STDERR "                           alpha2 is the transitional substitution rate between C and T,\n";
	##print STDERR "                           beta is the transversional substitution rate.\n";
	print STDERR "                           beta is the transversional substitution rate;\n\n"; # Modified in creation of 'srd_map.ver0.6.pl' on 2020/03/06. (ARGV)

			# (ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (ARGV) #
	print STDERR "            --DIS_LEN_SRD=integer     for the maximum number of sites in a line of the SRD Map (default=$DIS_LEN_SRD);\n\n";
	print STDERR "            --CT_EFF_DGTS=integer     for the number of effective digits when displaying each SRD (default=$CT_EFF_DGTS).\n";
			# End of "(ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (ARGV)" #

	die "$!\n\n";
}
my ($inaln_file, $intree_file, $out_dir) = @ARGV[0 .. 2];

	# NOTE: The output file system will be specified & prepared below. #

	### END of "Take in the Command-Line Arguments." ###

	##### END of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (ARGV)" #####
	##### ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PL 1) ####


	### Check the Command-Line Options. ###

for (my $k=3; $k <= $#ARGV; $k++) { # Preliminary for-loop (over additional command-line arguments). #

	my $clarg = $ARGV[$k];
	unless ($clarg =~ s/^(\-)+//) { next; } # Skip if the argument lacks preceding '-'s. #

	my ($option, $value) = split /\=/, $clarg;
	unless ((defined $value) and ($value ne '')) {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED FORMAT of the commmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument in an unexpected format.
	}
	my $org_option = $option;
	$option = uc $option; # Capitalize the option, in advance.

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (CL)
	if ($option eq 'MIN_WD_SIZE') {
	    $MIN_WDSIZE = $value;
	} elsif ($option eq 'MAX_WD_SIZE') {
	    $MAX_WDSIZE = $value;
	} elsif ($option eq 'THRSH_PV') {
	    $THRSH_PV = $value;
	} elsif ($option eq 'IF_MODPV') {
	    $IF_MODPV = $value;
	} elsif ($option eq 'THRSH_PV0') {
	    $THRSH_PV0 = $value;

	} elsif ($option eq 'BASE_FREQS') {
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (CL)"

	if ($option eq 'BASE_FREQS') { # MODIFIED in creation of 'srd_map.ver0.6.pl'. (on 2020/03/04). (CL)
	    my @new_root_resfreqs = split /\,/, $value;
	    if (@new_root_resfreqs == 4) {
		@root_resfreqs = @new_root_resfreqs;
	    } else {
		print STDOUT "[CAUTION '$plscript_name'] An UNEXPECTED NUMBER of residue frequencies are specified by the argument, ", $ARGV[$k], ": $! !\n";
		print STDOUT "                         ... So, Use the DEFAULT residue frquencies, ", '@root_resfreqs =(', join (', ', @root_resfreqs), ") ...\n\n";
	    }

	} elsif ($option eq 'TN_PARAMS') {
	    my ($alpha1, $alpha2, $beta) = split /\,/, $value;    
	    if ((defined $beta) and ($beta>0)) {
		$SV_RATIO_R = $alpha1/$beta;
		$SV_RATIO_Y = $alpha2/$beta;
	    } else {
		print STDOUT "[CAUTION '$plscript_name'] The Tamura-Nei parameter, Beta, is ZERO or UNDEFINED  within the argument, ", $ARGV[$k], ": $! !\n";
		print STDOUT "                          ... So, USE the DEFAULT T-N ratios, (\$SV_RATIO_R, \$SV_RATIO_Y) = ($SV_RATIO_R, $SV_RATIO_Y) ...\n\n\n";
	    }

		# (ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (CL) #
	} elsif ($option eq 'DIS_LEN_SRD') {
		$DIS_LEN_SRD = $value;

	} elsif  ($option eq 'CT_EFF_DGTS') {
		$CT_EFF_DGTS = $value;

		# End of "(ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (CL)" #

	} else {
	    print STDOUT "[CAUTION '$plscript_name'] Found an UNEXPECTED OPTION, '$org_option', within the cmmmand-line argument, ", $ARGV[$k], ": $! ! ... I will SKIP it ...\n\n";
	    next; # Skip an argument with an unexpected option.
	}

}	# End of the preliminary for-loop (over additional command-line arguments. #

	### End of "Check the Command-Line Options." ###

	##### End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PL 1)" ####


	### Input files. ... Already Given as Command-Line Arguments. ###

## ($inaln_file, $intree_file, $out_dir) = @ARGV[0 .. 2];

	##### ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PL 2) ####
	##### => Modified in creation of 'srd_map.ver0.6.pl' on 2020/03/04.) (PL 2) ####

	### Output files. ###


	## Main outputs. ##

my $name_out_br_ids = 'branch_identifiers.txt'; # Added on 2020/03/07. (OF-1) #
my $name_out_exp = 'expected_SRD_Map.txt';
my $name_out_obs = 'observed_SRD_Map.txt';

my $out_br_ids = "$out_dir/$name_out_br_ids"; # Added on 2020/03/07. (OF-2) #
my $out_exp = "$out_dir/$name_out_exp";
my $out_obs = "$out_dir/$name_out_obs";

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (OF)
	## Main output. ##
my $name_out_main = 'list_purge_candidates.txt';
my $out_main = "$out_dir/$name_out_main";
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (OF)"

	##### End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PL 2) ####
	##### => Modified in creation of 'srd_map.ver0.6.pl' on 2020/03/04.) (PL 2)" ####

	### Check the input files. ###

foreach my $infile ($inaln_file, $intree_file) {
   unless (( -f $infile ) and ( -r $infile )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file '$infile' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
   }
}
	##### ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PL 3) ####

	### Check the output directory. ###

foreach my $dir ($out_dir) { # Preliminary foreach-loop (over output directories). #
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
    unless ( -w $dir ) {
	unless (0 == system (" chmod -R u+w $dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
} 	# End of the preliminary foreach-loop (over output directories). #

	##### END of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PL 3)" ####


	### Declare the beginning of the program. ###

print STDOUT "\n<<<< Program name: $plscript_name (path: $0) >>>>\n\n";

	## ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PR 1) ##
print STDOUT "<< Input/Output Files >>\n\n";

print STDOUT "Input Alignment: $inaln_file ,\n";
print STDOUT "Input Tree:      $intree_file .\n\n";

print STDOUT "Output Directory: {Out_dir}=\"$out_dir\" .\n";

##my $name_out_exp = 'expected_SRD_Map.txt';
##my $name_out_obs = 'observed_SRD_Map.txt';
##my $out_exp = "$out_dir/$name_out_exp";
##my $out_obs = "$out_dir/$name_out_obs";

		## (Modified in creation of 'srd_map.ver0.6.pl' (on 2020/03/04).) (PR 1) ##
print STDOUT "Main Output Files:\n";
print STDOUT "  (for branch identifiers) {Out_dir}/$name_out_br_ids ,\n"; # Added in creation of 'srd_map.ver0.6.pl' (on 2020/03/07). (PR 1+) #
print STDOUT "  (for Expected SRD Map) {Our_dir}/$name_out_exp ,\n";
print STDOUT "  (for Observed SRD Map) {Our_dir}/$name_out_obs .\n";
##print STDOUT "Main Output File: {Our_dir}/$name_out_main .\n";
		## End of "(Modified in creation of 'srd_map.ver0.6.pl' (on 2020/03/04).) (PR 1)" ##

print STDOUT "\n\n";

	## End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PR 1)" ##


print STDOUT "<< Global Parameters >>\n\n";

print STDOUT '$DIS_LEN = ', $DIS_LEN, " ,\n"; # Added in revision (on 2019/06/18). #
print STDOUT '$MyANEX_IO::DIS_LEN = ', $MyANEX_IO::DIS_LEN, " .\n\n"; # Added in REVISION (on 2019/07/07). #

print STDOUT '$GAP = ', $GAP, " .\n\n";
print STDOUT '$CHAR_ABS  = ', $CHAR_ABS, " ,\n";
print STDOUT '$CHAR_PRES = ', $CHAR_PRES, " .\n\n";

print STDOUT '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT '$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca = ', $MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cpdca, " ,\n";
print STDOUT '$MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa  = ', $MyLikelihoods_respattern_msa_dna::TEST_LEVEL_cjpaaa, " .\n\n";
print STDOUT '$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE  = ', $MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_PRE , " ,\n";
print STDOUT '$MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_1ST = ', $MyTreeMap_indels_spt_odr_finer::TEST_LEVEL_1ST, " .\n\n";

	## ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PR 2) ##
print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs = ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_br_psm_did_hs, " ,\n";
print STDOUT '$MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs = ', $MyTreeMap_indels_spt_odr2::TEST_LEVEL_bch_br_psm_did_hs, " .\n\n";
	## End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PR 2)" ##

print STDOUT '$MyDetect_purge_cands::TEST_LEVEL = ', $MyDetect_purge_cands::TEST_LEVEL, " .\n\n";


print STDOUT "\n<<< Input Variables >>>\n\n";

		# (ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (PIV) #
print STDOUT '$DIS_LEN_SRD = ', $DIS_LEN_SRD, " .\n\n";
print STDOUT '$CT_EFF_DGTS = ', $CT_EFF_DGTS, " .\n\n\n";
		# End of "(ADDED in creation of 'srd_map.ver0.6.pl' on 2020/03/06.) (PIV)" #

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (PIV) 

print STDOUT '$MIN_WDSIZE = ', $MIN_WDSIZE, " ,\n";
print STDOUT '$MAX_WDSIZE = ', $MAX_WDSIZE, " .\n\n";

print STDOUT '$THRSH_PV = ', $THRSH_PV, " ,\n";
print STDOUT '$IF_MODPV = ', "$IF_MODPV .\n\n";

print STDOUT '$THRSH_PV0 = ', $THRSH_PV0, " .\n\n\n";


my @out_sbj_rgs = ();
foreach my $rg (@SBJ_REGIONS) { push @out_sbj_rgs, sprintf ("[%d, %d]", @{$rg}); }

		## MODIFIED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PR 1) ##
if (@SBJ_REGIONS>0) {
	print STDOUT "Subject regions:   ", join ('; ', @out_sbj_rgs), " .\n\n\n";
} else {
	print STDOUT "Subject regions:   THE WHOLE ALIGNMENT .\n\n\n";
}
##print STDOUT "Subject regions:   ", join ('; ', @out_sbj_rgs), " .\n\n\n";
		## End of "MODIFIED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (PR 1)" ##
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (PIV)"



print STDOUT "Substitution Model = Tamura-Nei with \n\n";
print STDOUT '     ($SV_RATIO_R, $SV_RATIO_Y) = ', "($SV_RATIO_R, $SV_RATIO_Y) ,\n\n";
print STDOUT ' and Freq[',  join (', ', @RESIDUES), "]_root = (", join (', ', @root_resfreqs), ") .\n\n\n";



	### Read in the input MSA. ###

print STDOUT "\n<<< Reading in the input MSA >>>\n\n";

(my $in_msa, my $seq_ids, undef) = extract_alignment ($inaln_file); 

unless (defined $in_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input MSA from '$inaln_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_seqs = @{$in_msa};

unless (@{$seq_ids} == $ct_seqs) {
	print STDOUT "[ERROR '$program_name'] The number of sequence IDs (=", scalar (@{$seq_ids}), ") CLASHES with the number of sequences in the input MSA !\n"; 
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $lenaln = length ${$in_msa->[0]};

my $seqnames = $seq_ids; # Added to suppress error messages (on 2019/06/18). #

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (IN 1)
	### ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (IN 1) ###
unless (@SBJ_REGIONS>0) {
	push @SBJ_REGIONS, [0, $lenaln -1];
}
	### End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/16). (IN 1)" ###
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (IN 1)"

print STDOUT "\n<< Input Alignment ( Size = $ct_seqs (#{sequences}) x $lenaln (#{sites}) ) >>\n\n";

if ($TEST_LEVEL> 3) { # Modified in creation of 'detect_purge_cands.ver0.5.pl'. #
##if ($TEST_LEVEL>0) {
   unless ( pr_ODEN_input (@{$seq_ids}, @{$in_msa}, $DIS_LEN, *STDOUT ) ) { # Modified in revision (on 2019/06/18). #
   ##unless ( pr_ODEN_input (@{$seq_ids}, @{$in_msa}, 50, *STDOUT ) ) {
	print STDOUT "[ERROR '$plscript_name'] 'pr_ODEN_input' FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n";
	die ;
   }
   print STDOUT "\n\n";
} # End of "if ($TEST_LEVEL> 3) {...}"


	### Read in the input tree. ###

print STDOUT "\n<<< Reading in the input tree >>>\n\n";

my $info_dawg_tree = readin_dawg_tree2 ($intree_file);
unless (defined $info_dawg_tree) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'readin_dawg_tree2' FAILED on the input file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'readin_dawg_tree2', Ended ...\n\n") if ($TEST_LEVEL>0);

my ($node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$info_dawg_tree};
my $topnode_id = $name2id->{$TOP_NODE};

if ($TEST_LEVEL> 3) { # Modified in creation of 'detect_purge_cands.ver0.5.pl'. #
#if ($TEST_LEVEL>1) {

   print STDOUT "\n<< Phylogenetic Tree of the Aligned Sequences (Top node: $topnode_id) >>\n\n";

   unless ( output_tree ($topnode_id, %{$node2ch}, %{$id2name}, %{$br_attr}, *STDOUT, 1, 1) ) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'output_tree' FAILED to print out the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
   }

} # End of "if ($TEST_LEVEL> 3) {...}"


if ($TEST_LEVEL>3) {
   print STDOUT "\n\n< Relationships between Nodes >\n\n";
   print STDOUT join ("\t", 'Node_ID', 'parent', '[child,node,IDs]'), "\n\n";

   foreach my $node (sort {$a <=> $b} keys %{$node2pa}, $topnode_id) { # 1st pre-preliminary foreach-loop (over nodes). #
	my $pa = $node2pa->{$node};
	(defined $pa) or ($pa = 'undef');
	my $children = $node2ch->{$node};
	my $cnct_children = (defined $children) ? join (',', @{$children}) : '';
	print STDOUT join ("\t", $node, $pa, "[$cnct_children]"), "\n";
   }						# End of the 1st pre-preliminary foreach-loop (over nodes). #
   print STDOUT "\n\n";
} # End of "if ($TEST_LEVEL>3) {...}"



		## IMPORTED with modification from (Pre 1.5) of 'anex.ver0.5.pl', in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). ##
print STDOUT "\n<<< (Preliminary 0.5:) Copy the input MSA (i.e., ", '@{$in_msa}', ") into an array, ", '@cp_in_msa', " . >>>\n\n\n";

my @cp_in_msa = ();
foreach my $seq (@{$in_msa}) {
	my $cp_seq = ${$seq};
	push @cp_in_msa, \$cp_seq;
}
(print STDOUT "\n", ' ... Finished Copying the MSA ...', "\n\n") if ($TEST_LEVEL>0);

		## End of "IMPORTED with modification from (Pre 1.5) of 'anex.ver0.5.pl', in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17)." ##


print STDOUT "\n<<< (Preliminary 1:) Assign a substitution transition matrix to each branch >>>\n\n\n";

my $br2tr_mtrx = calc_br2tr_mtrx_TN (%{$br_attr}, @root_resfreqs, $SV_RATIO_R, $SV_RATIO_Y, 1);
unless (defined $br2tr_mtrx) {
	print STDOUT "[ERROR '$program_name'] 'calc_br2tr_mtrx_TN' FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_br2tr_mtrx_TN', Ended ...\n\n") if ($TEST_LEVEL>0);

if ($TEST_LEVEL>2) {
   print STDOUT "\n<< Preliminary output (1): Substitution Transition Matrices Assigned to the Branches >>\n\n";

   foreach my $br_id (sort {$a <=> $b} keys %{$br2tr_mtrx}) {	# 1st preliminary foreach-loop (over branches). #

	my $tr_mtrx = $br2tr_mtrx->{$br_id};

	print STDOUT ' ' x 3, "( Branch ID: $br_id )\n\n";
	print STDOUT ' ' x 3, join ("\t", 'From\To', @RESIDUES), "\n\n";
	for (my $r=0; $r<$CT_NUCLS; $r++) { print STDOUT ' ' x 3, join ("\t", $RESIDUES[$r], @{$tr_mtrx->[$r]}), "\n"; }
	print STDOUT "\n\n";
   }						# End of the 1st preliminary foreach-loop (over branches). #
}


print STDOUT "\n<<< (Preliminary 2:) Calculate the residue frequences at ALL nodes. >>>\n\n\n";

my $node2ntfreqs = calc_node2ntfreqs_td (@root_resfreqs, %{$br2tr_mtrx}, %{$node2ch}, $topnode_id);

unless (defined $node2ntfreqs) {
	print STDOUT "[ERROR '$plscript_name'] The Preliminary Subroutine 'calc_node2ntfreqs_td' FAILED !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_node2ntfreqs_td', Ended ...\n\n") if ($TEST_LEVEL>0);

if ($TEST_LEVEL>2) {
   print STDOUT "\n<< Contents of the Preliminary Output (2), ", '%{$node2ntfreqs}', " >>\n\n";

   print STDOUT join ("\t", 'Node_ID', @labels_resfreqs), "\n\n";

   foreach my $node_id (sort {$a <=> $b} keys %{$node2ntfreqs}) { # 2nd preliminary outer foreach-loop (over nodes). #

	my $ntfreqs = $node2ntfreqs->{$node_id};

	print STDOUT join ("\t", $node_id, @{$ntfreqs}), "\n\n";


   }			# End of the 2nd preliminary outer foreach-loop (over nodes). #
   print STDOUT "\n";
}


print STDOUT "\n<<< (Preliminary 3:) Calculate probabilities of the residue states of extant sequences on the descendant side of each branch, conditioned on the residue state at either end of the branch. >>>\n\n\n";

my ($node2set_sw_cond_probs, $node2set_sw_extcond_probs)
 = calc_prob_descresstats_cond_ancstat2 (@{$in_msa}, @{$seq_ids}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $GAP);

unless (defined $node2set_sw_cond_probs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_prob_descresstats_cond_ancstat2', Ended ...\n\n") if ($TEST_LEVEL>0);

if ($TEST_LEVEL>3) {
   print STDOUT "\n<< Contents of the Preliminary Output (3-1): ", '%{$node2set_sw_cond_probs}', " >>\n\n";

   foreach my $node_id (sort {$a <=> $b} keys %{$node2set_sw_cond_probs}) { # 3rd-1st preliminary outer foreach-loop (over nodes). #

	my $set_sw_cond_probs = $node2set_sw_cond_probs->{$node_id};
	my $ct_sites = @{$set_sw_cond_probs};

	print STDOUT ' ' x 2, "{ Node ID: $node_id }\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Site\Cond', @NUCLEOTIDES), "\n\n";

		# A BUG fixed on Jan 13 (Sat), 2018. #
	for (my $s = 0; $s < $ct_sites; $s++) { print STDOUT ' ' x 2, join ("\t", $s, @{$set_sw_cond_probs->[$s]}), "\n"; }
	##for (my $s = 0; $s < $ct_sites; $s++) { print STDOUT ' ' x 2, join ("\t", $s, @{$set_sw_cond_probs}), "\n"; }

	print STDOUT "\n\n";
   }					# End of the 3rd-1st preliminary outer foreach-loop (over nodes). #

   print STDOUT "\n<< Contents of the Preliminary Output (3-1): ", '%{$node2set_sw_extcond_probs}', " >>\n\n";

   foreach my $node_id (sort {$a <=> $b} keys %{$node2set_sw_extcond_probs}) { # 3rd-2nd preliminary outer foreach-loop (over nodes). #

	my $set_sw_extcond_probs = $node2set_sw_extcond_probs->{$node_id};
	my $ct_sites = @{$set_sw_extcond_probs};

	print STDOUT ' ' x 2, "{ Node ID: $node_id }\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Site\Cond', @NUCLEOTIDES), "\n\n";

		# A BUG fixed on Jan 13 (Sat), 2018. #
	for (my $s = 0; $s < $ct_sites; $s++) { print STDOUT ' ' x 2, join ("\t", $s, @{$set_sw_extcond_probs->[$s]}), "\n"; }
	##for (my $s = 0; $s < $ct_sites; $s++) { print STDOUT ' ' x 2, join ("\t", $s, @{$set_sw_extcond_probs}), "\n"; }

	print STDOUT "\n\n";
   }					# End of the 3rd-2nd preliminary outer foreach-loop (over nodes). #

}


print STDOUT "\n<<< (Preliminary 5:) Calculating the joint probabilities of the residue state of the upper-end of each branch and the residue states of extant sequences on the ancestor side of the branch. >>>\n\n\n"; 

my ($br2set_sw_jnt_probs, $br2set_sw_extjnt_probs)
    = calc_joint_prob_ancresstats_and_ancstat2 (@{$in_msa}, @{$seq_ids}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $GAP, @root_resfreqs, %{$node2set_sw_extcond_probs}) ;

unless (defined $br2set_sw_jnt_probs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_prob_descresstats_cond_ancstats2' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_joint_prob_ancresstats_and_ancstat2', Ended ...\n\n") if ($TEST_LEVEL>0);


if ($TEST_LEVEL>3) {
   print STDOUT "\n<< Contents of the Preliminary Output (5-1): ", '%{$br2set_sw_jnt_probs}', " >>\n\n";

   foreach my $br_id (sort {$a <=> $b} keys %{$br2set_sw_jnt_probs}) { # 5th-1st preliminary outer foreach-loop (over branches). #

	my $set_sw_jnt_probs = $br2set_sw_jnt_probs->{$br_id};
	my $ct_sites = @{$set_sw_jnt_probs};

	print STDOUT ' ' x 2, "{ Branch ID: $br_id }\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Site\Jnt', @NUCLEOTIDES), "\n\n";

	for (my $s = 0; $s < $ct_sites; $s++) { print STDOUT ' ' x 2, join ("\t", $s, @{$set_sw_jnt_probs->[$s]}), "\n"; }

	print STDOUT "\n\n";
   }					# End of the 5th-1st preliminary outer foreach-loop (over branches). #


   print STDOUT "\n<< Contents of the Preliminary Output (5-2): ", '%{$br2set_sw_extjnt_probs}', " >>\n\n";

   foreach my $br_id (sort {$a <=> $b} keys %{$br2set_sw_extjnt_probs}) { # 5th-2nd preliminary outer foreach-loop (over branches). #

	my $set_sw_extjnt_probs = $br2set_sw_extjnt_probs->{$br_id};
	my $ct_sites = @{$set_sw_extjnt_probs};

	print STDOUT ' ' x 2, "{ Branch ID: $br_id }\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Site\Jnt', @NUCLEOTIDES), "\n\n";

	for (my $s = 0; $s < $ct_sites; $s++) { print STDOUT ' ' x 2, join ("\t", $s, @{$set_sw_extjnt_probs->[$s]}), "\n"; }

	print STDOUT "\n\n";
   }					# End of the 5th-2nd preliminary outer foreach-loop (over branches). #
}


print STDOUT "\n<<< (Preliminary 6:) Calculating the probability of each column (under the given substitution model) >>>\n\n\n";

my $set_clm_probs = calc_clm_probs (@root_resfreqs, %{$node2set_sw_cond_probs}, $topnode_id);
unless (defined $set_clm_probs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'calc_clm_probs' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'calc_clm_probs', Ended ...\n\n") if ($TEST_LEVEL>0);

my $ct_clms = @{$set_clm_probs};

if ($TEST_LEVEL>3) {
   print STDOUT "\n<< Contents of the Preliminary output (6): ", '@{$set_clm_probs}', " (#{columns} = $ct_clms) >>\n\n";

   print STDOUT join ("\t", 'Column', 'Probability'), "\n\n";

   for (my $c = 0; $c < $ct_clms; $c++) { print STDOUT join ("\t", $c, $set_clm_probs->[$c]), "\n"; }
   print STDOUT "\n\n";
}


print STDOUT "\n<<< (Pre-Preliminary 7:) Calculating some ingredients of the Dollo-parsimonious indel history  >>>\n\n\n";

my ($node2depth, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes)
    = enum_psm_indel_hsts_for_msa_finer_PRE ($topnode_id, %{$node2ch}, %{$node2pa});

unless (defined $br2nn) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'enum_psm_indel_hsts_for_msa_finer_PR', Ended ...\n\n") if ($TEST_LEVEL>0);

if ($TEST_LEVEL > 5) {
    print STDOUT "\n<< Contents of the Pre-Preliminary output (7) >>\n\n";
    print STDOUT '%{$br2eq} = (', "\n";
    foreach my $br (sort {$a <=> $b} keys %{$br2eq}) { print STDOUT "\t", $br, ' => ', $br2eq->{$br}, ",\n"; }
    print STDOUT "\n\n";
    print STDOUT '%{$br2nn} = (', "\n";
    foreach my $br (sort {$a <=> $b} keys %{$br2nn}) { print STDOUT "\t", $br, ' => [', join (", ", @{$br2nn->{$br}}), "],\n"; }
    print STDOUT "\n\n\n";
}

print STDOUT "\n<<< (Preliminary 7:) Calculating the Dollo-parsimonious indel history that results in the input MSA >>>\n\n\n";

my ($set_gpatterns, $set_node2dparsstat, $set_br2change, $info_psm_indel_hsts0)
  = enum_psm_indel_hsts_for_msa_finer_1ST_HF (%{$node2ch}, %{$node2pa}, %{$name2id}, @{$seq_ids}, @{$in_msa}, $CHAR_ABS, $CHAR_PRES, %{$node2depth}, %{$node2ct_exoffs});

unless ((defined $info_psm_indel_hsts0) and (@{$info_psm_indel_hsts0}>2) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' FAILED on the input tree !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'enum_psm_indel_hsts_for_msa_finer_1ST_HF', Ended ...\n\n") if ($TEST_LEVEL>0);

my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};
my $ct_cls = @{$set_cls_gpblocks};
my $ct_gpd_segments = $ct_cls; # Added to suppress error messages (on 2019/06/18). #

if ($TEST_LEVEL>2) {
    print STDOUT "\n<< Main Contents of the Preliminary output (7) ", '@{$set_cls_gpblocks} & @{$psm_hsts0_for_cls}', " >>\n\n";

    for (my $cl = 0; $cl < $ct_cls; $cl++) { # The 7 th preliminary  for-loop (over gapped segments). #
	my $cls_gpblks = $set_cls_gpblocks->[$cl];
	my $ct_gpblks = @{$cls_gpblks};

	print STDOUT ' ' x 2, "{ The $cl th gapped segment (#{gap-pattern blocks} = $ct_gpblks) }\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal)'"), "\n\n";
	for (my $bl=0; $bl < $ct_gpblks; $bl++) { # 1st inner for-loop (over blocks). #
	    my ($indx_gp, $len, $beg, $end) = @{$cls_gpblks->[$bl]};
	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print STDOUT ' ' x 2, join ("\t", $bl, $len, $beg, $end, $indx_gp, "'$gpattern'"), "\n";
	}				# End of the 1st inner for-loop (over blocks). #
	print STDOUT "\n\n";

	my ($br2indels0, $node2bw_gstats0) = @{$psm_hsts0_for_cls->[$cl]};

	print STDOUT ' ' x 2, "( Map of INITIAL block-wise gap states onto the nodes )\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Node_ID', "'block_wise_gap_states0(concatenated)'"), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats0}) { # 2nd inner foreach-loop (over nodes). #
	    my $cnct_bw_gstats0 = join ('', @{$node2bw_gstats0->{$node}});
	    print STDOUT ' ' x 2, join ("\t", $node, "'$cnct_bw_gstats0'"), "\n";
	}                           # End of the 2nd inner foreach-loop (over nodes). #
	print STDOUT "\n\n";
	
    }					# End of the 7th preliminary for-loop (over gapped segments). #

} # End of "if ($TEST_LEVEL>2) {...}"


=DISABLED in MAJOR REVISION of 'detect_purge_cands.ver0.5.pl' (on 2020/03/04). (Pre 8)

		## IMPORTED with modification from 'anex.ver0.5.pl', in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). ##
print STDOUT "\n<<< (Preliminary 8:) Briefly Attempt to Parsimonize the Dollo parsimony-based indel history for each gapped segment, via 'bch_br_parsimonize_dollo_indel_histories' >>>\n\n\n";

	# Prepare an additional input. #
my $info_init_psm_hst = $info_psm_indel_hsts0;

my @set_exnodes = ();
foreach my $node (@{$set_all_nodes}) { 
	unless (defined $node2ch->{$node}) { push @set_exnodes, $node; } 
}
(print STDOUT 'An Additional Input: @set_exnodes = (', join (', ', @set_exnodes), ") .\n\n") if ($TEST_LEVEL>2);
	# End of "Prepare an additional input." #

my ($new_info_init_psm_hst, $new_in_msa, $set_new_lists_gblocks, $sets_places_gpbs, $sets_gpbs_at_spec_places)
   = bch_br_parsimonize_dollo_indel_histories (@cp_in_msa, @{$info_init_psm_hst}, @set_exnodes, $topnode_id, %{$node2pa}, %{$node2ch}, $CHAR_ABS, $CHAR_PRES, %{$br2eq});
unless (defined $new_info_init_psm_hst) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'bch_br_parsimonize_dollo_indel_histories' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'bch_br_parsimonize_dollo_indel_histories', Ended ...\n\n") if ($TEST_LEVEL >0);

my ($new_set_cls_gpbs, $new_init_totcts_indels_for_cls, $new_init_psm_hsts_for_cls) = @{$new_info_init_psm_hst} ;

unless (@{$new_set_cls_gpbs} == $ct_gpd_segments) {
	print STDOUT "[ERROR '$plscript_name'] The number of gapped segments AFTER the brief parsimonization (=", scalar (@{$new_set_cls_gpbs}), ") CLASHES that BEFORE the brief parsimonization (=$ct_gpd_segments) !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n";
	die ;
}


if ($TEST_LEVEL > 3) { # (7) #

	print STDOUT "\n ... Check the outputs of 'bch_br_parsimonize_dollo_indel_histories', by performing the subroutine, 'print_out_bch_br_parsimonize_dollo_indel_histories' ...\n\n";

	my $outdir_name_psmnzd = "Init_psm_cands_parsimonized";
	my $new_aln_filename = "new_alignment.txt";

		# REVISED on 2019/06/17. #
		# => Further REVISED on 2019/07/07. #
	unless ( print_out_bch_br_parsimonize_dollo_indel_histories ($out_dir, $outdir_name_psmnzd, $new_aln_filename, @{$seqnames}, @{$new_info_init_psm_hst}, @{$new_in_msa}, @{$set_new_lists_gblocks}, @{$sets_places_gpbs}, @{$sets_gpbs_at_spec_places}, @{$set_gpatterns})  ) {
	##unless ( print_out_bch_br_parsimonize_dollo_indel_histories ($out_dir, $outdir_name_psmnzd, $new_aln_filename, @{$seqnames}, @{$new_info_init_psm_hst}, @{$new_in_msa}, @{$set_new_lists_gblocks}, @{$sets_places_gpbs}, @{$sets_gpbs_at_spec_places})  ) {
	##unless ( print_out_bch_br_parsimonize_dollo_indel_histories ($out_interim, @{$seqnames}, @{$new_info_init_psm_hst}, @{$new_in_msa}, @{$set_new_lists_gblocks}, @{$sets_places_gpbs}, @{$sets_gpbs_at_spec_places})  ) {
		print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_out_bch_br_parsimonize_dollo_indel_histories' FAILED !\n";
		print STDERR "[ERROR '$plscript_name'] $! !\n";
		die ;
	}
	print STDOUT "\n ... The subroutine, 'print_out_bch_br_parsimonize_dollo_indel_histories', Ended ...\n\n";

} # End of "if ($TEST_LEVEL > 3) { # (7) # ...}"

		## End of "IMPORTED with modification from 'anex.ver0.5.pl', in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17)." ##

=cut # END of "DISABLED in MAJOR REVISION of 'detect_purge_cands.ver0.5.pl' (on 2020/03/04). (Pre 8)"


		## MODIFIED in creation of 'srd_map.ver0.6.pl' (on 2020/03/07). (Pre 9) ##

print STDOUT "\n<<< (Preliminary 9:) Create a hash, ", '%{$br2exoffs_msa_rows}', " , via the subroutine, 'wrapper_mk_node2exoffs_msa_rows' >>>\n\n\n";

my ($br2exoffs_msa_rows, $node2depth_W, $nodeids_in_msa, $nodeid2msa_row)
	= wrapper_mk_node2exoffs_msa_rows ($topnode_id, %{$node2ch}, @{$seqnames}, %{$name2id});
unless (defined $br2exoffs_msa_rows) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'wrapper_mk_node2exoffs_msa_rows' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'wrapper_mk_node2exoffs_msa_rows', Ended ...\n\n") if ($TEST_LEVEL >0);


##my $fh_out_pre = open_output_file ($out_br_ids, 1);
my $fh_out_pre = open_output_file ($out_br_ids, 0); # MODIFIED on 2020/03/10. #
unless (defined $fh_out_pre) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_br_ids\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'open_output_file', Ended on \"$out_br_ids\" ...\n\n") if ($TEST_LEVEL>0);


print $fh_out_pre "\n", join ("\t", 'br_ID', '[msa,rows,under,br]'), "\n\n";

my %br2cnct_exoffs_msa_rows;
foreach my $br (sort {$a <=> $b} keys %{$br2exoffs_msa_rows}) { # Preliminary foreach-loop (over branches). #
	my $cnct_exoffs_msa_rows = join (',', @{$br2exoffs_msa_rows->{$br}});
	$cnct_exoffs_msa_rows = "[$cnct_exoffs_msa_rows]";
	$br2cnct_exoffs_msa_rows{$br} = $cnct_exoffs_msa_rows;

	print $fh_out_pre join ("\t", $br, $cnct_exoffs_msa_rows), "\n";

} # End of the preliminary foreach-loop (over branches). #

print $fh_out_pre "\n\n";

close ($fh_out_pre);

	### RESTART FROM HERE ###

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (Pre 9)
		## ADDED in revision of 'detect_purge_cands.ver0.5.pl' (on 2019/06/19). (Pre 9) ##

print STDOUT "\n<<< (Preliminary 9:) Create a hash, ", '%{$br2exoffs_msa_rows}', " , via the subroutine, 'wrapper_mk_node2exoffs_msa_rows' >>>\n\n\n";

my ($br2exoffs_msa_rows, $node2depth_W, $nodeids_in_msa, $nodeid2msa_row)
	= wrapper_mk_node2exoffs_msa_rows ($topnode_id, %{$node2ch}, @{$seqnames}, %{$name2id});
unless (defined $br2exoffs_msa_rows) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'wrapper_mk_node2exoffs_msa_rows' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'wrapper_mk_node2exoffs_msa_rows', Ended ...\n\n") if ($TEST_LEVEL >0);

(print STDOUT "\n", join ("\t", 'br_ID', '[msa,rows,under,br]'), "\n\n") if ($TEST_LEVEL>2);

my %br2cnct_exoffs_msa_rows;
foreach my $br (sort {$a <=> $b} keys %{$br2exoffs_msa_rows}) { # Preliminary foreach-loop (over branches). #
	my $cnct_exoffs_msa_rows = join (',', @{$br2exoffs_msa_rows->{$br}});
	$cnct_exoffs_msa_rows = "[$cnct_exoffs_msa_rows]";
	$br2cnct_exoffs_msa_rows{$br} = $cnct_exoffs_msa_rows;
    (print STDOUT join ("\t", $br, $cnct_exoffs_msa_rows), "\n") if ($TEST_LEVEL>2);
} # End of the preliminary foreach-loop (over branches). #

(print STDOUT "\n\n") if ($TEST_LEVEL>2);

		## End of "ADDED in revision of 'detect_purge_cands.ver0.5.pl' (on 2019/06/19). (Pre 9)" ##
=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (Pre 9)"

		## END of "MODIFIED in creation of 'srd_map.ver0.6.pl' (on 2020/03/07). (Pre 9)" ##


		### ADDED in creation of 'srd_map.ver0.6.pl' (on 2020/03/04-). (MAIN) ###

print STDOUT "\n\n<<<< MAIN PROCESS 1:  Creating & Outputting the Observed Substitutional Residue-Difference Map of the input MSA. >>>>\n\n\n";

my @br_ids_for_display = ();
for (my $r=0; $r < $ct_seqs; $r++) {
	my $seqname = $seqnames->[$r];
	my $brid = $name2id->{$seqname};
	$br_ids_for_display[$r] = $brid;
}
foreach my $br (sort {$a <=> $b} keys %{$br_attr}) {
	unless (defined $id2name->{$br}) { push @br_ids_for_display, $br; }
}

our $IF_OBSERVED = 1;

my ($beg_msa, $end_msa) = (0, $lenaln -1);

my ($srd_map_obs, $br2sw_srds_obs)
  =  mk_srd_map ($IF_OBSERVED, @br_ids_for_display, $CHAR_ABS, $CHAR_PRES,
	$topnode_id, %{$node2pa}, %{$node2ntfreqs}, %{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs},
	@{$info_psm_indel_hsts0}, $beg_msa, $end_msa, %{$br_attr}, %{$br2eq});
unless (defined $srd_map_obs) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_srd_map' FAILED to create an OBSERVED SRD Map from the input MSA: $? !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'mk_srd_map', (w/ \$IF_OBSERVED = 1) ENDED ...\n\n") if ($TEST_LEVEL >0);

##my $fh_out_obs = open_output_file ($out_obs, 1);
my $fh_out_obs = open_output_file ($out_obs, 0); # MODIFIED on 2020/03/10. #
unless (defined $fh_out_obs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_obs\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'open_output_file', Ended on \"$out_obs\" ...\n\n") if ($TEST_LEVEL>0);


unless ( print_srd_map (*${$fh_out_obs}, @br_ids_for_display, %{$id2name}, @{$srd_map_obs}, $beg_msa, $end_msa, $CHAR_ABS, $DIS_LEN_SRD, $CT_EFF_DGTS) ) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_srd_map' FAILED to print out the OBSERVED SRD Map: $? !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'print_srd_map', ENDED to print out the OBSERVED SRD Map ...\n\n") if ($TEST_LEVEL >0);

close ($fh_out_obs);


print STDOUT "\n\n<<<< MAIN PROCESS 2:  Creating & Outputting the Expected Substitutional Residue-Difference Map of the input MSA. >>>>\n\n\n";

$IF_OBSERVED = 0;

my ($srd_map_exp, $br2sw_srds_exp)
  =  mk_srd_map ($IF_OBSERVED, @br_ids_for_display, $CHAR_ABS, $CHAR_PRES,
	$topnode_id, %{$node2pa}, %{$node2ntfreqs}, %{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs},
	@{$info_psm_indel_hsts0}, $beg_msa, $end_msa, %{$br_attr}, %{$br2eq});
unless (defined $srd_map_exp) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'mk_srd_map' FAILED to create an EXPECTED SRD Map from the input MSA: $? !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'mk_srd_map', (w/ \$IF_OBSERVED = 0) ENDED ...\n\n") if ($TEST_LEVEL >0);

##my $fh_out_exp = open_output_file ($out_exp, 1);
my $fh_out_exp = open_output_file ($out_exp, 0); # MODIFIED on 2020/03/10.
unless (defined $fh_out_exp) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_exp\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'open_output_file', Ended on \"$out_exp\" ...\n\n") if ($TEST_LEVEL>0);

unless ( print_srd_map (*${$fh_out_exp}, @br_ids_for_display, %{$id2name}, @{$srd_map_exp}, $beg_msa, $end_msa, $CHAR_ABS, $DIS_LEN_SRD, $CT_EFF_DGTS) ) {
	print STDOUT "[ERROR '$plscript_name'] The subroutine 'print_srd_map' FAILED to print out the EXPECTED SRD Map: $? !\n";
	print STDERR "[ERROR '$plscript_name'] $!\n\n";
	die ;
}
(print STDOUT "\n ... The subroutine, 'print_srd_map', ENDED to print out the EXPECTED SRD Map ...\n\n") if ($TEST_LEVEL >0);

close ($fh_out_obs);


(print STDOUT "\n\n === The MAIN Processes ENDED!! ===\n\n\n") if ($TEST_LEVEL>0);

print STDOUT "\n\n ===== ALL the PROCESSES ENDED !!! =====\n\n\n";

		### END of "ADDED in creation of 'srd_map.ver0.6.pl' (on 2020/03/04-). (MAIN)" ###

=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (MAIN)

print STDOUT "\n\n<<<< MAIN PROCESS:  Detecting 'purge candidates' in specified regions of the input MSA. >>>>\n\n\n";

my $if_test_output = (($TEST_LEVEL> 2) and ($MyDetect_purge_cands::TEST_LEVEL <= 3)) ? 1 : 0; # ADDED in revision of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #

my @list_mgd_purge_candidates = ();
	# [ element = ['beg_rg', 'end_rg', br_ID, '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'] ] ... OBSOLETE as of 2019/06/19.
	# element = ['beg_rg', 'end_rg', br_ID, '[msa,rows,under,br]', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'] # Modified in revision (on 2019/06/19).

my $maxsize_pcand = 0; # Added in revision (on 2019/06/18).

foreach my $rg (@SBJ_REGIONS) { # Main outer foreach-loop (over subject regions). #

	my ($beg_rg, $end_rg) = @{$rg};

	print STDOUT "<<< Region: [$beg_rg, $end_rg] >>>\n\n";


			## MODIFIED in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). (MAIN) ##

	my ($br2purge_cands2, $br2purge_cands, $br2sgn_wds_in_pcands)  # => Further MODIFIED (on 2019/06/16). #
	##my ($br2purge_cands2, $br2purge_cands) 
	    = detect_purge_cands2 ($MIN_WDSIZE, $MAX_WDSIZE, $THRSH_PV, $THRSH_PV0, $IF_MODPV, $CHAR_ABS, $CHAR_PRES,
	$topnode_id, %{$node2pa}, %{$node2ntfreqs}, %{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, 

		# MODIFIED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). (MAIN) #
		# => UNDID the modification in MAJOR REVISION of 'detect_purge_cands.ver0.5.pl' (on 2020/03/04). (MAIN) #
	##@{$set_clm_probs}, @{$new_info_init_psm_hst}, $beg_rg, $end_rg, %{$br_attr}, %{$br2eq});
	@{$set_clm_probs}, @{$info_psm_indel_hsts0}, $beg_rg, $end_rg, %{$br_attr}, %{$br2eq});
		# End of "MODIFIED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). (MAIN) #
		# => UNDID the modification in MAJOR REVISION of 'detect_purge_cands.ver0.5.pl' (on 2020/03/04). (MAIN)" #

	unless (defined $br2purge_cands2) {
	    print STDOUT "[ERROR '$program_name'] The MAIN subroutine 'detect_purge_cands2' FAILED on the region [$beg_rg, $end_rg] !!\n";
	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	    die ;
	}
    (print STDOUT "\n ... Subroutine, 'detect_purge_cands2', Ended ...\n\n") if ($TEST_LEVEL>0);

##=For the predecessor, 'test_detect_purge_cands.pl'. (MAIN)
##	##my $br2purge_cands = detect_purge_cands ($MIN_WDSIZE, $MAX_WDSIZE, $THRSH_PV, $IF_MODPV, $CHAR_ABS, $CHAR_PRES, # MODIFIED on Jan 13 (Sat), 2018. #
##	my $br2purge_cands = detect_purge_cands ($MIN_WDSIZE, $MAX_WDSIZE, $THRSH_PV, $THRSH_PV0, $IF_MODPV, $CHAR_ABS, $CHAR_PRES,
##	$topnode_id, %{$node2pa}, %{$node2ntfreqs}, %{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, 
##	@{$set_clm_probs}, @{$info_psm_indel_hsts0}, $beg_rg, $end_rg, %{$br_attr}, %{$br2eq});
##
##	unless (defined $br2purge_cands) {
##	    print STDOUT "[ERROR '$program_name'] The MAIN subroutine 'detect_purge_cands' FAILED on the region [$beg_rg, $end_rg] !!\n";
##	    print STDERR "[ERROR '$plscript_name'] $! !\n\n";
##	    die ;
##	}
##=cut # End of "For the predecessor, 'test_detect_purge_cands.pl'. (MAIN)"
			## END of "MODIFIED in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). (MAIN)" ##

   ##if (($TEST_LEVEL>0) and ($MyDetect_purge_cands::TEST_LEVEL <= 3)) { # DISABLED (on 2019/06/17).

    if ($if_test_output) { # (1) #
	print STDOUT "\n[  Contents of the MAIN output, ", '%{$br2purge_cands2} & %{$br2purge_cands} & %{$br2sgn_wds_in_pcands}', " ]\n\n"; # => Further Modified in revision (on 2019/06/16). (Out1m)
	##print STDOUT "[  Content of the MAIN output, ", '%{$br2purge_cands2} & %{$br2purge_cands}', " ]\n\n"; # Modified in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). (Out1)
	####print STDOUT "[  Content of the MAIN output, ", '%{$br2purge_cands}', " ]\n\n";
    } # End of "if ($if_test_output) { # (1) # ...}"

	foreach my $br (sort {$a <=> $b} keys %{$br2purge_cands}) { # Middle foreach-loop (over branches). #
	    my $purge_cands = $br2purge_cands->{$br};
	    my $purge_cands2 = $br2purge_cands2->{$br}; # Added in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). (Out)'
	    my $sgn_wds_in_pcands = $br2sgn_wds_in_pcands->{$br}; # Added in modification (on 2019/06/16). (Outm)
	    my $ct_pcands = @{$purge_cands};

    if ($if_test_output) { # (2) #
			# Modified in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). (Out2) #
	    print STDOUT ' ' x 3, "{ Branch ID: $br ({#purge candidate regions} = $ct_pcands) }\n\n";

	    print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_rg', 'end_rg', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'), "\n\n"; # Modified in revision (on 2019/06/16). (Out2m)
	    ##print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_rg', 'end_rg', '(sites,in,the,region)', 'pv', 'pv0'), "\n\n";
    } # End of "if ($if_test_output) { # (2) # ...}"

	    for (my $pc = 0; $pc < $ct_pcands; $pc++) { # Inner for-loop (over (merged) purge-candidate regions). #
		my ($sites, $pv, $pv0) = @{$purge_cands2->[$pc]};
		my $cnct_sites = join (',', @{$sites});

		my ($beg_pc, $end_pc) = @{$purge_cands->[$pc]}; # Added in revision (on 2019/06/18).

			# Further Modified in revision (on 2019/06/16). (Out3m) #
		my @out_sgn_wds = ();
		foreach my $info_swd (@{$sgn_wds_in_pcands->[$pc]}) { # Innermost foreach-loop (over significant windows in the region).
		     my ($sites_in_wd, $pv_wd, $pv0_wd) = @{$info_swd};
		     my $cnct_sites_in_wd = join (',', @{$sites_in_wd});
		     push @out_sgn_wds, "[$cnct_sites_in_wd]/$pv_wd/$pv0_wd";
		} # End of the innermost foreach-loop (over significant windows in the region).

    if ($if_test_output) { # (3) #
		print STDOUT ' ' x 3, join ("\t", $pc, $beg_pc, $end_pc, "($cnct_sites)", $pv, $pv0, @out_sgn_wds), "\n"; # Modified in revision (on 2019/06/18).
		##print STDOUT ' ' x 3, join ("\t", $pc, @{$purge_cands->[$pc]}, "($cnct_sites)", $pv, $pv0, @out_sgn_wds), "\n";
		####print STDOUT ' ' x 3, join ("\t", $pc, @{$purge_cands->[$pc]}, "($cnct_sites)", $pv, $pv0), "\n";
    } # End of "if ($if_test_output) { # (3) # ...}"
			# End of "Further Modified in revision (on 2019/06/16). (Out3m)" #

		push @list_mgd_purge_candidates, [ $beg_pc, $end_pc, $br, $br2cnct_exoffs_msa_rows{$br}, "($cnct_sites)", $pv, $pv0, @out_sgn_wds]; # Further Modified in revision (on 2019/06/19).
		##push @list_mgd_purge_candidates, [ $beg_pc, $end_pc, $br, "($cnct_sites)", $pv, $pv0, @out_sgn_wds]; # Modified in revision (on 2019/06/18).
		####push @list_mgd_purge_candidates, [ @{$purge_cands->[$pc]}, $br, "($cnct_sites)", $pv, $pv0, @out_sgn_wds];

			# Added in revision (on 2019/06/18). #
		my $size_pc = $end_pc - $beg_pc +1;
		($maxsize_pcand < $size_pc) and ($maxsize_pcand = $size_pc);
			# End of "Added in revision (on 2019/06/18)." #

	    } # End of the inner for-loop (over (merged) purge-candidate regions). #

	    #print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_rg', 'end_rg'), "\n\n";
	    #for (my $pc = 0; $pc < $ct_pcands; $pc++) { print STDOUT ' ' x 3, join ("\t", $pc, @{$purge_cands->[$pc]}), "\n"; }
			# End of "Modified in creation of 'test_detect_purge_cands2.pl' (on 2019/06/16). (Out2)" #

    (print STDOUT "\n\n") if ($if_test_output); # Modified in revision (on 2019/06/17). #
	    ##print STDOUT "\n\n";

	}					# End of the middle foreach-loop (over branches). #

    (print STDOUT "\n") if ($if_test_output); # Modified in revision (on 2019/06/17). #
	##print STDOUT "\n";

   ##} # END of "if (($TEST_LEVEL>0) and ($MyDetect_purge_cands::TEST_LEVEL <= 3)) {...}" # DISABLED (on 2019/06/17)

}			# End of the main outer foreach-loop (over subject regions). #

my $totct_mgd_pcands = @list_mgd_purge_candidates; # Added in revision (on 2019/06/18).

		### ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/19). (TRL) ###

(print STDOUT "\n\n === The MAIN Process ENDED!! ===\n\n\n") if ($TEST_LEVEL>0);

=cut # END of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (MAIN)"


=For the predecessor, 'detect_purge_cands.ver0.5.pl'. (TRL)

print STDOUT "\n<<< TRAILING: Sort the list of merged purge-candidate regions & print them out >>>\n\n\n";

my @indices_srt = (0, 1, 2);
my @methods_srt = ('+', '+', '+');

my $std_list_mgd_purge_candidates = multiple_sort (@list_mgd_purge_candidates, @indices_srt, @methods_srt);
unless (defined $std_list_mgd_purge_candidates) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'multiple_sort' FAILED on the list of merged purge-candidate regions !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'multiple_sort', Ended ...\n\n") if ($TEST_LEVEL>0);

# @list_mgd_purge_candidates = ();
	# element = ['beg_rg', 'end_rg', br_ID, '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...']

my $fh_out = open_output_file ($out_main, 1);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED on the output file, \"$out_main\" !\n";
	print STDERR "[ERROR '$plscript_name'] $! !\n\n";
	die ;
}
(print STDOUT " ... Subroutine, 'open_output_file', Ended ...\n\n") if ($TEST_LEVEL>0);

(print STDOUT " ... Now, Print out the purge-candidates to the file: \"$out_main\" ...\n\n") if ($TEST_LEVEL>0);


print $fh_out "< #{merged-candidate-regions}=$totct_mgd_pcands, Maximum_span=$maxsize_pcand >\n\n"; # Modified (on 2019/06/20). #
##print $fh_out "< #{merged-candidate-regions}=$totct_mgd_pcands, Maximum_size=$maxsize_pcand >\n\n"; # ADDED in revision (on 2019/06/18). #

print $fh_out join ("\t", 'beg_rg', 'end_rg', 'br_ID', '[msa,rows,under,br]', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'), "\n\n"; # Modified in revision (on 2019/06/19).
##print $fh_out join ("\t", 'beg_rg', 'end_rg', 'br_ID', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'), "\n\n";

foreach my $mgd_pcand (@{$std_list_mgd_purge_candidates}) { print $fh_out join ("\t", @{$mgd_pcand}), "\n"; }
print $fh_out "\n";

close ($fh_out);


print STDOUT "\n\n ===== ALL the PROCESSES ENDED !!! =====\n\n\n";

		### End of "ADDED in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/19). (TRL)" ###

=cut # End of "For the predecessor, 'detect_purge_cands.ver0.5.pl'. (TRL)"



